#ifndef AUTHLIST_H
#define AUTHLIST_H

#include "inclstdint.h"
#include "http.h"
#include "sha256.h"
#include "nfcuid.h"
#include "lock.h"
#include "prepaid.h"

#include <FreeRTOS.h>
#include <semphr.h>

enum EOperatingMode
{
  OperatingMode_NoList    = 0,
  OperatingMode_BlackList = 1,
  OperatingMode_WhiteList = 2,
};

#define AUTH_LIST_VALID   0x12345678
#define AUTH_LIST_INVALID 0x0
#define AUTH_LIST_ERASED  0xFFFFFFFF

struct SAuthListHeader
{
  /*! Magic word that indicates the state of the auth list,
      should always be the first member.*/
  uint32_t state;

  //! Number of bytes in the auth list.
  uint32_t size;

  //! Hash over the auth list.
  uint8_t  hash[SHA256_BLOCK_SIZE]; 

  //! Flexible array member which points to the UID list.
  uint64_t uidlist[];
}__attribute__((packed));

#define COPYBUF_SIZE 1024

/*! Number of right shifts needed to extract the type byte.*/
#define TYPE_BYTE_POSITION 56

/*! AND mask to apply to the type byte to get its card type.*/
#define CARD_TYPE_MASK     0x7F

/*! In the whitelist, the most significant bit of the type byte
    indicates if the UID is 4 byte or 7 byte.*/
#define FLAG_7_BYTE_UID    0x80

/*! Bitwise AND mask to extract a 7 byte UID.*/
#define UID_7_BYTE_MASK    0x00FFFFFFFFFFFFFFULL

/*! Bitwise AND mask to extract a 4 byte UID.*/
#define UID_4_BYTE_MASK    0x00000000FFFFFFFFULL

/*! Card type, stored in the 7 least significant bits of the type byte.
    \enum ECardType */
enum ECardType
{
  CardType_Unregistered          = 0,   //!< Card is not on the authorisations list.

  CardType_Prepaid               = 50,  /*!< Card which has credit and usage
                                             counters in the whitelist.*/

  CardType_Secured_Program       = 51,  //!< Secret key was programmed into card.

  /*! Start of secured card type range. This implements the same funtionality
      as the card types starting at CardType_Unsecured_Start, except 3DES
      authentication is performed first. To convert a secured card type first
      subtract CardType_Secured_Start, then add CardType_Unsecured_Start.*/
  CardType_Secured_Start         = 71,

  //TODO: implement functionality for all these types
  CardType_Unsecured_Start       = 101, //!< Start of unsecured card type range.
  CardType_Outbound_Call         = 101, //!< Set up the modem and connect to the server.
  CardType_Reset                 = 102, //!< Perform a warm boot, reset the microcontroller.
  CardType_Service               = 103, //!< TODO: figure out what to do with this.
  CardType_Emptying              = 104, //!< Reset fill level counter.
  CardType_Inbound_Call          = 105, //!< Not implemented.
  CardType_Registration          = 106, //!< Waste deposit, open the lock.
  CardType_Lock_Test             = 107, //!< Open the lock, don't register a deposit.
  CardType_Show_Current_Load     = 108, //!< Show the current fill level.
  CardType_Reset_Battery_counter = 109, //!< Not implemented, functionality unclear.
  CardType_Show_GPRS_Signal      = 110, //!< Show the modem RSSI on the display.
  CardType_Unsecured_End,               //!< Used to count number of unsecured card types.
};

class CAuthList : public CHTTPDownload
{
  public:
    CAuthList();

    void               Setup();
    bool               Prepare();

    virtual uint32_t   DownloadBlockSize();
    virtual void       StartDownload(const char* url, uint32_t httprspcode, uint32_t contentlength);
    virtual void       Download(const uint8_t* buf, uint32_t size);

    bool               Install();

    bool               Authenticate(NFCUID* uid, bool* keyfailed);

    static const char* CardTypeToStr(ECardType cardtype);
    static bool        IsSecuredCard(ECardType cardtype);
    static ECardType   FromSecuredType(ECardType cardtype);

  private:
    bool               EraseSPIFlash();
    bool               WriteHeader();
    bool               CopyToInternalFlash(uint32_t size);
    bool               CopyFlashData(uint32_t readaddr, uint32_t writeaddr, uint32_t size);
    bool               CheckHeaderOfInternalFlash(bool checkhash = true);
    bool               CheckHeaderOfSPIFlash(bool* erased, uint32_t* size);
    void               Recover();

    bool               AuthenticateSelfTest(NFCUID* uid);
    bool               AuthenticateResetReason(NFCUID* uid);
    bool               AuthenticateActivate(NFCUID* uid);
    bool               AuthenticateNoListPresent(NFCUID* uid, CLock& lock);
    bool               AuthenticateAuthList(NFCUID* uid, CLock& lock, bool* keyfailed);
    bool               AuthenticateFallback(NFCUID* uid);
    bool               ProgramSecuredCard(NFCUID* uid);
    bool               AuthenticateSecuredCard(NFCUID* uid, ECardType cardtype);
    void               Execute(NFCUID* uid, ECardType cardtype, uint64_t trailingdata = 0);
    const char*        OperatingModeToStr(EOperatingMode operatingmode);
    bool               ConvertUID(NFCUID* uid, uint64_t* output);
    bool               IsActivateUID(NFCUID* uid);
    bool               IsAlwaysActivateUID(NFCUID* uid);

    SAuthListHeader*   m_listptr;
    uint32_t           m_maxlistsize;
    uint32_t           m_spiflashpos;
    uint32_t           m_downloadsize;
    uint32_t           m_bytesdownloaded;
    uint8_t            m_copybuf[COPYBUF_SIZE];
    SHA256_CTX         m_shactx;

    SemaphoreHandle_t  m_mutex;

    bool               m_signaledcomm;

    CPrepaid           m_prepaid;
};

#endif //AUTHLIST_H
