#ifndef IMAGES_H
#define IMAGES_H

#include "inclstdint.h"
#include "http.h"
#include "cryptokeys.h"
#include "sha256.h"
#include "imagesbuiltin.h"
#include "cbortools.h"

#include <FreeRTOS.h>
#include <semphr.h>

/*! Same size as the max string length from CKeyValueStorage which stores the url.*/
#define URL_SIZE 256

/*! The version field in SImagesFile is set to this value. If it's not this value,
    the stored data in the internal flash is not a valid images file.
    This way, it can be detected if the internal flash sector is erased,
    and the layout of SImagesFile can be changed without problems.*/
#define IMAGES_FILE_VERSION      0x1

/*! Initial value of the version field, so that if the images file contains data
    that causes a crash, the sector is erased at the next boot.*/
#define IMAGES_FILE_VERSION_INIT 0x7FFFFFFF

/*! Value of the version field in SImagesFile if the sector is erased.*/
#define IMAGES_FILE_ERASED       0xFFFFFFFF

/*! Stored at the start of the internal flash sector, to indictate from where the
    images file has been downloaded, and how big it is.*/
struct SImagesFile
{
  uint32_t version;
  uint32_t size;
  char     url[URL_SIZE];
}__attribute__((packed));

/*! Represents the start of the images file.*/
struct SImagesFileHeader
{
  uint8_t  hmac[SHA256_BLOCK_SIZE]; //! SHA256 HMAC over headersize, header and the data.
  uint32_t headersize;              //! Size of the CBOR header.
  uint8_t  header[1];               //! CBOR header, at least one byte long.
}__attribute__((packed));

enum EImage
{
  Image_None = -1,
  Image_Locked = 0,
  Image_Unlocked,
  Image_Denied,
  Image_Success,
  Image_SuccessPrepaid,
  Image_LowCredit,
  Image_Error,
  Image_Full,
  Image_Empty,
  Image_ServerComm,
  Image_InsufficientCredit,
  Image_CardError,
};

class CImages : public CHTTPDownload
{
  public:
    CImages();

    void               Setup();
    bool               Prepare();

    virtual uint32_t   DownloadBlockSize();
    virtual void       StartDownload(const char* url, uint32_t httprspcode, uint32_t contentlength);
    virtual void       Download(const uint8_t* buf, uint32_t size);

    bool               EraseInstalled();
    bool               URLChanged(const char* settingsurl);
    bool               Install(const char* url);

    void               Get(EImage type, lv_img_dsc_t* img);

    //cast operator to a SemaphoreHandle_t reference so that CLock(*this); can be used
    operator           SemaphoreHandle_t&() { return m_mutex; }

  private:
    bool               EraseSPIFlash();
    bool               CheckDownloadedFile();
    bool               InstallDownloadedFile(const char* url);
    bool               CheckInstalledFile(bool init = false);
    void               FinalizeInstall();
    bool               OpenHeader(CborParser* parser, CborValue* root, CborValue* images);
    bool               VerifyImage(lv_img_dsc_t* img, uint32_t position);

    bool               GetImageExternal(EImage type, lv_img_dsc_t* img);
    void               GetImageInternal(EImage type, lv_img_dsc_t* img);

    uint32_t           m_spiflashpos;
    uint32_t           m_downloadsize;
    uint32_t           m_bytesdownloaded;

    static uint8_t     m_signkey[CRYPTO_KEY_SIZE];

    uint8_t*           m_intflashaddr;
    SImagesFile*       m_imagesfile;
    SImagesFileHeader* m_fileheader;
    bool               m_filecorrect;

    SemaphoreHandle_t  m_mutex;

    EImage             m_prevtype;
};

#endif //IMAGES_H
