#ifndef INPUTS_H
#define INPUTS_H

#include <FreeRTOS.h>
#include <semphr.h>
#include <event_groups.h>

#include "ioexpander.h"
#include "i2c.h"
#include "exti.h"

#define NUM_INPUTS       8         //!< Number of inputs on the IO expander.

#define INPUT_TOUCH1     IOEXP_GP0 //!< Output of middle touch sensor.
#define INPUT_TOUCH2     IOEXP_GP1 //!< Output of right touch sensor.
#define INPUT_TOUCH3     IOEXP_GP2 //!< Output of left touch sensor.
#define INPUT_IN1        IOEXP_GP3 //!< Power-rail compatible input 1.
#define INPUT_IN2        IOEXP_GP4 //!< Power-rail compatible input 2.
#define INPUT_IN3        IOEXP_GP5 //!< Power-rail compatible input 3.
#define INPUT_IN4        IOEXP_GP6 //!< Power-rail compatible input 4.
#define INPUT_ST25R3911B IOEXP_GP7 //!< IRQ output of the ST25R3911B, also connected to PE4.

#define INPUT_ALL       (INPUT_TOUCH1 | INPUT_TOUCH2 | INPUT_TOUCH3 | \
                         INPUT_IN1 | INPUT_IN2 | INPUT_IN3 | INPUT_IN4 | \
                         INPUT_ST25R3911B)

#define INPUT_ALL_IN    (INPUT_IN1 | INPUT_IN2 | INPUT_IN3 | INPUT_IN4)

#define INPUT_IN_ARRAY  ((const uint8_t[]){ INPUT_IN1, INPUT_IN2, INPUT_IN3, INPUT_IN4 })
#define INPUT_NUM_IN    (sizeof(INPUT_IN_ARRAY) / sizeof(INPUT_IN_ARRAY[0]))

/*! Inputs for which the IO expander's interrupt is enabled.
    It's not enabled for the ST25R3911B IRQ line, because that is read using PE4.*/
#define INPUT_INTERRUPT (INPUT_TOUCH1 | INPUT_TOUCH2 | INPUT_TOUCH3 | \
                         INPUT_IN1 | INPUT_IN2 | INPUT_IN3 | INPUT_IN4)

//! Input where the fill level reset switch is connected to.
#define INPUT_FILL_LEVEL_RESET INPUT_IN3

//! Input where a switch for a container door is connected to.
#define INPUT_DOOR_SWITCH INPUT_IN4

//! Input where the deposit1 switch is connected to.
#define INPUT_DEPOSIT1 INPUT_IN1

//! Input where the deposit2 switch is connected to.
#define INPUT_DEPOSIT2 INPUT_IN2

//! Number of deposit inputs.
#define NUM_DEPOSIT_INPUTS 2

//! Conversion of deposit input number to define, starts at 0.
#define DEPOSIT_INPUT(x) (((const uint8_t[]){ INPUT_DEPOSIT1, INPUT_DEPOSIT2 })[x])

//! Wakeup interrupt of the 6D sensor.
#define INPUT_6DSENSOR INPUT_TOUCH2

class CInputs : public CGPIOInterrupt
{
  public:
    CInputs();
    ~CInputs();

    bool               Setup(CI2C* i2c);
    bool               SetWakeupInterrupt();
    bool               ReadSingle(uint8_t input);
    uint8_t            ReadMultiple(uint8_t inputs);
    uint8_t            WaitForChange(uint8_t inputs, uint8_t startvalues, uint32_t timeout);
    bool               ReadStable(uint8_t inputs, uint32_t stabletime, uint32_t timeout, uint8_t* values);
    bool               WaitForStableLevel(uint8_t inputs, uint32_t stabletime, uint32_t timeout, uint8_t wantedlevels, uint8_t* readlevels = nullptr);
    bool               WaitForStableSignal(uint8_t inputs, uint32_t stabletime, uint32_t timeout, uint8_t wantedlevels, uint8_t* readlevels = nullptr);

    static const char* InputStr(uint8_t input);
    static const char* StateStr(bool state);

    //Cast operator to a SemaphoreHandle_t reference so that CLock(*this); can be used.
    //A task can use this, to prevent other tasks from reading inputs.
    operator           SemaphoreHandle_t&() { return m_mutex; }

  private:
    virtual void       GPIOIRQHandler(SGPIOPin pin);

    CIOExpander        m_ioexpander;
    uint8_t            m_bootstate;
    volatile bool      m_pinchanged;
    uint8_t            m_inputstate;

    SemaphoreHandle_t  m_mutex;
    EventGroupHandle_t m_signal;
    EventBits_t        m_listeners;
};

#endif //INPUTS_H
