#ifndef LOCKDRIVER_H
#define LOCKDRIVER_H

#include "inclstdint.h"

enum ELockType
{
  LockType_None            = -1,
  LockType_Solenoid        = 16,
  LockType_Start           = LockType_Solenoid,
  LockType_Solenoid_Toggle = 17,
  LockType_Motor           = 18,
  LockType_Bistable_2      = 19,
  LockType_ServoMotor      = 20,
  LockType_Bistable_1      = 21,
  LockType_Press_Engels    = 22,
  LockType_Solenoid_Double = 23,
  LockType_30_60_Liter     = 24,
  LockType_HoldLock        = 25, //Servo lock with a permanent electromagnet.
  LockType_End
};

#define LOCK_OPENED_FLAG 0x20000000

/*! Default time in milliseconds until the permanent electromagnet
    on a hold lock is powered to release the deposit lid.*/
#define DEFAULT_HOLD_LOCK_DELAY 15000

/*! Default time in milliseconds that the permanent electromagnet
    on a hold lock is powered to release the deposit lid.*/
#define DEFAULT_HOLD_LOCK_RELEASE 1000

class CLockDriver
{
  public:
    CLockDriver();

    void        PowerUp();
    void        PowerDown();

    void        PowerUpBoostConverter();
    void        PowerDownBoostConverter();

    float       MeasurePowerRail();

    void        Open();
    void        Close();
    void        ReleaseHold();

    bool        Opened();
    void        ClearOpenedState();
    bool        IsBiStable();
    bool        HasRelease();

    ELockType   LockType();
    const char* LockTypeStr() { return LockTypeStr(LockType()); }

  private:
    void        SetType();
    const char* LockTypeStr(ELockType locktype);

    void        ReadServoSettings();

    void        PowerUpServoLock();
    void        OpenServoLock();
    void        CloseServoLock();
    void        PowerDownServoLock();

    void        PowerUpSolenoidLock();
    void        OpenSolenoidLock();
    void        CloseSolenoidLock();
    void        PowerDownSolenoidLock();

    void        PowerUpBistableLock();
    void        OpenBistableLock();
    void        CloseBistableLock();
    void        PowerDownBistableLock();

    void        PowerUp3060LiterLock();
    void        Open3060LiterLock();
    void        Close3060LiterLock();
    void        PowerDown3060LiterLock();

    void        PowerUpHoldLock();
    void        OpenHoldLock();
    void        CloseHoldLock();
    void        ReleaseHoldLock();
    void        PowerDownHoldLock();

    void        EnableTimer(uint32_t interval);
    void        DisableTimer();

    ELockType   m_locktype;

    uint32_t    m_servo1timeopen;
    uint32_t    m_servo1timeclose;

    uint32_t    m_servo2timeopen;
    uint32_t    m_servo2timeclose;

    int64_t     m_lockclosemoment;

    bool        m_powered;
    uint32_t    m_interval;

    float       m_powerrail;
};

#endif //LOCKDRIVER_H
