#ifndef LOWPOWER_H
#define LOWPOWER_H

#include "linkedlist.h"

#include <FreeRTOS.h>
#include <semphr.h>

#if configUSE_TRACE_FACILITY && configUSE_STATS_FORMATTING_FUNCTIONS
  #define PRINT_TASK_STATS
#endif

/*! Flag in FLAGS_REGISTER that indicates low power mode
    should be entered immediately after booting.*/
#define ENTER_LOW_POWER_MODE_FLAG 0x40000000

/*! Classes that want to enable or disable low power mode should derive from
    this and call RegisterLowPowerTask("taskname");
    The default is to disable entering low power mode, if a class wants to
    allow entering low power mode, it should call SetAllowLowPower(true),
    to deny going into low power mode, it should call SetAllowLowPower(false).
    Low power mode is entered when all registered classes allow low power mode.
    */
class CLowPowerTask : public CListElement
{
  public:
    CLowPowerTask();

    void         RegisterLowPowerTask(const char* name);
    void         AllowLowPower(bool allow);

  private:
    friend class CLowPower;

    bool         GetAllowLowPower()           { return m_allowlowpower;  }
    void         SetAllowLowPower(bool allow) { m_allowlowpower = allow; }

    void         SetName(const char* name)    { m_name = name;           }
    const char*  GetName()                    { return m_name;           }

    bool         m_allowlowpower;
    const char*  m_name;
};

class CLowPower
{
  public:
    CLowPower();

    void              Setup();

    void              Register(CLowPowerTask* lowpowertask, const char* name);
    void              AllowLowPower(CLowPowerTask* lowpowertask, bool allow);

  private:
    void              EnterLowPowerMode();
    void              PrintLowPowerTasks();

    CLinkedList<CLowPowerTask>
                      m_tasklist;
    SemaphoreHandle_t m_mutex;

#ifdef PRINT_TASK_STATS
    char              m_statsbuf[512]; //Needs about 40 bytes per task.
#endif
};

#endif //LOWPOWER_H
