#ifndef PLATFORM_H
#define PLATFORM_H

/* This file is the interface between rfal and the system code.
   It tells rfal what functions to call using macros.*/

#include <stdbool.h>
#include "inclstdint.h"
#include "limits.h"
#include "gpiodescriptions.h"
#include "reader.h"
#include "clock.h"

#ifdef __cplusplus
  extern "C"
  {
#endif

#include "timer.h"

#ifdef __cplusplus
  }
#endif

#define ST25R391X_INT_PORT                     ST25R3911B_INT_PORT //!< ST25R3911B interrupt pin.
#define ST25R391X_INT_PIN                      ST25R3911B_INT_PIN  //!< ST25R3911B interrupt pin.

#define ST25R_INT_PORT                         ST25R3911B_INT_PORT //!< ST25R3911B interrupt pin.
#define ST25R_INT_PIN                          ST25R3911B_INT_PIN  //!< ST25R3911B interrupt pin.

#define PLATFORM_LED_FIELD_PORT                LED_BLUE_PORT //!< LED that signals if the field is enabled.
//uncomment this define to enable a led when the field is on
//#define PLATFORM_LED_FIELD_PIN                 LED_BLUE_PIN  //!< LED that signals if the field is enabled.

#define platformProtectST25R391xComm()         ReaderLock()   //!< Locks access to the spi peripheral in reader.cpp.
#define platformUnprotectST25R391xComm()       ReaderUnlock() //!< Unlocks access to the spi peripheral in reader.cpp.

#define platformProtectST25RComm()             ReaderLock()   //!< Locks access to the spi peripheral in reader.cpp.
#define platformUnprotectST25RComm()           ReaderUnlock() //!< Unlocks access to the spi peripheral in reader.cpp.

#define platformIrqST25R3911SetCallback( cb )
#define platformIrqST25RSetCallback( cb )
#define platformIrqST25R3911PinInitialize()
#define platformIrqST25RPinInitialize()

#define platformProtectST25R391xIrqStatus()    platformProtectST25R391xComm()  //!< Locks access to the irq status variable.
#define platformUnprotectST25R391xIrqStatus()  platformUnprotectST25R391xComm()//!< Unlocks access to the irq status variable.

#define platformProtectST25RIrqStatus()        platformProtectST25R391xComm()  //!< Locks access to the irq status variable.
#define platformUnprotectST25RIrqStatus()      platformUnprotectST25R391xComm()//!< Unlocks access to the irq status variable.

#define platformProtectWorker()                                                /* Protect RFAL Worker/Task/Process from concurrent execution on multi thread platforms   */
#define platformUnprotectWorker()                                              /* Unprotect RFAL Worker/Task/Process from concurrent execution on multi thread platforms */

#define platformLedsInitialize()
#define platformLedOff( port, pin )            platformGpioClear(port, pin) //!< Turns the selected led off.
#define platformLedOn( port, pin )             platformGpioSet(port, pin)   //!< Turns the selected led on.
#define platformLedToogle( port, pin )         platformGpioToogle(port, pin)//!< Toggles the state of the selected led.

#define platformGpioSet( port, pin )           (GPIO_SetBits(port, pin))          //!< Makes the selected gpio high.
#define platformGpioClear( port, pin )         (GPIO_ResetBits(port, pin))        //!< Makes the selected gpio low.
#define platformGpioToogle( port, pin )        (GPIO_ToggleBits(port, pin))       //!< Toggles the state of the gpio.
#define platformGpioIsHigh( port, pin )        (GPIO_ReadInputDataBit(port, pin)) //!< Reads the state of the gpio.
#define platformGpioIsLow( port, pin )         (!GPIO_ReadInputDataBit(port, pin))//!< Reads the state of the gpio.

//time functions in milliseconds
#define platformTimerCreate( t )               timerCalculateTimer(t)
#define platformTimerIsExpired( timer )        timerIsExpired(timer)
#define platformDelay( t )                     DelayMS(t)
#define platformGetSysTick()                   ((uint32_t)Millis())
#define platformTimerDestroy( t )

//Spi functions
#define platformSpiSelect()                    do { DelayUS(1); GPIO_Reset(ST25R3911B_CS); DelayUS(1); } while(0)
#define platformSpiDeselect()                  do { DelayUS(1); GPIO_Set(ST25R3911B_CS);   DelayUS(1); } while(0)
#define platformSpiTxRx( txBuf, rxBuf, len )   ReaderTransceive(txBuf, rxBuf, len)

#define platformErrorHandle()

#define RFAL_FEATURE_NFCA                      true       /*!< Enable/Disable RFAL support for NFC-A (ISO14443A)                         */
#define RFAL_FEATURE_NFCB                      true       /*!< Enable/Disable RFAL support for NFC-B (ISO14443B)                         */
#define RFAL_FEATURE_NFCF                      true       /*!< Enable/Disable RFAL support for NFC-F (FeliCa)                            */
#define RFAL_FEATURE_NFCV                      true       /*!< Enable/Disable RFAL support for NFC-V (ISO15693)                          */
#define RFAL_FEATURE_T1T                       true       /*!< Enable/Disable RFAL support for T1T (Topaz)                               */
#define RFAL_FEATURE_ST25TB                    true       /*!< Enable/Disable RFAL support for ST25TB                                    */
#define RFAL_FEATURE_DYNAMIC_ANALOG_CONFIG     false      /*!< Enable/Disable Analog Configs to be dynamically updated (RAM)             */
#define RFAL_FEATURE_DPO                       false      /*!< Enable/Disable RFAL dynamic power support                                 */
#define RFAL_FEATURE_ISO_DEP                   true       /*!< Enable/Disable RFAL support for ISO-DEP (ISO14443-4)                      */
#define RFAL_FEATURE_ISO_DEP_POLL              true       /*!< Enable/Disable RFAL support for Poller mode (PCD) ISO-DEP (ISO14443-4)    */
#define RFAL_FEATURE_ISO_DEP_LISTEN            false      /*!< Enable/Disable RFAL support for Listen mode (PICC) ISO-DEP (ISO14443-4)   */
#define RFAL_FEATURE_NFC_DEP                   true       /*!< Enable/Disable RFAL support for NFC-DEP (NFCIP1/P2P)                      */
#define RFAL_FEATURE_T2T                       true       /*!< Enable/Disable RFAL support for T2T (type 2 tags)                         */ 
#define RFAL_FEATURE_T4T                       true       /*!< Enable/Disable RFAL support for T4T (type 4 tags)                         */ 
#define RFAL_FEATURE_WAKEUP_MODE               true
#define RFAL_FEATURE_LISTEN_MODE               true
#define RFAL_FEATURE_NFC_RF_BUF_LEN            258U       /*!< RF buffer length used by RFAL NFC layer                                   */
#define RFAL_FEATURE_NFC_DEP_BLOCK_MAX_LEN     254
#define RFAL_FEATURE_NFC_DEP_PDU_MAX_LEN       RFAL_FEATURE_NFC_DEP_BLOCK_MAX_LEN

#define RFAL_FEATURE_ISO_DEP_IBLOCK_MAX_LEN    256        /*!< ISO-DEP I-Block max length. Please use values as defined by rfalIsoDepFSx */
#define RFAL_FEATURE_ISO_DEP_APDU_MAX_LEN      1024       /*!< ISO-DEP APDU max length. Please use multiples of I-Block max length       */

#endif /* PLATFORM_H */
