#ifndef PREPAID_H
#define PREPAID_H

#include "inclstdint.h"
#include "aes.h"
#include "cryptokeys.h"
#include "nfcuid.h"

#include <cstring>

#define PREPAID_COUNTERS_VERSION 1

/*! Prepaid information that gets stored on the card.*/
class CPrepaidCounters
{
  public:
    CPrepaidCounters()
    {
      static_assert(sizeof(*this) == AES_BLOCK_SIZE,
                    "This class should be the size of one AES block.");
      Reset();
    }

    void Reset()
    {
      m_version     = PREPAID_COUNTERS_VERSION;
      m_padding[0]  = 0;
      m_padding[1]  = 0;
      m_padding[2]  = 0;
      m_transaction = 0;
      m_credit      = 0;
      m_used        = 0;
    }

    operator uint8_t*()                           { return (uint8_t*)this;       }

    uint8_t  Version()                            { return m_version;            }
    uint32_t Transaction()                        { return m_transaction;        }
    int32_t  Credit()                             { return m_credit;             }
    int32_t  Used()                               { return m_used;               }

    void     SetVersion(uint8_t version)          { m_version = version;         }
    void     SetTransaction(uint32_t transaction) { m_transaction = transaction; }
    void     SetCredit(int32_t credit)            { m_credit = credit;           }
    void     SetUsed(int32_t used)                { m_used = used;               }

  private:
    uint8_t  m_version;     //!< Version number of the information.
    uint8_t  m_padding[3];
    uint32_t m_transaction; //!< Transaction number of the prepaid information.
    int32_t  m_credit;      //!< Credit counter from the auth list.
    int32_t  m_used;        //!< Used counter, increased when doing a deposit.
}__attribute__((packed));

/*! Block of encrypted and signed data stored on the card.*/
class CPrepaidBlock
{
  public:
    CPrepaidBlock()
    {
      static_assert(sizeof(*this) == AES_BLOCK_SIZE * 3,
                    "This class should be the size of 3 AES blocks");
      Reset();
    }

    void Reset()
    {
      memset(m_iv,      0, sizeof(m_iv));
      memset(m_hmac,    0, sizeof(m_hmac));
      memset(m_payload, 0, sizeof(m_payload));
    }

    operator uint8_t*()    { return (uint8_t*)this;    } 

    uint8_t* IV()          { return m_iv;              }
    uint32_t IVSize()      { return sizeof(m_iv);      }

    uint8_t* HMAC()        { return m_hmac;            }
    uint32_t HMACSize()    { return sizeof(m_hmac);    }

    uint8_t* Payload()     { return m_payload;         }
    uint32_t PayloadSize() { return sizeof(m_payload); }

  private:
    uint8_t  m_iv[AES_BLOCK_SIZE];     //!< CBC initialization vector.
    uint8_t  m_hmac[AES_BLOCK_SIZE];   //!< Encrypted 16 byte HMAC.
    uint8_t  m_payload[AES_BLOCK_SIZE];//!< Encrypted CPrepaidCounters.

}__attribute__((packed));

enum EPrepaidResult
{
  EPrepaidResult_None = -1,
  EPrepaidResult_Success,
  EPrepaidResult_InsufficientCredit,
  EPrepaidResult_CardFailure,
  EPrepaidResult_ReaderError,
  EPrepaidResult_CardNotAuthenticated,
};

/*! Processes information about prepaid credit and prepaid used counters.*/
class CPrepaid
{
  public:
    CPrepaid();
    void                 Reset();

    void                 FromAuthList(NFCUID* uid, uint64_t trailingdata);
    void                 FromCard(NFCUID* uid, uint32_t bank, CPrepaidBlock* block);

    void                 SetCost(int32_t cost);
    int32_t              GetCost() { return m_cost; }

    bool                 Charge(NFCUID* uid, CPrepaidBlock* block);
    uint32_t             WriteBankIndex();

    int32_t              InitialValue()      { return m_initialvalue;                }

    bool                 ResultValid()       { return m_resultvalid;                 }
    uint32_t             ResultTransaction() { return m_resulttransaction;           }
    int32_t              ResultCredit()      { return m_resultcredit;                }
    int32_t              ResultUsed()        { return m_resultused;                  }
    int32_t              ResultCost()        { return m_resultcost;                  }
    int32_t              ResultValue()       { return m_resultcredit - m_resultused; }

    static void          Encrypt(NFCUID* uid, CPrepaidCounters* cardcounters,
                                 CPrepaidBlock* block);

  private:
    static void          GenerateSignKey(NFCUID* uid, uint8_t key[CRYPTO_KEY_SIZE]);
    static void          GenerateAESKey(NFCUID* uid, uint8_t key[CRYPTO_KEY_SIZE]);

    int32_t              m_authlistcredit;           //!< Credit counter according to the authlist.
    int32_t              m_authlistused;             //!< Used counter according to the authlist.

    CPrepaidCounters     m_cardcounters;             //!< Prepaid information from the card.
    int32_t              m_bank;                     //!< From which bank m_cardcounters was read.

    int32_t              m_cost;                     //!< Cost of the current deposit.

    int32_t              m_initialvalue;             //!< Initial value before the deposit has been charged.

    //These values are written in the registration for the deposit.
    bool                 m_resultvalid;              //!< Tried to charge, all the m_result fields are valid.
    uint32_t             m_resulttransaction;        //!< Number of the current card transaction.
    int32_t              m_resultcredit;             //!< Current credit counter.
    int32_t              m_resultused;               //!< Current used counter.
    int32_t              m_resultcost;               //!< Amount charged on card.

    static const uint8_t m_signkey[CRYPTO_KEY_SIZE]; //!< HMAC key used to sign data in the card.
    static const uint8_t m_aeskey[CRYPTO_KEY_SIZE];  //<! AES key used to encrypt data in the card.
};

#endif //PREPAID_H
