#ifndef READER_H
#define READER_H

#ifdef __cplusplus

#include "taskhandler.h"
#include "exti.h"
#include "spi.h"
#include "nfcuid.h"
#include "lowpower.h"
#include "prepaid.h"
#include "tomcrypt.h"

extern "C"
{
  #include "rfal_nfca.h"
}

#include <FreeRTOS.h>
#include <semphr.h>

//Set to 1 to program the WSS key into MIFARE Ultralight C cards that are not
//on the auth list.
#define ULTRALIGHT_C_PROGRAMMER_MODE 0

/* Default low power wakeup settings.*/
#define DEFAULT_LPW_PERIOD          800

/*! Amplitude low power wakeup is disabled because it causes a lot of unnecessary wakeups,*/
#define DEFAULT_AMP_ENABLED         false
#define DEFAULT_AMP_DELTA           6
#define DEFAULT_AMP_REFERENCE       RFAL_WUM_REFERENCE_AUTO
#define DEFAULT_AMP_AUTOAVG         true
#define DEFAULT_AMP_AUTOAVGWEIGHT   32
#define DEFAULT_AMP_INCLMEAS        false

#define DEFAULT_PHASE_ENABLED       true
#define DEFAULT_PHASE_DELTA         4
#define DEFAULT_PHASE_REFERENCE     RFAL_WUM_REFERENCE_AUTO
#define DEFAULT_PHASE_AUTOAVG       true
#define DEFAULT_PHASE_AUTOAVGWEIGHT 32
#define DEFAULT_PHASE_INCLMEAS      false

class CReader;

/*! Class that calls st25r3911Isr() from a FreeRTOS task,
    after a rising edge on the ST25R3911B interrupt pin is detected.
    Because the st25r3911Isr() function does communication over SPI,
    it can't be called directly from the ISR, because the SPI code
    does not work from an ISR.*/
class CST25R3911Isr : public CTaskHandler, public CGPIOInterrupt
{
  public:
    CST25R3911Isr(CReader& parent);

    void         Setup();

  private:
    virtual void Process();
    virtual void GPIOIRQHandler(SGPIOPin pin);

    CReader&     m_parent;
};

#define MAXDEVICES 16

class CReader : public CTaskHandler, public CLowPowerTask
{
  public:
    CReader();

    void                 Setup();
    void                 Start();
    virtual void         Process();

    void                 Transceive(const uint8_t* out, uint8_t* in, uint32_t numbytes);
    void                 Lock();
    void                 Unlock();

    void                 PreventLowPower();
    bool                 IsWorking() { return m_working; }

    EPrepaidResult       ChargePrepaidDeposit(NFCUID* uid, CPrepaid* prepaid);

    bool                 ProgramSecuredCard(NFCUID* uid);
    bool                 AuthenticateSecuredCard(NFCUID* uid);

  private:
    void                 InitializeReader();
    bool                 CheckAntennaCalibration();
    void                 EnableLowPowerWakeup();
    rfalWumPeriod        LowPowerWakeupPeriod(uint32_t detecttime);
    rfalWumAAWeight      AutoAvgWeight(uint32_t weight);

    void                 WaitLowPowerWakeup();

    bool                 ProcessRead(uint32_t timeoutms, rfalNfcaListenDevice* authenticatedcard,
                                     rfalNfcaListenDevice* presentcard, uint32_t* cardspresent = nullptr);
    bool                 InitCards();
    uint32_t             ReadCards(rfalNfcaListenDevice* nfcadevlist, uint32_t maxdevices);
    void                 PrintCardInfo(rfalNfcaListenDevice* nfcadevlist, uint32_t numdevices);
    void                 PrintCardUID(int loglevel, rfalNfcaListenDevice* dev);
    bool                 AuthenticateCard(rfalNfcaListenDevice* dev, uint32_t numdevices,
                                          rfalNfcaListenDevice* authenticatedcard, rfalNfcaListenDevice* presentcard);
    bool                 WaitForCardRemoval(uint32_t timeoutms, rfalNfcaListenDevice* authenticatedcard);

    void                 ClearDenyList();
    void                 AddToDenyList(NFCUID* uid, bool keyfailed);
    void                 RegisterDeniedCards();

    void                 IncreasePontCounter();
    void                 DecreasePontCounter();

    void                 GenerateUltralightCKey(NFCUID* uid, uint8_t* key);

    bool                 AuthenticateUltralightC(NFCUID* uid);
    bool                 Authenticate3DES(NFCUID* uid);
    void                 PrintUltralightCError(NFCUID* uid, ReturnCode retval,
                                               uint8_t*cmd, uint32_t cmdlen,
                                               uint8_t* reply, uint32_t replylen);

    bool                 SelectCard(NFCUID* uid);
    void                 TurnFieldOff(uint32_t delayms);
    bool                 ReadCardPrepaidData(NFCUID* uid, CPrepaid* prepaid,
                                             CPrepaidBlock* block);
    bool                 WriteCardPrepaidData(NFCUID* uid, CPrepaid* prepaid,
                                              CPrepaidBlock* block);
    bool                 ReadBank(NFCUID* uid, uint32_t bank, CPrepaidBlock* block);
    bool                 WriteBank(NFCUID* uid, uint32_t bank, CPrepaidBlock* block);

    bool                 ProgramUltralightC(NFCUID* uid);
    bool                 ProgramUltralightCMemory(NFCUID* uid);

    CSpi                 m_spi;
    SemaphoreHandle_t    m_spilock;

    CST25R3911Isr        m_irqhandler;

    bool                 m_woke;
    bool                 m_wokelowpower;
    bool                 m_working;
    bool                 m_signaledcomm;

    /*! Buffer for ProcessRead(), can't be on the stack because
        the FreeRTOS stack size per task is not very big.*/
    rfalNfcaListenDevice m_nfcadevlist[MAXDEVICES];

    NFCUID               m_denylist[MAXDEVICES];
    bool                 m_denykeyfailed[MAXDEVICES];
    uint32_t             m_denylistsize;

    des3_key             m_des3key; //des3_key instead of symmetric_key to save ram.
};

#endif //__cplusplus


#ifdef __cplusplus
extern "C" {
#endif

/*! Function for the platformSpiTxRx define in platform.h,
    it allows the reader library to communicate with the ST25R3911B using SPI.*/
void ReaderTransceive(const uint8_t* out, uint8_t* in, uint32_t len);
void ReaderLock();
void ReaderUnlock();

#ifdef __cplusplus
}
#endif

#endif //READER_H
