#ifndef SELFTEST_H
#define SELFTEST_H

#include <FreeRTOS.h>
#include <semphr.h>
#include "usart.h"

enum ESelfTest
{
  SelfTest_START = 0,
  SelfTest_InternalRTC = 0,       //!< Internal RTC is running.
  SelfTest_ExternalRTC,           //!< External RTC is running.
  SelfTest_SPIFlash,              //!< SPI flash is writeable and readable.
  SelfTest_Modem,                 //!< Modem connection has been set up and no fw update needed.
  SelfTest_ServerComm,            //!< The server has been contacted successfully.
  SelfTest_ReaderCoil,            //!< The NFC reader has calibrated the reader coil correctly.
  SelfTest_NFCReader,             //!< The NFC reader has read an NFC card.
  SelfTest_NFCLowPowerCardDetect, //!< The NFC reader has woken up the microcontroller from standby.
  SelfTest_BackupVoltage,         //!< The backup capacitor voltage is high enough.
  SelfTest_BatteryVoltage,        //!< The internal battery voltage is within limits.
  SelfTest_ExternalVoltage,       //!< The external charger voltage is within limits.
  SelfTest_ChargeCurrent,         //!< The battery charge current is within limits.
  SelfTest_ChargerChemistry,      //!< The battery charger is for LiFePO chemistry and has correct cell count.
  SelfTest_BoostConverter,        //!< The power rail is 5.5 or 12.3 volts.
  SelfTest_ExternalIO,            //!< External inputs, outputs, and H-bridges test passed with test circuit.
  SelfTest_RS485,                 //!< Data is echoed back by the RS485 transceiver.
  SelfTest_Display,               //!< Data can be written to and read from the display.
  SelfTest_GUILeds,               //!< The IO expander on the GUI communicates over I2C.
  SelfTest_VCC,                   //!< The microcontroller's VCC is between 3.2 and 3.4 volts.
  SelfTest_MAX
};

/*! Postfix increment overload to be used with for() loops.*/
static inline ESelfTest operator++ (ESelfTest& lhs, int)
{
  ESelfTest retval = lhs;
  lhs = (ESelfTest)((int)lhs + 1);
  return retval;
}

class CSelfTest
{
  public:
    CSelfTest();
    ~CSelfTest();

    void              Setup();
    void              Run();
    void              RunCableTest();

    bool              Enabled() { return m_enabled; }

    bool              IsTestPassed(ESelfTest test);
    void              SetTestPassed(ESelfTest test);
    bool              AllTestsPassed();

    void              Print();
    void              SignalAllTestsPassed();

  private:

    void              OnSelfTestStart();
    void              OnSelfTestEnd();

    const char*       TestStr(ESelfTest test);
    const char*       TestDescr(ESelfTest test);

    void              TestBatteryCharger();
    void              TestSpeaker();
    void              TestInternalRTC();
    void              TestExternalRTC();
    void              TestSPIFlash();
    void              TestBackupVoltage();
    void              TestExternalIO();
    void              TestBoostConverterVoltage();
    void              TestRS485();
    void              TestServerComm();
    void              TestGUILeds();
    void              TestVCC();
    bool              TestCable();

    void              SetupExternalIOPins();
    void              CloseExternalIOPins();
    bool              TestExternalIOPins(bool cabletest);
    uint8_t           ExpectedInputStateTestCable(uint32_t out, uint32_t en, uint32_t b);
    uint8_t           ExpectedInputStateTestPCB(uint32_t out, uint32_t en, uint32_t b);
    void              PrintExternalIOError(uint8_t inputs, uint8_t expected,
                                           uint32_t out, uint32_t en, uint32_t b);
    bool              TestCableGrounds();

    void              RunRS485Test(CUsart& usart);

    bool              m_enabled;
    SemaphoreHandle_t m_mutex;
};

#endif //SELFTEST_H
