#ifndef SERVERCOMMUNICATION_H
#define SERVERCOMMUNICATION_H

#include "taskhandler.h"
#include "modembg96.h"
#include "fwupdate.h"
#include "inclstdint.h"
// #include "exti.h"
#include "lowpower.h"
#include "settings.h"

#include <FreeRTOS.h>
#include <semphr.h>

/*! URL to do HTTP POST with while testing.
    The IP address is from the "Ziggo zakelijk" internet connection at:
    WSS Infocard Systems
    Siliciumweg 3,
    3812 SV, Amersfoort */
#define SERVERCOMM_TEST_URL "https://test.wss.eu/AccelloAPI/req"

/*! URL to to HTTP POST with when this device is active.*/
#define SERVERCOMM_LIVE_URL "https://rsoffice.net/accelloAPI/req"

#define SERVERCOMM_FALLBACK_URL SERVERCOMM_LIVE_URL

/*! URL to do a firmware update with if the server can't be reached.
    This URL should never use HTTPS, only HTTP.
    All firmware images are encrypted and signed,
    so no further security is needed.*/
#define SERVERCOMM_EMERGENCY_FIRMWARE_URL "http://rsoffice.net/firmware/Accello/emergency.bin"

#define SERVERCOMM_BUFSIZE 4096

#define SERVERCOMM_ACTIVE_SIM_INTERNAL 0
#define SERVERCOMM_ACTIVE_SIM_EXTERNAL 1
#define SERVERCOMM_ACTIVE_SIM_DEFAULT  SERVERCOMM_ACTIVE_SIM_INTERNAL

#define COMM_REASON_OUTBOUND_CALL_CARD          0
#define COMM_REASON_WARNING_FILL_LEVEL_REACHED  1
#define COMM_REASON_MAX_FILL_LEVEL_REACHED      2
#define COMM_REASON_DEPOSIT_OPENING_BLOCKED     3
#define COMM_REASON_EMPTIED                     4
#define COMM_REASON_RTC_ALARM                   5
#define COMM_REASON_DEPOSIT_OPENING_STAYED_OPEN 6
#define COMM_REASON_DEPOSIT_OPENING_CLOSED      7
#define COMM_REASON_INSUFFICIENT_PREPAID_CREDIT 8
#define COMM_REASON_LOW_PREPAID_CREDIT          9
#define COMM_REASON_SELF_TEST                   10
#define COMM_REASON_FACTORY_RESET               11
#define COMM_REASON_MADE_ACTIVE                 12
#define COMM_REASON_DOOR_STAYED_OPEN            13
#define COMM_REASON_DOOR_CLOSED                 14

class CServerComm : public CTaskHandler, public CRTCInterrupt, public CLowPowerTask
{
  public:
    CServerComm();

    void              Setup();
    void              Start();

    void              SignalCommAttempt(uint32_t reason);
    bool              IsContacting() { return m_contacting;   }
    int32_t           RSSI()         { return m_modem.RSSI(); }

  private:
    virtual void      Process();

    bool              CheckActiveMode();
    void              RegisterCommReason(uint32_t reason);
    void              WaitNextAttempt();
    void              CheckAttemptsReset();
    void              EndAttempts();
    void              OnSuccessfulAttempt();
    void              OnFailedAttempt();
    uint32_t          MaxNextAttemptDelay(uint32_t* commfailures);
    void              SetNextAttemptAlarm();
    virtual void      RTCIRQHandler(ERTCIRQType irqtype);
    bool              CheckBatteryVoltage();
    void              SetActiveSIM();
    void              RecordModemStatus();
    void              RecordLastAttempt(bool success);
    void              RecordLocation();
    void              RecordInactiveAttempt();
    void              RecordInactiveDataCounters();

    bool              ProcessCommands();

    void              GetURLFromSettings(uint32_t numfails);
    bool              GetCommands();
    bool              ExecuteCommands();
    bool              ExecuteCommand(CborValue* command);

    bool              ProductionTest();
    bool              PostSettings();
    bool              GetSettings();
    bool              GetAuthorisationslist(CborValue* command);
    bool              PostStatus();
    bool              PostRegistrations();
    void              GetFirmware(CborValue* command);
    void              UpdateFirmware(const char* url);
    bool              GetImages();

    bool              Transceive(uint32_t txsize, bool* txcborcorrect = nullptr);
    bool              CheckServerDeviceID(CborValue* root);
    void              AddCommandDefaultFields(CborEncoder* root, const char* command);
    bool              CheckReplyStatus();

    SemaphoreHandle_t m_signal;
    bool              m_factoryreset;
    uint32_t          m_commreason;
    bool              m_contacting;

    CModemBG96        m_modem;
    CFwUpdate         m_fwupdate;
    uint32_t          m_activesim;

    char              m_url[STRING_VALUE_MAX_LENGTH];

    uint8_t           m_txbuf[SERVERCOMM_BUFSIZE];
    uint8_t           m_rxbuf[SERVERCOMM_BUFSIZE];
    uint8_t           m_commandbuf[SERVERCOMM_BUFSIZE];
};

#endif //SERVERCOMMUNICATION_H
