#include "6dsensor.h"
#include "clock.h"
#include "log.h"
// #include "speaker.h"
#include "controller.h"
#include "lock.h"

// #include <cstring>
#include <cmath>
#include <algorithm>
#include <cinttypes>
C6DSensor::C6DSensor() : m_spi(nullptr), m_int2signal(0)
{
}

void C6DSensor::Setup(CSpi* spi, SGPIOPin cs, SGPIOPin int2)
{
  m_enabled       = false;
  m_spi           = spi;
  m_cs            = cs;
  m_int2          = int2;
  m_wakeup        = false;
  m_acceldetected = false;
  m_int2signal    = xSemaphoreCreateBinary();

  m_sensor.write_reg = SWrite;
  m_sensor.read_reg  = SRead;
  m_sensor.handle    = this;

  g_exti.EnableGPIOIrq(m_int2, EXTI_Trigger_Rising, this);

  RegisterLowPowerTask("6dsensor");
}

void C6DSensor::Start()
{
 Create("6dsensor", HIGHEST_PRIORITY);
}

/*! Called on a rising edge of the INT2 pin on the LSM6DSOX sensor.*/
void C6DSensor::GPIOIRQHandler(SGPIOPin pin)
{
  if (m_int2signal)
  {
    BaseType_t taskwoken = pdFALSE;
    xSemaphoreGiveFromISR(m_int2signal, &taskwoken);
    portYIELD_FROM_ISR(taskwoken);
  }
}

void C6DSensor::Process()
{
  Initialize();
  if (!m_enabled)
  {
    PrintLog(LOG6DSENSOR, "Accelerometer wakeup disabled\n");
    AllowLowPower(true);
    for(;;)
      DelayMS(5000); //TODO: wake up when settings change
  }

  bool wakeup = m_wakeup;
  for(;;)
  {
    if (wakeup)
    {
      wakeup = false;
      if (ReadWakeup())
      {
        PrintLog(LOG6DSENSOR, "Accelerometer wakeup triggered\n");

        //Indicate to MeasureRotation() that a Status_AccelerationDetect should be
        //written. This can't be done immediately, because it can take a few seconds
        //to initialize the status array right after the microcontroller has woken
        //up from standby.
        m_acceldetected = true;
      }

      //If the accelerometer detected enough acceleration to trigger a wakeup,
      //measure rotation for some time to detect if the container has been emptied.
      MeasureRotation();
    }
    else
    {
      //Set up the accelerometer to generate an interrupt when it detects movement.
      SetupWakeupOnAcceleration();
      DisableGyro();
      AllowLowPower(true);

      //Wait for the accelerometer to generate an interrupt on the interrupt pin
      //that is connected to the IO expander. If the microcontroller is in low
      //power mode when this happens, it will wake up.
      while (!wakeup)
      {
        uint8_t input = g_controller.Inputs().WaitForChange(INPUT_6DSENSOR, 0, 5000);
        if (input == INPUT_6DSENSOR)
        {
          AllowLowPower(false);
          PrintLog(LOG6DSENSOR, "Woke up\n");
          wakeup = true;
        }
      }
    }
  }
}

#define INIT_TIMEOUT 5000000

/*! Resets the LSM6DSOX to default settings.
    \return true Communication with LSM6DSOX succeeded and reset was done.
    \return false On communication error or reset timeout.*/
bool C6DSensor::Initialize()
{
  m_enabled       = false;
  m_wakeup        = false;
  m_acceldetected = true;

  //Check if the sensor is present on the GUI pcb.
  int64_t switchtype = g_controller.Settings().GetInteger(Setting_TurnSwitchEnable);
  if (!CheckPresent(switchtype == SENSOR_ENABLED))
  { //If the sensor is not present, execute a reset command anyway, in case the
    //sensor is present but not working correctly.
    lsm6dsox_reset_set(&m_sensor, PROPERTY_ENABLE);
    return false;
  }

  PrintLog(LOG6DSENSOR, "Initializing LSM6DSOX\n");

  //Read the interrupt registers to clear any latched interrupt.
  if (ReadWakeup())
  {
    PrintLog(LOG6DSENSOR, "Device woke up because of accelerometer wakeup interrupt\n");
    m_wakeup = true;
  }

  //Reset all registers to factory defaults.
  lsm6dsox_reset_set(&m_sensor, PROPERTY_ENABLE);

  //Setting_TurnSwitchEnable decides how this device detects the emptying
  //of a container.
  if (switchtype != SENSOR_ENABLED)
  {
    PrintLog(LOG6DSENSOR, "Empty detect sensor disabled in settings\n");
    return false;
  }

  //Wait for the reset to factory default to complete.
  uint32_t start = Micros();
  uint8_t  swreset = 0;
  for(;;)
  {
    lsm6dsox_reset_get(&m_sensor, &swreset);
    if (swreset == 0)
    {
      break;
    }
    else
    {
      uint32_t now = Micros();
      if (now - start > INIT_TIMEOUT)
      {
        PrintLog(LOG6DSENSOR, "ERROR: timeout initializing LSM6DSOX\n");
        return false;
      }
    }

    DelayMS(1);
  }

  //Disable the I3C interface, it's not used.
  lsm6dsox_i3c_disable_set(&m_sensor, LSM6DSOX_I3C_DISABLE);

  m_enabled = true;

  PrintLog(LOG6DSENSOR, "LSM6DSOX initialization completed\n");

  return true;
}

/*! Checks if the 6D sensor is present.
    \param settingenabled true when Setting_TurnSwitchEnable enables the sensor.
    \return true There is a GUI PCB present with 6D sensor.
    \return false The 6D sensor is not present.*/
bool C6DSensor::CheckPresent(bool settingenabled)
{
  //A GUI PCB with LSM6DSOX sensor has a 1k pullup connected to PB13.
  //The pin has its internal pulldown enabled. A high level on PB13 indicates
  //that a GUI PCB with sensor is installed.
  bool present = GPIO_ReadInput(SENSOR_PRESENT);

  if (present)
  {
    PrintLog(LOG6DSENSOR, "GUI PCB installed with LSM6DSOX\n");
    return CheckDeviceID(present);
  }
  else
  {
    PrintLog(LOG6DSENSOR, "GUI PCB installed without LSM6DSOX\n");
    if (settingenabled)
    {
      const char* error = "empty detect sensor enabled but not installed";
      PrintLog(LOG6DSENSOR, "ERROR: %s\n", error);
      g_controller.Status().Write(Status_HardwareError, "%s", error);
    }

    //The device should not be present, check that the device ID can't be read.
    //If it can be read, there is probably something wrong with the hardware.
    CheckDeviceID(present);

    return false;
  }
}

/*! Reads the device ID and checks if it's LSM6DSOX_ID if the sensor is present,
    or that it's not LSM6DSOX_ID if the sensor is not present.
    \param present true if the sensor should be present.
    \return true The device ID is correct.
    \return false The device ID is wrong.*/
bool C6DSensor::CheckDeviceID(bool present)
{
  uint8_t deviceid = 0;
  lsm6dsox_device_id_get(&m_sensor, &deviceid);

  if (present)
  {
    if (deviceid != LSM6DSOX_ID)
    {
      PrintLog(LOG6DSENSOR, "ERROR: wrong device ID: %02" PRIX8 ", expected %02"
               PRIX8 "\n", deviceid, LSM6DSOX_ID);
      g_controller.Status().Write(Status_HardwareError, "6D sensor wrong device ID: %02"
                                  PRIX8 ", expected %02" PRIX8, deviceid, LSM6DSOX_ID);
      return false;
    }
  }
  else
  {
    if (deviceid == LSM6DSOX_ID)
    {
      PrintLog(LOG6DSENSOR, "ERROR: device ID is %02" PRIX8 " but sensor should "
               "not be present\n", deviceid);
      g_controller.Status().Write(Status_HardwareError, "6D sensor device ID is %02"
                                  PRIX8 ", should not be present",
                                  deviceid);
      return false;
    }
  }

  return true;
}

/*! Minimum value for the wakeup threshold in milli-g. If the sensor's resolution is
    set to 2/256, this is the lowest threshold in milli-g.*/
#define MIN_THRESHOLD         ((int64_t)8)

/*! Maximum value for the wakeup threshold in milli-g. If the sensor's resolution is
    set to 2/64, this is the highest threshold in milli-g.*/
#define MAX_THRESHOLD         ((int64_t)1968.75f)

#define THRESHOLD_BITS        6
#define THRESHOLD_MAX         ((1 << THRESHOLD_BITS) - 1)

#define SENSOR_RESOLUTION_64  (2.0f / 64.0f)
#define SENSOR_RESOLUTION_256 (2.0f / 256.0f)

#define MIN_DURATION          ((int64_t)1)
#define MAX_DURATION          ((int64_t)4)

/*! Sets up a high level interrupt on the INT1 pin when the acceleration
    exceeds a certain level.*/
void C6DSensor::SetupWakeupOnAcceleration()
{
  int64_t threshold = g_controller.Settings().GetInteger(Setting_AccelWakeupThreshold);
  threshold = std::min(std::max(MIN_THRESHOLD, threshold), MAX_THRESHOLD);

  int64_t duration = g_controller.Settings().GetInteger(Setting_AccelWakeupDuration);
  duration = std::min(std::max(MIN_DURATION, duration), MAX_DURATION);

  PrintLog(LOG6DSENSOR, "Enabling accelerometer wakeup, threshold: %" PRIi64
           " milli-g, duration: %" PRIi64 " sample%s\n", threshold, duration,
           duration == 1 ? "" : "s");

  //Set the accelerometer to measure at 12.5 hertz with a +-2g scale.
  lsm6dsox_xl_data_rate_set(&m_sensor, LSM6DSOX_XL_ODR_12Hz5);
  lsm6dsox_xl_full_scale_set(&m_sensor, LSM6DSOX_2g);

  //Pass the measured accelerometer data through a digital high-pass filter,
  //to filter out gravity.
  lsm6dsox_xl_hp_path_internal_set(&m_sensor, LSM6DSOX_USE_HPF);
  lsm6dsox_xl_hp_path_on_out_set(&m_sensor, LSM6DSOX_HP_ODR_DIV_20);
  lsm6dsox_xl_filter_lp2_set(&m_sensor, PROPERTY_ENABLE);

  //The wakeup threshold value is 6 bits, the range can be configured to 2/64
  //per lsb, or 2/256 per lsb. Set the wakeup threshold range so that it can
  //best approximate the AccelWakeupThreshold setting.
  int64_t max256 = lroundf((float)THRESHOLD_MAX * SENSOR_RESOLUTION_256 * 1000.0f);
  if (threshold <= max256)
  {
    lsm6dsox_wkup_ths_weight_set(&m_sensor, LSM6DSOX_LSb_FS_DIV_256);

    int32_t wakeup = lroundf((float)threshold / 1000.0f / SENSOR_RESOLUTION_256);
    lsm6dsox_wkup_threshold_set(&m_sensor, wakeup);
  }
  else
  {
    lsm6dsox_wkup_ths_weight_set(&m_sensor, LSM6DSOX_LSb_FS_DIV_64);

    int32_t wakeup = lroundf((float)threshold / 1000.0f / SENSOR_RESOLUTION_64);
    lsm6dsox_wkup_threshold_set(&m_sensor, wakeup);
  }

  //Set the wakeup duration in samples, it's -1 encoded.
  lsm6dsox_wkup_dur_set(&m_sensor, duration - 1);

  //Enable the wakeup interrupt on pin INT1.
  lsm6dsox_pin_int1_route_t int1_route;
  lsm6dsox_pin_int1_route_get(&m_sensor, &int1_route);
  int1_route.wake_up = PROPERTY_ENABLE;
  lsm6dsox_pin_int1_route_set(&m_sensor, int1_route);

  //Set the accelerometer to ultra low power mode, to save battery life.
  lsm6dsox_xl_power_mode_set(&m_sensor, LSM6DSOX_ULTRA_LOW_POWER_MD);

  //Set all interrupts to latched mode, so that the interrupt pin stays high
  //until the interrupt status has been read.
  lsm6dsox_int_notification_set(&m_sensor, LSM6DSOX_ALL_INT_LATCHED);

  //Give the high-pass filter some time to settle.
  PrintLog(LOG6DSENSOR, "Waiting for high-pass filter to settle\n");
  DelayMS(5000);

  //Read the interrupt status to clear any currently pending wakeup interrupt.
  ReadWakeup();

  PrintLog(LOG6DSENSOR, "Accelerometer wakeup enabled\n");
}

#define ACCELERATION_TIMEOUT 5000000

/*! Reads acceleration from the accelerometer, and converts it to G.
    \param acceleration The acceration in G, in order pitch, roll, raw.
    \return true Reading accelerometer data successful.
    \return false Timeout waiting for accelerometer data.*/
bool C6DSensor::ReadAcceleration(float acceleration[3])
{
  for (uint32_t i = 0; i < 3; i++)
    acceleration[i] = 0.0f;

  //Wait for the LSM6DSOX to read the accelerometer.
  uint32_t start = Micros();
  uint8_t  reg   = 0;
  do
  {
    lsm6dsox_xl_flag_data_ready_get(&m_sensor, &reg);
  }
  while (!reg && Micros() - start < ACCELERATION_TIMEOUT);

  if (!reg)
  {
    PrintLog(LOG6DSENSOR, "ERROR: timeout waiting for accelerometer data\n");
    return false;
  }

  //Read the accelerometer data and convert to G.
  int16_t raw[3] = {};
  lsm6dsox_acceleration_raw_get(&m_sensor, raw);

  for (uint32_t i = 0; i < 3; i++)
    acceleration[i] = lsm6dsox_from_fs2_to_mg(raw[i]) * 0.001f;

  return true;
}

/*! Reads all interrupts, and returns the state of the wakeup bit.
    This also clears all latched interrupts.*/
bool C6DSensor::ReadWakeup()
{
  lsm6dsox_all_sources_t all_source;
  lsm6dsox_all_sources_get(&m_sensor, &all_source);

  //For an unknown reason, the wake_up bit is not always set when one of the
  //wake_up_[x,y,z] is set.
  return all_source.wake_up   ||
         all_source.wake_up_x ||
         all_source.wake_up_y ||
         all_source.wake_up_z;
}

/*! Sets up the gyroscope in high performance mode, to get measurements with the
    least amount of noise.*/
void C6DSensor::EnableGyro()
{
  lsm6dsox_gy_data_rate_set(&m_sensor, LSM6DSOX_GY_ODR_12Hz5);
  lsm6dsox_gy_full_scale_set(&m_sensor, LSM6DSOX_250dps);
  lsm6dsox_gy_power_mode_set(&m_sensor, LSM6DSOX_GY_HIGH_PERFORMANCE);

  EnableGyroDataReadyInterrupt();
}

/* Sets the gyroscope in power-down mode, to minimize power consumption.*/
void C6DSensor::DisableGyro()
{
  lsm6dsox_gy_data_rate_set(&m_sensor, LSM6DSOX_GY_ODR_OFF);

  DisableGyroDataReadyInterrupt();
}

/*! Enables signaling a high level on INT2 when there is new gyro data to read.*/
void C6DSensor::EnableGyroDataReadyInterrupt()
{
  lsm6dsox_pin_int2_route_t int2_route;
  lsm6dsox_pin_int2_route_get(&m_sensor, nullptr, &int2_route);

  int2_route.drdy_g = PROPERTY_ENABLE;

  lsm6dsox_pin_int2_route_set(&m_sensor, nullptr, int2_route);
}

void C6DSensor::DisableGyroDataReadyInterrupt()
{
  lsm6dsox_pin_int2_route_t int2_route;
  lsm6dsox_pin_int2_route_get(&m_sensor, nullptr, &int2_route);

  int2_route.drdy_g = PROPERTY_DISABLE;

  lsm6dsox_pin_int2_route_set(&m_sensor, nullptr, int2_route);
}

#define GYRO_INTERRUPT_TIMEOUT 5000

/*! Waits for new data from the gyroscope.
    \return true There is new data to read from the gyroscope.
    \return false There is no new data to read from the gyroscope.*/
bool C6DSensor::WaitGyroDataRready()
{
  //The LSM6DSOX sensor makes INT2 high when there is new gyroscope data to read.
  //m_int2signal is then signaled from the GPIO interrupt.
  if (!xSemaphoreTake(m_int2signal, pdMS_TO_TICKS(GYRO_INTERRUPT_TIMEOUT)))
  {
    PrintLog(LOG6DSENSOR, "ERROR: INT2 interrupt timed out\n");
    g_controller.Status().Write(Status_HardwareError, "6D sensor INT2 timed out");
  }

  //Check if there is new gyroscope data to read.
  uint8_t dataready = 0;
  lsm6dsox_gy_flag_data_ready_get(&m_sensor, &dataready);

  return dataready;
}

/*! Reads the angular velocity from the gyro, and integrates it. Detects if the
    rotation is larger than the set threshold.
    \param threshold Threshold in degrees to check for.
    \param rotation Buffer into which angular velocity is integrated.
    \param previoustime Previous timestamp is stored here.
    \param ignore When above zero, ignore sample. Is decremented each function call.
    \return true The rotation is above the threshold.
    \return false The rotation is below the threshold.*/
bool C6DSensor::CheckEmptied(int64_t threshold, float rotation[3],
                             uint32_t* previoustime, uint32_t* ignore)
{
  uint32_t now = Micros();
  int16_t  rawrate[3] = {};
  lsm6dsox_angular_rate_raw_get(&m_sensor, rawrate);

  bool changed = false;
  if (*ignore == 0)
  { //Integrate the angular rate from the gyro, multiplied by the difference in
    //time from the last sample, to get the angular rotation.
    float timediff = (float)(now - *previoustime) * 0.000001f;
    float rate[3] = {};
    for (uint32_t i = 0; i < 3; i++)
    {
      rate[i] = lsm6dsox_from_fs250_to_mdps(rawrate[i]) * 0.001f;

      //Ignore angular rate below 1.5 degrees per second, to filter out the
      //zero-rate value from the sensor.
      if (fabsf(rate[i]) > 1.5f)
      {
        rotation[i] += rate[i] * timediff;
        changed = true;
      }
    }
  }
  else
  {
    (*ignore)--;
  }

  *previoustime = now;

  if (changed)
  {
    PrintLog(LOG6DSENSOR, "x:%.3f y:%.3f z:%.3f\n",
             rotation[0], rotation[1], rotation[2]);
  }

  //Check if the rotation is more than the threshold, if so,
  //assume the container has been emptied.
  for (uint32_t i = 0; i < 3; i++)
  {
    if (fabsf(rotation[i]) > (float)threshold)
      return true;
  }

  return false;
}

#define ROTATION_THRESHOLD_MIN ((int64_t)1)
#define ROTATION_THRESHOLD_MAX ((int64_t)360)

#define ROTATION_TIMEOUT_MIN   ((int64_t)1000)
#define ROTATION_TIMEOUT_MAX   ((int64_t)600000)

/*! Detects rotation by integrating angular rate from the gyroscope, signals that
    the container has been emptied if a certain threshold is crossed.*/
void C6DSensor::MeasureRotation()
{
  int64_t threshold = g_controller.Settings().GetInteger(Setting_RotationThreshold);
  int64_t timeout   = g_controller.Settings().GetInteger(Setting_RotationTimeout);
  threshold = std::min(std::max(threshold, ROTATION_THRESHOLD_MIN), ROTATION_THRESHOLD_MAX);
  timeout   = std::min(std::max(timeout, ROTATION_TIMEOUT_MIN), ROTATION_TIMEOUT_MAX);

  PrintLog(LOG6DSENSOR, "Detecting rotation using gyro, threshold %" PRIi64
           " degrees, timeout %" PRIi64 " milliseconds\n", threshold, timeout);

  EnableGyro();

  float    rotation[3] = {};
  int64_t  start = Millis();
  uint32_t previoustime;
  uint32_t ignore = 5; //Ignore the first few samples after the gyro has started.
  bool     emptied = false;

  do
  {
    //Wait for the gyro to signal on its INT2 pin when there is new data to read.
    if (!WaitGyroDataRready())
      continue;

    //Integrate the angular velocity from the gyro and check if the rotation
    //threshold has been reached.
    if (CheckEmptied(threshold, rotation, &previoustime, &ignore))
    {
      emptied = true;
      break;
    }
  }
  while (Millis() - start < timeout);

  PrintLog(LOG6DSENSOR, "Detecting rotation ended\n");

  if (m_acceldetected)
  {
    m_acceldetected = false;
    g_controller.Status().Write(Status_AccelerationDetect, "1");
  }

  g_controller.Status().Write(Status_RotationDetect, "%.1f %.1f %.1f",
                              rotation[0], rotation[1], rotation[2]);

  //Contact the server if the container has been emptied.
  if (emptied)
    g_controller.Deposit().ResetFillLevel();
}

#define CS_DELAY 10 //!< 10 microseconds chip select delay.

void C6DSensor::Select()
{
  DelayUS(CS_DELAY);
  GPIO_Reset(m_cs);
  DelayUS(CS_DELAY);
}

void C6DSensor::Deselect()
{
  DelayUS(CS_DELAY);
  GPIO_Set(m_cs);
  DelayUS(CS_DELAY);
}

int32_t C6DSensor::SWrite(void *handle, uint8_t reg, const uint8_t *bufp, uint16_t len)
{
  C6DSensor& sensor = *((C6DSensor*)handle);
  return sensor.Write(reg, bufp, len);
}

int32_t C6DSensor::SRead(void *handle, uint8_t reg, uint8_t *bufp, uint16_t len)
{
  C6DSensor& sensor = *((C6DSensor*)handle);
  return sensor.Read(reg, bufp, len);
}

#define LSM6DSOX_READ 0x80

int32_t C6DSensor::Write(uint8_t reg, const uint8_t *bufp, uint16_t len)
{
  Select();

  uint8_t cmd = reg & (~LSM6DSOX_READ);
  m_spi->Transceive(&cmd, nullptr, sizeof(reg));
  m_spi->Transceive(bufp, nullptr, len);

  Deselect();

  return 0;
}

int32_t C6DSensor::Read(uint8_t reg, uint8_t *bufp, uint16_t len)
{
  Select();

  uint8_t cmd = reg | LSM6DSOX_READ;
  m_spi->Transceive(&cmd, nullptr, sizeof(reg));
  m_spi->Transceive(nullptr, bufp, len);

  Deselect();

  return 0;
}
