#include "active.h"
#include "inclstdint.h"
#include "log.h"
#include "internalflash.h"
#include "backupregisters.h"
#include "controller.h"

#include <cstdio>

/*! Set to 1 to remember the active mode in a battery backed flag.
    DO NOT COMMIT THIS WHEN SET TO 1.*/
#define ACTIVE_TEST_MODE 0

#if ACTIVE_TEST_MODE
  #warning ACTIVE_TEST_MODE enabled, do not release this way!
#endif

//! Battery backed flag that indicates the active mode when ACTIVE_TEST_MODE is 1.
#define ACTIVE_TEST_FLAG 0x2000000

//! Address of OTP memory where the active state is stored.
#define ACTIVE_STATE_ADDR 0x1FFF79E0

//! Address to write 0 to to lock OTP data block 15.
#define ACTIVE_STATE_LOCK_ADDR 0x1FFF7A0F

/*! \return true if the device is in active mode.
    \return false if the device is in inactive mode.*/
bool DeviceActive()
{
#if ACTIVE_TEST_MODE
  return ReadBackupFlag(ACTIVE_TEST_FLAG);
#else
  return *((uint32_t*)ACTIVE_STATE_ADDR) != 0xFFFFFFFF;
#endif
}

/*! Writes data to OTP to switch the device to active mode,
    and locks the corresponding OTP block.
    \param uid The UID of the card that activated the device,
               or nullptr when the device is activated another way.*/
void SetDeviceActive(NFCUID* uid)
{
  if (DeviceActive())
  {
    PrintLog(LOGACTIVE, "ERROR: asked to switch device to active mode, "
             "but device is already active\n");
    return;
  }

  PrintLog(LOGACTIVE, "Switching device to active mode\n");

#if ACTIVE_TEST_MODE
  SetBackupFlag(ACTIVE_TEST_FLAG, true);
#else
  //Write the active state to the OTP, and lock the OTP block.
  uint32_t active = 0;
  uint8_t  lock   = 0;
  FlashWrite(ACTIVE_STATE_ADDR, (uint8_t*)&active, sizeof(active));
  FlashWrite(ACTIVE_STATE_LOCK_ADDR, &lock, sizeof(lock));
#endif

  //Convert the UID to a printable string.
  char uidstr[NFC_UID_MAX_SIZE * 2 + 1] = {};
  if (uid)
  {
    for (uint32_t i = 0; i < uid->size; i++)
      sprintf(uidstr + i * 2, "%02" PRIX8, uid->uid[i]);
  }

  //Write a MadeActive status, with the UID of the card as argument.
  g_controller.Status().Write(Status_MadeActive, "%s", uidstr);

  //Contact the server to indicate that the device has been made active.
  g_controller.ServerComm().SignalCommAttempt(COMM_REASON_MADE_ACTIVE);

  //Reset any stored inactive statuses.
  g_controller.Settings().SetBoolean(InactiveStatus_ServerComm_Tested,   false);
  g_controller.Settings().SetInteger(InactiveStatus_ServerComm_Attempts, 0);
  g_controller.Settings().SetInteger(InactiveStatus_ServerComm_TxBytes,  0);
  g_controller.Settings().SetInteger(InactiveStatus_ServerComm_RxBytes,  0);
  g_controller.Settings().SetBoolean(InactiveStatus_SelfTest_Passed,     false);
  g_controller.Settings().SetBoolean(InactiveStatus_CableTest_Passed,    false);
}

void PrintDeviceActive()
{
  PrintLog(LOGACTIVE, "Device is %sactive\n", DeviceActive() ? "" : "in");
  PrintDeviceActiveTest();
}

void PrintDeviceActiveTest()
{
#if ACTIVE_TEST_MODE
  PrintLog(LOGACTIVE, "\n*** ACTIVE TEST MODE ENABLED, DO NOT RELEASE THIS ***\n\n");
#endif
}

/*! Checks if the firmware has been updated, if so, activates the device if not
    activated yet. This prevents that a device in the field will switch to
    inactive mode after updating from older firmware which does not support
    activating yet.*/
void ActivateDeviceAfterFwUpdate()
{
  if (FirmwareUpdatedFlagSet())
  {
    SetFirmwareUpdatedFlag(false);

    if (!DeviceActive())
    {
      PrintLog(LOGACTIVE, "Activating device because firmware has been updated\n");
      SetDeviceActive(nullptr);
    }
  }
}
