#include "authlist.h"
#include "controller.h"
#include "accellospiflashaddrs.h"
#include "accellointflashaddrs.h"
#include "log.h"
#include "internalflash.h"
#include "active.h"
#include "speaker.h"

#include <cassert>
#include <cstddef>
#include <algorithm>
#include <cstring>
#include "stm32wss.h"

//make sure that state is the first member of SAuthListHeader
static_assert(offsetof(SAuthListHeader, state) == 0);

CAuthList::CAuthList()
{
  m_listptr         = nullptr;
  m_maxlistsize     = 0;
  m_spiflashpos     = 0;
  m_downloadsize    = 0;
  m_bytesdownloaded = 0;
  m_mutex           = 0;
  m_signaledcomm    = false;
}

void CAuthList::Setup()
{
  m_listptr = (SAuthListHeader*)FlashSectorAddress(AUTHLIST_START_SECTOR);

  //create a mutex to synchronize access to the internal auth list
  m_mutex = xSemaphoreCreateRecursiveMutex();

  //calculate the maximum size of auth list that can be stored in the internal flash
  m_maxlistsize = FlashSectorsSize(AUTHLIST_START_SECTOR, AUTHLIST_NUM_SECTORS) -
                  sizeof(SAuthListHeader);

  
  //calculate where in the SPI flash the reserved space for the auth list is,
  //it's reserved from the end of the SPI flash, because the stm32wss library
  //reserves space from the start of the SPI flash
  m_spiflashpos = g_controller.SPIFlash().Size() -
                  AUTHLIST_DOWNLOAD_SIZE - AUTHLIST_DOWNLOAD_ADDRESS;

  //the reserved space in the SPI flash should map exactly to one or more sectors
  uint32_t sectorsize = g_controller.SPIFlash().SectorSize();
  assert((m_spiflashpos % sectorsize) == 0);
  assert((AUTHLIST_DOWNLOAD_SIZE % sectorsize) == 0);

  //check if the internal flash has a correct authlist, without checking
  //the SHA256, because that takes too long
  if (!CheckHeaderOfInternalFlash(false))
    Recover();
}
    
/*! Checks if the reserved space for the auth list is correct.
    Erases the reserved space in the SPI flash.
    \return true if the reserved space was correctly erased.
    \return false if anything went wrong.*/
bool CAuthList::Prepare()
{
  m_bytesdownloaded = 0;
  WSS_sha256_init(&m_shactx);

  if (!g_controller.SPIFlash().Enabled())
  {
    PrintLog(LOGAUTHLIST, "ERROR: SPI FLASH disabled\n");
    return false;
  }

  //erase the section of SPI flash that is reserved for downloading the auth list
  if (!EraseSPIFlash())
    return false;

  return true;
}

/*! Erases the space in the SPI flash reserved for downloading the auth list.
    \return true if erasing the SPI flash succeeded.
    \return false if anything failed.*/
bool CAuthList::EraseSPIFlash()
{
  if (!g_controller.SPIFlash().Enabled())
  {
    PrintLog(LOGAUTHLIST, "ERROR: SPI FLASH disabled\n");
    return false;
  }

  uint32_t sectorsize  = g_controller.SPIFlash().SectorSize();
  uint32_t startsector = m_spiflashpos / sectorsize;
  uint32_t numsectors  = AUTHLIST_DOWNLOAD_SIZE / sectorsize;
  if (!g_controller.SPIFlash().Erase(startsector, numsectors))
  {
    PrintLog(LOGAUTHLIST, "ERROR: erasing flash failed\n");
    return false;
  }

  return true;
}

/*! Returns the maximum block size this class can handle through the Download() method,
    set to the page size of the SPI flash, this has the best throughput, because the
    spi flash is writing while the firmware is getting data from the modem.
    \return The maximum block size this class can handle through the Download method.*/
uint32_t CAuthList::DownloadBlockSize()
{
  return g_controller.SPIFlash().PageSize();
}

void CAuthList::StartDownload(const char* url, uint32_t httprspcode, uint32_t contentlength)
{
  m_downloadsize = contentlength;
}

/*! Called for each block of auth list that is downloaded by the modem.
    \param buf The buffer that holds the downloaded data.
    \param size The size of the downloaded data.*/
void CAuthList::Download(const uint8_t* buf, uint32_t size)
{
  assert(size <= DownloadBlockSize());

  if (size + m_bytesdownloaded > m_maxlistsize)
  {
    PrintLog(LOGAUTHLIST, "ERROR: being asked to store %" PRIu32
             " bytes, but max download size is %" PRIu32 " bytes\n",
             m_bytesdownloaded + size, m_maxlistsize);
    m_bytesdownloaded += size;
    return;
  }
  
  //feed the downloaded data to the SHA256 generator
  WSS_sha256_update(&m_shactx, buf, size);

  if (!g_controller.SPIFlash().Enabled())
  {
    PrintLog(LOGAUTHLIST, "ERROR: SPI FLASH disabled\n");
    return;
  }

  //write the buffer to the flash 
  uint32_t writepos = m_spiflashpos + sizeof(SAuthListHeader) + m_bytesdownloaded;
  if (!g_controller.SPIFlash().Write(writepos, buf, size, false))
  {
    PrintLog(LOGAUTHLIST, "ERROR: unable to write downloaded auth list data to SPI flash\n");
    return;
  }

  m_bytesdownloaded += size;
}

/*! Copies the auth list from the SPI flash to the internal flash.
    \return true if the copy succeeded.
    \return false if anything failed.*/
bool CAuthList::Install()
{
  if (m_bytesdownloaded > m_maxlistsize)
  {
    PrintLog(LOGAUTHLIST, "ERROR: not installing auth list, download is too large\n");
    return false;
  }

  //write a header describing the downloaded data to the SPI flash
  if (!WriteHeader())
  {
    PrintLog(LOGAUTHLIST, "Erasing auth list from SPI flash\n");
    EraseSPIFlash();
    return false;
  }

  //lock access to the internal auth list flash
  CLock lock(m_mutex);

  //copy everything from the SPI flash to the internal flash and
  //check if all data has been correcly copied to the internal flash
  if (!CopyToInternalFlash(m_bytesdownloaded) || !CheckHeaderOfInternalFlash())
  { //if the internal flash does not contain correct data at this point,
    //erase the auth list from the internal and SPI flash
    PrintLog(LOGAUTHLIST, "Erasing auth list from internal flash and SPI flash\n");
    FlashEraseSectors(AUTHLIST_START_SECTOR, AUTHLIST_NUM_SECTORS);
    EraseSPIFlash();
    return false;
  }

  return true;
}

/*! Writes a header of the auth list to the SPI flash.*/
bool CAuthList::WriteHeader()
{
  if (!g_controller.SPIFlash().Enabled())
  {
    PrintLog(LOGAUTHLIST, "ERROR: SPI FLASH disabled\n");
    return false;
  }

  //create a header with the auth list size and sha256 hash
  SAuthListHeader header;
  header.state = AUTH_LIST_VALID;
  header.size  = m_downloadsize;
  WSS_sha256_final(&m_shactx, header.hash);

  CLogLock loglock;
  PrintLog(LOGAUTHLIST, "Writing auth list header with hash ");
  PrintLogHexBuf(LOGAUTHLIST, header.hash, SHA256_BLOCK_SIZE, "\n");
  loglock.Leave();

  if (!g_controller.SPIFlash().Write(m_spiflashpos, (uint8_t*)&header, sizeof(header)))
  {
    PrintLog(LOGAUTHLIST, "ERROR: unable to write auth list header to SPI flash\n");
    return false;
  }

  return true;
}

/*! Copies the auth list and header from the SPI flash to the internal flash.
    \param size Size of the auth list in the SPI flash.
    \return true if copying everything succeeded.
    \return false on any error.*/
bool CAuthList::CopyToInternalFlash(uint32_t size)
{
  PrintLog(LOGAUTHLIST, "Copying auth list of %" PRIu32
           " bytes from SPI flash to internal flash\n", size);

  //lock access to the internal auth list flash
  CLock lock(m_mutex);

  //mark the current internal auth list as invalid, by setting the state field to 0
  //if erasing or writing the internal flash is interrupted, it can be detected at the next
  //boot by checking the state field, it will most likely not be AUTH_LIST_VALID
  SAuthListHeader header = *m_listptr;
  header.state = AUTH_LIST_INVALID;
  if (!FlashWrite((uint32_t)m_listptr, (uint8_t*)&header, sizeof(header)))
  {
    PrintLog(LOGAUTHLIST, "ERROR: unable to mark internal auth list as invalid\n");
    return false;
  }

  //erase the internal flash reserved for the auth list
  if (!FlashEraseSectors(AUTHLIST_START_SECTOR, AUTHLIST_NUM_SECTORS))
  {
    PrintLog(LOGAUTHLIST, "ERROR: unable to erase internal flash\n");
    return false;
  }

  //copy the header and auth list from the SPI flash to the internal flash,
  //skipping over the state field at the start
  uint32_t statesize = sizeof(header.state);
  if (!CopyFlashData(m_spiflashpos + statesize, (uint32_t)m_listptr + statesize,
                     size + sizeof(SAuthListHeader) - statesize))
    return false;

  //copy the state field to the internal flash last, this indicates that
  //the internal flash now holds a valid copy of the auth list
  if (!CopyFlashData(m_spiflashpos, (uint32_t)m_listptr, statesize))
    return false;

  return true;
}

/*! Copies data from the SPI flash to the internal flash, does not erase.
    \param readaddr The address to read from in the SPI flash.
    \param writeaddr The address to write to in the internal flash.
    \param size The number of bytes to copy from the SPI flash to the internal flash.
    \return if the copy was successful.
    \return false on any error.*/
bool CAuthList::CopyFlashData(uint32_t readaddr, uint32_t writeaddr, uint32_t size)
{
  if (!g_controller.SPIFlash().Enabled())
  {
    PrintLog(LOGAUTHLIST, "ERROR: SPI FLASH disabled\n");
    return false;
  }

  //sanity checks
  assert(readaddr >= m_spiflashpos);
  assert((readaddr + size) <= (m_spiflashpos + m_maxlistsize + sizeof(SAuthListHeader)));
  assert(writeaddr >= (uint32_t)m_listptr);
  assert((writeaddr + size) <= ((uint32_t)m_listptr + m_maxlistsize + sizeof(SAuthListHeader)));

  //lock access to the internal auth list flash
  CLock lock(m_mutex);

  //copy the header and auth list from the SPI flash to the internal flash
  uint32_t bytescopied = 0;
  while (bytescopied < size)
  {
    uint32_t copysize = std::min((uint32_t)sizeof(m_copybuf), size - bytescopied);
    uint32_t readpos = readaddr + bytescopied;
    if (!g_controller.SPIFlash().Read(readpos, m_copybuf, copysize))
    {
      PrintLog(LOGAUTHLIST, "ERROR: unable to read %" PRIu32
               " bytes from SPI flash address %" PRIu32 "\n",
               copysize, readpos);
    }

    uint32_t writepos = writeaddr + bytescopied;
    if (!FlashWrite(writepos, m_copybuf, copysize))
    {
      PrintLog(LOGAUTHLIST, "ERROR: unable to write %" PRIu32 " bytes"
               " to internal flash address %" PRIu32 "\n", copysize, writepos);
      return false;
    }

    bytescopied += copysize;
  }

  return true;
}

/*! Checks if the header of the internal flash is correct, and if the hash in
    the header matches the hash generated over the stored auth list.
    \param checkhash When true, checks the SHA256 stored in the header,
                     this can take up to 500 milliseconds.
    \return true if everything is correct.
    \return false if anything is incorrect.*/
bool CAuthList::CheckHeaderOfInternalFlash(bool checkhash /*= true*/)
{
  //lock access to the internal auth list flash
  CLock lock(m_mutex);

  //check the state field, it indicates if the internal auth list is valid or not
  if (m_listptr->state != AUTH_LIST_VALID)
  {
    PrintLog(LOGAUTHLIST, "ERROR: auth list in internal flash invalid, state field: %"
             PRIX32 "\n", m_listptr->state);
    return false;
  }

  //make sure a correct size has been written
  if (m_listptr->size > m_maxlistsize)
  {
    PrintLog(LOGAUTHLIST, "ERROR: auth list in internal flash invalid, size field: %"
             PRIu32 ", maximum allowed size: %" PRIu32 "\n", m_listptr->size, m_maxlistsize);
    return false;
  }

  if (!checkhash)
  {
    PrintLog(LOGAUTHLIST, "Auth list of %" PRIu32 " bytes "
             "stored in internal flash\n", m_listptr->size);
    return true;
  }

  //generate a SHA256 over the auth list stored in the internal flash
  uint8_t hash[SHA256_BLOCK_SIZE];
  SHA256_CTX shactx;
  WSS_sha256_init(&shactx);
  WSS_sha256_update(&shactx, (uint8_t*)m_listptr + sizeof(SAuthListHeader), m_listptr->size);
  WSS_sha256_final(&shactx, hash);

  //compare the generated has with the hash stored in the internal flash
  if (!std::equal(hash, hash + SHA256_BLOCK_SIZE, m_listptr->hash))
  {
    CLogLock loglock;
    PrintLog(LOGAUTHLIST, "ERROR: generated hash ");
    PrintLogHexBuf(LOGAUTHLIST, hash, SHA256_BLOCK_SIZE);
    PrintLog(LOGAUTHLIST, " does not match stored hash ");
    PrintLogHexBuf(LOGAUTHLIST, m_listptr->hash, SHA256_BLOCK_SIZE, "\n");
    return false;
  }

  CLogLock loglock;
  PrintLog(LOGAUTHLIST, "Hash of auth list of %" PRIu32 " bytes "
           "stored in internal flash is correct: ", m_listptr->size);
  PrintLogHexBuf(LOGAUTHLIST, hash, SHA256_BLOCK_SIZE, "\n");
  loglock.Leave();

  return true;
}

/*! Checks if the header of the auth list in the SPI flash is correct,
    and if the hash in the header matches the hash generated over the
    auth list stored in the SPI flash.
    \param erased Set to true when the state field in the header is set to AUTH_ERASED.
    \param size The size of the auth list in the SPI flash is returned here.
    \return true if everything is correct.
    \return false if anything is incorrect.*/
bool CAuthList::CheckHeaderOfSPIFlash(bool* erased, uint32_t* size)
{
  *erased = false;
  *size   = 0;

  if (!g_controller.SPIFlash().Enabled())
  {
    PrintLog(LOGAUTHLIST, "ERROR: SPI FLASH disabled\n");
    return false;
  }

  //read the header from the SPI flash
  SAuthListHeader header;
  if (!g_controller.SPIFlash().Read(m_spiflashpos, (uint8_t*)&header, sizeof(header)))
  {
    PrintLog(LOGAUTHLIST, "ERROR: unable to read header from SPI flash\n");
    return false;
  }

  //check the state field, it indicates if the auth list is valid or not
  if (header.state == AUTH_LIST_ERASED)
  {
    *erased = true;
    PrintLog(LOGAUTHLIST, "No auth list stored in SPI flash\n");
    return false;
  }
  else if (header.state != AUTH_LIST_VALID)
  {
    PrintLog(LOGAUTHLIST, "ERROR: auth list in SPI flash invalid, state field: %"
             PRIX32 "\n", header.state);
    return false;
  }

  //check if the size in the header is correct
  if (header.size > m_maxlistsize)
  {
    PrintLog(LOGAUTHLIST, "ERROR: auth list in SPI flash invalid, size field: %"
             PRIu32 ", maximum allowed size: %" PRIu32 "\n", header.size, m_maxlistsize);
    return false;
  }

  //generate a hash over the auth list stored in the SPI flash
  SHA256_CTX shactx;
  WSS_sha256_init(&shactx);
  uint32_t bytesread = 0;
  uint32_t listpos = m_spiflashpos + sizeof(header);
  while (bytesread < header.size)
  {
    uint32_t readbytes = std::min(header.size - bytesread, (uint32_t)sizeof(m_copybuf));
    uint32_t readpos   = listpos + bytesread;
    if (!g_controller.SPIFlash().Read(readpos, m_copybuf, readbytes))
    {
      PrintLog(LOGAUTHLIST, "ERROR: unable to read %" PRIu32
               " bytes from SPI flash address %" PRIu32 "\n",
               readbytes, readpos);
      return false;
    }

    WSS_sha256_update(&shactx, m_copybuf, readbytes);

    bytesread += readbytes;
  }

  uint8_t hash[SHA256_BLOCK_SIZE]; 
  WSS_sha256_final(&shactx, hash);

  //check if the hash generated over the auth list is the same as the hash stored in the header
  if (!std::equal(hash, hash + SHA256_BLOCK_SIZE, header.hash))
  {
    CLogLock loglock;
    PrintLog(LOGAUTHLIST, "ERROR: hash of auth list of %" PRIu32
             " bytes in SPI flash is incorrect, expected ", header.size);
    PrintLogHexBuf(LOGAUTHLIST, header.hash, SHA256_BLOCK_SIZE);
    PrintLog(LOGAUTHLIST, ", calulated ");
    PrintLogHexBuf(LOGAUTHLIST, hash, SHA256_BLOCK_SIZE, "\n");
    return false;
  }

  CLogLock loglock;
  PrintLog(LOGAUTHLIST, "Auth list of %" PRIu32
           " bytes in SPI flash has correct hash ", header.size);
  PrintLogHexBuf(LOGAUTHLIST, hash, SHA256_BLOCK_SIZE, "\n");
  loglock.Leave();

  *size = header.size;

  return true;
}

/*! Will check if the SPI has a valid auth list stored, if so it will copy
    it to the internal flash.
    Is used if after an auth list download, the copying of the auth list was
    interrupted, and the internal flash has been left in an intermediate state.*/
void CAuthList::Recover()
{
  PrintLog(LOGAUTHLIST, "Trying to recover auth list from SPI flash\n");

  //lock access to the internal auth list flash
  CLock lock(m_mutex);

  //check if there is a valid auth list stored in the SPI flash
  bool     erased;
  uint32_t size;
  if (!CheckHeaderOfSPIFlash(&erased, &size))
  {
    PrintLog(LOGAUTHLIST, "Unable to recover auth list from SPI flash\n");
    if (!erased)
    { //if the auth list is not in the erased state, there is an invalid
      //auth list stored, erase that from the SPI flash
      PrintLog(LOGAUTHLIST, "Erasing invalid auth list from SPI flash\n");
      EraseSPIFlash();
    }
    return;
  }

  //copy the auth list from the SPI flash to the internal flash,
  //and check afterwards if it's correct
  if (!CopyToInternalFlash(size) || !CheckHeaderOfInternalFlash())
  { //if the internal flash does not contain correct data at this point,
    //erase the auth list from the internal and SPI flash
    PrintLog(LOGAUTHLIST, "Erasing auth list from internal flash and SPI flash\n");
    FlashEraseSectors(AUTHLIST_START_SECTOR, AUTHLIST_NUM_SECTORS);
    EraseSPIFlash();
  }
}

/*! Checks if the UID of the card is present on the auth list.
    It will then execute the function that is linked with the card on the auth list.
    If the card is not present on the auth list, it will show a denied image on the display.
    \param uid The UID of the card to authenticate.
    \param keyfailed Set to true if secure card key authentication fails.
    \return true if the UID of the card is present on the auth list.
    \return false if the UID of the card is not present on the auth list.*/
bool CAuthList::Authenticate(NFCUID* uid, bool* keyfailed)
{
  *keyfailed = false;

  //Check if this card should be treated differently because self test is enabled.
  if (AuthenticateSelfTest(uid))
    return true;

  //Check if the device is in inactive mode, and should be activated on this UID.
  if (AuthenticateActivate(uid))
    return true;

  //Check if this card should be treated differently because of the reset reason.
  if (AuthenticateResetReason(uid))
    return true;

  //Lock access to the internal auth list flash.
  CLock lock(m_mutex);

  //Check if the auth list is present, if not, do a regular deposit.
  if (AuthenticateNoListPresent(uid, lock))
    return true;

  //Check if the UID is present on the auth list, if so, execute the action
  //that is coupled to it.
  if (AuthenticateAuthList(uid, lock, keyfailed))
    return true;

  lock.Leave();

  //Check if a MIFARE Ultralight C card should be programmed.
  if (ProgramSecuredCard(uid))
    return true;

  //Check what to do when the UID is not present on the auth list.
  if (AuthenticateFallback(uid))
    return true;

  return false; //This card is not allowed to execute any action.
}

/*! Called before checking the auth list, checks if the card should be treated
    as a status card, because self test mode is enabled.
    \param uid The UID of the card that has been read.
    \return true if this card has been treated as a status card.
    \return false if no action was performed.*/
bool CAuthList::AuthenticateSelfTest(NFCUID* uid)
{
  if (g_controller.SelfTest().Enabled())
  {
    Execute(uid, CardType_Show_GPRS_Signal);
    return true;
  }

  return false;
}

/*! Called before checking the auth list, checks if the card should be treated
    as an outbound call card.
    \param uid The UID of the card that has been read.
    \return true if this card has been treated as an outbound call card.
    \return false if no action was performed.*/
bool CAuthList::AuthenticateResetReason(NFCUID* uid)
{
  //If the microcontroller has been reset by the reset pin only,
  //or there has been a power on reset, and a card
  //is read within 5 seconds after boot, communicate with the server.
  //This way, server communication can be started using the reset pin,
  //or by doing a power cycle by holding a magnet to the reed switch,
  //while an nfc card is held to the reader coil.
  if (!m_signaledcomm && Millis() < 5000 &&
      ((ResetFlags() == ResetFlagMask(RCC_FLAG_PINRST)) ||
       (ResetFlags() & ResetFlagMask(RCC_FLAG_PORRST))))
  {
    m_signaledcomm = true;
    Execute(uid, CardType_Outbound_Call);
    return true;
  }

  return false;
}

/* Checks if the UID of an NFC card is on the list of built-in UIDs that are
   allowed to activate the device.
   \param uid The UID of the card that has just been read.
   \return true The UID is on the activation list.
   \return false The UID is not on the activation list.*/
bool CAuthList::IsActivateUID(NFCUID* uid)
{
  //Built-in UID list of activate cards.
  static const NFCUID uids[] =
  {
    {{0x99, 0x99, 0xB0, 0x9A}, 4}, //Bob.
    {{0x99, 0xA3, 0xA5, 0xDA}, 4}, //RPD.
    {{0x99, 0xA8, 0x77, 0xCA}, 4}, //RPD.
    {{0xBC, 0x0C, 0x39, 0xF4}, 4}, //TJ.
    {{0xBC, 0x0C, 0xB2, 0x94}, 4}, //GP
    {{0xBC, 0x0F, 0x76, 0x64}, 4}, //GJL
    {{0xBC, 0x11, 0x3D, 0x04}, 4}, //TJ
    {{0xBC, 0x11, 0x51, 0xA4}, 4},
    {{0xBC, 0x1A, 0x81, 0xD4}, 4},
    {{0xBC, 0x1A, 0xB9, 0xF4}, 4},
    {{0xBC, 0x1B, 0x22, 0x84}, 4},
    {{0xBC, 0x1C, 0x6A, 0xA4}, 4},
    {{0xBC, 0x1D, 0xE5, 0x44}, 4},
    {{0xCB, 0xB0, 0x04, 0x44}, 4},
    {{0xCB, 0xB2, 0x75, 0x34}, 4},
    {{0xE5, 0x8A, 0x8F, 0x83}, 4},
    {{0xE5, 0x91, 0x33, 0x63}, 4},
    {{0xE5, 0x94, 0x85, 0xE3}, 4},
    {{0xE5, 0x95, 0x22, 0x73}, 4},
    {{0xEB, 0xAE, 0xB3, 0xA4}, 4},
  };

  //Check if the UID is on the list of built-in UIDs.
  for (uint32_t i = 0; i < sizeof(uids) / sizeof(uids[0]); i++)
  {
    if (uids[i].size == uid->size &&
        memcmp(uids[i].uid, uid->uid, uids[i].size) == 0)
      return true;
  }

  return false;
}

/* Checks if the UID of an NFC card is on the list of built-in UIDs that are
   allowed to always activate the device, even if tests have failed.
   \param uid The UID of the card that has just been read.
   \return true The UID is on the always activation list.
   \return false The UID is not on the always activation list.*/
bool CAuthList::IsAlwaysActivateUID(NFCUID* uid)
{
    // EDWIN
    // always activate
    // return true;

  //Built-in UID list of always activate cards.
  static const NFCUID uids[] =
  {
    {{0x44, 0x82, 0xCC, 0x12}, 4}, //Bob.
    {{0x44, 0x84, 0xB7, 0x12}, 4}, //RPD.
    {{0x44, 0x85, 0x5C, 0x12}, 4}, //RPD.
    {{0x44, 0x85, 0xBE, 0xC2}, 4}, //TJ.
    {{0x44, 0x85, 0xFE, 0xF2}, 4}, //GP
    {{0x7C, 0x3E, 0x1F, 0xC9}, 4}, //GJL
    {{0x7C, 0x3E, 0x20, 0x89}, 4}, //TJ
    {{0x7C, 0x3E, 0x27, 0xF9}, 4},
    {{0x7C, 0x3E, 0x28, 0x79}, 4},
    {{0x7C, 0x45, 0xBB, 0xD9}, 4},
    {{0x7C, 0x45, 0xC5, 0x79}, 4},
    {{0xBC, 0x0E, 0xDA, 0x74}, 4},
    {{0xBC, 0x0F, 0x18, 0x04}, 4},
    {{0xBC, 0x18, 0x49, 0x24}, 4},
    {{0xBC, 0x1B, 0x06, 0x84}, 4},
    {{0xDC, 0x21, 0x26, 0xE1}, 4},
    {{0xDC, 0x37, 0x78, 0xE1}, 4},
    {{0xDC, 0x6B, 0x0B, 0xB1}, 4},
    {{0xDC, 0x7C, 0xA1, 0x21}, 4},
    {{0xE5, 0x93, 0xEC, 0x13}, 4},
  };

  //Check if the UID is on the list of built-in UIDs.
  for (uint32_t i = 0; i < sizeof(uids) / sizeof(uids[0]); i++)
  {
    if (uids[i].size == uid->size &&
        memcmp(uids[i].uid, uid->uid, uids[i].size) == 0)
      return true;
  }

  return false;
}

/*! Checks if the device is in inactive mode, if so, activates the device if the
    read UID is on a hardcoded list.
    \param uid The UID of the card that has been read.
    \return true The device is in inactive mode, and has been switched to active
                 mode, or the inactive status screen is showing.
    \return false The device is in active mode.*/
bool CAuthList::AuthenticateActivate(NFCUID* uid)
{
  if (DeviceActive())
    return false; //Device already in active mode, so card not authenticated.

  //Check if the UID is on the list of built-in UIDs.
  bool isactivate = IsActivateUID(uid);
  bool isalwaysactivate = IsAlwaysActivateUID(uid);

  if (!isactivate && !isalwaysactivate)
  {
    //If the cable test has not been passed yet, run it.
    if (!g_controller.Settings().GetBoolean(InactiveStatus_CableTest_Passed))
      g_controller.Deposit().CableTest();

    //If the card is not allowed to switch the device to active mode,
    //show the inactive display.
    g_controller.UserInterface().Set(UIState_Inactive);
    DelayMS(10000);

    return true;
  }

  bool selftest  = g_controller.Settings().GetBoolean(InactiveStatus_SelfTest_Passed);
  bool cabletest = g_controller.Settings().GetBoolean(InactiveStatus_CableTest_Passed);

  //If the tests are not passed, activation of the device is not allowed,
  //unless the card is on the always activate list.
  if ((!selftest || !cabletest) && !isalwaysactivate)
  {
    if (!selftest)
      PrintLog(LOGAUTHLIST, "ERROR: not allowing activation because self test "
               "not passed\n");

    if (!cabletest)
      PrintLog(LOGAUTHLIST, "ERROR: not allowing activation because cable test "
               "not passed\n");

    const char* displaystr = "";
    if (!selftest && !cabletest)
      displaystr = "Not passed:\nSelf test\nCable test";
    else if (!selftest)
      displaystr = "Not passed:\nSelf test";
    else if (!cabletest)
      displaystr = "Not passed:\nCable test";

    //Show on the display which tests have not been passed yet.
    g_controller.UserInterface().SetString(displaystr);
    g_controller.GUILeds().Off(GUILED_ALL_GRN);
    g_controller.GUILeds().On(GUILED_ALL_RED);
    CSpeaker::Beep(5000);
    DelayMS(5000);
    g_controller.GUILeds().Off(GUILED_ALL_RED);

    return true;
  }

  //Activate the device.
  g_controller.UserInterface().Set(UIState_Success);
  CSpeaker::Beep(50, 20);
  SetDeviceActive(uid);
  DelayMS(1000);

  return true;
}

/*! Checks if the auth list is present, if not, will treat this card as
    a standard deposit.
    \param uid The UID of the card that has been read.
    \return true if no list is present and a deposit has been done.
    \return false if no action has been performed.*/
bool CAuthList::AuthenticateNoListPresent(NFCUID* uid, CLock& lock)
{
  //if no auth list is stored in the internal flash,
  //give any card access as a normal waste deposit
  if (!CheckHeaderOfInternalFlash(false))
  {
    CLogLock loglock;
    PrintLog(LOGAUTHLIST, "Unable to authenticate UID ");
    PrintNFCUID(LOGAUTHLIST, uid);
    PrintLog(LOGAUTHLIST, " because auth list is not present, assuming %s(%i) type card\n",
             CardTypeToStr(CardType_Registration), (int)CardType_Registration);
    loglock.Leave();

    lock.Leave(); //Don't hold the lock while executing the action.
    Execute(uid, CardType_Registration);

    return true;
  }

  return false;
}

/*! Checks if a UID is present on the auth list, and performs the action that
    is coupled with the UID.
    \param uid the UID of the card that has been read.
    \param lock The lock on m_mutex, will be released before executing an action.
    \param keyfailed Set to true if the card fails key authentication.
    \return true if the UID is present on the auth list.
    \return false if the UID is not present on the auth list.*/
bool CAuthList::AuthenticateAuthList(NFCUID* uid, CLock& lock, bool* keyfailed)
{
  //convert the UID of the card to uint64_t
  //abort if the UID is too long
  uint64_t carduid = 0;
  if (!ConvertUID(uid, &carduid))
    return false;

  //UIDs can be 4 or 7 bytes
  bool card7bytes = uid->size == 7;

  //the auth list contains UIDs, 8 bytes per UID in little endian format
  //when converted to uint64_t, the most significant byte is the type byte
  uint32_t  numuids = m_listptr->size / sizeof(uint64_t);
  uint64_t* listptr = m_listptr->uidlist;
  uint64_t* endptr  = listptr + numuids;
  while (listptr != endptr)
  {
    uint64_t  listuid    = *listptr;
    uint8_t   type       = listuid >> TYPE_BYTE_POSITION;
    bool      list7bytes = (type & FLAG_7_BYTE_UID) == FLAG_7_BYTE_UID;

    if (list7bytes)
      listuid &= UID_7_BYTE_MASK;
    else
      listuid &= UID_4_BYTE_MASK;

    //Check if the UID of the read card is the same as the UID of this list entry.
    if (card7bytes == list7bytes && carduid == listuid)
    {
      //Leave the lock before executing, so that it doesn't block other tasks
      //from updating the auth list.
      lock.Leave();

      //If the card is a secured type, perform 3DES key authentication.
      ECardType cardtype = (ECardType)(type & CARD_TYPE_MASK);
      bool authenticated = AuthenticateSecuredCard(uid, cardtype);

      //Abort if the authentication failed.
      if (!authenticated)
      {
        CLogLock loglock;
        PrintLog(LOGAUTHLIST, "ERROR: UID ");
        PrintNFCUID(LOGAUTHLIST, uid);
        PrintLog(LOGAUTHLIST, " is present on auth list as %s(%i) "
                 "but failed key authentication\n",
                 CardTypeToStr(cardtype), (int)cardtype);
        loglock.Leave();

        *keyfailed = true;

        return false;
      }

      CLogLock loglock;
      PrintLog(LOGAUTHLIST, "UID ");
      PrintNFCUID(LOGAUTHLIST, uid);
      PrintLog(LOGAUTHLIST, " authenticated successfully as %s(%i)\n",
               CardTypeToStr(cardtype), (int)cardtype);
      loglock.Leave();

      //read 8 bytes after the list entry, if available
      //prepaid counters are stored here for prepaid card types
      uint64_t trailingdata = listptr + 1 != endptr ? *(listptr + 1) : 0;

      //Execute the function linked to the card.
      Execute(uid, cardtype, trailingdata);

      return true;
    }

    listptr++;
  }

  CLogLock loglock;
  PrintLog(LOGAUTHLIST, "UID ");
  PrintNFCUID(LOGAUTHLIST, uid);
  PrintLog(LOGAUTHLIST, " is not present on the auth list\n");
  loglock.Leave();

  return false;
}

/*! Checks the operating mode setting, if set to NoList, all cards are treated
    as a standard deposit.
    \param uid The UID of the card that has been read.
    \return true if a regular deposit has been done.
    \return false if no action has been performend.*/
bool CAuthList::AuthenticateFallback(NFCUID* uid)
{
  //The operatingmode setting can be set to NoList, BlackList and WhiteList.
  //If the operatingmode setting is not set to WhiteList, all cards are accepted
  //as registration type, which allows a standard deposit.
  EOperatingMode operatingmode = (EOperatingMode)g_controller.Settings().GetInteger(Setting_Operatingmode);
  if (operatingmode == OperatingMode_NoList)
  {
    PrintLog(LOGAUTHLIST, "Operating mode is set to %s, accepting unknown cards as"
             " registration type\n", OperatingModeToStr(operatingmode));
    Execute(uid, CardType_Registration);
    return true;
  }
  else if (operatingmode != OperatingMode_WhiteList)
  {
    PrintLog(LOGAUTHLIST, "ERROR: operating mode %" PRIu32 ":%s not supported, "
             "accepting unknown cards as registration type\n",
             (uint32_t)operatingmode, OperatingModeToStr(operatingmode));
    Execute(uid, CardType_Registration);
    return true;
  }

  return false;
}

/*! Programs a secret key into a secured card. This can only be
    enabled at compile time.
    \param uid The UID of the card to program
    \return true An attempt was made to program the card.
    \return false Secured card programming is disabled.*/
bool CAuthList::ProgramSecuredCard(NFCUID* uid)
{
  return g_controller.Reader().ProgramSecuredCard(uid);
}

/*! Performs 3DES authentication on a secured card.
    \param uid the UID of the card to authenticate.
    \param cardtype The type of the card from the auth list.
    \return true The card is not a secured card type.
    \return true Authentication succeeded.
    \return false Authentication failed.*/
bool CAuthList::AuthenticateSecuredCard(NFCUID* uid, ECardType cardtype)
{
  if (IsSecuredCard(cardtype))
    return g_controller.Reader().AuthenticateSecuredCard(uid);
  else
    return true;
}

/*! Executes the function for the card type.
    \param uid The UID of the card.
    \param cardtype The type of the card from the auth list.
    \param trailingdata 8 bytes of data stored after the UID on the authlist.*/
void CAuthList::Execute(NFCUID* uid, ECardType cardtype, uint64_t trailingdata /*= 0*/)
{
  EDepositResult result = DepositResult_Closed;
  bool registered = false;

  m_prepaid.Reset();

  ECardType exectype = FromSecuredType(cardtype);

  if (exectype == CardType_Registration)
  { //waste deposit
    result = g_controller.Deposit().Process(uid, false, &registered, exectype);
  }
  else if (exectype == CardType_Prepaid)
  { //waste deposit with prepaid card
    m_prepaid.FromAuthList(uid, trailingdata);
    result = g_controller.Deposit().Process(uid, false, &registered, exectype, &m_prepaid);
  }
  else if (exectype == CardType_Emptying)
  {
    //lock the user interface so that no other task interrupts the fill level animation
    CLock lock(g_controller.UserInterface());

    //start the fill level reset animation
    g_controller.UserInterface().Set(UIState_ResetFillLevelAnim);

    //fill level reset without waiting for the animation
    g_controller.Deposit().ResetFillLevel(false);

    //register the fill level reset card first, before the animation
    //is shown, because that takes some time, which in case of low battery
    //might mean that a power-off will happen, the registration should not be
    //lost in that case
    g_controller.Registrations().Register(uid, result, exectype);
    registered = true;

    //wait for the fill level reset animation to finish
    g_controller.UserInterface().WaitStateFinish(UIState_ResetFillLevelAnim);
  }
  else if (exectype == CardType_Service || exectype == CardType_Lock_Test)
  { //service opening
    result = g_controller.Deposit().Process(uid, true, &registered, exectype);
  }
  else if (exectype == CardType_Outbound_Call)
  { //communicate with the server
    g_controller.UserInterface().Set(UIState_ServerComm);
    g_controller.ServerComm().SignalCommAttempt(COMM_REASON_OUTBOUND_CALL_CARD);
  }
  else if (exectype == CardType_Show_GPRS_Signal)
  { //show the service information on the display
    g_controller.UserInterface().Set(UIState_Service);
    DelayMS(10000);
  }
  else
  {
    PrintLog(LOGAUTHLIST, "ERROR: no functionality implemented for card type %s\n",
             CardTypeToStr(cardtype));
    g_controller.UserInterface().SetError(EC_CardFunctionNotImplemented);
  }

  //register the UID of the card, and the state of the deposit opening
  if (!registered)
    g_controller.Registrations().Register(uid, result, cardtype, &m_prepaid);
}

#define CARD_STR(cardstr) (issecuredcard ? "secured " cardstr : cardstr)

/*! Converts a value of ECardType to a printable string.
    \param cardtype A value of ECardType.
    \return A string literal describing the value of ECardType.*/
const char* CAuthList::CardTypeToStr(ECardType cardtype)
{
  bool      issecuredcard = IsSecuredCard(cardtype);
  ECardType unsecuredtype = FromSecuredType(cardtype);

  if (unsecuredtype == CardType_Prepaid)
    return CARD_STR("prepaid");
  else if (unsecuredtype == CardType_Outbound_Call)
    return CARD_STR("outbound call");
  else if (unsecuredtype == CardType_Reset)
    return CARD_STR("reset");
  else if (unsecuredtype == CardType_Service)
    return CARD_STR("service");
  else if (unsecuredtype == CardType_Emptying)
    return CARD_STR("empyting");
  else if (unsecuredtype == CardType_Inbound_Call)
    return CARD_STR("inbound call");
  else if (unsecuredtype == CardType_Registration)
    return CARD_STR("registration");
  else if (unsecuredtype == CardType_Lock_Test)
    return CARD_STR("lock test");
  else if (unsecuredtype == CardType_Show_Current_Load)
    return CARD_STR("show current load");
  else if (unsecuredtype == CardType_Reset_Battery_counter)
    return CARD_STR("reset battery counter");
  else if (unsecuredtype == CardType_Show_GPRS_Signal)
    return CARD_STR("show gprs signal");
  else
    return "ERROR UNKNOWN CARD TYPE";
}

/*! Tests if a card type is a secured type.
    \param cardtype The type of the card on the auth list.
    \return true The card type is a secured type.
    \return false The card type is not a secured type.*/
bool CAuthList::IsSecuredCard(ECardType cardtype)
{
  uint32_t numcardtypes = (uint32_t)CardType_Unsecured_End -
                          (uint32_t)CardType_Unsecured_Start;

  return cardtype >= CardType_Secured_Start &&
         cardtype < CardType_Secured_Start + numcardtypes;
}

/*! Converts a card type in the secured range to the same card type in
    the unsecured range.
    \param cardtype The cardtype to convert.
    \return The converted cardtype.*/
ECardType CAuthList::FromSecuredType(ECardType cardtype)
{
  if (IsSecuredCard(cardtype))
  { //Convert only a secured card type.
    return (ECardType)((uint32_t)cardtype - (uint32_t)CardType_Secured_Start +
                       (uint32_t)CardType_Unsecured_Start);
  }
  else
  { //Don't do a conversion on other card types.
    return cardtype;
  }
}

/*! Converts a value of EOperatingMode to a printable string.
    \param operatingmode A value of EOperatingMode.
    \return A string literal describing the value of EOperatingMode.*/
const char* CAuthList::OperatingModeToStr(EOperatingMode operatingmode)
{
  if (operatingmode == OperatingMode_NoList)
    return "no list";
  else if (operatingmode == OperatingMode_BlackList)
    return "blacklist";
  else if (operatingmode == OperatingMode_WhiteList)
    return "whitelist";
  else
    return "ERROR UNKNOWN OPERATING MODE";
}

//! Maximum size in bytes that this class supports.*/
#define UID_MAX_SUPPORTED_SIZE 7

/*! Converts a UID to a uint64_t for efficient matching to the auth list.
    \param uid The UID to convert.
    \param output The converted UID.
    \return true if the conversion was successful.
    \return false if the UID is too long.*/
bool CAuthList::ConvertUID(NFCUID* uid, uint64_t* output)
{
  //check if the size of the UID is supported
  if (uid->size > UID_MAX_SUPPORTED_SIZE)
  {
    CLogLock loglock;
    PrintLog(LOGAUTHLIST, "ERROR: UID ");
    PrintNFCUID(LOGAUTHLIST, uid);
    PrintLog(LOGAUTHLIST, " of size %" PRIu8 " not supported\n", uid->size);
    return false;
  }

  //convert the UID of the card to uint64_t
  *output = 0;
  for (uint32_t i = 0; i < uid->size; i++)
  {
    *output <<= 8;
    *output |= uid->uid[i];
  }

  return true;
}
