#include "controller.h"
#include "reader.h"
#include "gpiodescriptions.h"
#include "spiflashlog.h"
#include "stm32wss.h"
#include "internalflash.h"
#include "accellointflashaddrs.h"
#include "log.h"
#include "adc.h"
#include "lock.h"
#include "backupregisters.h"
#include "factoryreset.h"
#include "active.h"
// #include <cinttypes> // Edwin was here
/*! I2C address of the LTC4162-F battery charger.*/
#define BATTERY_CHARGER_ADDRESS 0xD0

/*! I2C address of the guileds IO expander.*/
#define GUILEDS_ADDRESS 0x40

CController g_controller;

CController::CController()
{
}

void CController::Setup()
{

    //set up the SPI flash
  m_spiflash.Setup(SPI1, FLASH_CS, FLASH_RESET, FLASH_WP, FLASH_HOLD);

  //Check if a factory reset should be done.
  CFactoryReset::Check();

  //set up the SPI logger task
  g_spiflashlog.Setup(&m_spiflash);

  //print information about this boot, after the SPI flash log has been set up
  PrintBootInfo();
  PrintDeviceActive();

  //Read the option resistors on the pcb, this determines the pcb version.
  m_pcbversion.Read();

  //initialize the settings before any task is set up and started
  m_settings.Init(g_settingsinfo, g_numsettings, SETTINGS_SECTOR_1,
                  SETTINGS_SECTOR_2, LOGSETTINGS);

  SetupSoftwareWatchdog();

  m_i2c1.Setup(I2C1, I2C1_SCL);
  m_i2c2.Setup(I2C2, I2C2_SCL);
  m_spi5.Setup(SPI5, SPI_BaudRatePrescaler_256, SPI_NSS_Soft, SPI_CPOL_High,
               SPI_CPHA_2Edge); // EDWIN zelfde als de Card reader

  CheckExternalRTC();

  //Set up the RTC scheduler, it uses the internal RTC to schedule waking up the
  //microcontroller from low power mode.
  m_rtcscheduler.Setup();

  //set up reading the battery voltage from the charger
  m_batterycharger.Setup(&m_i2c1, BATTERY_CHARGER_ADDRESS);
  m_batterycharger.SetForceTelemetry(true);

  //set up if the charger is powered from a power supply or a battery
  m_batterycharger.SetChargerPowerSource(
      (EChargerPowerSource)m_settings.GetInteger(Setting_ChargerPowerSource));

  //set up reading the inputs from the IO expander
  m_inputs.Setup(&m_i2c1);

  //call the setup method for every task, this is to make sure that any
  //resources are allocated before any task is started
  m_inputmonitor.Setup();
  m_userinterface.Setup();

  m_reader.Setup();
  m_servercomm.Setup();

  //Should be called after setting up the user interface, because it's powered
  //by the same mosfet.
  m_guileds.Setup(&m_i2c2, GUILEDS_ADDRESS, m_userinterface.Orientation()); 

  m_registrations.Setup();
  m_status.Setup();
  m_authlist.Setup();
  m_lowpower.Setup();
  m_deposit.Setup();
  m_deposittimer.Setup();
  m_adc.Setup(ADC1);
  m_6dsensor.Setup(&m_spi5, SENSOR_CS, SENSOR_INT2);

  //Check if the device has had a firmware update, if so, activate the device
  //if not already active.
  ActivateDeviceAfterFwUpdate();

  //read the battery voltage and external voltage,
  //after the charger has been given time to measure it
  m_batterycharger.ReadChargerInfo(&m_batteryinfo, true, 100);

  //Set up self testing, if enabled via the BOOT1 pin.
  m_selftest.Setup();

  //start a task for the user interface
  m_userinterface.Start();

  //Run boot tests, if enabled.
  m_selftest.Run();

  //check if the lock should be opened or closed before reading a card
  m_deposit.CheckLockState();

  //Start a task for monitoring the deposit opening state,
  //and operating the hold lock magnet.
  m_deposittimer.Start();

  //start a task to monitor the hardware inputs
  m_inputmonitor.Start();

  //start a task for reading NFC cards
 // m_reader.Setup(); // edwin
  m_reader.Start();

  //start a task for communication to the server over the modem
  m_servercomm.Start();

  //Start a task to run scheduled alarms based on the RTC time.
  m_rtcscheduler.Start();

  //Start a task that detects using the 6D sensor if the container is being emptied.
 m_6dsensor.Start();

}

/*! Called from the servercomm task, so that it doesn't block the
    FreeRTOS startup task.*/
void CController::PostStartup()
{
  //Setting up the ringbuffer takes a lot of processing time. Use a delay so
  //that reading and processing an NFC card is faster.
  DelayMS(1000);

  //lock the status object, so that after the ringbuffer has been set up,
  //the first thing that is always written is the boot reason
  CLock lock(m_status);
  m_status.SetupRingBuffer();
  g_controller.Status().RegisterBootReason();
  lock.Leave();

  //write the voltage measured at boot as unloaded voltage
  g_controller.Status().Write(Status_InternalBatteryVoltageNoLoad, "%.0f",
                              m_batteryinfo.m_intvoltage * 1000.0f);

  //write the charge current of the internal battery as a status
  if (m_batteryinfo.m_charging)
    g_controller.Status().Write(Status_InternalBatteryChargeCurrent, "%.0f",
                                m_batteryinfo.m_chargecurrent * 1000.0f);

  //write the external voltage as a status
  uint32_t powersource = m_settings.GetInteger(Setting_ChargerPowerSource);
  if (powersource == ChargerPowerSource_Solar_PowerSupply)
    g_controller.Status().Write(Status_SolarVoltage, "%.0f",
                                m_batteryinfo.m_extvoltage * 1000.0f);
  else if (powersource == ChargerPowerSource_Battery ||
           powersource == ChargerPowerSource_Solar_12vBattery)
    g_controller.Status().Write(Status_ExternalBatteryVoltageNoLoad, "%.0f",
                                m_batteryinfo.m_extvoltage * 1000.0f);

  m_registrations.SetupRingBuffer();

  //read the temperature using the ADC and write it as a status
  float temperature = m_adc.ReadTempSensor();
  g_controller.Status().Write(Status_Temperature, "%.1f", temperature);

  //Contact the server if a factory reset happened. The server is contacted
  //always because of the backup domain reset, but this way a comm reason is
  //written as a status, which can later be viewed in TACT NE.
  //WARNING: Because the date is reset to 1-1-2000, the data in TACT NE should
  //be sorted by serverdate to make the contact reason visible.
  if (CFactoryReset::Happened())
    m_servercomm.SignalCommAttempt(COMM_REASON_FACTORY_RESET);
}

/*! Sets the software reboot time to the Setting_MaxAwakeTime setting.
    This will reboot the microcontroller if it's awake for too long.
    During normal use, the microcontroller should go into low power mode
    after it has completed all its task. If it stays awake for too long
    one of its tasks is probably frozen, in that case a reboot is done from
    a FreeRTOS timer.*/
void CController::SetupSoftwareWatchdog()
{
  int64_t maxawake = m_settings.GetInteger(Setting_MaxAwakeTime);
  if (maxawake < LOWEST_MAX_AWAKE_TIME)
    maxawake = LOWEST_MAX_AWAKE_TIME;
  else if (maxawake > HIGHEST_MAX_AWAKE_TIME)
    maxawake = HIGHEST_MAX_AWAKE_TIME;

  PrintLog(LOGCONTROLLER, "Setting maximum awake time to %" PRIi64 "h%02" PRIi64 "m\n",
           maxawake / 60, maxawake % 60);

  SetRebootTime(maxawake * 60 * 1000);
}

/*! Checks if the external RTC should be reset.*/
void CController::CheckExternalRTC()
{
  //Set up the external RTC to a defined state.
  //Reset the external RTC if a power on reset or factory reset happened.
  bool por = ResetFlags() & ResetFlagMask(RCC_FLAG_PORRST);
  m_externalrtc.Setup(&m_i2c1, REBOOT_EN, true, por || CFactoryReset::Happened());

  //Allow the external RTC to power-cycle the microcontroller.
  GPIO_Reset(REBOOT_EN);
}
