#define _GNU_SOURCE //for strptime in C++
#include <ctime>
#include <cmath>

#include "deposit.h"
#include "log.h"
#include "controller.h"
#include "speaker.h"
#include "clock.h"
#include "gpiodescriptions.h"
#include "lock.h"
#include "backupregisters.h"
#include "timeconversion.h"
#include "lock.h"

#include <algorithm>

/*! Time in milliseconds that a beep should sound when the lock is opened.*/
#define DEPOSIT_BEEP_TIME              1000

/*! Maximum debounce time in milliseconds for the deposit input.
    The debounce time is set by the settings, but there should be a sane maximum.*/
#define DEPOSIT_INPUT_DEBOUNCE_MAX     60000

/*! Minimum debounce time in milliseconds for the deposit input.*/
#define DEPOSIT_INPUT_DEBOUNCE_MIN     10

/*! Maximum time that the test for the deposit input can last.*/
#define DEPOSIT_INPUT_TEST_TIMEOUT     1000

/*! Time in milliseconds during the deposit input test before
    its level is considered stable.*/
#define DEPOSIT_INPUT_TEST_STABLE_TIME 100

/*! Maximum time in milliseconds to wait for a deposit to happen
    after the lock has been closed.*/
#define DEPOSIT_TIMEOUT_AFTER_CLOSE    1000

/*! USART that is connected to the fill level sensor.*/
#define SENSOR_USART USART3

CDeposit::CDeposit()
{
  m_restingstate        = false;
  m_startstate          = 0;
  m_readfilllevelsensor = false;
}

void CDeposit::Setup()
{
  m_mutex = xSemaphoreCreateRecursiveMutex();
}

/*! Checks if the lock should be opened or closed based on the OpenStart and OpenStop settings.
    It will open or close the lock if the lock is in the incorrect state.*/
void CDeposit::CheckLockState()
{
  bool wantedstate = WantedLockState();

  //If the lock is not a bistable type, it will always return to the closed
  //state when power is lost, so no further action is needed.
  if (!m_lockdriver.IsBiStable())
  {
    m_lockdriver.ClearOpenedState();
    if (wantedstate)
      PrintLog(LOGDEPOSIT, "ERROR: lock type does not support bistable operation, not opening\n");

    return;
  }

  //If the wanted lock state is different from the current state, open/close the lock.
  if (wantedstate != m_lockdriver.Opened())
  {
    if (wantedstate)
      PrintLog(LOGDEPOSIT, "Lock should be opened, but is closed, opening\n");
    else
      PrintLog(LOGDEPOSIT, "Lock should be closed, but is opened, closing\n");

    //If the lock should be opened, but the battery voltage is too low, don't open the lock.
    //Closing the lock is always allowed.
    if (wantedstate && !CheckBatteryVoltage())
      return;

    //Actuate the lock to get it into the wanted state. It's important that
    //m_lockdriver.PowerUp() is not called before m_lockdriver.Open(), because
    //the lock state is stored in battery backed ram, which should be set before
    //the boost converter is activated. Otherwise a reboot loop happens which
    //can blow the fuse in the battery pack.
    if (wantedstate)
      m_lockdriver.Open();
    else
      m_lockdriver.Close();

    m_lockdriver.PowerDown();
  }
  else
  {
    PrintLog(LOGDEPOSIT, "Lock is %s, which is the correct state\n",
             wantedstate ? "opened" : "closed");
  }
}

/*! Checks the OpenStart and OpenStop settings against the current time.
    \return true if the lock should be opened by default.
    \return false if the lock should be closed by default.*/
bool CDeposit::WantedLockState()
{
  if (!ReadBackupFlag(RTC_VALID_FLAG))
  {
    PrintLog(LOGDEPOSIT, "RTC not set to a valid time, "
             "not checking OpenStart and OpenStop settings\n");
    g_controller.RTCScheduler().DisableAlarm(LOCK_OPEN_ALARM_INDEX);
    return false; //RTC time not valid, lock should be closed.
  }

  //Read and parse the ISO8601 OpenStart and OpenStop settings.
  char openstartstr[DATESTR_MAX_LENGTH];
  char openstopstr[DATESTR_MAX_LENGTH];
  tm openstart;
  tm openstop;
  if (!LockStateSettings(openstartstr, openstopstr, &openstart, &openstop))
  {
    g_controller.RTCScheduler().DisableAlarm(LOCK_OPEN_ALARM_INDEX);
    return false;
  }

  tm now;
  GetRTC(&now);

  //Convert OpenStart, OpenStop and the current time to seconds since 1-1-1970.
  int64_t openstartsecs = TmToSecs(&openstart);
  int64_t openstopsecs  = TmToSecs(&openstop);
  int64_t nowsecs       = TmToSecs(&now);

  //Sanity check
  if (openstartsecs > openstopsecs)
  {
    PrintLog(LOGDEPOSIT, "ERROR: OpenStart \"%s\" is after OpenStop \"%s\"\n",
             openstartstr, openstopstr);
    g_controller.RTCScheduler().DisableAlarm(LOCK_OPEN_ALARM_INDEX);
    return false;
  }

  //If the current time is after OpenStart, and before OpenStop, the lock should
  //be opend by default, so that anyone can do a deposit.
  //If the current time is before OpenStart, or after OpenStop,
  //the lock should be closed by default.
  bool wantedstate = nowsecs >= openstartsecs && nowsecs < openstopsecs;

  char nowstr[DATESTR_MAX_LENGTH];
  ToISO8601(&now, nowstr, sizeof(nowstr), "");
  PrintLog(LOGDEPOSIT, "OpenStart is \"%s\", OpenStop is \"%s\", "
           "current time is \"%s\", lock should be %s\n",
           openstartstr, openstopstr, nowstr, wantedstate ? "opened" : "closed");

  //If the OpenStart time or OpenStop time is in the future, set the alarm
  //to wake up from low power mode at that time, otherwise disable the alarm.
  if (openstartsecs > nowsecs)
    g_controller.RTCScheduler().SetAlarm(LOCK_OPEN_ALARM_INDEX, &openstart);
  else if (openstopsecs > nowsecs)
    g_controller.RTCScheduler().SetAlarm(LOCK_OPEN_ALARM_INDEX, &openstop);
  else
    g_controller.RTCScheduler().DisableAlarm(LOCK_OPEN_ALARM_INDEX);

  return wantedstate;
}

/*! Reads the OpenStart and OpenStop settings, and parses them to struct tm.
    \param openstartstr A char array of size DATESTR_MAX_LENGTH where OpenStart is written to.
    \param openstopstr A char array of size DATESTR_MAX_LENGTH where OpenStop is written to.
    \param openstart The OpenStart setting is parsed and written here.
    \param openstop The OpenStop setting is parsed and written here.
    \return true if OpenStart and OpenStop were written and parsed correctly.
    \return false on any error.*/
bool CDeposit::LockStateSettings(char* openstartstr, char* openstopstr,
                                 tm* openstart, tm* openstop)
{
  memset(openstartstr, 0, DATESTR_MAX_LENGTH);
  memset(openstopstr, 0, DATESTR_MAX_LENGTH);
  memset(openstart, 0, sizeof(tm));
  memset(openstop, 0, sizeof(tm));

  //Read OpenStart from settings.
  g_controller.Settings().GetString(Setting_OpenStart, openstartstr, DATESTR_MAX_LENGTH);
  if (strlen(openstartstr) == 0)
  {
    PrintLog(LOGDEPOSIT, "OpenStart not set\n");
    return false;
  }

  //Read OpenStop from settings.
  g_controller.Settings().GetString(Setting_OpenStop, openstopstr, DATESTR_MAX_LENGTH);
  if (strlen(openstopstr) == 0)
  {
    PrintLog(LOGDEPOSIT, "OpenStop not set\n");
    return false;
  }

  //Convert OpenStart from ISO8601 to a struct tm.
  if (!FromISO8601(openstartstr, openstart))
  {
    PrintLog(LOGDEPOSIT, "ERROR: invalid OpenStart: \"%s\"\n", openstartstr);
    return false;
  }

  //Convert OpenStop from ISO8601 to a struct tm.
  if (!FromISO8601(openstopstr, openstop))
  {
    PrintLog(LOGDEPOSIT, "ERROR: invalid OpenStop: \"%s\"\n", openstopstr);
    return false;
  }

  return true;
}

/* Does a cable test while holding m_mutex.*/
void CDeposit::CableTest()
{
  CLock lock(m_mutex);

  //Testing the cable with the cable circuit attached will connect outputs
  //to inputs, this might trigger a fill level reset and other effects.
  //Lock the Inputs() object to prevent other tasks from reading the inputs.
  CLock inputslock(g_controller.Inputs());

  //Power down the lock if currently powered.
  PowerDown();

  g_controller.SelfTest().RunCableTest();
}

/*! Processes a deposit, should be called after an NFC card is accepted.
    \param uid The UID of the NFC card that was authenticated to do the deposit.
    \param service When true, assume it's a service card, always open and
                   don't increase the fill level.
    \param registered This value is set to true if this function has registered
                      the deposit itself, in that case the function that calls
                      this should not register the deposit.
    \param cardtype The type of card from the auth list.
    \param prepaid Should be a pointer to a CPrepaid instance initialized with
                   authlist counters if cardtype is CardType_Prepaid.*/
EDepositResult CDeposit::Process(NFCUID* uid, bool service, bool* registered,
                                 ECardType cardtype, CPrepaid* prepaid /*= nullptr*/)
{
  CLock lock(m_mutex);

  //If the lock is already opened because of the OpenStart/OpenStop settings,
  //indicate that a deposit can be done, and don't register the deposit.
  if (!service && m_lockdriver.Opened())
  {
    CSpeaker::Beep(DEPOSIT_BEEP_TIME);
    g_controller.UserInterface().Set(UIState_Unlocked);
    DelayMS(DEPOSIT_BEEP_TIME);
    return DepositResult_DontRegister;
  }

  //check if the battery voltage is high enough to open the lock
  if (!CheckBatteryVoltage())
  {
    g_controller.UserInterface().SetError(EC_BatteryVoltageTooLow);
    g_controller.Status().Write(Status_DepositRefused, "battery voltage too low");
    return DepositResult_Refused;
  }

  //check if depositing is currently allowed based on settings
  if (!DepositAllowed(service))
  {
    g_controller.UserInterface().Set(UIState_Locked);
    g_controller.Status().Write(Status_DepositRefused, "container locked by settings");
    return DepositResult_Refused;
  }

  PrintLog(LOGDEPOSIT, "Beginning %s deposit\n", service ? "service" : "regular");

  //check the current fill level, deny access if it's equal or higher to the maximum
  //and the card is not a service card
  if (!CheckFillLevel(service))
  {
    g_controller.Status().Write(Status_DepositRefused, "container full");
    return DepositResult_Refused;
  }

  //inform the user as quickly as possible that the NFC card has been accepted
  if (cardtype != CardType_Prepaid)
    g_controller.UserInterface().Set(UIState_Success);

  //If the deposit is done using a prepaid card, try to charge the deposit on the card.
  if (!ChargePrepaidDeposit(uid, cardtype, prepaid))
    return DepositResult_Refused;

  //enable power to the lock drive circuitry and the deposit switch
  PowerUp();

  //Test if the deposit inputs are at the resting state.
  bool inputerror = false;
  if (!TestDepositInput())
    inputerror = true;

  //Abort the monitoring of the registration switch, and powering the hold lock magnet.
  //This also indicates to the deposit timer that the deposit input is closed.
  g_controller.DepositTimer().BeforeDeposit(inputerror);

  //Power up the fill level sensor if enabled.
  CheckFillLevelSensorSetting();
  PowerUpFillLevelSensor();

  //get the debounce time for the deposit input
  uint32_t debouncetime = DebounceTime();

  //Get the timeout waiting for the registration switch.
  uint32_t deposittimeout = DepositTimeout();

  //Open the lock.
  m_lockdriver.Open();

  //Don't show the unlocked symbol for prepaid cards,
  //because users need enough time to read the value of the card.
  if (cardtype != CardType_Prepaid)
    g_controller.UserInterface().Set(UIState_Unlocked);

  CSpeaker::Beep(DEPOSIT_BEEP_TIME);

  //If the deposit input is at the resting state, wait for the deposit opening to open.
  //Otherwise wait for the timeout.
  uint8_t opened = 0;
  if (!inputerror)
    opened = WaitDepositInput(!m_restingstate, debouncetime, deposittimeout);
  else
    DelayMS(deposittimeout);

  //Indicate to the user that the container has not been opened.
  if (inputerror)
    g_controller.UserInterface().SetError(EC_ContainerAlreadyOpened);
  else if (!opened)
    g_controller.UserInterface().SetError(EC_ContainerNotOpened);

  //Close the lock.
  m_lockdriver.Close();

  //After closing the lock, check the deposit inputs again,
  //in case the deposit opening was opened while closing the lock.
  if (!inputerror)
  {
    debouncetime = std::min(debouncetime, (uint32_t)DEPOSIT_TIMEOUT_AFTER_CLOSE / 2);
    opened |= WaitDepositInput(!m_restingstate, debouncetime, DEPOSIT_TIMEOUT_AFTER_CLOSE);
  }

  //Signal the deposit timer task that a deposit has been done, so that it can
  //start monitoring the registration switch, and release the hold magnet for
  //a hold lock.
  g_controller.DepositTimer().AfterDeposit();

  //Read and parse buffered data from the fill level sensor.
  ReadFillLevelSensor();

  //If the registration inputs are not at the correct resting state at start,
  //record a status.
  if (inputerror)
    RecordDepositInputError();

  if (opened)
  {
    //If the deposit opening was opened, indicate that the deposit has succeeded.
    //Don't show the unlocked symbol for prepaid cards,
    //because users need enough time to read the value of the card.
    if (cardtype != CardType_Prepaid)
      g_controller.UserInterface().Set(UIState_Success);

    //This function registers the deposit itself, the calling function should not.
    *registered = true;

    //Register a deposit for each input that a deposit has been detected on.
    RegisterDeposit(opened, service, uid, cardtype, prepaid);

    DecreaseUnLockedNotOpened();

    PrintLog(LOGDEPOSIT, "Deposit ended\n");
    return DepositResult_Opened;
  }
  else
  {
    IncreaseUnLockedNotOpened();

    PrintLog(LOGDEPOSIT, "Deposit ended\n");
    return DepositResult_Closed;
  }
}

/*! Minimum battery voltage required to open the lock.*/
#define MINIMUM_BATTERY_VOLTAGE 3.1

/*! Checks if the battery voltage is high enough to open the lock.
    \return true if the battery voltage is high enough.
    \return false if the battery voltage is too low.*/
bool CDeposit::CheckBatteryVoltage()
{
  return g_controller.BatteryCharger().CheckBatteryVoltage(MINIMUM_BATTERY_VOLTAGE, LOGDEPOSIT, "opening lock");
}

/*! Checks if a deposit is currently allowed based on time settings.
    \param service When true, indicates that his is a service deposit,
                   which is always allowed.
    \return true if a deposit is currently allowed.
    \return false if a deposit is currently not allowed.*/
bool CDeposit::DepositAllowed(bool service)
{
  if (service)
    return true; //a service deposit is always allowed

  //don't check any of the time settings if the RTC's time has not been set
  if (!ReadBackupFlag(RTC_VALID_FLAG))
  {
    PrintLog(LOGDEPOSIT, "RTC not set to a valid time, "
             "not checking time settings\n");
    return true;
  }

  tm now;
  GetRTC(&now);

  //check if the TimeOpen and TimeClose settings allow a deposit
  if (!TimeOpenCloseAllowed(now))
    return false;

  //check if the BlockStart and BlockStop settings allow a deposit
  if (!BlockStartStopAllowed(now))
    return false;

  return true;
}

/*! Checks if a deposit is allowed based on the TimeOpen and TimeClose settings,
    and the current RTC time.
    \param now The current time returned by the RTC.
    \return true if a deposit is allowed.
    \return false if a deposit is denied.*/
bool CDeposit::TimeOpenCloseAllowed(tm now)
{
  char timeopenstr[16];
  g_controller.Settings().GetString(Setting_TimeOpen, timeopenstr, sizeof(timeopenstr));
  if (strlen(timeopenstr) == 0)
  {
    PrintLog(LOGDEPOSIT, "TimeOpen not set, allowing deposit\n");
    return true; //allow deposit when timeopen is not set
  }

  char timeclosestr[16];
  g_controller.Settings().GetString(Setting_TimeClose, timeclosestr, sizeof(timeclosestr));
  if (strlen(timeclosestr) == 0)
  {
    PrintLog(LOGDEPOSIT, "TimeClose not set, allowing deposit\n");
    return true; //allow deposit when timeclose is not set
  }

  //convert the TimeOpen setting to a struct tm
  tm timeopen = {};
  if (strptime(timeopenstr, "%R", &timeopen) == nullptr)
  {
    PrintLog(LOGDEPOSIT, "ERROR: Invalid TimeOpen: \"%s\", allowing deposit\n", timeopenstr);
    return true; //allow deposit when the setting is invalid
  }

  //convert the TimeClose setting to a struct tm
  tm timeclose = {};
  if (strptime(timeclosestr, "%R", &timeclose) == nullptr)
  {
    PrintLog(LOGDEPOSIT, "ERROR: Invalid TimeClose: \"%s\", allowing deposit\n", timeclosestr);
    return true; //allow deposit when the setting is invalid
  }

  //convert the TimeOpen, TimeClose and current times from struct tm to minutes since 00:00
  uint32_t openmin  = timeopen.tm_hour  * 60 + timeopen.tm_min;
  uint32_t closemin = timeclose.tm_hour * 60 + timeclose.tm_min;
  uint32_t nowmin   = now.tm_hour       * 60 + now.tm_min;

  if (openmin == closemin)
  {
    PrintLog(LOGDEPOSIT, "TimeOpen and TimeClose both set to %02i:%02i, "
             "allowing deposit\n", timeopen.tm_hour, timeopen.tm_min);
    return true;
  }

  //check if nowmin falls in the openmin-closemin window, taking into
  //account the wrap that occurs at 00:00 hour
  if ((openmin < closemin && (nowmin >= openmin && nowmin <= closemin)) ||
      (openmin > closemin && (nowmin >= openmin || nowmin <= closemin)))
  {
    PrintLog(LOGDEPOSIT, "TimeOpen set to %02i:%02i, TimeClose set to "
             "%02i:%02i, current time is %02i:%02i, deposit allowed\n",
             timeopen.tm_hour, timeopen.tm_min,
             timeclose.tm_hour, timeclose.tm_min,
             now.tm_hour, now.tm_min);

    return true;
  }

  PrintLog(LOGDEPOSIT, "TimeOpen set to %02i:%02i, TimeClose set to "
           "%02i:%02i, current time is %02i:%02i, deposit denied\n",
           timeopen.tm_hour, timeopen.tm_min,
           timeclose.tm_hour, timeclose.tm_min,
           now.tm_hour, now.tm_min);

  return false;
}

/*! Checks if deposits should be blocked based on the BlockStart and BlockStop
    settings. Together they define a period where deposits are not allowed.
    \return true if deposits are allowed.
    \return false if deposits are blocked.*/
bool CDeposit::BlockStartStopAllowed(tm now)
{
  char blockstartstr[DATESTR_MAX_LENGTH];
  g_controller.Settings().GetString(Setting_BlockStart, blockstartstr, sizeof(blockstartstr));
  if (strlen(blockstartstr) == 0)
  {
    PrintLog(LOGDEPOSIT, "BlockStart not set, allowing deposit\n");
    return true;
  }

  char blockstopstr[DATESTR_MAX_LENGTH];
  g_controller.Settings().GetString(Setting_BlockStop, blockstopstr, sizeof(blockstopstr));
  if (strlen(blockstopstr) == 0)
  {
    PrintLog(LOGDEPOSIT, "BlockStop not set, allowing deposit\n");
    return true;
  }

  //convert the BlockStart setting to a struct tm
  tm blockstart;
  if (!FromISO8601(blockstartstr, &blockstart))
  {
    PrintLog(LOGDEPOSIT, "ERROR: Invalid BlockStart: \"%s\", allowing deposit\n",
             blockstartstr);
    return true;
  }

  //convert the BlockStop setting to a struct tm
  tm blockstop;
  if (!FromISO8601(blockstopstr, &blockstop))
  {
    PrintLog(LOGDEPOSIT, "ERROR: Invalid BlockStop: \"%s\", allowing deposit\n",
             blockstopstr);
    return true;
  }

  //convert blockstart, blockstop and the current time to seconds since 1-1-1970
  int64_t blockstartsecs = TmToSecs(&blockstart);
  int64_t blockstopsecs  = TmToSecs(&blockstop);
  int64_t nowsecs        = TmToSecs(&now);

  if (blockstartsecs > blockstopsecs)
  {
    PrintLog(LOGDEPOSIT, "ERROR: BlockStart \"%s\" is after BlockStop \"%s\", "
             "allowing deposit\n", blockstartstr, blockstopstr);
    return true;
  }

  char nowstr[DATESTR_MAX_LENGTH];
  ToISO8601(&now, nowstr, sizeof(nowstr), "");

  //check if the current time falls within the block period
  if (nowsecs >= blockstartsecs && nowsecs <= blockstopsecs)
  {
    PrintLog(LOGDEPOSIT, "BlockStart is \"%s\", BlockStop is \"%s\", "
             "current time is \"%s\", deposit denied\n",
             blockstartstr, blockstopstr, nowstr);
    return false;
  }

  PrintLog(LOGDEPOSIT, "BlockStart is \"%s\", BlockStop is \"%s\", "
           "current time is \"%s\", deposit allowed\n",
           blockstartstr, blockstopstr, nowstr);

  return true;
}

/*! Checks if the card is a prepaid type, if so, it will charge the deposit on the prepaid card.
    \param uid The UID of the card that is doing the deposit.
    \param cardtype The type of the card on the authlist that is doing the deposit.
    \param prepaid Pointer to a CPrepaid instance that holds the authlist counters.
    \return true If the card is not a prepaid card.
    \return true If the card is a prepaid card and the deposit was charged on the card.
    \return false If the card is a prepaid card but doesn't have enough credit for a deposit.
    \retrn false If the card is a prepaid card but the deposit could not be charged on it.*/
bool CDeposit::ChargePrepaidDeposit(NFCUID* uid, ECardType cardtype, CPrepaid* prepaid)
{
  if (cardtype != CardType_Prepaid)
  {
    PrintLog(LOGDEPOSIT, "Not checking prepaid credit for card type \"%s\"\n",
             CAuthList::CardTypeToStr(cardtype));
    return true; //No prepaid processing needed.
  }

  //Prepaid deposit is only supported for single opening deposit types.
  if (DepositInput() != INPUT_DEPOSIT1)
  {
    g_controller.UserInterface().SetError(EC_WrongLockType);
    PrintLog(LOGDEPOSIT, "ERROR: prepaid deposit not supported for lock type %s\n",
             m_lockdriver.LockTypeStr());
    return false;
  }

  assert(prepaid); //Pointer should always be valid for CardType_Prepaid.

  //Try to charge the deposit on the prepaid card.
  prepaid->SetCost(g_controller.Settings().GetInteger(Setting_CostInput1));
  EPrepaidResult result = g_controller.Reader().ChargePrepaidDeposit(uid, prepaid);
  if (result != EPrepaidResult_Success)
  {
    if (result == EPrepaidResult_InsufficientCredit)
    {
      g_controller.UserInterface().Set(UIState_InsufficientCredit);
      g_controller.Status().Write(Status_DepositRefused, "insufficient prepaid credit");
      //In case of insufficient prepaid credit, contact the server,
      //because the user might have bought extra credit.
      g_controller.ServerComm().SignalCommAttempt(COMM_REASON_INSUFFICIENT_PREPAID_CREDIT);
    }
    else if (result == EPrepaidResult_CardFailure)
    {
      g_controller.UserInterface().Set(UIState_CardError);
      g_controller.Status().Write(Status_DepositRefused, "prepaid card defective");
    }
    else if (result == EPrepaidResult_ReaderError)
    {
      g_controller.UserInterface().Set(UIState_CardError);
      g_controller.Status().Write(Status_DepositRefused, "prepaid NFC error");
    }
    else if (result == EPrepaidResult_CardNotAuthenticated)
    {
      g_controller.UserInterface().Set(UIState_CardError);
      g_controller.Status().Write(Status_DepositRefused, "prepaid card not authenticated");
    }
    else
    {
      PrintLog(LOGDEPOSIT, "ERROR: unhandled prepaid error %i\n", (int)result);
      g_controller.UserInterface().SetError(EC_None);
      g_controller.Status().Write(Status_DepositRefused,
                                  "unhandled prepaid error %i",
                                  (int)result);
    }

    CLogLock loglock;
    PrintLog(LOGDEPOSIT, "ERROR: could not charge deposit of %" PRIu32
             " credits on prepaid card ", prepaid->GetCost());
    PrintNFCUID(LOGDEPOSIT, uid, "\n");
    loglock.Leave();

    return false;
  }

  g_controller.UserInterface().SetPrepaidInfo(prepaid);

  //If the credit on the prepaid card is below a level needed for a set number of deposits,
  //contact the server, because the user might have bought extra credit.
  int32_t warninglevel = g_controller.Settings().GetInteger(Setting_CreditWarningLevel);
  if (warninglevel > 0 && prepaid->ResultValue() < (prepaid->GetCost() * warninglevel))
  {
    //Indicate to the user that the prepaid deposit has been charged successfully,
    //but that the credit the users has is below the warning limit.
    g_controller.UserInterface().Set(UIState_SuccessPrepaidLowCredit);

    //Contact the server in case the user has bought extra credit.
    g_controller.ServerComm().SignalCommAttempt(COMM_REASON_LOW_PREPAID_CREDIT);
  }
  else
  {
    //Indicate to the user that the prepaid deposit has been charged successfully.
    g_controller.UserInterface().Set(UIState_SuccessPrepaid);
  }


  CLogLock loglock;
  PrintLog(LOGDEPOSIT, "Charged deposit of %" PRIu32 
           " credits on prepaid card ", prepaid->GetCost());
  PrintNFCUID(LOGDEPOSIT, uid, "\n");
  loglock.Leave();

  return true;
}

/*! Gets the debounce time for the deposit input from the settings,
    and clamps them to sane limits.
    \return The debounce time in milliseconds for the deposit input.*/
uint32_t CDeposit::DebounceTime()
{
  int64_t debouncetime = g_controller.Settings().GetInteger(Setting_InterruptDelay);
  if (debouncetime > DEPOSIT_INPUT_DEBOUNCE_MAX)
    debouncetime = DEPOSIT_INPUT_DEBOUNCE_MAX;
  else if (debouncetime < DEPOSIT_INPUT_DEBOUNCE_MIN)
    debouncetime = DEPOSIT_INPUT_DEBOUNCE_MIN;

  return debouncetime;
}

/*! Gets the deposit timeout from the settings.
    \return Timeout in milliseconds waiting for the registration switch.*/
uint32_t CDeposit::DepositTimeout()
{
  int64_t deposittimeout = g_controller.Settings().GetInteger(Setting_DelayLock);
  if (deposittimeout == 0)
    deposittimeout = DEPOSIT_TIMEOUT_DEFAULT;
  else if (deposittimeout > DEPOSIT_TIMEOUT_MAX)
    deposittimeout = DEPOSIT_TIMEOUT_MAX;
  else if (deposittimeout < DEPOSIT_TIMEOUT_MIN)
    deposittimeout = DEPOSIT_TIMEOUT_MIN;

  return deposittimeout;
}

/*  Checks if the fill level stays below the set maximum if a deposit is done.
    \param service When true, the card that authenticated is a service card.
    \return true if access should be granted based on fill level.
    \return false if access should be denied based on fill level.*/
bool CDeposit::CheckFillLevel(bool service)
{
  float   filllevel  = g_controller.Settings().GetFloat(RunStatus_FillLevel);
  int32_t countermax = g_controller.Settings().GetInteger(Setting_CounterMax);

  if (countermax == 0)
  { //If no maximum is set, a deposit is always allowed.
    PrintLog(LOGDEPOSIT, "Fill level is %g, no maximum set\n",
             filllevel);
    return true;
  }
  else if (service)
  { //A service deposit is always allowed even if the container is full.
    PrintLog(LOGDEPOSIT, "Fill level is %g, max is %" PRIu32
             ", service deposit is always allowed\n",
             filllevel, countermax);
    return true;
  }

  //For each enabled input, check if the fill level would go beyond the
  //maximum if a deposit is done. If so, deny the deposit.
  uint32_t inputs = DepositInput();
  for (uint32_t i = 0; i < NUM_DEPOSIT_INPUTS; i++)
  {
    uint8_t input = DEPOSIT_INPUT(i);
    if (input & inputs)
    {
      //Calculate the new fill level.
      float increase = FillLevelIncrease(i);
      float newfilllevel = filllevel + increase;

      //Round to nearest integer then compare,
      //this way the container can be filled exactly.
      if (lroundf(newfilllevel) > countermax)
      {
        PrintLog(LOGDEPOSIT, "Fill level is %g, would increase to %g using input"
                 " %s, max is %" PRIu32 ", denying access\n",
                 filllevel, newfilllevel, CInputs::InputStr(input), countermax);
        g_controller.UserInterface().Set(UIState_Full);
        return false;
      }
    }
  }

  PrintLog(LOGDEPOSIT, "Fill level is %g, max is %" PRIu32 "\n",
           filllevel, countermax);

  return true;
}

/*! Gets from the settings how much the fill level should increase for a deposit input.
    \param input 0 for input1, 1 for input2.
    \return The fill level increase for the corresponding input.*/
float CDeposit::FillLevelIncrease(uint32_t input)
{
  assert(input == 0 || input == 1);

  EStorageKey setting;
  float       defval;
  if (input == 0)
  {
    setting = Setting_CounterInput1;
    defval  = INPUT1_DEFAULT_FILL_INC;
  }
  else
  {
    setting = Setting_CounterInput2;
    defval  = INPUT2_DEFAULT_FILL_INC;
  }

  float increase = g_controller.Settings().GetFloat(setting);

  //Return the setting if it's set to a valid value, otherwise return the default.
  if (std::isfinite(increase) && increase >= 0.0f)
  {
    return increase;
  }
  else
  {
    PrintLog(LOGDEPOSIT, "ERROR: setting %s has invalid value %g, using default %g\n", 
             g_controller.Settings().Name(setting), increase, defval);
    return defval;
  }
}

/*! Increases the fill level if not a service deposit, and contacts the server if
    the warning fill level or the maximum fill level is reached.
    \param service When true, this is a service opening, don't increase fill level.
    \param input The input that the deposit was detected on, 0 for input1, 1 for input2.*/
void CDeposit::IncreaseFillLevel(bool service, uint32_t input)
{
  if (service)
    return; //A service opening doesn't increase the fill level.

  //Increase the fill level as indicated by the setting for this input.
  float filllevel    = g_controller.Settings().GetFloat(RunStatus_FillLevel);
  float increase     = FillLevelIncrease(input);
  float newfilllevel = filllevel + increase;
  g_controller.Settings().SetFloat(RunStatus_FillLevel, newfilllevel);

  //Check if the fill level has risen above the warning level, if so, contact the server.
  int32_t counterwarning = g_controller.Settings().GetInteger(Setting_CounterWarning);
  if (counterwarning > 0 && lroundf(filllevel) < counterwarning && lroundf(newfilllevel) >= counterwarning)
  {
    PrintLog(LOGDEPOSIT, "Fill level increased from %g to %g, warning level"
             " is %" PRIi32 ", contacting server\n", filllevel, newfilllevel, counterwarning);
    g_controller.ServerComm().SignalCommAttempt(COMM_REASON_WARNING_FILL_LEVEL_REACHED);
  }

  //Check if the fill level has risen above the maximum level, if so, contact the server.
  int32_t countermax = g_controller.Settings().GetInteger(Setting_CounterMax);
  if (countermax > 0 && lroundf(filllevel) < countermax && lroundf(newfilllevel) >= countermax)
  {
    PrintLog(LOGDEPOSIT, "Fill level increased from %g to %g, max is %"
             PRIi32 ", contacting server\n", filllevel, newfilllevel, countermax);
    g_controller.ServerComm().SignalCommAttempt(COMM_REASON_MAX_FILL_LEVEL_REACHED);
  }
}

/*! Called when the deposit opening was unlocked, but not opened.
    Increases the RunStatus_UnlockedNotOpened counter, contacts the server
    when it reaches the Setting_BreakCounter warning level.*/
void CDeposit::IncreaseUnLockedNotOpened()
{
  PrintLog(LOGDEPOSIT, "ERROR: unlocked deposit opening but opening was not opened\n");

  uint32_t max = g_controller.Settings().GetInteger(Setting_BreakCounter);
  if (max > 0)
  {
    uint32_t notopened = g_controller.Settings().GetInteger(RunStatus_UnlockedNotOpened);
    notopened++;

    if (notopened >= max)
    { //record the number of times the deposit opening was not opened, and contact the server
      g_controller.Status().Write(Status_DepositOpeningBlocked, "%" PRIi32, notopened);
      g_controller.ServerComm().SignalCommAttempt(COMM_REASON_DEPOSIT_OPENING_BLOCKED);
      g_controller.Settings().SetInteger(RunStatus_UnlockedNotOpened, 0);
    }
    else
    {
      g_controller.Settings().SetInteger(RunStatus_UnlockedNotOpened, notopened);
    }
  }
}

/*! Called when the deposit opening was unlocked and opened.
    Decreases the notopened level.*/
void CDeposit::DecreaseUnLockedNotOpened()
{
  uint32_t notopened = g_controller.Settings().GetInteger(RunStatus_UnlockedNotOpened);
  if (notopened > 0)
  {
    notopened--;
    g_controller.Settings().SetInteger(RunStatus_UnlockedNotOpened, notopened);
  }
}

/*! Called from the CDepositTimer task, checks the state of the deposit input.
    \param releasehold Release the hold lock if the deposit input is opened.
    \param releasealways Release even if the deposit input is closed.
    \return true The deposit input is opened.
    \return false The deposit input is closed.*/
bool CDeposit::CheckDepositInputFromTimer(bool releasehold, bool releasealways)
{
  CLock lock(m_mutex);

  //Get the deposit inputs currently in use.
  uint32_t inputs = DepositInput();

  //Power up the boost converter and 12V_SW output to power the registration switch.
  PowerUp();

  //Read the registration switch.
  uint8_t values;
  bool stable = g_controller.Inputs().ReadStable(inputs, DEPOSIT_INPUT_TEST_STABLE_TIME,
                                                 DEPOSIT_INPUT_TEST_TIMEOUT, &values);

  //Read from the setting if the deposit input should be high or low when closed.
  ReadRegSwitchtypeSetting();

  //Check the values of the inputs, if any of the used inputs is not at the resting
  //state, the deposit opening corresponding to that input is still opened.
  bool opened = false;
  for (uint32_t i = 0; i < NUM_DEPOSIT_INPUTS; i++)
  {
    uint8_t input = DEPOSIT_INPUT(i);
    if ((input & inputs)) //Check if the deposit input is currently used.
    {
      if ((bool)(values & input) != m_restingstate) //Check if opened.
      {
        PrintLog(LOGDEPOSIT, "Deposit opening %s is still opened\n",
                 CInputs::InputStr(input));
        opened = true;
      }
      else
      {
        PrintLog(LOGDEPOSIT, "Deposit opening %s is closed\n",
                 CInputs::InputStr(input));
      }
    }
  }

  //Release the hold lock if needed.
  if (releasehold && ((stable && opened) || releasealways))
    m_lockdriver.ReleaseHold();

  PowerDown();

  return stable && opened;
}

/*! Disables power to the lock drive circuity and the deposit switch.*/
void CDeposit::PowerDown()
{
  CLock lock(m_mutex);

  //disable the H bridges and boost converter
  m_lockdriver.PowerDown();

  //disable power to the deposit switch
  GPIO_Reset(P12V_SW);
}

/*! This function can be called if only the fill level sensor should be
    powered and read, without powering the lock.
    This function is normally called from a different task than CDeposit::Process()*/
void CDeposit::RegisterFillLevelSensor()
{
  CLock lock(m_mutex);

  CheckFillLevelSensorSetting();
  if (!m_readfilllevelsensor)
    return;

  //Power down the lock if currently powered.
  PowerDown();

  //Enable the 5V boost converter.
  m_lockdriver.PowerUpBoostConverter();

  //Read data from the fill level sensor.
  PowerUpFillLevelSensor();
  ReadFillLevelSensor();

  //Power down the 5V boost converter.
  m_lockdriver.PowerDownBoostConverter();
}

/*! Sets m_readfilllevelsensor based on Setting_ReadFillLevelSensor.*/
void CDeposit::CheckFillLevelSensorSetting()
{
  m_readfilllevelsensor = g_controller.Settings().GetBoolean(Setting_ReadFillLevelSensor);

  if (!m_readfilllevelsensor)
    PrintLog(LOGDEPOSIT, "Fill level sensor not enabled\n");
}

/* If enabled through the setting, powers up the fill level sensor
   and sets up the usart to read data from it.
   The fill level sensor will continuously measure distance and send its measurements.
   The usart rx queue will buffer the data, which can be read back later.*/
void CDeposit::PowerUpFillLevelSensor()
{
  if (!m_readfilllevelsensor)
    return;

  PrintLog(LOGDEPOSIT, "Powering up fill level sensor\n");

  //Enable the RS485 transceiver in receive mode.
  GPIO_Set(RS485_PWR_ON);
  GPIO_Reset(RS485_RX_EN);

  //Enable the OUT1 power output, this powers the ultrasonic sensor.
  GPIO_Set(P12V_OUT1);

  //Set up the usart to read data from the fill level sensor,
  //this will be buffered in the allocated rx queue.
  m_filllevelsensor.Setup(SENSOR_USART);

  return;
}

/*! Powers down the fill level sensor and closes the usart.*/
void CDeposit::PowerDownFillLevelSensor()
{
  if (!m_readfilllevelsensor)
    return;

  GPIO_Reset(RS485_PWR_ON);
  GPIO_Reset(P12V_OUT1);
  m_filllevelsensor.Close();
  m_readfilllevelsensor = false;
}

/*! Maximum time in milliseconds waiting for a valid
    measurement from the fill level sensor.*/
#define FILL_LEVEL_SENSOR_TIMEOUT 5000

/*! Reads the data from the fill level sensor,
    which has been buffered by the usart rx queue.*/
void CDeposit::ReadFillLevelSensor()
{
  if (!m_readfilllevelsensor)
    return;

  PrintLog(LOGDEPOSIT, "Reading data from fill level sensor\n");

  //The first few measurements from the fill level sensor are 0.
  //Keep reading until a nonzero measurement is returned.
  bool    success = false;
  float   distance = 0.0f;
  int64_t start = Millis();
  do
  {
    if (m_filllevelsensor.Read(100))
    {
      success = true;
      distance = m_filllevelsensor.Distance();

      if (distance != 0.0f)
        break;
    }
  }
  while (Millis() - start < FILL_LEVEL_SENSOR_TIMEOUT);

  PowerDownFillLevelSensor();

  if (success)
  {
    g_controller.Status().Write(Status_FillLevelSensor, "%.3f", distance);
  }
  else
  {
    PrintLog(LOGDEPOSIT, "ERROR: did not receive a measurement from the fill level sensor\n");
    g_controller.Status().Write(Status_FillLevelSensor, "Not available");
  }
}

/*! Enables power to the deposit switch and the lock drive circuitry.*/
void CDeposit::PowerUp()
{
  //enable the 12V_SW output to power the deposit switch
  GPIO_Set(P12V_SW);

  //enable the boost converter to power the lock and the deposit switch
  m_lockdriver.PowerUp();
}

/*! Checks the lock type, and decides which inputs should be used for detecting a deposit.
    \return \ref INPUT_DEPOSIT1
    \return a bitwise OR of \ref INPUT_DEPOSIT1 and INPUT_DEPOSIT2 */
uint32_t CDeposit::DepositInput()
{
  if (m_lockdriver.LockType() == LockType_30_60_Liter)
    return INPUT_DEPOSIT1 | INPUT_DEPOSIT2;
  else
    return INPUT_DEPOSIT1;
}

/*! The deposit input has a pull down resistor, the registration switch connects
    a voltage to the deposit input when closed. The resting state is low for a
    normally open switch, and high for a normally closed switch.*/
void CDeposit::ReadRegSwitchtypeSetting()
{
  m_restingstate = g_controller.Settings().GetBoolean(Setting_RegSwitchNormallyClosed);
}

/*! Tests the deposit inputs.
    \return true if the deposit inputs are at its resting state.
    \return false if one of the deposit inputs is not at its resting state, or is unstable.*/
bool CDeposit::TestDepositInput()
{
  ReadRegSwitchtypeSetting();

  //Get the deposit inputs currently in use.
  uint32_t inputs = DepositInput();
  
  //Print what is about to be measured for each selected input.
  for (uint32_t i = 0; i < NUM_DEPOSIT_INPUTS; i++)
  {
    uint8_t input = DEPOSIT_INPUT(i);
    if (input & inputs)
      PrintLog(LOGDEPOSIT, "Testing state of %s switch on deposit input %s\n",
               m_restingstate ? "normally closed" : "normally open", CInputs::InputStr(input));
  }

  //Test that the deposit inputs are at its resting state.
  if (!g_controller.Inputs().WaitForStableLevel(inputs, DEPOSIT_INPUT_TEST_STABLE_TIME,
                                                DEPOSIT_INPUT_TEST_TIMEOUT,
                                                m_restingstate ? inputs : 0,
                                                &m_startstate))
  {
    //Print an error for each deposit input that is not at the resting state.
    for (uint32_t i = 0; i < NUM_DEPOSIT_INPUTS; i++)
    {
      uint8_t input = DEPOSIT_INPUT(i);
      if ((input & inputs) && ((bool)(m_startstate & input) != m_restingstate))
        PrintLog(LOGDEPOSIT, "ERROR: timeout detecting %s level on deposit input %s\n",
                 CInputs::StateStr(m_restingstate), CInputs::InputStr(input));
    }

    //Indicate that the registration input is not at the correct level.
    return false;
  }

  //Indicate that each deposit input is at the resting state.
  for (uint32_t i = 0; i < NUM_DEPOSIT_INPUTS; i++)
  {
    uint8_t input = DEPOSIT_INPUT(i);
    if (input & inputs)
      PrintLog(LOGDEPOSIT, "Deposit input %s is %s\n",
              CInputs::InputStr(input), CInputs::StateStr(m_startstate & input));
  }

  return true;
}

/*! Waits for one of the deposit inputs to become a certain level,
    and stay at this level for some time.
    \param wantedlevel return true if one of the deposit inputs is at this level.
    \param debouncetime return true if one of the deposit input is at wantedlevel for
                        this time in milliseconds.          
    \param timeout Time in milliseconds in which the deposit input should change.
    \return A bitmask of deposit inputs that have reached wantedlevel
            in debouncetime milliseconds.*/
uint8_t CDeposit::WaitDepositInput(bool wantedlevel, uint32_t debouncetime,
                                   uint32_t timeout)
{
  //Get the deposit inputs currently in use.
  uint32_t inputs = DepositInput();
  uint8_t  wantedinputlevels = 0;

  //Print which inputs are going to be read.
  for (uint32_t i = 0; i < NUM_DEPOSIT_INPUTS; i++)
  {
    uint8_t input = DEPOSIT_INPUT(i);
    if (input & inputs)
    {
      PrintLog(LOGDEPOSIT, "Waiting for deposit input %s to become %s for %" PRIu32
               " milliseconds, timeout %" PRIu32 " milliseconds\n", 
               CInputs::InputStr(input), CInputs::StateStr(wantedlevel),
               debouncetime, timeout);

      //Create a bitmask of the signal levels that should generate a trigger.
      if (wantedlevel)
        wantedinputlevels |= input;
    }
  }

  //Wait for one of the selected inputs to stay at the wanted level for debouncetime milliseconds.
  uint8_t readlevel;
  if (!g_controller.Inputs().WaitForStableSignal(inputs, debouncetime, timeout, wantedinputlevels, &readlevel))
  {
    for (uint32_t i = 0; i < NUM_DEPOSIT_INPUTS; i++)
    {
      uint8_t input = DEPOSIT_INPUT(i);
      if (input & inputs)
        PrintLog(LOGDEPOSIT, "ERROR: timeout detecting %s level on deposit input %s\n",
                 CInputs::StateStr(wantedlevel), CInputs::InputStr(input));
    }
    return 0;
  }

  //Indicate the state each input is in.
  for (uint32_t i = 0; i < NUM_DEPOSIT_INPUTS; i++)
  {
    uint8_t input = DEPOSIT_INPUT(i);
    if (input & inputs)
      PrintLog(LOGDEPOSIT, "Deposit input %s is %s\n",
              CInputs::InputStr(input), CInputs::StateStr(readlevel & input));
  }

  //Turn readlevel into a bitmask of inputs that are at wantedlevel.
  if (!wantedlevel)
    readlevel = ~readlevel & inputs;

  //Return a bitmask of inputs that are at wantedlevel.
  return readlevel;
}

/*! Records a Status_RegistrationSwitchError with the input state recorded
    in m_startstate.*/
void CDeposit::RecordDepositInputError()
{
  char inputstr[32] = {};

  //For each enabled deposit input, write the state as recorded in m_startstate.
  uint32_t inputs = DepositInput();
  bool     first  = true;
  for (uint32_t i = 0; i < NUM_DEPOSIT_INPUTS; i++)
  {
    uint8_t input = DEPOSIT_INPUT(i);
    if (input & inputs)
    {
      bool     state       = m_startstate & input;
      uint32_t inputstrlen = strlen(inputstr);
      char*    inputptr    = inputstr + inputstrlen;

      snprintf(inputptr, sizeof(inputstr) - inputstrlen, "%sInput%" PRIu32 ":%s",
               first ? "" : " ", i + 1, state ? "high" : "low");

      first = false;
    }
  }

  g_controller.Status().Write(Status_RegistrationSwitchError, "%s", inputstr);
}

/*! Sets the fill level to 0, and shows the user
    that the fill level has been reset.*/
void CDeposit::ResetFillLevel(bool anim /*= true*/)
{
  CLock lock(g_controller.UserInterface(), false);

  if (anim)
  {
    //lock the user interface so that no other task interrupts the fill level animation
    lock.Enter();

    //start the fill level reset animation before writing a status,
    //because writing the status might take a few seconds
    g_controller.UserInterface().Set(UIState_ResetFillLevelAnim);
  }

  //record the fill level when emptied
  uint32_t filllevel = lroundf(g_controller.Settings().GetFloat(RunStatus_FillLevel));
  g_controller.Status().Write(Status_Emptied, "%" PRIu32, filllevel);

  //set the fill level to 0
  g_controller.Settings().SetFloat(RunStatus_FillLevel, 0.0f);

  //communicate with the server that the fill level has been reset
  g_controller.ServerComm().SignalCommAttempt(COMM_REASON_EMPTIED);

  //wait for the fill level reset animation to finish
  if (anim)
    g_controller.UserInterface().WaitStateFinish(UIState_ResetFillLevelAnim);
}

/*! Registers a deposit for each input on which a deposit was detected.
    \param opened Bitmask of inputs on which a deposit was detected.
    \param service Set to true if this is a service deposit.
    \param uid The UID of the card that has done the deposit.
    \param cardtype The type of the card that has done the deposit.
    \param prepaid Pointer to a CPrepaid object that has
                   handled the prepaid transaction.*/
void CDeposit::RegisterDeposit(uint8_t opened, bool service, NFCUID* uid,
                               ECardType cardtype, CPrepaid* prepaid)
{
  //For each input that detected a deposit, increase the fill level and register
  //a deposit.
  for (uint32_t i = 0; i < NUM_DEPOSIT_INPUTS; i++)
  {
    uint8_t input = DEPOSIT_INPUT(i);
    if (opened & input)
    {
      PrintLog(LOGDEPOSIT, "Registered deposit on input %s\n",
               CInputs::InputStr(input));

      IncreaseFillLevel(service, i);
      g_controller.Registrations().Register(uid, DepositResult_Opened, cardtype,
                                            prepaid, i);
    }
  }
}
