#include "deposittimer.h"
#include "log.h"
#include "clock.h"
#include "backupregisters.h"
#include "lock.h"
#include "controller.h"

#include <algorithm>

#define MIN_INITIAL_RTC_DELAY          1    //!< Minimum initial delay in seconds.
#define MAX_INITIAL_RTC_DELAY          300  //!< Maximum initial delay in seconds.
#define MAX_RTC_DELAY                  (6 * 60 * 60) //!< Max RTC delay in seconds.
#define MAX_RTC_DELAY_MULTIPLIER_SHIFT 17 //!< Max 1 << count bitshift.

/*! Maximum number of times the permanent electromagnet is powered.*/
#define MAX_HOLDLOCK_COUNT             3

/*! Number of RTC wakeup events since the last deposit.*/
#define WAKEUP_COUNT_REGISTER          RTC_BKP_DR16

/*! Total number of seconds passed since the last deposit.*/
#define WAKEUP_SECONDS_PASSED_REGISTER RTC_BKP_DR13

/*! Status_DepositOpeningStayedOpen written with value 1.*/
#define STAYED_OPEN_FLAG               0x1000000

/*! Status_DepositOpeningStayedOpen written with value 0.*/
#define STAYED_CLOSED_FLAG             0x800000

CDepositTimer::CDepositTimer()
{
  m_signal = 0;
  m_cmd    = EDepositTimerCmd_None;
}

void CDepositTimer::Setup()
{
  //Create a binary semaphore to signal from the RTC wakeup interrupt
  //that the wakeup timer expired.
  m_signal = xSemaphoreCreateBinary();

  //Enable the wakeup timer interrupt, so that it's interrupt handler is called,
  //and so that the RTC wakeup timer can wake the microcontroller from low power mode.
  RTC_ITConfig(RTC_IT_WUT, ENABLE);

  //When an RTC wakeup event happens, call the RTCIRQHandler method.
  g_exti.EnableRTCIrq(RTCIRQWakeup, this);

  //Register this class as a task that can decide if low power mode can be entered.
  RegisterLowPowerTask("deposittimer");
}

void CDepositTimer::Start()
{
  Create("DepositTimer");
}

/*! Called from the RTC wakeup interrupt, via the g_exti object.*/
void CDepositTimer::RTCIRQHandler(ERTCIRQType irqtype)
{
  xSemaphoreGiveFromISR(m_signal, nullptr);
}

/*! Called from CDeposit before a deposit is done.
    \param opened When true, the container is currently opened.
                  When false, the container is currently closed.*/
void CDepositTimer::BeforeDeposit(bool opened)
{
  //Signal to the CDepositTimer task that any monitoring of the registration
  //switch, and powering of the hold lock magnet should be aborted.
  CCritical criticalsection;
  if (opened)
    m_cmd = EDepositTimerCmd_Stop_Opened;
  else
    m_cmd = EDepositTimerCmd_Stop_Closed;
  criticalsection.Leave();

  xSemaphoreGive(m_signal);
}

/*! Called from CDeposit after a deposit has been done.*/
void CDepositTimer::AfterDeposit()
{
  //Prevent entering low power mode until the timer has been set up.
  AllowLowPower(false);

  //Signal to the CDepositTimer task that a deposit has been done.
  CCritical criticalsection;
  m_cmd = EDepositTimerCmd_Deposit;
  criticalsection.Leave();

  xSemaphoreGive(m_signal);
}

void CDepositTimer::Process()
{
  for(;;)
  {
    //Don't allow entering low power mode while processing.
    AllowLowPower(false);

    //Check if this task has woken up because of a RTC wakeup timer event.
    bool wakeup = RTCWakeupFlagSet();

    //Check if this task has woken up because of a command passed by the CDeposit task.
    CCritical criticalsection;
    EDepositTimerCmd cmd = m_cmd;
    m_cmd                = EDepositTimerCmd_None;
    criticalsection.Leave();

    //A command from the CDeposit task always overrules an RTC wakeup timer event.
    if (cmd == EDepositTimerCmd_Stop_Closed || cmd == EDepositTimerCmd_Stop_Opened)
      Stop(cmd);
    else if (cmd == EDepositTimerCmd_Deposit)
      SetupTimer();
    else if (wakeup)
      OnRTCWakeup();

    //Wait for the timer to expire, or for a deposit to happen.
    //Allow entering low power mode while waiting.
    AllowLowPower(true);
    xSemaphoreTake(m_signal, portMAX_DELAY);
  }
}

/*! Called when the CDeposit task sends a stop command to the CDepositTimer task.*/
void CDepositTimer::Stop(EDepositTimerCmd cmd)
{
  uint32_t stayopen = g_controller.Settings().GetInteger(Setting_CheckStayOpen);
  stayopen = (stayopen + 500) / 1000;

  //If before a Status_DepositOpeningStayedOpen with value "1" has been written,
  //and the container is currently closed, write Status_DepositOpeningStayedOpen
  //with value "0" and contact the server.
  if (stayopen > 0 && ReadBackupFlag(STAYED_OPEN_FLAG) &&
      cmd == EDepositTimerCmd_Stop_Closed)
  {
    g_controller.Status().Write(Status_DepositOpeningStayedOpen, "0");
    g_controller.ServerComm().SignalCommAttempt(COMM_REASON_DEPOSIT_OPENING_CLOSED);
  }

  Reset();
}

/*! Resets the RTC wakeup timer and associated state.*/
void CDepositTimer::Reset()
{
  DisableRTCWakeup();
  WriteBackupRegister(WAKEUP_COUNT_REGISTER, 0);
  WriteBackupRegister(WAKEUP_SECONDS_PASSED_REGISTER, 0);
  SetBackupFlag(STAYED_OPEN_FLAG, false);
  SetBackupFlag(STAYED_CLOSED_FLAG, false);
}

/*! Sets up the RTC to wake up the CDepositTimer task after a set timeout.*/
void CDepositTimer::SetupTimer()
{
  //Convert from milliseconds to seconds.
  uint32_t stayopen = g_controller.Settings().GetInteger(Setting_CheckStayOpen);
  stayopen = (stayopen + 500) / 1000;

  //Enable the deposit timer if the lock has a Release() method,
  //or if CheckStayOpen is enabled.
  if (g_controller.Deposit().LockDriver().HasRelease() || stayopen > 0)
  { //Set up the RTC wakeup timer to wake up this task after the configured delay.
    WriteBackupRegister(WAKEUP_COUNT_REGISTER, 1);
    WriteBackupRegister(WAKEUP_SECONDS_PASSED_REGISTER, 0);
    SetBackupFlag(STAYED_OPEN_FLAG, false);
    SetBackupFlag(STAYED_CLOSED_FLAG, false);
    SetWakeup();
  }
  else
  {
    Reset(); //Stop the periodic wakeup.
  }
}

/*! Called by the CDepositTimer task when an RTC wakeup event has occurred.*/
void CDepositTimer::OnRTCWakeup()
{
  //Read from the battery backed register how many times the hold lock magnet
  //has been powered.
  uint32_t count = ReadBackupRegister(WAKEUP_COUNT_REGISTER);
  count = std::max(count, (uint32_t)1);

  //Disable the RTC wakeup timer first, because a fault in the wiring might cause
  //a reset while powering the hold lock magnet. If the RTC wakeup timer is enabled
  //while that happens, a reboot loop will occur.
  DisableRTCWakeup();

  //Power the release magnet if the lock has one,
  //and power it up to MAX_HOLDLOCK_COUNT times.
  bool release = count <= MAX_HOLDLOCK_COUNT &&
                 g_controller.Deposit().LockDriver().HasRelease();

  //If configured, power the release magnet always, up to MAX_HOLDLOCK_COUNT times,
  //even if the deposit input is closed.
  bool releasealways = g_controller.Settings().GetBoolean(Setting_HoldLockReleaseAlways);

  //Read the state of the deposit input, if it's still opened,
  //power the release magnet if needed.
  bool opened = g_controller.Deposit().CheckDepositInputFromTimer(release, releasealways);

  //If the Setting_CheckStayOpen setting is enabled, convert its value from
  //milliseconds to seconds, and check if the deposit opening is still opened
  //after this time has passed.
  uint32_t stayopen = g_controller.Settings().GetInteger(Setting_CheckStayOpen);
  stayopen = (stayopen + 500) / 1000;
  CheckStayedOpen(stayopen, opened);

  //If the release magnet has been powered the maximum number of times,
  //stop setting up the RTC wakeup timer.
  if (count == MAX_HOLDLOCK_COUNT)
    release = false;

  //Schedule another wakeup timer event, for as long as the release magnet
  //should be powered, or the open state should be checked.
  if ((release && (opened || releasealways)) || (opened && stayopen > 0))
  {
    WriteBackupRegister(WAKEUP_COUNT_REGISTER, count + 1);
    SetWakeup();
  }
  else
  {
    Reset();
  }
}

/*! Checks if the deposit opening stayed open after a deposit.
    \param stayopen The value of Setting_CheckStayOpen, converted to milliseconds.
    \param opened The return value of CheckDepositInputFromTimer().*/
void CDepositTimer::CheckStayedOpen(uint32_t stayopen, bool opened)
{
  uint32_t passed = ReadBackupRegister(WAKEUP_SECONDS_PASSED_REGISTER);

  PrintLog(LOGDEPOSITTMR, "%" PRIu32 " seconds passed, stay open check set to %"
           PRIu32 " seconds, deposit opening stayed open check %s\n",
           passed, stayopen, stayopen > 0 ? "enabled" : "disabled");

  if (stayopen > 0)
  {
    if (opened)
    { //If the deposit opening has stayed open for too long, write a status
      //and contact the server.
      if (passed >= stayopen && !ReadBackupFlag(STAYED_OPEN_FLAG))
      {
        SetBackupFlag(STAYED_OPEN_FLAG, true);
        SetBackupFlag(STAYED_CLOSED_FLAG, false);
        g_controller.Status().Write(Status_DepositOpeningStayedOpen, "1");
        g_controller.ServerComm().
          SignalCommAttempt(COMM_REASON_DEPOSIT_OPENING_STAYED_OPEN);
      }
    }
    else
    {
      //If no status has been written that the deposit opening stayed closed,
      //write it, but don't contact the server.
      if (!ReadBackupFlag(STAYED_CLOSED_FLAG))
      {
        SetBackupFlag(STAYED_CLOSED_FLAG, true);
        g_controller.Status().Write(Status_DepositOpeningStayedOpen, "0");
      }

      //If before a status has been written that the deposit opening was opened,
      //contact the server again.
      if (ReadBackupFlag(STAYED_OPEN_FLAG))
      {
        SetBackupFlag(STAYED_OPEN_FLAG, false);
        g_controller.ServerComm().
          SignalCommAttempt(COMM_REASON_DEPOSIT_OPENING_CLOSED);
      }
    }
  }
}

/*! Sets the RTC wakeup timer to trigger after the time returned by RTCDelayTime().
    Records the total number of seconds in WAKEUP_SECONDS_PASSED_REGISTER.*/
void CDepositTimer::SetWakeup()
{
  uint32_t passed = ReadBackupRegister(WAKEUP_SECONDS_PASSED_REGISTER);
  uint32_t delay  = RTCDelayTime();

  WriteBackupRegister(WAKEUP_SECONDS_PASSED_REGISTER, passed + delay);

  SetRTCWakeup(delay);
}

/*! \return The delay to pass to SetRTCWakeup(), based on if a hold lock is 
            in use, and if CheckStayOpen is active.*/
uint32_t CDepositTimer::RTCDelayTime()
{
  int64_t  stayopen      = g_controller.Settings().GetInteger(Setting_CheckStayOpen);
  int64_t  holdlockdelay = g_controller.Settings().GetInteger(Setting_HoldLockDelay);
  uint32_t wakeupcount   = ReadBackupRegister(WAKEUP_COUNT_REGISTER);

  //If a hold lock is used, use the delay from Setting_HoldLockDelay.
  //Otherwise use the delay from Setting_CheckStayOpen.
  uint32_t delay;
  if (g_controller.Deposit().LockDriver().HasRelease())
    delay = holdlockdelay;
  else
    delay = stayopen;

  //Round the delay from milliseconds to seconds, and clamp to a safe range.
  delay = (delay + 500) / 1000;
  delay = std::min(delay, (uint32_t)MAX_INITIAL_RTC_DELAY);
  delay = std::max(delay, (uint32_t)MIN_INITIAL_RTC_DELAY);

  //Generate a multiplier from the number of wakeups, clamped to a reasonable value.
  uint32_t multiplier = 1;
  if (wakeupcount > 1)
  {
    uint32_t shift = std::min(wakeupcount - 1, (uint32_t)MAX_RTC_DELAY_MULTIPLIER_SHIFT);
    multiplier = 1 << shift;
  }

  //Clamp the delay to a reasonable maximum.
  delay = std::min(delay * multiplier, (uint32_t)MAX_RTC_DELAY);

  return delay;
}
