#include "display.h"
#include "clock.h"
#include "log.h"
#include "lock.h"
#include "gpiodescriptions.h"
#include "controller.h"

#include <stm32f4xx.h>
#include <stm32f4xx_fmc.h>

/*
Driver for the JHD128128-G02BSW LCD, with ST7541 display driver.

The LCD should be configured to 8080 mode.
The FMC should be set to mode A.
When address 0x60000000 is used for reading or writing, the flexible memory controller (FMC) will
use the data pins and the LCD_E_RD or LCD_RW_WR pins to access the LCD registers.
When address 0x60000008 is used, address pin FMC_A3 will go high to access the LCD data memory.

LCD Pin:                LCD Function        MCU Pin:    MCU Function:       Header pin:
LCD_BACKLIGHT_ENABLE    Enables backlight   PG3         GPIO                35
LCD_CSB                 #Chip select        PD7         FMC_NE1/FMC_NCE2    11
LCD_D0                  Data0               PD14        FMC_D0              1
LCD_D1                  Data1               PD15        FMC_D1              2
LCD_D2                  Data2               PD0         FMC_D2              3
LCD_D3                  Data3               PD1         FMC_D3              4
LCD_D4                  Data4               PE7         FMC_D4              6
LCD_D5                  Data5               PE8         FMC_D5              5
LCD_D6                  Data6               PE9         FMC_D6              8
LCD_D7                  Data7               PE10        FMC_D7              7
LCD_E_RD                Read                PD4         FMC_NOE             14
LCD_PWR                 Enables LCD power   PF2         GPIO                10
LCD_RW_WR               Write               PD5         FMC_NWE             12
LCD_RESET               Reset               PF4         GPIO                13
LCD_RS                  H = data, L = ctrl  PF3         FMC_A3              16
*/

#define LCD_REG  (*(volatile uint8_t*)0x60000000) //!< FMC Bank 1 with the A3 line low
#define LCD_DATA (*(volatile uint8_t*)0x60000008) //!< FMC Bank 1 with the A3 line high

CDisplay::CDisplay() : m_dispbuf {}, m_buf {}, m_dispdrv {}, m_scr(nullptr)
{
  m_orientation = Orientation_Top;
}

void CDisplay::Setup(EOrientation orientation)
{
  PrintLog(LOGDISPLAY, "Setting up Display in orientation %i:%s\n",
           (int)orientation, OrientationStr(orientation));

  //Always use a known orientation.
  if (orientation >= Orientation_Start && orientation < Orientation_End)
    m_orientation = orientation;
  else
    m_orientation = Orientation_Top;

  //set up the memory controller so that it can drive the LCD
  SetupFMC();

  //initialize the LCD controller
  Initialize();

  //Test the display connectivity if self testing is enabled.
  Test();

  //set up the LittlevGL library
  SetupLvGL();
}

void CDisplay::Enable()
{
  //turn the LCD on
  LCD_REG = 0xAF;
}

void CDisplay::Disable()
{
  //turn the LCD off
  LCD_REG = 0xAE;
}

/*! Sets up the memory controller so that it can drive the LCD in 8080 mode.*/
void CDisplay::SetupFMC()
{
  //use a critical section to protect the RCC and FMC registers
  CCritical critsection;

  //enable the clock for the FMC
  RCC_AHB3PeriphClockCmd(RCC_AHB3Periph_FMC, ENABLE);

  //initialize the FMC in 8080 mode, with the slowest timings possible
  FMC_NORSRAMTimingInitTypeDef timinginit =
  {
    .FMC_AddressSetupTime      = 0x0f,
    .FMC_AddressHoldTime       = 0x0f,
    .FMC_DataSetupTime         = 0xff,
    .FMC_BusTurnAroundDuration = 0x0f,
    .FMC_CLKDivision           = 0x0f,
    .FMC_DataLatency           = 0x0f,
    .FMC_AccessMode            = FMC_AccessMode_A,
  };

  FMC_NORSRAMInitTypeDef fmcinit = {};
  FMC_NORSRAMStructInit(&fmcinit);

  fmcinit.FMC_Bank                  = FMC_Bank1_NORSRAM1;
  fmcinit.FMC_DataAddressMux        = FMC_DataAddressMux_Disable;
  fmcinit.FMC_MemoryDataWidth       = FMC_NORSRAM_MemoryDataWidth_8b;
  fmcinit.FMC_WaitSignal            = FMC_WaitSignal_Disable;
  fmcinit.FMC_ReadWriteTimingStruct = &timinginit;
  fmcinit.FMC_WriteTimingStruct     = &timinginit;

  FMC_NORSRAMInit(&fmcinit);
  FMC_NORSRAMCmd(FMC_Bank1_NORSRAM1, ENABLE);  
}

/* Sets up LittlevGL so that it can control the LCD.*/
void CDisplay::SetupLvGL()
{
  lv_init();
  lv_disp_buf_init(&m_dispbuf, m_buf, NULL, sizeof(m_buf));

  //initialze the LittlevGL driver
  lv_disp_drv_init(&m_dispdrv);
  m_dispdrv.flush_cb = Flush;
  m_dispdrv.rounder_cb = Rounder;
  m_dispdrv.buffer = &m_dispbuf;
  m_dispdrv.user_data = (lv_disp_drv_user_data_t)this;
  lv_disp_drv_register(&m_dispdrv);

  //set the current theme to mono, this theme is
  //best suited for monochrome displays
  lv_theme_t* theme = lv_theme_mono_init(0, NULL);
  lv_theme_set_current(theme);

  //store a pointer to the active screen
  m_scr = lv_disp_get_scr_act(NULL);
}

/*! Initializes the LCD controller for normal operation.
    This code has been copied from the Barrow 4.*/
void CDisplay::Initialize()
{
  //enable power to the LCD
  GPIO_Set(LCD_PWR);
  DelayMS(50);

  //bring the LCD out of reset
  GPIO_Set(LCD_RESET);
  DelayMS(1);

  LCD_REG = 0x38; /* mode (77 Hz booster level 2) */
  LCD_REG = 0x04;
  
  /* 180 degrees screen rotation */
  LCD_REG = 0xA1; /* ADC reverse direction */
  
  LCD_REG = 0xC0; /* SHL normal direction */
  LCD_REG = 0x44;
  LCD_REG = 0x00; /* com0 = 0 */
  LCD_REG = 0xAB; /* osc on */
  LCD_REG = 0x67; /* dc-dc step-up (step-up 4 times) */
  LCD_REG = 0x56; /* Set Bias (1/11) */
  LCD_REG = 0x40;
  LCD_REG = 0x00; /* initial display line */
  LCD_REG = 0x4C;
  LCD_REG = 0x00; /* N-line inversion = 0 */
  LCD_REG = 0x27; /* Ra/Rb */
  LCD_REG = 0x81;
  LCD_REG = 0x0A; /* Set EV */
  LCD_REG = 0xA8; /* Power normal (mot sleep) */
  
  LCD_REG = 0x93; /* gray scale / PWM mode */
  LCD_REG = 0x88; LCD_REG =  0x00; /* gray scale WHITE */
  LCD_REG = 0x89; LCD_REG =  0x00; /* gray scale WHITE */
  LCD_REG = 0x8A; LCD_REG =  0xAA; /* gray scale LIGHT */
  LCD_REG = 0x8B; LCD_REG =  0xAA; /* gray scale LIGHT */
  LCD_REG = 0x8C; LCD_REG =  0xCC; /* gray scale DARK */
  LCD_REG = 0x8D; LCD_REG =  0xCC; /* gray scale DARK */
  LCD_REG = 0x8E; LCD_REG =  0xFF; /* gray scale BLACK */
  LCD_REG = 0x8F; LCD_REG =  0xFF; /* gray scale BLACK */
  
  LCD_REG = 0x38; LCD_REG = 0x04;  /* mode (77 Hz booster level 2) */

  LCD_REG = 0x2C; //enable internal voltage converter (VC)
  DelayMS(100);

  LCD_REG = 0x2E; //enable internal voltage regulator (VR)
  DelayMS(100);

  LCD_REG = 0x2F; //enable internal voltage follower (VF)
  DelayMS(100);

  //clear the LCD display memory before the LCD is turned on
  Clear();
}

/*! Turns off all pixels of the LCD.*/
void CDisplay::Clear()
{
  //the pixels of the LCD are addressed as 16 rows of 8 pixels high
  for (uint32_t row = 0; row < LCD_ROWS; row++)
  {
    LCD_REG = 0xB0 + row; /* select first page */
    LCD_REG = 0x10;       /* set first column */
    LCD_REG = 0x00;

    //turn the entire row off
    for (uint32_t x = 0; x < LCD_WIDTH; x++)
    {
      LCD_DATA = 0x0;
      LCD_DATA = 0x0;
    }
  }
}

/*! Writes data to the display ram, and read the data back.
    If the display is working correctly, the data should be the same.*/
void CDisplay::Test()
{
  if (!g_controller.SelfTest().Enabled() ||
      g_controller.SelfTest().IsTestPassed(SelfTest_Display))
    return;

  PrintLog(LOGDISPLAY, "Testing display\n");

  //Write a test pattern to the display ram.
  for (uint32_t row = 0; row < LCD_ROWS; row++)
  {
    LCD_REG = 0xB0 + row; //Select first page.
    LCD_REG = 0x10;       //Set first column.
    LCD_REG = 0x00;

    for (uint32_t x = 0; x < LCD_WIDTH; x++)
    {
      LCD_DATA = row + 1;
      LCD_DATA = x + 1;
    }
  }

  //Read the test pattern back.
  bool success = true;
  for (uint32_t row = 0; row < LCD_ROWS; row++)
  {
    LCD_REG = 0xB0 + row; //Select first page.
    LCD_REG = 0x10;       //Set first column.
    LCD_REG = 0x00;
#ifdef GD32F427
   // (void)LCD_DATA;       // NO Dummy read needed after setting the read address. on GD chip EDWIN
#else
(void)LCD_DATA;       //Dummy read needed after setting the read address.for ST chip EDWIN
#endif
    for (uint32_t x = 0; x < LCD_WIDTH; x++)
    {
      uint8_t datarow = LCD_DATA;
      uint8_t datax = LCD_DATA;
      if (datarow != row  + 1 || datax != x + 1)
      {
        PrintLog(LOGDISPLAY, "ERROR: read back 0x%02" PRIX8 " 0x%02" PRIX8
                 " from row %" PRIu32 " column %" PRIu32 ", should have been 0x%02"
                 PRIX32 " 0x%02" PRIX32 "\n", datarow, datax, row, x, row + 1, x + 1);
        success = false;
        break;
      }
    }

    if (!success)
      break;
  }

  //Clear the test pattern.
  Clear();
// EDWIN FLAG TEST OK
 // success=1;
  if (success)
  {
    PrintLog(LOGDISPLAY, "Display is working correctly\n");
    g_controller.SelfTest().SetTestPassed(SelfTest_Display);
  }
}

void CDisplay::TransformTop(uint32_t* xcoord, uint32_t* ycoord)
{
  //Do nothing.
}

void CDisplay::TransformRight(uint32_t* xcoord, uint32_t* ycoord)
{
  //Swap X and Y coordinates, with Y coordinate inverted.
  uint32_t temp = *xcoord;
  *xcoord = LCD_HEIGHT - *ycoord - 1;
  *ycoord = temp;
}

void CDisplay::TransformBottom(uint32_t* xcoord, uint32_t* ycoord)
{
  //Invert X and Y coordinates.
  *xcoord = LCD_WIDTH - *xcoord - 1;
  *ycoord = LCD_HEIGHT - *ycoord - 1;
}

void CDisplay::TransformLeft(uint32_t* xcoord, uint32_t* ycoord)
{
  //Swap X and Y coordinates, with X coordinate inverted.
  uint32_t temp = LCD_WIDTH - *xcoord - 1;
  *xcoord = *ycoord;
  *ycoord = temp;
}

/*! Returns a function pointer to do coordinate transformation depending
    on the display orientation.
    \param orientation The display orientation.
    \return A pointer to a transform function.*/
void (*CDisplay::TransformFunction(EOrientation orientation))(uint32_t*, uint32_t*)
{
  if (orientation == Orientation_Top)
    return TransformTop;
  else if (orientation == Orientation_Right)
    return TransformRight;
  else if (orientation == Orientation_Bottom)
    return TransformBottom;
  else if (orientation == Orientation_Left)
    return TransformLeft;

  //For safety.
  PrintLog(LOGDISPLAY, "ERROR: invalid display orientation %" PRIu32 "\n",
           (uint32_t)orientation);
  return TransformTop;
}

/*! Writes an area from LittlevGL to the LCD.
    \param disp Pointer to the display driver.
    \param area The area to write to the LCD. This should cover the whole lcd.
    \param color The buffer to write to the LCD.*/
void CDisplay::Flush(lv_disp_drv_t* disp, const lv_area_t* area, lv_color_t* color)
{
  //Sanity check if updated area is the full LCD width.
  int32_t width = lv_area_get_width(area);
  if (width != LCD_WIDTH)
    PrintLog(LOGDISPLAY, "ERROR: updated area %" PRIu32 " is not lcd width %i\n",
        width, LCD_WIDTH);

  //Sanity check if updated area is the full LCD height.
  int32_t height = lv_area_get_height(area);
  if (height != LCD_HEIGHT)
    PrintLog(LOGDISPLAY, "ERROR: updated area %" PRIu32 " is not lcd height %i\n",
        height, LCD_HEIGHT);

  //Get a pointer to a coordinate transform function based on the display orientation.
  EOrientation orientation = ((CDisplay*)disp->user_data)->m_orientation;
  void (*Transform)(uint32_t*, uint32_t*) = TransformFunction(orientation);

  //The LCD is driven as rows of 8 pixels high.
  for (int32_t row = 0; row < LCD_ROWS; row++)
  {
    //Select page.
    LCD_REG = 0xB0 + row;

    //Select first column, selecting another column does not work right.
    LCD_REG = 0x10;
    LCD_REG = 0x00;

    //LittlevGL passes a buffer that has pixels addressed from left to right,
    //and from top to bottom, convert them to the format the LCD needs
    for (int32_t x = 0; x < LCD_WIDTH; x++)
    {
      //Create a byte that is 8 pixels above each other.
      uint8_t columnval = 0;
      for (int32_t pixel = LCD_ROWSIZE - 1; pixel >= 0; pixel--)
      {
        uint32_t xcoord = x;
        uint32_t ycoord = row * LCD_ROWSIZE + pixel;

        //Transform the coordinates according to the display orientation.
        //They will then point to the correct pixel in the color buffer.
        Transform(&xcoord, &ycoord);

        columnval <<= 1;
        columnval |= !color[ycoord * LCD_WIDTH + xcoord].full;
      }

      //write the byte twice, once for the light grayscale, once for the dark grayscale
      LCD_DATA = columnval;
      LCD_DATA = columnval;
    }
  }

  //inform LittlevGL that the display has been updated
  lv_disp_flush_ready(disp);
}

/*! Tells LittlevGL which area of the display needs to be updated.
    The whole area is always updated, to make calculating the
    screen rotation easier.*/
void CDisplay::Rounder(lv_disp_drv_t * disp, lv_area_t* area)
{
  area->x1 = 0;
  area->x2 = LCD_WIDTH - 1;

  area->y1 = 0;
  area->y2 = LCD_HEIGHT - 1;
}
