#include "factoryreset.h"
#include "log.h"
#include "stm32wss.h"
#include "backupregisters.h"
#include "accellointflashaddrs.h"
#include "internalflash.h"
#include "clock.h"
#include "controller.h"
#include "gpiodescriptions.h"
#include "active.h"

//! The battery backed register where the number of power on resets is counted.
#define POWER_ON_RESET_COUNTER RTC_BKP_DR14

//! The number of power on resets until a factory reset is done.
#define RESETS_TRIGGER_FACTORY_RESET 5

//! Battery backed flag that indicates that a factory reset happened.
#define FACTORY_RESET_FLAG 0x8000000

/*! Battery backed flag that indicates that a factory reset status should be
    written. This will be done on the first server contact, after the RTC has
    been set to the correct time.*/
#define FACTORY_RESET_WRITE_STATUS_FLAG 0x4000000

/*! Checks if a factory reset should be done,
    based on the microcontroller reset reason.*/
void CFactoryReset::Check()
{
  //Store in ram if a factory reset happened since the last reboot.
  if (ReadBackupFlag(FACTORY_RESET_FLAG))
  {
    SetBackupFlag(FACTORY_RESET_FLAG, false);
    m_happened = true;

    //Write a factory reset status on the next server contact.
    SetBackupFlag(FACTORY_RESET_WRITE_STATUS_FLAG, true);
  }

  uint32_t numresets = ReadBackupRegister(POWER_ON_RESET_COUNTER);

  //Check if a power on reset happened, if so, increase the reset counter by 2.
  //If not, decrease it by 1. This way, a factory reset can be done by quickly
  //doing power cycles with a reset magnet.
  //Check if a watchdog timer reset happened. This only happens on a freeze,
  //or when the FatalError() function is called. If this happens too much,
  //perform a factory reset.
  //Check if a software reboot has happened, this happens when the microcontroller
  //stays out of low power mode for too long. If this happens too much, perform
  //a factory reset.
  //Factory reset is not allowed when the device is inactive, because power
  //cycles might happen a lot, and it resets the inactive statuses that are kept
  //in the settings.
  if (DeviceActive() &&
      ((ResetFlags() & ResetFlagMask(RCC_FLAG_PORRST))  ||
       (ResetFlags() & ResetFlagMask(RCC_FLAG_IWDGRST)) ||
        SoftwareRebootTimeHappened()))
  {
    numresets += 2;
    if (numresets >= RESETS_TRIGGER_FACTORY_RESET * 2)
    {
      WriteBackupRegister(POWER_ON_RESET_COUNTER, 0);
      Perform();
    }
    else
    {
      WriteBackupRegister(POWER_ON_RESET_COUNTER, numresets);
      PrintLog(LOGFACTORYRST, "Increasing reset counter to %" PRIu32 "\n", numresets);
    }
  }
  else if (numresets > 0)
  {
    numresets--;
    WriteBackupRegister(POWER_ON_RESET_COUNTER, numresets);
    PrintLog(LOGFACTORYRST, "Decreasing reset counter to %" PRIu32 "\n", numresets);
  }
}

/*! Performs a factory reset.*/
void CFactoryReset::Perform()
{
  PrintLog(LOGFACTORYRST, "Performing factory reset\n");

  //Indicate to the user that a factory reset is in progress.
  GPIO_Set(LCD_PWR);
  g_controller.UserInterface().SetBacklight(true);

  //Erase all flash sectors, except 
  for (uint32_t i = 0; i < NUM_FACTORY_RESET_SECTORS; i++)
  {
    uint32_t sector = FACTORY_RESET_SECTORS[i];
    PrintLog(LOGFACTORYRST, "Erasing internal flash sector %" PRIu32 "\n", sector);
    FlashEraseSector(sector);
  }

  //Re-initialize the RTC, this also clears all battery backed registers by
  //performing a backup domain reset.
  RTCInit(true);

  uint32_t start = Micros();
  uint32_t prev  = start;

  //Erase the SPI flash.
  uint32_t numsectors = g_controller.SPIFlash().NumSectors();
  for (uint32_t i = 0; i < numsectors; i++)
  {
    PrintLog(LOGFACTORYRST, "Erasing SPI flash sector %" PRIu32 " of %" PRIu32
             "\n", i, numsectors - 1);
    g_controller.SPIFlash().Erase(i, 1);

    //Indicate to the user that something is happening.
    g_controller.UserInterface().SetBacklight(i % 2);

    //The watchdog timer is normally reset by the heartbeat timer, but that is
    //not running at this point. Reset the watchdog timer here instead, but only
    //if the Micros() counter is ticking forward, and this loop doesn't run for
    //too long.
    uint32_t now = Micros();
    if (now > prev && now - start < (uint32_t)60 * 60 * 1000000)
    {
      IWDG_ReloadCounter();
      prev = now;
    }
  }

  //Set the battery backed factory reset flag, so that after the next reboot,
  //it's known that a factory reset happened.
  SetBackupFlag(FACTORY_RESET_FLAG, true);

  //Reset the microcontroller after a factory reset.
  DelayMS(1000);
  NVIC_SystemReset();
}

/*! Checks if the FACTORY_RESET_WRITE_STATUS_FLAG is set in the backup register.
    If so, writes a Status_FactoryReset status. This is called during server
    communication after the RTC has been set to the correct time, because the RTC
    is reset to 1-1-2000 during a factory reset.*/
void CFactoryReset::WriteStatus()
{
  if (ReadBackupFlag(FACTORY_RESET_WRITE_STATUS_FLAG))
  {
    SetBackupFlag(FACTORY_RESET_WRITE_STATUS_FLAG, false);
    g_controller.Status().Write(Status_FactoryReset, "1");
  }
}

bool CFactoryReset::m_happened;
