#include "images.h"
#include "controller.h"
#include "log.h"
#include "internalflash.h"
#include "accellointflashaddrs.h"
#include "lock.h"
#include <algorithm>
#include <cctype>

/* Key with which the images file is signed, using a SHA256 HMAC.*/
uint8_t CImages::m_signkey[CRYPTO_KEY_SIZE] =
  { 0x54, 0xc6, 0xd6, 0xa0, 0x44, 0xf6, 0xa9, 0x2c,
    0x0f, 0x63, 0xfa, 0xe8, 0x43, 0x0b, 0x69, 0xde };

CImages::CImages()
{
  m_intflashaddr = nullptr;
  m_imagesfile   = nullptr;
  m_fileheader   = nullptr;
  m_filecorrect  = false;
  m_mutex        = nullptr;
  m_prevtype     = Image_None;
}

void CImages::Setup()
{
  m_mutex = xSemaphoreCreateRecursiveMutex();

  //Calculate where in the SPI flash the reserved space for the images file is.
  //It's reserved from the end of the SPI flash, because the stm32wss library
  //reserves space from the start of the SPI flash.
  m_spiflashpos = g_controller.SPIFlash().Size() -
                  IMAGES_DOWNLOAD_SIZE - IMAGES_DOWNLOAD_ADDRESS;

  //Sanity checks.
  assert((IMAGES_DOWNLOAD_SIZE / g_controller.SPIFlash().SectorSize()) >= 1);
  assert((IMAGES_DOWNLOAD_SIZE % g_controller.SPIFlash().SectorSize()) == 0);

  //Set up pointers to the internal flash.
  m_intflashaddr = (uint8_t*)FlashSectorAddress(IMAGES_SECTOR);
  m_imagesfile   = (SImagesFile*)m_intflashaddr;
  m_fileheader   = (SImagesFileHeader*)(m_intflashaddr + sizeof(SImagesFile));

  //Check if there is an images file installed in the internal flash.
  if (m_imagesfile->version != IMAGES_FILE_ERASED)
  {
    if (!CheckInstalledFile())
      EraseInstalled(); //Erase what is there if it's not correct.
  }
  else
  {
    PrintLog(LOGIMAGES, "Images file not installed\n");
  }
}

/*! Checks if the reserved space for the images is correct.
    Erases the reserved space in the SPI flash.
    \return true if the reserved space was correctly erased.
    \return false if anything went wrong.*/
bool CImages::Prepare()
{
  m_bytesdownloaded = 0;

  if (!g_controller.SPIFlash().Enabled())
  {
    PrintLog(LOGIMAGES, "ERROR: SPI FLASH disabled\n");
    return false;
  }

  //Erase the section of SPI flash that is reserved for downloading the images file.
  if (!EraseSPIFlash())
    return false;

  return true;
}

/*! Returns the maximum block size this class can handle through the Download() method,
    set to the page size of the SPI flash, this has the best throughput, because the
    spi flash is writing while the firmware is getting data from the modem.
    \return The maximum block size this class can handle through the Download method.*/
uint32_t CImages::DownloadBlockSize()
{
  return g_controller.SPIFlash().PageSize();
}

void CImages::StartDownload(const char* url, uint32_t httprspcode, uint32_t contentlength)
{
  m_downloadsize = contentlength;
}

/*! Maximum size the downloaded images file can be.
    Size of the internal flash sector used to store the images file,
    minus the stored SImagesFile struct.*/
#define MAX_IMAGES_FILE_SIZE ((64 * 1024) - sizeof(SImagesFile))

/*! Minimum size the downloaded images file should be.*/
#define MIN_IMAGES_FILE_SIZE (sizeof(SImagesFileHeader))

/*! Called for each block of images file that is downloaded by the modem.
    \param buf The buffer that holds the downloaded data.
    \param size The size of the downloaded data.*/
void CImages::Download(const uint8_t* buf, uint32_t size)
{
  assert(size <= DownloadBlockSize());

  if (size + m_bytesdownloaded > MAX_IMAGES_FILE_SIZE)
  {
    PrintLog(LOGIMAGES, "ERROR: being asked to store %" PRIu32
             " bytes, but max download size is %i bytes\n",
             m_bytesdownloaded + size, MAX_IMAGES_FILE_SIZE);
    m_bytesdownloaded += size;
    return;
  }
  
  //Write the downloaded data to the SPI flash.
  uint32_t writepos = m_spiflashpos + m_bytesdownloaded;
  if (!g_controller.SPIFlash().Write(writepos, buf, size, false))
  {
    PrintLog(LOGIMAGES, "ERROR: unable to write downloaded images file data to SPI flash\n");
    return;
  }

  m_bytesdownloaded += size;
}

/*! Checks if a different images file should be installed.
    \param settingsurl The url from Setting_ImagesURL.
    \return true if settingsurl is not the same as where the
                 installed images file was downloaded from.
    \return false if settingsurl is the same as the url where
                  the installed images file was downloaded from.*/
bool CImages::URLChanged(const char* settingsurl)
{
  CLock lock(m_mutex);

  if (!m_filecorrect)
    return true;

  if (strncmp(settingsurl, m_imagesfile->url, URL_SIZE) != 0)
    return true;

  return false;
}

/*! Copies the images file from the SPI flash to the internal flash.
    \param url The url from which the images file has been downloaded.
    \return true if the copy succeeded.
    \return false if anything failed.*/
bool CImages::Install(const char* url)
{
  //Check if the file has been downloaded correctly.
  if (!CheckDownloadedFile())
  {
    EraseSPIFlash(); //Erase what has just been downloaded.
    return false;
  }

  //Install the downloaded file into the internal flash,
  //and check if it has been installed correctly.
  if (InstallDownloadedFile(url) && CheckInstalledFile(true))
  {
    //Update the version field to mark the installation as complete.
    FinalizeInstall();
  }
  else
  {
    EraseSPIFlash();  //Erase what has just been downloaded.
    EraseInstalled(); //Erase what has just been installed.
    return false;
  }

  return true;
}

/*! Erases the space in the SPI flash reserved for downloading the images file.
    \return true if erasing the SPI flash succeeded.
    \return false if anything failed.*/
bool CImages::EraseSPIFlash()
{
  uint32_t sectorsize  = g_controller.SPIFlash().SectorSize();
  uint32_t startsector = m_spiflashpos / sectorsize;
  uint32_t numsectors  = IMAGES_DOWNLOAD_SIZE / sectorsize;
  if (!g_controller.SPIFlash().Erase(startsector, numsectors))
  {
    PrintLog(LOGIMAGES, "ERROR: erasing flash failed\n");
    return false;
  }

  return true;
}

/*! Checks if the downloaded images file is correct.
    \return true The downloaded images file is correct.
    \return false The downloaded images file is incorrect.*/
bool CImages::CheckDownloadedFile()
{
  PrintLog(LOGIMAGES, "Checking downloaded file of size %" PRIu32 "\n",
           m_bytesdownloaded);

  if (m_bytesdownloaded > MAX_IMAGES_FILE_SIZE)
  {
    PrintLog(LOGIMAGES, "ERROR: not installing images file, download is too large\n");
    return false;
  }
  else if (m_bytesdownloaded < MIN_IMAGES_FILE_SIZE)
  {
    PrintLog(LOGIMAGES, "ERROR: not installing images file, download of %"
             PRIu32 " bytes is too small, should be at least %" PRIu32 " bytes\n",
             m_bytesdownloaded, (uint32_t)MIN_IMAGES_FILE_SIZE);
    return false;
  }

  //Read the SHA256 HMAC at the start of the downloaded images file.
  uint8_t filehmac[SHA256_BLOCK_SIZE] = {};
  if (!g_controller.SPIFlash().Read(m_spiflashpos, filehmac, SHA256_BLOCK_SIZE))
  {
    PrintLog(LOGIMAGES, "ERROR: not installing images file, unable to read from SPI flash\n");
    return false;
  }

  //Generate a SHA256 HMAC over the rest of the downloaded images file.
  HMAC_SHA256Context hmacctx;
  HMAC_SHA256_Init(&hmacctx, m_signkey, CRYPTO_KEY_SIZE);
  uint8_t buf[512];
  for (uint32_t i = SHA256_BLOCK_SIZE; i < m_bytesdownloaded; i += sizeof(buf))
  {
    uint32_t readpos  = m_spiflashpos + i;
    uint32_t readsize = std::min((uint32_t)sizeof(buf), m_bytesdownloaded - i);

    if (!g_controller.SPIFlash().Read(readpos, buf, readsize))
    {
      PrintLog(LOGIMAGES, "ERROR: not installing images file, unable to read from SPI flash\n");
      return false;
    }

    HMAC_SHA256_Update(&hmacctx, buf, readsize);
  }

  //WARNING: because of security, do not print the generated HMAC to the log.
  uint8_t generatedhmac[SHA256_BLOCK_SIZE] = {};
  HMAC_SHA256_Final(&hmacctx, generatedhmac);

  //Check in linear time if the generated HMAC matches the downloaded HMAC.
  bool hmacmatch = true;
  for (uint32_t i = 0; i < SHA256_BLOCK_SIZE; i++)
    hmacmatch = hmacmatch && (generatedhmac[i] == filehmac[i]);

  //Clear the generated HMAC from memory as quickly as possible.
  memset(generatedhmac, 0, SHA256_BLOCK_SIZE);

  //If the HMAC is incorrect, don't use the downloaded images file.
  if (!hmacmatch)
  {
    CLogLock loglock;
    PrintLog(LOGIMAGES, "ERROR: HMAC in image file is incorrect: ");
    PrintLogHexBuf(LOGIMAGES, filehmac, sizeof(filehmac), "\n");
    loglock.Leave();

    return false;
  }

  PrintLog(LOGIMAGES, "Downloaded file of size %" PRIu32 " is correct\n",
           m_bytesdownloaded);

  return true;
}

/*! Erases the internal flash sector that holds the images file.*/
bool CImages::EraseInstalled()
{
  CLock lock(m_mutex);

  //Indicate that the images file that was installed is now invalid.
  m_filecorrect = false;

  if (!FlashEraseSector(IMAGES_SECTOR))
  {
    PrintLog(LOGIMAGES, "ERROR: unable to erase sector %i for images file\n", IMAGES_SECTOR);
    return false;
  }

  return true;
}

/*! Copies the downloaded images file from SPI flash to the internal flash.
    \param url This string is stored in the internal flash, so that if
               Setting_ImagesURL changes, the new file is downloaded.*/
bool CImages::InstallDownloadedFile(const char* url)
{
  PrintLog(LOGIMAGES, "Installing images file into internal flash\n");

  CLock lock(m_mutex);

  //Erase the images file that was installed before.
  if (!EraseInstalled())
    return false;

  //Initialize the version field with the initialization value, so that if the
  //images file contains data that causes a crash, the sector is erased at the next boot.
  SImagesFile imagesfile = {};
  imagesfile.version = IMAGES_FILE_VERSION_INIT;
  imagesfile.size = m_bytesdownloaded;
  strlcpy(imagesfile.url, url, URL_SIZE);

  //Write a header indicating that the flash sector holds a valid images file.
  FlashWrite((uint32_t)m_intflashaddr, (uint8_t*)&imagesfile, sizeof(imagesfile));

  //Copy the images file from the SPI flash to the internal flash.
  uint8_t buf[512];
  for (uint32_t i = 0; i < m_bytesdownloaded; i += sizeof(buf))
  {
    uint32_t readpos  = m_spiflashpos + i;
    uint32_t writepos = sizeof(imagesfile) + i;
    uint32_t size     = std::min((uint32_t)sizeof(buf), m_bytesdownloaded - i);

    if (!g_controller.SPIFlash().Read(readpos, buf, size))
    {
      PrintLog(LOGIMAGES, "ERROR: unable to read from SPI flash\n");
      return false;
    }

    if (!FlashWrite((uint32_t)m_intflashaddr + writepos, buf, size))
    {
      PrintLog(LOGIMAGES, "ERROR: unable to write to internal flash\n");
      return false;
    }
  }

  PrintLog(LOGIMAGES, "Installed images file into internal flash\n");

  return true;
}

/*! Checks if the installed images file is correct.
    \param init When true, this is the initial check after installation.*/
bool CImages::CheckInstalledFile(bool init /*= false*/)
{
  CLock lock(m_mutex);

  m_filecorrect = false;

  //Check if the version field is set to the correct value.
  if ((init  && m_imagesfile->version != IMAGES_FILE_VERSION_INIT) ||
      (!init && m_imagesfile->version != IMAGES_FILE_VERSION))
  {
    PrintLog(LOGIMAGES, "ERROR: images file has incorrect version field: %" PRIX32 "\n",
             m_imagesfile->version);
    return false;
  }

  PrintLog(LOGIMAGES, "Checking images file in internal flash\n");

  //Do a sanity check on the file size.
  if (m_imagesfile->size < MIN_IMAGES_FILE_SIZE ||
      m_imagesfile->size > MAX_IMAGES_FILE_SIZE)
  {
    PrintLog(LOGIMAGES, "ERROR: installed images file has incorrect size %"
             PRIu32 ", should be between %" PRIu32 " and %" PRIu32 " bytes\n",
             m_imagesfile->size, (uint32_t)MIN_IMAGES_FILE_SIZE,
             (uint32_t)MAX_IMAGES_FILE_SIZE);
    return false;
  }

  //Do a sanity check on the url length.
  uint32_t urlsize = strnlen(m_imagesfile->url, URL_SIZE);
  if (urlsize == 0)
  {
    PrintLog(LOGIMAGES, "ERROR: installed images file has no url set\n");
    return false;
  }
  else if (urlsize >= URL_SIZE)
  {
    PrintLog(LOGIMAGES, "ERROR: installed images file has url of length %" PRIu32
             ", which is larger than the max %i\n", urlsize, URL_SIZE);
    return false;
  }

  //Do a sanity check on the url characters.
  for (uint32_t i = 0; i < urlsize; i++)
  {
    if (!isprint(m_imagesfile->url[i]))
    {
      PrintLogDecode(LOGIMAGES, "ERROR: URL has a non-printable character"
                     " at position %" PRIu32 ": \"%s\"", i, m_imagesfile->url);
      return false;
    }
  }

  //Generate a SHA256 HMAC over the installed images file, starting after
  //the HMAC in the images file.
  HMAC_SHA256Context hmacctx;
  HMAC_SHA256_Init(&hmacctx, m_signkey, CRYPTO_KEY_SIZE);
  HMAC_SHA256_Update(&hmacctx, (uint8_t*)&m_fileheader->headersize,
                     m_imagesfile->size - SHA256_BLOCK_SIZE);

  //WARNING: because of security, do not print the generated HMAC to the log.
  uint8_t generatedhmac[SHA256_BLOCK_SIZE] = {};
  HMAC_SHA256_Final(&hmacctx, generatedhmac);

  //Check in linear time if the generated HMAC matches the installed HMAC.
  bool hmacmatch = true;
  for (uint32_t i = 0; i < SHA256_BLOCK_SIZE; i++)
    hmacmatch = hmacmatch && (generatedhmac[i] == m_fileheader->hmac[i]);

  //Clear the generated HMAC from memory as quickly as possible.
  memset(generatedhmac, 0, SHA256_BLOCK_SIZE);

  //If the HMAC is incorrect, the installed images file is not correct.
  if (!hmacmatch)
  {
    CLogLock loglock;
    PrintLog(LOGIMAGES, "ERROR: HMAC in installed image file is incorrect: ");
    PrintLogHexBuf(LOGIMAGES, m_fileheader->hmac, SHA256_BLOCK_SIZE, "\n");
    loglock.Leave();

    return false;
  }

  //Do a sanity check on the header size of the installed images file.
  if (m_fileheader->headersize < 1 ||
      m_fileheader->headersize >= (MAX_IMAGES_FILE_SIZE - sizeof(SImagesFileHeader)) ||
      m_fileheader->headersize > m_imagesfile->size - SHA256_BLOCK_SIZE - sizeof(uint32_t))
   {
    PrintLog(LOGIMAGES, "ERROR; images file has incorrect header size %" PRIu32 "\n",
             m_fileheader->headersize);
    return false;
  }

  //Check if the installed images file has a correct CBOR header.
  CborParser parser;
  CborValue  root;
  CborValue  images;
  if (!OpenHeader(&parser, &root, &images))
  {
    PrintLog(LOGIMAGES, "ERROR: images file has incorrect CBOR header\n");
    return false;
  }

  PrintLog(LOGIMAGES, "Images file installed in internal flash from url %s"
           " is correct\n", m_imagesfile->url);

  //Indicate that the installed images file is correct.
  m_filecorrect = true;

  return true;
}

/*! Sets the version field in the internal flash sector to IMAGES_FILE_VERSION
    to indicate that the data is correct.*/
void CImages::FinalizeInstall()
{
  //Copy the SImagesFile struct from the internal flash, and set the version field.
  SImagesFile imagesfile = *m_imagesfile;
  imagesfile.version = IMAGES_FILE_VERSION;

  //Write the SImageFile struct back to the internal flash.
  FlashWrite((uint32_t)m_intflashaddr, (uint8_t*)&imagesfile, sizeof(imagesfile));
}

/*! Opens the CBOR header of the installed images file.
    \param parser The parser struct that the CBOR header is opened with.
    \param root The first element of the CBOR header, this should be a map.
    \param images The "I" array in the root map, this should be present.
    \return true The CBOR header was opened successfully.
    \return false The CBOR header was not opened successfully.*/
bool CImages::OpenHeader(CborParser* parser, CborValue* root, CborValue* images)
{
  //Check if the installed images file has a correct CBOR header.
  CborError retval;
  retval = cbor_parser_init(m_fileheader->header, m_fileheader->headersize,
                            CborValidateBasic, parser, root);

  if (retval != CborNoError)
  {
    PrintLog(LOGIMAGES, "ERROR: cbor_parser_init returned %i:%s on images file header\n",
             (int)retval, cbor_error_string(retval));
    return false;
  }

  //Check if the header contains valid CBOR.
  retval = cbor_value_validate_basic(root);
  if (retval != CborNoError)
  {
    PrintLog(LOGIMAGES, "ERROR: cbor_value_validate_basic returned %i:%s "
             "on images file header\n", (int)retval, cbor_error_string(retval));
  }

  if (!cbor_value_is_map(root))
  {
    CLogLock loglock;
    PrintLog(LOGIMAGES, "ERROR: the first CBOR element in the"
             " images file header is not a map: ");
    cbor_value_to_pretty(stdout, root);
    PrintLog(LOGIMAGES, "\n");
    return false;
  }

  //Check if the CBOR root map has an "I" array.
  if (!CBORFindMapArray(LOGIMAGES, "images", root, "I", images))
  {
    PrintLog(LOGIMAGES, "ERROR: \"I\" array not present in root map");
    return false;
  }

  return true;
}

/*! Verifies that the image is in a valid position, and has a correct size,
    then fills the img struct with the image information.
    \param img The image data read from the internal flash.
    \param position The position of the image in the internal flash.
    \return true The image has a correct position and size.
    \return false On any error.*/
bool CImages::VerifyImage(lv_img_dsc_t* img, uint32_t position)
{
  //Check if the image position is inside the images file.
  uint32_t imgfilepos = position + m_fileheader->headersize;
  if (imgfilepos >= m_imagesfile->size)
  {
    PrintLog(LOGIMAGES, "ERROR: image has incorrection position %" PRIu32 "\n",
             position);
    return false;
  }

  //Copy the image header from the images file into the img struct.
  uint8_t* imgptr = m_fileheader->header + imgfilepos;
  memcpy(img, imgptr, sizeof(lv_img_dsc_t));
  img->data = imgptr + sizeof(img->header) + sizeof(img->data_size);

  //Check if the image size fits in the images file.
  if (img->data_size > m_imagesfile->size - imgfilepos)
  {
    PrintLog(LOGIMAGES, "ERROR: image has incorrection size %" PRIu32 "\n",
             img->data_size);
    return false;
  }

  //Check if the data_size member corresponds with the needed datasize for an image
  //of this format and these dimensions. This check is not complete, since LVGL
  //might store some color information at the start of the image data.
  uint32_t neededsize = lv_img_color_format_get_px_size(img->header.cf);
  neededsize = (neededsize * img->header.w * img->header.h) / 8;
  if (neededsize > img->data_size)
  {
    PrintLog(LOGIMAGES, "ERROR: image needs size %" PRIu32 ", but data_size is only %"
             PRIu32 "\n", neededsize, img->data_size);
    return false;
  }

  return true;
}

/*! Gets the image corresponding with the type. If the image is present in the
    downloaded file, then that image is used. Otherwise a builtin image is returned.
    \param type The type of image to get.
    \param img The image is returned here.*/
void CImages::Get(EImage type, lv_img_dsc_t* img)
{
  memset(img, 0, sizeof(lv_img_dsc_t));

  CLock lock(m_mutex);

  if (!GetImageExternal(type, img))
    GetImageInternal(type, img);
}

/*! Gets an image from the downloaded images file.
    \param type The type of image to get.
    \param img The image is returned here.
    \return true The image was found in the external file.
    \return false The image was not found in the external file.*/
bool CImages::GetImageExternal(EImage type, lv_img_dsc_t* img)
{
  CLock lock(m_mutex);

  if (!m_filecorrect)
    return false; //No images file installed in internal flash.

  //Open the CBOR header of the installed images file.
  CborParser parser;
  CborValue  root;
  CborValue  images;
  if (!OpenHeader(&parser, &root, &images))
  {
    PrintLog(LOGIMAGES, "ERROR: images file has incorrect CBOR header\n");
    return m_filecorrect = false;
  }

  //Enter the images array, on any error mark the installed file as invalid.
  CborValue it;
  if (!CBOREnterContainer(LOGIMAGES, "images array", &images, &it))
    return m_filecorrect = false;

  //The Images array is an array of maps, find an array element that has a "T"
  //integer member with the value that matches the type argument.
  while (!cbor_value_at_end(&it))
  {
    int64_t imgtype;
    if (!CBORFindMapInt(LOGIMAGES, "image", &it, "T", &imgtype))
      return m_filecorrect = false;

    if (imgtype == (int64_t)type)
    {
      int64_t position;
      if (!CBORFindMapInt(LOGIMAGES, "image", &it, "P", &position))
        return m_filecorrect = false;

      //Check if the image is correct, then fill the img struct with
      //data from the images file in the internal flash.
      if (!VerifyImage(img, position))
        return m_filecorrect = false;

      return true; //The image has been found in the images file.
    }
    
    if (!CBORValueAdvance(LOGIMAGES, "images array", &it))
      return m_filecorrect = false;
  }

  return false;
}

/*! Gets a builtin image, if the image type is unknown the error image is returned.
    \param type The type of image to get.
    \param img The image is returned here.*/
void CImages::GetImageInternal(EImage type, lv_img_dsc_t* img)
{
  CLock lock(m_mutex);

  if (type == Image_Locked)
    memcpy(img, &locked_img, sizeof(lv_img_dsc_t));
  else if (type == Image_Unlocked)
    memcpy(img, &unlocked_img, sizeof(lv_img_dsc_t));
  else if (type == Image_Denied)
    memcpy(img, &denied_img, sizeof(lv_img_dsc_t));
  else if (type == Image_Success)
    memcpy(img, &success_img, sizeof(lv_img_dsc_t));
  else if (type == Image_SuccessPrepaid)
    memcpy(img, &successprepaid_img, sizeof(lv_img_dsc_t));
  else if (type == Image_LowCredit)
    memcpy(img, &lowcredit_img, sizeof(lv_img_dsc_t));
  else if (type == Image_Error)
    memcpy(img, &error_img, sizeof(lv_img_dsc_t));
  else if (type == Image_Full)
    memcpy(img, &full_img, sizeof(lv_img_dsc_t));
  else if (type == Image_Empty)
    memcpy(img, &empty_img, sizeof(lv_img_dsc_t));
  else if (type == Image_ServerComm)
    memcpy(img, &servercomm_img, sizeof(lv_img_dsc_t));
  else if (type == Image_InsufficientCredit)
    memcpy(img, &insufficientcredit_img, sizeof(lv_img_dsc_t));
  else if (type == Image_CardError)
    memcpy(img, &carderror_img, sizeof(lv_img_dsc_t));
  else
  {
    if (m_prevtype != type)
      PrintLog(LOGIMAGES, "ERROR: asked to get image of unknown type %i\n", (int)type);

    memcpy(img, &error_img, sizeof(lv_img_dsc_t));
  }

  m_prevtype = type;
}
