#include "inputmonitor.h"
#include "log.h"
#include "backupregisters.h"
#include "controller.h"
#include "clock.h"

/*! How long the fill level reset input should be at a different
    level before it's considered to be at a stable level.*/
#define SWITCH_STABLE_TIME 1000
#define SWITCH_TIMEOUT     5000

CInputMonitor::CInputMonitor()
{
}

void CInputMonitor::Setup()
{
  //Register this class as a task that can decide if low power mode can be entered.
  RegisterLowPowerTask("inputmonitor");

  //The RTCSchedulerCallback() method should be called when the door alarm expires.
  g_controller.RTCScheduler().SetCallback(DOOR_ALARM_INDEX, this);
}

void CInputMonitor::Start()
{
  Create("InputMonitor");
}

/*! \return true If checking the door switch is enabled, and a change on
                 INPUT_DOOR_SWITCH should wake the microcontroller from low
                 power mode.*/
bool CInputMonitor::CheckDoorSwitch()
{
  if (g_controller.Settings().GetInteger(Setting_CheckDoorStayOpen) > 0)
    return true; //Check has been enabled in settings.
  else if (ReadBackupFlag(DOOR_STATUS_FLAG))
    return true; //Status_DoorStayedOpen with "0" should be written.
  else
    return false;
}

void CInputMonitor::Process()
{
  for(;;)
  {
    bool checkdoorswitch = CheckDoorSwitch();

    //The previous states of the level reset and door switch inputs are stored
    //in the battery backed ram.
    bool oldlevelreset = ReadBackupFlag(FILL_LEVEL_SWITCH_FLAG);
    bool olddoorstate  = ReadBackupFlag(DOOR_SWITCH_FLAG);

    uint8_t inputs = INPUT_FILL_LEVEL_RESET;

    //Only monitor INPUT_DOOR_SWITCH if the door switch check is enabled.
    //If it's disabled, reset DOOR_SWITCH_FLAG to false.
    if (checkdoorswitch)
      inputs |= INPUT_DOOR_SWITCH;
    else if (olddoorstate)
      SetBackupFlag(DOOR_SWITCH_FLAG, false);

    uint8_t wantedstates = (oldlevelreset ? 0 : INPUT_FILL_LEVEL_RESET) |
                           (olddoorstate ? 0 : INPUT_DOOR_SWITCH);
    uint8_t readlevels;

    //Wait for any of the selected inputs to change.
    if (g_controller.Inputs().WaitForStableSignal(inputs,
                                                  SWITCH_STABLE_TIME,
                                                  SWITCH_TIMEOUT,
                                                  wantedstates, &readlevels))
    {
      //Deny going into low power mode when an input has changed.
      AllowLowPower(false);

      CheckLevelReset(oldlevelreset, readlevels);

      if (checkdoorswitch)
        CheckDoor(olddoorstate, readlevels);
    }
    else
    {
      //Allow the microcontroller to go into low power mode if the inputs are
      //at a stable level for some time.
      AllowLowPower(true);
    }
  }
}

/*! This is called from the CRTCScheduler task when the alarm expires.
    \param index The index of the alarm, should be DOOR_ALARM_INDEX.*/
void CInputMonitor::RTCSchedulerCallback(uint32_t index)
{
  uint8_t inputstate;
  bool stable = g_controller.Inputs().
                  ReadStable(INPUT_DOOR_SWITCH, SWITCH_STABLE_TIME,
                             SWITCH_TIMEOUT, &inputstate);

  //If the door switch input it still high after the RTC alarm expires,
  //write a Status_DoorStayedOpen with value 1 and contact the server.
  //This informs the server that someone should go the container and close the
  //door.
  if (stable && (inputstate & INPUT_DOOR_SWITCH))
  {
    PrintLog(LOGINPUTMNTR, "Door stayed open, contacting server\n");

    g_controller.Status().Write(Status_DoorStayedOpen, "1");

    //Remember that Status_DoorStayedOpen has been written to 1, so that
    //when the door closes, the same status should be written with value 0.
    SetBackupFlag(DOOR_STATUS_FLAG, true);

    g_controller.ServerComm().SignalCommAttempt(COMM_REASON_DOOR_STAYED_OPEN);
  }
}

void CInputMonitor::CheckLevelReset(bool oldlevelreset, uint8_t readlevels)
{
  //Reset the fill level if a signal on the fill level reset input has
  //been detected.
  bool levelreset = readlevels & INPUT_FILL_LEVEL_RESET;
  if (levelreset != oldlevelreset)
  {
    PrintLog(LOGINPUTMNTR, "Fill level reset input changed from %s to %s\n",
             CInputs::StateStr(oldlevelreset), CInputs::StateStr(levelreset));
    SetBackupFlag(FILL_LEVEL_SWITCH_FLAG, levelreset);

    //Only reset the fill level on a rising edge.
    if (levelreset)
      g_controller.Deposit().ResetFillLevel();
  }
}

void CInputMonitor::CheckDoor(bool olddoorstate, uint8_t readlevels)
{
  bool doorstate = readlevels & INPUT_DOOR_SWITCH;
  if (doorstate != olddoorstate)
  {
    PrintLog(LOGINPUTMNTR, "Door input changed from %s to %s\n",
             CInputs::StateStr(olddoorstate), CInputs::StateStr(doorstate));
    SetBackupFlag(DOOR_SWITCH_FLAG, doorstate);

    if (doorstate)
    { //If CheckDoorStayOpen is enabled, and the door has been opened,
      //set a timer to check after the configured
      //time if it's still opened, if so the server will be contacted.
      int64_t checkdelay =
        g_controller.Settings().GetInteger(Setting_CheckDoorStayOpen);

      if (checkdelay > 0)
      { //Convert from millisconds to seconds.
        checkdelay = (checkdelay + 500) / 1000;

        PrintLog(LOGINPUTMNTR, "Checking if door is still opened after %" PRIi64
                 " seconds\n", checkdelay);
        g_controller.RTCScheduler().SetCountDown(DOOR_ALARM_INDEX, checkdelay);
      }
      else
      {
        PrintLog(LOGINPUTMNTR, "Checking door input disabled\n");
      }
    }
    else
    {
      g_controller.RTCScheduler().DisableAlarm(DOOR_ALARM_INDEX);

      //If the door is closed, and previously the Status_DoorStayedOpen has been
      //written with value 1, write it again with value 0 and contact the server,
      //to inform that the door has closed.
      if (ReadBackupFlag(DOOR_STATUS_FLAG))
      {
        PrintLog(LOGINPUTMNTR, "Door closed after staying open, contacting server\n");

        SetBackupFlag(DOOR_STATUS_FLAG, false);
        g_controller.Status().Write(Status_DoorStayedOpen, "0");

        g_controller.ServerComm().
          SignalCommAttempt(COMM_REASON_DOOR_CLOSED);
      }
    }
  }
}
