#include "inputs.h"
#include "inputmonitor.h"
#include "log.h"
#include "gpiodescriptions.h"
#include "lock.h"
#include "clock.h"
#include "backupregisters.h"
#include "controller.h"

#include <cassert>
#include <algorithm>

CInputs::CInputs()
{
  m_bootstate  = 0;
  m_pinchanged = false;
  m_inputstate = 0;
  m_mutex      = nullptr;
  m_signal     = nullptr;
  m_listeners  = 0;
}

CInputs::~CInputs()
{
  if (m_mutex)
    vSemaphoreDelete(m_mutex);

  if (m_signal)
    vEventGroupDelete(m_signal);
}

/*! I2C address of the input IO expander.*/
#define INPUT_IOEXPANDER_ADDRESS 64

/*! Sets up the IO expander to generate an interrupt on all inputs except the ST25R3911B IRQ.
    It will then generate a rising edge on WAKEUP(PA0) when one of the input changes.
    When a pin read is requested by a task, the GPIO register is only read if a
    pin interrupt on WAKEUP(PA0) has happened.
    \return true if setting up the IO expander succeeded.
    \return false if anything failed.*/
bool CInputs::Setup(CI2C* i2c)
{
  if (m_mutex == nullptr)
    m_mutex = xSemaphoreCreateRecursiveMutex();

  //create an event group so that each task waiting for an input to change
  //can be signaled using a single bit
  if (m_signal == nullptr)
    m_signal = xEventGroupCreate();
  m_listeners = 0;

  if (!m_ioexpander.Setup(i2c, INPUT_IOEXPANDER_ADDRESS, "input"))
    return false;

  //at boot, read which input pin has generated an interrupt,
  //this is used to check why the microcontroller came out of standby mode
  if (!m_ioexpander.GetInterruptCapture(&m_bootstate))
  {
    m_bootstate = 0;
    return false;
  }

  //reset the IO expander to its defaults, in case it's
  //somehow not configured correctly
  if (!m_ioexpander.ResetToDefaults())
    return false;

  //set the level of the IO expander INT pin to active high
  if (!m_ioexpander.SetOutputPolarity(true))
    return false;

  //enable the rising edge interrupt on WAKEUP(PA0), so that an interrupt
  //is triggered when the IO expander detects a pin change
  g_exti.EnableGPIOIrq(WAKEUP, EXTI_Trigger_Rising, this);

  //enable interrupt on change in the IO expander, so that it generates
  //an interrupt when any input pin except ST25R3911B IRQ changes
  if (!m_ioexpander.SetInterruptOnChange(INPUT_INTERRUPT))
    return false;
  
  //enable interrupt generation in the IO expander
  if (!m_ioexpander.EnableInterrupt(INPUT_INTERRUPT))
    return false;

  //read the GPIO register at least once to clear any pending interrupt state
  m_pinchanged = true;
  ReadSingle(INPUT_IN1);

  return true;
}

/*! Sets the IO expander in such a way that a rising edge on INPUT_ST25R3911B
    for low power card detect or a change on INPUT_FILL_LEVEL_RESET or
    INPUT_DOOR_SWITCH wakes up the microcontroller from low power mode using
    a rising edge on PA0.*/
bool CInputs::SetWakeupInterrupt()
{
  CLock lock(m_mutex);

  bool success = true;

  if (!m_ioexpander.DisableInterrupt(INPUT_ALL))
    success = false;

  //set the pin connected to the ST25R3911B NFC reader to generate an interrupt
  //when it's high
  if (!m_ioexpander.SetDefaultCompare(INPUT_ST25R3911B, false))
    success = false;

  //set the pin connected to the fill level reset switch to generate
  //an interrrupt when it's different from its last value read by the
  //input monitor task
  if (!m_ioexpander.SetDefaultCompare(INPUT_FILL_LEVEL_RESET,
                                      ReadBackupFlag(FILL_LEVEL_SWITCH_FLAG)))
    success = false;

  //set the pin connected to the door switch to generate
  //an interrrupt when it's different from its last value read by the
  //input monitor task
  if (!m_ioexpander.SetDefaultCompare(INPUT_DOOR_SWITCH,
                                      ReadBackupFlag(DOOR_SWITCH_FLAG)))
    success = false;

  //Set the pin connected to the 6D sensor to generate an interrupt when it's high.
  if (!m_ioexpander.SetDefaultCompare(INPUT_6DSENSOR, false))
    success = false;

  //Set the ST25R3911B, fill level reset, door switch and 6d sensor pins to
  //generate an interrupt when they are different from the default compare level.
  uint8_t irqinputs = INPUT_ST25R3911B | INPUT_FILL_LEVEL_RESET |
                      INPUT_DOOR_SWITCH | INPUT_6DSENSOR;
  if (!m_ioexpander.SetInterruptOnNotDefault(irqinputs))
    success = false;

  //always enable the fill level reset input to generate an interrupt,
  //only allow the ST25R3911B input to generate an interrupt if the reader is
  //working correctly
  uint8_t pins = INPUT_FILL_LEVEL_RESET;
  if (g_controller.Reader().IsWorking())
    pins |= INPUT_ST25R3911B;

  //Enable the door switch interrupt only when enabled, to prevent waking up
  //the microcontroller when checking the door switch is disabled.
  if (g_controller.InputMonitor().CheckDoorSwitch())
    pins |= INPUT_DOOR_SWITCH;

  //Enable the 6D sensor interrupt only when the sensor wakeup is enabled.
  if (g_controller.SixDSensor().Enabled())
    pins |= INPUT_6DSENSOR;

  if (!m_ioexpander.EnableInterrupt(pins))
    success = false;

  if (!success)
    PrintLog(LOGINPUTS, "ERROR: unable to configure IO expander for low power "
                        "wakeup interrupt\n");

  return success;
}

/*! Called on a rising edge on WAKEUP(PA0), signals that an input pin has changed.*/
void CInputs::GPIOIRQHandler(SGPIOPin pin)
{
  m_pinchanged = true;
  xEventGroupSetBitsFromISR(m_signal, m_listeners, nullptr);
}

/*! Reads the GPIO register from the IO expander if an interrupt on change was signaled.
    \param input The input to return the state of.
    \return true if the input is high.
    \return false if the input is low.*/
bool CInputs::ReadSingle(uint8_t input)
{
  CLock lock(m_mutex);

  if (m_pinchanged)
  {
    m_pinchanged = false;

    //reading the GPIO register clears the interrupt state in the IO expander,
    //this makes its INT pin low, after this it can detect another pin change
    if (!m_ioexpander.ReadGPIO(&m_inputstate))
    {
      PrintLog(LOGINPUTS, "Unable to read value of %s pin from input IO expander\n",
               InputStr(input));
    }
  }

  return m_inputstate & input;
}

/*! Reads the GPIO register from the IO expander if an interrupt on change was signaled.
    \param input Bitwise-OR of the inputs to read.
    \return Bitwise-AND of the GPIO register and the inputs argument.*/
uint8_t CInputs::ReadMultiple(uint8_t inputs)
{
  CLock lock(m_mutex);

  if (m_pinchanged)
  {
    m_pinchanged = false;

    //reading the GPIO register clears the interrupt state in the IO expander,
    //this makes its INT pin low, after this it can detect another pin change
    if (!m_ioexpander.ReadGPIO(&m_inputstate))
      PrintLog(LOGINPUTS, "Unable to read value of input pins from input IO expander\n");
  }

  return m_inputstate & inputs;
}

/*! Waits for an input to change state.
    \param inputs Bitwise-OR of the inputs to wait for.
    \param startvalues The initial values of the inputs, should have been read with ReadMultiple().
    \param timeout Timeout in milliseconds waiting for the input to change.
    \return Bitwise-OR of the current values of the inputs.*/
uint8_t CInputs::WaitForChange(uint8_t inputs, uint8_t startvalues, uint32_t timeout)
{
  //each bit in m_listeners represents a task that is waiting for a change
  //reserve a bit in m_listeners for the current task
  CCritical criticalsection;
  EventBits_t listenerbit = 0;
  for (EventBits_t i = 0; i < sizeof(m_listeners) * 8; i++)
  {
    EventBits_t bit = 1 << i;
    if ((bit & m_listeners) == 0)
    { //reserve the first bit in m_listeners that is 0 for this task
      listenerbit = bit;
      m_listeners |= listenerbit;
      break;
    }
  }
  criticalsection.Leave();
  assert(listenerbit); //make sure that one bit has been reserved

  //wait for the input to change
  int64_t start   = Millis();
  int64_t elapsed = 0;
  uint8_t values  = ReadMultiple(inputs); //the inputs might have changed already

  startvalues &= inputs;

  while (values == startvalues && elapsed < timeout)
  {
    elapsed = Millis() - start;
    if (elapsed < timeout)
      xEventGroupWaitBits(m_signal, listenerbit, pdTRUE, pdFALSE, timeout - elapsed);

    values = ReadMultiple(inputs);
  }

  //remove this task from the listeners
  criticalsection.Enter();
  m_listeners &= ~listenerbit;
  criticalsection.Leave();

  return values;
}

/*! Checks if one or more inputs is stable for some period.
    \param inputs The inputs to check.
    \param stabletime Number of milliseconds that an input should not change before it's considered stable.
    \param timeout Number of milliseconds before giving up if an input does not become stable.
    \param values A bitwise-OR of the inputs values is returned here.
    \return true if the input is considered stable.
    \return false if the input has not been stable within the timeout.*/
bool CInputs::ReadStable(uint8_t inputs, uint32_t stabletime, uint32_t timeout, uint8_t* values)
{
  int64_t start = Millis();
  while (Millis() - start < timeout)
  {
    //read the state of the input
    int64_t readstart   = Millis();
    uint8_t startlevels = ReadMultiple(inputs);
    uint8_t endlevels   = WaitForChange(inputs, startlevels, stabletime);
    int64_t duration    = Millis() - readstart;

    //if the level of the input stays the same for some time,
    //assume the input is stable
    if (startlevels == endlevels && duration >= stabletime - 1)
    {
      *values = startlevels;
      return true;
    }
    else
    {
      PrintLog(LOGINPUTS, "ERROR: %" PRIX8 " input unstable, start:%" PRIX8 " end:%" PRIX8 " duration:%"
               PRIi64 "\n", inputs, startlevels, endlevels, duration);
    }
  }

  PrintLog(LOGINPUTS, "ERROR: Tested %" PRIX8 " input for %" PRIi64 " milliseconds, "
           " without reading a stable state for %" PRIu32 " milliseconds, giving up\n",
           inputs, Millis() - start, stabletime);

  return false;
}

/*! Waits until all selected inputs are at the wanted level for the selected time.
    \param inputs Bitwise-OR of inputs to check.
    \param stabletime How long the inputs should be stable for.
    \param timeout How long to check the inputs before giving up.
    \param wantedlevels Bitwise-OR of wanted input states, return if the inputs are at this level for stabletime milliseconds.
    \param readlevels When not nullptr, the latest read levels of the inputs are returned here.
    \return true if the inputs are at wantedlevels for stabletime milliseconds.
    \return false if timeout milliseconds elapsed without the inputs being at wantedlevels for stabletime milliseconds.*/
bool CInputs::WaitForStableLevel(uint8_t inputs, uint32_t stabletime, uint32_t timeout, uint8_t wantedlevels, uint8_t* readlevels /*= nullptr*/)
{
  int64_t start = Millis();

  //read the state of the inputs
  int64_t readstart   = Millis();
  uint8_t startlevels = ReadMultiple(inputs);

  if (readlevels)
    *readlevels = startlevels;

  wantedlevels &= inputs;

  do
  {
    //if the inputs are already at the wanted level, make sure that they don't
    //change for stabletime milliseconds, if they're not at the wanted level,
    //wait for them to change up to the timeout
    int64_t elapsed  = Millis() - start;
    int64_t waittime = stabletime;
    if (startlevels != wantedlevels)
      waittime = std::max((int64_t)timeout - elapsed, (int64_t)1);

    uint8_t endlevels = WaitForChange(inputs, startlevels, waittime);
    int64_t duration  = Millis() - readstart;

    if (readlevels)
      *readlevels = endlevels;

    //if the levels did not change for at least stabletime milliseconds,
    //and the levels are at the wanted levels, consider the inputs stable
    if (startlevels == endlevels &&
        endlevels   == wantedlevels &&
        duration    >= stabletime - 1)
      return true;

    startlevels = endlevels;
    readstart += duration;
  }
  while (Millis() - start < timeout);

  return false;
}

/*! Waits for one of the selected inputs to stay at a certain level for some time,
    while also guaranteeing that the other selected inputs are also at a stable level.
    \param inputs Bitwise-OR of inputs to check.
    \param stabletime How long the input should be stable for.
    \param timeout How long to check the inputs before giving up.
    \param wantedlevels Bitwise-OR of wanted state of inputs.
    \param readlevels When not nullptr, the latest read levels of the inputs are returned here.
    \return true if one of the inputs is the same as its bit in wantedlevels for stabletime milliseconds.
    \return false if timeout milliseconds elapsed without one of the inputs being at its bit in wantedlevels for stabletime milliseconds.*/
bool CInputs::WaitForStableSignal(uint8_t inputs, uint32_t stabletime, uint32_t timeout, uint8_t wantedlevels, uint8_t* readlevels /*= nullptr*/)
{
  int64_t start = Millis();

  //Read the state of the inputs.
  int64_t readstart   = Millis();
  uint8_t startlevels = ReadMultiple(inputs);

  if (readlevels)
    *readlevels = startlevels;

  do
  {
    //Check if one of the inputs is already at the wanted level.
    bool atwantedlevel = false;
    for (uint32_t i = 0; i < NUM_INPUTS; i++)
    {
      uint8_t inputmask = 1 << i;

      if (!(inputs & inputmask))
        continue; //Input not selected.

      if ((startlevels & inputmask) == (wantedlevels & inputmask))
      {
        atwantedlevel = true;
        break;
      }
    }

    //If one of the selected inputs is already at the wanted level, make sure
    //that all selected inputs stay stable for stabletime milliseconds.
    //If no input is at the wanted level, wait for the inputs to change
    //up to the timeout.
    int64_t elapsed  = Millis() - start;
    int64_t waittime = stabletime;
    if (!atwantedlevel)
      waittime = std::max((int64_t)timeout - elapsed, (int64_t)1);

    uint8_t endlevels = WaitForChange(inputs, startlevels, waittime);
    int64_t duration  = Millis() - readstart;

    if (readlevels)
      *readlevels = endlevels;

    //If one of the inputs was already at the wanted level, and all inputs
    //stayed the same for stabletime milliseconds, regard that a stable signal
    //has been presented at the input.
    if (atwantedlevel &&
        startlevels == endlevels &&
        duration    >= stabletime - 1)
      return true;

    startlevels = endlevels;
    readstart += duration;
  }
  while (Millis() - start < timeout);

  return false;
}

/*! Converts an input pin from the input IO expander to a printable string.
    \param input The input pin to convert to a printable string.
    \return A pointer to a string literal.*/
const char* CInputs::InputStr(uint8_t input)
{
  if (input == INPUT_TOUCH1)
    return "touch1";
  else if (input == INPUT_TOUCH2)
    return "touch2";
  else if (input == INPUT_TOUCH3)
    return "touch3";
  else if (input == INPUT_IN1)
    return "in1";
  else if (input == INPUT_IN2)
    return "in2";
  else if (input == INPUT_IN3)
    return "in3";
  else if (input == INPUT_IN4)
    return "in4";
  else if (input == INPUT_ST25R3911B)
    return "ST25R3911B";
  else
    return "UNKNOWN";
}

const char* CInputs::StateStr(bool state)
{
  if (state)
    return "high";
  else
    return "low";
}
