#include "lockdriver.h"
#include "lock.h"
#include "log.h"
#include "clock.h"
#include "gpiodescriptions.h"
#include "settings.h"
#include "controller.h"
#include "backupregisters.h"

#include <stm32f4xx.h>
#include <algorithm>

CLockDriver::CLockDriver()
{
  m_locktype        = LockType_None;
  m_servo1timeopen  = 0;
  m_servo1timeclose = 0;
  m_servo2timeopen  = 0;
  m_servo2timeclose = 0;
  m_lockclosemoment = 0;
  m_powered         = false;
  m_interval        = 0;
  m_powerrail       = -1.0f;
}

/*! Returns if the lock is opened or closed, based on the battery backed ram flag.*/
bool CLockDriver::Opened()
{
  return ReadBackupFlag(LOCK_OPENED_FLAG);
}

/*! Used to clear the opened state on monostable lock types.*/
void CLockDriver::ClearOpenedState()
{
  SetBackupFlag(LOCK_OPENED_FLAG, false);
}

/*! \return true if the lock type operates as a bistable lock, it will keep
                 its last position without power.
    \return false if the lock type operates as a monostable lock, it will always
                  return to the closed state when it loses power.*/
bool CLockDriver::IsBiStable()
{
  SetType();

  return m_locktype == LockType_Motor           ||
         m_locktype == LockType_Bistable_2      ||
         m_locktype == LockType_ServoMotor      ||
         m_locktype == LockType_Bistable_1      ||
         m_locktype == LockType_30_60_Liter     ||
         m_locktype == LockType_HoldLock;
}

/*! \return true If the lock type has a release method.
    \return false If the lock type does not have a release method.*/
bool CLockDriver::HasRelease()
{
  return LockType() == LockType_HoldLock;
}

/*! Reads the lock type from the settings if not already done.
    \return A value of \ref ElockType.*/
ELockType CLockDriver::LockType()
{
  if (m_locktype == LockType_None)
    SetType();

  return m_locktype;
}

/*! Reads the lock type from the settings, and stores it in m_locktype.*/
void CLockDriver::SetType()
{
  ELockType locktype = (ELockType)g_controller.Settings().GetInteger(Setting_AccelloType);
  if (locktype == m_locktype)
    return;

  m_locktype = locktype;

  if (m_locktype >= LockType_Start && m_locktype < LockType_End)
  {
    PrintLog(LOGLOCK, "Lock type set to %i: %s\n", (int)m_locktype,
             LockTypeStr(m_locktype));
  }
  else
  {
    PrintLog(LOGLOCK, "ERROR: invalid lock type %i\n", (int)m_locktype);
    m_locktype = LockType_None;
  }
}

/*! Converts a value of \ref ELockType to a printable string.
    \param locktype A value of \ref ELockType, read from settings.
    \return A pointer to a string literal.*/
const char* CLockDriver::LockTypeStr(ELockType locktype)
{
  if (locktype == LockType_Solenoid)
    return "solenoid";
  else if (locktype == LockType_Solenoid_Toggle)
    return "solenoid toggle";
  else if (locktype == LockType_Motor)
    return "motor";
  else if (locktype == LockType_Bistable_2)
    return "bistable 2";
  else if (locktype == LockType_ServoMotor)
    return "servo";
  else if (locktype == LockType_Bistable_1)
    return "bistable 1";
  else if (locktype == LockType_Press_Engels)
    return "press Engels";
  else if (locktype == LockType_HoldLock)
    return "hold lock";
  else if (locktype == LockType_30_60_Liter)
    return "30/60 liter";
  else
    return "ERROR UNKNOWN";
}

void CLockDriver::ReadServoSettings()
{
  m_servo1timeopen  = g_controller.Settings().GetInteger(Setting_Servomotor1TimeOpen);
  m_servo1timeclose = g_controller.Settings().GetInteger(Setting_Servomotor1TimeClose);

  m_servo2timeopen  = g_controller.Settings().GetInteger(Setting_Servomotor2TimeOpen);
  m_servo2timeclose = g_controller.Settings().GetInteger(Setting_Servomotor2TimeClose);
}

void CLockDriver::PowerUp()
{
  if (m_powered)
    return;

  SetType();

  PrintLog(LOGLOCK, "Powering up lock drive circuit\n");

  //Enable the boost converter to power the lock.
  PowerUpBoostConverter();

  //Measure the voltage of the power rail after the boost converter has been powered.
  m_powerrail = MeasurePowerRail();
  PrintLog(LOGLOCK, "Power rail: %.3f volt\n", m_powerrail);

  if (m_locktype == LockType_ServoMotor)
    PowerUpServoLock();
  else if (m_locktype == LockType_Solenoid)
    PowerUpSolenoidLock();
  else if (m_locktype == LockType_Bistable_1 || m_locktype == LockType_Bistable_2)
    PowerUpBistableLock();
  else if (m_locktype == LockType_30_60_Liter)
    PowerUp3060LiterLock();
  else if (m_locktype == LockType_HoldLock)
    PowerUpHoldLock();
  else
    PrintLog(LOGLOCK, "ERROR: lock type %i:%s not supported\n",
             (int)m_locktype, LockTypeStr(m_locktype));

  m_powered = true;
}

/*! Measures the power rail and the microcontroller reference voltage using the ADC.
    \return The voltage of the power rail, in volts.*/
float CLockDriver::MeasurePowerRail()
{
  //The power rail is measured via a voltage divider made from two 12k resistors
  //and one 3k9 resistor. The ADC measures using the microcontroller's power supply
  //as a reference. The measured value is scaled to represent the power rail voltage.*/
  float powerrail = g_controller.Adc().Read(ADC_Channel_8);
  float vref      = g_controller.Adc().ReadVrefint();
  powerrail       = powerrail * vref * (12.0f + 12.0f + 3.9f) / 3.9f;

  return powerrail;
}

/*! Calls the power down function that corresponds with the lock type.*/
void CLockDriver::PowerDown()
{
  if (!m_powered)
    return;

  PrintLog(LOGLOCK, "Powering down lock drive circuit\n");

  if (m_locktype == LockType_ServoMotor)
    PowerDownServoLock();
  else if (m_locktype == LockType_Solenoid)
    PowerDownSolenoidLock();
  else if (m_locktype == LockType_Bistable_1 || m_locktype == LockType_Bistable_2)
    PowerDownBistableLock();
  else if (m_locktype == LockType_30_60_Liter)
    PowerDown3060LiterLock();
  else if (m_locktype == LockType_HoldLock)
    PowerDownHoldLock();
  else
    PrintLog(LOGLOCK, "ERROR: lock type %i:%s not supported\n",
             (int)m_locktype, LockTypeStr(m_locktype));

  PowerDownBoostConverter();

  m_powered = false;
}

/*! Powers up the 5v boost converter.*/
void CLockDriver::PowerUpBoostConverter()
{
  //Force a battery telemetry read before the boost converter is powered on,
  //so that the voltage drop on the battery doesn't cause the charger to
  //start charging the battery.
  g_controller.BatteryCharger().SetForceTelemetry(true, false, true);
  DelayMS(20);

  //Since PCB version 1.6, the capacitors of the 5 volt boost converter should
  //be precharged by making pin P5V0_WARMUP high, this prevents a large dip in
  //the battery voltage when the boost converter is enabled. Enabling this pin on
  //previous PCB versions does nothing.
  GPIO_Set(P5V0_WARMUP);
  DelayMS(25);

  //Turn off the precharge and enable the boost converter as quickly as possible.
  CCritical criticalsection;
  GPIO_Reset(P5V0_WARMUP);
  GPIO_Set(P5V0_ENABLE);
  criticalsection.Leave();

  DelayMS(175);
}

/*! Powers down the 5v boost converter.*/
void CLockDriver::PowerDownBoostConverter()
{
  GPIO_Reset(P5V0_ENABLE);
}

/*! Reads the lock type and calls the corresponding lock open function.*/
void CLockDriver::Open()
{
  if (Opened())
  {
    PrintLog(LOGLOCK, "ERROR: lock already opened\n");
    return;
  }

  //Set the lock opened flag in battery backed ram before opening the lock,
  //so that if anything goes wrong when opening the lock, it will be closed
  //at the next boot.
  SetBackupFlag(LOCK_OPENED_FLAG, true);

  //Force a battery telemetry read before the lock is actuated, so that the
  //voltage drop on the battery doesn't cause the charger to start charging
  //the battery.
  g_controller.BatteryCharger().SetForceTelemetry(true, false, true);
  DelayMS(20);

  //Make sure the lock is powered.
  PowerUp();

  if (m_locktype == LockType_ServoMotor)
    OpenServoLock();
  else if (m_locktype == LockType_Solenoid)
    OpenSolenoidLock();
  else if (m_locktype == LockType_Bistable_1 || m_locktype == LockType_Bistable_2)
    OpenBistableLock();
  else if (m_locktype == LockType_30_60_Liter)
    Open3060LiterLock();
  else if (m_locktype == LockType_HoldLock)
    OpenHoldLock();
  else
    PrintLog(LOGLOCK, "ERROR: lock type %i:%s not supported\n",
             (int)m_locktype, LockTypeStr(m_locktype));
}

/*! Calls the lock close function that corresponds with the lock type. */
void CLockDriver::Close()
{
  if (!Opened())
  {
    PrintLog(LOGLOCK, "ERROR: lock already closed\n");
    return;
  }

  //Clear the lock opened flag before closing the lock, so that if anything
  //goes wrong, no reboot loop is created trying to close the lock at boot.
  SetBackupFlag(LOCK_OPENED_FLAG, false);

  //Force a battery telemetry read before the lock is actuated, so that the
  //voltage drop on the battery doesn't cause the charger to start charging
  //the battery.
  g_controller.BatteryCharger().SetForceTelemetry(true, false, true);
  DelayMS(20);

  //Make sure the lock is powered.
  PowerUp();

  if (m_locktype == LockType_ServoMotor)
    CloseServoLock();
  else if (m_locktype == LockType_Solenoid)
    CloseSolenoidLock();
  else if (m_locktype == LockType_Bistable_1 || m_locktype == LockType_Bistable_2)
    CloseBistableLock();
  else if (m_locktype == LockType_30_60_Liter)
    Close3060LiterLock();
  else if (m_locktype == LockType_HoldLock)
    CloseHoldLock();
  else
    PrintLog(LOGLOCK, "ERROR: lock type %i:%s not supported\n",
             (int)m_locktype, LockTypeStr(m_locktype));

  //If m_powerrail has been measured before, store it as a status, in millivolts.
  //This can't be stored immediately after measuring, because it takes
  //too long to initialize the status ringbuffer.
  if (m_powerrail != -1.0f)
  {
    g_controller.Status().Write(Status_PWRVoltageNoLoad, "%.0f",
                                m_powerrail * 1000.0f);
    m_powerrail = -1.0f;
  }
}

/*! When a hold magnet is present, powers the hold magnet for a set amount of time. */
void CLockDriver::ReleaseHold()
{
  if (m_locktype == LockType_HoldLock)
  {
    //Force a battery telemetry read before the magnet is powered, so that the
    //voltage drop on the battery doesn't cause the charger to start charging
    //the battery.
    g_controller.BatteryCharger().SetForceTelemetry(true, false, true);
    DelayMS(20);

    //Power the permanent electromagnet, to release its hold on the deposit lid.
    ReleaseHoldLock();
  }
  else
  {
    PrintLog(LOGLOCK, "ERROR: ReleaseHold called but lock type is %i:%s\n",
             (int)m_locktype, LockTypeStr());
  }
}

//! Time in milliseconds waiting for the servo to open the lock.
#define SERVO_ACTUATE_OPEN_TIME  500

//! Time in milliseconds waiting for the servo to close the lock.
#define SERVO_ACTUATE_CLOSE_TIME 1000

/*! Minimum time in milliseconds between closing the lock and powering down the servo.*/
#define SERVO_POWER_DOWN_TIME    1500

/*! Interval in microseconds the servo is driven at.*/
#define SERVO_INTERVAL           20000

/*! Enables the H-bridges and timer so that a servo lock can be driven.*/
void CLockDriver::PowerUpServoLock()
{
  //Enable the timer so that it can generate a servo timing signal.
  EnableTimer(SERVO_INTERVAL);

  //enable both H-bridges
  GPIO_Set(B1_EN);
  GPIO_Set(B2_EN);

  //The timings for the servo signal are controlled by the settings.
  ReadServoSettings();
}

/*! Sends a signal to open the servo lock.*/
void CLockDriver::OpenServoLock()
{
  PrintLog(LOGLOCK, "Opening servo lock, servo time %" PRIu32 "\n", m_servo1timeopen);

  //Set the timer to generate a servo signal that opens the lock.
  TIM_SetCompare1(TIM8, m_servo1timeopen);

  //Enable power to the servo, by forcing timer 8 channel 4 to active mode.
  //This makes PC9/B2_IN2 high, which drives BRIDGE3 on CON2 high.
  TIM_ForcedOC4Config(TIM8, TIM_ForcedAction_Active);

  //Give the servo some time to open the lock.
  DelayMS(SERVO_ACTUATE_OPEN_TIME);

  PrintLog(LOGLOCK, "Servo lock opened\n");
}

/*! Sends a signal to close the lock, and disables both H-bridges.
    WARNING: does not disable the internal boost converter.*/
void CLockDriver::CloseServoLock()
{
  PrintLog(LOGLOCK, "Closing servo lock, servo time %" PRIu32 "\n", m_servo1timeclose);

  //Make sure the servo has power.
  TIM_ForcedOC4Config(TIM8, TIM_ForcedAction_Active);

  //Set the timer to generate a servo signal that closes the lock.
  TIM_SetCompare1(TIM8, m_servo1timeclose);

  m_lockclosemoment = Millis();

  //give the servo some time to close the lock
  DelayMS(SERVO_ACTUATE_CLOSE_TIME);

  PrintLog(LOGLOCK, "Servo lock closed\n");
}

/*! Disables servo drive, should be called after CloseLock().
    Checks how long ago the call to CloseLock() was, will delay if it's
    less than SERVO_POWER_DOWN_TIME milliseconds.*/
void CLockDriver::PowerDownServoLock()
{
  //Give the lock time to fully close before powering down.
  int64_t elapsed = Millis() - m_lockclosemoment;
  if (elapsed > 0 && elapsed < SERVO_POWER_DOWN_TIME)
    DelayMS(SERVO_POWER_DOWN_TIME - elapsed);

  //Disable power to the servo, by forcing timer 8 channel 4 to inactive mode.
  //This makes PC9/B2_IN2 low, which drives BRIDGE3 on CON2 low.
  TIM_ForcedOC4Config(TIM8, TIM_ForcedAction_InActive);

  //Disable both H-bridges.
  GPIO_Reset(B1_EN);
  GPIO_Reset(B2_EN);

  DisableTimer();
}

//! Time in milliseconds waiting for the solenoid to open the lock.
#define SOLENOID_ACTUATE_OPEN_TIME  500

//! Time in milliseconds waiting for the solenoid to close the lock.
#define SOLENOID_ACTUATE_CLOSE_TIME 500

/*! Interval in microseconds the solenoid is driven at.*/
#define SOLENOID_INTERVAL           10000

/*! Percentage of PWM that the solenoid is driven at after SOLENOID_ACTUATE_OPEN_TIME has passed.*/
#define SOLENOID_PWM                20

/*! Powers up the H-bridges so that a solenoid lock can be driven.*/
void CLockDriver::PowerUpSolenoidLock()
{
  //Set the timer so that it can drive the solenoid with PWM.
  EnableTimer(SOLENOID_INTERVAL);

  //enable both H-bridges
  GPIO_Set(B1_EN);
  GPIO_Set(B2_EN);
}

/*! Drives the H-bridges so that it will open a solenoid lock.*/
void CLockDriver::OpenSolenoidLock()
{
  PrintLog(LOGLOCK, "Opening solenoid lock\n");

  //Set the timer to drive the BRIDGE3 output with 100% PWM
  TIM_SetCompare4(TIM8, SOLENOID_INTERVAL);

  //Give the solenoid some time to open the lock.
  DelayMS(SOLENOID_ACTUATE_OPEN_TIME);

  //Drive the solenoid with PWM to save battery power.
  TIM_SetCompare4(TIM8, SOLENOID_INTERVAL * SOLENOID_PWM / 100);

  PrintLog(LOGLOCK, "Solenoid lock opened\n");
}

/*! Makes the output of both H-bridges low so that the solenoid lock closes.*/
void CLockDriver::CloseSolenoidLock()
{
  PrintLog(LOGLOCK, "Closing solenoid lock\n");

  //Make the bridge output low, this will create a short across the solenoid's
  //terminals using the H-bridge outputs.
  TIM_SetCompare4(TIM8, 0);

  //give the solenoid some time to close the lock
  DelayMS(SOLENOID_ACTUATE_CLOSE_TIME);

  PrintLog(LOGLOCK, "Solenoid Lock closed\n");
}

/*! Disables power to the solenoid lock and H-bridges.*/
void CLockDriver::PowerDownSolenoidLock()
{
  //disable both H-bridges
  GPIO_Reset(B1_EN);
  GPIO_Reset(B2_EN);

  DisableTimer();
}

//! Time in milliseconds waiting for the bistable solenoid to open the lock.
#define BISTABLE_ACTUATE_OPEN_TIME  500

//! Time in milliseconds waiting for the bistable solenoid to close the lock.
#define BISTABLE_ACTUATE_CLOSE_TIME 500

/*! Interval in microseconds the solenoid is driven at.*/
#define BISTABLE_INTERVAL           10000

/*! Sets up the H-bridges so that it can drive a bistable lock.*/
void CLockDriver::PowerUpBistableLock()
{
  //Enable the timer so that it can drive a bistable solenoid.
  EnableTimer(BISTABLE_INTERVAL);

  //enable both H-bridges
  GPIO_Set(B1_EN);
  GPIO_Set(B2_EN);
}

/*! Delivers a pulse to open a bistable solenoid lock.*/
void CLockDriver::OpenBistableLock()
{
  PrintLog(LOGLOCK, "Opening bistable solenoid lock\n");

  //Make PC9/B2_IN2 high, which drives BRIDGE3 on CON2 high.
  //This opens the bistable solenoid lock.
  TIM_ForcedOC4Config(TIM8, TIM_ForcedAction_Active);

  //give the lock some time to open
  DelayMS(BISTABLE_ACTUATE_OPEN_TIME);

  //Make the drive pin low, the bistable lock will hold its last position.
  TIM_ForcedOC4Config(TIM8, TIM_ForcedAction_InActive);

  PrintLog(LOGLOCK, "Bistable solenoid lock opened\n");
}

/*! Delivers a pulse to close a bistable solenoid lock.*/
void CLockDriver::CloseBistableLock()
{
  PrintLog(LOGLOCK, "Closing bistable solenoid lock\n");

  //Make PC6/B1_IN1 high, which drives BRIDGE2 on CON2 high.
  //This closes the bistable solenoid lock.
  TIM_ForcedOC1Config(TIM8, TIM_ForcedAction_Active);

  //Give the lock some time to close.
  DelayMS(BISTABLE_ACTUATE_CLOSE_TIME);

  //Disable the drive pin after the lock has closed
  TIM_ForcedOC1Config(TIM8, TIM_ForcedAction_InActive);

  PrintLog(LOGLOCK, "Bistable solenoid lock closed\n");
}

/*! Powers down drive circuitry needed for a bistable solenoid lock.*/
void CLockDriver::PowerDownBistableLock()
{
  //disable both H-bridges
  GPIO_Reset(B1_EN);
  GPIO_Reset(B2_EN);

  DisableTimer();
}

/*! Sets up the H-bridges so that it can drive two servo locks.*/
void CLockDriver::PowerUp3060LiterLock()
{
  //Enable the timer so that it can generate a servo timing signal.
  EnableTimer(SERVO_INTERVAL);

  //enable both H-bridges
  GPIO_Set(B1_EN);
  GPIO_Set(B2_EN);

  //the timings for the servo signal are controlled by the settings
  ReadServoSettings();
}

//! Time in milliseconds waiting for each servo to open the lock.
#define DUAL_SERVO_ACTUATE_OPEN_TIME  250

//! Time in milliseconds waiting for each servo to close the lock.
#define DUAL_SERVO_ACTUATE_CLOSE_TIME 250

/*! Sends a signal to open two servo locks.*/
void CLockDriver::Open3060LiterLock()
{
  PrintLog(LOGLOCK, "Opening 30/60 liter lock, servo1 time:%" PRIu32
           ", servo2 time:%" PRIu32 "\n", m_servo1timeopen, m_servo2timeopen);

  //Set the timer to generate a servo signal that opens the lock.
  TIM_SetCompare1(TIM8, m_servo1timeopen);
  TIM_SetCompare2(TIM8, m_servo2timeopen);

  //Powering up two servos at the same time draws too much current,
  //so one is powered after the other.
  //Enable power to the first servo, by forcing timer 8 channel 4 to active mode.
  //This makes PC9/B2_IN2 high, which drives BRIDGE3 on CON2 high.
  TIM_ForcedOC4Config(TIM8, TIM_ForcedAction_Active);
  DelayMS(DUAL_SERVO_ACTUATE_OPEN_TIME);

  //Enable power to the second servo, by forcing timer 8 channel 3 to active mode.
  //This makes PC8/B2_IN1 high, which drives BRIDGE4 on CON2 high.
  TIM_ForcedOC3Config(TIM8, TIM_ForcedAction_Active);
  DelayMS(DUAL_SERVO_ACTUATE_OPEN_TIME);

  PrintLog(LOGLOCK, "30/60 liter lock opened\n");
}

/*! Sends a signal to close two servo locks.*/
void CLockDriver::Close3060LiterLock()
{
  PrintLog(LOGLOCK, "Closing 30/60 liter lock, servo1 time:%" PRIu32
           ", servo2 time:%" PRIu32 "\n", m_servo1timeclose, m_servo2timeclose);

  //Set the timer to generate a servo signal that closes the lock.
  //Only one servo at a time should be turning, otherwise the current
  //draw from the boost converter is too high.
  TIM_ForcedOC4Config(TIM8, TIM_ForcedAction_Active);
  TIM_SetCompare1(TIM8, m_servo1timeclose);
  DelayMS(DUAL_SERVO_ACTUATE_CLOSE_TIME);

  TIM_ForcedOC3Config(TIM8, TIM_ForcedAction_Active);
  TIM_SetCompare2(TIM8, m_servo2timeclose);
  DelayMS(DUAL_SERVO_ACTUATE_CLOSE_TIME);

  PrintLog(LOGLOCK, "30/60 liter lock closed\n");
}

/*! Disables servo drive, should be called after CloseLock().
    Checks how long ago the call to CloseLock() was, will delay if it's
    less than SERVO_POWER_DOWN_TIME milliseconds.*/
void CLockDriver::PowerDown3060LiterLock()
{
  //give the lock time to fully close before powering down
  int64_t elapsed = Millis() - m_lockclosemoment;
  if (elapsed > 0 && elapsed < SERVO_POWER_DOWN_TIME)
    DelayMS(SERVO_POWER_DOWN_TIME - elapsed);

  //Disable power to the servos, by forcing timer 8 channels 3 and 4 to inactive mode.
  //This makes PC8/B2_IN1 and PC9/B2_IN2 low, which drive BRIDGE4 and BRIDGE3 on CON2 low.
  TIM_ForcedOC3Config(TIM8, TIM_ForcedAction_InActive);
  TIM_ForcedOC4Config(TIM8, TIM_ForcedAction_InActive);

  //disable both H-bridges
  GPIO_Reset(B1_EN);
  GPIO_Reset(B2_EN);

  DisableTimer();
}

/*! Sets up the H-bridges so that it can drive a servo lock and
    power a hold magnet.*/
void CLockDriver::PowerUpHoldLock()
{
  //Enable the timer so that it can generate a servo timing signal.
  EnableTimer(SERVO_INTERVAL);

  //Enable both H-bridges.
  GPIO_Set(B1_EN);
  GPIO_Set(B2_EN);

  //The timings for the servo signal are controlled by the settings.
  ReadServoSettings();
}

/*! Sends a signal to open a hold lock. */
void CLockDriver::OpenHoldLock()
{
  PrintLog(LOGLOCK, "Opening hold lock, servo time:%" PRIu32 "\n",
           m_servo1timeopen);

  //Set the timer to generate a servo signal that opens the lock.
  TIM_SetCompare1(TIM8, m_servo1timeopen);

  //Enable power to the servo, by forcing timer 8 channel 4 to active mode.
  //This makes PC9/B2_IN2 high, which drives BRIDGE3 on CON2 high.
  TIM_ForcedOC4Config(TIM8, TIM_ForcedAction_Active);

  //Give the servo some time to open the lock.
  DelayMS(SERVO_ACTUATE_OPEN_TIME);

  PrintLog(LOGLOCK, "Hold lock opened\n");
}

/*! Sends a signal that closes a hold lock.*/
void CLockDriver::CloseHoldLock()
{
  PrintLog(LOGLOCK, "Closing hold lock, servo time %" PRIu32 "\n", m_servo1timeclose);

  //Make sure the servo has power.
  TIM_ForcedOC4Config(TIM8, TIM_ForcedAction_Active);

  //Set the timer to generate a servo signal that closes the lock.
  TIM_SetCompare1(TIM8, m_servo1timeclose);

  m_lockclosemoment = Millis();

  //Give the servo some time to close the lock.
  DelayMS(SERVO_ACTUATE_CLOSE_TIME);

  PrintLog(LOGLOCK, "Hold lock closed\n");
}

/*! Time in milliseconds to drive the H-bridge output low after the permanent
    electromagnet has been powered, to clamp the inductive current.*/
#define RELEASE_CLAMP_TIME 100

/*! Releases the hold magnet.*/
void CLockDriver::ReleaseHoldLock()
{
  uint32_t releasetime = g_controller.Settings().GetInteger(Setting_HoldLockRelease);
  releasetime = std::max(std::min(releasetime, (uint32_t)300000), (uint32_t)100);

  PrintLog(LOGLOCK, "Releasing hold magnet for %" PRIu32 " milliseconds\n",
           releasetime);

  //Toggle channel 3 on timer 8, this toggles PC8/B2_IN1,
  //which drives BRIDGE4 high and low.
  //This powers the permanent electromagnet, which releases the deposit lid.
  TIM_ForcedOC3Config(TIM8, TIM_ForcedAction_Active);
  DelayMS(releasetime);
  TIM_ForcedOC3Config(TIM8, TIM_ForcedAction_InActive);
  DelayMS(RELEASE_CLAMP_TIME); //Clamp inductive current.

  PrintLog(LOGLOCK, "Hold magnet released\n");
}

/*! Disables servo drive, should be called after CloseLock().
    Checks how long ago the call to CloseLock() was, will delay if it's
    less than SERVO_POWER_DOWN_TIME milliseconds.*/
void CLockDriver::PowerDownHoldLock()
{
  //Give the lock time to fully close before powering down.
  int64_t elapsed = Millis() - m_lockclosemoment;
  if (elapsed > 0 && elapsed < SERVO_POWER_DOWN_TIME)
    DelayMS(SERVO_POWER_DOWN_TIME - elapsed);

  //Disable power to the servos, by forcing timer 8 channels 3 and 4 to inactive mode.
  //This makes PC8/B2_IN1 and PC9/B2_IN2 low, which drive BRIDGE4 and BRIDGE3 on CON2 low.
  TIM_ForcedOC3Config(TIM8, TIM_ForcedAction_InActive);
  TIM_ForcedOC4Config(TIM8, TIM_ForcedAction_InActive);

  //Disable both H-bridges.
  GPIO_Reset(B1_EN);
  GPIO_Reset(B2_EN);

  DisableTimer();
}

/*! Sets up timer 8 to drive channel 1 using output compare.
    This channel is connected to the BRIDGE2 output, which is pin 5 on J3.
    \param interval The timer interval in microseconds.*/
void CLockDriver::EnableTimer(uint32_t interval)
{
  m_interval = interval;

  //use a critical section to protect the RCC and timer registers
  CCritical criticalsection;

  //get the clock frequency of the APB2, which is what clocks timer8,
  RCC_ClocksTypeDef clockfreqs;
  RCC_GetClocksFreq(&clockfreqs);

  //enable the clock for timer8
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_TIM8, ENABLE);

  //reset timer8 to its default state
  RCC_APB2PeriphResetCmd(RCC_APB2Periph_TIM8, ENABLE);
  RCC_APB2PeriphResetCmd(RCC_APB2Periph_TIM8, DISABLE);

  //initialize timer8 as an upcounter clock at 1 tick per microsecond,
  TIM_TimeBaseInitTypeDef timinit = {};
  timinit.TIM_Prescaler           = (clockfreqs.SYSCLK_Frequency / 1000000) - 1;
  timinit.TIM_CounterMode         = TIM_CounterMode_Up;
  timinit.TIM_Period              = interval;
  timinit.TIM_ClockDivision       = TIM_CKD_DIV1;
  TIM_TimeBaseInit(TIM8, &timinit);

  //set outputs 1, 2, 3 and 4 of timer8 to PWM mode
  TIM_OCInitTypeDef ocinit = {};
  ocinit.TIM_OCMode        = TIM_OCMode_PWM1;
  ocinit.TIM_OutputState   = TIM_OutputState_Enable;
  ocinit.TIM_OutputNState  = TIM_OutputNState_Disable;
  ocinit.TIM_Pulse         = 0;
  ocinit.TIM_OCPolarity    = TIM_OCPolarity_High;
  ocinit.TIM_OCNPolarity   = TIM_OCNPolarity_Low;
  ocinit.TIM_OCIdleState   = TIM_OCIdleState_Reset;
  ocinit.TIM_OCNIdleState  = TIM_OCNIdleState_Reset;
  TIM_OC1Init(TIM8, &ocinit);
  TIM_OC2Init(TIM8, &ocinit);
  TIM_OC3Init(TIM8, &ocinit);
  TIM_OC4Init(TIM8, &ocinit);

  //enable the timer outputs and the timer itself
  TIM_CtrlPWMOutputs(TIM8, ENABLE);
  TIM_Cmd(TIM8, ENABLE);
}

/*! Disables and resets timer8.*/
void CLockDriver::DisableTimer()
{
  //make sure the drive signal is low
  TIM_SetCompare1(TIM8, 0);
  TIM_SetCompare2(TIM8, 0);
  TIM_SetCompare3(TIM8, 0);
  TIM_SetCompare4(TIM8, 0);
  DelayMS(m_interval * 15 / 10000 + 1);

  //use a critical section to protect the RCC and timer registers
  CCritical criticalsection;

  //disable the PWM output of timer8
  TIM_CtrlPWMOutputs(TIM8, DISABLE);
  TIM_Cmd(TIM8, DISABLE);

  //reset timer8 to its default state
  RCC_APB2PeriphResetCmd(RCC_APB2Periph_TIM8, ENABLE);
  RCC_APB2PeriphResetCmd(RCC_APB2Periph_TIM8, DISABLE);

  //disable clock to timer8
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_TIM8, DISABLE);
}
