#include "lowpower.h"
#include "controller.h"
#include "lock.h"
#include "log.h"
#include "stm32f4xx.h"
#include "backupregisters.h"
#include "active.h"

CLowPowerTask::CLowPowerTask()
{
  m_allowlowpower = false;
  m_name          = "ERROR NAME NOT SET";
}

/*! Registers a task as a low power mode task.
    \param name The name of the task, used for log printing.*/
void CLowPowerTask::RegisterLowPowerTask(const char* name)
{
  g_controller.LowPower().Register(this, name);
}

/*! Allows or denies entering low power mode from a low power task.
    \param allow When true, allow entering low power mode, when false, denies it.*/
void CLowPowerTask::AllowLowPower(bool allow)
{
  g_controller.LowPower().AllowLowPower(this, allow);
}

CLowPower::CLowPower()
{
  m_mutex = 0;
}

void CLowPower::Setup()
{
  m_mutex = xSemaphoreCreateRecursiveMutex();
}

/*! Registers a task as a low power mode task.
    \param lowpowertask A pointer to the low power mode task.
    \param name The name of the task, used for log printing.*/
void CLowPower::Register(CLowPowerTask* lowpowertask, const char* name)
{
  lowpowertask->SetName(name);
  lowpowertask->AddToListAtEnd(m_tasklist);
  PrintLog(LOGLOWPOWER, "Registered low power task %s\n", name);
}

/*! Allows or denies entering low power mode from a low power task.
    \param lowpowertask A pointer to the task that denies/allows low power mode.
    \param allow When true, allow entering low power mode, when false, denies it.*/
void CLowPower::AllowLowPower(CLowPowerTask* lowpowertask, bool allow)
{
  CLock lock(m_mutex);

  if (allow != lowpowertask->GetAllowLowPower())
  {
    //print to the log if a task changed its low power mode setting
    PrintLog(LOGLOWPOWER, "Low power mode set to %s by %s\n", allow ? "allowed" : "denied",
             lowpowertask->GetName());
    lowpowertask->SetAllowLowPower(allow);

    //print which tasks allow or deny low power mode
    PrintLowPowerTasks();

    //check if all tasks allow entering low power mode
    if (allow)
    {
      bool enterlowpowermode = true;
      CLowPowerTask* it = m_tasklist.Start();
      while (it)
      {
        if (!it->GetAllowLowPower())
        {
          enterlowpowermode = false;
          break;
        }

        it = (CLowPowerTask*)it->Next();
      }

      //enter low power mode if all low power tasks allow it
      if (enterlowpowermode)
        EnterLowPowerMode();
    }
  }
}

void CLowPower::EnterLowPowerMode()
{
  CLock lock(m_mutex);

  //disable forced telemetry readings on the battery charger to save idle current
  g_controller.BatteryCharger().SetForceTelemetry(false);

  //check if the lock should be opened or closed based on settings
  g_controller.Deposit().CheckLockState();

  //Print the results of the self test before entering low power mode.
  g_controller.SelfTest().Print();

  //If all tests are passed, signal it on the display.
  g_controller.SelfTest().SignalAllTestsPassed();

#ifdef PRINT_TASK_STATS
  //If enabled, write statistics about every task.
  //Set configUSE_STATS_FORMATTING_FUNCTIONS in FreeRTOSConfig.h to 1 to enable this.
  vTaskList(m_statsbuf);
  PrintLog(LOGLOWPOWER, "Current task: %s, task statistics:\n",
           pcTaskGetName(xTaskGetCurrentTaskHandle()));
  PrintLog(LOGLOWPOWER, "Name           \tState\tPrio\tFree\tNum\n");
  PrintLog(LOGLOWPOWER, "%s", m_statsbuf);
#endif

  PrintLog(LOGLOWPOWER, "Entering low power mode\n");
  PrintDeviceActiveTest(); //Warn if active test mode is enabled.
  WaitLogWriteComplete(100);

  //configure the IO expander to generate a rising edge on PA0 when
  //the microcontroller should wake up from low power mode
  g_controller.Inputs().SetWakeupInterrupt();

  //disable interrupts to prevent anything from interrupting this section
  __disable_irq();

  //when entering low power mode, the watchdog timer should be disabled,
  //otherwise a wakeup wil occur every 32.7 seconds, however the watchdog timer
  //is enabled from the boot firmware, and can only be disabled by a full reset
  //so set the DISABLE_WATCHDOG_FLAG which tells the boot firmware to not enable
  //the watchdog timer, and set the ENTER_LOW_POWER_MODE_FLAG which tells the
  //main firmware to directly enter low power mode at boot
  SetBackupFlag(ENTER_LOW_POWER_MODE_FLAG, true);
  SetBackupFlag(DISABLE_WATCHDOG_FLAG, true);

  NVIC_SystemReset();
}

/*! Prints to the log which low power tasks deny/allow entering low power mode.
    The line looks like:
    LOWPOWER:Low power mode denied by: task1, task2, allowed by: task3, task4
    */
void CLowPower::PrintLowPowerTasks()
{
  CLogLock loglock;

  bool linestarted = false;

  //print all low power tasks that deny entering low power mode
  CLowPowerTask* it = m_tasklist.Start();
  bool first = true;
  while (it)
  {
    if (!it->GetAllowLowPower())
    {
      if (first)
      {
        PrintLog(LOGLOWPOWER, "Low power mode denied by: %s", it->GetName());
        first = false;
        linestarted = true;
      }
      else
      {
        PrintLog(LOGLOWPOWER, ", %s", it->GetName());
      }
    }

    it = (CLowPowerTask*)it->Next();
  }

  //print all low power tasks that allow entering low power mode
  it = m_tasklist.Start();
  first = true;
  while (it)
  {
    if (it->GetAllowLowPower())
    {
      if (first)
      {
        if (!linestarted)
          PrintLog(LOGLOWPOWER, "Low power mode allowed by:");
        else
          PrintLog(LOGLOWPOWER, ", allowed by:");

        PrintLog(LOGLOWPOWER, " %s", it->GetName());
        first = false;
        linestarted = true;
      }
      else
      {
        PrintLog(LOGLOWPOWER, ", %s", it->GetName());
      }
    }

    it = (CLowPowerTask*)it->Next();
  }

  //terminate with a newline if anything has been printed
  if (linestarted)
    PrintLog(LOGLOWPOWER, "\n");
}
