#include "prepaid.h"
#include "log.h"
#include "lock.h"
#include "sha256.h"
#include "random.h"
#include <algorithm>
#include <cstring>

const uint8_t CPrepaid::m_signkey[] = 
{
  0xaf, 0xd3, 0x28, 0x34, 0xc8, 0xe6, 0x95, 0x1b,
  0x7c, 0x92, 0x5b, 0xc7, 0x94, 0x1c, 0x99, 0x94
};

const uint8_t CPrepaid::m_aeskey[] =
{
  0x4b, 0x45, 0x3b, 0x18, 0x73, 0xa7, 0xa3, 0xc3,
  0x39, 0xaa, 0x92, 0x6a, 0x9d, 0xf4, 0xee, 0x22
};

CPrepaid::CPrepaid()
{
  Reset();
}

void CPrepaid::Reset()
{
  m_authlistcredit    = 0;
  m_authlistused      = 0;
  m_cardcounters.Reset();
  m_bank              = -1;
  m_cost              = 0;
  m_initialvalue      = 0;
  m_resultvalid       = false;
  m_resulttransaction = 0;
  m_resultcredit      = 0;
  m_resultused        = 0;
  m_resultcost        = 0;
}

/*! Initializes m_authlistcredit and m_authlistused from 8 bytes stored
    after the UID on the authlist.
    \param uid The UID of the card, used for logging.
    \param trailingdata The 64 bytes after the UID in the authlist.*/
void CPrepaid::FromAuthList(NFCUID* uid, uint64_t trailingdata)
{
  m_authlistcredit = trailingdata & 0xFFFFFFFF;
  m_authlistused   = (trailingdata >> 32) & 0xFFFFFFFF;

  CLogLock loglock;
  PrintLog(LOGPREPAID, "Card ");
  PrintNFCUID(LOGPREPAID, uid);
  PrintLog(LOGPREPAID, " authlist credit:%" PRIi32 ", authlist used:% " PRIi32
           ", authlist value:%" PRIi32 "\n", m_authlistcredit, m_authlistused,
           m_authlistcredit - m_authlistused);
  loglock.Leave();
}

/*! Decrypts a bank of data stored in the card, checks the HMAC,
    stores the information from the bank into m_cardcounters if its transaction
    number is higher.
    \param uid The uid of the card from which the block is read.
    \param bank From which bank the block was read.
    \param block The block of data read from the bank of the card.*/
void CPrepaid::FromCard(NFCUID* uid, uint32_t bank, CPrepaidBlock* block)
{
  //Generate an AES key that is unique for this card.
  uint8_t aeskey[CRYPTO_KEY_SIZE];
  GenerateAESKey(uid, aeskey);

  //Decrypt the information from the card,
  //decryptbuf will then hold the HMAC and the payload.
  uint8_t decryptbuf[block->HMACSize() + block->PayloadSize()];
  AES_CBC_decrypt_buffer(decryptbuf, block->HMAC(), block->HMACSize() + block->PayloadSize(),
                         aeskey, block->IV());

  //Generate a sign key that is unique for this card.
  uint8_t signkey[CRYPTO_KEY_SIZE];
  GenerateSignKey(uid, signkey);

  //Generate an HMAC over the decrypted payload.
  uint8_t hmac[SHA256_BLOCK_SIZE];
  HMAC_SHA256_CTX hmacctx;
  HMAC_SHA256_Init_rfc2104(&hmacctx, m_signkey, CRYPTO_KEY_SIZE);
  HMAC_SHA256_Update(&hmacctx, decryptbuf + block->HMACSize(), block->PayloadSize());
  HMAC_SHA256_Final(&hmacctx, hmac);

  //Check if the HMAC is correct, in a loop that runs in constant time.
  bool hmaccorrect = true;
  for (uint32_t i = 0; i < std::min(block->HMACSize(), (uint32_t)SHA256_BLOCK_SIZE); i++)
    hmaccorrect = hmaccorrect && (decryptbuf[i] == hmac[i]);

  //If the HMAC of this block is incorrect, the data should not be used.
  if (!hmaccorrect)
  {
    CLogLock loglock;
    PrintLog(LOGPREPAID, "ERROR: HMAC of bank %" PRIu32 " in card ", bank);
    PrintNFCUID(LOGPREPAID, uid, " is incorrect\n");
    loglock.Leave();

    return;
  }

  CLogLock loglock;
  PrintLog(LOGPREPAID, "HMAC of bank %" PRIu32 " in card ", bank);
  PrintNFCUID(LOGPREPAID, uid, " is correct\n");
  loglock.Leave();

  //Interpret the decrypted information from the card.
  CPrepaidCounters cardcounters;
  memcpy(cardcounters, decryptbuf + block->HMACSize(),
         std::min((uint32_t)sizeof(cardcounters), block->PayloadSize()));

  //Print the counters stored in the card to the log.
  loglock.Enter();
  PrintLog(LOGPREPAID, "Counters in card ");
  PrintNFCUID(LOGPREPAID, uid, ": ");
  PrintLog(LOGPREPAID, " bank %" PRIu32 " version:%" PRIu8 ", transaction:%" PRIu32
           ", credit:%" PRIi32 ", used:%" PRIi32 ", value:%" PRIi32 "\n", bank,
           cardcounters.Version(), cardcounters.Transaction(), cardcounters.Credit(),
           cardcounters.Used(), cardcounters.Credit() - cardcounters.Used());
  loglock.Leave();

  //Use the information from this bank if its transaction number is higher.
  if (cardcounters.Transaction() > m_cardcounters.Transaction())
  {
    m_cardcounters = cardcounters;
    m_bank = bank;
  }
}

/*! Sets the cost of the prepaid transaction.
    \param cost The cost of the prepaid transaction.*/
void CPrepaid::SetCost(int32_t cost)
{
  if (cost >= 0)
  {
    m_cost = cost;
  }
  else
  {
    PrintLog(LOGPREPAID, "ERROR: cost is %" PRIi32
             ", should not be negative, setting to 0\n", cost);

    m_cost = 0;
  }
}

/*! Checks the credit and used counters from the whitelist and the counters
    stored on the card. If the card has enough value, the use counter will
    be increased by m_cost and block is filled with data that can be written
    to the card.
    \param uid   The uid of the card on which the deposit is charged.
    \param block Pointer to a CPrepaidBlock instance that is filled with data
                 that can be written to the card in case of a successfull charge.
    \return true If the deposit has been charged on the card.
    \return false If the card does not have enough credit for the deposit.
                  Block still holds prepaid data with a new transaction number
                  that should be written to the card.*/
bool CPrepaid::Charge(NFCUID* uid, CPrepaidBlock* block)
{
  //Indicate that the result fields will become valid.
  m_resultvalid = true;

  //Create a new card transaction number.
  m_resulttransaction = m_cardcounters.Transaction() + 1;

  //There are credit and used counters stored in the authlist, and on the card.
  //From these counters, use the highest value of each.
  m_resultcredit = std::max(m_authlistcredit, m_cardcounters.Credit());
  m_resultused   = std::max(m_authlistused, m_cardcounters.Used());

  //Calculate the value before the deposit is charged.
  m_initialvalue = m_resultcredit - m_resultused;

  //Check if the card has enough value to pay for a deposit.
  int32_t value  = m_resultcredit - m_resultused;
  bool    enough = value >= m_cost;
  if (enough)
  {
    CLogLock loglock;
    PrintLog(LOGPREPAID, "Card ");
    PrintNFCUID(LOGPREPAID, uid);
    PrintLog(LOGPREPAID, " has value %" PRIi32 " which is enough to pay for a deposit of cost %"
             PRIi32 "\n", value, m_cost);
    loglock.Leave();

    //Charge the deposit on the card.
    m_resultcost = m_cost;
    m_resultused += m_cost;
  }
  else
  {
    m_resultcost = 0;

    CLogLock loglock;
    PrintLog(LOGPREPAID, "ERROR: card ");
    PrintNFCUID(LOGPREPAID, uid);
    PrintLog(LOGPREPAID, " does not have enough value to pay for a deposit of %"
             PRIi32 " credits\n", m_cost);
    PrintLog(LOGPREPAID, "Value:%" PRIi32 ", credit:%" PRIi32 ", used:%" PRIi32 "\n",
             value, m_resultcredit, m_resultused);
    loglock.Leave();
  }

  //Create new data to write to the card.
  CPrepaidCounters cardcounters;
  cardcounters.SetTransaction(m_resulttransaction);
  cardcounters.SetCredit(m_resultcredit);
  cardcounters.SetUsed(m_resultused);

  PrintLog(LOGPREPAID, "Writing to card: transaction:%" PRIu32
           ", credit:%" PRIi32 ", used:%" PRIi32 "\n",
           cardcounters.Transaction(), cardcounters.Credit(), cardcounters.Used());

  //Sign and encrypt prepaid data.
  Encrypt(uid, &cardcounters, block);

  //Return if the card has enough value to do the deposit.
  return enough;
}

/*! Signs and encypts cardcounters and stores it in block. This function is
    declared static because it's also used to initialize prepaids cards during
    programming.
    \param uid The UID to generate sign and encryption keys from.
    \param cardcounters The prepaid counters to encrypt.
    \param block The encrypted data is stored here.*/
void CPrepaid::Encrypt(NFCUID* uid, CPrepaidCounters* cardcounters,
                       CPrepaidBlock* block)
{
  //Generate a sign key that is unique for this card.
  uint8_t signkey[CRYPTO_KEY_SIZE];
  GenerateSignKey(uid, signkey);

  //Generate an HMAC over the unencrypted data.
  uint8_t hmac[SHA256_BLOCK_SIZE];
  HMAC_SHA256_CTX hmacctx;
  HMAC_SHA256_Init_rfc2104(&hmacctx, m_signkey, CRYPTO_KEY_SIZE);
  HMAC_SHA256_Update(&hmacctx, *cardcounters, sizeof(*cardcounters));
  HMAC_SHA256_Final(&hmacctx, hmac);

  //Generate an AES key that is unique for this card.
  uint8_t aeskey[CRYPTO_KEY_SIZE];
  GenerateAESKey(uid, aeskey);

  //Encrypt the HMAC and card counters.
  uint8_t buf[block->HMACSize() + block->PayloadSize()];
  memcpy(buf, hmac, block->HMACSize());
  memcpy(buf + block->HMACSize(), *cardcounters, block->PayloadSize());
  CreateIV(block->IV());
  AES_CBC_encrypt_buffer(block->HMAC(), buf, block->HMACSize() + block->PayloadSize(),
                         aeskey, block->IV());
}

/*! Generates an HMAC over the UID, using m_signkey as key. This is then copied into key.
    This creates a unique signing key for each NFC card.
    \param uid The UID of the card.
    \param key The generated key is stored here.*/
void CPrepaid::GenerateSignKey(NFCUID* uid, uint8_t key[CRYPTO_KEY_SIZE])
{
  //Generate an HMAC.
  uint8_t hmac[SHA256_BLOCK_SIZE];
  HMAC_SHA256_CTX hmacctx;
  HMAC_SHA256_Init_rfc2104(&hmacctx, m_signkey, CRYPTO_KEY_SIZE);
  HMAC_SHA256_Update(&hmacctx, uid->uid, uid->size);
  HMAC_SHA256_Final(&hmacctx, hmac);

  //Use the HMAC as signing key.
  memset(key, 0, CRYPTO_KEY_SIZE);
  memcpy(key, hmac, std::min(SHA256_BLOCK_SIZE, CRYPTO_KEY_SIZE));
}

/*! Generates an HMAC over the UID, using m_aeskey as key. This is then copied into key.
    This creates a unique AES key for each NFC card.
    \param uid The UID of the card.
    \param key The generated key is stored here.*/
void CPrepaid::GenerateAESKey(NFCUID* uid, uint8_t key[CRYPTO_KEY_SIZE])
{
  //Generate an HMAC.
  uint8_t hmac[SHA256_BLOCK_SIZE];
  HMAC_SHA256_CTX hmacctx;
  HMAC_SHA256_Init_rfc2104(&hmacctx, m_aeskey, CRYPTO_KEY_SIZE);
  HMAC_SHA256_Update(&hmacctx, uid->uid, uid->size);
  HMAC_SHA256_Final(&hmacctx, hmac);

  //Use the HMAC as AES key.
  memset(key, 0, CRYPTO_KEY_SIZE);
  memcpy(key, hmac, std::min(SHA256_BLOCK_SIZE, CRYPTO_KEY_SIZE));
}

/*! \return To which bank on the card data should be written.*/
uint32_t CPrepaid::WriteBankIndex()
{
  if (m_bank == 0)
    return 1;
  else
    return 0;
}
