#include "reader.h"
#include "inclstdint.h"
#include "log.h"
#include "clock.h"
#include "gpiodescriptions.h"
#include "platform.h"
#include "controller.h"
#include "speaker.h"
#include "stm32wss.h"
#include "random.h"
#include "sha256.h"

extern "C"
{
  #include "st25r3911_interrupt.h"
  #include "rfal_rf.h"
  #include "rfal_analogConfig.h"
  #include "rfal_nfc.h"
  #include "rfal_t2t.h"
  #include "st25r3911_com.h"
  #include "st25r3911.h"
}

#include <cassert>
#include <algorithm>

CST25R3911Isr::CST25R3911Isr(CReader& parent) : m_parent(parent)
{
}

/*! Sets up the ISR task for the ST25R3911B interrupt pin.*/
void CST25R3911Isr::Setup()
{
  //enable the interrupt for the rising edge on the ST25R3911B interrupt pin
  //when it happens, the GPIOIRQHandler method is called on this class
  g_exti.EnableGPIOIrq(ST25R3911B_INT, EXTI_Trigger_Rising, this);

  //create a high priority task to handle the ST25R3911B pin interrupt
  Create("ReaderISR", HIGHEST_PRIORITY);
}

void CST25R3911Isr::Process()
{
  for(;;)
  {
    //wait until the CST25R3911Isr task is signaled, this means a rising edge
    //on the STB25R3911B interrupt pin has happened
  //  EDWIN SAYS NO
    WaitNotificationCounting();
    //  vTaskDelay(1000);
    //the rfal library assumes that st25r3911Isr() is called from an interrupt handler,
    //and that it will never be interrupted,
    //however because it's called from a FreeRTOS task, it could be interrupted by the CReader task,
    //so lock the spi lock, and suspend the CReader task,
    //this guarantees that the CReader task will not interrupt it
    ReaderLock();
    m_parent.Suspend();
    m_parent.PreventLowPower(); //in case this is a low power wakeup interrupt
    st25r3911Isr();
    m_parent.Resume();
    ReaderUnlock();

    //notify the reader task that an interrupt happened,
    //in case it's waiting on the low power wakeup
    m_parent.Notify();
  }
}

/*! Called when a rising edge on @ref ST25R3911B_INT is detected.*/
void CST25R3911Isr::GPIOIRQHandler(SGPIOPin pin)
{
  //signal the CST25R3911Isr FreeRTOS task that an interrupt has happened
  NotifyFromISR();
}

static CReader* g_reader;

/*! Used in platform.h as the bridge between rfal and CReader.
    Does a transceive on the spi bus.
    \param out A pointer to a buffer of bytes to send.
    \param in A pointer to a buffer where received bytes are stored.
    \param len The number of bytes to transceive.*/
void ReaderTransceive(const uint8_t* out, uint8_t* in, uint32_t len)
{
  assert(g_reader);
  g_reader->Transceive(out, in, len);
}

/*! Used in platform.h as the bridge between rfal and CReader.
    Locks access to the spi bus using a recursive mutex.*/
void ReaderLock()
{
  assert(g_reader);
  g_reader->Lock();
}

/*! Used in platform.h as the bridge between rfal and CReader.
    Unlocks access to the spi bus using a recursive mutex.*/
void ReaderUnlock()
{
  assert(g_reader);
  g_reader->Unlock();
}

CReader::CReader() : m_irqhandler(*this)
{
  //set a pointer for the rfal library
  assert(g_reader == nullptr);
  g_reader = this;

  m_spilock      = 0;
  m_woke         = false;
  m_wokelowpower = false;
  m_working      = false;
  m_signaledcomm = false;
  m_denylistsize = 0;
}

/*! Sets up everything needed for the nfc card reader.*/
void CReader::Setup()
{
  //read the ST25R3911B interrupt pin, if it's high, it probably signaled
  //a low power card detect
  m_woke = GPIO_ReadInput(ST25R3911B_INT);
  m_wokelowpower = m_woke;

  //create a mutex to lock access to m_spi, because two tasks are accessing it
  //this task, and the CST25R3911Isr task can access it at any time
  m_spilock = xSemaphoreCreateRecursiveMutex();

  //set up the spi peripheral in mode 1, the clock line is low on idle, the MOSI and MISO lines
  //are allowed to change on the rising edge, and they should be stable on the falling edge
 // m_spi.Setup(SPI4, SPI_BaudRatePrescaler_256, SPI_NSS_Soft, /* SPI_CPOL_Low  EDWIN */ SPI_CPOL_High , SPI_CPHA_2Edge);
/* 23 aug 2022 Edwin TEST of SPI0 (GD nummering) (dat is SPI1 in STM) wel werkt op de APB2 BUS, ja dat werkt */
/* SPI2 werkt niet, registers wel updated in memory view van SPI2 */
/* SPI3 werkt wel incl registers */
/* SPI4 werkt niet , geen register updates */
/* SPI5 werkt niet , geen register updates */
/* SPI6 werkt niet , geen register updates */

  m_spi.Setup(SPI4, SPI_BaudRatePrescaler_256, SPI_NSS_Soft, /* SPI_CPOL_Low  EDWIN */ SPI_CPOL_Low , SPI_CPHA_2Edge);

  //register this class as a task that can decide if low power mode can be entered
  RegisterLowPowerTask("reader");
}

/*! Starts the FreeRTOS reader task.*/
void CReader::Start()
{
  Create("Reader");
}

/*! The main process loop for the CReader task, it waits for a low power card detect,
    then it reads all cards in the field, and passes them to the auth list for processing,
    then waits for the card to disappear before entering low power card detect mode again.*/
void CReader::Process()
{
    printf("aap\n");
  m_irqhandler.Setup();
printf("noot\n");
  bool lowpowerwakeup = true;

  for(;;)
  {
    //start with a full reset and initialize of the reader
    //this is done every iteration, for stability,
    InitializeReader();
// printf("EDWIN:ERR\n");
    bool calerror = CheckAntennaCalibration();

    ClearDenyList();

    //try to read and authenticate a card for 5 seconds
    rfalNfcaListenDevice authenticatedcard;
    if (ProcessRead(5000, &authenticatedcard, nullptr))
    {
      //wait 10 seconds for the card to disappear after it has been authenticated
      //if during this, a different card appears, don't enable low power wakeup,
      //instead try to authenticate the other card
      lowpowerwakeup = WaitForCardRemoval(10000, &authenticatedcard);

      //if a deposit action was executed, power down the lock drive circuitry after
      //the deposit has been processed, because the power down needs a delay
      g_controller.Deposit().PowerDown();

      DecreasePontCounter();

      //A card has been read, mark test as passed.
      g_controller.SelfTest().SetTestPassed(SelfTest_NFCReader);

      //If the microcontroller has been woken up from low power mode by the
      //reader, and a card has been read, mark low power test as passed.
      if (m_wokelowpower && ResetFlags() == 0)
        g_controller.SelfTest().SetTestPassed(SelfTest_NFCLowPowerCardDetect);
    }
    else
    {
      IncreasePontCounter();

      //if no card has been granted access, register the UIDs of all cards
      //that have been denied access
      RegisterDeniedCards();

      //if no cards can be read or authenticated, go into low power wakeup
      lowpowerwakeup = true;
    }

    //If the reader had an error calibrating its NFC coil, record a HardwareError
    //status, but only after trying to read cards, because recording a status right
    //after booting might take few seconds, and cards should be read immediately
    //after boot.
    if (calerror)
      g_controller.Status().Write(Status_HardwareError, "NFC coil calibration");

    //wait for an NFC card to appear
    if (lowpowerwakeup)
      WaitLowPowerWakeup();
    else
      m_woke = false;

    //turn off the LCD when the processing the card is done,
    //and the card has been removed
    g_controller.UserInterface().Set(UIState_Off);
  }
}

void CReader::Transceive(const uint8_t* out, uint8_t* in, uint32_t numbytes)
{
  m_spi.Transceive(out, in, numbytes);
}

void CReader::Lock()
{
  xSemaphoreTakeRecursive(m_spilock, portMAX_DELAY);
}

void CReader::Unlock()
{
  xSemaphoreGiveRecursive(m_spilock);
}

/*! Time in milliseconds that the reader initialization can take. If this
    timeout expires, there is probably something wrong with the reader.
    In that case, the initialization should be aborted, and low power
    mode should be entered, without enabling the interrupt on the IRQ pin of
    the ST25R3911B NFC reader.*/
#define READER_INIT_TIMEOUT 300000

/*! Initializes the ST25R3911B NFC reader and library.*/
void CReader::InitializeReader()
{
  int64_t start = Millis();
  for(;;)
  {
   //   printf("XXPP\n");
    ReturnCode err = rfalNfcInitialize();
    //  printf("XXQQ\n");
    if (err == ERR_NONE)
    {
      m_working = true;
      return; //successfully initialized
    }
    else
    {
      //signal that the ST25R3911B IRQ pin should not wake the microcontroller
      //from low power mode, because it's not working correctly
      m_working = false;
//printf("ERROR: rfalInitialize() returned error %i\n", (int)err);
      PrintLog(LOGREADER, "ERROR: rfalInitialize() returned error %i\n", (int)err);

      if (Millis() - start > READER_INIT_TIMEOUT)
      {
        PrintLog(LOGREADER, "ERROR: timeout initializing reader\n");
        printf("ERROR: timeout initializing reader");
        WaitLowPowerWakeup(); //enter low power mode to save battery power

      }

      DelayMS(1000);
    }
  }
}

/*! The ST25R3911B tunes the antenna using capacitors connected to pins.
    This function checks the value of the calibration register, and prints it to the log.
    \return true if the antenna is calibrated correctly.
    \return false if calibrating the antenna failed.*/
bool CReader::CheckAntennaCalibration()
{
  uint8_t calibration;
  st25r3911ReadRegister(ST25R3911_REG_ANT_CAL_RESULT, &calibration);

  uint8_t calvalue = calibration >> 4;
  bool    calerror = !!(calibration & (1 << 3));

  PrintLog(LOGREADER, "Antenna calibration value (0-15): %" PRIu8 ", calibration error: %s\n",
           calvalue, calerror ? "yes" : "no");

  if (!calerror)
    g_controller.SelfTest().SetTestPassed(SelfTest_ReaderCoil);

  return calerror;
}

/*! Reads parameters for low power wakeup from the settings,
    and sets up the NFC reader in low power wakeup mode.*/
void CReader::EnableLowPowerWakeup()
{
  //Read low power wakeup parameters from the settings.
  uint32_t period           = g_controller.Settings().GetInteger(Setting_LPW_Period);

  bool     ampenabled       = g_controller.Settings().GetBoolean(Setting_LPW_AmplitudeEnabled);
  uint32_t ampdelta         = g_controller.Settings().GetInteger(Setting_LPW_AmplitudeDelta);
  uint32_t ampref           = g_controller.Settings().GetInteger(Setting_LPW_AmplitudeReference);
  bool     ampautoavg       = g_controller.Settings().GetBoolean(Setting_LPW_AmplitudeAutoAvg);
  uint32_t ampautoavgweight = g_controller.Settings().GetInteger(Setting_LPW_AmplitudeAutoAvgWeight);
  bool     ampinclmeas      = g_controller.Settings().GetBoolean(Setting_LPW_AmplitudeAutoAvgInclMeas);

  bool     phaenabled       = g_controller.Settings().GetBoolean(Setting_LPW_PhaseEnabled);
  uint32_t phadelta         = g_controller.Settings().GetInteger(Setting_LPW_PhaseDelta);
  uint32_t pharef           = g_controller.Settings().GetInteger(Setting_LPW_PhaseReference);
  bool     phaautoavg       = g_controller.Settings().GetBoolean(Setting_LPW_PhaseAutoAvg);
  uint32_t phaautoavgweight = g_controller.Settings().GetInteger(Setting_LPW_PhaseAutoAvgWeight);
  bool     phainclmeas      = g_controller.Settings().GetBoolean(Setting_LPW_PhaseAutoAvgInclMeas);

  //Set up the rfalWakeUpConfig with the settings, some need to be translated.
  rfalWakeUpConfig cfg      = {};
  cfg.period                = LowPowerWakeupPeriod(period);     //Card detect interval.
  cfg.irqTout               = false;                            //Don't generate an interrupt every measurement.
  
  cfg.indAmp.enabled        = ampenabled;                       //Enables inductive amplitude measurement.
  cfg.indAmp.delta          = std::min(ampdelta, (uint32_t)15); //Difference between measurements that triggers a wakeup.
  cfg.indAmp.reference      = std::min(ampref, (uint32_t)255);  //Zero reference, 255 for auto measurement.
  cfg.indAmp.autoAvg        = ampautoavg;                       //Sets hardware auto averaging.
  cfg.indAmp.aaWeight       = AutoAvgWeight(ampautoavgweight);  //Sets the weight for auto averaging.
  cfg.indAmp.aaInclMeas     = ampinclmeas;                      //Sets if the wakeup measurement is included in the auto averaging.

  cfg.indPha.enabled        = phaenabled;                       //Enables inductive phase measurement.
  cfg.indPha.delta          = std::min(phadelta, (uint32_t)15); //Difference between measurements that triggers a wakeup.
  cfg.indPha.reference      = std::min(pharef, (uint32_t)255);  //Zero reference, 255 for auto measurement.
  cfg.indPha.autoAvg        = phaautoavg;                       //Sets hardware auto averaging.
  cfg.indPha.aaWeight       = AutoAvgWeight(phaautoavgweight);  //Sets the weight for auto averaging.
  cfg.indPha.aaInclMeas     = phainclmeas;                      //Sets if the wakeup measurement is included in the auto averaging.

  cfg.cap.enabled           = false;                            //Disable capacitive measurement

  //Enable low power wakeup.
  rfalWakeUpModeStart(&cfg);

  //Print the settings to the log.
  CLogLock loglock;
  PrintLog(LOGREADER, "Low power card detect enabled, period:%" PRIu32 "ms\n", period);

  PrintLog(LOGREADER, "Amplitude detection:%s, amplitude delta:%" PRIu32", "
           "amplitude reference:%" PRIu32 ", amplitude auto averaging:%s, "
           "amplitude auto averaging weight:%" PRIu32 ", "
           "amplitude include wakeup measurement:%s\n",
           ampenabled ? "enabled" : "disabled",
           ampdelta,
           ampref,
           ampautoavg ? "enabled" : "disabled",
           ampautoavgweight,
           ampinclmeas ? "enabled" : "disabled");

  PrintLog(LOGREADER, "Phase detection:%s, phase delta:%" PRIu32", "
           "phase reference:%" PRIu32 ", phase auto averaging:%s, "
           "phase auto averaging weight:%" PRIu32 ", "
           "phase include wakeup measurement:%s\n",
           phaenabled ? "enabled" : "disabled",
           phadelta,
           pharef,
           phaautoavg ? "enabled" : "disabled",
           phaautoavgweight,
           phainclmeas ? "enabled" : "disabled");
}
/*! Reads the low power wakeup period from the settings and converts it.
    \return The period field for the rfalWakeUpConfig struct.*/
rfalWumPeriod CReader::LowPowerWakeupPeriod(uint32_t detecttime)
{
  //Conversion table to convert a period into an enum rfalWumPeriod.
  static const struct
  {
    uint32_t      period;
    rfalWumPeriod setting;
  }
  periods[] =
  {
    { 10,  RFAL_WUM_PERIOD_10MS  },
    { 20,  RFAL_WUM_PERIOD_20MS  },
    { 30,  RFAL_WUM_PERIOD_30MS  },
    { 40,  RFAL_WUM_PERIOD_40MS  },
    { 50,  RFAL_WUM_PERIOD_50MS  },
    { 60,  RFAL_WUM_PERIOD_60MS  },
    { 70,  RFAL_WUM_PERIOD_70MS  },
    { 80,  RFAL_WUM_PERIOD_80MS  },
    { 100, RFAL_WUM_PERIOD_100MS },
    { 200, RFAL_WUM_PERIOD_200MS },
    { 300, RFAL_WUM_PERIOD_300MS },
    { 400, RFAL_WUM_PERIOD_400MS },
    { 500, RFAL_WUM_PERIOD_500MS },
    { 600, RFAL_WUM_PERIOD_600MS },
    { 700, RFAL_WUM_PERIOD_700MS },
    { 800, RFAL_WUM_PERIOD_800MS },
  };

  //uint32_t period = g_controller.Settings().GetInteger(Setting_LPW_Period);
  for (uint32_t i = 0; i < sizeof(periods) / sizeof(periods[0]); i++)
  {
    if (periods[i].period >= detecttime)
      return periods[i].setting;
  }

  return RFAL_WUM_PERIOD_800MS; //Safe default.
}

/*! Converts a weight from a number to a value of enum rfalWumAAWeight.
    \return a value for aaWeight in struct rfalWakeUpConfig.*/
rfalWumAAWeight CReader::AutoAvgWeight(uint32_t weight)
{
  //Conversion table to convert from a number to enum rfalWumAAWeight.*/
  static const struct
  {
    uint32_t        weight;
    rfalWumAAWeight setting;
  }
  weights[] =
  {
    { 4,  RFAL_WUM_AA_WEIGHT_4 },
    { 8,  RFAL_WUM_AA_WEIGHT_8 },
    { 16, RFAL_WUM_AA_WEIGHT_16 },
    { 32, RFAL_WUM_AA_WEIGHT_32 }
  };

  for (uint32_t i = 0; i < sizeof(weights) / sizeof(weights[0]); i++)
  {
    if (weights[i].weight >= weight)
      return weights[i].setting;
  }

  return RFAL_WUM_AA_WEIGHT_32; //Safe default.
}

void CReader::WaitLowPowerWakeup()
{
  //re-initialize the reader before going into low power wakeup, for stability
  InitializeReader();

  //clear any task notification from the CST25R3911Isr task
  //before enabling low power card detect
  ClearNotification();

  //enable low power card detect from inside a lock, in case a
  //low power card detect happens right after it's enabled
  //this prevents a race condition with the CST25R3911Isr::Process() method
  ReaderLock();
  EnableLowPowerWakeup();

  //delay at least 500 milliseconds before turning off the display, this way
  //if a user presents a card right after the display turns off, the reader
  //will wake from low power mode immediately
  DelayMS(500);
  g_controller.UserInterface().Set(UIState_Off);

  AllowLowPower(true); //allow the microcontroller to go into low power mode
  ReaderUnlock();

  for(;;)
  {
    //the rfal worker should run regularly
    rfalWorker();

    if(rfalWakeUpModeHasWoke())
    { //stop when a wakeup event has been detected
      PrintLog(LOGREADER, "Woke up\n");
      rfalFieldOff();
      rfalWakeUpModeStop();
      m_woke = true;
      break;
    }

    //wait until the CST25R3911Isr task notifies the CReader task
    //of a rising edge on the ST25R3911B interrupt pin
    WaitNotificationBinary();
    PrintLog(LOGREADER, "Reader task notified\n");
  }
}

/*! Called from the CST25R3911Isr task, when the ST25R3911B reader signals
    an interrupt via its interrupt pin. If the ST25R3911B was in low power
    wakeup mode before the interrupt, the microcontroller should not enter
    low power mode, instead it should read a card via NFC.*/
void CReader::PreventLowPower()
{
  AllowLowPower(false);
}

/*! Tries to read cards until a card is authenticated, or until the timeout expires.
    \param timeoutms Timeout in milliseconds until giving up.
    \param authenticatedcard Returns the information of the card that was authenticated on the auth list, or against presentcard.
    \param presentcard When not nullptr, will check all of the read cards against this, and true is returned when this card is present.
    \param cardspresent Returns the number of cards detected in the field during the last read.
    \return true if a card was read that is on the auth list, or has the same uid as presentcard if it's not nullptr.*/
bool CReader::ProcessRead(uint32_t timeoutms, rfalNfcaListenDevice* authenticatedcard, rfalNfcaListenDevice* presentcard, uint32_t* cardspresent)
{
  if (presentcard == nullptr)
  {
    PrintLog(LOGREADER, "Trying to detect ISO14443A card on the auth list for %" PRIu32 " milliseconds\n", timeoutms);
  }
  else
  {
    CLogLock loglock;
    PrintLog(LOGREADER, "Trying to detect ISO14443A card with UID ");
    PrintCardUID(LOGREADER, presentcard);
    PrintLog(LOGREADER, " for %" PRIu32 " milliseconds\n", timeoutms);
  }

  if (cardspresent != nullptr)
    *cardspresent = 0;

  int64_t start = Millis();
  bool detected = false;
  for(;;)
  {
    int64_t iterationstart = Millis();

    //the rfal worker should run regularly
    rfalWorker();

    //try to read all cards in the field,
    uint32_t numdevices = ReadCards(m_nfcadevlist, MAXDEVICES);
    rfalFieldOff();
    assert(numdevices <= MAXDEVICES);

    //return the number of cards present in the field
    if (cardspresent != nullptr)
      *cardspresent = numdevices;

    PrintCardInfo(m_nfcadevlist, numdevices);

    //check if one of the cards can be authenticated
    //if presentcard is nullptr, the cards should be checked against the auth list,
    //if none are present on it, show the denied screen on the display
    if (AuthenticateCard(m_nfcadevlist, numdevices, authenticatedcard, presentcard))
      return true;
    else if (presentcard == nullptr && numdevices > 0)
      g_controller.UserInterface().Set(UIState_Denied);

    //reset the timeout after the first detect,
    //this keeps the timing of the denied screen constant
    if (!detected && numdevices > 0)
    {
      start = Millis();
      detected = true;
    }

    if (Millis() - start >= timeoutms)
      break;

    //limit the speed at which cards are read
    int64_t duration = Millis() - iterationstart;
    if (duration < 300)
      DelayMS(300 - duration);
  }

  PrintLog(LOGREADER, "Timeout detecting ISO14443A card\n");
  return false;
}

/*! Macro which executes func and checks its return value,
    prints a log message if the return value is not the same as logignore,
    and returns 0 if the return value is not ERR_NONE.
    \param func The function to execute.
    \param args The arguments to pass to func.
    \param logignore The return value to ignore, set to ERR_NONE to print all error messages.
    \param functreturn The value to return from the function this is called from.*/
#define RETURN_ON_ERROR(func, args, logignore, funcreturn) \
  do \
  { \
    ReturnCode retval = func args; \
    if (retval != ERR_NONE) \
    { \
      if (retval != logignore) \
        PrintLog(LOGREADER, "ERROR: %s returned error %i\n", #func "()", (int)retval); \
      return funcreturn; \
    } \
  } \
  while (false)

/*! Initializes the cards in the field so that they can be read.
    \return true on success.
    \return false on failure.*/
bool CReader::InitCards()
{
  RETURN_ON_ERROR(rfalNfcaPollerInitialize, (), ERR_NONE, false); //initialize the reader in NFC-A mode
  RETURN_ON_ERROR(rfalFieldOnAndStartGT, (), ERR_NONE, false);    //enable the field and start the GT timer

  //detect if there is any NFC device in the field
  rfalNfcaSensRes sensres;
  RETURN_ON_ERROR(rfalNfcaPollerTechnologyDetection, (RFAL_COMPLIANCE_MODE_NFC, &sensres), ERR_TIMEOUT, false);

  //TODO: figure out if this is useful information
  PrintLog(LOGREADER, "rfalNfcaPollerTechnologyDetection() anticollisionInfo:%u "
           "platformInfo:%u\n", sensres.anticollisionInfo, sensres.platformInfo);

  return true;
}

/*! Reads all devices in the field, and stores the info in nfcadevlist.
    \param nfcadevlist An array of type rfalNfcaListenDevice and of size maxdevices.
    \param maxdevices The maximum number of devices to read.
    \return The number of NFC devices read.*/
uint32_t CReader::ReadCards(rfalNfcaListenDevice* nfcadevlist, uint32_t maxdevices)
{
  //Initialize NFC cards in the field so that they can be read.
  if (!InitCards())
    return false;

  //Detect all devices in the field.
  uint8_t numdevices;
  RETURN_ON_ERROR(rfalNfcaPollerFullCollisionResolution, (RFAL_COMPLIANCE_MODE_NFC, maxdevices, nfcadevlist, &numdevices), ERR_NONE, false);
  assert(numdevices <= maxdevices);

  //In the WSS database, the byte order of all UIDs is reversed,
  //so reverse the byte order of each UID after reading to keep compatibility.
  for (uint32_t device = 0; device < numdevices; device++)
  {
    uint8_t buf[sizeof(nfcadevlist[device].nfcId1)];
    memcpy(buf, nfcadevlist[device].nfcId1, sizeof(nfcadevlist[device].nfcId1));

    uint32_t uidlen = nfcadevlist[device].nfcId1Len;
    for (uint32_t i = 0; i < uidlen; i++)
      nfcadevlist[device].nfcId1[i] = buf[uidlen - i - 1];
  }

  return numdevices;
}

void CReader::PrintCardInfo(rfalNfcaListenDevice* nfcadevlist, uint32_t numdevices)
{
  if (numdevices == 0)
    return;

  PrintLog(LOGREADER, "Detected %" PRIu32 " NFC-A card%s\n", numdevices, numdevices != 1 ? "s" : "");

  for (uint32_t c = 0; c < numdevices; c++)
  {
    PrintLog(LOGREADER, "\n");
    PrintLog(LOGREADER, "Card %" PRIu32 "\n", c + 1);

    if (nfcadevlist[c].type == RFAL_NFCA_T1T)
      PrintLog(LOGREADER, "Type: Type 1 Tag\n");
    else if (nfcadevlist[c].type == RFAL_NFCA_T2T)
      PrintLog(LOGREADER, "Type: Type 2 Tag\n");
    else if (nfcadevlist[c].type == RFAL_NFCA_T4T)
      PrintLog(LOGREADER, "Type: Type 4 Tag\n");
    else if (nfcadevlist[c].type == RFAL_NFCA_NFCDEP)
      PrintLog(LOGREADER, "Type: RFAL_NFCA_NFCDEP\n");
    else if (nfcadevlist[c].type == RFAL_NFCA_T4T_NFCDEP)
      PrintLog(LOGREADER, "Type: RFAL_NFCA_T4T_NFCDEP\n");

    CLogLock loglock;
    PrintLog(LOGREADER, "UID: ");
    PrintCardUID(LOGREADER, nfcadevlist + c);
    PrintLog(LOGREADER, "\n");
  }

  if (numdevices > 0)
    PrintLog(LOGREADER, "\n");
}

/*! Prints the UID of a card, or XXXXXXXX when Setting_Data_anonymous is enabled.*/
void CReader::PrintCardUID(int loglevel, rfalNfcaListenDevice* dev)
{
  if (!g_controller.Settings().GetBoolean(Setting_Data_anonymous))
  {
    for (uint32_t i = 0; i < dev->nfcId1Len; i++)
      PrintLog(loglevel, "%02X", dev->nfcId1[i]);
  }
  else
  {
    PrintLog(loglevel, "XXXXXXXX");
  }
}

/*! Reads a card for one second, if it fails it assumes there is no card present.
    It will try to read until the timeout has been reached.
    \param timeoutms The timeout in milliseconds before giving up waiting for the card to be removed.
    \param authenticatedcard The card that was previously authenticated to open the lock,
    this function waits until this card has been removed.
    \return true if low power card detect can be enabled,
    false if a different card is now present that should be authenticated.*/
bool CReader::WaitForCardRemoval(uint32_t timeoutms, rfalNfcaListenDevice* authenticatedcard)
{
  PrintLog(LOGREADER, "Waiting %" PRIu32 " milliseconds for card to be removed\n", timeoutms);

  uint32_t cardspresent = 0;

  int64_t start = Millis();
  for(;;)
  {
    //try to read a card for one second, if it fails, it has been removed
    if (!ProcessRead(1000, nullptr, authenticatedcard, &cardspresent))
    {
      PrintLog(LOGREADER, "Card removed\n");
      //if a different card is now present in the field, low power wakeup should not be enabled
      //if no cards are present, low power wakeup can be enabled
      return cardspresent == 0;
    }
    else if (Millis() - start >= timeoutms)
    {
      PrintLog(LOGREADER, "Timeout waiting for card to be removed\n");
      break;
    }
    else
    {
      //indicate to the user that the card should be removed
      CSpeaker::Beep(100);
      //if reading the card succeeded, try again in one second, to save battery life
      DelayMS(1000);
    }
  }

  //low power wakeup can be enabled
  return true;
}

/*! If presentcard is nullptr, checks if a card is present on the auth list.
    If presentcard is not nullptr, checks if a card is the same as presentcard.
    \param dev The card(s) currently read.
    \param numdevices The number of cards currently read.
    \param authenticatedcard When not nullptr, the card that was authenticated is returned here.
    \param presentcard When not nullptr, the card to compare with.
    \return true if presentcard is nullptr and a card is on the auth list.
    \return false if presentcard is nullptr and none of the cards is on the auth list.
    \return true if presentcard is not nullptr and one of the cards is the same as presentcard.
    \return false if presentcard is not nullptr and none of the card is the same as presentcard.*/
bool CReader::AuthenticateCard(rfalNfcaListenDevice* dev, uint32_t numdevices,
                               rfalNfcaListenDevice* authenticatedcard, rfalNfcaListenDevice* presentcard)
{
  if (presentcard == nullptr)
  { //if presentcard is null, check all devices against the auth list
    for (uint32_t i = 0; i < numdevices; i++)
    {
      if (dev[i].nfcId1Len > NFC_UID_MAX_SIZE)
      {
        PrintLog(LOGREADER, "ERROR: NFC UID with length %" PRIu8 " not supported\n",
                 dev[i].nfcId1Len);
        continue;
      }

      NFCUID uid;
      uid.size = dev[i].nfcId1Len;
      memcpy(uid.uid, dev[i].nfcId1, dev[i].nfcId1Len);

      //feed the UID of this card to the auth list, if it's on the list,
      //the function linked to the card will be executed
      bool keyfailed;
      if (g_controller.AuthList().Authenticate(&uid, &keyfailed))
      {
        if (authenticatedcard)
          *authenticatedcard = dev[i];

        return true;
      }
      else
      { //keep a list of cards that have been denied access
        AddToDenyList(&uid, keyfailed);
      }
    }
  }
  else
  {
    //check all devices against presentcard if its not null
    for (uint32_t i = 0; i < numdevices; i++)
    {
      if (dev[i].nfcId1Len == presentcard->nfcId1Len &&
          std::equal(dev[i].nfcId1, dev[i].nfcId1 + dev[i].nfcId1Len, presentcard->nfcId1))
      {
        if (authenticatedcard)
          *authenticatedcard = dev[i];

        return true;
      }
    }
  }

  return false;
}

void CReader::ClearDenyList()
{
  memset(m_denylist,      0, sizeof(m_denylist));
  memset(m_denykeyfailed, 0, sizeof(m_denykeyfailed));
  m_denylistsize = 0;
}

/*! Used to keep a list of UIDs that have been denied access.
    \param uid The UID to add to the deny list.
    \param keyfailed If true, the card has failed key authentication.*/
void CReader::AddToDenyList(NFCUID* uid, bool keyfailed)
{
  if (m_denylistsize >= MAXDEVICES)
    return;

  //abort if the uid is already present on the deny list
  for (uint32_t i = 0; i < m_denylistsize; i++)
  {
    if (m_denylist[i].size == uid->size)
    {
      if (std::equal(uid->uid, uid->uid + uid->size, m_denylist[i].uid))
        return;
    }
  }

  //add the UID to the deny list
  m_denylist[m_denylistsize]      = *uid;
  m_denykeyfailed[m_denylistsize] = keyfailed;
  m_denylistsize++;
}

/*! Creates a registration for each UID on the deny list.*/
void CReader::RegisterDeniedCards()
{
  for (uint32_t i = 0; i < m_denylistsize; i++)
  {
    g_controller.Registrations().Register(m_denylist + i, DepositResult_Refused,
                                          CardType_Unregistered);

    //If the card failed key authentication, write a DepositRefused status.
    //TODO: this should be written with the card registration instead, but
    //the server doesn't support this.
    if (m_denykeyfailed[i])
      g_controller.Status().Write(Status_DepositRefused,
                                  "secured card failed key authentication");
    else
      g_controller.Status().Write(Status_DepositRefused,
                                  "card not authenticated");
  }
}

/*! Called after a failed read, if the reader signaled a low power card
    detect, but no UID was read, increase the power on no tag counter.*/
void CReader::IncreasePontCounter()
{
  if (m_woke && m_denylistsize == 0)
  {
    if (g_controller.ServerComm().IsContacting())
    { //The GSM antenna might cause interference with the reader coil, therefore
      //power on no tag is not recorded when the server is being contacted.
      PrintLog(LOGREADER, "Not recording power on no tag because server is being contacted\n");
      return;
    }

    PrintLog(LOGREADER, "ERROR: Received low power card detect from reader but no cards read\n");

    //if the reader signaled a low power card detect, but no valid UID was read,
    //and a maximum level for power on no tag is set, increase the power on no tag counter
    uint32_t pontmax = g_controller.Settings().GetInteger(Setting_PowerOnNoTagCounter);
    if (pontmax > 0)
    {
      uint32_t pontcounter = g_controller.Settings().GetInteger(RunStatus_PowerOnNoTag);
      pontcounter++;

      if (pontcounter < pontmax)
      {
        g_controller.Settings().SetInteger(RunStatus_PowerOnNoTag, pontcounter);
      }
      else
      { //record a status of this power on no tag level
        g_controller.Status().Write(Status_PowerOnNoTag, "%" PRIu32, pontcounter);
        g_controller.Settings().SetInteger(RunStatus_PowerOnNoTag, 0);
      }
    }
  }
}

/*! Called after a successful read, if the reader signaled a low power card
    detect, and a UID was read, the reader is working correctly.*/
void CReader::DecreasePontCounter()
{
  if (m_woke)
  {
    uint32_t pontcounter = g_controller.Settings().GetInteger(RunStatus_PowerOnNoTag);
    if (pontcounter > 0)
    {
      pontcounter--;
      g_controller.Settings().SetInteger(RunStatus_PowerOnNoTag, pontcounter);
    }
  }
}

#define DES_KEY_SIZE 16

/*! Generates an HMAC over the UID, using a builtin key. The first part of this
    is then copied into key.
    \param uid The UID of the card to generate a key for.
    \param key Array of 16 bytes to store the key in.*/
void CReader::GenerateUltralightCKey(NFCUID* uid, uint8_t* key)
{
  static const uint8_t builtinkey[] =
    { 0x1f, 0x85, 0xcf, 0xff, 0x54, 0xbb, 0x3e, 0xdd,
      0x36, 0x41, 0xde, 0xe9, 0x95, 0x0e, 0xc0, 0xca };

  //Generate an HMAC.
  uint8_t hmac[SHA256_BLOCK_SIZE];
  HMAC_SHA256_CTX hmacctx;
  HMAC_SHA256_Init_rfc2104(&hmacctx, builtinkey, sizeof(builtinkey));
  HMAC_SHA256_Update(&hmacctx, uid->uid, uid->size);
  HMAC_SHA256_Final(&hmacctx, hmac);

  //Copy the first 16 bytes of the hash and use as a key.
  //The DES parity bits are not calculated,
  //because they're not used by the MIFARE Ultralight C.
  memcpy(key, hmac, DES_KEY_SIZE);
}

//! Number of times to try a prepaid card transaction.
#define PREPAID_CARD_RETRY 10

//! Time in milliseconds to wait after a prepaid transaction failed.
#define PREPAID_FAIL_DELAY 200

/*! Called from CDeposit if a deposit should be done using a prepaid card.
    \param uid The UID of the card that was read.
    \param prepaid Pointer to a CPrepaid instance that has the authlist
                   usage and credit counters loaded.
    \return EPrepaidResult_Success if the deposit was successfully charged
                                   on the prepaid card.
    \return EPrepaidResult_InsufficientCredit if there was not enough credit
                                              to do a deposit.
    \return EPrepaidResult_CardFailure if there was an error charging the deposit
                                       on the prepaid card.
    \return EPrepaidResult_ReaderError if there was an error selecting the card
                                       in the NFC field.
    \return EPrepaidResult_CardNotAuthenticated if the MIFARE Ultralight C card
                                                failed 3DES authentication.*/
EPrepaidResult CReader::ChargePrepaidDeposit(NFCUID* uid, CPrepaid* prepaid)
{
  PrintLog(LOGREADER, "Charging prepaid deposit on card\n");

  CPrepaidBlock  block;
  bool           charged          = false;
  bool           sufficientcredit = false;
  EPrepaidResult result           = EPrepaidResult_None;

  //Try to charge a prepaid deposit on the card a number of times.
  //If there is a failure communicating with the card, turn the field off
  //and try again after a delay.
  for (uint32_t i = 0; i < PREPAID_CARD_RETRY; i++)
  {
    //Enable the NFC field and select the card.
    if (!SelectCard(uid))
    {
      TurnFieldOff(PREPAID_FAIL_DELAY);
      result = EPrepaidResult_ReaderError;
      continue;
    }

    //Perform 3DES authentication.
    if (!AuthenticateUltralightC(uid))
    {
      TurnFieldOff(PREPAID_FAIL_DELAY);
      result = EPrepaidResult_CardNotAuthenticated;
      continue;
    }

    //If no prepaid data has been read from the card yet, read the prepaid data
    //and charge a prepaid deposit on the card.
    if (!charged)
    {
      if (!ReadCardPrepaidData(uid, prepaid, &block))
      {
        TurnFieldOff(PREPAID_FAIL_DELAY);
        result = EPrepaidResult_CardFailure;
        continue;
      }

      //Charge the deposit on the card. Block now holds encrypted data that should
      //be written back to the card. Even if the card does not have enough value,
      //a new transaction number is generated that should be written to the card.
      block.Reset();
      if (prepaid->Charge(uid, &block))
        sufficientcredit = true;

      //A prepaid deposit should only be charged once, and should not be repeated
      //if writing data to the card fails.
      charged = true;
    }

    //Write block to the oldest bank on the card. If this succeeds, the prepaid
    //transaction is completed.
    if (WriteCardPrepaidData(uid, prepaid, &block))
    {
      rfalFieldOff();
      if (sufficientcredit)
      {
        PrintLog(LOGREADER, "Charging prepaid deposit on card successful\n");
        return EPrepaidResult_Success;
      }
      else
      {
        PrintLog(LOGREADER, "Prepaid card does not have enough value for deposit\n");
        return EPrepaidResult_InsufficientCredit;
      }
    }
    else
    {
      TurnFieldOff(PREPAID_FAIL_DELAY);
      result = EPrepaidResult_CardFailure;
    }
  }

  rfalFieldOff();

  PrintLog(LOGREADER, "ERROR: charging prepaid deposit on card failed\n");

  return result; //Return last error code if the number of retries ended.
}

//!< Number of times to try 3DES authentication with a secured card.
#define AUTH_CARD_RETRY 10

//! Time in milliseconds to wait after authentication with a secured card failed.
#define AUTH_FAIL_DELAY 200

/*! Activates the reader, performs 3DES authentication with a
    MIFARE Ultralight C card, and turns off the reader.
    \param uid The UID of the card to authenticate.
    \return true The card authenticated successfully.
    \return false The card failed authentication.*/
bool CReader::AuthenticateSecuredCard(NFCUID* uid)
{
  PrintLog(LOGREADER, "Authenticating secured card\n");

  for (uint32_t i = 0; i < AUTH_CARD_RETRY; i++)
  {
    //Enable the NFC field and select the card.
    if (!SelectCard(uid))
    {
      TurnFieldOff(PREPAID_FAIL_DELAY);
      continue;
    }

    //Perform 3DES authentication.
    if (!AuthenticateUltralightC(uid))
    {
      TurnFieldOff(PREPAID_FAIL_DELAY);
      continue;
    }

    rfalFieldOff();
    return true;
  }

  PrintLog(LOGREADER, "ERROR: authenticating secured card failed\n");

  return false;
}

/*! Calls Authenticate3DES() to perform 3DES authentication with the MIFARE
    Ultralight C card. Prints to the log if the authentication succeeded or failed.
    \param uid The UID of the card to perform authentication with.
    \return true The authentication succeeded.
    \return false The authentication failed.*/
bool CReader::AuthenticateUltralightC(NFCUID* uid)
{
  CLogLock loglock;
  PrintLog(LOGREADER, "Authenticating card ");
  PrintNFCUID(LOGREADER, uid);
  PrintLog(LOGREADER, " using 3DES authentication\n");
  loglock.Leave();

  if (Authenticate3DES(uid))
  {
    loglock.Enter();
    PrintLog(LOGREADER, "Card ");
    PrintNFCUID(LOGREADER, uid);
    PrintLog(LOGREADER, " authenticated successfully\n");
    loglock.Leave();

    return true;
  }
  else
  {
    loglock.Enter();
    PrintLog(LOGREADER, "ERROR: card ");
    PrintNFCUID(LOGREADER, uid);
    PrintLog(LOGREADER, " did not authenticate\n");
    loglock.Leave();

    return false;
  }
}

#define DES_BLOCK_SIZE      8

#define AUTH_CMD            0x1A
#define AUTH_RESPONSE       0xAF
#define AUTH_RESPONSE_SIZE (DES_BLOCK_SIZE + 1)
#define AUTH_COMPLETE       0x00

/*! Authenticates a MIFARE Ultralight C card using the AUTHENTICATE command.
    \param uid The UID of the card to authenticate.
    \return true The card authenticated successfully.
    \return false The card did not authenticate.*/
bool CReader::Authenticate3DES(NFCUID* uid)
{
  //Generate a 3DES key from the UID of the card.
  uint8_t key[DES_KEY_SIZE] = {};
  GenerateUltralightCKey(uid, key);

  //Set up 3DES crypto using the standard MIFARE Ultralight C key.
  des3_setup(key, sizeof(key), 0, (symmetric_key*)&m_des3key);

  uint8_t  authcmd[2]                = {AUTH_CMD, 0x00};
  uint32_t timeout                   = rfalConvMsTo1fc(5U);
  uint8_t  reply[AUTH_RESPONSE_SIZE] = {};
  uint16_t replylen                  = 0;

  //Send the MIFARE Ultralight C AUTHENTICATE command to the card.
  ReturnCode retval;
  retval = rfalTransceiveBlockingTxRx(authcmd, sizeof(authcmd), reply,
                                      sizeof(reply), &replylen,
                                      RFAL_TXRX_FLAGS_DEFAULT, timeout);
  //Check if the card replied correctly.
  if (retval != ERR_NONE || replylen != AUTH_RESPONSE_SIZE || reply[0] != AUTH_RESPONSE)
  {
    PrintUltralightCError(uid, retval, authcmd, sizeof(authcmd), reply,
                          std::min(replylen, (uint16_t)sizeof(reply)));
    PrintLog(LOGREADER, "ERROR: card is probably not a MIFARE Ultralight C card\n");
    return false;
  }

  //The card has generated a random number RndB,
  //and has encrypted it with the key to create ek(RndB).
  uint8_t ekRndB[DES_BLOCK_SIZE] = {};
  memcpy(ekRndB, reply + 1, DES_BLOCK_SIZE);

  //Decrypt ekRndB to create RndB.
  uint8_t RndB[DES_BLOCK_SIZE] = {};
  des3_ecb_decrypt(ekRndB, RndB, (symmetric_key*)&m_des3key);

  //Rotate RndB one byte to the left to create RndB'
  uint8_t RndBd[DES_BLOCK_SIZE] = {};
  for (uint32_t i = 0; i < DES_BLOCK_SIZE; i++)
    RndBd[i] = RndB[(i + 1) % DES_BLOCK_SIZE];

  //Create a random number RndA.
  uint8_t RndA[DES_BLOCK_SIZE] = {};
  CreateIV(RndA, DES_BLOCK_SIZE);

  //Encrypt RndA and RndB in CBC mode using ekRndB as IV,
  //and store it in the ekRndApRndBd buffer (ek(RndA plus RndB')).
  //The ekRndApRndBd buffer starts with AUTH_RESPONSE, so that it can be used
  //as a command to send to the MIFARE Ultralight C card.
  uint8_t  ekRndApRndBd[DES_BLOCK_SIZE * 2 + 1] = {AUTH_RESPONSE};
  uint8_t* IV = ekRndB;
  uint8_t  buf[DES_BLOCK_SIZE] = {};

  //XOR RndA with the IV and encrypt.
  for (uint32_t i = 0; i < DES_BLOCK_SIZE; i++)
    buf[i] = IV[i] ^ RndA[i];

  des3_ecb_encrypt(buf, ekRndApRndBd + 1, (symmetric_key*)&m_des3key);

  //Xor the previous ciphertext with RndBd and encrypt.
  for (uint32_t i = 0; i < DES_BLOCK_SIZE; i++)
    buf[i] = ekRndApRndBd[i + 1] ^ RndBd[i];

  des3_ecb_encrypt(buf, ekRndApRndBd + DES_BLOCK_SIZE + 1, (symmetric_key*)&m_des3key);

  //Send the command to the card. The card will use this data to verify if
  //this device has used the same 3DES key as programmed into the card.
  memset(reply, 0, sizeof(reply));
  replylen = 0;
  retval = rfalTransceiveBlockingTxRx(ekRndApRndBd, sizeof(ekRndApRndBd), reply,
                                      sizeof(reply), &replylen,
                                      RFAL_TXRX_FLAGS_DEFAULT, timeout);

  //Check if the card replied correctly.
  if (retval != ERR_NONE || replylen != AUTH_RESPONSE_SIZE || reply[0] != AUTH_COMPLETE)
  {
    PrintUltralightCError(uid, retval, ekRndApRndBd, sizeof(ekRndApRndBd), reply,
                          std::min(replylen, (uint16_t)sizeof(reply)));
    PrintLog(LOGREADER, "ERROR: card probably has the wrong 3DES key\n");
    return false;
  }

  //The card has replied with ek(RndA'), encrypted in CBC mode using the last
  //8 bytes of ekRndApRndBd as IV. Decrypt it and XOR with the IV to get back
  //RndA' generated by the card.
  IV = ekRndApRndBd + DES_BLOCK_SIZE + 1;
  uint8_t RndAd[DES_BLOCK_SIZE] = {};
  des3_ecb_decrypt(reply + 1, RndAd, (symmetric_key*)&m_des3key);
  for (uint32_t i = 0; i < DES_BLOCK_SIZE; i++)
    RndAd[i] ^= IV[i];

  //Check if the reply from the card RndA' is the same as the generated RndA.
  //This code should run in constant time. If the reply is correct, this card
  //has the correct 3DES key programmed in.
  bool correct = true;
  for (uint32_t i = 0; i < DES_BLOCK_SIZE; i++)
    correct = (RndAd[i] == RndA[(i + 1) % DES_BLOCK_SIZE]) && correct;

  if (!correct)
    PrintLog(LOGREADER, "ERROR: card failed 3DES authentication\n");

  return correct;
}

/*! Prints a log error describing the command sent to the card,
    and the reply received.
    \param uid The UID of the card to which a command was sent.
    \param retval The return value of rfalTransceiveBlockingTxRx().
    \param cmd The command that was sent to the card.
    \param cmdlen The length of cmd.
    \param reply The reply the card sent.
    \param replylen the length of reply.*/
void CReader::PrintUltralightCError(NFCUID* uid, ReturnCode retval, uint8_t* cmd,
                                    uint32_t cmdlen, uint8_t* reply,
                                    uint32_t replylen)
{
  CLogLock loglock;
  PrintLog(LOGREADER, "ERROR: card ");
  PrintNFCUID(LOGREADER, uid);

  if (retval != ERR_NONE)
  {
    PrintLog(LOGREADER, " aborted with error code %i", (int)retval);
  }
  else if (replylen > 0)
  {
    PrintLog(LOGREADER, " replied with incorrect reply ");
    PrintLogHexBuf(LOGREADER, reply, replylen);
  }
  else
  {
    PrintLog(LOGREADER, " replied with empty reply");
  }

  PrintLog(LOGREADER, " to command ");
  PrintLogHexBuf(LOGREADER, cmd, cmdlen, "\n");
}

/*! Enables the field, and selects the card in the field so that it can be used
    for reading and writing.
    \param uid The UID of the card to select.
    \return true Selecting the card succeeded.
    \return false Selecting the card failed.*/
bool CReader::SelectCard(NFCUID* uid)
{
  PrintLog(LOGREADER, "Activating NFC field and selecting card\n");

  //Initialize the reader.
  RETURN_ON_ERROR(rfalNfcInitialize, (), ERR_NONE, false);
  RETURN_ON_ERROR(rfalNfcaPollerInitialize, (), ERR_NONE, false);
  RETURN_ON_ERROR(rfalFieldOnAndStartGT, (), ERR_NONE, false);

  rfalWorker();

  //Check for any devices in the field.
  rfalNfcaSensRes sensres;
  RETURN_ON_ERROR(rfalNfcaPollerCheckPresence, (RFAL_14443A_SHORTFRAME_CMD_REQA, &sensres),
                  ERR_NONE, false);

  //The UID is reversed by the ReadCards() method.
  //Reverse it again so that it's correct for the rfalNfcaPollerSelect() function.
  uint8_t buf[NFC_UID_MAX_SIZE];
  ReverseNFCUID(uid, buf);

  //Select the card so that its memory can be read and written.
  rfalNfcaSelRes selres = {};
  RETURN_ON_ERROR(rfalNfcaPollerSelect, (buf, uid->size, &selres), ERR_NONE, false);

  return true;
}

/*! Turns off the NFC field, then waits for the requested time.
    \param delayms Wait for this number of milliseconds.*/
void CReader::TurnFieldOff(uint32_t delayms)
{
  rfalFieldOff();

  if (delayms > 0)
    DelayMS(delayms);
}

/*! Reads both banks from the MIFARE Ultralight C, and interprets the prepaid data.
    param uid The UID of the card to read from.
    \param prepaid Pointer to CPrepaid that will interpret the card data.
    \param block Buffer to read card prepaid data into.
    \return true Reading both banks succeeded.
    \return false Reading a bank failed.*/
bool CReader::ReadCardPrepaidData(NFCUID* uid, CPrepaid* prepaid, CPrepaidBlock* block)
{
  //Data on the card is divided into 2 banks by this firmware.
  for (uint32_t bank = 0; bank < 2; bank++)
  {
    if (ReadBank(uid, bank, block))
    {
      //Decrypt and interpret the bank data. If the bank data is incorrect it's
      //discarded, and instead the data from the auth list is used. This way if the
      //card memory is corrupted, the prepaid card will still work.
      prepaid->FromCard(uid, bank, block);
    }
    else
    {
      return false;
    }
  }

  return true;
}

/*! Writes a block of prepaid data to the bank on the card that holds the oldest
    prepaid transaction.
    \param uid The UID of the card to write to.
    \param prepaid Pointer to CPrepaid that holds prepaid information
    \param block Pointer to CPrepaidBlock that holds encrypted prepaid data that
                 should be written to the card.*/
bool CReader::WriteCardPrepaidData(NFCUID* uid, CPrepaid* prepaid, CPrepaidBlock* block)
{
  uint32_t bank = prepaid->WriteBankIndex();
  return WriteBank(uid, bank, block);
}

/* The Mifare Ultralight C can store userdata in page 4 to 39, each page is 4 bytes.
   A total of 144 bytes of userdata can be stored. The userdata is divided by firmware
   into 2 banks, each bank is 72 bytes. On each bank, a CPrepaidBlock is stored,
   which holds encrypted and signed prepaid data.*/

//! Start page of bank 0 and 1.
#define PREPAID_STARTPAGE(bank) (bank == 0 ? 4 : 22)

//! Number of bytes per page.
#define PREPAID_PAGESIZE 4

//! Number of pages per bank.
#define PREPAID_NUMPAGES (sizeof(CPrepaidBlock) / PREPAID_PAGESIZE)

//! Number of pages read per read request.
#define PREPAID_READSIZE 4

/*! Reads a bank of data from a card.
    \param uid The uid of the card from which should be read.
    \param bank The bank to read, can only be 0 or 1.
    \param block The data read from the card is returned here.
    \return true If the data was successfully read.
    \return false If the data could not be read.*/
bool CReader::ReadBank(NFCUID* uid, uint32_t bank, CPrepaidBlock* block)
{
  assert(bank == 0 || bank == 1);

  //Read a full bank from the card and store it in *block.
  block->Reset();
  uint32_t startpage = PREPAID_STARTPAGE(bank);
  for (uint32_t i = 0; i < PREPAID_NUMPAGES; i += PREPAID_READSIZE)
  {
    //Read data from the card.
    uint16_t   rcvlen = 0;
    RETURN_ON_ERROR(rfalT2TPollerRead, (i + startpage, *block + i * PREPAID_PAGESIZE,
                                        RFAL_T2T_READ_DATA_LEN, &rcvlen), ERR_NONE, false);

    //Check if enough data was read.
    if (rcvlen != RFAL_T2T_READ_DATA_LEN)
    {
      CLogLock loglock;
      PrintLog(LOGREADER, "ERROR: requested to read %i bytes from card ",
               RFAL_T2T_READ_DATA_LEN);
      PrintNFCUID(LOGREADER, uid);
      PrintLog(LOGREADER, " but read only %" PRIu16 " bytes\n", rcvlen);
      loglock.Leave();

      return false;
    }

    rfalWorker();
  }

  CLogLock loglock;
  PrintLog(LOGREADER, "Read bank %" PRIu32 " from card ", bank);
  PrintNFCUID(LOGREADER, uid, "\n");
  loglock.Leave();

  return true;
}

/*! Writes a bank of data to the card.
    \param uid The UID of the card to which a bank should be written.
    \param bank To which bank should be written, can only be 0 or 1.
    \param block The block of data that will be written to the card.
    \return true If the data was written successfully.
    \return false If the data could not be written.*/
bool CReader::WriteBank(NFCUID* uid, uint32_t bank, CPrepaidBlock* block)
{
  //Write a full bank to the card.
  uint32_t startpage = PREPAID_STARTPAGE(bank);
  for (uint32_t i = 0; i < PREPAID_NUMPAGES; i++)
  {
    RETURN_ON_ERROR(rfalT2TPollerWrite, (i + startpage,
                    *block + i * RFAL_T2T_WRITE_DATA_LEN), ERR_NONE, false);

    rfalWorker();
  }

  CLogLock loglock;
  PrintLog(LOGREADER, "Wrote bank %" PRIu32 " to card ", bank);
  PrintNFCUID(LOGREADER, uid, "\n");
  loglock.Leave();

  return true;
}

//! Number of times to try programming a MIFARE Ultralight C.
#define PROGRAM_RETRY 10

/*! Programs a MIFARE Ultralight Card to be used as a secured card.
    Will show to the user if the programming succeeded or failed.
    \param uid The UID of the card to program.
    \return true Programming a card was attempted.
    \return false Card programming is disabled.*/
bool CReader::ProgramSecuredCard(NFCUID* uid)
{
#if !ULTRALIGHT_C_PROGRAMMER_MODE
  //Return false because secure card programming is disabled.
  return false;
#endif

  CLogLock loglock;
  PrintLog(LOGREADER, "Programming secured card ");
  PrintNFCUID(LOGREADER, uid, "\n");
  loglock.Leave();

  CSpeaker::Beep(100);

  //Program a MIFARE Ultralight C as prepaid card.
  bool success = false;
  for (uint32_t i = 0; i < PROGRAM_RETRY; i++)
  {
    success = ProgramUltralightC(uid);
    if (success)
      break;
    else
      TurnFieldOff(100);
  }

  //Show to the user if the programming succeeded or failed.
  if (success)
  {
    CLogLock loglock;
    PrintLog(LOGREADER, "Programmed secured card ");
    PrintNFCUID(LOGREADER, uid, " successfully\n");
    loglock.Leave();

    g_controller.UserInterface().Set(UIState_Success);
    CSpeaker::Beep(100);
    DelayMS(100);
  }
  else
  {
    CLogLock loglock;
    PrintLog(LOGREADER, "ERROR: programming secured card ");
    PrintNFCUID(LOGREADER, uid, " failed\n");
    loglock.Leave();

    g_controller.UserInterface().SetError(EC_CantProgramCard);
    CSpeaker::Beep(10000);
    DelayMS(10000);
  }

  //Register the programming as a deposit.
  EDepositResult result = success ? DepositResult_Opened : DepositResult_Closed;
  g_controller.Registrations().Register(uid, result, CardType_Secured_Program);

  //Return true because an attempt at programming the card was made.
  //Doesn't matter if it succeeded or failed.
  return true;
}

#define ULTRALIGHT_C_KEY_STARTPAGE  0x2C //!< Start page where 3DES key is stored.
#define ULTRALIGHT_C_KEY_NUMPAGES      4

#define ULTRALIGHT_C_LOCK_PAGE      0x02 //!< Page of lock bytes 0 and 1.

#define ULTRALIGHT_C_AUTH0_PAGE     0x2A //!< First byte is AUTH0.
#define ULTRALIGHT_C_AUTH1_PAGE     0x2B //!< First byte is AUTH1.

#define ULTRALIGHT_C_DATA_STARTPAGE 0x04 //<! First page of user definable memory.
#define ULTRALIGHT_C_DATA_NUMPAGES    36

/*! Programs a 3DES key into a MIFARE Ultralight C card.
    \param uid The UID of the card to program.
    \return true Programming the card succeeded.
    \return false Programming the card failed.*/
bool CReader::ProgramUltralightC(NFCUID* uid)
{
  //Enable the NFC field and select the card.
  if (!SelectCard(uid))
    return false;

  //Authenticate the card, in case this card is already programmed,
  //this will unlock the card.
  //If the card is not programmed, the authentication will fail.
  if (!AuthenticateUltralightC(uid))
  {
    //If the authentication failed, reset the card by turning the field off and on,
    //and selecting the card again. Otherwise the card will not work.
    TurnFieldOff(100);
    if (!SelectCard(uid))
      return false;
  }

  //Initialize its memory so that it can be used for prepaid.
  if (!ProgramUltralightCMemory(uid))
    return false;

  //Generate a 3DES key from the UID of the card.
  uint8_t key[DES_KEY_SIZE] = {};
  GenerateUltralightCKey(uid, key);

  //The MIFARE Ultralight C uses two DES keys, which are stored in reverse
  //byte order on the card.
  uint8_t cardkey[DES_KEY_SIZE] = {};
  for (uint32_t i = 0; i < DES_BLOCK_SIZE; i++)
  {
    cardkey[i] = key[DES_BLOCK_SIZE - i - 1];
    cardkey[DES_BLOCK_SIZE + i] = key[DES_BLOCK_SIZE + DES_BLOCK_SIZE - i - 1];
  }

  PrintLog(LOGREADER, "Programming 3DES key\n");

  //Write the 3DES key to the card.
  for (uint32_t i = 0; i < ULTRALIGHT_C_KEY_NUMPAGES; i++)
  {
    RETURN_ON_ERROR(rfalT2TPollerWrite, (ULTRALIGHT_C_KEY_STARTPAGE + i,
                    cardkey + RFAL_T2T_WRITE_DATA_LEN * i), ERR_NONE, false);
    rfalWorker();
  }

  PrintLog(LOGREADER, "Programming lock bytes\n");

  //Set lock byte 0 to 0x07, and lock byte 1 to 0x00.
  //This locks all memory except the OTP to read/write mode.
  //The reason for this is that page 2 can't be protected with authentication,
  //and only an authenticated device should be allowed to
  //read and write to the card.
  uint8_t lock[4] = {0x00, 0x00, 0x07, 0x00};
  RETURN_ON_ERROR(rfalT2TPollerWrite, (ULTRALIGHT_C_LOCK_PAGE, lock),
                  ERR_NONE, false);

  PrintLog(LOGREADER, "Programming AUTH1\n");

  //Set AUTH1 to 0x00, this restricts read and write access to pages starting at
  //the address programmed in AUTH0.
  uint8_t auth1[4] = {0x00, 0x00, 0x00, 0x00};
  RETURN_ON_ERROR(rfalT2TPollerWrite, (ULTRALIGHT_C_AUTH1_PAGE, auth1),
                  ERR_NONE, false);

  PrintLog(LOGREADER, "Programming AUTH0\n");

  //Set AUTH0 to 0x03, this is the lowest possible value for AUTH0. This means
  //that any devices that wants to read/write at address 0x03 or higher needs to
  //to 3DES authentication with the card first.
  //For this reason AUTH0 should always be written last.
  uint8_t auth0[4] = {0x03, 0x00, 0x00, 0x00};
  RETURN_ON_ERROR(rfalT2TPollerWrite, (ULTRALIGHT_C_AUTH0_PAGE, auth0),
                  ERR_NONE, false);

  //Reset the card so that the authentication can be tested.
  DelayMS(10);
  TurnFieldOff(100);
  if (!SelectCard(uid))
    return false;

  //Test if the card can be authenticated, if that fails the card has not been
  //programmed correctly, or the card is not a MIFARE Ultralight C card.
  if (!AuthenticateUltralightC(uid))
    return false;

  rfalFieldOff();

  return true;
}

/*! First initializes all user memory to zero, then initializes the user memory
    as two prepaid banks.
    \param uid The UID of the card to write the memory of.
    \return true Programming the card succeeded.
    \return false Programming the card failed.*/
bool CReader::ProgramUltralightCMemory(NFCUID* uid)
{
  //The user memory is in an undefined state from the factory.
  //Set all user memory to zero.
  PrintLog(LOGREADER, "Erasing card user memory\n");
  static const uint8_t emptypage[RFAL_T2T_WRITE_DATA_LEN] = {};
  for (uint32_t i = 0; i < ULTRALIGHT_C_DATA_NUMPAGES; i++)
  {
    RETURN_ON_ERROR(rfalT2TPollerWrite, (ULTRALIGHT_C_DATA_STARTPAGE + i,
                    emptypage), ERR_NONE, false);
    rfalWorker();
  }

  //Initialize the user memory as two prepaid banks.
  PrintLog(LOGREADER, "Initializing card user memory for prepaid\n");
  for (uint32_t bank = 0; bank <= 1; bank++)
  {
    //Initialize, sign and encrypt a prepaid bank.
    CPrepaidBlock    writeblock;
    CPrepaidCounters cardcounters;
    cardcounters.SetTransaction(bank); //Use bank number as transaction number.
    CPrepaid::Encrypt(uid, &cardcounters, &writeblock);

    //Write the bank to the card.
    if (!WriteBank(uid, bank, &writeblock))
    {
      CLogLock loglock;
      PrintLog(LOGREADER, "ERROR: unable to write to bank %" PRIu32 " on card ", bank);
      PrintNFCUID(LOGREADER, uid, "\n");
      loglock.Leave();

      return false;
    }

    //Read the bank back from the card.
    CPrepaidBlock readblock;
    if (!ReadBank(uid, bank, &readblock))
    {
      CLogLock loglock;
      PrintLog(LOGREADER, "ERROR: unable to write to bank %" PRIu32 " on card ", bank);
      PrintNFCUID(LOGREADER, uid, "\n");
      loglock.Leave();

      return false;
    }

    //Check if the bank written to the card is the same as the bank read back
    //from the card.
    if (memcmp(writeblock, readblock, sizeof(writeblock)) != 0)
    {
      CLogLock loglock;
      PrintLog(LOGREADER, "ERROR: bank %" PRIu32 " written to card is not the "
               "same as read back from the card:\n", bank);

      PrintLog(LOGREADER, "Written: ");
      PrintLogHexBuf(LOGREADER, writeblock, sizeof(writeblock), "\n");
      PrintLog(LOGREADER, "Read:    ");
      PrintLogHexBuf(LOGREADER, readblock, sizeof(readblock), "\n");
      loglock.Leave();

      return false;
    }

    PrintLog(LOGREADER, "Data written to bank %" PRIu32
             " is the same as what has been read back\n", bank);
  }

  return true;
}
