#include "registrations.h"
#include "controller.h"
#include "log.h"
#include "clock.h"
#include "cbortools.h"
#include "lock.h"
#include "settings.h"
#include "active.h"

#include <cstring>

CRegistrations::CRegistrations()
{
}

void CRegistrations::Setup()
{
  //create a mutex to protect m_flashringbuffer and m_buf
  m_mutex = xSemaphoreCreateRecursiveMutex();
}

/*! Sets up the SPI flash ringbuffer, this might take a few seconds.*/
void CRegistrations::SetupRingBuffer()
{
  CLock lock(m_mutex);

  if (!m_flashringbuffer.Enabled())
  {
    //put the SPI flash ringbuffer at the end of the SPI flash
    //the stm32wss library has some space reserved at the start of the SPI flash
    //by putting this at the end, a collision is avoided
    uint32_t position = g_controller.SPIFlash().Size() - REG_RINGBUFFER_TOTAL_SIZE - REG_RINGBUFFER_ADDRESS;
    m_flashringbuffer.Setup(&g_controller.SPIFlash(), position,
                            REG_RINGBUFFER_TOTAL_SIZE, REG_RINGBUFFER_ELEMENT_SIZE);
  }
}

/*! After each deposit, a registration of the deposit is written to the SPI flash.
    A counter is kept in internal flash, which is increased after the deposit.
    That counter value is written to the registration and is finally sent to the server.
    This counter is used to prove that no registration is ever missed.*/
void CRegistrations::IncreaseRegistrationCounter()
{
  uint32_t regcount = g_controller.Settings().GetInteger(RunStatus_RegistrationCounter);
  g_controller.Settings().SetInteger(RunStatus_RegistrationCounter, regcount + 1);
}

/*! Writes a registration of an authenticated card to the SPI flash.
    \param uid The UID of the NFC tag used to authenticate.
    \param result The result of the deposit.
    \param cardtype The type of card this is stored as in the auth list.
    \param prepaid Pointer to a CPrepaid instance if a prepaid deposit has been done.
    \param input The input a deposit is detected on, 0 for input1, 1 for input2.*/
void CRegistrations::Register(NFCUID* uid, EDepositResult result, ECardType cardtype, CPrepaid* prepaid /*= nullptr*/, int32_t input /* = -1*/)
{
  if (result == DepositResult_DontRegister || !DeviceActive())
  {
    CLogLock loglock;
    PrintLog(LOGREG, "Not registering deposit for uid ");
    PrintNFCUID(LOGREG, uid);
    PrintLog(LOGREG, "\n");
    return;
  }

  IncreaseRegistrationCounter();

  CLock lock(m_mutex);

  //set up the SPI flash ringbuffer here, because it might take a few seconds,
  //at this point, the lock has been opened and closed, and the user has opened
  //the deposit opening, so any delay will not be noticed
  SetupRingBuffer();

  memset(m_buf, 0, sizeof(m_buf));

  uint32_t maxbufbytes = MAX_RINGBUFFER_PAYLOAD_SIZE(REG_RINGBUFFER_ELEMENT_SIZE);
  CborEncoder encoder;
  cbor_encoder_init(&encoder, m_buf, maxbufbytes, 0);

  //create a registration map to add all other elements
  CborEncoder registration;
  cbor_encoder_create_map(&encoder, &registration, CborIndefiniteLength);

  //For regular deposits, if Setting_Data_anonymous is enabled,
  //the UID of the card should not be registered.
  bool anonymous = false;
  if (CAuthList::FromSecuredType(cardtype) == CardType_Registration &&
      g_controller.Settings().GetBoolean(Setting_Data_anonymous))
    anonymous = true;

  //Convert the UID to a hex string,
  //or an empty string in case it should be anonymous.
  char uidstr[NFC_UID_MAX_SIZE * 2 + 1] = {};
  if (!anonymous)
  {
    for (uint32_t i = 0; i < uid->size; i++)
      sprintf(uidstr + i * 2, "%02" PRIX8, uid->uid[i]);
  }

  //add the UID as the T field
  cbor_encode_text_stringz(&registration, "T");
  cbor_encode_text_stringz(&registration, uidstr);

  //add the deposit result as the S field
  cbor_encode_text_stringz(&registration, "S");
  cbor_encode_int(&registration, result);

  //add the date as the D field, don't add a timezone to the date
  char datebuf[DATESTR_MAX_LENGTH] = {};
  GetRTCISO8601(datebuf, DATESTR_MAX_LENGTH, "");
  cbor_encode_text_stringz(&registration, "D");
  cbor_encode_text_stringz(&registration, datebuf);

  //add the card type as the A field
  cbor_encode_text_stringz(&registration, "A");
  cbor_encode_int(&registration, cardtype);

  //add the registration counter as the N field
  uint32_t regcount = g_controller.Settings().GetInteger(RunStatus_RegistrationCounter);
  cbor_encode_text_stringz(&registration, "N");
  cbor_encode_int(&registration, regcount);

  if (input != -1)
  {
    //if a deposit was detected on an input, register on which input the
    //deposit happened, 1 for input1, 2 for input2
    cbor_encode_text_stringz(&registration, "I");
    cbor_encode_int(&registration, input + 1);

    //register the fill level increase belonging to the deposit input
    cbor_encode_text_stringz(&registration, "V");
    cbor_encode_float(&registration, g_controller.Deposit().FillLevelIncrease(input));
  }

  //If a deposit has been done using a prepaid card, and credit was charged on the card,
  //add fields to the registration to inform the server about the charge.
  if (cardtype == CardType_Prepaid && prepaid && prepaid->ResultValid())
  {
    //Add the card transaction number as the X field.
    cbor_encode_text_stringz(&registration, "X");
    cbor_encode_int(&registration, prepaid->ResultTransaction());

    //Add the credit of the card as the R field.
    cbor_encode_text_stringz(&registration, "R");
    cbor_encode_int(&registration, prepaid->ResultCredit());

    //Add the used counter as the U field.
    cbor_encode_text_stringz(&registration, "U");
    cbor_encode_int(&registration, prepaid->ResultUsed());

    //Add the cost of the deposit as the C field.
    cbor_encode_text_stringz(&registration, "C");
    cbor_encode_int(&registration, prepaid->ResultCost());
  }

  //close the registration object and check if everything was encoded correctly
  if (!CBORCloseContainer(LOGREG, "registration", &encoder, &registration))
    return;

  //print the generated CBOR to the log
  uint32_t cborsize = cbor_encoder_get_buffer_size(&encoder, m_buf);
  if (!CBORPrint(LOGREG, "Registration: ", m_buf, cborsize))
    return;

  //store the generated CBOR registration in the flash ringbuffer
  if (!m_flashringbuffer.Write(m_buf, cborsize))
    PrintLog(LOGREG, "ERROR: storing CBOR registration of size %"
             PRIu32 " failed\n", cborsize);

  uint32_t numstored = NumStored();
  PrintLog(LOGREG, "%" PRIu32 " registration%s stored\n",
           numstored, (numstored == 0 || numstored > 1) ? "s" : "");
}

/*! Called from the servercomm task, reads registrations from the flash ringbuffer,
    and adds them to the registrations array.
    \param registrations The CBOR array to add the registrations to.
    \param buf The buffer that the CBOR encoder was created with.
    \param bufsize The size of buf.
    \return The number of registrations written to the registrations array.*/
uint32_t CRegistrations::Read(CborEncoder* registrations, uint8_t* buf, uint32_t bufsize)
{
  CLock lock(m_mutex);
  SetupRingBuffer();

  uint32_t numregistrations = 0;
  for(;;)
  {
    //check how much free space there is in the CBOR buffer
    //always make sure there is at least one byte free for the CBOR break control code
    uint32_t freespace = bufsize - cbor_encoder_get_buffer_size(registrations, buf);
    if (freespace > 0)
      freespace--;

    //read one registration from the ringbuffer
    memset(m_buf, 0, sizeof(m_buf));
    uint32_t size;
    if (!m_flashringbuffer.Read(m_buf, &size, false))
      break;

    //if the registration doesn't fit in the CBOR buffer, stop
    if (size >= freespace)
      break;

    //copy the registration directly into the CBOR array
    memcpy(registrations->data.ptr, m_buf, size);
    registrations->data.ptr += size;
    numregistrations++;
    m_flashringbuffer.AdvanceReadPosition();
  }

  return numregistrations;
}

/*! Returns the number of registrations stored in the flash ringbuffer.*/
uint32_t CRegistrations::NumStored()
{
  CLock lock(m_mutex);
  SetupRingBuffer();
  return m_flashringbuffer.NumStoredElements();
}

/*! Called from the servercomm task, clears registrations that have been read
    and advanced beyond in the flash ringbuffer.
    This is called after the servercomm task has sent the registrations to the server,
    and the server has received them.*/
void CRegistrations::Clear()
{
  CLock lock(m_mutex);
  m_flashringbuffer.ClearReadElements();
}

/*! Called from the servercomm task, resets the read counter of the
    flash ringbuffer, in case registrations have been read, but the server did not
    receive them.*/
void CRegistrations::Reset()
{
  CLock lock(m_mutex);
  m_flashringbuffer.ResetReadCounter();
}
