#include "selftest.h"
#include "gpiodescriptions.h"
#include "log.h"
#include "backupregisters.h"
#include "lock.h"
#include "clock.h"
#include "controller.h"
#include "speaker.h"
#include "active.h"

#include <cstring>

#define SELF_TEST_ENABLED_FLAG 0x10000000
#define SELF_TEST_PASSED_REG   RTC_BKP_DR15

//When a board has a fault, certain self tests will cause the microcontroller
//to reboot before the log line describing the self test is written.
//For those log lines, WaitLogWriteComplete(LOG_WAIT_TIMEOUT) is called afterwards,
//to make sure that the log line is printed.
#define LOG_WAIT_TIMEOUT 1000

//Maximum bits in a backup register.
static_assert(SelfTest_MAX <= 32);

CSelfTest::CSelfTest()
{
  m_enabled = false;
  m_mutex   = nullptr;
}

CSelfTest::~CSelfTest()
{
}

/*! Reads the BOOT1 pin, and enables self testing if it's high.*/
void CSelfTest::Setup()
{
  m_enabled = GPIO_ReadInput(BOOT1);

  if (!m_enabled)
  {
    PrintLog(LOGSELFTEST, "Self testing disabled\n");

    if (ReadBackupFlag(SELF_TEST_ENABLED_FLAG))
      OnSelfTestEnd();

    return;
  }

  PrintLog(LOGSELFTEST, "Self testing enabled\n");

  m_mutex = xSemaphoreCreateRecursiveMutex();

  //Indicate that test mode is enabled.
  g_controller.GUILeds().On(GUILED_ALL_RED);

  if (!ReadBackupFlag(SELF_TEST_ENABLED_FLAG))
    OnSelfTestStart();
}

/*! Runs tests at boot when test mode is enabled.*/
void CSelfTest::Run()
{
  if (!m_enabled)
    return;

  TestBatteryCharger();
  TestSpeaker();
  TestInternalRTC();
  TestExternalRTC();
  TestSPIFlash();
  TestBackupVoltage();
  TestBoostConverterVoltage();
 // TestExternalIO();
  TestRS485();
  TestServerComm();
  TestGUILeds();
  TestVCC();

  g_controller.UserInterface().Set(UIState_Off);
}

/*! Runs a test on the cable. A test circuit should be attached to the cable.
    This test is not part of the normal self tests, instead it can be used
    to test a cable of a device in the field.*/
void CSelfTest::RunCableTest()
{
  g_controller.UserInterface().SetString("Testing:\ncable");

  bool passed = TestCable();

  //Indicate if the cable passed or failed the test.
  if (passed)
  {
    g_controller.UserInterface().SetString("Cable test:\npassed");
    g_controller.GUILeds().Off(GUILED_ALL_RED);
    g_controller.GUILeds().On(GUILED_ALL_GRN);

    g_controller.Status().Write(Status_CableTestPassed, "1");

    //When device is inactive mode, store that the cable test has been passed.
    if (!DeviceActive())
      g_controller.Settings().SetBoolean(InactiveStatus_CableTest_Passed, true);
  }
  else
  {
    g_controller.UserInterface().SetString("Cable test:\nFAILED");
    g_controller.GUILeds().Off(GUILED_ALL_GRN);
    g_controller.GUILeds().On(GUILED_ALL_RED);
    CSpeaker::Beep(5000);
  }

  DelayMS(5000);

  g_controller.GUILeds().Off(GUILED_ALL);

  g_controller.UserInterface().Set(UIState_Off);
}

/*! Called on the first boot after self testing mode has been enabled.*/
void CSelfTest::OnSelfTestStart()
{
  PrintLog(LOGSELFTEST, "Self test started\n");

  SetBackupFlag(SELF_TEST_ENABLED_FLAG, true);
  WriteBackupRegister(SELF_TEST_PASSED_REG, 0);
}

/*! Called on the first boot after self testing mode has been disabled.*/
void CSelfTest::OnSelfTestEnd()
{
  PrintLog(LOGSELFTEST, "Self test ended\n");

  SetBackupFlag(SELF_TEST_ENABLED_FLAG, false);
  WriteBackupRegister(SELF_TEST_PASSED_REG, 0);
}

/*! \param test The self test value to turn into a string.
    \return A C-string that describes the self test.*/
const char* CSelfTest::TestStr(ESelfTest test)
{
  static const char* teststrs[] =
  {
    "internal RTC",
    "external RTC",
    "SPI flash",
    "modem",
    "server communication",
    "reader coil",
    "NFC reader",
    "NFC low power card detect",
    "backup voltage",
    "battery voltage",
    "external charger voltage",
    "battery charge current",
    "battery charger chemistry",
    "boost converter",
    "external IO",
    "RS485",
    "display",
    "gui leds",
    "VCC",
  };

  if (test >= 0 && (int32_t)test < (int32_t)(sizeof(teststrs) / sizeof(teststrs[0])))
    return teststrs[test];
  else
    return "ERROR UNKNOWN TEST";
}

/*! \param test The self test value to return a description of.
    \return A C-string that describes what is tested.*/
const char* CSelfTest::TestDescr(ESelfTest test)
{
  static const char* teststrs[] =
  {
    "The internal RTC counts forward in time.",
    "The external RTC counts forward in time.",
    "The SPI flash can be erased, written to and read from.",
    "The modem can be turned on and off.",
    "Communication with rsoffice.net has succeeded.",
    "The NFC reader has calibrated the NFC coil correctly.",
    "The NFC reader has read the UID of an NFC card.",
    "The microcontroller has woken from low power mode by the NFC reader, and a card has been read.",
    "The backup capacitor voltage is above 2 volts.",
    "The internal battery voltage is between 3.0 and 3.8 volts.",
    "The external charger voltage is between 5 and 15 volts.",
    "The internal battery charge current is between 0.5 and 1.5 amps.",
    "The battery charger is for LiFePO and has correct cell count",
    "The boost converter makes 5.5 or 12.3 volts on the power rail line",
    "The external inputs, outputs and H-bridges are working correctly.",
    "The RS485 transceiver echoes back transmitted data.",
    "Data can be written to and read from the display memory.",
    "The IO expander that drives the GUI leds communicates over I2C.",
    "The microcontroller's VCC is between 3.2 and 3.4 volts.",
  };

  if (test >= 0 && (int32_t)test < (int32_t)(sizeof(teststrs) / sizeof(teststrs[0])))
    return teststrs[test];
  else
    return "ERROR UNKNOWN TEST";
}

/*! \param test For which test the bit should be returned.
    \return true if the bit in SELF_TEST_PASSED_REG is set for this test.
    \return false if the bit in SELF_TEST_PASSED_REG is not set for this test.*/
bool CSelfTest::IsTestPassed(ESelfTest test)
{
  if (!m_enabled)
    return false;

  CLock lock(m_mutex);
  uint32_t testreg = ReadBackupRegister(SELF_TEST_PASSED_REG);
  return testreg & (1 << test);
}

void CSelfTest::SetTestPassed(ESelfTest test)
{
  if (!m_enabled)
    return;

  CLock lock(m_mutex);

  //Set the bit in SELF_TEST_PASSED_REG that corresponds with the test.
  uint32_t testreg = ReadBackupRegister(SELF_TEST_PASSED_REG);
  if (!(testreg & (1 << test)))
  {
    PrintLog(LOGSELFTEST, "Marking test %s as passed\n", TestStr(test));
    testreg |= (1 << test);
    WriteBackupRegister(SELF_TEST_PASSED_REG, testreg);
  }

  //Record a status if all tests are passed.
  if (AllTestsPassed())
  {
    g_controller.Status().Write(Status_SelfTestPassed, "1");

    //Record for the inactive display that the self test has been passed.
    if (!DeviceActive())
      g_controller.Settings().SetBoolean(InactiveStatus_SelfTest_Passed, true);
  }
}

/* \return true if all tests are passed.
   \return false if one or more tests not passed.*/
bool CSelfTest::AllTestsPassed()
{
  CLock lock(m_mutex);

  if (!m_enabled)
    return false;

  for (ESelfTest test = SelfTest_START; test < SelfTest_MAX; test++)
  {
    if (!IsTestPassed(test))
      return false;
  }

  return true;
}

/*! Prints the results of the self test.*/
void CSelfTest::Print()
{
  if (!m_enabled)
  {
    PrintLog(LOGSELFTEST, "Self testing disabled\n");
    return;
  }

  //Find the longest self test string.
  int fieldlength = 0;
  for (ESelfTest test = SelfTest_START; test < SelfTest_MAX; test++)
  {
    int length = strlen(TestStr(test));
    if (length > fieldlength)
      fieldlength = length;
  }

  //Print the result for each test, with the self test string padded with spaces.
  for (ESelfTest test = SelfTest_START; test < SelfTest_MAX; test++)
  {
    PrintLog(LOGSELFTEST, "%-*s: %-12s %s\n", fieldlength, TestStr(test),
             IsTestPassed(test) ? "passed," : "not passed,",
             TestDescr(test));
  }
}

//! Max time in milliseconds to signal that all tests have passed.
#define SIGNAL_TIME (300 * 1000)

//! Time in milliseconds to blink the signal with.
#define SIGNAL_INTERVAL (500)

/*! Shows the OK image on the display, and blinks the backlight, to indicate
    to the user that all tests have been passed successfully.*/
void CSelfTest::SignalAllTestsPassed()
{
  if (!m_enabled)
    return;

  CLock lock(m_mutex);

  if (!AllTestsPassed())
    return;

  //Indicate that all tests have passed.
  g_controller.GUILeds().Off(GUILED_ALL_RED);
  g_controller.GUILeds().On(GUILED_ALL_GRN);

  bool    backlight = true;
  int64_t start = Millis();
  do
  {
    //Keep signaling while test mode is enabled.
    if (GPIO_ReadInput(BOOT1) == LOW)
      break;

    //Make sure that the success display is visible.
    g_controller.UserInterface().Set(UIState_Success);

    //Blink the backlight to inform the user.
    g_controller.UserInterface().SetBacklight(backlight);
    backlight = !backlight;

    DelayMS(SIGNAL_INTERVAL);
  }
  while (Millis() - start < SIGNAL_TIME);

  g_controller.UserInterface().Set(UIState_Off);
}

/*! Plays a sound on the speaker.*/
void CSelfTest::TestSpeaker()
{
  PrintLog(LOGSELFTEST, "Testing speaker\n");
  WaitLogWriteComplete(LOG_WAIT_TIMEOUT);

  //Beep the speaker 5 times for 100 milliseconds.
  CSpeaker::Beep(100, 5);
}

/*! Tests if the internal battery voltage, the external charger voltage,
    and the battery charge current are within the correct limits.*/
void CSelfTest::TestBatteryCharger()
{
  //Test if the internal battery voltage is within limits.
  if (!IsTestPassed(SelfTest_BatteryVoltage))
  {
    const float vmin = 3.0f;
    const float vmax = 3.8f;
    float       vbat = g_controller.BatteryInfo().m_intvoltage;

    PrintLog(LOGSELFTEST, "Internal battery voltage: %.3f, minimum required: %.3f"
             ", maximum allowed: %.3f\n", vbat, vmin, vmax);

    if (vbat >= vmin && vbat <= vmax)
      SetTestPassed(SelfTest_BatteryVoltage);
  }

  //Test if the external charger voltage is within limits.
  if (!IsTestPassed(SelfTest_ExternalVoltage))
  {
    const float vmin = 5.0f;
    const float vmax = 15.0f;
    float       vext = g_controller.BatteryInfo().m_extvoltage;

    PrintLog(LOGSELFTEST, "External charger voltage: %.3f, minimum required: %.3f"
             ", maximum allowed: %.3f\n", vext, vmin, vmax);

    if (vext >= vmin && vext <= vmax)
      SetTestPassed(SelfTest_ExternalVoltage);
  }

  //Test if the battery is being charged,
  //and that the charge current is within limits.
  if (!IsTestPassed(SelfTest_ChargeCurrent))
  {
    const float imin = 0.5f;
    const float imax = 1.5f;
    float       ibat = g_controller.BatteryInfo().m_chargecurrent;
    bool        chrg = g_controller.BatteryInfo().m_charging;

    PrintLog(LOGSELFTEST, "Battery charge current: %.3f, minimum required: %.3f"
             ", maximum allowed: %.3f, currently charging: %s\n",
             ibat, imin, imax, chrg ? "yes" : "no");

    if (chrg && ibat >= imin && ibat <= imax)
      SetTestPassed(SelfTest_ChargeCurrent);
  }

  //Test if the battery charger is for LiFePO chemistry, and that it
  //has the correct cell count.
  if (!IsTestPassed(SelfTest_ChargerChemistry))
  {
    if (g_controller.BatteryCharger().CheckChemistry())
      SetTestPassed(SelfTest_ChargerChemistry);
  }
}

/*! Tests if the internal RTC runs forward in time.*/
void CSelfTest::TestInternalRTC()
{
  if (IsTestPassed(SelfTest_InternalRTC))
    return;

  g_controller.UserInterface().SetString("Testing:\ninternal RTC");

  if (TestRTC())
    SetTestPassed(SelfTest_InternalRTC);
}

/*! Tests if the external RTC runs forward in time.*/
void CSelfTest::TestExternalRTC()
{
  if (IsTestPassed(SelfTest_ExternalRTC))
    return;

  g_controller.UserInterface().SetString("Testing:\nexternal RTC");

  if (g_controller.ExternalRTC().Test())
    SetTestPassed(SelfTest_ExternalRTC);
}

/*! Tests if the SPI flash can be erased, written to and read from.*/
void CSelfTest::TestSPIFlash()
{
  if (IsTestPassed(SelfTest_SPIFlash))
    return;

  g_controller.UserInterface().SetString("Testing:\nSPI flash");

  if (g_controller.SPIFlash().Test())
    SetTestPassed(SelfTest_SPIFlash);
}

/*! Tests if the backup capacitor voltage is high enough.*/
void CSelfTest::TestBackupVoltage()
{
  if (IsTestPassed(SelfTest_BackupVoltage))
    return;

  float backupvoltage = g_controller.Adc().ReadVbat();
  const float minvoltage = 2.0f;

  PrintLog(LOGSELFTEST, "Backup capacitor voltage: %.3f, minimum required: %.3f\n",
           backupvoltage, minvoltage);

  if (backupvoltage >= minvoltage)
    SetTestPassed(SelfTest_BackupVoltage);
}

/*! Signals a communication attempt if the servercomm test has not passed yet.*/
void CSelfTest::TestServerComm()
{
 if (IsTestPassed(SelfTest_ServerComm))
    return;

  g_controller.ServerComm().SignalCommAttempt(COMM_REASON_SELF_TEST);
}

/*! Tests if the IO expander that drives the GUI LEDs can be communicated with.*/
void CSelfTest::TestGUILeds()
{
  if (IsTestPassed(SelfTest_GUILeds))
    return;

  if (g_controller.GUILeds().Enabled())
    SetTestPassed(SelfTest_GUILeds);
}

#define RS485_USART     USART3
#define RS485_TEST_SIZE 6

/*! Transmits data out the RS485 transceiver,
    and checks if that same data is received back.*/
void CSelfTest::TestRS485()
{
  if (IsTestPassed(SelfTest_RS485))
    return;

  g_controller.UserInterface().SetString("Testing:\nRS485 bus");

  PrintLog(LOGSELFTEST, "Testing RS485\n");
  WaitLogWriteComplete(LOG_WAIT_TIMEOUT);

  CUsart usart;

  USART_InitTypeDef usartinit;
  USART_StructInit(&usartinit);
  usartinit.USART_BaudRate = 115200;
  usart.Setup(RS485_USART, &usartinit, RS485_TEST_SIZE * 2);

  RunRS485Test(usart);

  //Turn off the RS485 transceiver.
  GPIO_WritePin(RS485_PWR_ON, LOW);
  GPIO_WritePin(RS485_RX_EN,  HIGH);
  GPIO_WritePin(RS485_TX_EN,  LOW);
}

/*! Runs a loopback test with all possible states of the 
    RS485_PWR_ON, RS485_RX_EN and RS485_TX_EN pins.*/
void CSelfTest::RunRS485Test(CUsart& usart)
{
  //Test pattern that will be transmitted over RS485.
  static const uint8_t txbuf[RS485_TEST_SIZE] = {0xAA, 0x55, 0x88, 0x44, 0x22, 0x11};

  for (uint32_t i = 0; i < 8; i++)
  {
    bool pwr = i & 1;
    bool rx  = i & 2;
    bool tx  = i & 4;

    PrintLog(LOGSELFTEST, "Testing RS485 with: power enabled:%s rx enabled:%s"
             " tx enabled:%s\n", pwr ? "yes," : "no, ", rx ? "yes," : "no, ",
             tx ? "yes" : "no");

    GPIO_WritePin(RS485_PWR_ON, pwr);
    GPIO_WritePin(RS485_RX_EN,  !rx); //Receive is enabled low.
    GPIO_WritePin(RS485_TX_EN,  tx);
    DelayMS(10);

    //Write the test pattern on the RS485 bus.
    usart.ClearInput();
    usart.Write(txbuf, sizeof(txbuf));

    //Read the test pattern back from the RS485 bus.
    uint8_t  rxbuf[RS485_TEST_SIZE] = {};
    uint32_t bytesread;
    uint32_t result = usart.Read(rxbuf, RS485_TEST_SIZE, &bytesread, 100);

    //Only when the RS485 power, rx and tx pins are enabled, should the uart
    //receive the data it transmitted on the bus.
    if (pwr && rx && tx)
    {
      //Check that all transmitted bytes have been received correctly.
      if (result != READRESULT_SUCCESS)
      {
        PrintLog(LOGSELFTEST, "ERROR: reading back RS485 data: %s\n",
                 usart.ResultToStr(result));
        return;
      }

      //Check that the received pattern is identical to the one that was transmitted.
      if (memcmp(txbuf, rxbuf, RS485_TEST_SIZE) != 0)
      {
        CLogLock loglock;
        PrintLog(LOGSELFTEST, "ERROR: transmitted on RS485: ");
        PrintLogHexBuf(LOGSELFTEST, txbuf, RS485_TEST_SIZE, "\n");
        PrintLog(LOGSELFTEST, "       received on RS485:    ");
        PrintLogHexBuf(LOGSELFTEST, rxbuf, RS485_TEST_SIZE, "\n");
        return;
      }

      CLogLock loglock;
      PrintLog(LOGSELFTEST, "Read back test pattern: ");
      PrintLogHexBuf(LOGSELFTEST, rxbuf, RS485_TEST_SIZE, "\n");
    }
    else
    {
      //No data should have been received. (Edwin says framing error is fine too)
      if (result != READRESULT_ERROR_TIMEOUT && result != READRESULT_ERROR_FRAMING)
      {
        CLogLock loglock;
        PrintLog(LOGSELFTEST, "ERROR: read back %" PRIu32 " bytes over RS485 "
                 "with RS485 bus disabled, uart result: %s, data received: ",
                 bytesread, CUsart::ResultToStr(result)); // EDWIN MOD
        PrintLogHexBuf(LOGSELFTEST, rxbuf, bytesread, "\n");
        return;
      }
    }
  }

  PrintLog(LOGSELFTEST, "RS485 is working correctly\n");
  SetTestPassed(SelfTest_RS485);
}

/* Wiring needed for external IO test:

   BRIDGE1 connected to I1 
   BRIDGE2 connected to I2 
   BRIDGE3 connected to I3 
   BRIDGE4 connected to I4 

   PR_OUT1 connected to BRIDGE1/I1 using a 2k2 resistor. 
   PR_OUT2 connected to BRIDGE2/I2 using a 2k2 resistor. 

   POWERRAIL_OUT/P12V_SW connected to BRIDGE3/I3 and BRIDGE4/I4 using
   2k2 resistors and diodes in series. 
*/

/*! Tests the H-bridges, the switched output and the inputs.
    Expects a test circuit board attached to the IO pins.*/
void CSelfTest::TestExternalIO()
{
  if (IsTestPassed(SelfTest_ExternalIO))
    return;

  g_controller.UserInterface().SetString("Testing:\nexternal IO");

  PrintLog(LOGSELFTEST, "Testing external IO\n");
  WaitLogWriteComplete(LOG_WAIT_TIMEOUT);

  g_controller.Deposit().LockDriver().PowerUpBoostConverter();
  SetupExternalIOPins();

  //Run the external IO test with the test circuit board attached.
  if (TestExternalIOPins(false))
  {
    PrintLog(LOGSELFTEST, "External IO is working correctly\n");
    SetTestPassed(SelfTest_ExternalIO);
  }

  g_controller.Deposit().LockDriver().PowerDownBoostConverter();
  CloseExternalIOPins();
}

/*! Tests if the boost converter voltage is within the correct range.
    It's required that the boost converter is enabled.*/
void CSelfTest::TestBoostConverterVoltage()
{
  if (IsTestPassed(SelfTest_BoostConverter))
    return;

  g_controller.UserInterface().SetString("Testing:\nboost\nconverter");

  PrintLog(LOGSELFTEST, "Testing boost converter\n");
  WaitLogWriteComplete(LOG_WAIT_TIMEOUT);

  g_controller.Deposit().LockDriver().PowerUpBoostConverter();
  DelayMS(1000);

  //Measure POWER_RAIL, which is connected to the ADC with a voltage diviver,
  //made from two 12k resistors and one 3k9 resistor.
  float pwrvoltage = g_controller.Deposit().LockDriver().MeasurePowerRail();

  //Check if the boost converter is within 10% of one of the allowed voltages.
  const float allowed[] = {5.5f, 12.3f};
  bool success = false;
  for (uint32_t i = 0; i < sizeof(allowed) / sizeof(allowed[0]); i++)
  {
    if (pwrvoltage <= allowed[i] * 1.1f && pwrvoltage >= allowed[i] * 0.9f)
    {
      PrintLog(LOGSELFTEST, "Boost converter voltage %.3f is within 10%% of"
               " allowed voltage %.1f\n", pwrvoltage, allowed[i]);

      success = true;
      SetTestPassed(SelfTest_BoostConverter);
    }
  }

  if (!success)
  {
    CLogLock loglock;
    PrintLog(LOGSELFTEST, "ERROR: boost converter voltage %.3f is not within "
             "10%% of allowed voltages:", pwrvoltage);

    for (uint32_t i = 0; i < sizeof(allowed) / sizeof(allowed[0]); i++)
      PrintLog(LOGSELFTEST, " %.1f", allowed[i]);

    PrintLog(LOGSELFTEST, "\n");
    loglock.Leave();
  }

  g_controller.Deposit().LockDriver().PowerDownBoostConverter();
}

/*! Tests the microcontroller's VCC, by measuring the internal ADC reference
    voltage. The VREF+ pin on the microcontroller is connected to 3.3 volts.
    By measuring the internal 1.21 volts reference voltage, the VCC voltage
    can be measured.*/
void CSelfTest::TestVCC()
{
  if (IsTestPassed(SelfTest_VCC))
    return;

  float vcc  = g_controller.Adc().ReadVCC();
  float vmin = 3.2f;
  float vmax = 3.4f;

  PrintLog(LOGSELFTEST, "VCC: %.3f, minimum required: %.3g, maximum allowed: %.3g\n",
           vcc, vmin, vmax);

  if (vcc >= vmin && vcc <= vmax)
    SetTestPassed(SelfTest_VCC);
}

/* Wiring needed for cable test:

   BRIDGE2 connected to I1
   BRIDGE3 connected to I3

   POWERRAIL_OUT/P12V_SW connected to BRIDGE2/I1 and BRIDGE3/I3 using
   2k2 resistors and diodes in series. 

   12 volts power supply connected to SB_IN, and to two grounds via a relay's
   normally closed and normally open contacts.

   The relay coil is powered by POWERRAIL_OUT, and grounded by BRIDGE3/I.
*/

/*! Runs a test on the external IO pins,
    expects a test circuit attached to the cable.*/
bool CSelfTest::TestCable()
{
  PrintLog(LOGSELFTEST, "Testing cable IO\n");
  WaitLogWriteComplete(LOG_WAIT_TIMEOUT);

  g_controller.Deposit().LockDriver().PowerUpBoostConverter();
  SetupExternalIOPins();

  bool success = true;

  //Run the external IO test with the test circuit attached to the cable.
  if (TestExternalIOPins(true))
    PrintLog(LOGSELFTEST, "Cable IO is working correctly\n");
  else
    success = false;

  //Test if the charger connection and the two cable grounds are working.
  if (!TestCableGrounds())
    success = false;

  g_controller.Deposit().LockDriver().PowerDownBoostConverter();
  CloseExternalIOPins();

  if (success)
    PrintLog(LOGSELFTEST, "Cable is correct\n");
  else
    PrintLog(LOGSELFTEST, "ERROR: Cable is not correct\n");
// EDWIN TEST
// success=true;
  return success;
}

static const SGPIOStateInit g_pininit[] =
{
  {LOW, { B1_IN1_PIN, GPIO_Mode_OUT, GPIO_Speed_2MHz, GPIO_OType_PP, GPIO_PuPd_NOPULL }, -1},
  {LOW, { B1_IN2_PIN, GPIO_Mode_OUT, GPIO_Speed_2MHz, GPIO_OType_PP, GPIO_PuPd_NOPULL }, -1},
  {LOW, { B2_IN1_PIN, GPIO_Mode_OUT, GPIO_Speed_2MHz, GPIO_OType_PP, GPIO_PuPd_NOPULL }, -1},
  {LOW, { B2_IN2_PIN, GPIO_Mode_OUT, GPIO_Speed_2MHz, GPIO_OType_PP, GPIO_PuPd_NOPULL }, -1}
};
static const SGPIOPin g_pins[] =
{
  { B1_IN1_PORT, B1_IN1_PIN },
  { B1_IN2_PORT, B1_IN2_PIN },
  { B2_IN1_PORT, B2_IN1_PIN },
  { B2_IN2_PORT, B2_IN2_PIN }
};

/*! Sets up the bridge driver pins as outputs with no alternate function.*/
void CSelfTest::SetupExternalIOPins()
{
  for (uint32_t i = 0; i < sizeof(g_pins) / sizeof(g_pins[0]); i++)
    GPIOInitPin(g_pins[i].port, g_pins[i].pin, g_pininit + i);
}

/*! Sets external IO pins to their default initialization state.*/
void CSelfTest::CloseExternalIOPins()
{
  for (uint32_t i = 0; i < sizeof(g_pins) / sizeof(g_pins[0]); i++)
    GPIOInitPin(g_pins[i].port, g_pins[i].pin);

  //Disable the H-bridges.
  GPIO_Reset(B1_EN);
  GPIO_Reset(B2_EN);

  //Disable the powered outputs.
  GPIO_Reset(P12V_SW);
  GPIO_Reset(P12V_OUT1);
  GPIO_Reset(P12V_OUT2);
}

/*! Tests external inputs, output and H-bridges, with the test circuit attached.
    \param cabletest When true, runs a test with a test circuit attached to the cable.
                     When false, runs a test with a test pcb attached to the IO pins.
    \return true All external IOs are working correctly.
    \return false There is a problem with an external IO.*/
bool CSelfTest::TestExternalIOPins(bool cabletest)
{
  uint32_t oldpercentage = 0;

  //Loop with B1_EN and B2_EN disabled, one of either enabled, and both enabled.
  for (uint32_t en = 0; en < 4; en++)
  {
    GPIO_WritePin(B1_EN, (en & 1) ? HIGH : LOW);
    GPIO_WritePin(B2_EN, (en & 2) ? HIGH : LOW);

    //Loop with all possibilities of the P12V outputs.
    for (uint32_t out = 0; out < 8; out++)
    {
      GPIO_WritePin(P12V_OUT1, (out & 1) ? HIGH : LOW);
      GPIO_WritePin(P12V_OUT2, (out & 2) ? HIGH : LOW);
      GPIO_WritePin(P12V_SW,   (out & 4) ? HIGH : LOW);

      //Loop with all possibilities of the H-bridge outputs.
      for (uint32_t b = 0; b < 16; b++)
      {
        GPIO_WritePin(B1_IN2, (b & 1) ? HIGH : LOW); //B1_IN2 enables BRIDGE1.
        GPIO_WritePin(B1_IN1, (b & 2) ? HIGH : LOW); //B1_IN1 enables BRIDGE2.
        GPIO_WritePin(B2_IN2, (b & 4) ? HIGH : LOW); //B2_IN2 enables BRIDGE3.
        GPIO_WritePin(B2_IN1, (b & 8) ? HIGH : LOW); //B2_IN1 enables BRIDGE4.

        //Give the outputs and inputs some time to settle, then read the inputs.
        DelayMS(50);
        uint8_t inputs = g_controller.Inputs().ReadMultiple(INPUT_ALL_IN);

        //Compare the state of the inputs with the expected state.
        uint8_t expected;
        if (cabletest)
          expected = ExpectedInputStateTestCable(out, en, b);
        else
          expected = ExpectedInputStateTestPCB(out, en, b);

        if (inputs != expected)
        {
          PrintExternalIOError(inputs, expected, out, en, b);
          return false;
        }

        //Because this test takes some time, print every 10% completion.
        uint32_t percentage = (b + out * 16 + en * 8 * 16) * 100 / 511;
        if (percentage / 10 != oldpercentage / 10)
        {
          PrintLog(LOGSELFTEST, "%" PRIu32 "%% done\n", percentage);
          oldpercentage = percentage;
          
          //Reset the watchdog timer here, because the
          //heartbeat timer is not running yet.
          IWDG_ReloadCounter();
        }
      }
    }
  }

  return true;
}

/*! Calculates the expected state of the inputs with a test PCB attached,
    based on the state of the outputs and H-bridges.
    \param en The state of the H-bridge enable pins.
    \param out The state of the output pins.
    \param b The state of the H-bridge output pins.
    \return The expected state of the inputs.*/
uint8_t CSelfTest::ExpectedInputStateTestPCB(uint32_t out, uint32_t en, uint32_t b)
{
  uint8_t expected = 0;

  if (out & 1) //P12V_OUT1 high, acts as pullup for IN1 via a resistor.
    expected |= INPUT_IN1;
  if (out & 2) //P12V_OUT2 high, acts as pullup for IN2 via a resistor.
    expected |= INPUT_IN2;
  if (out & 4) //P12V_SW high, acts as pullup for IN3 and IN4 via resistors.
    expected |= INPUT_IN3 | INPUT_IN4;

  //B1_EN high, drive enabled for BRIDGE1 and BRIDGE2, connected to I1 and I2.
  if (en & 1) 
  {
    if (b & 1) //If B1_IN2/BRIDGE1 is high, IN1 should be high, otherwise low.
      expected |= INPUT_IN1;
    else
      expected &= ~INPUT_IN1;

    if (b & 2) //If B1_IN1/BRDIGE2 is high, IN2 should be high, otherwise low.
      expected |= INPUT_IN2;
    else
      expected &= ~INPUT_IN2;
  }

  //B2_EN high, drive enabled for BRIDGE3 and BRIDGE4, connected to I3 and I4.
  if (en & 2)
  {
    if (b & 4) //If B2_IN2/BRIDGE3 is high, IN3 should be high, otherwise low.
      expected |= INPUT_IN3;
    else
      expected &= ~INPUT_IN3;

    if (b & 8) //If B2_IN1/BRIDGE4 is high, IN4 should be high, otherwise low.
      expected |= INPUT_IN4;
    else
      expected &= ~INPUT_IN4;
  }

  return expected;
}

/*! Calculates the expected state of the inputs with a test circuit attached
    to the cable, based on the state of the outputs and H-bridges.
    \param en The state of the H-bridge enable pins.
    \param out The state of the output pins.
    \param b The state of the H-bridge output pins.
    \return The expected state of the inputs.*/
uint8_t CSelfTest::ExpectedInputStateTestCable(uint32_t out, uint32_t en, uint32_t b)
{
  uint8_t expected = 0;

  if (out & 4) //P12V_SW high, acts as pullup for IN1 and IN3 via resistors.
    expected |= INPUT_IN1 | INPUT_IN3;

  //B1_EN high, drive enabled for BRIDGE1 and BRIDGE2.
  if (en & 1) 
  {
    if (b & 2) //If B1_IN1/BRIDGE2 is high, IN1 should be high, otherwise low.
      expected |= INPUT_IN1;
    else
      expected &= ~INPUT_IN1;
  }

  //B2_EN high, drive enabled for BRIDGE3 and BRIDGE4.
  if (en & 2)
  {
    if (b & 4) //If B2_IN2/BRIDGE3 is high, IN3 should be high, otherwise low.
      expected |= INPUT_IN3;
    else
      expected &= ~INPUT_IN3;
  }

  return expected;
}

/*! Prints the expected state of IN1 to IN4, the measured state of IN1 to IN4,
    and the state of the outputs.
    \param inputs The measured input state.
    \param expected The expected input state.
    \param en The state of the H-bridge enable pins.
    \param out The state of the output pins.
    \param b The state of the H-bridge outputs.*/
void CSelfTest::PrintExternalIOError(uint8_t inputs, uint8_t expected,
                                     uint32_t out, uint32_t en, uint32_t b)
{
  PrintLog(LOGSELFTEST, "\n");
  PrintLog(LOGSELFTEST, "ERROR: expected inputs to be: "
           "IN1:%-4s       IN2:%-4s       IN3:%-4s       IN4:%-4s\n",
           CInputs::StateStr(expected & INPUT_IN1),
           CInputs::StateStr(expected & INPUT_IN2),
           CInputs::StateStr(expected & INPUT_IN3),
           CInputs::StateStr(expected & INPUT_IN4));

  PrintLog(LOGSELFTEST, "       but inputs are:        "
           "IN1:%-4s       IN2:%-4s       IN3:%-4s       IN4:%-4s\n",
           CInputs::StateStr(inputs & INPUT_IN1),
           CInputs::StateStr(inputs & INPUT_IN2),
           CInputs::StateStr(inputs & INPUT_IN3),
           CInputs::StateStr(inputs & INPUT_IN4));

  PrintLog(LOGSELFTEST, "\n");
  PrintLog(LOGSELFTEST, "       Bridge enable:         "
           "B1_EN:%-4s                    B2_EN:%-4s\n",
           CInputs::StateStr(en & 1),
           CInputs::StateStr(en & 2));

  PrintLog(LOGSELFTEST, "       Bridge outputs:        "
           "BRIDGE1:%-4s   BRIDGE2:%-4s   BRIDGE3:%-4s   BRIDGE4:%-4s\n",
           CInputs::StateStr(b & 1),
           CInputs::StateStr(b & 2),
           CInputs::StateStr(b & 4),
           CInputs::StateStr(b & 8));

  PrintLog(LOGSELFTEST, "\n");
  PrintLog(LOGSELFTEST, "       Switched outputs:      "
           "P12V_OUT1:%-4s P12V_OUT2:%-4s P12V_SW:%-4s\n",
           CInputs::StateStr(out & 1),
           CInputs::StateStr(out & 2),
           CInputs::StateStr(out & 4));

  PrintLog(LOGSELFTEST, "\n");
}

/*! Tests if the battery charger reports an external voltage.
    Then switches on the relay connected between POWERRAIL_OUT and BRIDGE3,
    and tests again, the relay contacts switch the charger input to the other
    ground wire on the cable.
    The SetupExternalIOPins() method should be called before using this.*/
bool CSelfTest::TestCableGrounds()
{
  //Enable BRIDGE3 as GND for the relay coil.
  GPIO_Set(B2_EN);
  GPIO_Reset(B2_IN2);

  bool success = true;

  //Measure the external voltage connected to the battery charger,
  //with the relay connected to POWERRAIL_OUT off, and on.
  //This tests the two grounds on the cable.
  for (uint32_t i = 0; i < 2; i++)
  {
    const char* relaystr = i ? "on" : "off";
    PrintLog(LOGSELFTEST, "Testing charger connection with ground relay %s\n",
             relaystr);

    //Switch the relay to select a ground wire on the cable.
    if (i)
      GPIO_Set(P12V_SW);
    else
      GPIO_Reset(P12V_SW);

    DelayMS(100);

    //Trigger a recharge on the battery charger to discharge the capacitor.
    g_controller.BatteryCharger().TriggerRecharge();
    DelayMS(2000);

    //Give the onboard capacitor some time to discharge, force a telemetry
    //update, and read the telemetry data back.
    g_controller.BatteryCharger().SetForceTelemetry(true, false, true);
    DelayMS(100);

    SBatteryInfo chargerinfo;
    g_controller.BatteryCharger().ReadChargerInfo(&chargerinfo);

    if (chargerinfo.m_extvoltage > 8.0f)
    {
      PrintLog(LOGSELFTEST, "External voltage with ground relay %s: %.1f, "
               "connection is working\n", relaystr, chargerinfo.m_extvoltage);
    }
    else
    {
      PrintLog(LOGSELFTEST, "ERROR: External voltage with ground relay %s: %.1f, "
               "connection is not working\n", relaystr, chargerinfo.m_extvoltage);
      success = false;
    }
  }

  GPIO_Reset(P12V_SW);
  GPIO_Reset(B2_EN);

  return success;
}
