#include <cmath>

#include "clock.h"
#include "servercomm.h"
#include "log.h"
#include "modembg96.h"
#include "fwupdate.h"
#include "controller.h"
#include "gpiodescriptions.h"
#include "cbortools.h"
#include "http.h"
#include "stm32wss.h"
#include "backupregisters.h"
#include "lock.h"
#include "timeconversion.h"
#include "factoryreset.h"
#include "active.h"

/*! Timeout for the HTTP POST.*/
#define SERVERCOMM_TIMEOUT 120

#define MODEM_USART USART6

CServerComm::CServerComm() : m_url {}
{
  m_signal       = 0;
  m_factoryreset = false;
  m_commreason   = 0;
  m_contacting   = false;
  m_activesim    = 0;
}

void CServerComm::Setup()
{
  m_modem.SetPeripherals(MODEM_USART, MODEM_PWRKEY, MODEM_STATUS, MODEM_RESET);
  m_fwupdate.Init(&g_controller.SPIFlash());

  //create a semaphore to signal that an RTC interrupt has happened
  m_signal = xSemaphoreCreateBinary();

  //Check if the modem should have a factory reset during the first setup.
  m_factoryreset = CFactoryReset::Happened();

  //enable the RTC alarm interrupt, when that interrupt happens,
  //the RTCIRQHandler() method is called from g_exti
  RTC_ITConfig(RTC_IT_ALRA, ENABLE);

  g_exti.EnableRTCIrq(RTCIRQAlarmA, this);

  //register this class as a task that can decide if low power mode can be entered
  RegisterLowPowerTask("servercomm");
}

void CServerComm::Start()
{
  Create("servercomm");
}

//! Number of milliseconds to wait before setting up the modem again.
#define SERVERCOMM_MODEM_FAILED_DELAY (10 * 1000)

void CServerComm::Process()
{
  g_controller.PostStartup();

  //make sure the modem is powered off, in case of a crash the modem might
  //stay powered on, this will drain the battery
  m_modem.PowerOff();

  for(;;)
  {
    //wait until it's time to communicate with the server
    WaitNextAttempt();
    
    //Do not contact the server unneeded when the device is inactive.
    if (!CheckActiveMode())
      continue;

    //check if the battery has enough charge to power the modem
    if (!CheckBatteryVoltage())
      continue;

    //Keep track of number of attempts in inactive mode.
    RecordInactiveAttempt();

    //make the internal soldered SIM, or the external socketed SIM active
    SetActiveSIM();

    //Perform a factory reset if requested.
    if (m_factoryreset)
    {
      m_modem.Setup(true, false);
      m_factoryreset = false;
    }

    //Set up an internet connection, only allow a modem firmware update when
    //the device is in active mode.
    bool success = m_modem.Setup(false, DeviceActive());

    //record the modem RSSI and SIM ICCID as statuses
    RecordModemStatus();

    //Communicate with the server.
    if (success)
      success = ProcessCommands();

    //Keep track of number of bytes sent and received in inactive mode.
    RecordInactiveDataCounters();

    if (m_modem.Close())
    { //Mark the modem test as passed, when server communication is successful,
      //and the modem could be powered off afterwards.
      if (success)
      {
        g_controller.SelfTest().SetTestPassed(SelfTest_Modem);

        //If the device is in inactive mode, mark that server communication has
        //succeeded, so that it will not be tried again.
        if (!DeviceActive())
          g_controller.Settings().SetBoolean(InactiveStatus_ServerComm_Tested, true);
      }
    }

    if (success)
      OnSuccessfulAttempt();
    else
      OnFailedAttempt();
  }
}

/*!If the device is in inactive mode, there should be as little traffic on
   the SIM as possible, to prevent activating the subscription.
   So only contact the server if self testing is enabled, and the server has
   not been contacted once before.
   \return true The server should be contacted.
   \return false The server should not be contacted.*/
bool CServerComm::CheckActiveMode()
{
  if (DeviceActive())
    return true;

  if (!g_controller.SelfTest().Enabled())
  {
    PrintLog(LOGSERVERCOMM, "Device is inactive and self testing is disabled, "
             "not contacting server\n");
    EndAttempts();
    return false;
  }

  if (g_controller.Settings().GetBoolean(InactiveStatus_ServerComm_Tested))
  {
    PrintLog(LOGSERVERCOMM, "Device is inactive and self testing is enabled, "
             "but server contact has succeeded before, not contacting server\n");

    //Mark these tests as passed, because they have been passed before.
    //This makes sure that all self tests are passed.
    g_controller.SelfTest().SetTestPassed(SelfTest_ServerComm);
    g_controller.SelfTest().SetTestPassed(SelfTest_Modem);

    EndAttempts();
    return false;
  }

  //Self test enabled, and server communication has not succeeded yet.
  return true;
}

/*! Registers a status indicating why communication to the server has been started.
    \param reason A value of COMM_REASON.*/
void CServerComm::RegisterCommReason(uint32_t reason)
{
  const char* reasons[] =
  {
    "outbound call card",
    "warning fill level reached",
    "max fill level reached",
    "deposit opening blocked",
    "emptied",
    "RTC alarm",
    "deposit opening stayed open",
    "deposit opening closed",
    "insufficient prepaid credit",
    "low prepaid credit",
    "self test",
    "factory reset",
    "made active",
    "door stayed open",
    "door closed",
  };

  //write the contact reason as a status
  if (reason < sizeof(reasons) / sizeof(reasons[0]))
    g_controller.Status().Write(Status_ContactReason, "%s", reasons[reason]);
  else
    g_controller.Status().Write(Status_ContactReason, "ERROR UNKNOWN REASON %"
                                PRIu32 , reason );
}

/*! Number of communication attempts since the last successful contact with the server.
    Stored in battery backed ram, so that even after a crash or power loss,
    trying to make contact with the server will be retried, but only for a maximum
    number of times.*/
#define SERVERCOMM_ATTEMPTS_REGISTER RTC_BKP_DR19

//! Number of times the modem is set up before giving up.
#define SERVERCOMM_MODEM_MAX_ATTEMPTS 4

//! Number of failed communications, used to set the retry interval.
#define SERVERCOMM_FAILED_COMMS_REGISTER RTC_BKP_DR17

/*! Waits until it's time to communicate with the server. Will return immdiately
    if SERVERCOMM_ATTEMPTS_REGISTER is lower than SERVERCOMM_MODEM_MAX_ATTEMPTS.
    Will allow going into low power mode when SERVERCOMM_ATTEMPTS_REGISTER is
    equal to or higher than SERVERCOMM_MODEM_MAX_ATTEMPTS.*/
void CServerComm::WaitNextAttempt()
{
  for(;;)
  {
    //check if SERVERCOMM_ATTEMPTS_REGISTER should be reset
    CheckAttemptsReset();

    uint32_t numattempts = RTC_ReadBackupRegister(SERVERCOMM_ATTEMPTS_REGISTER);

    if (numattempts >= SERVERCOMM_MODEM_MAX_ATTEMPTS)
    {
      PrintLog(LOGSERVERCOMM, "Waiting for next communication attempt\n");

      //set the RTC alarm to the next time a communication with the server
      //should be attempted, this alarm will wake the mcu from low power mode
      SetNextAttemptAlarm();

      //allow going into low power mode if the maximum number of attempts is reached,
      //an RTC alarm can wake the microcontroller from low power mode
      //for the next communicaton attempt
      AllowLowPower(true);

      //signal on the service display that the server is not being contacted
      m_contacting = false;

      //wait for an event after which SERVERCOMM_ATTEMPTS_REGISTER should be reset
      xSemaphoreTake(m_signal, portMAX_DELAY);

      //deny going into low power mode if communication with the server is needed
      AllowLowPower(false);
    }
    else
    {
      //increase SERVERCOMM_ATTEMPTS_REGISTER by 1, this is done before the
      //communication attempt, so that even in case of a crash or power loss,
      //the number of communication attempts is limited
      RTC_WriteBackupRegister(SERVERCOMM_ATTEMPTS_REGISTER, numattempts + 1);

      PrintLog(LOGSERVERCOMM, "Starting communication attempt %" PRIu32 " of max %i\n",
               numattempts + 1, SERVERCOMM_MODEM_MAX_ATTEMPTS);

      //signal on the service display that the server is being contacted
      m_contacting = true;

      return;
    }
  }
}

/*! Checks if an RTC alarm has happened, or if a communication
    attempt is requested.
    If so resets SERVERCOMM_ATTEMPTS_REGISTER to 0.*/
void CServerComm::CheckAttemptsReset()
{
  bool alarm = RTCAlarmFlagSet(RTC_Alarm_A);

  CCritical criticalsection;
  uint32_t commreason = m_commreason;
  m_commreason = 0;
  criticalsection.Leave();

  //If RTC alarm A happened, the server should be contacted.
  bool resetattempts = false;
  if (alarm)
  {
    PrintLog(LOGSERVERCOMM, "RTC alarm event happened, "
             "resetting communication attempts to 0\n");
    ClearRTCAlarmFlag(RTC_Alarm_A);
    resetattempts = true;
    RegisterCommReason(COMM_REASON_RTC_ALARM);
  }

  //If another task set a bit in m_commreason, the server should be contacted.
  if (commreason != 0)
  {
    PrintLog(LOGSERVERCOMM, "Communication attempt requested, "
             "resetting communication attempts to 0\n");
    resetattempts = true;

    //Register all recorded comm reasons. This should be done from the servercomm
    //task so that the reader task does not block on initializing the status ringbuffer.
    for (uint32_t i = 0; i < sizeof(commreason) * 8; i++)
    {
      if (commreason & (1 << i))
        RegisterCommReason(i);
    }

    //On a requested communication attempt, reset the retry delay.
    RTC_WriteBackupRegister(SERVERCOMM_FAILED_COMMS_REGISTER, 0);
  }

  if (resetattempts)
    RTC_WriteBackupRegister(SERVERCOMM_ATTEMPTS_REGISTER, 0);
}

/*! Sets SERVERCOMM_ATTEMPTS_REGISTER to SERVERCOMM_MODEM_MAX_ATTEMPTS so that
    no further communication with the server is done.*/
void CServerComm::EndAttempts()
{
  RTC_WriteBackupRegister(SERVERCOMM_ATTEMPTS_REGISTER, SERVERCOMM_MODEM_MAX_ATTEMPTS);
}

/*! Called after successful communication with the server.*/
void CServerComm::OnSuccessfulAttempt()
{
  g_controller.SelfTest().SetTestPassed(SelfTest_ServerComm);

  uint32_t numattempts = RTC_ReadBackupRegister(SERVERCOMM_ATTEMPTS_REGISTER);
  RTC_WriteBackupRegister(SERVERCOMM_ATTEMPTS_REGISTER, SERVERCOMM_MODEM_MAX_ATTEMPTS);
  RTC_WriteBackupRegister(SERVERCOMM_FAILED_COMMS_REGISTER, 0); //Reset the retry delay.

  PrintLog(LOGSERVERCOMM, "Successfully communicated with the server after %"
           PRIu32 " attempt%s\n", numattempts, numattempts > 1 ? "s" : "");

  //Record the last attempt for the service display.
  RecordLastAttempt(true);
}

/*! Delays when the maximum number of failed attempts has not been reached.*/
void CServerComm::OnFailedAttempt()
{
  uint32_t numattempts = RTC_ReadBackupRegister(SERVERCOMM_ATTEMPTS_REGISTER);

  //Keep track of the number of failed attempts for status reporting.
  uint32_t numfailed = g_controller.Settings().GetInteger(RunStatus_ServerContactFailed);
  g_controller.Settings().SetInteger(RunStatus_ServerContactFailed, numfailed + 1);

  //Record the last attempt for the service display.
  RecordLastAttempt(false);

  if (numattempts < SERVERCOMM_MODEM_MAX_ATTEMPTS)
  {
    PrintLog(LOGSERVERCOMM, "Setting up an internet connection and connecting "
             "to the server failed %" PRIu32 " time%s, retrying after %g seconds\n",
             numattempts, numattempts > 1 ? "s" : "",
             (float)SERVERCOMM_MODEM_FAILED_DELAY / 1000.0f);

    DelayMS(SERVERCOMM_MODEM_FAILED_DELAY);
  }
  else
  {
    PrintLog(LOGSERVERCOMM, "Setting up an internet connection and connecting "
             "to the server failed %" PRIu32 " times, giving up\n", numattempts);

    //Keep track of the number of failed communications for the retry interval.
    uint32_t failedreg = RTC_ReadBackupRegister(SERVERCOMM_FAILED_COMMS_REGISTER);
    RTC_WriteBackupRegister(SERVERCOMM_FAILED_COMMS_REGISTER, failedreg + 1);
  }
}

/*! Called from other tasks to signal that a communication attempt is needed.*/
void CServerComm::SignalCommAttempt(uint32_t reason)
{
  PrintLog(LOGSERVERCOMM, "Task %s signaled that a communication attempt is needed\n",
           pcTaskGetName(xTaskGetCurrentTaskHandle()));

  CCritical criticalsection;
  m_commreason |= 1 << reason;
  criticalsection.Leave();

  xSemaphoreGive(m_signal);
}

/*! Returns the maximum allowed delay in seconds for the next communication attempt.
    This is based on the number of failed communications.*/
uint32_t CServerComm::MaxNextAttemptDelay(uint32_t* commfailures)
{
  *commfailures = RTC_ReadBackupRegister(SERVERCOMM_FAILED_COMMS_REGISTER);

  uint32_t retval;

  if (*commfailures == 0)
    retval = 36 * 60 * 60; //Communication should happen within 36 hours.
  else if (*commfailures == 1)
    retval = 60 * 60;      //Communication should happen within 1 hour after 1 failure.
  else if (*commfailures == 2)
    retval = 2 * 60 * 60;  //Communication should happen within 2 hours after 2 failures.
  else if (*commfailures == 3)
    retval = 4 * 60 * 60;  //Communication should happen within 4 hours after 3 failures.
  else if (*commfailures == 4)
    retval = 8 * 60 * 60;  //Communication should happen within 8 hours after 4 failures.
  else if (*commfailures == 5)
    retval = 8 * 60 * 60;  //Communication should happen within 8 hours after 5 failures.
  else
    retval = 19 * 60 * 60; //Communication should happen within 19 hours.
                           //Some mobile operators reduce their transmit power
                           //during the night, which might mean the connection
                           //won't work. Set the retry to 19 hours so that it's
                           //asynchronous to their 24 hour cycle.

  PrintLog(LOGSERVERCOMM, "Communication failures: %" PRIu32 ", maximum allowed "
           "time until next communication attempt: %02" PRIu32 "h%02" PRIu32 "m\n",
           *commfailures, retval / 3600, (retval % 3600) / 60);

  return retval;
}

/*! Checks the Setting_TimeComm setting, if it has a reasonable value it will
    set RTC alarm A to it, otherwise will set RTC alarm A to a reasonable time
    in the future.*/
void CServerComm::SetNextAttemptAlarm()
{
  //Check when the next attempt should happen based on the number of failed
  //communication attempts.
  uint32_t commfailures;
  uint32_t maxattemptdelay = MaxNextAttemptDelay(&commfailures);

  char buf[64];
  g_controller.Settings().GetString(Setting_TimeComm, buf, sizeof(buf));

  //check if Setting_TimeComm can be interpreted as ISO8601
  bool commtimevalid = true;
  tm timeval = {};
  if (!FromISO8601(buf, &timeval))
  {
    PrintLog(LOGSERVERCOMM, "ERROR: invalid ISO8601 TimeComm setting \"%s\"\n", buf);
    commtimevalid = false;
  }

  int64_t rtctime = GetRTCSeconds(nullptr);
  if (commtimevalid)
  { //if Setting_TimeComm is valid ISO8601, check if it's not in the past and
    //not too far in the future
    int64_t commtime = TmToSecs(&timeval);
    if (commtime < rtctime)
    {
      CLogLock loglock;
      PrintLog(LOGSERVERCOMM, "Communication time %s is in the past", buf);
      GetRTCISO8601(buf, sizeof(buf), "");
      PrintLog(LOGSERVERCOMM, ", current rtc time: %s\n", buf);
      commtimevalid = false;
    }
    else if (commtime > rtctime + maxattemptdelay)
    {
      CLogLock loglock;
      PrintLog(LOGSERVERCOMM, "Communication time %s is too far in the future", buf);
      GetRTCISO8601(buf, sizeof(buf), "");
      PrintLog(LOGSERVERCOMM, ", current rtc time: %s\n", buf);
      commtimevalid = false;
    }
    else
    {
      PrintLog(LOGSERVERCOMM, "Next communication time: %s\n", buf);
    }
  }

  //If the communication time is invalid, set the next communication time to
  //the current time plus the maximum allowed delay.
  if (!commtimevalid)
  {
    //If no valid communication time is set, but server communication did not fail,
    //Retry communication after 24 hours.
    if (commfailures == 0)
      maxattemptdelay = 24 * 60 * 60;

    int64_t commtime = rtctime + maxattemptdelay;
    SecsToTm(commtime, &timeval);
    ToISO8601(&timeval, buf, sizeof(buf), "");
    PrintLog(LOGSERVERCOMM, "Setting communication time to current time plus %02"
             PRIu32 "h%02" PRIu32 "m: %s\n",
             maxattemptdelay / 3600, (maxattemptdelay % 3600) / 60, buf);
    g_controller.Settings().SetString(Setting_TimeComm, buf);
  }

  //set RTC alarm A to the wanted value, when this alarm expires,
  //the microcontroller will wake from low power mode, the servercomm task
  //will read the RTC_IT_ALRA bit from the RTC, and will start communicating
  SetRTCAlarm(RTC_Alarm_A, &timeval);
}

/*! Called from CExti when an RTC alarm has happened.
    It signals m_signal to indicate that a new communication attempt
    with the server should be started.*/
void CServerComm::RTCIRQHandler(ERTCIRQType irqtype)
{
  xSemaphoreGiveFromISR(m_signal, nullptr);
}

/*! Minimum battery voltage required to contact the server, below this
    the server is not contacted, to save battery power.*/
#define MINIMUM_BATTERY_VOLTAGE 3.0

/*! Number of times the battery voltage is checked before giving up.*/
#define NUM_BATTERY_VOLTAGE_CHECKS 3

/*! Checks if the battery voltage is high enough to contact the server.
    \return true if the battery voltage is high enough.
    \return false if the battery voltage is too low.*/
bool CServerComm::CheckBatteryVoltage()
{
  //the battery charger updates its telemetry readings every 5 seconds when the
  //battery is not being charged, if the lock is actuated while the battery
  //voltage is read the reading will be a lot lower, so check several times
  for (uint32_t i = 0; i < NUM_BATTERY_VOLTAGE_CHECKS; i++)
  {
    if (g_controller.BatteryCharger().CheckBatteryVoltage(MINIMUM_BATTERY_VOLTAGE,
                                                LOGSERVERCOMM, "contacting server"))
      return true;
    else if (i + 1 < NUM_BATTERY_VOLTAGE_CHECKS)
      DelayMS(5100);
  }

  //Disable further attempts to contact the server until the next time
  //the server should be contacted.
  RecordLastAttempt(false);
  EndAttempts();

  PrintLog(LOGSERVERCOMM, "Battery voltage is too low to contact the server,"
           " giving up\n");

  return false;
}

/*! Selects the active SIM using the MODEM_SIM_SELECT gpio.
    The active SIM is chosen from the settings.
    On the last communication attempt, the SIM is switched to the other SIM.
    This is a last resort in case the wrong SIM is selected in the settings.*/
void CServerComm::SetActiveSIM()
{
  //Read the number of connection attempts from the battery backed ram.
  uint32_t numattempts = RTC_ReadBackupRegister(SERVERCOMM_ATTEMPTS_REGISTER);

  //Read the active SIM from the settings.
  m_activesim = g_controller.Settings().GetInteger(Setting_ActiveSIM);
  if (m_activesim != SERVERCOMM_ACTIVE_SIM_INTERNAL &&
      m_activesim != SERVERCOMM_ACTIVE_SIM_EXTERNAL)
  {
    PrintLog(LOGSERVERCOMM, "ERROR: invalid value %" PRIu32
             " for active SIM setting, defaulting to external SIM\n", m_activesim);
    m_activesim = SERVERCOMM_ACTIVE_SIM_EXTERNAL; //Safe default.
  }

  //If this is the last communication attempt before giving up, switch the SIM.
  if (numattempts == SERVERCOMM_MODEM_MAX_ATTEMPTS)
  {
    if (m_activesim == SERVERCOMM_ACTIVE_SIM_INTERNAL)
      m_activesim = SERVERCOMM_ACTIVE_SIM_EXTERNAL;
    else
      m_activesim = SERVERCOMM_ACTIVE_SIM_INTERNAL;

    PrintLog(LOGSERVERCOMM, "Last communication attempt, switching SIM\n");
  }

  //Select the active SIM using the gpio.
  if (m_activesim == SERVERCOMM_ACTIVE_SIM_INTERNAL)
  {
    PrintLog(LOGSERVERCOMM, "Using internal SIM\n");
    GPIO_Set(MODEM_SIM_SELECT);
  }
  else if (m_activesim == SERVERCOMM_ACTIVE_SIM_EXTERNAL)
  {
    PrintLog(LOGSERVERCOMM, "Using external SIM\n");
    GPIO_Reset(MODEM_SIM_SELECT);
  }
}

/*! Records the modem RSSI and SIM ICCID as statuses.*/
void CServerComm::RecordModemStatus()
{
  //record the modem RSSI as a status
  g_controller.Status().Write(Status_ModemRSSI, "%" PRIi32, m_modem.RSSI());

  //select which ICCID status to write based on which SIM is selected
  EStatus iccidstatus;
  if (m_activesim == SERVERCOMM_ACTIVE_SIM_INTERNAL)
    iccidstatus = Status_InternalSIMICCID;
  else
    iccidstatus = Status_ExternalSIMICCID;

  //record the SIM ICCID as a status
  const char* iccid = m_modem.ICCID();
  if (strlen(iccid) == 0)
    iccid = "Not present";
  g_controller.Status().Write(iccidstatus, "%s", iccid);

  //Record the modem type and firmware version as a status.
  g_controller.Status().Write(Status_ModemType, "%s %s %s %s",
                              m_modem.Manufacturer(), m_modem.Model(),
                              m_modem.Revision(), m_modem.AppVer());
}

/*! Records the last attempt as a runstatus, used for the service display.
    \param success Indicates if the last attempt was successful.*/
void CServerComm::RecordLastAttempt(bool success)
{
  tm timeval = {};
  GetRTC(&timeval);

  int32_t rssi = m_modem.RSSI();
  char rssistr[16] = {};
  if (rssi == -1) //If the RSSI is -1, it's not available.
    strcpy(rssistr, "NA");
  else
    sprintf(rssistr, "%-2" PRIi32, rssi);

  char attempt[64] = {};
  snprintf(attempt, sizeof(attempt) - 1, "%02i:%02i R:%s S:%s",
           timeval.tm_hour, timeval.tm_min, rssistr, success ? "Y" : "N");

  //RunStatus_LastContactIndex points to the RunStatus_LastContactAttempt string
  //which holds information about the most recent contact attempt.
  uint32_t lastcontactindex =
    g_controller.Settings().GetInteger(RunStatus_LastContactIndex);
  lastcontactindex = (lastcontactindex + 1) % 3;

  EStorageKey attemptindexes[] =
  {
    RunStatus_LastContactAttempt1,
    RunStatus_LastContactAttempt2,
    RunStatus_LastContactAttempt3
  };

  g_controller.Settings().SetString(attemptindexes[lastcontactindex], attempt);
  g_controller.Settings().SetInteger(RunStatus_LastContactIndex, lastcontactindex);
}

/*! Keeps a counter of number of communication attempts when device is inactive.*/
void CServerComm::RecordInactiveAttempt()
{
  if (DeviceActive())
    return;

  uint32_t attempts =
    g_controller.Settings().GetInteger(InactiveStatus_ServerComm_Attempts);

  attempts++;
  g_controller.Settings().SetInteger(InactiveStatus_ServerComm_Attempts, attempts);
}

/*! Requests number of bytes transferred by the modem, stores it in counters.*/
void CServerComm::RecordInactiveDataCounters()
{
  if (DeviceActive())
    return;

  if (!m_modem.Activated())
    return;

  uint64_t tx;
  uint64_t rx;

  if (!m_modem.GetBytesTransferred(&tx, &rx))
    return;

  PrintLog(LOGSERVERCOMM, "Bytes transmitted: %" PRIu64 ", received: %" PRIu64
           ", total: %" PRIu64 "\n", tx, rx, tx + rx);

  int64_t oldtx = g_controller.Settings().GetInteger(InactiveStatus_ServerComm_TxBytes);
  int64_t oldrx = g_controller.Settings().GetInteger(InactiveStatus_ServerComm_RxBytes);

  g_controller.Settings().SetInteger(InactiveStatus_ServerComm_TxBytes, oldtx + tx);
  g_controller.Settings().SetInteger(InactiveStatus_ServerComm_RxBytes, oldrx + rx);
}

/*! Records the cell phone tower the modem is connected to, and records the
    location by the modem's GNSS receiver if requested.*/
void CServerComm::RecordLocation()
{
  //Record to which tower the modem is connected to by recording:
  //Public Land Mobile Network (PLMN)
  //Location Area code (LAC)
  //Cell ID (CID)
  g_controller.Status().Write(Status_ConnectedCell, "PLMN:%s LAC:0x%s(%" PRIu32
                              ") CID:0x%s(%" PRIu32 ")",
                              m_modem.RPLMN(), m_modem.LACStr(), m_modem.LAC(),
                              m_modem.CIDStr(), m_modem.CID());

  //The GPSFixTime decides how long to wait for a location lock in seconds.
  int32_t fixtime = g_controller.Settings().GetInteger(Setting_GPSFixTime);
  if (fixtime > 0)
  {
    //Read the GNSS location from the modem, and record it as a status.
    double latitude;
    double longitude;
    if (m_modem.GetLocation(fixtime, &latitude, &longitude))
      g_controller.Status().Write(Status_GPSLocation, "%f,%f", latitude, longitude);
    else
      g_controller.Status().Write(Status_GPSLocation, "Not available");
  }
}

//! Maximum number of connection failures before resetting the modem.
#define SERVERCOMM_MAX_FAILS 10

//! Number of connection failures before trying the fallback url.
#define SERVERCOMM_FAILS_USE_FALLBACK 5

//! Number of milliseconds before retrying after executing commands failed.
#define SERVERCOMM_COMMAND_FAILED_DELAY 10000

/*! Gets a list of commands from the server and executes them
    \return true if getting the list of command and executing all of them succeeded.
    \return false on any error, getting and executing the commands should be retried.*/
bool CServerComm::ProcessCommands()
{
  uint32_t numfails = 0;
  for(;;)
  {
    //get the url from the settings, process all commands with this url until completed
    //because executing a GetSettings command can change the url
    GetURLFromSettings(numfails);

    //In active mode, get commands from the server and execute them.
    //In inactive mode, only send a production test message.
    bool success;
    if (DeviceActive())
    {
      success = GetCommands();
      if (success)
        success = ExecuteCommands();
    }
    else
    {
      success = ProductionTest();
    }

    //if all commands are executed successfully, this function is done
    if (success)
      return true;
    else
      numfails++;

    if (m_modem.NeedsReset())
    {
      PrintLog(LOGSERVERCOMM, "ERROR: Modem reset needed\n");
      return false;
    }
    else if (numfails >= SERVERCOMM_MAX_FAILS)
    {
      //if the server can't be reached at all, check if an emergency firmware
      //update can be done, this can be used as a last resort if all else fails
      PrintLog(LOGSERVERCOMM, "%" PRIu32 " failures trying to connect to server"
               ", checking for emergency firmware update\n", numfails);
      UpdateFirmware(SERVERCOMM_EMERGENCY_FIRMWARE_URL);

      PrintLog(LOGSERVERCOMM, "ERROR: %" PRIu32 " failures trying to connect to "
               "the server, giving up\n", numfails);
      return false;
    }

    PrintLog(LOGSERVERCOMM, "Connecting to the server failed %" PRIu32
             " time%s, retrying after %g seconds\n",
             numfails, numfails > 1 ? "s" : "",
             (float)SERVERCOMM_COMMAND_FAILED_DELAY / 1000.0f);

    DelayMS(SERVERCOMM_COMMAND_FAILED_DELAY);
  }
}

/*! Gets the url of the server from the settings and stores it in m_url.
    If the url is an empty string or does not start with http://, it will
    copy SERVERCOMM_FALLBACK_URL into m_url.
    \param numfails When higher or equal to SERVERCOMM_FAILS_USE_FALLBACK,
                    copy SERVERCOMM_FALLBACK_URL into m_url.*/
void CServerComm::GetURLFromSettings(uint32_t numfails)
{
  //if connecting to the server has failed a number of times,
  //it's possible that the wrong url is stored in the settings,
  //so instead try the fallback url
  if (numfails >= SERVERCOMM_FAILS_USE_FALLBACK)
  {
    PrintLog(LOGSERVERCOMM, "Using fallback url \"%s\" because communication "
             "failed %" PRIu32 " time%s\n", SERVERCOMM_FALLBACK_URL,
             numfails, numfails > 1 ? "s" : "");

    strcpy(m_url, SERVERCOMM_FALLBACK_URL);
    return;
  }

  g_controller.Settings().GetString(Setting_ServerIp, m_url, sizeof(m_url));

  const char* httpprefix = "http://";
  const char* httpsprefix = "https://";
  if (strncmp(m_url, httpprefix, strlen(httpprefix)) != 0 &&
      strncmp(m_url, httpsprefix, strlen(httpsprefix)) != 0)
  {
    PrintLog(LOGSERVERCOMM, "ERROR: URL from settings \"%s\" does not start "
             "with \"%s\" or \"%s\", using fallback url \"%s\"\n",
             m_url, httpprefix, httpsprefix, SERVERCOMM_FALLBACK_URL);
    strcpy(m_url, SERVERCOMM_FALLBACK_URL);
  }
  else
  {
    PrintLog(LOGSERVERCOMM, "Using URL from settings: \"%s\"\n", m_url);
  }
}

/*! Sends a GetCommands command to the server, and downloads the response.
    \return true if everything succeeded, false if anything failed.*/
bool CServerComm::GetCommands()
{
  PrintLog(LOGSERVERCOMM, "Getting commands from server\n");

  memset(m_txbuf, 0, SERVERCOMM_BUFSIZE);
  CborEncoder encoder;
  cbor_encoder_init(&encoder, m_txbuf, SERVERCOMM_BUFSIZE, 0);
 
  //create a root map where all other fields will be added to
  CborEncoder root;
  cbor_encoder_create_map(&encoder, &root, 4);

  //add default fields
  AddCommandDefaultFields(&root, "GetCommands");

  //add the software version
  cbor_encode_text_stringz(&root, "V");
  cbor_encode_text_stringz(&root, FirmwareVersion());

  //close the container and check if everything was encoded correctly
  if (!CBORCloseContainer(LOGSERVERCOMM, "GetCommands message", &encoder, &root))
    return false;

  //send the generated CBOR to the HTTP server and receive its reply
  size_t cborsize = cbor_encoder_get_buffer_size(&encoder, m_txbuf);
  if (!Transceive(cborsize))
  {
    PrintLog(LOGSERVERCOMM, "ERROR: GetCommands failed\n");
    return false;
  }

  //Copy the reply to a separate buffer, because it holds the commands
  //that should be executed, if a command executes Transceive(),
  //m_rxbuf is overwritten.
  memcpy(m_commandbuf, m_rxbuf, SERVERCOMM_BUFSIZE);

  return true;
}

/*! Parses the CBOR in m_commandbuf, executes every command in the "C" array.
    \return true if everything succeeded, false if something failed.*/
bool CServerComm::ExecuteCommands()
{
  CborError  retval;
  CborParser parser;
  CborValue  root;

  retval = cbor_parser_init(m_commandbuf, SERVERCOMM_BUFSIZE,
                            CborValidateBasic, &parser, &root);

  //should have been checked by Transceive();
  assert(retval == CborNoError);
  retval = cbor_value_validate_basic(&root);
  assert(retval == CborNoError);
  assert(cbor_value_is_map(&root));

  //check if the "I" element sent by the server holds the correct device ID
  if (!CheckServerDeviceID(&root))
    return false;

  //if the server reply has a date, set the internal RTC
  CBORSetRTCFromDateField(LOGSERVERCOMM, "root", &root, "D");

  //find the C array, which holds the commands to execute
  CborValue C;
  if (!CBORFindMapArray(LOGSERVERCOMM, "root", &root, "C", &C))
    return false;

  //open the C array
  CborValue command;
  if (!CBOREnterContainer(LOGSERVERCOMM, "C", &C, &command))
    return false;

  //execute all commands in the C array
  while (!cbor_value_at_end(&command))
  {
    if (!ExecuteCommand(&command))
      return false;

    if (!CBORValueAdvance(LOGSERVERCOMM, "C", &command))
      return false;
  }

  //close the C array
  if (!CBORLeaveContainer(LOGSERVERCOMM, "C", &C, &command))
    return false;

  //Download the images file from the url stored in the settings.
  if (!GetImages())
    return false;

  PrintLog(LOGSERVERCOMM, "Done executing commands\n");

  return true;
}

/*! Executes a command in a CBOR map.
    \param command The CBOR map to execute the command from.
    \return true if executing the command succeeded.
    \return false if executing the command failed.*/
bool CServerComm::ExecuteCommand(CborValue* command)
{
  CLogLock loglock;
  PrintLog(LOGSERVERCOMM, "Executing command:");
  cbor_value_to_pretty(stdout, command);
  PrintLog(LOGSERVERCOMM, "\n");
  loglock.Leave();

  //find the C string, which holds the command to execute
  CborValue C;
  if (!CBORFindMapString(LOGSERVERCOMM, "command", command, "C", &C))
    return false;

  bool result;

  //post the settings to the server if requested
  if (!CBORCompareString(LOGSERVERCOMM, "C", &C, "PostSettings", &result))
    return false;
  if (result)
    return PostSettings();

  //get the settings from the server if requested
  if (!CBORCompareString(LOGSERVERCOMM, "C", &C, "GetSettings", &result))
    return false;
  if (result)
    return GetSettings();

  //download a new authorizations list from the server if requested
  if (!CBORCompareString(LOGSERVERCOMM, "C", &C, "GetAuthorisationslist", &result))
    return false;
  if (result)
    return GetAuthorisationslist(command);

  //post the status to the server if requested
  if (!CBORCompareString(LOGSERVERCOMM, "C", &C, "PostStatus", &result))
    return false;
  if (result)
    return PostStatus();

  //post the registrations to the server if requested
  if (!CBORCompareString(LOGSERVERCOMM, "C", &C, "PostRegistrations", &result))
    return false;
  if (result)
    return PostRegistrations();

  //update the firmware if requested
  if (!CBORCompareString(LOGSERVERCOMM, "C", &C, "GetFirmware", &result))
    return false;
  if (result)
  { //GetFirmware never returns if it succeeds, instead it reboots the microcontroller
    GetFirmware(command);
    return false;
  }

  loglock.Enter();
  PrintLog(LOGSERVERCOMM, "ERROR: unknown command: ");
  cbor_value_to_pretty(stdout, &C);
  PrintLog(LOGSERVERCOMM, "\n");
  loglock.Leave();

  return true;
}

/*! Used to test modem/internet connectivity during production,
    when in inactive mode.
    \return true Communication succeeded.
    \return false Communication failed.*/
bool CServerComm::ProductionTest()
{
  PrintLog(LOGSERVERCOMM, "Testing internet connectivity\n");

  CborEncoder encoder;
  cbor_encoder_init(&encoder, m_txbuf, SERVERCOMM_BUFSIZE, 0);

  //Create a root map where all other fields will be added to.
  CborEncoder root;
  cbor_encoder_create_map(&encoder, &root, CborIndefiniteLength);

  //Add default fields.
  AddCommandDefaultFields(&root, "ProductionTest");

  //Add the software version.
  cbor_encode_text_stringz(&root, "V");
  cbor_encode_text_stringz(&root, FirmwareVersion());

  //Add the modem IMEI.
  cbor_encode_text_stringz(&root, "M");
  cbor_encode_text_stringz(&root, m_modem.IMEI());

  //Add the SIM ICCID.
  cbor_encode_text_stringz(&root, "D");
  cbor_encode_text_stringz(&root, m_modem.ICCID());

  //Add the modem RSSI.
  cbor_encode_text_stringz(&root, "R");
  cbor_encode_int(&root, m_modem.RSSI());

  //close the root map and check if everything was encoded correctly
  if (!CBORCloseContainer(LOGSERVERCOMM, "ProductionTest message", &encoder, &root))
    return false;

  //Send the generated CBOR to the HTTP server and receive its reply.
  size_t cborsize = cbor_encoder_get_buffer_size(&encoder, m_txbuf);
  if (!Transceive(cborsize))
  {
    PrintLog(LOGSERVERCOMM, "ERROR: production test failed\n");
    return false;
  }

  CborError  retval;
  CborParser parser;
  CborValue  replyroot;

  retval = cbor_parser_init(m_rxbuf, SERVERCOMM_BUFSIZE, CborValidateBasic,
                            &parser, &replyroot);

  //Should have been checked by Transceive();
  assert(retval == CborNoError);
  retval = cbor_value_validate_basic(&replyroot);
  assert(retval == CborNoError);
  assert(cbor_value_is_map(&replyroot));

  //Check if the "I" element sent by the server holds the correct device ID.
  if (!CheckServerDeviceID(&replyroot))
    return false;

  //If the server reply has a date, set the internal RTC.
  CBORSetRTCFromDateField(LOGSERVERCOMM, "root", &replyroot, "D");

  return true;
}

/*! Posts the settings that are stored to the server.
    \return true if posting the settings succeeded.
    \return false if posting the settings failed.*/
bool CServerComm::PostSettings()
{
  PrintLog(LOGSERVERCOMM, "Sending settings\n");

  CborEncoder encoder;
  cbor_encoder_init(&encoder, m_txbuf, SERVERCOMM_BUFSIZE, 0);

  //create a root map where all other fields will be added to
  CborEncoder root;
  cbor_encoder_create_map(&encoder, &root, CborIndefiniteLength);

  //add default fields
  AddCommandDefaultFields(&root, "PostSettings");

  //add all settings to the root map
  g_controller.Settings().ToServerCBOR(&root);

  //close the root map and check if everything was encoded correctly
  if (!CBORCloseContainer(LOGSERVERCOMM, "PostSettings message", &encoder, &root))
    return false;

  //send the generated CBOR to the HTTP server and receive its reply
  size_t cborsize = cbor_encoder_get_buffer_size(&encoder, m_txbuf);
  if (!Transceive(cborsize))
  {
    PrintLog(LOGSERVERCOMM, "ERROR: Sending settings failed\n");
    return false;
  }

  return true;
}

/*! Gets the settings from the server, and stores them in the internal flash.
    \return true if getting the settings succeeded.
    \return false if getting the settings failed.*/
bool CServerComm::GetSettings()
{
  PrintLog(LOGSERVERCOMM, "Downloading settings\n");

  CborEncoder encoder;
  cbor_encoder_init(&encoder, m_txbuf, SERVERCOMM_BUFSIZE, 0);

  //create a root map where all other fields will be added to
  CborEncoder root;
  cbor_encoder_create_map(&encoder, &root, 3);

  //add default fields
  AddCommandDefaultFields(&root, "GetSettings");

  //close the root map and check if everything was encoded correctly
  if (!CBORCloseContainer(LOGSERVERCOMM, "GetSettings message", &encoder, &root))
    return false;

  //send the generated CBOR to the HTTP server and receive its reply
  size_t cborsize = cbor_encoder_get_buffer_size(&encoder, m_txbuf);
  if (!Transceive(cborsize))
  {
    PrintLog(LOGSERVERCOMM, "ERROR: Downloading settings failed\n");
    return false;
  }

  //parse the downloaded CBOR
  CborError  retval;
  CborParser parser;
  CborValue  settings;
  retval = cbor_parser_init(m_rxbuf, SERVERCOMM_BUFSIZE,
                            CborValidateBasic, &parser, &settings);

  //should have been checked by Transceive();
  assert(retval == CborNoError);
  retval = cbor_value_validate_basic(&settings);
  assert(retval == CborNoError);
  assert(cbor_value_is_map(&settings));

  //store the downloaded settings into the internal flash
  g_controller.Settings().FromServerCBOR(&settings);

  return true;
}

/*! Timeout in seconds for an auth list download,
    this is the timeout for making the connection,
    and the timeout per block of data, so that a download will complete even
    on a very slow connection.*/
#define AUTH_LIST_DOWNLOAD_TIMEOUT 300

/*! Downloads an authorizations list file to the SPI flash,
    copies it to the internal flash after the download is done.
    \param command The CBOR object that holds the url in the U element.
    \return true if the CBOR object is correct and downloading the list succeeded.
    \return false if anything went wrong.*/
bool CServerComm::GetAuthorisationslist(CborValue* command)
{
  //copy the U field from the command map into the url string
  CborValue U;
  if (!CBORFindMapString(LOGSERVERCOMM, "command", command, "U", &U))
    return false;

  char     url[256] = {};
  uint32_t length = sizeof(url) - 1;
  if (!CBORCopyString(LOGSERVERCOMM, "U", &U, url, &length))
    return false;

  PrintLogDecode(LOGSERVERCOMM, "Downloading auth list from \"%s\"", url);

  //prepare the auth list, it will erase the SPI flash
  if (!g_controller.AuthList().Prepare())
  {
    PrintLog(LOGSERVERCOMM, "Peparing auth list failed\n");
    return false;
  }

  //download the auth list to the SPI flash
  //the CModemBG96 class downloads the file, and calls into the CAuthList class
  //through the CHTTPDownload interface
  if (!m_modem.HTTPGet(url, AUTH_LIST_DOWNLOAD_TIMEOUT, &g_controller.AuthList(), true))
  {
    PrintLogDecode(LOGSERVERCOMM, "Downloading auth list from %s failed", url);
    return false;
  }

  //Check if the HTTP server returned a response code
  //indicating the correct file was downloaded.
  if (!g_controller.AuthList().HTTPRspCodeGood())
  {
    PrintLog(LOGSERVERCOMM, "ERROR: HTTP GET returned %" PRIu32 " on \"%s\"\n",
             g_controller.AuthList().HTTPRspCode(), url);
    return false;
  }

  //copy the auth list from the SPI flash to the internal flash
  return g_controller.AuthList().Install();
}

/*! Maximum number of milliseconds that statuses can be posted until aborting.*/
#define POSTSTATUS_TIMEOUT 1800000

/*! Posts recorded statuses to the server.
    \return true if posting the statuses succeeded, or if there are none to post.
    \return false if posting the statuses failed.*/
bool CServerComm::PostStatus()
{
  //Write a Status_FactoryReset status, if this is the first communication
  //after a factory reset.
  CFactoryReset::WriteStatus();

  //add the current status as status records
  g_controller.Status().Write(Status_SoftwareVersion, "%s", FirmwareVersion());
  g_controller.Status().Write(Status_PCBVersion, "%s",
                              g_controller.PCBVersion().VersionStr());

  uint32_t filllevel = lroundf(g_controller.Settings().GetFloat(RunStatus_FillLevel));
  g_controller.Status().Write(Status_FillLevel, "%" PRIu32, filllevel);

  //record a status about how many attempts it took before
  //the server was reached successfully
  uint32_t numfailed = g_controller.Settings().GetInteger(RunStatus_ServerContactFailed);
  g_controller.Status().Write(Status_NrContactAttempts, "%" PRIi32, numfailed + 1);
  g_controller.Settings().SetInteger(RunStatus_ServerContactFailed, 0);

  //record the location reported by the modem's GNSS receiver if requested,
  //and record the cell phone tower the modem is connected to
  RecordLocation();

  //Record the fill level sensor, if enabled in the settings.
  g_controller.Deposit().RegisterFillLevelSensor();

  //keep posting statuses until there are no more statuses stored in the flash
  int64_t start = Millis();
  for(;;)
  {
    //check if there are any statuses to send to the server
    uint32_t numstatuses = g_controller.Status().NumStored();
    if (numstatuses == 0)
    {
      PrintLog(LOGSERVERCOMM, "No status to post\n");
      return true;
    }

    //failsafe, if posting statuses takes too long, abort
    int64_t elapsed = Millis() - start;
    if (elapsed > POSTSTATUS_TIMEOUT)
    {
      PrintLog(LOGSERVERCOMM, "ERROR: posted statuses for %" PRIi64
               " milliseconds, aborting\n", elapsed);
      return false;
    }

    PrintLog(LOGSERVERCOMM, "Posting status, %" PRIu32
             " statuses stored\n", numstatuses);

    memset(m_txbuf, 0, SERVERCOMM_BUFSIZE);
    CborEncoder encoder;
    cbor_encoder_init(&encoder, m_txbuf, SERVERCOMM_BUFSIZE, 0);

    //create a root map where all other fields will be added to
    //this map must be of a fixed size, it can't be CborIndefiniteLength,
    //because the CStatus::Read() method only leaves free space for
    //one CBOR break byte
    CborEncoder root;
    cbor_encoder_create_map(&encoder, &root, 4);

    //add default fields
    AddCommandDefaultFields(&root, "PostStatus");

    //add the statuses array
    //WARNING: the statuses array must always be the last element in the CBOR map,
    //because the CStatus::Read() method copies messages directly
    //into the buffer, leaving only one byte free for the CBOR break byte
    CborEncoder R;
    cbor_encode_text_stringz(&root, "R");
    cbor_encoder_create_array(&root, &R, CborIndefiniteLength);

    //add any statuses from the flash ringbuffer to the statuses array
    numstatuses = g_controller.Status().Read(&R, m_txbuf, SERVERCOMM_BUFSIZE);

    //close the statuses array and check if everything was encoded correctly
    if (!CBORCloseContainer(LOGSERVERCOMM, "statuses array", &root, &R))
      return false;

    //close the root map and check if everything was encoded correctly
    if (!CBORCloseContainer(LOGSERVERCOMM, "Poststatuses message", &encoder, &root))
      return false;

    PrintLog(LOGSERVERCOMM, "Sending %" PRIu32 " status%s\n",
             numstatuses, numstatuses > 1 ? "es" : "");

    //send the generated CBOR to the HTTP server and receive its reply
    size_t cborsize = cbor_encoder_get_buffer_size(&encoder, m_txbuf);
    bool txcborcorrect;
    if (!Transceive(cborsize, &txcborcorrect))
    {
      PrintLog(LOGSERVERCOMM, "ERROR: Posting statuses failed\n");

      //if sending the statuses failed, but the generated CBOR is correct,
      //reset the read counter, so that the statuses are sent again next time
      //if the generated CBOR is not correct, then there is probably wrong CBOR
      //stored in the SPI flash ringbuffer, so clear all statuses read from it
      if (txcborcorrect)
        g_controller.Status().Reset();
      else
        g_controller.Status().Clear();

      return false;
    }

    //if the statuses has been sent successfully, clear them from
    //the flash ringbuffer
    g_controller.Status().Clear();
  }
}

/*! Maximum number of milliseconds that registrations can be posted until aborting.*/
#define POSTREGISTRATIONS_TIMEOUT 1800000

/*! Posts recorded registrations to the server.
    \return true if posting the registrations succeeded, or if there are none to post.
    \return false if posting the registrations failed.*/
bool CServerComm::PostRegistrations()
{
  //keep posting registrations until there are no more registrations stored in the flash
  int64_t start = Millis();
  for(;;)
  {
    //check if there are any registrations to send to the server
    uint32_t numregistrations = g_controller.Registrations().NumStored();
    if (numregistrations == 0)
    {
      PrintLog(LOGSERVERCOMM, "No registations to post\n");
      return true;
    }

    //failsafe, if posting registrations takes too long, abort
    int64_t elapsed = Millis() - start;
    if (elapsed > POSTREGISTRATIONS_TIMEOUT)
    {
      PrintLog(LOGSERVERCOMM, "ERROR: posted registrations for %" PRIi64
               " milliseconds, aborting\n", elapsed);
      return false;
    }

    PrintLog(LOGSERVERCOMM, "Posting registrations, %" PRIu32
             " registrations stored\n", numregistrations);

    memset(m_txbuf, 0, sizeof(m_txbuf));
    CborEncoder encoder;
    cbor_encoder_init(&encoder, m_txbuf, SERVERCOMM_BUFSIZE, 0);

    //create a root map where all other fields will be added to
    //this map must be of a fixed size, it can't be CborIndefiniteLength,
    //because the CRegistrations::Read() method only leaves free space for
    //one CBOR break byte
    CborEncoder root;
    cbor_encoder_create_map(&encoder, &root, 4);

    //add default fields
    AddCommandDefaultFields(&root, "PostRegistrations");

    //add the registrations array
    //WARNING: the registrations array must always be the last element in the CBOR map,
    //because the CRegistrations::Read() method copies messages directly
    //into the buffer, leaving only one byte free for the CBOR break byte
    CborEncoder R;
    cbor_encode_text_stringz(&root, "R");
    cbor_encoder_create_array(&root, &R, CborIndefiniteLength);

    //add any registrations from the flash ringbuffer to the registrations array
    numregistrations = g_controller.Registrations().Read(&R, m_txbuf, SERVERCOMM_BUFSIZE);

    //close the registrations array and check if everything was encoded correctly
    if (!CBORCloseContainer(LOGSERVERCOMM, "registrations array", &root, &R))
      return false;

    //close the root map and check if everything was encoded correctly
    if (!CBORCloseContainer(LOGSERVERCOMM, "PostRegistrations message", &encoder, &root))
      return false;

    PrintLog(LOGSERVERCOMM, "Sending %" PRIu32 " registration%s\n",
             numregistrations, numregistrations > 1 ? "s" : "");

    //send the generated CBOR to the HTTP server and receive its reply
    size_t cborsize = cbor_encoder_get_buffer_size(&encoder, m_txbuf);
    bool txcborcorrect;
    if (!Transceive(cborsize, &txcborcorrect))
    {
      PrintLog(LOGSERVERCOMM, "ERROR: Posting registrations failed\n");

      //if sending the registrations failed, but the generated CBOR is correct,
      //reset the read counter, so that the registrations are sent again next time
      //if the generated CBOR is not correct, then there is probably wrong CBOR
      //stored in the SPI flash ringbuffer, so clear all registrations read from it
      if (txcborcorrect)
        g_controller.Registrations().Reset();
      else
        g_controller.Registrations().Clear();

      return false;
    }

    //if the registrations has been sent successfully, clear them from
    //the flash ringbuffer
    g_controller.Registrations().Clear();
  }
}

/*! Extracts the URL from the U field in a CBOR map and uses it to update the firmware.
    \param command The CBOR map to get the U field from.*/
void CServerComm::GetFirmware(CborValue* command)
{
  //copy the U field from the command map into the url string
  CborValue U;
  if (!CBORFindMapString(LOGSERVERCOMM, "command", command, "U", &U))
    return;

  char     url[256] = {};
  uint32_t length = sizeof(url) - 1;
  if (!CBORCopyString(LOGSERVERCOMM, "U", &U, url, &length))
    return;

  //do the firmware update with the url
  UpdateFirmware(url);
}

/*! Timeout in seconds for a firmware download,
    this is the timeout for making the connection,
    and the timeout per block of data, so that a firmware download will
    complete even on a very slow connection.*/
#define FIRMWARE_DOWNLOAD_TIMEOUT 300

/*! Erases the reserved section of the SPI flash, downloads a firmware image,
    checks it and installs it.
    \param url URL to do the firmware update with.*/
void CServerComm::UpdateFirmware(const char* url)
{
  PrintLogDecode(LOGSERVERCOMM, "Doing a firmware update with \"%s\"", url);

  //prepare the firmware updater, it will erase the SPI flash
  if (!m_fwupdate.Prepare())
  {
    PrintLog(LOGSERVERCOMM, "Peparing firmware update failed\n");
    return;
  }

  //Download the firmware update image to the SPI flash.
  //The CModemBG96 class downloads the firmware file,
  //and calls into the CFwUpdate class through the CHTTPDownload interface.
  if (!m_modem.HTTPGet(url, FIRMWARE_DOWNLOAD_TIMEOUT, &m_fwupdate, true))
  {
    PrintLogDecode(LOGSERVERCOMM, "Downloading firmware image from %s failed", url);
    return;
  }

  //Check if the HTTP server returned a response code
  //indicating the correct file was downloaded.
  if (!m_fwupdate.HTTPRspCodeGood())
  {
    PrintLog(LOGSERVERCOMM, "ERROR: HTTP GET returned %" PRIu32 " on \"%s\"\n",
             m_fwupdate.HTTPRspCode(), url);
    return;
  }

  //check if the firmware image is correct
  if (!m_fwupdate.CheckUpdateImage())
  {
    PrintLog(LOGSERVERCOMM, "Updating firmware failed\n");
    return;
  }

  //power off the modem before installing the firmware update image,
  //because that will reset the microcontroller, otherwise the modem will stay on
  m_modem.Close();

  //install the firmware update image, this call will never return if succeeded,
  //instead, it will reset the entire microcontroller and it will boot the updater firmware
  if (!m_fwupdate.InstallUpdateImage())
  {
    PrintLog(LOGSERVERCOMM, "Installing firmware update failed\n");
    return;
  }
}

/*! Timeout in seconds for an images file download,
    this is the timeout for making the connection,
    and the timeout per block of data, so that a download will complete even
    on a very slow connection.*/
#define IMAGES_FILE_DOWNLOAD_TIMEOUT 300

/*! Downloads the file that Setting_ImagesURL points to, and installs that
    as the images used for the display.*/
bool CServerComm::GetImages()
{
  char url[STRING_VALUE_MAX_LENGTH] = {};
  g_controller.Settings().GetString(Setting_ImagesURL, url, sizeof(url));

  if (strlen(url) == 0)
  { //No images file to download, remove whatever is installed.
    g_controller.UserInterface().Images().EraseInstalled();
    return true;
  }

  if (!g_controller.UserInterface().Images().URLChanged(url))
  {
    PrintLogDecode(LOGSERVERCOMM, "Images file has already been installed"
                   " from \"%s\"", url);
    return true; //File is the same as what is installed.
  }

  PrintLogDecode(LOGSERVERCOMM, "Downloading images file from \"%s\"", url);

  //Erase the SPI flash.
  if (!g_controller.UserInterface().Images().Prepare())
  {
    PrintLog(LOGSERVERCOMM, "ERROR: Peparing images file failed\n");
    return false;
  }

  //Download the images file to the SPI flash.
  //The CModemBG96 class downloads the file, and calls into the CImages class
  //through the CHTTPDownload interface.
  if (!m_modem.HTTPGet(url, IMAGES_FILE_DOWNLOAD_TIMEOUT,
                       &g_controller.UserInterface().Images(), true))
  {
    PrintLogDecode(LOGSERVERCOMM, "ERROR: Downloading images file from %s failed", url);
    return false;
  }

  //Check if the HTTP server returned a response code
  //indicating the correct file was downloaded.
  if (!g_controller.UserInterface().Images().HTTPRspCodeGood())
  {
    PrintLog(LOGSERVERCOMM, "ERROR: HTTP GET returned %" PRIu32 " on \"%s\"\n",
             g_controller.UserInterface().Images().HTTPRspCode(), url);
    return false;
  }

  //Copy the images file from the SPI flash to the internal flash.
  return g_controller.UserInterface().Images().Install(url);
}

/*! Uploads generated CBOR in m_txbuf to the HTTP server, and downloads its CBOR reply.
    \param txsize The size of the generated CBOR in m_txbuf.
    \param txcborcorrect The CBOR that should be transmitted is checked.
                         When not nullptr, the result of that check is returned here.
    \return true if the upload and download succeeded, false if anything failed.*/
bool CServerComm::Transceive(uint32_t txsize, bool* txcborcorrect /*= nullptr*/)
{
  //check if the CBOR that should be uploaded is ok, and print it
  if (!CBORPrint(LOGSERVERCOMM, "Send: ", m_txbuf, txsize))
  {
    PrintLog(LOGSERVERCOMM, "ERROR: CBOR that should be transmitted is incorrect\n");

    if (txcborcorrect)
      *txcborcorrect = false;

    return false;
  }

  if (txcborcorrect)
    *txcborcorrect = true;

  //upload the generated CBOR to the server and download the CBOR reply
  memset(m_rxbuf, 0, SERVERCOMM_BUFSIZE);
  CHTTPTransceiver transceiver(m_txbuf, txsize, m_rxbuf, SERVERCOMM_BUFSIZE);
  if (!m_modem.HTTPPost(m_url, SERVERCOMM_TIMEOUT, &transceiver, true))
  {
    PrintLog(LOGSERVERCOMM, "HTTP POST failed\n");
    return false;
  }

  //check if the return code is in the 200 range
  if (!transceiver.HTTPRspCodeGood())
  {
    PrintLog(LOGSERVERCOMM, "ERROR: HTTP POST returned %" PRIu32 "\n",
             transceiver.HTTPRspCode());
    return false;
  }

  //check if not more bytes were downloaded than fit in m_rxbuf
  if (transceiver.Overflow() > 0)
  {
    PrintLog(LOGSERVERCOMM, "ERROR: RX buffer overflowed by %" PRIu32
             " bytes\n", transceiver.Overflow());
    return false;
  }

  //check if the downloaded CBOR is ok, and print it
  if (!CBORPrint(LOGSERVERCOMM, "Received: ", m_rxbuf, transceiver.Downloaded()))
  {
    PrintLog(LOGSERVERCOMM, "ERROR: downloaded CBOR is incorrect\n");
    return false;
  }

  //check if the S field in the reply is set to 0
  if (!CheckReplyStatus())
    return false;

  return true;
}

/*! Checks if the device ID in the "I" element is the same
    as the device ID of this device.
    \param root The root map that holds the "I" element.
    \return true if the "I" element matches the device ID of this device.
    \return false if the "I" element does not match the device ID of this device.
    \return false on any kind of error.*/
bool CServerComm::CheckServerDeviceID(CborValue* root)
{
  //the I element holds the device ID
  CborValue I;
  if (!CBORFindMapString(LOGSERVERCOMM, "root", root, "I", &I))
    return false;

  //check if the I element returned by the server is the same as the one sent to it
  bool result;
  if (!CBORCompareString(LOGSERVERCOMM, "I", &I, DeviceIDStr(), &result))
    return false;

  if (!result)
  { //if the device ID is different, abort
    CLogLock loglock;
    PrintLog(LOGSERVERCOMM, "ERROR: server returned device ID ");
    cbor_value_to_pretty(stdout, &I);
    PrintLog(LOGSERVERCOMM, ", but the ID of this device is \"%s\"\n",
             DeviceIDStr());
    return false;
  }

  return true;
}

/*! Adds default fields necessary for each command.
    \param root The root map to add the fields to.
    \param command The name of the command for the "C" field.*/
void CServerComm::AddCommandDefaultFields(CborEncoder* root, const char* command)
{
  //add the I field which holds the device ID
  cbor_encode_text_stringz(root, "I");
  cbor_encode_text_stringz(root, DeviceIDStr());

  //add command
  cbor_encode_text_stringz(root, "C");
  cbor_encode_text_stringz(root, command);

  //add the communication sequence number
  cbor_encode_text_stringz(root, "N");
  cbor_encode_int(root, GetServerCommSequenceNr());
}

/*! Reads the "S" field from a CBOR map, checks if its value is 0.
    \param root The CBOR map to get the S field from.
    \param name A string that describes the CBOR map, for logging.
    \return true if the S field is 0.
    \return false on any error, or if the S field is not 0.*/
bool CServerComm::CheckReplyStatus()
{
  CborParser parser;
  CborValue  root;
  cbor_parser_init(m_rxbuf, SERVERCOMM_BUFSIZE, CborValidateBasic, &parser, &root);

  //get the S field from the root map
  CborValue S;
  if (!CBORFindMapValue(LOGSERVERCOMM, "server reply", &root, "S", &S))
    return false;

  //check if the S is an integer
  if (!cbor_value_is_integer(&S))
  {
    CLogLock loglock;
    PrintLog(LOGSERVERCOMM, "ERROR: S field in server reply is not an integer: ");
    cbor_value_to_pretty(stdout, &S);
    PrintLog(LOGSERVERCOMM, "\n");
    return false;
  }

  //read the integer value of S
  int64_t intval;
  CborError retval = cbor_value_get_int64(&S, &intval);
  if (retval != CborNoError)
  {
    PrintLog(LOGSERVERCOMM, "ERROR: cbor_value_get_int64 returned %i:%s on S "
             "field in server reply\n", (int)retval, cbor_error_string(retval));
    return false;
  }

  //if S is not 0, the server had an error, abort
  if (intval != 0)
  {
    PrintLog(LOGSERVERCOMM, "ERROR: S field in server reply is %"
             PRIi64 ", should be 0\n", intval);
    return false;
  }

  return true;
}
