#include "speaker.h"
#include "lock.h"

#include <FreeRTOS.h>
#include <stm32f4xx.h>

/*! Enables or disables the timer1 update interrupt.
    \param newstate ENABLE or DISABLE.*/
static void SetTimer1UpdateInterrupt(FunctionalState newstate)
{
  NVIC_InitTypeDef nvicsettings =
  {
    .NVIC_IRQChannel                   = TIM1_UP_TIM10_IRQn,
    .NVIC_IRQChannelPreemptionPriority = configLIBRARY_MAX_SYSCALL_INTERRUPT_PRIORITY,
    .NVIC_IRQChannelSubPriority        = 0,
    .NVIC_IRQChannelCmd                = newstate
  };
  NVIC_Init(&nvicsettings);
  TIM_ITConfig(TIM1, TIM_IT_Update, newstate);
}

/*! Number of timer1 periods the timer should run.*/
static volatile uint32_t g_periods;

/*! Interval of timer1 periods in which to turn the speaker on and off.*/
static volatile uint32_t g_interval;

/*! IRQ handler of the timer1 update interrupt. */
extern "C" void TIM1_UP_TIM10_IRQHandler()
{
  TIM_ClearITPendingBit(TIM1, TIM_IT_Update);

  if (g_periods > 0)
  {
    g_periods--;

    //Turn the square wave output on and off based on the interval.
    //This creates the number of requested beeps.
    bool disabled = (g_periods / g_interval) & 1;
    TIM_CtrlPWMOutputs(TIM1, disabled ? DISABLE : ENABLE);
  }
  else
  {
    //disable timer1
    SetTimer1UpdateInterrupt(DISABLE);
    TIM_CtrlPWMOutputs(TIM1, DISABLE);
    TIM_Cmd(TIM1, DISABLE);

    //reset timer1 to its default state
    RCC_APB2PeriphResetCmd(RCC_APB2Periph_TIM1, ENABLE);
    RCC_APB2PeriphResetCmd(RCC_APB2Periph_TIM1, DISABLE);

    //disable clock to timer1
    RCC_APB2PeriphClockCmd(RCC_APB2Periph_TIM1, DISABLE);
  }
}

/*! Resonance frequency of the KCVG084B16 speaker.*/
#define SPEAKER_FREQUENCY 2730

/*! Drivers the speaker with a square wave.
    \param time Number of milliseconds each beep should last.
    \param count Number of times to beep the speaker.*/
void CSpeaker::Beep(uint32_t time, uint32_t count /*= 1*/)
{
  if (time < 1)
    time = 1;

  if (count < 1)
    count = 1;

  //use a critical section to protect the RCC and timer registers
  CCritical criticalsection;

  //get the clock frequency of the APB2, which is what clocks timer1,
  RCC_ClocksTypeDef clockfreqs;
  RCC_GetClocksFreq(&clockfreqs);

  //enable the clock for timer1
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_TIM1, ENABLE);

  //reset timer1 to its default state
  RCC_APB2PeriphResetCmd(RCC_APB2Periph_TIM1, ENABLE);
  RCC_APB2PeriphResetCmd(RCC_APB2Periph_TIM1, DISABLE);
  SetTimer1UpdateInterrupt(DISABLE);

  //initialize timer1 as an upcounter clock at 1 tick per microsecond,
  //it overflows at the resonance frequency of the speaker
  TIM_TimeBaseInitTypeDef timinit = {};
  timinit.TIM_Prescaler           = (clockfreqs.SYSCLK_Frequency / 1000000) - 1;
  timinit.TIM_CounterMode         = TIM_CounterMode_Up;
  timinit.TIM_Period              = (1000000 / SPEAKER_FREQUENCY) - 1;
  timinit.TIM_ClockDivision       = TIM_CKD_DIV1;
  TIM_TimeBaseInit(TIM1, &timinit);

  //set output 4 of timer1 to PWM mode with 50% duty cycle
  TIM_OCInitTypeDef ocinit = {};
  ocinit.TIM_OCMode        = TIM_OCMode_PWM2;
  ocinit.TIM_OutputState   = TIM_OutputState_Enable;
  ocinit.TIM_OutputNState  = TIM_OutputState_Disable;
  ocinit.TIM_Pulse         = (500000 / SPEAKER_FREQUENCY) - 1;
  ocinit.TIM_OCPolarity    = TIM_OCPolarity_High;
  ocinit.TIM_OCNPolarity   = TIM_OCPolarity_High;
  ocinit.TIM_OCIdleState   = TIM_OCIdleState_Reset;
  ocinit.TIM_OCNIdleState  = TIM_OCIdleState_Reset;
  TIM_OC4Init(TIM1, &ocinit);

  //Calculate number of timer1 periods needed for 1 beep.
  g_interval = SPEAKER_FREQUENCY * time / 1000;

  //Calculate number of beeps, with number of silences in between.
  uint32_t numintervals = count * 2 - 1;

  //Calculate how long to keep timer1 running.
  g_periods = g_interval * numintervals;

  SetTimer1UpdateInterrupt(ENABLE);

  //enable the timer outputs and the timer itself, this turns on the beep.
  TIM_CtrlPWMOutputs(TIM1, ENABLE);
  TIM_Cmd(TIM1, ENABLE);
}
