#include "status.h"
#include "log.h"
#include "lock.h"
#include "controller.h"
#include "backupregisters.h"
#include "clock.h"
#include "cbortools.h"
#include "stm32wss.h"
#include "gpio.h"
#include "active.h"

#include <cstdio>
#include <cstdarg>

CStatus::CStatus() : m_buf {}, m_value {}
{
  m_mutex = 0;
}

void CStatus::Setup()
{
  //create a mutex to protect m_flashringbuffer and m_buf
  m_mutex = xSemaphoreCreateRecursiveMutex();
}

void CStatus::SetupRingBuffer()
{
  CLock lock(m_mutex);

  if (!m_flashringbuffer.Enabled())
  {
    //put the SPI flash ringbuffer at the end of the SPI flash
    //the stm32wss library has some space reserved at the start of the SPI flash
    //by putting this at the end, a collision is avoided
    uint32_t position = g_controller.SPIFlash().Size() - STATUS_RINGBUFFER_TOTAL_SIZE - STATUS_RINGBUFFER_ADDRESS;
    m_flashringbuffer.Setup(&g_controller.SPIFlash(), position,
                            STATUS_RINGBUFFER_TOTAL_SIZE, STATUS_RINGBUFFER_ELEMENT_SIZE);
  }
}

/*! Sequence number stored in battery backed ram, increased by 1 for every
    status that is written, to prove that no status write was ever missed.*/
#define STATUS_SEQUENCE_REGISTER RTC_BKP_DR18

/*! \return A new sequence number to write in the status object.*/
uint32_t CStatus::GetSequenceNumber()
{
  CCritical criticalsection;

  uint32_t sequencenr = RTC_ReadBackupRegister(STATUS_SEQUENCE_REGISTER);
  RTC_WriteBackupRegister(STATUS_SEQUENCE_REGISTER, sequencenr + 1);

  return sequencenr;
}

/*! Creates a status of the microcontroller boot reason.*/
void CStatus::RegisterBootReason()
{
  char bootreasonstr[256] = {};

  //If the SoftwareRebootTimeHappened() returns true,
  //the microcontroller has been rebooted because the it stayed awake for
  //too long. This indicates that one of the FreeRTOS tasks has frozen.
  if (SoftwareRebootTimeHappened())
    strlcat(bootreasonstr, "software watchdog timer", sizeof(bootreasonstr));

  //add the boot reason status, there might be multiple boot reasons,
  //add all of them as a single string
  uint32_t    resetflags = ResetFlags();
  const char* flagstr;
  while ((flagstr = ResetFlagStr(&resetflags)))
  {
    if (bootreasonstr[0] != 0)
      strlcat(bootreasonstr, ", ", sizeof(bootreasonstr));

    strlcat(bootreasonstr, flagstr, sizeof(bootreasonstr));
  }

  //Check if an RTC event happened, if so add it to the boot reason string,
  //because the microcontroller probably woke up from standby because of this.
  uint32_t rtcflags = GetRTCAlarmBootFlags();
  while ((flagstr = RTCAlarmBootFlagStr(&rtcflags)))
  {
    if (bootreasonstr[0] != 0)
      strlcat(bootreasonstr, ", ", sizeof(bootreasonstr));

    strlcat(bootreasonstr, flagstr, sizeof(bootreasonstr));
  }

  //If the wakeup pin was high during boot, record it as a boot reason.
  if (WakeupPinBootState() == Bit_SET)
  {
    if (bootreasonstr[0] != 0)
      strlcat(bootreasonstr, ", ", sizeof(bootreasonstr));

    strlcat(bootreasonstr, "wakeup pin", sizeof(bootreasonstr));
  }

  Write(Status_SoftReset, "%s", bootreasonstr);
}

/*! Writes a status string using vsnprintf to a buffer, and stores it as a CBOR
    object in the flash ringbuffer.
    \param status A value of \ref EStatus.
    \param format A printf style formatted string.*/
void CStatus::Write(EStatus status, const char* format, ...)
{
  CLock lock(m_mutex);

  //set up the SPI flash ringbuffer before the first write,
  //because it might take a few seconds,
  SetupRingBuffer();

  //print the value string of the status to m_value
  memset(m_value, 0, sizeof(m_value));
  va_list args;
  va_start(args, format);
  vsnprintf(m_value, sizeof(m_value), format, args);
  va_end(args);
  m_value[sizeof(m_value) - 1] = 0;

  uint32_t maxbufbytes = MAX_RINGBUFFER_PAYLOAD_SIZE(STATUS_RINGBUFFER_ELEMENT_SIZE);
  memset(m_buf, 0, sizeof(m_buf));
  CborEncoder encoder;
  cbor_encoder_init(&encoder, m_buf, maxbufbytes, 0);

  //create a root map to add all other elements
  CborEncoder root;
  cbor_encoder_create_map(&encoder, &root, 4);

  //add the status number as the T field
  cbor_encode_text_stringz(&root, "T");
  cbor_encode_int(&root, status);

  //Add the date as the D field, don't add a timezone to the date.
  //Write the status 1 second earlier in case of Status_SoftReset,
  //so that that status is always the first one in TACT NE.
  char datebuf[DATESTR_MAX_LENGTH] = {};
  GetRTCISO8601(datebuf, DATESTR_MAX_LENGTH, "", status == Status_SoftReset ? -1 : 0);
  cbor_encode_text_stringz(&root, "D");
  cbor_encode_text_stringz(&root, datebuf);

  //add the value of the status as the V field
  cbor_encode_text_stringz(&root, "V");
  cbor_encode_text_stringz(&root, m_value);

  //add the sequence number as the N field
  cbor_encode_text_stringz(&root, "N");
  cbor_encode_int(&root, GetSequenceNumber());

  //close the root object and check if everything was encoded correctly
  if (!CBORCloseContainer(LOGSTATUS, "status", &encoder, &root))
    return;

  //print the generated CBOR to the log
  uint32_t cborsize = cbor_encoder_get_buffer_size(&encoder, m_buf);

  //If the device is inactive, only certain statuses should be stored.
  //this prevents sending a large number of statuses when the device is made active.
  bool store = StoreInactive(status);

  CLogLock loglock;

  if (store)
    PrintLog(LOGSTATUS, "Storing status %s(%i): ", StatusToStr(status), (int)status);
  else
    PrintLog(LOGSTATUS, "Device inactive, not storing status %s(%i): ",
             StatusToStr(status), (int)status);

  if (!CBORPrint(LOGSTATUS, nullptr, m_buf, cborsize))
    return;

  loglock.Leave();

  if (store)
  {
    //Store the generated CBOR status in the flash ringbuffer.
    if (!m_flashringbuffer.Write(m_buf, cborsize))
      PrintLog(LOGSTATUS, "ERROR: storing CBOR status of size %"
               PRIu32 " failed\n", cborsize);
  }

  uint32_t numstored = m_flashringbuffer.NumStoredElements();
  PrintLog(LOGSTATUS, "%" PRIu32 " status%s stored\n", numstored,
           numstored != 1 ? "es" : "");
}

/*! Called from the servercomm task, reads statuses from the flash ringbuffer,
    and adds them to the statuses array.
    \param statuses The CBOR array to add the statuses to.
    \param buf The buffer that the CBOR encoder was created with.
    \param bufsize The size of buf.
    \return The number of statuses written to the statuses array.*/
uint32_t CStatus::Read(CborEncoder* statuses, uint8_t* buf, uint32_t bufsize)
{
  CLock lock(m_mutex);
  SetupRingBuffer();

  uint32_t numstatuses = 0;
  for(;;)
  {
    //check how much free space there is in the CBOR buffer
    //always make sure there is at least one byte free for the CBOR break control code
    uint32_t freespace = bufsize - cbor_encoder_get_buffer_size(statuses, buf);
    if (freespace > 0)
      freespace--;

    //read one registration from the ringbuffer
    memset(m_buf, 0, sizeof(m_buf));
    uint32_t size;
    if (!m_flashringbuffer.Read(m_buf, &size, false))
      break;

    //if the registration doesn't fit in the CBOR buffer, stop
    if (size >= freespace)
      break;

    //copy the registration directly into the CBOR array
    memcpy(statuses->data.ptr, m_buf, size);
    statuses->data.ptr += size;
    numstatuses++;
    m_flashringbuffer.AdvanceReadPosition();
  }

  return numstatuses;
}

/*! Returns the number of statuses stored in the flash ringbuffer.*/
uint32_t CStatus::NumStored()
{
  CLock lock(m_mutex);
  SetupRingBuffer();
  return m_flashringbuffer.NumStoredElements();
}

/*! Called from the servercomm task, clears statuses that have been read
    and advanced beyond in the flash ringbuffer.
    This is called after the servercomm task has sent the statuses to the server,
    and the server has received them.*/
void CStatus::Clear()
{
  CLock lock(m_mutex);
  m_flashringbuffer.ClearReadElements();
}

/*! Called from the servercomm task, resets the read counter of the
    flash ringbuffer, in case statuses have been read, but the server did not
    receive them.*/
void CStatus::Reset()
{
  CLock lock(m_mutex);
  m_flashringbuffer.ResetReadCounter();
}

/*! \return true if the status should be stored.
    \return false if the status should not be stored,
                  because the device is inactive.*/
bool CStatus::StoreInactive(EStatus status)
{
  if (!DeviceActive())
  {
    if (status != Status_MadeActive &&
        status != Status_SelfTestPassed &&
        status != Status_CableTestPassed)
      return false;
  }

  return true;
}
