#include "userinterface.h"
#include "clock.h"
#include "log.h"
#include "lock.h"
#include "gpiodescriptions.h"
#include "imagesbuiltin.h"
#include "speaker.h"
#include "controller.h"
#include "stm32wss.h"
#include "active.h"

/*! Interval in milliseconds that the service and inactive displays are updated.*/
#define SERVICE_UPDATE_TIME 1000

//! Time in milliseconds that a power on reset is indicated.
#define POR_SHOW_TIME 100

CUserInterface::CUserInterface()
{
  m_orientation         = Orientation_Top;
  m_state               = UIState_Off;
  m_newstate            = UIState_Off;
  m_resetfilllevelstate = 0;
  m_resetfillleveltime  = 0;
  m_updatetime          = 0;
  m_label               = nullptr;
  m_oldprepaidvalue     = 0;
  m_newprepaidvalue     = 0;
  m_str                 = "";
  m_errorcode           = EC_None;
  m_mutex               = nullptr;
  m_clientmutex         = nullptr;
  m_signal              = nullptr;
}

void CUserInterface::Setup()
{
  m_orientation = (EOrientation)
    g_controller.Settings().GetInteger(Setting_DisplayOrientation);

  m_mutex       = xSemaphoreCreateRecursiveMutex();
  m_clientmutex = xSemaphoreCreateRecursiveMutex();
  m_signal      = xSemaphoreCreateBinary();

  IndicatePowerOn();

  m_images.Setup();
}

/*! Briefly turns on the backlight and beeps the speaker to indicate
    that a power on reset has happened.
    This is useful when a power cycle is done using the reed contact.*/
void CUserInterface::IndicatePowerOn()
{
  if (ResetFlags() & ResetFlagMask(RCC_FLAG_PORRST))
  {
    CSpeaker::Beep(POR_SHOW_TIME);

    GPIO_Set(LCD_PWR);
    SetBacklight(true);
    DelayMS(POR_SHOW_TIME);
    SetBacklight(false);
    GPIO_Reset(LCD_PWR);
  }
}

void CUserInterface::Start()
{
  Create("UserInterface");
}

void CUserInterface::Process()
{
  //Set up the display, with the orientation controlled by the settings.
  m_display.Setup(m_orientation);

  //create a background image to show on the LCD, make it hidden initially
  lv_obj_t* image = lv_img_create(m_display.Screen(), NULL);
  lv_obj_set_hidden(image, true);

  //Create a label over the entire LCD, where the service text can be shown.
  m_label = lv_label_create(m_display.Screen(), NULL);
  lv_label_set_text_fmt(m_label, "");

  //Lock m_images so that the internal flash won't be overwritten while an image is in use.
  CLock imglock(m_images);

  for(;;)
  {
    CLock lock(m_mutex);
    EUIState newstate   = m_newstate;
    const char* str     = m_str;
    ErrorCode errorcode = m_errorcode;
    lock.Leave();

    //Set the background image based on which state is selected.
    //This should be done every iteration, because the data member of imgdsc
    //might become invalid when imglock is released.
    EImage       stateimg = StateImg(newstate);
    lv_img_dsc_t imgdsc;
    m_images.Get(stateimg, &imgdsc);
    lv_img_set_src(image, &imgdsc);

    if (newstate != m_state)
    {
      m_state = newstate;

      if (m_state == UIState_Off)
      {
        //turn off the LCD
        m_display.Disable();

        //in the off state, hide the background image
        lv_obj_set_hidden(image, true);
      }
      else
      {
        //Show the selected image only when the service state or
        //string state is not selected.
        lv_obj_set_hidden(image, m_state == UIState_Service ||
                                 m_state == UIState_Inactive ||
                                 m_state == UIState_String);

        //Show the label when text should be shown.
        lv_obj_set_hidden(m_label, m_state != UIState_Service &&
                                   m_state != UIState_Inactive &&
                                   m_state != UIState_SuccessPrepaid &&
                                   m_state != UIState_SuccessPrepaidLowCredit &&
                                   m_state != UIState_String &&
                                   m_state != UIState_Error);

        //turn on the LCD
        m_display.Enable();
      }
    }

    //Fill m_label with service information.
    if (m_state == UIState_Service)
      ProcessService();
    else if (m_state == UIState_Inactive)
      ProcessInactive();
    else if (m_state == UIState_SuccessPrepaid ||
             m_state == UIState_SuccessPrepaidLowCredit)
      ProcessPrepaid();
    else if (m_state == UIState_String)
      ProcessString(str);
    else if (m_state == UIState_Error)
      ProcessError(errorcode);

    //call the LittlevGL process cycle
    lv_task_handler();

    //Wait for the display state to change. During this time, the display images
    //might be updated by the servercomm task.
    imglock.Leave();
    xSemaphoreTake(m_signal, pdMS_TO_TICKS(100));
    imglock.Enter();

    //blink the backlight and beep the speaker if necessary
    if (m_state == UIState_ResetFillLevelAnim)
      ProcessResetFillLevelAnimation();
  }
}

/*! Can be called from different tasks to set what should be displayed on the LCD.
    Should not be called from the CUserInterface task, SetState() can be called
    from the CUserInterface task.
    \param state The state the user interface should be set at.*/
void CUserInterface::Set(EUIState state)
{
  //lock m_clientmutex to make sure that no other client has exclusive access,
  //a client might lock m_clientmutex while waiting for the fill level reset
  //animation, to prevent another task from changing what it shown on the display
  CLock lock(m_clientmutex);

  SetState(state);
}

/*! Sets the state of the user interface, without locking m_clientmutex.
    This can be called from the CUserInterface task.
    \param state The state the user interface should be set at.*/
void CUserInterface::SetState(EUIState state)
{
  CLock lock(m_mutex);
  if (state != m_newstate)
  {
    //change the state of the backlight as quickly as possible, so that the
    //user sees it turning on immediately when a card is held to the NFC reader
    SetBacklight(!(state == UIState_Off));

    m_updatetime = Millis() - SERVICE_UPDATE_TIME;

    if (state == UIState_ResetFillLevelAnim)
      StartResetFillLevelAnimation();

    PrintLog(LOGUSRINTRFCE, "State changed from %s to %s\n",
             StateStr(m_newstate), StateStr(state));
    m_newstate = state;
    xSemaphoreGive(m_signal);
  }
}

/*! Maximum time a task can wait for the user interface state to change.*/
#define MAX_WAIT_TIME (5 * 60 * 1000000)

/*  \param state When m_newstate is not this value, this function returns.*/
void CUserInterface::WaitStateFinish(EUIState state)
{
  uint32_t start = Micros();

  //Uses a polling loop instead of a semaphore because multiple tasks might be
  //waiting on this at the same time.
  CLock lock(m_mutex);
  while (m_newstate == state)
  {
    lock.Leave();
    DelayMS(100);

    //Timeout for safety.
    if (Micros() - start > MAX_WAIT_TIME)
    {
      PrintLog(LOGUSRINTRFCE, "ERROR: timeout waiting for user interface state"
               " to change away from %i:%s\n", (int)state, StateStr(state));
      return;
    }

    lock.Enter();
  }
}

/*! Time in milliseconds between changes in the fill level reset animation.*/
#define FILL_LEVEL_ANIMATION 500

/*! Number of times the fill level reset animation should run.*/
#define FILL_LEVEL_ROUNDS 18

/*! Initializes a fill level reset animation.*/
void CUserInterface::StartResetFillLevelAnimation()
{
  m_resetfilllevelstate = 0;
  m_resetfillleveltime  = Millis();

  //Sound the initial beep as soon as possible.
  CSpeaker::Beep(FILL_LEVEL_ANIMATION);
}

/*! Flashes the backlight and beeps the speaker for as long as the fill level
    reset animation should run.*/
void CUserInterface::ProcessResetFillLevelAnimation()
{
  CLock lock(m_mutex);

  //Calculate which state the fill level reset animation is in,
  //by calculating how long the animation is running, and dividing that
  //by how long each iteration of the animation should take.
  int64_t  elapsed = Millis() - m_resetfillleveltime;
  uint32_t state = elapsed / FILL_LEVEL_ANIMATION;

  if (state > m_resetfilllevelstate)
  {
    m_resetfilllevelstate = state;

    //Turn on the backlight and beep the speaker on even rounds,
    //turn the backlight off on odd rounds.
    bool on = !(state % 2);
    if (on)
    {
      SetBacklight(!(state % 2));
      CSpeaker::Beep(FILL_LEVEL_ANIMATION);
    }
    else
    {
      SetBacklight(false);
      if (state >= FILL_LEVEL_ROUNDS)
      {
        m_resetfilllevelstate = 0;
        m_resetfillleveltime  = 0;
        SetState(UIState_Off);
      }
    }
  }
}

/*! Creates the service display.*/
void CUserInterface::ProcessService()
{
  int64_t now = Millis();

  CLock lock(m_mutex);
  if (now - m_updatetime >= SERVICE_UPDATE_TIME)
  {
    m_updatetime = now;
    lock.Leave(); //No need to hold the lock when creating the service info.

    //Align the label to the top left of the display.
    lv_label_set_align(m_label, LV_LABEL_ALIGN_LEFT);
    lv_obj_align(m_label, NULL, LV_ALIGN_IN_TOP_LEFT, 0, 0);

    //Read the battery state.
    SBatteryInfo batteryinfo;
    g_controller.BatteryCharger().ReadChargerInfo(&batteryinfo);

    //The battery current measurement is only meaningful when charging.
    if (!batteryinfo.m_charging)
      batteryinfo.m_chargecurrent = 0.0f;

    tm timeval = {};
    GetRTC(&timeval);

    //Get the modem RSSI and if the server is currently being contacted.
    int32_t     rssi        = g_controller.ServerComm().RSSI();
    const char* contacting  = g_controller.ServerComm().IsContacting() ? "Y" : "N";

    char rssistr[16] = {};
    if (rssi == -1) //If the RSSI is -1, it's not available yet.
      strcpy(rssistr, "NA");
    else
      sprintf(rssistr, "%-2" PRIi32, rssi);

    //Get the information stored about the previous server contact attempt.
    //RunStatus_LastContactIndex points to the most recent server contact attempt information.
    int32_t lastcontactindex = g_controller.Settings().GetInteger(RunStatus_LastContactIndex);
    EStorageKey attemptindexes[] =
    { RunStatus_LastContactAttempt1, RunStatus_LastContactAttempt2, RunStatus_LastContactAttempt3 };

    //Copy the server contact attempts into char arrays.
    char  lastattempts[3][16] = {};
    for (int32_t i = 0; i < 3; i++)
    {
      int32_t index = lastcontactindex - i;
      if (index < 0)
        index += 3;

      index = (uint32_t)index % 3; //Safeguard.

      g_controller.Settings().GetString(attemptindexes[index], lastattempts[i],
                                        sizeof(lastattempts[i]) - 1);
    }

    //Print the service into m_label.
    //7 lines can be shown with 14 characters per line.
    lv_label_set_text_fmt(
      m_label,

      //Device ID, firmware version
      "%s %s\n"
      //B = internal battery voltage, E = external input voltage
      "B:%.2f E:%.2f\n"
      //C = currently charging internal batttery, I = internal battery charge current
      "C:%s I:%.3f\n"
      //Current time, R = last known RSSI, D = currently contacting server
      "%02i:%02i R:%s D:%s\n"
      //Last 3 known server contact attempts
      "%s\n"
      "%s\n"
      "%s\n",

      //Arguments for each line in the service display.
      DeviceIDStr(), FirmwareVersion(),
      batteryinfo.m_intvoltage, batteryinfo.m_extvoltage,
      batteryinfo.m_charging ? "Y" : "N", batteryinfo.m_chargecurrent,
      timeval.tm_hour, timeval.tm_min, rssistr, contacting,
      lastattempts[0],
      lastattempts[1],
      lastattempts[2]
      );
  }
}

/*! Shows information when the device is inactive mode.*/
void CUserInterface::ProcessInactive()
{
  int64_t now = Millis();

  CLock lock(m_mutex);
  if (now - m_updatetime >= SERVICE_UPDATE_TIME)
  {
    m_updatetime = now;
    lock.Leave(); //No need to hold the lock when creating the service info.

    //Align the label to the top left of the display.
    lv_label_set_align(m_label, LV_LABEL_ALIGN_LEFT);
    lv_obj_align(m_label, NULL, LV_ALIGN_IN_TOP_LEFT, 0, 0);

    //Read the battery state.
    SBatteryInfo batteryinfo;
    g_controller.BatteryCharger().ReadChargerInfo(&batteryinfo);

    //The battery current measurement is only meaningful when charging.
    if (!batteryinfo.m_charging)
      batteryinfo.m_chargecurrent = 0.0f;

    uint32_t tx = g_controller.Settings().GetInteger(InactiveStatus_ServerComm_TxBytes);
    uint32_t rx = g_controller.Settings().GetInteger(InactiveStatus_ServerComm_RxBytes);
    uint32_t at = g_controller.Settings().GetInteger(InactiveStatus_ServerComm_Attempts);

    bool selftest = g_controller.Settings().GetBoolean(InactiveStatus_SelfTest_Passed);
    bool cabletest = g_controller.Settings().GetBoolean(InactiveStatus_CableTest_Passed);

    //Show on the leds if all tests have been passed.
    if (selftest && cabletest)
      g_controller.GUILeds().On(GUILED_ALL_GRN);
    else
      g_controller.GUILeds().On(GUILED_ALL_RED);

    //Print the service into m_label.
    //7 lines can be shown with 14 characters per line.
    lv_label_set_text_fmt(
      m_label,
      "INACTIVE\n"
      //Device ID, firmware version
      "%s %s\n"
      //B = internal battery voltage, E = external input voltage
      "B:%.2f E:%.2f\n"
      //C = currently charging internal batttery, I = internal battery charge current
      "C:%s I:%.3f\n"
      //D = data usage , A = communication attempts
      "A:%" PRIu32 " D:%.1fkB\n"
      //St = self test passed, Ct = cable test passed
      "St:%s Ct:%s\n",

      //Arguments for each line in the service display.
      DeviceIDStr(), FirmwareVersion(),
      batteryinfo.m_intvoltage, batteryinfo.m_extvoltage,
      batteryinfo.m_charging ? "Y" : "N", batteryinfo.m_chargecurrent,
      at, (float)(tx + rx)/ 1024.0f,
      selftest ? "Y" : "N", cabletest ? "Y" : "N"
      );
  }
}

/*! Sets the befor and after prepaid values to be shown on the display.
    \param prepaid Pointer to the CPrepaid instance belonging to the deposit.*/
void CUserInterface::SetPrepaidInfo(CPrepaid* prepaid)
{
  m_oldprepaidvalue = prepaid->InitialValue();
  m_newprepaidvalue = prepaid->ResultValue();
}

/*! Shows the old and new prepaid values on the display.*/
void CUserInterface::ProcessPrepaid()
{
  //Align the label to the top left of the display.
  lv_label_set_align(m_label, LV_LABEL_ALIGN_LEFT);

  lv_label_set_text_fmt(m_label, "Tegoed:\n"
                                 "Huidig: %" PRIi32 "\n"
                                 "Vorig : %" PRIi32,
                        m_newprepaidvalue,
                        m_oldprepaidvalue);

  lv_obj_align(m_label, NULL, LV_ALIGN_IN_BOTTOM_LEFT, 0, 0);
}

/*! Shows a custom text string on the display.
    \param str The text string to show on the display, a copy of m_str.*/
void CUserInterface::ProcessString(const char* str)
{
  //Align the label to the top left of the display.
  lv_label_set_align(m_label, LV_LABEL_ALIGN_LEFT);
  lv_obj_align(m_label, NULL, LV_ALIGN_IN_TOP_LEFT, 0, 0);

  lv_label_set_text(m_label, str);
}

/*! Shows a numeric error code on the top left of the display.
    \param errorcode The error code to show on the display.*/
void CUserInterface::ProcessError(ErrorCode errorcode)
{
  //Align the label to the top left of the display.
  lv_label_set_align(m_label, LV_LABEL_ALIGN_LEFT);
  lv_obj_align(m_label, NULL, LV_ALIGN_IN_TOP_LEFT, 0, 0);

  if (errorcode != EC_None)
    lv_label_set_text_fmt(m_label, "%" PRIi32, (int32_t)errorcode);
  else
    lv_label_set_text(m_label, "");
}

/*! Sets the error code to display and sets the state to UIState_Error.
    \param errorcode the error code to display.*/
void CUserInterface::SetError(ErrorCode errorcode)
{
  CLock lock(m_mutex);
  m_errorcode = errorcode;
  lock.Leave();

  Set(UIState_Error);
}

/*! Turns the backlight on or off.
    \param state The state the backlight should be set to.*/
void CUserInterface::SetBacklight(bool state)
{
  if (state)
    GPIO_Set(LCD_BACKLIGHT);
  else
    GPIO_Reset(LCD_BACKLIGHT);
}

/*! Sets the string to display in UIState_String,
    and switches the UI state to UIState_String.*/
void CUserInterface::SetString(const char* str)
{
  CLock lock(m_mutex);
  m_str = str;
  lock.Leave();

  SetState(UIState_String);
}

/*! Converts an EUIState value to a string.
    \param state The EUIState to convert.
    \return A null terminated string.*/
const char* CUserInterface::StateStr(EUIState state)
{
  if (state == UIState_Off)
    return "off";
  else if (state == UIState_Locked)
    return "locked";
  else if (state == UIState_Unlocked)
    return "unlocked";
  else if (state == UIState_Denied)
    return "denied";
  else if (state == UIState_Success)
    return "success";
  else if (state == UIState_SuccessPrepaid)
    return "success prepaid";
  else if (state == UIState_SuccessPrepaidLowCredit)
    return "success prepaid low credit";
  else if (state == UIState_Error)
    return "error";
  else if (state == UIState_Full)
    return "full";
  else if (state == UIState_Empty)
    return "empty";
  else if (state == UIState_ServerComm)
    return "server communication";
  else if (state == UIState_ResetFillLevelAnim)
    return "fill level reset animation";
  else if (state == UIState_InsufficientCredit)
    return "insufficient credit";
  else if (state == UIState_CardError)
    return "card defective";
  else if (state == UIState_Service)
    return "service";
  else if (state == UIState_String)
    return "text string";
  else if (state == UIState_Inactive)
    return "inactive mode";
  else
    return "UNKNOWN";
}

/*! Returns the image to show for the specific user interface state.
    \param state The state of the user interface.
    \return The image type to show for the user interface state.*/
EImage CUserInterface::StateImg(EUIState state)
{
  static const struct
  {
    EUIState state;
    EImage   image;
  }
  stateimgs[] =
  {
    { UIState_Locked,                  Image_Locked             },
    { UIState_Unlocked,                Image_Unlocked           },
    { UIState_Denied,                  Image_Denied             },
    { UIState_Success,                 Image_Success            },
    { UIState_SuccessPrepaid,          Image_SuccessPrepaid     },
    { UIState_SuccessPrepaidLowCredit, Image_LowCredit          },
    { UIState_Error,                   Image_Error              },
    { UIState_Full,                    Image_Full               },
    { UIState_Empty,                   Image_Empty              },
    { UIState_ServerComm,              Image_ServerComm         },
    { UIState_ResetFillLevelAnim,      Image_Empty              },
    { UIState_InsufficientCredit,      Image_InsufficientCredit },
    { UIState_CardError,               Image_CardError          },
  };

  for (uint32_t i = 0; i < sizeof(stateimgs) / sizeof(stateimgs[0]); i++)
  {
    if (stateimgs[i].state == state)
      return stateimgs[i].image;
  }

  return Image_Error; //Fallback.
}
