#ifndef BACKUPREGISTERS_H
#define BACKUPREGISTERS_H

#include <stm32f4xx.h>
#include <stdbool.h>

//if BOOT is defined, this header is included in the boot firmware,
//which does not have FreeRTOS
#ifndef BOOT
  #include <FreeRTOS.h>
  #include <task.h>
#endif //BOOT

//battery backed registers can be reserved for special purposes
//the stm32wss library starts at RTC_BKP_DR0 and goes up, the derived
//projects start at RTC_BKP_DR19 and go down

//flags in the stm32wss library start at the least significant bit
//flags in derived projects start at the most significant bit
#define FLAGS_REGISTER                        RTC_BKP_DR0
#define BOOT_FLAG_START_UPDATER               0x1  //!< Boot firmware should start the update firmware.
#define RTC_VALID_FLAG                        0x2  //!< RTC has been set to a valid time.
#define RTC_INITIALIZED_FLAG                  0x4  //!< RTC has been activated.
#define DISABLE_WATCHDOG_FLAG                 0x8  //!< Boot firmware should not turn on the IWDG, once.
#define SOFTWARE_REBOOTTIME_FLAG              0x10 //!< System has been rebooted because the reboot time expired.
#define FIRMWARE_UPDATED_FLAG                 0x20 //!< New firmware has been written to the internal flash.
//0x80000000 on the Accello is reserved for the fill level switch state
//0x40000000 on the Accello is reserved for enabling low power wakeup at boot
//0x20000000 on the Accello is reserved for remembering the lock state
//0x10000000 on the Accello is reserved for remembering the previous self test state
//0x8000000  on the Accello is reserved for remembering that a factory reset happened.
//0x4000000  on the Accello is reserved for remembering to write a factory reset status.
//0x2000000  on the Accello is reserved for remembering the active test state.
//0x1000000  on the Accello is reserved for remembering the stayed open detection.
//0x800000   on the Accello is reserved for remembering the stayed open detection.
//0x400000   on the Accello is reserved for the door switch state.
//0x200000   on the Accello is reserved for the door status state.

#define RANDOM_SEED_REGISTER                  RTC_BKP_DR1

#define SPI_FLASH_LOG_WRITE_POSITION_REGISTER RTC_BKP_DR2
#define SPI_FLASH_LOG_FREE_SPACE_REGISTER     RTC_BKP_DR3

#define SERVERCOMM_SEQUENCE_NUMBER_REGISTER   RTC_BKP_DR4

// RTC_BKP_DR13 on the Accello is reserved for storing the deposit timer seconds.
// RTC_BKP_DR14 on the Accello is reserved for storing the number of power on resets.
// RTC_BKP_DR15 on the Accello is reserved for storing passed self tests.
// RTC_BKP_DR16 on the Accello is reserved for the deposit timer wakeup counter.
// RTC_BKP_DR17 on the Accello is reserved for the number of failed servercomm attempts.
// RTC_BKP_DR18 on the Accello is reserved for the status sequence number.
// RTC_BKP_DR19 on the Accello is reserved for the number of servercomm attempts.

// RTC_BKP_DR16 on the eCat is reserved for the received trap sequence number.
// RTC_BKP_DR17 on the eCat was reserved for the servercomm sequence number, is now free, but not guaranteed to be 0.
// RTC_BKP_DR18 on the eCat is reserved for the LoRa event number
// RTC_BKP_DR19 on the eCat is reserved for the LoRa sequence number

// RTC_BKP_DR18 on the TMT is reserved for the status sequence number.
// RTC_BKP_DR19 on the TMT is reserved for the registrations sequence number.

/*! In reset state, the RCC_BDCR register, the RTC registers (including the
    backup registers), and the BRE bit of the PWR_CSR register, are protected
    against parasitic write access.
    The DBP bit in the PWR_CR register should be set to allow RTC access,
    but for access to the DBP bit, the clock to the power block should be
    enabled first.*/
static inline void EnableBackupAccess()
{
  //Enable clock to the power block so that the backup access can be enabled.
  RCC_APB1PeriphClockCmd(RCC_APB1Periph_PWR, ENABLE); 

  //Set the DBP bit to enable access to the backup domain.
  PWR_BackupAccessCmd(ENABLE);

  //If not enabled, enable the backup regulator so that the backup SRAM keeps its
  //contents in standby mode. After enabling it, wait for the BRR flag that
  //indicates that the backup regulator is ready.
  if (!PWR_GetFlagStatus(PWR_FLAG_BRR))
  {
    PWR_BackupRegulatorCmd(ENABLE);
    for (volatile uint32_t i = 0; i < 1000000; i++)
    {
      if (PWR_GetFlagStatus(PWR_FLAG_BRR))
        break;
    }
  }

  //Enable the backup SRAM clock.
  RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_BKPSRAM, ENABLE);
}

/*! Sets or clears a flag in FLAGS_REGISTER
    \param flag The flag to set to clear.
    \param enabled true to set the flag, false to clear the flag.*/
static inline void SetBackupFlag(uint32_t flag, bool enabled)
{
#ifndef BOOT
  //Only enter a critical section when the FreeRTOS kernel has been started.
  bool lock = xTaskGetSchedulerState() != taskSCHEDULER_NOT_STARTED;
  UBaseType_t uxSavedInterruptStatus = 0;
  if (lock)
  {
    if (xPortIsInsideInterrupt())
      uxSavedInterruptStatus = taskENTER_CRITICAL_FROM_ISR();
    else
      taskENTER_CRITICAL();
  }
#endif

  uint32_t flagsreg = RTC_ReadBackupRegister(FLAGS_REGISTER);

  if (enabled)
    flagsreg |= flag;
  else
    flagsreg &= ~flag;

  RTC_WriteBackupRegister(FLAGS_REGISTER, flagsreg);

#ifndef BOOT
  if (lock)
  {
    if (xPortIsInsideInterrupt())
      taskEXIT_CRITICAL_FROM_ISR(uxSavedInterruptStatus);
    else
      taskEXIT_CRITICAL();
  }
#endif
}

/*! Reads a flag from FLAGS_REGISTER.
    \param flag The flag to read from FLAGS_REGISTER.
    \return true if the flag is set.
    \return false if the flag is cleared.*/
static inline bool ReadBackupFlag(uint32_t flag)
{
  uint32_t flagsreg = RTC_ReadBackupRegister(FLAGS_REGISTER);

  return (flagsreg & flag) == flag;
}

//sets or clears BOOT_FLAG_START_UPDATER in FLAGS_REGISTER
static inline void SetBootUpdaterFlag(bool enabled)
{
  SetBackupFlag(BOOT_FLAG_START_UPDATER, enabled);
}

//returns if BOOT_FLAG_START_UPDATER is set in FLAGS_REGISTER
static inline bool BootUpdateFlagSet()
{
  return ReadBackupFlag(BOOT_FLAG_START_UPDATER);
}

//sets or clears RTC_VALID_FLAG in FLAGS_REGISTER 
static inline void SetRTCValidFlag(bool enabled)
{
  SetBackupFlag(RTC_VALID_FLAG, enabled);
}

//returns if RTC_VALID_FLAG is set in FLAGS_REGISTER
static inline bool RTCValidFlagSet()
{
  return ReadBackupFlag(RTC_VALID_FLAG);
}

//sets or clears RTC_INITIALIZED_FLAG in FLAGS_REGISTER 
static inline void SetRTCInitializedFlag(bool enabled)
{
  SetBackupFlag(RTC_INITIALIZED_FLAG, enabled);
}

//returns if RTC_INITIALIZED_FLAG is set in FLAGS_REGISTER
static inline bool RTCInitializedFlagSet()
{
  return ReadBackupFlag(RTC_INITIALIZED_FLAG);
}

//sets or clears DISABLE_WATCHDOG_FLAG in FLAGS_REGISTER
static inline void SetDisableWatchdogFlag(bool enabled)
{
  SetBackupFlag(DISABLE_WATCHDOG_FLAG, enabled);
}

//returns if DISABLE_WATCHDOG_FLAG is set in FLAGS_REGISTER
static inline bool DisableWatchdogFlagSet()
{
  return ReadBackupFlag(DISABLE_WATCHDOG_FLAG);
}

//sets or clears FIRMWARE_UPDATED_FLAG in FLAGS_REGISTER
static inline void SetFirmwareUpdatedFlag(bool enabled)
{
  SetBackupFlag(FIRMWARE_UPDATED_FLAG, enabled);
}

//returns if FIRMWARE_UPDATED_FLAG is set in FLAGS_REGISTER
static inline bool FirmwareUpdatedFlagSet()
{
  return ReadBackupFlag(FIRMWARE_UPDATED_FLAG);
}

static inline void WriteBackupRegister(uint32_t reg, uint32_t data)
{
  RTC_WriteBackupRegister(reg, data);
}

static inline uint32_t ReadBackupRegister(uint32_t reg)
{
  return RTC_ReadBackupRegister(reg);
}

static inline void SetRandomSeed(uint32_t randomseed)
{
  RTC_WriteBackupRegister(RANDOM_SEED_REGISTER, randomseed);
}

static inline uint32_t GetRandomSeed()
{
  return RTC_ReadBackupRegister(RANDOM_SEED_REGISTER);
}

static inline void SetSPIFlashLogBackupWriteAddr(uint32_t writeaddr)
{
  RTC_WriteBackupRegister(SPI_FLASH_LOG_WRITE_POSITION_REGISTER, writeaddr);
}

static inline uint32_t GetSPIFlashLogBackupWriteAddr()
{
  return RTC_ReadBackupRegister(SPI_FLASH_LOG_WRITE_POSITION_REGISTER);
}

static inline void SetSPIFlashLogBackupFreeSpace(uint32_t freespace)
{
  RTC_WriteBackupRegister(SPI_FLASH_LOG_FREE_SPACE_REGISTER, freespace);
}

static inline uint32_t GetSPIFlashLogBackupFreeSpace()
{
  return RTC_ReadBackupRegister(SPI_FLASH_LOG_FREE_SPACE_REGISTER);
}

/*! Returns a new sequence number for each time a communication with the server is done.
    This way, the server can track if any HTTP POST requests did not reach the server.*/
static inline uint32_t GetServerCommSequenceNr()
{
  uint32_t sequencenumber = RTC_ReadBackupRegister(SERVERCOMM_SEQUENCE_NUMBER_REGISTER);
  RTC_WriteBackupRegister(SERVERCOMM_SEQUENCE_NUMBER_REGISTER, sequencenumber + 1);

  return sequencenumber;
}

/*! \return The memory address of the backup SRAM.*/
static inline uint8_t* BackupSRAM()
{
  return (uint8_t*)BKPSRAM_BASE; /* ST=0x40024000 GD=*/
}

/*! \return The size of the backup SRAM, in bytes.*/
static inline uint32_t BackupSRAMSize()
{
  return 4 * 1024;
}

#endif //BACKUPREGISTRS_H
