#ifndef BATTERYCHARGER_H
#define BATTERYCHARGER_H

#include <FreeRTOS.h>
#include <semphr.h>

#include "i2c.h"

enum EChargerPowerSource
{
  /*! Charger draws power from a power supply or solar panel.*/
  ChargerPowerSource_Solar_PowerSupply,

  /*! Charger draws power from a battery pack.*/
  ChargerPowerSource_Battery,

  /*! Charger draws power from a 12 volt lead acid battery,
      which is charged by a solar panel.*/
  ChargerPowerSource_Solar_12vBattery,
};

/*! Time in milliseconds waiting for the battery telemetry to become valid.
    If the charger input voltage is around 3.6 volts, the telemetry valid
    bit might never be set, but the readings will still be correct.*/
#define TELEMETRY_TIMEOUT 500

struct SBatteryInfo
{
  SBatteryInfo() : m_intvoltage(0.0f), m_extvoltage(0.0f),
                   m_charging(false), m_chargecurrent(0.0f)
  {
  }

  float m_intvoltage;    //!< Internal battery voltage.
  float m_extvoltage;    //!< External applied voltage.
  bool  m_charging;      //!< True when the internal battery is being charged.
  float m_chargecurrent; //!< Internal battery charge current.
};

class CBatteryCharger
{
  public:
    CBatteryCharger();
    ~CBatteryCharger();

    bool              Setup(CI2C* i2c, uint8_t addr);
    bool              ReadBatteryVoltage(float* voltage, bool print = false, float* input = nullptr,
                                         bool* charging = nullptr, float* batterycurrent = nullptr,
                                         uint32_t timeout = TELEMETRY_TIMEOUT);
    bool              CheckBatteryVoltage(float minimum, uint32_t loglevel, const char* task);
    bool              ReadChargerInfo(SBatteryInfo* batteryinfo, bool print = false, uint32_t timeout = TELEMETRY_TIMEOUT);
    bool              SetForceTelemetry(bool enabled, bool fast = false, bool reset = false);
    bool              CheckChemistry();

    bool              SetChargerPowerSource(EChargerPowerSource powersource);
    bool              TriggerRecharge();

  private:
    bool              ReadRegister(uint8_t reg, uint16_t* value);
    bool              WriteRegister(uint8_t reg, uint16_t value);

    bool              SetCharger(bool enabled);
    bool              TelemetryValid();
    bool              SetMPPT(bool enabled);
    bool              CheckMaxChargeTimeFault(float mininput);

    SemaphoreHandle_t m_mutex;
    CI2C*             m_i2c;
    uint8_t           m_addr;
};

#endif //BATTERYCHARGER_H
