#ifndef CLOCK_H
#define CLOCK_H

#include "inclstdint.h"
#include <stm32f4xx.h>

#ifdef __cplusplus
  extern "C" {
  #include <ctime>
#else
  #include <time.h>
#endif

//These functions must remain C-compatible.
int64_t  Millis();
void     DelayMS(uint32_t milliseconds);
void     DelayUS(uint32_t microseconds);

/*! Returns the counter value of timer2, which is set up as a microsecond
    counter in ClockInit().
    This function is interrupt safe, and should always be kept interrupt safe.*/
static inline uint32_t Micros()
{
  return TIM2->CNT;
}

#ifdef __cplusplus
//These functions only have to be C++ compatible.

/*! If a callback is needed when the RTC is adjusted, derive a class from this
    and install a pointer using SetRTCAdjustCallback(). The RTCAdjustCallback
    method will be called after the RTC is adjusted by SetRTC(). The argument
    timediff is the difference in seconds between the old and new values of the
    RTC. A positive value means the RTC has been set forward, a negative value
    means the RTC has been set backwards.*/
class CRTCAdjustCallback
{
  public:
    enum RTCAdjustCallbackType
    {
      PreAdjustCallback,
      PostAdjustCallback,
    };
    virtual void RTCAdjustCallback(int64_t timediff,
                                   RTCAdjustCallbackType callbacktype) = 0;
};

uint32_t TimerClockFrequency();

void     ClockInit();
void     RTCInit(bool force = false);
void     PrintRTCTime();
bool     TestRTC();

void     SetRTC(struct tm* timeval);
int64_t  GetRTCSeconds(bool* valid);
void     GetRTC(struct tm* timeval);
void     SetRTCSeconds(int64_t unixtime);
void     SetRTCAdjustCallback(CRTCAdjustCallback* callback);

/*! Maximum value for the RTC wakeup counter, this is about 36.4 hours.*/
#define RTC_WAKEUP_MAX 0x1FFFF
void     SetRTCWakeup(uint32_t count);
void     DisableRTCWakeup();
bool     RTCWakeupFlagSet();
void     ClearRTCWakeupFlag();

void     SetRTCAlarm(uint32_t rtcalarm, struct tm* timeval);
void     DisableRTCAlarm(uint32_t rtcalarm);
bool     RTCAlarmFlagSet(uint32_t rtcalarm);
void     ClearRTCAlarmFlag(uint32_t rtcalarm);

uint32_t RTCAlarmFlags();

void     StoreRTCAlarmBootFlags();
uint32_t GetRTCAlarmBootFlags();
const char* RTCAlarmBootFlagStr(uint32_t* flags);

#define DATESTR_MAX_LENGTH 64
void     GetRTCISO8601(char* buf, uint32_t bufsize, const char* timezone = "Z",
                       int64_t offset = 0);
bool     SetRTCISO8601(char* buf);

#endif //__cplusplus

#ifdef __cplusplus
}
#endif

#endif //CLOCK_H
