#ifndef FLASHRINGBUFFER_H
#define FLASHRINGBUFFER_H

#include "inclstdint.h"
#include "spiflash.h"
#include "crc32hw.h"
#include "sha256.h"
#include "aes.h"
#include <cstring>

#define RINGBUFFER_ELEMENT_STARTCODE 0x55AACC33
#define RINGBUFFER_ELEMENT_NULLCODE  0x0
#define RINGBUFFER_ERASED_VALUE      0xFFFFFFFF

enum ElementState
{
  ElementState_Null,
  ElementState_Invalid,
  ElementState_Free,
  ElementState_Used,
};

const char* ElementStateStr(ElementState state);

struct SRingbufferElementHeader
{
  SRingbufferElementHeader() { startcode = 0; size = 0; type = 0; crc = 0;     }
  SRingbufferElementHeader(uint8_t* data) { memcpy(this, data, sizeof(*this)); }

  bool        CRCValid()     { return crc == CalculateCRC();              }
  uint32_t    CalculateCRC() { return CCRC32Hw::Calculate(&startcode, 3); }
  void        AssignCRC()    { crc = CalculateCRC();                      }

  /*! Checks if the element is used, empty, or invalid.
      \return ElementState_Used if the start code, crc and size fields are valid,
      \return ElementState_Free if all bits of the header are 1.
      \return ElementState_Invalid if the header contains invalid data.*/
  ElementState State(uint32_t elementsize)
  {
    if (startcode == RINGBUFFER_ELEMENT_STARTCODE && CRCValid() && size + sizeof(SRingbufferElementHeader) <= elementsize)
      return ElementState_Used;
    else if (startcode == RINGBUFFER_ERASED_VALUE && type == RINGBUFFER_ERASED_VALUE &&
             size == RINGBUFFER_ERASED_VALUE && crc == RINGBUFFER_ERASED_VALUE)
      return ElementState_Free;
    else
      return ElementState_Invalid;
  }

  void Init(uint32_t asize, uint32_t atype = 0)
  {
    startcode = RINGBUFFER_ELEMENT_STARTCODE;
    size      = asize;
    type      = atype;
    AssignCRC();
  }

  uint32_t startcode;
  uint32_t size;
  uint32_t type;
  uint32_t crc;
} __attribute__((packed));

//everything in this struct is a multiple of 16 byte, because the S25FL512S
//stores an error correction code per 16 bytes, however this mandates that
//each block of 16 bytes is written only once
struct SRingbufferElement
{
  SRingbufferElement() : hmac {}, iv {} {/*empty function*/}
  SRingbufferElementHeader header;
  uint8_t                  hmac[SHA256_BLOCK_SIZE];
  uint8_t                  iv[AES_BLOCK_SIZE];
} __attribute__((packed));

/*! Per ringbuffer element, some storage is needed for the header.
    This macro converts the element size to the maximum usable payload size.*/
#define MAX_RINGBUFFER_PAYLOAD_SIZE(elementsize) (elementsize - sizeof(SRingbufferElement))

class CFlashRingbuffer
{
  public:
    CFlashRingbuffer();
    ~CFlashRingbuffer();

    bool       Enabled() { return m_enabled; }
    void       Setup(CSPIFlash* spiflash, uint32_t position, uint32_t size, uint32_t elementsize);
    bool       Write(uint8_t* data, uint32_t size);
    bool       Read(uint8_t* data, uint32_t* size, bool advance = true);
    bool       AdvanceReadPosition();

    uint32_t   NumStoredElements();

    void       ResetReadCounter();
    void       ClearReadElements();

  private:
    bool       CalculateBufferPositions();
    uint32_t   FreeSpace();
    uint32_t   ElementAddr(uint32_t elementindex);
    bool       ReadElementHeader(uint32_t elementindex, SRingbufferElementHeader* header);
    bool       NextReadElement();
    bool       WriteSizeFits(uint32_t size);
    bool       CheckWriteable();
    bool       WriteElement(uint8_t* data, uint32_t size);
    bool       WriteIV(uint8_t* iv);
    bool       WriteHMAC(uint8_t* hmac);
    bool       WriteDataEncrypted(SRingbufferElementHeader* header, uint8_t* data, uint32_t size);
    bool       VerifyDecryptData(SRingbufferElementHeader* header, uint8_t* data, uint32_t* size);
    bool       ClearElement(uint32_t elementpos);

    bool       m_enabled;

    CSPIFlash* m_spiflash;
    uint32_t   m_position;    //!< Position of the ringbuffer inside the flash.
    uint32_t   m_elementsize; //!< Size in bytes of each element.

    uint32_t   m_numelements; //!< Number of elements the ringbuffer can hold.
    uint32_t   m_writepos;    //!< Write position, should always point to a free element.
    uint32_t   m_readpos;     //!< Read position, when the same as m_writepos, the ringbuffer is empty.
    uint32_t   m_readoffset;  //!< m_readpos + m_readoffset is the currently read element.
};

#endif //FLASHRINGBUFFER_H
