#ifndef FWUPDATE_H
#define FWUPDATE_H

#include "inclstdint.h"
#include "http.h"
#include "spiflash.h"
#include "fwimageparams.h"
#include "cbor.h"
#include "spiflashaddrs.h"
#include "internalflashaddrs.h"

/*! Place of the internal flash where the main firmware lives.*/
#define MAIN_FIRMWARE_START_SECTOR    5
#define MAIN_FIRMWARE_NUM_SECTORS     5

/*! Place of the internal flash where the update firmware lives.*/
#define UPDATER_FIRMWARE_START_SECTOR 10
#define UPDATER_FIRMWARE_NUM_SECTORS  2

/*! Sector where the flag is stored that signals that the main firmware has been erased.*/
#define MAIN_FIRMWARE_ERASED_SECTOR   1

enum EFWFileType
{
  EFWFileType_None,
  EFWFileType_Main_Firmware,
  EFWFileType_Updater_Firmware,
};

#define NUM_FILE_MAP_ELEMENTS 4

struct SFileInfo
{
  SFileInfo() : pos(0), size(0), hash {}, name {}, present(false)
  {
  }

  uint32_t pos;
  uint32_t size;
  uint8_t  hash[FW_HASH_SIZE];
  char     name[FW_FILENAME_MAX_LENGTH];
  bool     present;
};

/*! Receives a firmware update file via the CHTTPDownload interface,
    stores it in the SPI flash, and executes the firmware update.*/
class CFwUpdate : public CHTTPDownload
{
  public:
    CFwUpdate();

    void             Init(CSPIFlash* spiflash);
    bool             Prepare();

    virtual uint32_t DownloadBlockSize();
    virtual void     StartDownload(const char* url, uint32_t httprspcode, uint32_t contentlength);
    virtual void     Download(const uint8_t* buf, uint32_t size);

    bool             CheckUpdateImage();
    bool             InstallUpdateImage();

  private:
    void             Reset();

    bool             CheckFlashState();
    bool             CheckSignature();
    bool             CheckBlocks();
    bool             CheckHeader();

    bool             ReadVerifyDecryptBlock(uint32_t blocknum, SEncryptedBlock* encryptedblock, SEncryptedBlock* decryptedblock);
    bool             ReadBlock(uint32_t blocknum, SEncryptedBlock* block);
    bool             VerifyBlockHMAC(uint32_t blocknum, SEncryptedBlock* block);
    bool             CompareHMAC(uint8_t hmac1[FW_HASH_SIZE], uint8_t hmac2[FW_HASH_SIZE]);
    void             DecryptBlock(SEncryptedBlock* encryptedblock, SEncryptedBlock* decryptedblock);
    bool             VerifyBlockNum(SEncryptedBlock* decryptedblock, uint32_t num);
    bool             CheckFiles(CborValue* files);
    bool             CheckFile(CborValue* file);
    bool             CheckFileMapName(CborValue* value, EFWFileType* filetype, char filename[FW_FILENAME_MAX_LENGTH]);
    bool             CheckFileMapPosition(CborValue* value, uint32_t* filepos);
    bool             CheckFileMapSize(CborValue* value, uint32_t* filesize);
    bool             CheckFileMapHash(CborValue* value, uint8_t filehash[FW_HASH_SIZE]);
    bool             CheckFileElementsPresent(bool present[NUM_FILE_MAP_ELEMENTS]);
    bool             CheckFilePositionSize(SFileInfo* fileinfo, EFWFileType filetype);
    bool             VerifyFileHash(SFileInfo* fileinfo);
    bool             WriteFileToInternalFlash(SFileInfo* fileinfo, uint32_t startsector, uint32_t numsectors);
    bool             SetMainFirmwareErasedFlag(bool erased);

    CSPIFlash*       m_spiflash;
    uint32_t         m_fwupdatereserved;
    uint32_t         m_bytesdownloaded;

    SEncryptedBlock  m_encryptedblock;
    SEncryptedBlock  m_decryptedblock;

    uint8_t          m_signatureblockhash[FW_HASH_SIZE];
    SSignature       m_signature;

    uint8_t          m_header[FW_HEADER_SIZE];
    CborParser       m_parser;

    SFileInfo        m_updaterfile;
    SFileInfo        m_mainfile;
};

#endif //FWUPDATE_H
