#ifndef HTTPDOWNLOAD_H
#define HTTPDOWNLOAD_H

#include "inclstdint.h"

class CHTTPDownload
{
  public:
    CHTTPDownload() { m_httprspcode = 0; }

    /*! Callback to inform the object that does the download what size should be used when calling
        the Download() callback, the buffer passed to Download() should not exceed this size.*/
    virtual uint32_t DownloadBlockSize() = 0;

    /*! Callback to inform the object that receives the download that the download has started.
        \param url The url of the file that is downloaded.
        \param httprspcode The http response code from the server.
        \param contentlength The content-length field from the server.*/
    virtual void     StartDownload(const char* url, uint32_t httprspcode, uint32_t contentlength) = 0;

    /*! Callback to pass downloaded data to the object that receives the download.
        \param buf Pointer to the buffer of downloaded bytes.
        \param size The number of bytes in buf.*/
    virtual void     Download(const uint8_t* buf, uint32_t size) = 0;

    /*! Called by the object that does the download to inform that the download started.*/
    void             DownloadBegin(const char* url, uint32_t httprspcode, uint32_t contentlength)
    {
      m_httprspcode = httprspcode;
      StartDownload(url, httprspcode, contentlength);
    }

    /*! \return The http response code from the server.*/
    uint32_t         HTTPRspCode()     { return m_httprspcode; }

    /*! \return If the http response code is in the 200 range, meaning the data was received ok.*/
    bool             HTTPRspCodeGood() { return m_httprspcode >= 200 && m_httprspcode < 300; }

  protected:
    uint32_t         m_httprspcode; //!< Response code of the HTTP POST.
};

class CHTTPPost : public CHTTPDownload
{
  public:
    virtual uint32_t UploadSize() = 0;
    virtual void     StartUpload(const char* url) = 0;
    virtual uint32_t Upload(uint8_t* buf, uint32_t bufsize) = 0;
};

/*! Class for using a buffer to hold POST upload and download data.*/
class CHTTPTransceiver : public CHTTPPost
{
  public:
    CHTTPTransceiver(uint8_t* txbuf, uint32_t txbufsize, uint8_t* rxbuf, uint32_t rxbufsize);

    virtual uint32_t UploadSize();
    virtual void     StartUpload(const char* url);
    virtual uint32_t Upload(uint8_t* buf, uint32_t bufsize);

    virtual uint32_t DownloadBlockSize();
    virtual void     StartDownload(const char* url, uint32_t httprspcode, uint32_t contentlength);
    virtual void     Download(const uint8_t* buf, uint32_t size);

    void             SetTxBuf(uint8_t* txbuf, uint32_t txbufsize);

    uint8_t*         TxBuf()           { return m_txbuf;       }
    uint32_t         TxBufSize()       { return m_txbufsize;   }
    uint8_t*         RxBuf()           { return m_rxbuf;       }
    uint32_t         RxBufSize()       { return m_rxbufsize;   }
    uint32_t         Uploaded()        { return m_uploaded;    }
    uint32_t         Downloaded()      { return m_downloaded;  }
    uint32_t         Overflow()        { return m_overflow;    }

  private:
    uint8_t*         m_txbuf;       //!< Pointer to the buffer to upload, may be the same as m_rxbuf.
    uint32_t         m_txbufsize;   //!< Number of bytes to upload.
    uint8_t*         m_rxbuf;       //!< Pointer to the buffer to download data into, may be the same as m_txbuf.
    uint32_t         m_rxbufsize;   //!< Size of m_rxbuf.

    uint32_t         m_uploaded;    //!< Number of bytes that have been uploaded.
    uint32_t         m_downloaded;  //!< Number of bytes that have been downloaded into m_rxbuf.
    uint32_t         m_overflow;    //!< Excess bytes downloaded but not copied into m_rxbuf because it was too small.
};

#endif //HTTPDOWNLOAD_H
