#ifndef KEYVALUESTORAGE_H
#define KEYVALUESTORAGE_H

#include "inclstdint.h"
#include "cbor.h"
#include "sha256.h"

#include <FreeRTOS.h>
#include <semphr.h>

struct SKeyValueInfo
{
  uint32_t    key;         //!< Key of the key-value pair.
  const char* name;        //!< Name of the key-value.
  const char* serverfield; //!< Name of the key-value pair when communicating with the server.
  CborType    type;        //!< Type of the key-value pair, CborIntegerType, CborTextStringType, CborBooleanType or CborFloatType.

  /*! Default value of the key-value pair, which field to use depends on the type.*/
  union
  {
    int64_t     i; //!< Default for CborIntegerType.
    const char* s; //!< Default for CborTextStringType.
    bool        b; //!< Default for CborBooleanType.
    float       f; //!< Default for CborFloatType.
  }
  defval;

  const char* ServerField() const
  {
    if (serverfield)
      return serverfield;
    else
      return "null";
  }
};

/*! The first 8 bytes of a SHA256 hash are used, this saves space
    compared to using a full 32 byte SHA256 hash.*/
#define KEY_VALUE_HASH_SIZE 8

struct SKeyValueHash
{
  uint32_t size;                      //!< Number of bytes that the hash is made from.
  uint8_t  hash[KEY_VALUE_HASH_SIZE]; //!< Hash of stored key-value pairs.

  /*! \return true if all bytes are set to the 0xFF erased state.
      \return false if any byte in this struct is not 0xFF.*/
  bool IsErased()
  {
    uint8_t* hashptr = (uint8_t*)this;
    uint8_t* endptr  = hashptr + sizeof(*this);
    bool     erased  = true;

    while (hashptr != endptr)
    {
      if (*(hashptr++) != 0xFF)
      {
        erased = false;
        break;
      }
    }

    return erased;
  }
}__attribute__((packed));

//! Maximum length of a string value, including the null terminator.
#define STRING_VALUE_MAX_LENGTH 256

//! Size of a CBOR encode buffer, for encoding key-value pairs.
#define KEY_VALUE_ENCODE_BUF_SIZE (32 + STRING_VALUE_MAX_LENGTH)

class CKeyValueStorage
{
  public:
    CKeyValueStorage();
    ~CKeyValueStorage();

    void                 Init(const SKeyValueInfo* keyvalueinfo, uint32_t numkeyvalues,
                              uint32_t sector1, uint32_t sector2, int32_t loglevel);

    bool                 GetBoolean(uint32_t key);
    int64_t              GetInteger(uint32_t key);
    void                 GetString(uint32_t key, char* buf, uint32_t bufsize);
    float                GetFloat(uint32_t key);

    bool                 SetBoolean(uint32_t key, bool value);
    bool                 SetInteger(uint32_t key, int64_t value);
    bool                 SetString(uint32_t key, const char* value);
    bool                 SetFloat(uint32_t key, float value);

    void                 ToServerCBOR(CborEncoder* root);
    void                 FromServerCBOR(CborValue* root);

    void                 PrintKeyValuePairs();

    const char*          Name(uint32_t key);

  private:
    void                 ResetState();
    uint32_t             KeyIndex(uint32_t key);
    const SKeyValueInfo& KeyInfo(uint32_t key);
    bool                 CheckStoredKeyValuePairs();
    bool                 ChooseSector();
    bool                 CheckStoredHash();
    bool                 InitDecoder();
    bool                 CheckWritePosition();
    bool                 GetRecoverSectors(int32_t sectors[2]);
    bool                 CopySectorData(int32_t source, uint32_t size);
    bool                 RecoverStoredKeyValuePairs();
    bool                 InitStoredKeyValuePairs();
    bool                 InitSector(int32_t sector, uint32_t sequencenr);
    bool                 StoreKeyValuePair(const SKeyValueInfo& keyinfo, CborValue* it, const char* source);
    bool                 CanWrite(uint32_t size);
    bool                 WriteCBOR(uint8_t* buf, uint32_t size);
    bool                 WriteHash();
    bool                 SwitchSector();
    bool                 CopySectorKeyValuePairs();
    uint8_t*             EncodeBuf();
    void                 CreateHash(uint8_t* address, uint32_t size, SKeyValueHash* hash);

    const SKeyValueInfo* m_keyvalueinfo;
    uint32_t             m_numkeyvalues;
    int32_t              m_sector1;
    int32_t              m_sector2;
    int32_t              m_loglevel;

    SemaphoreHandle_t    m_mutex;

    int32_t              m_sector;
    uint8_t*             m_addr;
    uint32_t             m_size;
    uint32_t             m_writepos;
    SKeyValueHash*       m_hashptr;
    uint8_t              m_encodebuf[2][KEY_VALUE_ENCODE_BUF_SIZE];
    char                 m_stringbuf[STRING_VALUE_MAX_LENGTH];
    char                 m_oldstrvalue[STRING_VALUE_MAX_LENGTH];
    bool                 m_writehash;
    bool                 m_inswitch;

    CborParser           m_parser;
    CborValue            m_root;
};

#endif //KEYVALUESTORAGE_H
