#ifndef LINKEDLIST_H
#define LINKEDLIST_H

#include "inclstdint.h"

/*! Implements a doubly linked list. Classes can be derived
    from CListElement to store custom objects.
    This way, only one allocation is needed per element, this is
    more efficient than using std::list with pointers.

usage: Derive a class from CListElement, create an instance of CList.
       Or add a CListElement as the first member of a struct, in case
       a struct with predictable memory layout is needed.
       Allocate elements on the heap, add them to the list using AddToListAtStart() or
       AddToListAtEnd(), remove them from the list using RemoveFromList().
       The list can be iterated over forwards and backwards using the value
       returned by the Start() and End() methods of CList.*/

class CListElement;

class CList
{
  friend class CListElement;

  public:
    CList()
    {
      m_start = nullptr;
      m_end   = nullptr;
      m_size  = 0;
    }

    CListElement* Start() { return m_start;     }
    CListElement* End()   { return m_end;       }
    uint32_t      Size()  { return m_size;      }
    bool          Empty() { return m_size == 0; }

  protected:
    CListElement* m_start; //!< Pointer to the first element, nullptr on an empty list.
    CListElement* m_end;   //!< Pointer to the last element, nullptr on an empty list.
    uint32_t      m_size;  //!< Number of elements in the list.
};

/*! Template class to do pointer casting from CListElement to the type derived
    from CListElement. This saves a lot of pointer casts in the code.*/
template <class ElementType> class CLinkedList : public CList
{
  public:
    ElementType* Start() { return (ElementType*)m_start; }
    ElementType* End()   { return (ElementType*)m_end;   }
};

class CListElement
{
  public:
    CListElement();
    ~CListElement();

    void          AddToListAtStart(CList& list);
    void          AddToListAtEnd(CList& list);
    void          AddToListBeforeElement(CList& list, CListElement* element);
    void          AddToListAfterElement(CList& list, CListElement* element);
    void          RemoveFromList(CList& list);

    CListElement* Prev() { return m_prev; }
    CListElement* Next() { return m_next; }

  protected:
    CListElement* m_prev; //!< Pointer to the previous element, nullptr when this is the first element.
    CListElement* m_next; //!< Pointer to the next element, nullptr when this is the last element.
}__attribute__((packed));

#endif //LINKEDLIST_H
