#ifndef LOCK_H
#define LOCK_H

#include <FreeRTOS.h>
#include <semphr.h>
#include <task.h>
#include <cassert>

/*! This class locks a SemaphoreHandle_t recursive mutex in its constructor,
    and will unlock it in its destructor.
    This creates a scoped lock, where the lock is always released when the
    instance of this class goes out of scope. */
class CLock
{
  public:

    /*! Constructor, locks the mutex.
      \param mutex Should be a handle to a recursive mutex, a reference to this
                   handle is stored in m_mutex.
      \param enter Set to true to lock the mutex from this constructor.*/
    CLock(SemaphoreHandle_t& mutex, bool enter = true) : m_mutex(mutex)
    {
      m_haslock = false;

      if (enter)
        Enter();
    }

    /*! Destructor, always executes Leave()*/
    ~CLock()
    {
      Leave();
    }

    /*! Enters the lock by calling xSemaphoreTakeRecursive(), but only if this
        object doesn't have the lock.*/
    void Enter()
    {
      if (!m_haslock)
      {
        assert(m_mutex);
        xSemaphoreTakeRecursive(m_mutex, portMAX_DELAY);
        m_haslock = true;
      }
    }

    /*! Enters the lock by calling xSemaphoreTakeRecursive() with a timeout of 0,
        but only if this object doesn't have the lock.
     \return true if this object has the lock, false if it doesn't.*/
    bool TryEnter()
    {
      if (!m_haslock)
      {
        assert(m_mutex);
        if (xSemaphoreTakeRecursive(m_mutex, 0))
        {
          m_haslock = true;
          return true;
        }
        else
        {
          return false;
        }
      }
      else
      {
        return true;
      }
    }

    /*! Leaves the lock by calling xSemaphoreGiveRecursive(),
        but only if it previously got the lock with xSemaphoreGiveRecursive().*/
    void Leave()
    {
      if (m_haslock)
      {
        assert(m_mutex);
        xSemaphoreGiveRecursive(m_mutex);
        m_haslock = false;
      }
    }

    /*! \return true if this object has the lock, false if it doesn't.*/
    bool HasLock()
    {
      return m_haslock;
    }

  private:
    SemaphoreHandle_t& m_mutex;   /*!< Reference to the mutex that should be
                                       locked and unlocked.*/
    bool               m_haslock; //!< Set to true when m_mutex is locked.
};

/*! This class enters a critital section in its constructor, and leaves the
    critial section in its destructor.
    This creates a scoped critial section.*/
class CCritical
{
  public:
    /*! Constructor, enters the critical section.*/
    CCritical()
    {
      m_interrupt              = xPortIsInsideInterrupt();
      m_uxSavedInterruptStatus = 0;
      m_hascritical            = false;
      Enter();
    }

    /*! Destructor, leaves the critical section, but only if the critical
        section was not left before using Leave().*/
    ~CCritical()
    {
      Leave();
    }

    /*! Enters the critical section using taskENTER_CRITICAL(), but only if the
        critical section was not entered before.*/
    void Enter()
    {
      if (!m_hascritical)
      {
        if (m_interrupt)
          m_uxSavedInterruptStatus = taskENTER_CRITICAL_FROM_ISR();
        else
          taskENTER_CRITICAL();

        m_hascritical = true;
      }
    }

    /*! Leaves the critical section using taskEXIT_CRITICAL(), but only if the
        critical section was not left before.*/
    void Leave()
    {
      if (m_hascritical)
      {
        if (m_interrupt)
          taskEXIT_CRITICAL_FROM_ISR(m_uxSavedInterruptStatus);
        else
          taskEXIT_CRITICAL();

        m_hascritical = false;
      }
    }

  private:
    /*! Indicates if the critical section is held.*/
    bool        m_hascritical;

    /*! Interrupt status in case the critical section is entered
        from an interrupt.*/
    bool        m_interrupt;
    UBaseType_t m_uxSavedInterruptStatus;
};

#endif //LOCK_H
