#ifndef LOG_H
#define LOG_H

#include "loglevels.h"
#include "inclstdint.h"

#ifdef __cplusplus
extern "C" {
#endif

#define DEBUGUSART USART1

void SetupDebugUsart();
void SetDebugUsartAsync(bool async);

int  PrintLog(int level, const char* fmt, ...) __attribute__ ((format (printf, 2, 3)));
int  PrintLogDecode(int level, const char* fmt, ...) __attribute__ ((format (printf, 2, 3)));
void PrintLogHexBuf(int level, const uint8_t* buf, uint32_t size, const char* suffix);
void DEBUGUSARTprintf(const char* fmt, ...) __attribute__ ((format (printf, 1, 2)));
void WaitLogWriteComplete(uint32_t timeout);

void LockLog();
void UnlockLog();

/*! The stm32wss library has negative log levels, the application has positive log levels that start at 0.
    This way, log levels can be added to stm32wss and each application that depends on it,
    without ever having a log level collision.*/
#define LOGSTDOUT     -1
#define LOGMAIN       -2
#define LOGINFO       -3
#define LOGERROR      -4
#define LOGCLOCK      -5
#define LOGTASK       -6
#define LOGATCMD      -7
#define LOGAPNLIST    -8
#define LOGSPIFLASH   -9
#define LOGFWUPDATE   -10
#define LOGINTFLASH   -11
#define LOGFLASHLOG   -12
#define LOGFLASHBUF   -13
#define LOGMODEM      -14
#define LOGRFM9xW     -15
#define LOGEXTRTC     -16
#define LOGOLEDDPY    -17
#define LOGIOEXPANDER -18
#define LOGBTTRYCHRGR -19
#define LOGFILLSENSOR -20
#define LOGRTCSCHED   -21

//! Each log level string will be padded to this length with spaces.
#define LOGLEVELPADLEN 10

void SetApplicationLogLevels(const char* const* levels, int numlevels);

#ifdef __cplusplus
}
#endif

#ifdef __cplusplus
 /*! Locks the log in the constructor, and unlocks the log in the destructor,
     useful for printing a single line with multiple calls to PrintLog(),
     within a single scope.*/
 class CLogLock
 {
   public:
     CLogLock()
     {
       m_haslock = false;
       Enter();
     }

     ~CLogLock()
     {
       Leave();
     }

     void Enter()
     {
       if (!m_haslock)
       {
         LockLog();
         m_haslock = true;
       }
     }

     void Leave()
     {
       if (m_haslock)
       {
         UnlockLog();
         m_haslock = false;
       }
     }

   private:
     bool m_haslock;
 };
#endif

#ifdef __cplusplus
/*! Overload for PrintLogHexBuf that doesn't need the suffix argument.*/
static inline void PrintLogHexBuf(int level, const uint8_t* buf, uint32_t size)
{
  PrintLogHexBuf(level, buf, size, nullptr);
}
#endif

#endif //LOG_H
