#ifndef MODEMBG96_H
#define MODEMBG96_H

/*
Connections between MCU and BG96 modem on the Accello pcb:

Pin nr  MCU function    Modem function  Description

PG15    DCD             DTR             Data terminal ready.
PG14    TX              RX              Data transmission from modem to mcu.
PG13    CTS             CTS             Flow control modem to mcu, the modem has CTS and RTS reversed.
PG12    RTS             RTS             Flow control from mcu to modem, the modem has CTS and RTS reversed.
PG11    DTR             DCD             Data carrier detect.
PG10    RI              RI              Ring indicator.
PG9     RX              TX              Data transmission from modem to mcu.
PG8     SIM select                      Select between soldered chip SIM and socketed SIM.
PG7     Reset                           Reset of the modem via a transistor.
PG6     Status                          Modem status pin, via transistor, uses resistor and led as pullup,
                                        the internal pullup of the GPIO should be enabled.
PG5     On/Off                          Connected to PWRKEY pin via a transistor.
PG4     Com disable                     Turns off the level shifters between the modem and the mcu.
PF5     PSM indicator                   When low, indicates that power saving mode is active.
*/

#include "usart.h"
#include "exti.h"
#include "atcmd.h"
#include "http.h"
#include "gpiopin.h"

#include <FreeRTOS.h>
#include <semphr.h>

enum EModemType
{
  ModemType_UNKNOWN,
  ModemType_BG96,
  ModemType_BG95,
};

#define SIMINFO_SIZE    32
#define FNN_SIZE        64
#define SNN_SIZE        64
#define SPN_SIZE        64
#define RPLMN_SIZE      64
#define IMEI_SIZE       64
#define MANUFACT_SIZE   32
#define MODEL_SIZE      32
#define REVISION_SIZE   32
#define APPVER_SIZE     32
#define LAC_SIZE        16
#define CI_SIZE         16

class CModemBG96 : public CGPIOInterrupt
{
  public:
    CModemBG96();
    ~CModemBG96();

    void              SetPeripherals(USART_TypeDef* usartptr, const SGPIOPin pwrkey, const SGPIOPin status, const SGPIOPin reset);
    bool              Setup(bool factoryreset = false, bool allowfwupdate = true);
    bool              Close();
    void              PowerOff();
    bool              Activated()    { return m_activated;     }
    bool              NeedsReset()   { return m_needsreset;    }

    bool              HTTPGet(const char* url, uint32_t timeout, CHTTPDownload* httpdownload, bool logpercentages = false);
    bool              HTTPPost(const char* url, uint32_t timeout, CHTTPPost* httppost, bool logpercentages = false);

    const char*       IMEI()         { return m_imei;          }
    const char*       IMSI()         { return m_siminfo.imsi;  }
    const char*       ICCID()        { return m_siminfo.iccid; }
    const char*       Manufacturer() { return m_manufact;      }
    const char*       Model()        { return m_model;         }
    const char*       Revision()     { return m_revision;      }
    const char*       AppVer()       { return m_appver;        }
    const char*       RPLMN()        { return m_rplmn;         }
    const char*       LACStr()       { return m_lacstr;        }
    uint32_t          LAC()          { return m_lac;           }
    const char*       CIDStr()       { return m_cistr;         }
    uint32_t          CID()          { return m_ci;            }
    int32_t           RSSI()         { return m_rssi;          }
    bool              UpdateRSSI();
    bool              NeedsFwUpdate(){ return m_needsfwupdate; }

    bool              GetLocation(uint32_t locktime, double* latitude, double* longitude);
    bool              GetBytesTransferred(uint64_t* tx, uint64_t* rx);

  private:
    void              ResetState();

    void              PreSetup();
    bool              SetupModem(bool factoryreset, bool allowfwupdate);

    virtual void      GPIOIRQHandler(SGPIOPin pin);
    bool              SetPower(bool on, bool waitfirst = false);

    bool              SetupBaudrate();
    uint32_t          DetectBaudrate(uint32_t baudratefirst);
    bool              CheckBaudrate(uint32_t baudrate);
    bool              SetFlowControl();

    bool              SetErrorMessageFormat();

    bool              FactoryReset(bool factoryreset);
    bool              SetPhoneFunctionality(bool on);
    bool              SetNetworkScanMode();
    bool              SetOperatorSelection();
    bool              GetModemInfo();
    bool              ReadSIMIMSI();
    bool              ReadSIMICCID();
    bool              ReadSIMSPN();
    bool              EnterSIMPIN();
    bool              CheckSIMPIN(bool* ready);
    bool              GetNumPINRetries(uint32_t* pincounter);
    bool              WaitNetworkRegister();
    bool              WaitPacketServiceAttach();
    bool              GetSignalQuality();
    bool              GetNetworkInfo();
    bool              GetNetworkType();
    bool              ActivatePDPContext();
    bool              DeactivatePDPContext();
    bool              GetIPAddress();
    bool              UpdateFirmware(bool allowfwupdate);
    bool              WaitFirmwareUpdateComplete();
    bool              CheckFirmwareUpdateInProgress();

    bool              SetupHTTPRequest(const char* url);
    bool              SetHTTPContentType(uint32_t content_type);

    bool              StartHTTPGET(uint32_t timeout, uint32_t* httprspcode, uint32_t* contentlength);

    bool              StartHTTPPOST(uint32_t size, uint32_t timeout);
    bool              FinishHTTPPOST(uint32_t timeout, uint32_t* httprspcode, uint32_t* contentlength);

    bool              HTTPUpload(const char* url, uint32_t timeout, CHTTPPost* httppost, bool logpercentages, uint32_t size);
    bool              HTTPDownload(const char* url, uint32_t timeout, CHTTPDownload* httpdownload, bool logpercentages,
                                   uint32_t httprspcode, uint32_t contentlength);
    bool              FinishHTTPDownload(uint32_t* httpresult);

    bool              EnableGNSS();
    bool              DisableGNSS();
    
    EModemType        m_type;                    //!< Indicates the type of modem.

    bool              m_activated;               //!< Set to true when the modem is set up successfully.
    bool              m_needsreset;              //!< Set to true when communicating with the modem has failed.

    CATCmd            m_atcmd;

    SGPIOPin          m_pwrkey;                  //!< PWRKEY pin, for turning the modem on and off.
    SGPIOPin          m_status;                  //!< Status pin, indicates if the modem is on or off.
    SGPIOPin          m_reset;                   //!< Reset pin, can be used to fully reset the modem.

    SemaphoreHandle_t m_statussignal;            //!< Set from the GPIO interrupt when the status pin changes.

    USART_InitTypeDef m_usartinit;               //!< Initialization parameters for m_usart.
    USART_TypeDef*    m_usartptr;                //!< Pointer to the usart to use.
    CUsart            m_usart;                   //!< Usart object to communicate with the modem.

    int32_t           m_rssi;                    //!< Received signal strength indication, 0 to 31, or 99 for unknown.
    uint32_t          m_ber;                     //!< Bit error rate, 0 to 7 where 0 is the best quality, or 99 for unknown.

    //! Everything in struct this should always be a null terminated C-string.
    struct
    {
      char            imsi[SIMINFO_SIZE];        //!< IMSI of the SIM.
      char            iccid[SIMINFO_SIZE];       //!< ICCID of the SIM.
      char            spn[SIMINFO_SIZE];         //!< SPN of the SIM
    }
    m_siminfo;

    char              m_imei[IMEI_SIZE];         //!< IMEI of the modem, C-string.
    char              m_manufact[MANUFACT_SIZE]; //!< Manufacturer of the modem, C-string.
    char              m_model[MODEL_SIZE];       //!< Model of the modem, C-string.
    char              m_revision[REVISION_SIZE]; //!< Revision of the modem, C-string.
    char              m_appver[APPVER_SIZE];     //!< Firmware version of the modem, C-string.
    char              m_fnn[FNN_SIZE + 1];       //!< Full network name, C-string.
    char              m_snn[SNN_SIZE + 1];       //!< Short network name, C-string.
    char              m_spn[SPN_SIZE + 1];       //!< Service provider name, C-string.
    char              m_rplmn[RPLMN_SIZE + 1];   //!< Registered PLMN (public land mobile network), C-string.
    char              m_lacstr[LAC_SIZE + 1];    //!< Hexadecimal location area code of the cell tower, C-String.
    uint32_t          m_lac;                     //!< m_lacstr converted to binary.
    char              m_cistr[CI_SIZE + 1];      //!< Hexadecimal cell tower id, C-string.
    uint32_t          m_ci;                      //!< m_cistr converted to binary.

    bool              m_needsfwupdate;           //!< Indicates if the modem needs a firmware update.
};

#endif //MODEMBG96_H
