#ifndef RTCSCHEDULER_H
#define RTCSCHEDULER_H

#include <FreeRTOS.h>
#include <semphr.h>

#include "exti.h"
#include "taskhandler.h"
#include "clock.h"

#include <ctime>


//Flags to set in m_flags of CRTCScheduleEntry.
#define RTCSCHEDULEFLAG_VALID     (1 << 0) //!< Entry has valid information.
#define RTCSCHEDULEFLAG_EXPIRED   (1 << 1) //!< Timestamp of entry has expired.
#define RTCSCHEDULEFLAG_HANDLED   (1 << 2) //!< Entry has been handled by client code.
#define RTCSCHEDULEFLAG_COUNTDOWN (1 << 3) //!< Countdown timer instead of fixed time.

//Each alarm of the RTC scheduler has an index. When it's used in an stm32wss
//based project, it should be documented in this header file.
//Accello alarms:
//0: OpenStart and OpenStop settings.
//1: Checking the door switch input.

#define MAX_RTCSCHEDULE_ENTRIES 32

class CRTCScheduleEntry
{
  public:
    void     Set(int64_t timestamp, bool countdown);
    void     Reset();
    void     SetCRC();
    bool     Valid(uint32_t index = MAX_RTCSCHEDULE_ENTRIES);
    void     AdjustTimestamp(int64_t timediff);

    int64_t  Timestamp()  { return m_timestamp;                         }

    bool     CountDown()  { return m_flags & RTCSCHEDULEFLAG_COUNTDOWN; }

    bool     Expired()    { return m_flags & RTCSCHEDULEFLAG_EXPIRED;   }
    void     SetExpired();

    bool     Handled()    { return m_flags & RTCSCHEDULEFLAG_HANDLED;   }
    void     SetHandled();

  private:
    //Operator overload to feed to the hardware CRC32 generator.
    operator uint32_t*()  { return (uint32_t*)this;                     }

    //Size in 4 byte words to feed to the harware CRC32 generator.
    uint32_t CRCGenSize() { return (sizeof(*this) - sizeof(m_crc)) / 4; }

    uint32_t GenerateCRC();

    int64_t  m_timestamp; //Unix timestamp of when the entry expires.
    uint32_t m_flags;

    uint32_t m_crc;       //CRC of the data, used to check that the entry in the
                          //battery backed SRAM is correct. This field should
                          //always be the last of this object.
};

//The hardware CRC generator can only work with 4 bytes at a time.
static_assert((sizeof(CRTCScheduleEntry) % sizeof(uint32_t)) == 0,
              "CRTCScheduleEntry size should be a multiple of 4");

class CRTCSchedulerCallback
{
  public:
    virtual ~CRTCSchedulerCallback() {}

    virtual void RTCSchedulerCallback(uint32_t index) = 0;
};


class CRTCScheduler : public CRTCInterrupt, public CTaskHandler,
                      public CRTCAdjustCallback
{
  public:
    CRTCScheduler();

    void               Setup();
    void               Start();

    void               SetAlarm(uint32_t index, tm* timeval);
    void               SetAlarm(uint32_t index, int64_t timestamp);
    void               SetCountDown(uint32_t index, int64_t seconds);
    void               DisableAlarm(uint32_t index);

    void               SetCallback(uint32_t index, CRTCSchedulerCallback* callback);

    bool               AlarmExpired(uint32_t index);

  private:
    virtual void       PostSetup()   {}
    virtual void       PreProcess()  {}
    virtual void       Process();
    virtual void       PostProcess() {}
    virtual void       RTCIRQHandler(ERTCIRQType irqtype);
    virtual void       RTCAdjustCallback(int64_t timediff,
                                         RTCAdjustCallbackType callbacktype);
    void               ProcessAlarms();
    void               SetNextAlarm();
    void               AdjustCountDownAlarms(int64_t timediff);

    void               SetAlarm(uint32_t index, tm* timeval, int64_t timestamp,
                                bool countdown);

    CRTCScheduleEntry* m_entries;

    CRTCSchedulerCallback*
                       m_callbacks[MAX_RTCSCHEDULE_ENTRIES];

    SemaphoreHandle_t  m_signal; //Semaphore to wake up the CRTCScheduler task.
    SemaphoreHandle_t  m_mutex;  //Mutex for locking access to member variables.
};

#endif //RTCSCHEDULER_H
