#ifndef TASKHANDLER_H
#define TASKHANDLER_H

#include <FreeRTOS.h>
#include <task.h>

#define DEFAULT_PRIORITY   ((configMAX_PRIORITIES - 1) / 2)
#define DEFAULT_STACKDEPTH configMINIMAL_STACK_SIZE

#define HIGHEST_PRIORITY   (configMAX_PRIORITIES - 1)
#define LOWEST_PRIORITY    (0)

/*! This class creates a FreeRTOS task using the xTaskCreate() function.
    In the created task the virtual Process() function is called.
    Anything using this class should derive from it, and create their own Process() function.*/
class CTaskHandler
{
  public:
    CTaskHandler();
    ~CTaskHandler();

    /*! This class creates a FreeRTOS task using the xTaskCreate() function.
        \param name Name of the task, will be truncated to configMAX_TASK_NAME_LEN including the null terminator.
        \param priority Priority of the task, will be capped to configMAX_PRIORITIES-1, a higher number is a higher priority task.
        \param stackdepth Stack size of the created task.*/
    void         Create(const char* name, uint32_t priority = DEFAULT_PRIORITY, uint32_t stackdepth = DEFAULT_STACKDEPTH);

    /*! Calls vTaskSuspend on m_task to suspend the task.*/
    void         Suspend();
    /*! Calls vTaskResume on m_task to resume the task.*/
    void         Resume();

    /*! Calls xTaskNotifyGive() for a direct to task notification.*/
    void         Notify();
    /*! Calls vTaskNotifyGiveFromISR() for a direct to task notification from an ISR.*/
    void         NotifyFromISR();
    /*! Waits on a task notification using ulTaskNotifyTake(), and decreases the count on exit.*/
    void         WaitNotificationCounting();
    /*! Waits on a task notification using ulTaskNotifyTake(), and clears the count on exit.*/
    void         WaitNotificationBinary();
    /*! Clears the task notification using ulTaskNotifyTake() with a zero delay.*/
    void         ClearNotification();

  protected:
    /*! Called by the Task() function, should be declared by any class deriving from this class.*/
    virtual void Process() = 0;

    /*! Callback function for the xTaskCreate() function, calls Process(), and calls vTaskDelete() when Process() exits.*/
    static  void Task(void *pvParameters);

    /*! Handle to the created task.*/
    TaskHandle_t m_task;
};

/*! Sets the priority of the current task in the constructor,
    and restores the old priority in the destructor.*/
class CScopedTaskPriority
{
  public:
    CScopedTaskPriority(uint32_t priority)
    {
      m_task        = xTaskGetCurrentTaskHandle();
      m_oldpriority = uxTaskPriorityGet(m_task);

      vTaskPrioritySet(m_task, priority);
    }

    ~CScopedTaskPriority()
    {
      vTaskPrioritySet(m_task, m_oldpriority);
    }

  private:
    TaskHandle_t m_task;
    uint32_t     m_oldpriority;
};

#endif //TASKHANDLER_H

