#ifndef USART_H
#define USART_H

#include <stm32f4xx.h>
#include <FreeRTOS.h>
#include <queue.h>
#include <semphr.h>

#include "inclstdint.h"

#define USART_QUEUE_SIZE                 1024 //!< Number of bytes that can be buffered by the usart queues.

#define READDATA_MASK                  0x00FF
#define READRESULT_MASK                0xFF00

#define READRESULT_SUCCESS                0x1
#define READRESULT_ERROR_TIMEOUT          0x2
#define READRESULT_ERROR_TOO_LARGE        0x4

//these are picked to be bits on the most significant byte of a 16 bit variable,
//so that they can be passed through the rx queue
#define READRESULT_ERROR_OVERRUN        0x100
#define READRESULT_ERROR_NOISE          0x200
#define READRESULT_ERROR_FRAMING        0x400
#define READRESULT_ERROR_PARITY         0x800
#define READRESULT_ERROR_RX_QUEUE_FULL 0x1000

#define WRITERESULT_SUCCESS            0x1
#define WRITERESULT_ERROR_TIMEOUT      0x2

class CUsart
{
  public:
    CUsart();
    ~CUsart();

    void               Setup(USART_TypeDef* usart, USART_InitTypeDef* init, uint32_t queuesize = USART_QUEUE_SIZE);
    void               SetOneBitMode(FunctionalState newstate);
    void               Close();

    void               ClearInput();

    uint32_t           Read(uint8_t* buf, uint32_t bufsize, uint32_t* bytesread, uint32_t timeout, int end = -1);
    uint32_t           ReadModemReply(char* buf, uint32_t bufsize, uint32_t* bytesread, uint32_t timeout,
                                      const char* start = nullptr, const char* end = nullptr);

    uint32_t           Write(const uint8_t* buf, uint32_t bufsize, uint32_t timeout = 0);
    uint32_t           WriteStr(const char* str, uint32_t timeout = 0);
    bool               WaitWriteComplete(uint32_t timeout = 0);

    bool               IsEnabled() { return m_usart != nullptr; }

    static const char* ResultToStr(uint32_t result);

    void               IRQHandler();

  private:
    IRQn               UsartIrqn();
    void               SetIrq(FunctionalState newstate);
    void               SetClock(FunctionalState newstate);
    BaseType_t         RxQueueReceive(void *pvBuffer, TickType_t xTicksToWait);

    uint32_t           WaitForModemReplyStart(const char* start, uint32_t timeout);
    bool               EndsWithModemResult(const char* buf, uint32_t bytesread, const char* end);

    USART_TypeDef*     m_usart;

    uint32_t           m_queuesize;
    QueueHandle_t      m_rxqueue; //!< Queue where the usart interrupt writes bytes received by the usart.
    QueueHandle_t      m_txqueue; //!< Queue where the task writes bytes to be transmitted over the usart.

    SemaphoreHandle_t  m_txcompletesignal;

    bool               m_rtsflowcontrol;
};

void InstallUsartIRQHandler(USART_TypeDef* usart, CUsart* irqhandler);
void RemoveUsartIRQHandler(USART_TypeDef* usart);

#endif //USART_H
