#include "adc.h"
#include "lock.h"

#include <cassert>

static CAdcInterrupt* g_adcirqhandlers[3];

/* Checks the end of conversion (EOC) interrupt bit on each ADC,
   and calls the corresponding interrupt handler.*/
extern "C" void ADC_IRQHandler()
{
  if (ADC_GetITStatus(ADC1, ADC_IT_EOC))
  {
    if (g_adcirqhandlers[0])
      g_adcirqhandlers[0]->AdcIRQHandler();
    else
      ADC_ClearITPendingBit(ADC1, ADC_IT_EOC);
  }

  if (ADC_GetITStatus(ADC2, ADC_IT_EOC))
  {
    if (g_adcirqhandlers[1])
      g_adcirqhandlers[1]->AdcIRQHandler();
    else
      ADC_ClearITPendingBit(ADC2, ADC_IT_EOC);
  }

  if (ADC_GetITStatus(ADC3, ADC_IT_EOC))
  {
    if (g_adcirqhandlers[2])
      g_adcirqhandlers[2]->AdcIRQHandler();
    else
      ADC_ClearITPendingBit(ADC3, ADC_IT_EOC);
  }
}

CAdc::CAdc()
{
  m_mutex      = 0;
  m_adc        = nullptr;
  m_value      = 0;
  m_numsamples = 0;
  m_maxsamples = 0;
  m_done       = 0;
}

CAdc::~CAdc()
{
  vSemaphoreDelete(m_mutex);
  vSemaphoreDelete(m_done);
}

/*! Allocates mutexes and sets member variables.
    \param adc The ADC to take samples from, ADC1, ADC2 or ADC3.*/
void CAdc::Setup(ADC_TypeDef* adc)
{
  m_adc = adc;

  //Create a mutex to lock access between tasks.
  m_mutex = xSemaphoreCreateRecursiveMutex();

  //Create a semaphore to signal from the interrupt handler that
  //all samples have been taken.
  m_done = xSemaphoreCreateBinary();
}

/*! Takes 100 samples from the ADC channel, this takes about a millisecond.
    \param channel The ADC channel to take samples from, ADC_Channel_0 to ADC_Channel_18.
    \return The average of 100 samples as a float between 0 and 3.3.*/
float CAdc::Read(uint8_t channel)
{
  CLock lock(m_mutex);

  //Make sure the semaphore is cleared.
  xSemaphoreTake(m_done, pdMS_TO_TICKS(0));

  //enable the ADC irq callback, this will call the AdcIRQHandler()
  //method when an ADC interrupt happens
  InstallIRQHandler();

  //enable the clock for the ADC
  CCritical criticalsection;
  RCC_APB2PeriphClockCmd(ADCToRCC(), ENABLE); 
  criticalsection.Leave();

  //initialize the ADC as 12 bit, running in continuous mode
  ADC_InitTypeDef adcinit =
  {
    .ADC_Resolution           = ADC_Resolution_12b,
    .ADC_ScanConvMode         = DISABLE,
    .ADC_ContinuousConvMode   = ENABLE,
    .ADC_ExternalTrigConvEdge = ADC_ExternalTrigConvEdge_None,
    .ADC_ExternalTrigConv     = ADC_ExternalTrigConv_T1_CC1,
    .ADC_DataAlign            = ADC_DataAlign_Right,
    .ADC_NbrOfConversion      = 1
  };
  ADC_Init(m_adc, &adcinit);

  //enable the specified ADC channel, at the slowest sample time
  ADC_RegularChannelConfig(m_adc, channel, 1, ADC_SampleTime_480Cycles);

  //enable the end of conversion (EOC) interrupt for this ADC
  ADC_ITConfig(m_adc, ADC_IT_EOC, ENABLE);

  //reset the sample state variables
  m_value      = 0;
  m_numsamples = 0;
  m_maxsamples = 100;

  //enable the ADC and start the conversion
  ADC_Cmd(m_adc, ENABLE);
  ADC_SoftwareStartConv(m_adc);

  //wait until all samples have been taken
  xSemaphoreTake(m_done, portMAX_DELAY);

  //disable everything
  ADC_Cmd(m_adc, DISABLE);
  ADC_ITConfig(m_adc, ADC_IT_EOC, DISABLE);
  RCC_APB2PeriphClockCmd(ADCToRCC(), DISABLE); 
  RemoveIRQHandler();

  //return the average of the samples as a float between 0 and 3.3
  return (float)m_value / (float)m_numsamples / (4095.0f / 3.3f);
}

/* Reads the internal reference, it's assumed that the VREF+ pin
   is connected to the 3.3v power supply.
   This only works on ADC1.
   \return A value of around 1.0, by which any measurement can be multiplied
           to convert it from using the power supply as reference, to using
           the internal reference.*/
float CAdc::ReadVrefint()
{
  assert(m_adc == ADC1);

  CLock lock(m_mutex);

  //enable the vrefint bit so that the internal reference can be measured
  //on ADC1 channel 17
  CCritical criticalsection;
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_ADC1, ENABLE); 
  ADC_VBATCmd(DISABLE);
  ADC_TempSensorVrefintCmd(ENABLE);
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_ADC1, DISABLE); 
  criticalsection.Leave();

  //read the reference channel, and convert it to a value around 1.0
  float vref = Read(ADC_Channel_Vrefint);
  if (vref > 0.0f) //prevent division by 0
    vref = 1.21f / vref;
  else
    vref = 1.0f;

  //disable the vrefint bit
  criticalsection.Enter();
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_ADC1, ENABLE); 
  ADC_TempSensorVrefintCmd(DISABLE);
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_ADC1, DISABLE); 

  return vref;
}

/*! \return The value of the internal reference, scaled to 3.3 volts.
            It's assumed that the VREF+ pin of the ADC is connected to VCC.
            The measurement of the internal reference voltage can then be used
            to measure VCC.*/
float CAdc::ReadVCC()
{
  return ReadVrefint() * 3.3f;
}

/*! Reads the value of the internal temperature sensor, and scales it according
    to the factory calibration values at 30C and 110C degrees.
    This only works on ADC1.
    \return The temperature in degrees celsius.*/
float CAdc::ReadTempSensor()
{
  assert(m_adc == ADC1);

  CLock lock(m_mutex);

  //read the internal reference voltage
  float vref = ReadVrefint();

  //enable the vrefint bit so that the internal temperature sensor
  //can be measured on ADC1 channel 18
  CCritical criticalsection;
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_ADC1, ENABLE); 
  ADC_VBATCmd(DISABLE);
  ADC_TempSensorVrefintCmd(ENABLE);
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_ADC1, DISABLE); 
  criticalsection.Leave();

  //read the temperature sensor
  float vtemp = Read(ADC_Channel_TempSensor) * vref;

  criticalsection.Enter();
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_ADC1, ENABLE); 
  ADC_TempSensorVrefintCmd(DISABLE);
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_ADC1, DISABLE); 
  criticalsection.Leave();

  //convert the voltage to a temperature in degrees celsius, using the voltage
  //at 25 degrees (0.76V) and the slope (2.5mV per degree)
  return ((vtemp - 0.76f) / 0.0025f) + 25.0f;

  /* The built-in calibration values are currently not used, because for an
     unknown reason, it always produces incorrect results.

  //the STM32F427 has calibrations for its internal temperature sensor
  //written into it at the factory at addresses 0x1FFF7A2C and 0x1FFF7A2E

  //TS ADC raw data acquired at temperature of 30 °C, V DDA = 3.3 V
  uint16_t* ts_cal1 = (uint16_t*)0x1FFF7A2C;

  //TS ADC raw data acquired at temperature of 110 °C, V DDA = 3.3 V
  uint16_t* ts_cal2 = (uint16_t*)0x1FFF7A2E;

  //calculate factory voltages at 30 and 110 degrees
  float v30  = (float)*ts_cal1 / (4095.0f / 3.3f);
  float v110 = (float)*ts_cal2 / (4095.0f / 3.3f);

  //calculate voltage per degree
  float slope = (v110 - v30) / (110.0f - 30.0f);

  //calculate the current temperature based on the 30C calibration
  //temperature and the slope from 30C to 110C
  return (vtemp - v30) / slope + 30.0f;
  */
}

/* Reads the voltage on the VBAT pin, using the internal reference voltage.
   This only works on ADC1.
   \return The voltage on VBAT.*/
float CAdc::ReadVbat()
{
  assert(m_adc == ADC1);

  CLock lock(m_mutex);

  //sample the internal reference voltage
  float vref = ReadVrefint();

  //enable the vbat bit so that the vbat voltage can be measured
  //on ADC1 channel 18
  CCritical criticalsection;
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_ADC1, ENABLE); 
  ADC_VBATCmd(ENABLE);
  ADC_TempSensorVrefintCmd(DISABLE);
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_ADC1, DISABLE); 
  criticalsection.Leave();

  //read the vbat channel, multiply it by 4 because the microcontroller
  //has a voltage divider when measuring VBAT
  //then scale it by the value from ReadVrefint()
  float vbat = Read(ADC_Channel_Vbat) * 4.0f * vref;

  criticalsection.Enter();
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_ADC1, ENABLE); 
  ADC_VBATCmd(DISABLE);
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_ADC1, DISABLE); 

  return vbat;
}

/* \return The corresponding RCC value for m_adc.*/
uint32_t CAdc::ADCToRCC()
{
  assert(IS_ADC_ALL_PERIPH(m_adc));

  if (m_adc == ADC1)
    return RCC_APB2Periph_ADC1;
  else if (m_adc == ADC2)
    return RCC_APB2Periph_ADC2;
  else if (m_adc == ADC3)
    return RCC_APB2Periph_ADC3;

  return 0;
}

/* Installs a pointer in g_adcirqhandlers, so that the AdcIRQHandler() method of
   this class is called when an ADC interrupt happens.*/
void CAdc::InstallIRQHandler()
{
  assert(IS_ADC_ALL_PERIPH(m_adc));

  CCritical criticalsection;

  //check if this is the first irq handler to be installed
  bool enablenvic = true;
  for (uint32_t i = 0; i < sizeof(g_adcirqhandlers) / sizeof(g_adcirqhandlers[0]); i++)
  {
    if (g_adcirqhandlers[i])
    {
      enablenvic = false;
      break;
    }
  }

  //if this is the first irq handler, enable the ADC interrupt in the NVIC
  if (enablenvic)
  {
    NVIC_InitTypeDef nvicsettings =
    {
      .NVIC_IRQChannel                   = (uint8_t)ADC_IRQn,
      .NVIC_IRQChannelPreemptionPriority = (uint8_t)configLIBRARY_MAX_SYSCALL_INTERRUPT_PRIORITY,
      .NVIC_IRQChannelSubPriority        = 0,
      .NVIC_IRQChannelCmd                = ENABLE,
    };
    NVIC_Init(&nvicsettings);
  }

  //set the IRQ handler pointer
  if (m_adc == ADC1)
    g_adcirqhandlers[0] = this;
  else if (m_adc == ADC2)
    g_adcirqhandlers[1] = this;
  else if (m_adc == ADC3)
    g_adcirqhandlers[2] = this;
}

/* Removes the pointer installed in g_adcirqhandlers by InstallIRQHandler().*/
void CAdc::RemoveIRQHandler()
{
  assert(IS_ADC_ALL_PERIPH(m_adc));

  CCritical criticalsection;

  //remove the IRQ handler pointer
  if (m_adc == ADC1)
  {
    assert(g_adcirqhandlers[0] == this);
    g_adcirqhandlers[0] = nullptr;
  }
  else if (m_adc == ADC2)
  {
    assert(g_adcirqhandlers[1] == this);
    g_adcirqhandlers[1] = nullptr;
  }
  else if (m_adc == ADC3)
  {
    assert(g_adcirqhandlers[2] == this);
    g_adcirqhandlers[2] = nullptr;
  }

  //check if this is the last pointer to be removed
  bool disablenvic = true;
  for (uint32_t i = 0; i < sizeof(g_adcirqhandlers) / sizeof(g_adcirqhandlers[0]); i++)
  {
    if (g_adcirqhandlers[i])
    {
      disablenvic = false;
      break;
    }
  }

  //if this was the last pointer removed, disable the ADC interrupt in the NVIC
  if (disablenvic)
  {
    NVIC_InitTypeDef nvicsettings =
    {
      .NVIC_IRQChannel                   = (uint8_t)ADC_IRQn,
      .NVIC_IRQChannelPreemptionPriority = (uint8_t)configLIBRARY_MAX_SYSCALL_INTERRUPT_PRIORITY,
      .NVIC_IRQChannelSubPriority        = 0,
      .NVIC_IRQChannelCmd                = DISABLE,
    };
    NVIC_Init(&nvicsettings);
  }
}

/*! Called from ADC_IRQHandler() when an end of conversion (EOC) interrupt happens.
    It will read the ADC value, and add it to m_value.
    This method must clear the ADC_IT_EOC bit, by reading the data register,
    or by clearing it directly.*/
void CAdc::AdcIRQHandler()
{
  m_value += ADC_GetConversionValue(m_adc);
  m_numsamples++;
  if (m_numsamples >= m_maxsamples)
  {
    ADC_Cmd(m_adc, DISABLE);

    //the task that initialized the conversion should be blocked on the semaphore
    //use portYIELD_FROM_ISR to unblock it immediately when this ISR is finished
    BaseType_t taskwoken = pdFALSE;
    xSemaphoreGiveFromISR(m_done, &taskwoken);
    portYIELD_FROM_ISR(taskwoken);
  }
}
