#include "apnlist.h"
#include "inclstdint.h"
#include "log.h"

#include <cstring>
#include <cctype>

/*! Array of custom apns to match on, matched by the first part of the SIM's ICCID.*/
static const struct
{
  const char* iccid; //!< First part of the ICCID of the SIM to match on.
  const char* pin;   //!< PIN to use with this SIM.
  _apn        apn;   //!< APN to use with this SIM.
}
g_localapns[] =
{
  {
    .iccid = "89460800",
    .pin   = "1111",
    .apn   =
    {
      .carrier         = (char*)"Telenor Connexion",
      .mcc             = (char*)"",
      .mnc             = (char*)"",
      .apn             = (char*)"internet.cxn",
      .user            = (char*)"",
      .password        = (char*)"",
      .mvno_match_data = (char*)"",
      .mvno_type       = (char*)""
    }
  },
  {
    .iccid = "89430301",
    .pin   = "0000",
    .apn   =
    {
      .carrier         = (char*)"T-Mobile LTE",
      .mcc             = (char*)"",
      .mnc             = (char*)"",
      .apn             = (char*)"m2m.business",
      .user            = (char*)"m2m",
      .password        = (char*)"sim",
      .mvno_match_data = (char*)"",
      .mvno_type       = (char*)""
    }
  },
  { //KPN cards from New Line Mobile.
    //This ICCID probably matches a lot of other KPN SIMs too.
    .iccid = "893108",
    .pin   = "0000",
    .apn   =
    {
      .carrier         = (char*)"New Line Mobile - KPN",
      .mcc             = (char*)"",
      .mnc             = (char*)"",
      .apn             = (char*)"m2m.newlinemobile",
      .user            = (char*)"",
      .password        = (char*)"",
      .mvno_match_data = (char*)"",
      .mvno_type       = (char*)""
    }
  },
  { //GSIM cards from T-Mobile.
    .iccid = "89882280",
    .pin   = "0000",
    .apn   =
    {
      .carrier         = (char*)"T-Mobile GSIM",
      .mcc             = (char*)"",
      .mnc             = (char*)"",
      .apn             = (char*)"m2m.public.nl",
      .user            = (char*)"",
      .password        = (char*)"",
      .mvno_match_data = (char*)"",
      .mvno_type       = (char*)""
    }
  },
};

#define NUM_LOCAL_APNS (sizeof(g_localapns) / sizeof(g_localapns[0]))

CAPNList::CAPNList()
{
  Reset();
}

void CAPNList::Reset()
{
  m_localindex   = 0;
  m_globalindex  = 0;
  m_defaultindex = 0;
}

/*! Tries to get an APN from the local list, then the global list, then the default APN.
    Can be called repeatedly to get a different APN to try.
    \param spn The service provider name of the SIM.
    \param imsi The IMSI of the SIM.
    \param iccid The ICCID of the SIM.
    \return A pointer to an apn to use, or nullptr when no apn is found.*/
const _apn* CAPNList::GetApn(const char* spn, const char* imsi, const char* iccid)
{
  const _apn* apn;

  apn = GetApnFromLocal(iccid);
  if (apn == nullptr)
    apn = GetApnFromGlobal(spn, imsi);
  if (apn == nullptr)
    apn = GetDefaultApn();

  return apn;
}

/*! Picks an entry from g_localapns based on the ICCID of the SIM.
    Because there might be multiple matches in the apn list,
    this function can be called repeatedly to get all matches.
    \param iccid The ICCID of the SIM.
    \return A pointer to the apn entry, or nullptr if it fails*/
const _apn* CAPNList::GetApnFromLocal(const char* iccid)
{
  const _apn* apn = nullptr;
  for (uint32_t i = m_localindex + 1; i < NUM_LOCAL_APNS; i++)
  { //check if the first part of the ICCID in the local list matches the first part of the SIM's ICCID
    if (strncmp(g_localapns[i].iccid, iccid, strlen(g_localapns[i].iccid)) == 0)
    {
      apn = &g_localapns[i].apn;
      m_localindex = i;

      PrintLog(LOGAPNLIST, "Picked carrier \"%s\" because of match on ICCID %s\n", 
               apn->carrier, g_localapns[i].iccid);
      break;
    }
  }

  if (apn == nullptr)
    m_localindex = NUM_LOCAL_APNS;

  return apn;
}

/*! Picks an entry from apns in apns-full-conf.c based on the SPN and the IMSI of the SIM.
    Because there might be multiple matches in the apn list,
    this function can be called repeatedly to get all matches.
    \param spn  Service provider name, the provider string from the SIM.
    \param imsi International mobile subscriber identity, a numerical code from the SIM.
    \return A pointer to the apn entry, or nullptr if it fails.*/
const _apn* CAPNList::GetApnFromGlobal(const char* spn, const char* imsi)
{
  //make sure the imsi can hold at least the mcc and the mnc
  if (strlen(imsi) < 5)
    return nullptr;

  const _apn* apn = nullptr;

  //start iterating over the apn list, one position after where the previous call to this function ended
  for (uint32_t i = m_globalindex + 1; i < NUM_APNS; i++)
  {
    //check if the MCC matches the first three characters of the IMSI
    if (strncmp(apns.apn[i].mcc, imsi, 3) != 0)
      continue;

    //check if the MNC matches after the MCC on the IMSI
    if (strncmp(apns.apn[i].mnc, imsi + 3, strlen(apns.apn[i].mnc)) != 0)
      continue;

    //if there is a match, check if mvno_type points to a string, if so, a second check is needed
    //if the second check succeeds, the apn entry is selected
    //mvno_type and mvno_match_data are used together to select virtual carriers on the same network
    if (apns.apn[i].mvno_type)
    {
      if (MatchesMVNO(apns.apn + i, imsi, spn))
      {
        apn = apns.apn + i;
        m_globalindex = i;
        break;
      }
    }
    else
    { //if mvno_type is null, no futher checks are needed
      apn = apns.apn + i;
      m_globalindex = i;
      PrintLog(LOGAPNLIST, "Picked carrier \"%s\" because of match on MCC %s and MNC %s\n",
               apn->carrier, apn->mcc, apn->mnc);
      break;
    }
  }

  return apn;
}

/*! Returns the default APN, but only on the first call.
    \return A pointer to the default APN, or nullptr after the first call.*/
const _apn* CAPNList::GetDefaultApn()
{
  if (m_defaultindex == 0)
  {
    PrintLog(LOGAPNLIST, "Picked default APN\n");

    m_defaultindex = 1;

    static const _apn defaultapn =
    {
      .carrier  = (char*)"Default APN",
      .mcc      = (char*)"0",
      .mnc      = (char*)"0",
      .apn      = (char*)"internet",
      .user     = (char*)"",
      .password = (char*)"",
    };

    return &defaultapn;
  }

  return nullptr;
}

bool CAPNList::MatchesMVNO(const _apn* apn, const char* imsi, const char* spn)
{
  if (apn->mvno_match_data == NULL)
  {
    PrintLog(LOGAPNLIST, "Warning: carrier \"%s\" has mvno_type set but mvno_match_data is NULL\n", apn->carrier);
    return false;
  }

  if (strcasecmp(apn->mvno_type, "spn") == 0)
  { //for an SPN match, check if mvno_match_data matches the start of the SPN
    if (strncmp(apn->mvno_match_data, spn, strlen(apn->mvno_match_data)) == 0)
    {
      PrintLog(LOGAPNLIST, "Picked carrier \"%s\" because of SPN match on \"%s\"\n",
               apn->carrier, apn->mvno_match_data);

      return true;
    }
  }
  else if (strcasecmp(apn->mvno_type, "imsi") == 0)
  { //for an IMSI match, the IMSI is matched against a pattern in mvno_match_data
    bool     match = true;
    uint32_t len   = strlen(apn->mvno_match_data);

    //don't check an IMSI shorter than mvno_match_data
    if (len <= strlen(imsi))
    {
      //check the IMSI for a match, an x in mvno_match_data is a wildcard, it can be any character
      for (uint32_t c = 0; c < len; c++)
      {
        char imsichr = apn->mvno_match_data[c];
        if (isdigit(imsichr))
        {
          if (imsichr != imsi[c])
          {
            match = false;
            break;
          }
        }
        else if (imsichr == 'x')
        {
          continue;
        }
        else
        {
          PrintLogDecode(LOGAPNLIST, "Warning: invalid character \"%c\" in mvno_match_data of carrier \"%s\"",
                         imsichr, apn->carrier);
          match = false;
          break;
        }
      }
    }

    if (match)
    {
      PrintLog(LOGAPNLIST, "Picked carrier \"%s\" because of IMSI match on \"%s\"\n",
               apn->carrier, apn->mvno_match_data);

      return true;
    }
  }
  else
  {
    //TODO: implement GID matching, it's not clear what exactly mvno_match_data describes when mvno_type is set to gid
    PrintLog(LOGAPNLIST, "Warning: Match type \"%s\" of carrier \"%s\" not implemented\n", apn->mvno_type, apn->carrier);
  }

  return false;
}

/*! Selects a PIN from g_localapns based on the ICCID of the SIM,
    returns "0000" if the ICCID of the card can't be matched to anything in g_localapns.
    \param iccid A pointer to the ICCID of the SIM.
    \return A pointer to the PIN to use.*/
const char* CAPNList::GetPIN(const char* iccid)
{
  for (uint32_t i = 0; i < NUM_LOCAL_APNS; i++)
  { //check if the first part of the ICCID in the local list matches the first part of the SIM's ICCID
    if (strncmp(g_localapns[i].iccid, iccid, strlen(g_localapns[i].iccid)) == 0)
    {
      PrintLog(LOGAPNLIST, "Picked PIN %s based on ICCID %s\n", g_localapns[i].pin, g_localapns[i].iccid);
      return g_localapns[i].pin;
    }
  }

  const char* defaultpin = "0000";
  PrintLog(LOGAPNLIST, "Picked default pin %s\n", defaultpin);

  return defaultpin;
}
