#include "atcmd.h"
#include "log.h"

#include <cstring>
#include <cassert>
#include <cstdio>
#include <cstdarg>

#include <algorithm>

CATCmd::CATCmd(CUsart& usart) : m_usart(usart)
{
  ResetState();
}

/*! Resets all state variables.*/
void CATCmd::ResetState()
{
  m_writetimeout = DEFAULT_WRITE_TIMEOUT;
  m_echotimeout  = DEFAULT_ECHO_TIMEOUT;
  m_readtimeout  = DEFAULT_READ_TIMEOUT;
  m_cmdlen       = 0;
  m_readlen      = 0;
  m_result       = ATCmdResult_Success;
  m_usartresult  = 0;
  memset(m_readbuf, 0, sizeof(m_readbuf));
  memset(m_cmdbuf, 0, sizeof(m_cmdbuf));
}

/*! Writes cmd to the usart, and reads the echo back.
    \param cmd A C-string to write to the usart.
    \return ATCmdResult_Success if the echo was read back successfully.
            ATCmdResult_Write_Timeout if the write timed out.
            ATCmdResult_Echo_Timeout if reading the echo timed out
            ATCmdResult_Echo_Usart_Error if a usart error was detected when reading the echo
            ATCmdResult_Echo_Fail if the echo was incorrect.
            ATCmdResult_Cmd_Too_Large if the command does not fit into m_buf.*/
EATCmdResult CATCmd::SendCmd(const char* fmt, ...)
{
  //use vsnprintf to print the command to the buffer
  va_list args;
  va_start(args, fmt);
  m_cmdlen = vsnprintf(m_cmdbuf, AT_BUFSIZE, fmt, args);
  va_end(args);

  //check if the command fits in the buffer
  //if not, turn it into a C-string, adjust m_cmdlen to the length of the C-string,
  //and return an error, this way, m_cmdbuf holds a valid string for error printing
  if (m_cmdlen >= AT_BUFSIZE)
  {
    m_cmdlen = AT_BUFSIZE - 1;
    m_cmdbuf[AT_BUFSIZE - 1] = 0; 
    return ATCmdResult_Cmd_Too_Large;
  }

  //clear anything that was previously received from the modem
  m_usart.ClearInput();

  //write the command string to the usart
  if (m_usart.WriteStr(m_cmdbuf, m_writetimeout) == WRITERESULT_ERROR_TIMEOUT)
    return ATCmdResult_Write_Timeout;

  //if the echo timeout is set to 0, don't read the echo at all
  if (m_echotimeout == 0)
    return ATCmdResult_Success;

  //try to read the echo back
  m_usartresult = m_usart.Read((uint8_t*)m_readbuf, m_cmdlen, &m_readlen, m_echotimeout);
  m_readbuf[m_readlen] = 0; //turn the echo into a C-string
  if (m_usartresult != READRESULT_SUCCESS)
  { //if the read failed, return if it timed out, or the usart detected an error
    if (m_usartresult == READRESULT_ERROR_TIMEOUT)
      return ATCmdResult_Echo_Timeout;
    else
      return ATCmdResult_Echo_Usart_Error;
  }

  //the number of received bytes should be the same as the command length at this point
  assert(m_readlen == m_cmdlen);

  //compare the echo, and return if it's the same as the command that was just sent
  if (strncmp((const char*)m_cmdbuf, (const char*)m_readbuf, m_cmdlen) == 0)
    return ATCmdResult_Success;
  else
    return ATCmdResult_Echo_Fail;
}

/*! Reads the modem reply from the usart, starts and ends with \r\n.
    \param start When not nullptr, the string that signals the start of the reply.
                 When nullptr, the CUsart class waits for the default \r\n start.
    \param end When not nullptr, the string that signals the end of the reply.
               When nullptr, the CUsart class waits for a default modem reply.
    \return ATCmdResult_Success on success,
            ATCmdResult_Read_Timeout on a read timeout,
            ATCmdResult_Read_Usart_Error when the usart detectd an error.*/
EATCmdResult CATCmd::ReadReply(const char* start /*= nullptr*/, const char* end /*= nullptr*/)
{
  m_usartresult = m_usart.ReadModemReply(m_readbuf, AT_BUFSIZE, &m_readlen, m_readtimeout, start, end);
  return TranslateUsartResult(m_usartresult);
}

/*! Checks if the last two characters of m_readbuf is OK.
    \return ATCmdResult_Success if OK was received as the last two characters.
            ATCmdResult_Reply_Not_OK if the last two characters are not OK,
                                     or less than two characters are received.*/
EATCmdResult CATCmd::CheckOKReply()
{
  const char* okreply = "\r\nOK\r\n";
  uint32_t    okreplylen = strlen(okreply);

  if (m_readlen >= okreplylen)
  {
    if (strncmp(okreply, m_readbuf + m_readlen - okreplylen, okreplylen) == 0)
      return ATCmdResult_Success;
  }

  return ATCmdResult_Reply_Not_OK;
}

/*! Checks if m_readbuf ends with \r\nCONNECT\r\n.
    \return ATCmdResult_Success if OK was received as the last two characters.
            ATCmdResult_Reply_Not_Connect if the last two characters are not OK,
                                          or less than two characters are received.*/
EATCmdResult CATCmd::CheckConnectReply()
{
  const char* connectreply = "\r\nCONNECT\r\n";
  uint32_t    connectreplylen = strlen(connectreply);

  if (m_readlen >= connectreplylen)
  {
    if (strncmp(connectreply, m_readbuf + m_readlen - connectreplylen, connectreplylen) == 0)
      return ATCmdResult_Success;
  }

  return ATCmdResult_Reply_Not_CONNECT;
}

/*! Converts a result from CUsart::ReadModemReply() to a EATCmdResult.
    \param result The value returned by CUsart::ReadModemReply().
    \return A value of EATCmdResult.*/
EATCmdResult CATCmd::TranslateUsartResult(uint32_t result)
{
  if (result != READRESULT_SUCCESS)
  {
    if (result == READRESULT_ERROR_TIMEOUT)
      return ATCmdResult_Read_Timeout;
    else if (result == READRESULT_ERROR_TOO_LARGE)
      return ATCmdResult_Reply_Too_Large;
    else
      return ATCmdResult_Read_Usart_Error;
  }

  return ATCmdResult_Success;
}

/*! If m_result is not ATCmdResult_Success, this function will print the error.*/
void CATCmd::PrintError()
{
  if (m_result != ATCmdResult_Success)
  {
    if (m_result == ATCmdResult_Reply_Not_OK)
      PrintLogDecode(LOGATCMD, "ERROR: Did not receive OK reply to command: \"%s\", received \"%s\"", m_cmdbuf, m_readbuf);
    else if (m_result == ATCmdResult_Write_Timeout)
      PrintLogDecode(LOGATCMD, "ERROR: Timout writing command: \"%s\"", m_cmdbuf);
    else if (m_result == ATCmdResult_Echo_Timeout)
      PrintLogDecode(LOGATCMD, "ERROR: Receiving echo to command timed out: \"%s\", received \"%s\"", m_cmdbuf, m_readbuf);
    else if (m_result == ATCmdResult_Echo_Usart_Error)
      PrintLogDecode(LOGATCMD, "ERROR: Usart error \"%s\" receiving echo to command: \"%s\", received \"%s\"",
                     m_usart.ResultToStr(m_usartresult), m_cmdbuf, m_readbuf);
    else if (m_result == ATCmdResult_Echo_Fail)
      PrintLogDecode(LOGATCMD, "ERROR: Did not receive correct echo to command: \"%s\", received \"%s\"", m_cmdbuf, m_readbuf);
    else if (m_result == ATCmdResult_Cmd_Too_Large)
      PrintLogDecode(LOGATCMD, "ERROR: Command too large for buffer: \"%s\"", m_cmdbuf);
    else if (m_result == ATCmdResult_Read_Timeout)
      PrintLogDecode(LOGATCMD, "ERROR: Timeout reading reply to command: \"%s\", received \"%s\"", m_cmdbuf, m_readbuf);
    else if (m_result == ATCmdResult_Read_Usart_Error)
      PrintLogDecode(LOGATCMD, "ERROR: USART error \"%s\" reading reply to command: \"%s\", received \"%s\"",
                     m_usart.ResultToStr(m_usartresult), m_cmdbuf, m_readbuf);
    else if (m_result == ATCmdResult_Reply_Too_Large)
      PrintLogDecode(LOGATCMD, "ERROR: Reply to command too large: \"%s\", received \"%s\"", m_cmdbuf, m_readbuf);
    else if (m_result == ATCmdResult_Missing_Start_Quote)
      PrintLogDecode(LOGATCMD, "ERROR: Missing start quote in reply to command: \"%s\", received \"%s\"", m_cmdbuf, m_readbuf);
    else if (m_result == ATCmdResult_Reply_Too_Short)
      PrintLogDecode(LOGATCMD, "ERROR: Reply too short to command: \"%s\", received \"%s\"", m_cmdbuf, m_readbuf);
    else if (m_result == ATCmdResult_Reply_Not_CONNECT)
      PrintLogDecode(LOGATCMD, "ERROR: Did not receive CONNECT reply to command: \"%s\", received \"%s\"", m_cmdbuf, m_readbuf);
    else
      PrintLogDecode(LOGATCMD, "ERROR: UNKNOWN ERROR %i, command: \"%s\"", (int)m_result, m_cmdbuf);
  }
}

/*! Counts the number of lines in the reply, by counting the number of \r\n after the first \r\n.
    \return The number of lines in the reply.*/
uint32_t CATCmd::ReplyNumLines()
{
  if (m_readlen > 2)
  {
    uint32_t numlines  = 0;
    const char* bufptr = m_readbuf + 2;
    while ((bufptr = strstr(bufptr, "\r\n")))
    {
      bufptr++;
      numlines++;
    }

    return numlines;
  }

  return 0;
}

/*  Return a pointer to a specific line, and the length of the line.
    The line is not 0 terminated.
    The reply received from the modem starts with \r\n, and each line ends with \r\n.
    This function assumes that each line starts with \r\n, so it finds the
    \r\n corresponding with the line, and returns a pointer to it.
    \param line The number of the line to find, starting at 0.
    \return A pointer to the line in m_readbuf, or nullptr.*/
const char* CATCmd::ReplyLine(uint32_t line, uint32_t* linelength)
{
  //state 0 = find \r, state 1 = find \n, state2 = save pointer to first character of the line
  uint32_t state      = 0;
  uint32_t linenr     = 0;
  const char* lineptr = nullptr;
  uint32_t bufpos;

  //find the \r\n that the line starts with
  for (bufpos = 0; bufpos < m_readlen; bufpos++)
  {
    if (state == 0)
    { //check for carriage return
      if (m_readbuf[bufpos] == '\r')
        state = 1;
    }
    else if (state == 1)
    { //find the first newline after the carriage return
      if (m_readbuf[bufpos] == '\n')
      {
        state = 2;
      }
      else if (m_readbuf[bufpos] != '\r')
      { //reset the state if there was no newline after the carriage return,
        //and this character is also not a carriage return
        state = 0;
      }
    }
    else if (state == 2)
    { //if this is the first character after \r\n,
      //save the pointer if it's the requested line
      if (linenr == line)
      {
        lineptr = m_readbuf + bufpos;
        break;
      }
      else
      {
        linenr++;
        state = 0;
        //this is the first character of the current line, however,
        //if the length of the line is 0, there is a carriage return here which should be checked
        bufpos--;
      }
    }
  }

  if (lineptr == nullptr)
  { //return nullptr if the requested line was not found
    *linelength = 0;
    return nullptr;
  }

  //if the requested line was found, its length should be calculated
  *linelength = 0;
  state = 0;

  //find the lenght of the line by searching for the next \r\n
  for (; bufpos < m_readlen; bufpos++)
  {
    if (state == 0)
    {
      if (m_readbuf[bufpos] == '\r')
        state = 1;
    }
    else if (state == 1)
    {
      if (m_readbuf[bufpos] == '\n')
      {
        //don't count the previous \r as part of the line
        if (*linelength > 0)
          (*linelength)--;

        // \r\n is found, stop reading
        break;
      }
      else if (m_readbuf[bufpos] != '\r')
      {
        state = 0;
      }
    }

    (*linelength)++;
  }

  return lineptr;
}

/*! Reads from m_usart directly into m_readbuf.
    \param len The maximum number of bytes to read.
    \param timeout The timeout of the entire write in seconds.
    \return ATCmdResult_Success if everything succeeded, or something else if it failed.*/
EATCmdResult CATCmd::ReadUsart(uint32_t len, uint32_t timeout)
{
  //reset the read buffer
  m_readbuf[0] = 0;
  m_readlen    = 0;

  //decide the maximum number of bytes to read
  len = std::min(len, (uint32_t)AT_BUFSIZE - 1);

  //read bytes from the usart into m_readbuf
  m_usartresult = m_usart.Read((uint8_t*)m_readbuf, len, &m_readlen, timeout * 1000);

  //always terminate as a C-string
  m_readbuf[m_readlen] = 0;

  return m_result = TranslateUsartResult(m_usartresult);
}

/*! Writes from buf directly into m_usart.
    \param buf The buffer to write to the usart, this buffer might point to m_readbuf.
    \param len The number of bytes to write to the usart.
    \param timeout The timeout in seconds for the entire write.*/
EATCmdResult CATCmd::WriteUsart(uint8_t* buf, uint32_t len, uint32_t timeout)
{
  m_result = ATCmdResult_Success;

  //write the buffer to the usart
  if (m_usart.Write(buf, len, timeout * 1000) == WRITERESULT_ERROR_TIMEOUT)
    m_result = ATCmdResult_Write_Timeout;

  return m_result;
}

/*! Function:
    Test command.
    Sends:
    AT
    Expects:
    OK
*/
EATCmdResult CATCmd::AT()
{
  m_result = SendCmd("AT\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

#define CFUN_TIMEOUT 60000

/*! Function:
    The command controls the functionality level. It can also be used to reset the ME.
    <fun> 0 Minimum functionality
          1 Full functionality (Default)
          4 Disable the ME from both transmitting and receiving RF signals

    <rst> 0 Do not reset the ME before setting it to <fun> functionality level.
            This is the default setting when <rst> is not given.
          1 Reset the ME. The device is fully functional after the reset. This value is available
            only for <fun>=1.

    Sends:
    AT+CFUN=<fun>[,<rst>]
    Expects:
    OK
*/
EATCmdResult CATCmd::ATCFUN(uint32_t fun, uint32_t rst)
{
  m_result = SendCmd("AT+CFUN=%" PRIu32 ",%" PRIu32 "\r", fun, rst);
  if (m_result != ATCmdResult_Success)
    return m_result;

  uint32_t readtimeout = m_readtimeout;
  m_readtimeout = CFUN_TIMEOUT;
  m_result = ReadReply();
  m_readtimeout = readtimeout;

  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Attachment or Detachment of PS.
    <state> Indicates the state of PS attachment
    0       Detached
    1       Attached
    Sends:
    AT+CGATT?
    Expects:
    +CGATT: <state>
    OK
*/
EATCmdResult CATCmd::ATCGATT_Query()
{
  m_result = SendCmd("AT+CGATT?\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Attachment or Detachment of PS.
    <state> Indicates the state of PS attachment
    0       Detached
    1       Attached
    Other values are reserved and will result in an ERROR response to the Write Command.
    Sends:
    AT+CGATT=<state>
    Expects:
    OK
*/
EATCmdResult CATCmd::ATCGATT_Set(uint32_t state)
{
  m_result = SendCmd("AT+CGATT=%" PRIu32 "\r", state);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Specifies PDP context parameters for a specific context.
    <cid>       PDP context identifier. A numeric parameter which specifies a particular PDP context
                definition. The parameter is local to the TE-MT interface and is used in other PDP
                context-related commands. The range of permitted values (minimum value=1) is
                returned by the test form of the command.
    <PDP_type>  Packet data protocol type. A string parameter which specifies the type of packet data
                protocol.
                "IP" IPv4
                "PPP"
                "IPV6"
                "IPV4V6"
    <APN>       Access point name. A string parameter that is a logical name used to select the GGSN
                or the external packet data network. If the value is null or omitted, then the
                subscription value will be requested. 
    <PDP_addr>  A string parameter identifies the MT in the address space applicable to the PDP. If the
                value is null or omitted, then a value may be provided by the TE during the PDP startup
                procedure or, failing that, a dynamic address will be requested. The allocated address
                may be read using the AT+CGPADDR command.
    <data_comp> A numeric parameter that controls PDP data compression (applicable for SNDCP only)
                (refer to 3GPP TS 44.065).
                0 OFF (Default value if the parameter is omitted)
                1 ON (Manufacturer preferred compression)
                2 V.42bis
                3 V.44 (Not supported currently)
    <head_comp> A numeric parameter that controls PDP header compression (refer to 3GPP TS 44.065
                and 3GPP TS 25.323).
                0 OFF
                1 ON
                2 RFC1144
                3 RFC2507
                4 RFC3095
    Sends:
    AT+CGDCONT=<cid>[,<PDP_type>[,<APN>[,<PDP_addr>[,<data_comp>[,<head_comp>]]]]]
    Expects:
    OK
*/
EATCmdResult CATCmd::ATCGDCONT(uint32_t cid, const char* PDP_type /*= nullptr*/, const char* APN /*= nullptr*/,
                               const char* PDP_addr /*= nullptr*/, uint32_t* data_comp /*= nullptr*/, uint32_t* head_comp /*= nullptr*/)
{
  if (PDP_type == nullptr)
  { //if PDP_type is null, only send the PDP context ID, this makes all values of this PDP context undefined
    m_result = SendCmd("AT+CGDCONT=%" PRIu32 "\r", cid);
  }
  else
  {
    //APN and PDP_addr are mandatory
    assert(APN);
    assert(PDP_addr);

    //data_comp and head_comp are optional, only pass them if their pointers
    //are not null
    if (data_comp == nullptr)
    {
      m_result = SendCmd("AT+CGDCONT=%" PRIu32 ",\"%s\",\"%s\",\"%s\"\r",
                         cid, PDP_type, APN, PDP_addr);
    }
    else if (head_comp == nullptr)
    {
      m_result = SendCmd("AT+CGDCONT=%" PRIu32 ",\"%s\",\"%s\",\"%s\",%" PRIu32 "\r",
                         cid, PDP_type, APN, PDP_addr, *data_comp);
    }
    else
    {
      m_result = SendCmd("AT+CGDCONT=%" PRIu32 ",\"%s\",\"%s\",\"%s\",%" PRIu32 ",%" PRIu32 "\r",
                         cid, PDP_type, APN, PDP_addr, *data_comp, *head_comp);
    }
  }

  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Request international mobile subscriber identity (IMSI).
    Sends:
    AT+CIMI
    Expects:
    <IMSI>
    OK
*/
EATCmdResult CATCmd::ATCIMI()
{
  m_result = SendCmd("AT+CIMI\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Sets the error message format.
    format:
      0 Disable result code
      1 Enable result code and use numeric values
      2 Enable result code and use verbose values
    Sends:
    AT+CMEE=<format>
    Expects:
    OK
*/
EATCmdResult CATCmd::ATCMEE(uint32_t format)
{
  m_result = SendCmd("AT+CMEE=%" PRIu32 "\r", format);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    The command returns the list of operator names from the ME. Each operator code <numericn> that has
    an alphanumeric equivalent <alphan> in the ME memory is returned.

    Sends:
    AT+COPN
    Expects:
    +COPN: <numeric1>,<alpha1>
    [+COPN: <numeric2>,<alpha2>
    [...]]
    OK
    */
EATCmdResult CATCmd::ATCOPN()
{
  m_result = SendCmd("AT+COPN\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Operator selection query.
    <mode>   0 Automatic mode. <oper> field is ignored.
             1 Manual operator selection. <oper> field shall be presented and <Act> optionally.
             2 Manual deregister from network.
             3 Set only <format> (for AT+COPS? Read Command), and do not attempt to
               register/deregister (<oper> and <Act> fields are ignored). This value is invalid in
               the response of the Read Command.
             4 Manual/automatic selection. <oper> field shall be presented. If manual selection
               fails, automatic mode (<mode>=0) is entered.
    <format> 0 Long format alphanumeric <oper> which can be up to 16 characters long
             1 Short format alphanumeric <oper>
             2 Numeric <oper>. GSM location area identification number.
    <oper>   Operator in format as per <mode>
    <Act>    0 GSM
             8 LTE Cat M1
             9 LTE Cat NB1
    Sends:
    AT+COPS?
    Expects:
    +COPS: <mode>[,<format>[,<oper>][,<Act>]]
    OK
*/
EATCmdResult CATCmd::ATCOPS_Query()
{
  m_result = SendCmd("AT+COPS?\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Operator selection query.
    <mode>   0 Automatic mode. <oper> field is ignored.
             1 Manual operator selection. <oper> field shall be presented and <Act> optionally.
             2 Manual deregister from network.
             3 Set only <format> (for AT+COPS? Read Command), and do not attempt to
               register/deregister (<oper> and <Act> fields are ignored). This value is invalid in
               the response of the Read Command.
             4 Manual/automatic selection. <oper> field shall be presented. If manual selection
               fails, automatic mode (<mode>=0) is entered.
    <format> 0 Long format alphanumeric <oper> which can be up to 16 characters long
             1 Short format alphanumeric <oper>
             2 Numeric <oper>. GSM location area identification number.
    <oper>   Operator in format as per <mode>
    <Act>    0 GSM
             8 LTE Cat M1
             9 LTE Cat NB1
    Sends:
    AT+COPS=<mode>[,<format>[,<oper>[,<Act>]]]
    Expects:
    OK
*/
EATCmdResult CATCmd::ATCOPS_Set(uint32_t mode, const uint32_t* format /*= nullptr*/, const uint32_t* oper /*= nullptr*/, const uint32_t* act /*= nullptr*/)
{
  if (format == nullptr)
    m_result = SendCmd("AT+COPS=%" PRIu32 "\r", mode);
  else if (oper == nullptr)
    m_result = SendCmd("AT+COPS=%" PRIu32 ",%" PRIu32 "\r", mode, *format);
  else if (act == nullptr)
    m_result = SendCmd("AT+COPS=%" PRIu32 ",%" PRIu32 ",%" PRIu32 "\r", mode, *format, *oper);
  else
    m_result = SendCmd("AT+COPS=%" PRIu32 ",%" PRIu32 ",%" PRIu32 ",%" PRIu32 "\r", mode, *format, *oper, *act);

  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

#define COPS_SET_QUERY_TIMEOUT 300000

/*! Function:
    Operator selection query.
    <stat>   0 Unknown
             1 Operator available
             2 Current operator
             3 Operator forbidden
    <oper>   Operator in format as per <mode>
    <mode>   0 Automatic mode. <oper> field is ignored.
             1 Manual operator selection. <oper> field shall be presented and <Act> optionally.
             2 Manual deregister from network.
             3 Set only <format> (for AT+COPS? Read Command), and do not attempt to
               register/deregister (<oper> and <Act> fields are ignored). This value is invalid in
               the response of the Read Command.
             4 Manual/automatic selection. <oper> field shall be presented. If manual selection
               fails, automatic mode (<mode>=0) is entered.
    <format> 0 Long format alphanumeric <oper> which can be up to 16 characters long
             1 Short format alphanumeric <oper>
             2 Numeric <oper>. GSM location area identification number.
    <Act>    0 GSM
             8 LTE Cat M1
             9 LTE Cat NB1
    Sends:
    AT+COPS=?
    Expects:
    +COPS: (list of supported<stat>, long alphanumeric <oper>,
    short alphanumeric <oper>, numeric <oper>s)[,< Act>])s]
    [,,(list of supported <mode>s),(list of supported <format>s )]
    OK
*/
EATCmdResult CATCmd::ATCOPS_Set_Query()
{
  m_result = SendCmd("AT+COPS=?\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  uint32_t readtimeout = m_readtimeout;
  m_readtimeout = COPS_SET_QUERY_TIMEOUT;
  m_result = ReadReply();
  m_readtimeout = readtimeout;

  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Test if the SIM requires a PIN.
    reply codes:
      READY       MT is not pending for any password
      SIM PIN     MT is waiting for (U)SIM PIN to be given
      SIM PUK     MT is waiting for (U)SIM PUK to be given
      SIM PIN2    MT is waiting for (U)SIM PIN2 to be given
      SIM PUK2    MT is waiting for (U)SIM PUK2 to be given
      PH-NET PIN  MT is waiting for network personalization password to be given
      PH-NET PUK  MT is waiting for network personalization unblocking password
    Sends:
    AT+CPIN?
    Expects:
    +CPIN: <code>
    OK
*/
EATCmdResult CATCmd::ATCPIN_Query()
{
  m_result = SendCmd("AT+CPIN?\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Enters the SIM PIN.
    Sends:
    AT+CPIN=<pin>
    Expects:
    OK
*/
EATCmdResult CATCmd::ATCPIN_Set(const char* pin)
{
  m_result = SendCmd("AT+CPIN=%s\r", pin);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Query the list of preferred operators.
    Sends:
    AT+CPOL?
    Expects:
    +CPOL: <index>,<format>,<oper>[,<GSM>,<GSM_compact,<UTRAN>,<E-UTRAN>]<index>,<format>,<oper>[,<GSM>,<GSM_compact,<UTRAN>,<E-UTRAN>]
    OK
*/
EATCmdResult CATCmd::ATCPOL_Query()
{
  m_result = SendCmd("AT+CPOL?\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Query the network registration status.
    <n>    0 Disable network registration unsolicited result code
           1 Enable network registration unsolicited result code: +CREG:<stat>
           2 Enable network registration and location information unsolicited result code: +CREG: <stat>[,<lac>,<ci>[,<Act>]]
    <stat> 0 Not registered. MT is not currently searching an operator to register to.
           1 Registered, home network.
           2 Not registered, but MT is currently trying to attach or searching an operator to register to.
           3 Registration denied.
           4 Unknown.
           5 Registered, roaming
    <lac>  String type. Two-byte location area code in hexadecimal format.
    <ci>   String type. Four-byte GERAN/E-UTRAN cell ID in hexadecimal format.
    <Act>  Access technology selected
           0 GSM
           8 LTE Cat M1
           9 LTE Cat NB1
    Sends:
    AT+CREG?
    Expects:
    +CREG: <n>,<stat>[,<lac>,<ci>[,<Act>]]
    OK
*/
EATCmdResult CATCmd::ATCREG_Query()
{
  m_result = SendCmd("AT+CREG?\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Sets the network registration status.
    <n>    0 Disable network registration unsolicited result code
           1 Enable network registration unsolicited result code: +CREG:<stat>
           2 Enable network registration and location information unsolicited result code: +CREG: <stat>[,<lac>,<ci>[,<Act>]]
    Sends:
    AT+CREG=<n>
    Expects:
    OK
*/
EATCmdResult CATCmd::ATCREG_Set(uint32_t n)
{
  m_result = SendCmd("AT+CREG=%" PRIu32 "\r", n);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Restricted (U)SIM Access.
    <command>        (U)SIM command number
                     176 READ BINARY
                     178 READ RECORD
                     192 GET RESPONSE
                     214 UPDATE BINARY
                     220 UPDATE RECORD
                     242 STATUS
    <fileId>         Integer type. Identifier for an elementary data file on (U)SIM, if used by <command>.
    <P1>, <P2>, <P3> Integer type. Parameters transferred by the MT to the (U)SIM. These parameters are
                     mandatory for every command, except GET RESPONSE and STATUS. The values are described in 3GPP TS 51.011.
    <data>           Information which shall be written to the (U)SIM (hexadecimal character format; refer to AT+CSCS).
    <pathId>         The directory path of an elementary file on a UICC in hexadecimal format.
    <sw1>, <sw2>     Integer type. Information from the (U)SIM about the execution of the actual command.
                     These parameters are delivered to the TE in both cases, on successful or failed execution of the command.
    <response>       Response of a successful completion of the command previously issued (hexadecimal character format;
                     refer to AT+CSCS). STATUS and GET RESPONSE return data, which gives information about the current
                     elementary data field. The information includes the type of file and its size (refer to 3GPP TS 51.011).
                     After READ BINARY, READ RECORD or RETRIEVE DATA command, the requested data will be returned.
                     <response> is not returned after a successful UPDATE BINARY, UPDATE RECORD or SET DATA command.
   Sends:
    AT+CRSM=<command>[,<fileld>[,<P1>,<P2>,<P3>[,<data>][,<pathld>]]]
    Expects:
    +CRSM: <sw1>,<sw2>[,<response>]
    OK
*/
EATCmdResult CATCmd::ATCRSM(uint32_t command, const uint32_t* fileid /*= nullptr*/, const uint32_t* p /*= nullptr*/,
                            const char* data /*= nullptr*/, const char* pathid /*= nullptr*/)
{
  if (fileid == nullptr)
    m_result = SendCmd("AT+CRSM=%" PRIu32 "\r", command);
  else if (p == nullptr)
    m_result = SendCmd("AT+CRSM=%" PRIu32 ",%" PRIu32 "\r", command, *fileid);
  else if (data == nullptr)
    m_result = SendCmd("AT+CRSM=%" PRIu32 ",%" PRIu32 ",%" PRIu32 ",%" PRIu32 ",%" PRIu32 "\r",
                       command, *fileid, p[0], p[1], p[2]);
  else if (pathid == nullptr)
    m_result = SendCmd("AT+CRSM=%" PRIu32 ",%" PRIu32 ",%" PRIu32 ",%" PRIu32 ",%" PRIu32 ",%s\r",
                       command, *fileid, p[0], p[1], p[2], data);
  else
    m_result = SendCmd("AT+CRSM=%" PRIu32 ",%" PRIu32 ",%" PRIu32 ",%" PRIu32 ",%" PRIu32 ",%s,%s\r",
                       command, *fileid, p[0], p[1], p[2], data, pathid);

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Uses AT+CRSM to do a read binary command (176) on a specific location and of a specific length.
    \param location The SIM location to do the binary read at.
    \param length The length in bytes of the binary read.
    \param dataptr A pointer to a buffer holding the hex data is returned here.
    \param datalen The length of the string holding the hex data.
    \return ATCmdResult_Success if the command succeeded, false if it failed.*/
EATCmdResult CATCmd::ATCRSM_ReadBinary(uint32_t location, uint32_t length, const char** dataptr, uint32_t* datalen)
{
  *dataptr = nullptr;
  *datalen = 0;

  const char* line;
  uint32_t    linelength;

  //use the AT+CRSM command to do a binary read (176)
  uint32_t p[3] = {0, 0, length};
  if (ATCRSM(176, &location, p) != ATCmdResult_Success)
    return m_result;

  //the reply is something like:
  //+CRSM: 144,0,"00542D4D6F62696C6520204E4CFFFFFFFF"
  //find the first quote on the first line, and create a pointer to the first character after that
  line = ReplyLine(0, &linelength);
  char* data = (char*)memchr(line, '"', linelength);
  if (data == nullptr || data[0] != '"')
  { //if the starting quote is missing, return an error
    m_result = ATCmdResult_Missing_Start_Quote;
    return m_result;
  }
  data++; //skip over the initial "

  //get the length of the hex string after the starting quote, limited to the number of bytes times 2
  uint32_t hexstrlen = strspn(data, "0123456789abcdefABCDEF");
  if (hexstrlen != length * 2)
  { //return an error if the length of the hex string is not correct
    m_result = ATCmdResult_Reply_Too_Short;
    return m_result;
  }

  *dataptr = data;
  *datalen = hexstrlen;

  m_result = ATCmdResult_Success;
  return m_result;
}

/*! Function:
    Signal Quality Report.
    Sends:
    AT+CSQ
    Expects:
    +CSQ: <rssi>,<ber>
    OK
*/
EATCmdResult CATCmd::ATCSQ()
{
  m_result = SendCmd("AT+CSQ\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Set all current parameters to manufacturer defaults.
    Sends:
    AT&F
    Expects:
    OK
*/
EATCmdResult CATCmd::ATF()
{
  m_result = SendCmd("AT&F\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Request Manufacturer Identification.
    Sends:
    AT+GMI
    Expects:
    <manufacturer>
    OK
*/
EATCmdResult CATCmd::ATGMI()
{
  m_result = SendCmd("AT+GMI\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Request TA Model Identification.
    Sends:
    AT+GMM
    Expects:
    <modem identification>
    OK
*/
EATCmdResult CATCmd::ATGMM()
{
  m_result = SendCmd("AT+GMM\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Request TA Revision Identification of Software Release.
    Sends:
    AT+GMR
    Expects:
    <modem revision>
    OK
*/
EATCmdResult CATCmd::ATGMR()
{
  m_result = SendCmd("AT+GMR\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Request international mobile equipment identity (IMEI).
    Sends:
    AT+GSN
    Expects:
    <IMEI>
    OK
*/
EATCmdResult CATCmd::ATGSN()
{
  m_result = SendCmd("AT+GSN\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Display product identification information.
    Sends:
    ATI
    Expects:
    <modem information>
    OK
*/
EATCmdResult CATCmd::ATI()
{
  m_result = SendCmd("ATI\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    The command determines the flow control behavior of the serial port.
    Note:
    Flow control is only applicable for data mode.
    <dce_by_dte>
      Specifies the method that will be used by TE when receiving data from TA
      0 None
      2 RTS flow control
    <dte_by_dce>
      Specifies the method that will be used by TA when receiving data from TE
      0 None
      2 CTS flow control
    Sends:
    AT+IFC=<dce_by_dte>,<dte_by_dce>
    Expects:
    OK
*/
EATCmdResult CATCmd::ATIFC(uint32_t dce_by_dte, uint32_t dte_by_dce)
{
  m_result = SendCmd("AT+IFC=%" PRIu32 ",%" PRIu32 "\r", dce_by_dte, dte_by_dce);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Set TE-TA fixed local rate.
    Sends:
    AT+IPR=<baudrate>
    Expects:
    OK
*/
EATCmdResult CATCmd::ATIPR(uint32_t baudrate)
{
  m_result = SendCmd("AT+IPR=%" PRIu32 "\r", baudrate);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    The command delivers the application firmware version number.
    Sends:
    AT+QAPPVER
    Expects:
    <revision>
    OK
*/
EATCmdResult CATCmd::ATQAPPVER()
{
  m_result = SendCmd("AT+QAPPVER\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    The command returns the ICCID (Integrated Circuit Card Identifier) number of the (U)SIM card.
    Sends:
    AT+QCCID
    Expects:
    +QCCID: <iccid>
    OK
*/
EATCmdResult CATCmd::ATQCCID()
{
  m_result = SendCmd("AT+QCCID\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Extended configuration.
    Sends:
    AT+QFCG=cfg
    Expects:
    OK
*/
EATCmdResult CATCmd::ATQCFG(const char* cfg)
{
  m_result = SendCmd("AT+QCFG=%s\r", cfg);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    The command allows the application to check how much bytes
    are sent to or received by the module.
    <bytes_sent> A numeric parameter. The amount of sent bytes.
    <bytes_recv> A numeric parameter. The amount of received bytes.
    Sends:
    AT+QGDCNT?
    Expects:
    +QGDCNT: <bytes_sent>,<bytes_recv>
    OK
*/
EATCmdResult CATCmd::ATQGDCNT_Query()
{
  m_result = SendCmd("AT+QGDCNT?\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Read current GNSS state
    <gnssstate> GNSS state
                0 GNSS OFF
                1 GNSS ON
    Sends:
    AT+QGPS?
    Expects:
    +QGPS: <gnssstate>
    OK
*/
EATCmdResult CATCmd::ATQGPS_Query()
{
  m_result = SendCmd("AT+QGPS?\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Signal Quality Report.
    Sends:
    AT+QCSQ
    Expects:
    +QCSQ: <sysmode>,[,<value1>[,<value2>[,<value3>[,<value4>]]]]
    OK

    <sysmode>       <value1>   <value2>   <value3>   <value4>
    "NOSERVICE"
    "GSM"           <gsm_rssi>
    "CAT-M1"        <lte_rssi> <lte_rsrp> <lte_sinr> <lte_rsrq>
    "CAT-NB1"       <lte_rssi> <lte_rsrp> <lte_sinr> <lte_rsrq>

    <gsm_rssi>,<lte_rssi> An integer indicating the received signal strength.
                          These parameters are available for GSM and LTE mode respectively.
    <lte_rsrp> An integer indicating the reference signal received power (RSRP).
               This parameter is available for LTE mode.
    <lte_sinr> An integer indicating the signal to interference plus noise ratio (SINR).
               Logarithmic value of SINR. Values are in 1/5th of a dB. The range is 0-250 which translates to -20dB - +30dB.
    <lte_rsrq> An integer indicating the reference signal received quality (RSRQ) in dB.
*/
EATCmdResult CATCmd::ATQCSQ()
{
  m_result = SendCmd("AT+QCSQ\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    AT+QENG
    Switch on/off Engineering Mode
    The engineering mode is designed to report the information of serving cells, neighbour cells and Packet
    Switch parameters. The command is used to switch on/off the mode.

    Query the information of serving cell:
    AT+QENG="servingcell"

    Response:

    In the case of GSM mode:
    +QENG: "servingscell",<state>,"GSM",<mcc>,<mnc>,<lac>,
    <cellid>,<bsic>,<arfcn>,<band>,<rxlev>,<txp>,<rla>,<drx>,
    <c1>,<c2>,<gprs>,<tch>,<ts>,<ta>,<maio>,<hsn>,<rxlevsub>,
    <rxlevfull>,<rxqualsub>,<rxqualfull>,<voicecodec>
    OK

    In the case of LTE Cat M1/Cat NB1 mode:
    +QENG: "servingcell",<state>,"rat",<is_tdd>,<mcc>,<mnc>,
    <cellid>,<pcid>,<earfcn>,<freq_band_ind>,<ul_bandwidth>,
    <dl_bandwidth>,<tac>,<rsrp>,<rsrq>,<rssi>,<sinr>,<srxlev>
    OK

    Query the information of neighbour cells:
    AT+QENG="neighbourcell"

    Response:

    In the case of GSM mode:
    [+QENG: "neighbourcell","GSM",<mcc>,<mnc>,<lac>,<cellid>,
    <bsic>,<arfcn>,<rxlev>,<c1>,<c2>,<c31>,<c32>
    [...]]
    OK

    In the case of LTE Cat M1/Cat NB1 mode:
    [+QENG: "neighbourcell intra","rat",<earfcn>,<pcid>,<rsrq>,
    <rsrp>,<rssi>,<sinr>,<srxlev>,<cell_resel_priority>,
    <s_non_intra_search>,<thresh_serving_low>,<s_intra_search>
    [...]]
    [+QENG: "neighbourcell inter","rat",<earfcn>,<pcid>,<rsrq>,
    <rsrp>,<rssi>,<sinr>,<srxlev>,<threshX_low>,<threshX_high>,
    <cell_resel_priority>
    [...]]
    OK
    */
EATCmdResult CATCmd::ATQENG(const char* cell)
{
  m_result = SendCmd("AT+QENG=\"%s\"\r", cell);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Performs a firmware over the air upgrade.
    <httpURL>         String format. The max length is 255 bytes. It should be started with "HTTP://" or
                      "HTTPS://", for example:
                      "HTTP://<http_server_URL>:<http_port>/<http_file_path>".
    <http_server_URL> String type. The IP address or domain name of the HTTP(S) server.
    <http_port>       Integer type. The port of the HTTP(S) server. The default value is 80. The range is
                      1-65535.
    <http_file_path>  String type. The file name in HTTP(S) server.
    <http_err>        Integer type. The HTTP(S) error code. 0 means upgraded successfully. Any other
                      value means an error. Please refer to Chapter 5 for more details.
    <percent>         Integer type. The upgrade progress in percentage.
    <err>             Integer type. 0 means upgraded successfully. Any other value means an error.
                      Please refer to Chapter 5 for more details.

    Sends:
    AT+QFOTADL=<httpURL>
    Expects:
    OK
    If there is any error, response:
    ERROR
    */
EATCmdResult CATCmd::ATQFOTADL(const char* url)
{
  m_result = SendCmd("AT+QFOTADL=\"%s\"\r", url);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*  Function:
    Checks the state of a firmware over the air upgrade,
    after the AT+QFOTADL command has been executed.
    Expects:
    +QIND: "FOTA","HTTPSTART"
    +QIND: "FOTA","HTTPEND",<http_err>
    +QIND: "FOTA","START"
    +QIND: "FOTA","UPDATING",<percent>
    +QIND: "FOTA","UPDATING",<percent>
    ...
    +QIND: "FOTA","RESTORE",<percent>
    +QIND: "FOTA","RESTORE",<percent>
    ...
    +QIND: "FOTA","END",<err>
    If there is any error, response:
    ERROR
*/
EATCmdResult CATCmd::ATQFOTADL_State(uint32_t timeout)
{
  //read a line that starts and ends with \r\n into m_readbuf
  uint32_t readtimeout = m_readtimeout;
  m_readtimeout = timeout;
  m_result = ReadReply(nullptr, "\r\n");
  m_readtimeout = readtimeout;

  return m_result;
}

/*! Function:
    Turn on GNSS on the Quectel BG96.
    <gnssmode>   GNSS working mode
                 1 Stand-alone
                 2 MS-based
                 3 MS-assisted
                 4 Speed-optimal
    <fixmaxtime> The maximum positioning time (unit: s). which indicate the response time of
                 GNSS receiver while measuring the GNSS pseudo range, and the upper time limit
                 of GNSS satellite searching. It also includes the time for demodulating the
                 ephemeris data and calculating the position.
                 1-30-255 Maximum positioning time
    <fixmaxdist> Accuracy threshold of positioning. Unit: m.
                 1-50-1000
    <fixcount>   Number of attempts for positioning.
                 0–1000 0 indicates continuous positioning. Non-zero values indicate the
                          actual number of attempts for positioning.
    <fixrate>    The interval time between the first and second time positioning. Unit: s.
                 1–65535
    Sends:
    AT+QGPS=<gnssmode>[,<fixmaxtime>[,<fixmaxdist>[,<fixcount>[,<fixrate>]]]]
    Expects:
    OK
*/
EATCmdResult CATCmd::ATQGPS_Set_BG96(uint32_t gnssmode, uint32_t* fixmaxtime /*= nullptr*/, uint32_t* fixmaxdist /*= nullptr*/,
                                     uint32_t* fixcount /*= nullptr*/, uint32_t* fixrate /*= nullptr*/)
{
  if (fixmaxtime == nullptr)
    m_result = SendCmd("AT+QGPS=%" PRIu32 "\r", gnssmode);
  else if (fixmaxdist == nullptr)
    m_result = SendCmd("AT+QGPS=%" PRIu32 ",%" PRIu32 "\r", gnssmode, *fixmaxtime);
  else if (fixcount == nullptr)
    m_result = SendCmd("AT+QGPS=%" PRIu32 ",%" PRIu32 ",%" PRIu32 "\r", gnssmode, *fixmaxtime, *fixmaxdist);
  else if (fixrate == nullptr)
    m_result = SendCmd("AT+QGPS=%" PRIu32 ",%" PRIu32 ",%" PRIu32 ",%" PRIu32 "\r", gnssmode, *fixmaxtime, *fixmaxdist, *fixrate);
  else
    m_result = SendCmd("AT+QGPS=%" PRIu32 ",%" PRIu32 ",%" PRIu32 ",%" PRIu32 ",%" PRIu32 "\r",
                       gnssmode, *fixmaxtime, *fixmaxdist, *fixcount, *fixrate);

  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Turn on GNSS on the Quectel BG95.
    <GNSS_mode> Integer type. GNSS working mode.
                1 Stand-alone mode
    <accuracy>  Integer type. The desired level of accuracy that is acceptable for fix computation.
                1 Low Accuracy (1000 m)
                2 Medium Accuracy (500 m)
                3 High Accuracy (50 m)
    <fix_count> Integer type. Number of attempts for positioning or continuous positioning.
                Range: 0–1000. Default value: 0.
                0 indicates continuous positioning. Other values indicate the number of attempts for
                positioning. When the value reaches the specified number of attempts, the GNSS
                will be stopped.
    <fix_rate>  The interval time between the first- and second-time positioning. Unit: second.
                If <fix_rate> < 1, it is a float type. Available options:
                0.1
                0.2
                0.5
                If <fix_rate> ≥ 1, it is an integer type.
                Range: 1–65535. Default value: 1.
    <HEPE>      Integer type. Accuracy threshold. GNSS NMEA sentences will not be outputted if the
                accuracy is less than <HEPE>. 0 means disable this function. Range: 0–1000.
                Default value: 0. Unit: meter.    Sends:
    AT+QGPS=<GNSS_mode>[,<accuracy>[,<fix_count>[,<fix_rate>[,<HEPE>]]]]
    Expects:
    OK
*/
EATCmdResult CATCmd::ATQGPS_Set_BG95(uint32_t gnssmode,
                                     uint32_t* accuracy /*= nullptr*/,
                                     uint32_t* fixcount /*= nullptr*/,
                                     uint32_t* fixrate /*= nullptr*/,
                                     uint32_t* hepe /*= nullptr*/)
{
  if (accuracy == nullptr)
    m_result = SendCmd("AT+QGPS=%" PRIu32 "\r", gnssmode);
  else if (fixcount == nullptr)
    m_result = SendCmd("AT+QGPS=%" PRIu32 ",%" PRIu32 "\r", gnssmode, *accuracy);
  else if (fixrate == nullptr)
    m_result = SendCmd("AT+QGPS=%" PRIu32 ",%" PRIu32 ",%" PRIu32 "\r", gnssmode, *accuracy, *fixcount);
  else if (hepe == nullptr)
    m_result = SendCmd("AT+QGPS=%" PRIu32 ",%" PRIu32 ",%" PRIu32 ",%" PRIu32 "\r", gnssmode, *accuracy, *fixcount, *fixrate);
  else
    m_result = SendCmd("AT+QGPS=%" PRIu32 ",%" PRIu32 ",%" PRIu32 ",%" PRIu32 ",%" PRIu32 "\r",
                       gnssmode, *accuracy, *fixcount, *fixrate, *hepe);

  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Read current GNSS configuration
    Settings:
      "outport",("none","usbnmea","uartnmea")
      "nmeasrc",(0,1)
      "gpsnmeatype",(0-31)
      "glonassnmeatype",(0-7)
      "galileonmeatype",(0,1)
      "beidounmeatype",(0-3)
      "gsvextnmeatype",(0,1)
      "gnssconfig",(0-6)
      "autogps",(0,1)
    Sends:
    AT+QGPSCFG="<setting>"
    Expects:
    +QGPSCFG: "<setting>",<value>
    OK
*/
EATCmdResult CATCmd::ATQGPSCFG_Query(const char* setting)
{
  m_result = SendCmd("AT+QGPSCFG=\"%s\"\r", setting);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Extended GPS configuration.
    Sends:
    AT+QGPSCFG=cfg
    Expects:
    OK
*/
EATCmdResult CATCmd::ATQGPSCFG_Set(const char* cfg)
{
  m_result = SendCmd("AT+QGPSCFG=%s\r", cfg);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Delete Assistance Data
    Sends:
    AT+QGPSDEL=<deletetype>
    Expects:
    OK
*/
EATCmdResult CATCmd::ATQGPSDEL(uint32_t deletetype)
{
  m_result = SendCmd("AT+QGPSDEL=%" PRIu32 "\r", deletetype);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Acquire positioning information.
    <mode>      Latitude and longitude display format.
                0 <latitude>,<longitude> format: ddmm.mmmm N/S,dddmm.mmmm E/W
                1 <latitude>,<longitude> format: ddmm.mmmmmm N/S,dddmm.mmmmmm E/W
                2 <latitude>,<longitude> format: (-)dd.ddddd,(-)ddd.ddddd
    <UTC>       UTC time.
                Format: hhmmss.sss (Quoted from GPGGA sentence).
    <latitude>  Latitude.
                If <mode> is 0:
                  Format: ddmm.mmmm N/S (Quoted from GPGGA sentence)
                  dd      00-89 (degree)
                  mm.mmmm 00.0000-59.9999 (minute)
                  N/S     North latitude/South latitude
                If <mode> is 1:
                  Format: ddmm.mmmmmm N/S (Quoted from GPGGA sentence)
                  dd        00-89 (degree)
                  mm.mmmmmm 00.000000-59.999999 (minute)
                  N/S North latitude/South latitude
                If <mode> is 2:
                  Format: (-)dd.ddddd (Quoted from GPGGA sentence)
                  dd.ddddd -89.99999-89.99999 (degree)
                  - South latitude
    <longitude> Longitude.
                If <mode> is 0:
                  Format: dddmm.mmmm E/W (Quoted from GPGGA sentence)
                  ddd     000-179 (degree)
                  mm.mmmm 00.0000-59.9999 (minute)
                  E/W     East longitude/West longitude
                If <mode> is 1:
                  Format: dddmm.mmmmmm E/W (Quoted from GPGGA sentence)
                  ddd       000-179 (degree)
                  mm.mmmmmm 00.000000-59.999999 (minute)
                  E/W       East longitude/West longitude
                If <mode> is 2:
                  Format: (-)dd.ddddd Quoted from GPGGA sentence)
                  dd.ddddd -179.99999-179.99999 (degree)
                  -        West longitude
    <hdop>      Horizontal precision: 0.5-99.9 (Quoted from GPGGA sentence).
    <altitude>  The altitude of the antenna away from the sea level (unit: m), accurate to one decimal
                place (Quoted from GPGGA sentence)
    <fix>       GNSS positioning mode (Quoted from GNGSA/GPGSA sentence).
                2 2D positioning
                3 3D positioning
    <cog>       Course Over Ground based on true north.
                Format: ddd.mm (Quoted from GPVTG sentence).
                ddd 000-359 (degree)
                mm  00-59 (minute)
    <spkm>      Speed over ground.
                Format: xxxx.x; unit: Km/h; accurate to one decimal place (Quoted from GPVTG sentence).
    <spkn>      Speed over ground.
                Format: xxxx.x; unit: knots; accurate to one decimal place (Quoted from GPVTG sentence).
    <date>      UTC time when fixing position.
                Format: ddmmyy (Quoted from GPRMC sentence).
    <nsat>      Number of satellites, from 00 (The first 0 should be retained) to 12 (Quoted from
                GPGGA sentence).
    Sends:
    AT+QGPSLOC=<mode>
    Expects:
    +QGPSLOC: <UTC>,<latitude>,<longitude>,<hdop>,<altitude>,<fix>,<cog>,<spkm>,<spkn>,<date>,<nsat>
    OK
*/
EATCmdResult CATCmd::ATQGPSLOC(uint32_t mode)
{
  m_result = SendCmd("AT+QGPSLOC=%" PRIu32 "\r", mode);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*
    This command turns off GNSS. It is used when GNSS is turned on and continuously fixes position
    (<fix_count> is 0). In such a case, GNSS can be turned off with AT+QGPSEND.

    Sends:
    AT+QGPSEND
    Expects:
    OK
*/
EATCmdResult CATCmd::ATQGPSEND()
{
  m_result = SendCmd("AT+QGPSEND\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! This command enables XTRA function.
    <XTRA_enable> Integer type. Enable/disable XTRA feature.
                  0 Disable
                  1 Enable */
EATCmdResult CATCmd::ATQGPSXTRA(uint32_t enable)
{
  m_result = SendCmd("AT+QGPSXTRA=%" PRIu32 "\r", enable);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Configure Parameters for HTTP(S).
    <contextID>       Numeric type. PDP context ID. The range is 1-16, and the default value is 1.
    <request_header>  Numeric type. Disable or enable to customize HTTP(S) request header.
                      0 Disable
                      1 Enable
    <response_header> Numeric type. Disable or enable to output HTTP(S) response header.
                      0 Disable
                      1 Enable
    <sslctxID>        Numeric type. SSL context ID used for HTTP(S). The range is 0-5, and the
                      default value is 1. Customers should configure the SSL parameters by
                      AT+QSSLCFG. For details, please refer to Quectel_BG96_SSL_AT_
                      Commands_Manual.
    <content_type>    Numeric type. Data type of HTTP(S) body.
                      0 application/x-www-form-urlencoded
                      1 text/plain
                      2 application/octet-stream
                      3 multipart/form-data
    Sends:
    AT+QHTTPCFG="parameter",<value>
    Expects:
    OK
*/
EATCmdResult CATCmd::ATQHTTPCFG(const char* parameter, uint32_t value)
{
  m_result = SendCmd("AT+QHTTPCFG=\"%s\",%" PRIu32 "\r", parameter, value);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Send GET Request to HTTP(S) Server.
    <rsptime>  Numeric type. The range is 1-65535, and the default value is 60. Unit: second. It is
               used  to  configure  the  timeout  for  the  HTTP(S)  GET  response  “+QHTTPGET:
               <err>[,<httprspcode>,<content_length>]” to be outputted after “OK” is returned. 
    Sends:
    AT+QHTTPGET=<rsptime>
    Expects:
    OK
    +QHTTPGET: <err>,[<httprspcode>,[<content_length>]]
*/
EATCmdResult CATCmd::ATQHTTPGET(uint32_t rsptime)
{
  m_result = SendCmd("AT+QHTTPGET=%" PRIu32 "\r", rsptime);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = CheckOKReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  //after the \r\nOK\r\n is received, another line will be sent when the HTTP request has been done
  //it looks something like:
  //\r\n+QHTTPGET: 0,200,1270\r\n
  //so use ReadReply() with custom end string \r\n to read this
  uint32_t readtimeout = m_readtimeout;
  m_readtimeout = rsptime * 1500;
  m_result = ReadReply(nullptr, "\r\n");
  m_readtimeout = readtimeout;

  return m_result;
}

/*! Function:
    Send POST Request to HTTP(S) Server via UART/USB.

    <data_length> Numeric type. If <request_header> is 0, it indicates the length of post body, 
                  and if <request_header> is 1, it indicates the length of HTTP(S) request
                  information, including HTTP(S) request header and HTTP(S) request body. 
    <input_time>  Numeric type. The maximum time for inputting post body or HTTP(S) request
                  information. The range is 1-65535, and the default value is 60. Unit: second.
    <rsptime>     Numeric  type.  The  range  is  1-65535,  and  the  default  value  is  60.  Unit:
                  second. It is used to configure the timeout for the HTTP(S) POST response
                  "+QHTTPPOST:  <err>[,<httprspcode>[,<content_length>]]" to be outputted
                  after "OK" is returned. 
    Sends:
    AT+QHTTPPOST=<data_length>,<input_time>,<rsptime>
    */
EATCmdResult CATCmd::ATQHTTPPOST(uint32_t data_length, uint32_t input_time /*= 60*/, uint32_t rsptime /*= 60*/)
{
  //send AT+QHTTPPOST which starts the HTTP POST
  m_result = SendCmd("AT+QHTTPPOST=%" PRIu32 ",%" PRIu32 ",%" PRIu32 "\r", data_length, input_time, rsptime);
  if (m_result != ATCmdResult_Success)
    return m_result;

  //read the reply, which should be \r\nCONNECT\r\n
  //use a custom read timeout which is 1.5 times the response time,
  //the modem should have responded within that time
  uint32_t readtimeout = m_readtimeout;
  m_readtimeout = rsptime * 1500;
  m_result = ReadReply();
  m_readtimeout = readtimeout;
  if (m_result != ATCmdResult_Success)
    return m_result;

  //wait for CONNECT, which means a TCP connection to the HTTP server has been made
  m_result = CheckConnectReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  //at this point, the POST body can be sent to the modem,
  //which means that data_length bytes should be written directly to the usart

  return m_result;
}

/*! The modem ends the AT+QHTTPPOST with \r\nOK\r\n+QHTTPPOST: <err>[,<httprspcode>[,<content_length>]]\r\n
    After the POST data has been send to the modem.
    \param timeout The timeout in seconds to wait for the modem reply.*/
EATCmdResult CATCmd::ATQHTTPPOST_Finish(uint32_t timeout)
{
  uint32_t readtimeout = m_readtimeout;
  m_readtimeout = timeout * 1500;

  //wait for \r\nOK\r\n from the modem
  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
  {
    m_readtimeout = readtimeout;
    return m_result;
  }

  m_result = CheckOKReply();
  if (m_result != ATCmdResult_Success)
  {
    m_readtimeout = readtimeout;
    return m_result;
  }

  //after the \r\nOK\r\n is received, another line will be sent when the HTTP request has been done
  //it looks something like:
  //\r\n+QHTTPPOST: 0,200,1270\r\n
  //so use ReadReply() with custom end string \r\n to read this
  m_result = ReadReply(nullptr, "\r\n");
  if (m_result != ATCmdResult_Success)
  {
    m_readtimeout = readtimeout;
    return m_result;
  }

  m_readtimeout = readtimeout;

  return m_result;
}

/*! The modem ends the http read with \r\nOK\r\n+QHTTPREAD: <err>\r\n
    where <err> is a numerical result code.*/
EATCmdResult CATCmd::ATQHTTPREAD_Finish()
{
  //wait for \r\nOK\r\n from the modem
  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = CheckOKReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  //wait for \r\n+QHTTPREAD: from the modem
  m_result = ReadReply(nullptr, "\r\n+QHTTPREAD: ");
  if (m_result != ATCmdResult_Success)
    return m_result;

  //read the result code and the terminating \r\n
  return m_result = ReadReply("", "\r\n");
}

/*! Function:
    Read Response from HTTP(S) Server via UART/USB.
    <wait_time> Numeric type. The maximum interval time between receiving two packets of data. 
                The range is 1-65535, and the default value is 60. Unit: second. 
    <err>       Integer type. The error code of the operation.
    Sends:
    AT+QHTTPREAD[=<wait_time>] 
    Expects:
    CONNECT
    <data>
    It will return after CONNECT is received, further data can be read with ReadUsart(),
    after all data is read, the command should be finished with ATQHTTPREAD_Finish();
*/
EATCmdResult CATCmd::ATQHTTPREAD_Start(uint32_t wait_time)
{
  m_result = SendCmd("AT+QHTTPREAD=%" PRIu32 "\r", wait_time);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckConnectReply();
}

/*! Function:
    Set HTTP URL.
    <URL_length> Numeric type. The length of URL. The range is 1-700. Unit: byte.
    <timeout>    Numeric type. The maximum time for inputting URL. The range is 1-65535, and
                 the default value is 60. Unit: second.
    Sends:
    AT+QHTTPURL=<URL_length>
    Expects:
    CONNECT
    Sends:
    <URL>
    Expects:
    OK
*/
EATCmdResult CATCmd::ATQHTTPURL(const char* url)
{
  //send the length of the URL
  m_result = SendCmd("AT+QHTTPURL=%" PRIu32 "\r", (uint32_t)strlen(url));
  if (m_result != ATCmdResult_Success)
    return m_result;

  //instead of OK, this command replies with CONNECT
  m_result = ReadReply(nullptr, "\r\nCONNECT\r\n");
  if (m_result != ATCmdResult_Success)
    return m_result;

  //the modem has now switched to transparent mode, meaning it
  //will read the number of bytes previously entered wit AT+QHTTPURL
  //and it will not echo the bytes, so m_echotimeout is set to 0 for
  //the next call to SendCmd(), to disable the check for the echo
  uint32_t echotimeout = m_echotimeout;
  m_echotimeout = 0;
  m_result = SendCmd("%s", url);
  m_echotimeout = echotimeout;
  if (m_result != ATCmdResult_Success)
    return m_result;

  //read the OK reply
  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Return the list of the current activated contexts and their IP addresses.
    <contextID>     Integer type. The context ID. The range is 1-16.
    <context_state> Integer type. The context state.
                    0 Deactivated
                    1 Activated
    <context_type>  Integer type. The context state.
                    1 IPV4
                    2 IPV6
    <IP_address>    The local IP address after the context is activated.
    Sends:
    AT+QIACT?
    Expects:
    +QIACT:
    <contextID>,<context_state>,<context_type>[,<IP_address>]
    OK
*/
EATCmdResult CATCmd::ATQIACT_Query()
{
  m_result = SendCmd("AT+QIACT?\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Activate a PDP context.
    Sends:
    AT+QIACT=<contextID>
    Expects:
    OK
*/
EATCmdResult CATCmd::ATQIACT_Set(uint32_t contextid)
{
  m_result = SendCmd("AT+QIACT=%" PRIu32 "\r", contextid);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Query the configuration of a TCP/IP context.
    Sends:
    AT+QICSGP=<contextID>
    Expects:
    +QICSGP: <context_type>,<APN>,<username>,<password>,<authentication>
    OK
*/
EATCmdResult CATCmd::ATQICSGP_Query(uint32_t contextid)
{
  m_result = SendCmd("AT+QICSGP=%" PRIu32 "\r", contextid);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

EATCmdResult CATCmd::ATQICSGP_Set(uint32_t contextid, const uint32_t* context_type /*= nullptr*/, const char* apn /*= nullptr*/,
                                  const char* username /*= nullptr*/, const char* password /*= nullptr*/, const uint32_t* authentication /*= nullptr*/)
{
  //if context_type is passed, apn should also be passed
  assert((context_type == nullptr && apn == nullptr) || (context_type != nullptr && apn != nullptr));
  //if username is passed, password should also be passed
  assert((username == nullptr && password == nullptr) || (username != nullptr && password != nullptr));

  if (context_type == nullptr)
    m_result = SendCmd("AT+QICSGP=%" PRIu32 "\r", contextid);
  else if (username == nullptr)
    m_result = SendCmd("AT+QICSGP=%" PRIu32 ",%" PRIu32 ",\"%s\"\r", contextid, *context_type, apn);
  else if (authentication == nullptr)
    m_result = SendCmd("AT+QICSGP=%" PRIu32 ",%" PRIu32 ",\"%s\",\"%s\",\"%s\"\r", contextid, *context_type, apn, username, password);
  else
    m_result = SendCmd("AT+QICSGP=%" PRIu32 ",%" PRIu32 ",\"%s\",\"%s\",\"%s\",%" PRIu32 "\r",
                       contextid, *context_type, apn, username, password, *authentication);

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Deactivate a PDP Context
    <contextID> Integer type. The context ID. The range is 1-16.
    Sends:
    AT+QIDEACT=<contextID>
    Expects:
    OK
*/
EATCmdResult CATCmd::ATQIDEACT(uint32_t contextid)
{
  m_result = SendCmd("AT+QIDEACT=%" PRIu32 "\r", contextid);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Query initialization status of sim card.
    status:
      Initialization status of (U)SIM card. Actual value is the sum of several of the following four
      kinds (e.g. 3=1+2 means CPIN READY & SMS DONE).
      0 Initial state
      1 CPIN READY. Operation like lock/unlock PIN is allowed
      2 SMS initialization completed
    Sends:
    AT+QINISTAT
    Expects:
    +QINISTAT: <status>
    OK
*/
EATCmdResult CATCmd::ATQINISTAT()
{
  m_result = SendCmd("AT+QINISTAT\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    The command indicates network information such as the access technology selected, the operator, and
    the band selected.

    <Act>     String type. Access technology selected.
              "NONE"
              "GSM"
              "GPRS"
              "EDGE"
              "Cat-M1"
              "Cat-NB1"

    <oper>    String type. Operator in numeric format.

    <band>    String type. Band selected.
              "GSM 850"
              "GSM 900"
              "GSM 1800"
              "GSM 1900"
              "LTE BAND 1" – "LTE BAND 43"

    <channel> Integer type. Channel ID.

    Sends:
    AT+QNWINFO
    Expects:
    +QNWINFO: <Act>,<oper>,<band>,<channel>
    OK
*/
EATCmdResult CATCmd::ATQNWINFO()
{
  m_result = SendCmd("AT+QNWINFO\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
   AT+QNWLOCK="common/4g" can lock a sepcial cell of 4g.

   Sends:
   AT+QNWLOCK="common/4g",[action],[earfcn],[pci]
   Expects:
   +QNWLOCK:"common/4g",<action>,<earfcn>,<pci>,<status>
   OK
*/
EATCmdResult CATCmd::ATQNWLOCK(uint32_t* action /*= nullptr*/, uint32_t* earfcn /*= nullptr*/, uint32_t* pci /*= nullptr*/)
{
  if (action == nullptr)
    m_result = SendCmd("AT+QNWLOCK=\"common/4g\"\r");
  else if (earfcn == nullptr)
    m_result = SendCmd("AT+QNWLOCK=\"common/4g\",%" PRIu32 "\r", *action);
  else if (pci == nullptr)
    m_result = SendCmd("AT+QNWLOCK=\"common/4g\",%" PRIu32 ",%" PRIu32 "\r", *action, *earfcn);
  else
    m_result = SendCmd("AT+QNWLOCK=\"common/4g\",%" PRIu32 ",%" PRIu32 ",%" PRIu32 "\r", *action, *earfcn, *pci);

  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Display PIN remainder counter
    <facility>   "SC" (U)SIM PIN
                 "P2" (U)SIM PIN2
    <pincounter> Number of attempts left to enter the password of PIN.
    <pukcounter> Number of attempts left to enter the password of PUK.
    Sends:
    AT+QPINC=<facility>
    Expects:
    +QPINC: <facility>,<pincounter>,<pukcounter>
    OK
*/
EATCmdResult CATCmd::ATQPINC(const char* facility)
{
  m_result = SendCmd("AT+QPINC=\"%s\"\r", facility);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Power down.
    powerstate:
      0 Immediately power down
      1 Normal power down
    Sends:
    AT+QPOWD=<powerstate>
    Expects:
    OK
*/
#define POWERED_DOWN_TIMEOUT 60000
EATCmdResult CATCmd::ATQPOWD(uint32_t powerstate)
{
  m_result = SendCmd("AT+QPOWD=%" PRIu32 "\r", powerstate);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = CheckOKReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  //wait for \r\nPOWERED DOWN\r\n from the modem for 60 seconds
  m_usartresult = m_usart.ReadModemReply(m_readbuf, AT_BUFSIZE, &m_readlen, POWERED_DOWN_TIMEOUT, nullptr, "\r\nPOWERED DOWN\r\n");
  return m_result = TranslateUsartResult(m_usartresult);
}

/*! Function:
    Restore backup configuration.
    Sends:
    AT+QPRTPARA=3
    Expects:
    OK
*/
EATCmdResult CATCmd::ATQPRTPARA()
{
  m_result = SendCmd("AT+QPRTPARA=3\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    The Write Commands force the module to transmit in FTM (factory test mode).
    <band>      String type. Supported bands in GSM/LTE. The possible values are:
                For GSM:
                "GSM850"
                "GSM900"
                "GSM1800"
                "GSM1900"
                For LTE:
                "LTE BAND1"
                "LTE BAND2"
                "LTE BAND3"
                "LTE BAND4"
                "LTE BAND5"
                "LTE BAND8"
                "LTE BAND12"
                "LTE BAND13"
                "LTE BAND18"
                "LTE BAND19"
                "LTE BAND20"
                "LTE BAND25"
                "LTE BAND26"
                "LTE BAND27"
                "LTE BAND28"
                "LTE BAND31"
                "LTE BAND66"
                "LTE BAND71"
                "LTE BAND72"
                "LTE BAND73"
                "LTE BAND85"
    <channel>   Integer type. Supported uplink channels in GSM/LTE. The corresponding channels for
                different bands in GSM/LTE are as follows:
                GSM band   Uplink Channels
                GSM850     128–251
                GSM900     1–124, 975–1023
                GSM1800    512–885
                GSM1900    512–810
                LTE band   Uplink Channels
                LTE BAND1  18000–18599
                LTE BAND2  18600–19199
                LTE BAND3  19200–19949
                LTE BAND4  19950-20399
                LTE BAND5  20400–20649
                LTE BAND8  21450–21799
                LTE BAND12 23010–23179
                LTE BAND13 23180–23279
                LTE BAND18 23850–23999
                LTE BAND19 24000–24149
                LTE BAND20 24150–24449
                LTE BAND25 26040–26689
                LTE BAND26 26690–27039
                LTE BAND27 27040–27209
                LTE BAND28 27210–27659
                LTE BAND31 27760–27809
                LTE BAND66 131972–132671
                LTE BAND71 131122–133471
                LTE BAND72 133472–133521
                LTE BAND73 133522–133571
                LTE BAND85 134231–134280
    <tx_enable> String type. Enable/disable RF TX.
                "ON"  Enable RF TX
                "OFF" Disable RF TX
    <tx_burst>  Integer type.
                0 Continuous TX mode
    <rgi>       Integer type. LTE power level. Range: 0–100.
    <tx_gain>   Integer type. GSM power level (GSM power in dBm × 100). Range: 0–3300.
    <waveform>  Integer type.
                1 LTE modulated TX mode

    Sends:
    AT+QRFTEST=<band>,<channel>,<tx_enable>,<tx_burst>,<tx_gain>
    or:
    AT+QRFTEST=<band>,<channel>,<tx_enable>,<rgi>,<waveform>
    Expects:
    OK
*/
EATCmdResult CATCmd::ATQRFTEST(const char* band, uint32_t channel, bool tx_enable,
                               uint32_t tx_burst_or_rgi, uint32_t tx_gain_or_waveform)
{
  m_result = SendCmd("AT+QRFTEST=\"%s\",%" PRIu32 ",\"%s\",%" PRIu32 ",%" PRIu32 "\r",
                     band, channel, tx_enable ? "ON" : "OFF",
                     tx_burst_or_rgi, tx_gain_or_waveform);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    The Write Command makes the module enter/exit FTM (RF test mode).
    Parameter

    <mode> Integer type. Enter/exit FTM.
           0 Exit FTM
           1 Enter FTM

    Sends:
    QRFTESTMODE=<mode>
    Expects:
    OK
*/
EATCmdResult CATCmd::ATQRFTESTMODE(uint32_t mode)
{
  m_result = SendCmd("AT+QRFTESTMODE=%" PRIu32 "\r", mode);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    The command enables (U)SIM card hot-swap function. (U)SIM card is detected by GPIO interrupt. The
    level of (U)SIM card detection pin should also be set when the (U)SIM card is inserted. This command
    can be saved by AT&W.

    <enable>      Enable or disable (U)SIM card detection
                  0 Disable
                  1 Enable

    <insertlevel> The level of (U)SIM card detection pin when a (U)SIM card is inserted
                  0 Low level
                  1 High level
    Sends:
    AT+QSIMDET=<enable>,<insertlevel>
    Expects:
    OK
*/
EATCmdResult CATCmd::ATQSIMDET(uint32_t enable, uint32_t insertlevel)
{
  m_result = SendCmd("AT+QSIMDET=%" PRIu32 ",%" PRIu32 "\r", enable, insertlevel);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Display the Name of Registered Network.
    <SPN>      Service provider name
    <alphabet> Alphabet of full network name and short network name
               0 GSM 7 bit default alphabet
               1 UCS2
    <RPLMN>    Registered PLMN
    <FNN>      Full network name
    <SNN>      Short network name
    Sends:
    AT+QSPN
    Expects:
    +QSPN: <FNN>,<SNN>,<SPN>,<alphabet>,<RPLMN>
    OK
*/
EATCmdResult CATCmd::ATQSPN()
{
  m_result = SendCmd("AT+QSPN\r");
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Configure parameters of an SSL context.
    <sslctxID>         Numeric type. SSL context ID. The range is 0-5.
    <sslversion>       Numeric type. SSL Version.
                       0 SSL3.0
                       1 TLS1.0
                       2 TLS1.1
                       3 TLS1.2
                       4 All
    <ciphersuite>      Numeric type. SSL cipher suites.
                       0X0035 TLS_RSA_WITH_AES_256_CBC_SHA
                       0X002F TLS_RSA_WITH_AES_128_CBC_SHA
                       0X0005 TLS_RSA_WITH_RC4_128_SHA
                       0X0004 TLS_RSA_WITH_RC4_128_MD5
                       0X000A TLS_RSA_WITH_3DES_EDE_CBC_SHA
                       0X003D TLS_RSA_WITH_AES_256_CBC_SHA256
                       0XC011 TLS_ECDHE_RSA_WITH_RC4_128_SHA
                       0XC012 TLS_ECDHE_RSA_WITH_3DES_EDE_CBC_SHA
                       0XC013 TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA
                       0XC014 TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA
                       0XC027 TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256
                       0XC028 TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA384
                       0XC02F TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256
                       0XFFFF Support all
    <ignorelocaltime>  Numeric format. How to deal with expired certificate.
                       0 Care about validity check for certification
                       1 Ignore validity check for certification
    <cacert>           String format. The path of the trusted CA certificate.
    <clientcert>       String format. The path of the client certificate.
    <clientkey>        String format, the path of the client private key.
    <seclevel>         Numeric format. The authentication mode.
                       0 No authentication
                       1 Manage server authentication
                       2 Manage server and client authentication if requested by the remote server.
    <negotiatetime>    Numeric format. Indicates maximum timeout used in SSL negotiation stage. The
                       value range is 10-300, and the default value is 300. Unit: second.
    Sends:
    AT+QSSLCFG="<parameter>",<sslctxID>,"<value>"
    Expects:
    OK
*/
EATCmdResult CATCmd::ATQSSLCFG(const char* parameter, uint32_t sslctxID, const char* value)
{
  m_result = SendCmd("AT+QSSLCFG=\"%s\",%" PRIu32 ",\"%s\"\r", parameter, sslctxID, value);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*this function is explained in the overload above
    Sends:
    AT+QSSLCFG="<parameter>",<sslctxID>,<value>
    Expects:
    OK
*/
EATCmdResult CATCmd::ATQSSLCFG(const char* parameter, uint32_t sslctxID, uint32_t value)
{
  m_result = SendCmd("AT+QSSLCFG=\"%s\",%" PRIu32 ",%" PRIu32 "\r", parameter, sslctxID, value);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}

/*! Function:
    Receive data via SSL connection.
    <clientID>    Numeric type. Socket index. The range is 0-11.
    <readlen>     Numeric type. The length of data to be retrieved. The range is 1-1500. Unit: byte.
    <havereadlen> Numeric type. The actual data length obtained by AT+QSSLRECV. Unit: byte.
    <data>        The retrieved data.
    Sends:
    AT+QSSLRECV=<clientID>,<readlen>
    Expects:
    +QSSLRECV: <havereadlen><CR><LF><data>
    OK
*/
EATCmdResult CATCmd::ATQSSLRECV(uint32_t clientid, uint32_t readlen)
{
  m_result = SendCmd("AT+QSSLRECV=%" PRIu32 ",%" PRIu32 "\r", clientid, readlen);
  if (m_result != ATCmdResult_Success)
    return m_result;

  m_result = ReadReply();
  if (m_result != ATCmdResult_Success)
    return m_result;

  return m_result = CheckOKReply();
}
