#include "batterycharger.h"
#include "LTC4162FRegisters.h"
#include "lock.h"
#include "log.h"
#include "clock.h"
#include <algorithm>

CBatteryCharger::CBatteryCharger()
{
  m_mutex = nullptr;
  m_i2c   = nullptr;
  m_addr  = 0;
}

CBatteryCharger::~CBatteryCharger()
{
  if (m_mutex)
    vSemaphoreDelete(m_mutex);
}

/*! Sets up pointers and the address, and checks if the LTC4162-F battery charger
    responds at the I2C address.
    \param i2c Pointer to an instance of CI2C.
    \param addr The I2C address the LTC4162-F battery charger is configured to.*/
bool CBatteryCharger::Setup(CI2C* i2c, uint8_t addr)
{
  if (m_mutex == nullptr)
    m_mutex = xSemaphoreCreateRecursiveMutex();

  m_i2c  = i2c;
  m_addr = addr;

  CLock lock(*m_i2c);

  //start by sending a stop condition, to reset all communication
  m_i2c->Stop();

  //test if there is an I2C device at the correct address
  if (m_i2c->Write(m_addr, nullptr, 0))
  {
    PrintLog(LOGBTTRYCHRGR, "Battery charger responded at address %"
             PRIu8 "\n", m_addr);
    m_i2c->Stop();
  }
  else
  {
    PrintLog(LOGBTTRYCHRGR, "ERROR: battery charger is not responding at address %"
             PRIu8 "\n", m_addr);
    m_i2c->Stop();
    return false;
  }

  //Check if the charger is a LiFePO type with correct cell count.
  CheckChemistry();

  return true;
}

/*! Reads a register from the LTC4162-F battery charger.
    \param reg The register to read.
    \param value The register value is stored here after a successful read.
    \return true if reading the register succeeded.
    \return false if reading the register failed.*/
bool CBatteryCharger::ReadRegister(uint8_t reg, uint16_t* value)
{
  CLock lock(*m_i2c);

  //first write the register address to the LTC4162-F
  if (!m_i2c->Write(m_addr, &reg, 1))
  {
    PrintLog(LOGBTTRYCHRGR, "ERROR: Unable to write to battery charger, "
             "failed to ack at byte %" PRIu32 "\n", m_i2c->AckFail());
    m_i2c->Stop();
    return false;
  }

  //read from the LTC4162-F, its registers are 2 bytes wide,
  uint8_t reply[2];
  if (!m_i2c->Read(m_addr, reply, 2))
  {
    PrintLog(LOGBTTRYCHRGR, "ERROR: Unable to read from battery charger, "
             "failed to ack at byte %" PRIu32 "\n", m_i2c->AckFail());
    m_i2c->Stop();
    return false;
  }
  m_i2c->Stop();

  //convert to uint16_t, the LTC4162-F returns its registers as little endian
  *value = ((uint16_t)reply[1] << 8) | (uint16_t)reply[0];

  return true;
}

/*! Writes a register on the LTC4162-F battery charger.
    \param reg The register to write.
    \pararm value The value to write to the register.
    \return true if writing the register succeeded.
    \return false if writing the register failed.*/
bool CBatteryCharger::WriteRegister(uint8_t reg, uint16_t value)
{
  CLock lock(*m_i2c);

  //write the register address and the register value
  //this sets the register to the wanted value
  uint8_t regval[2] = {(uint8_t)value, (uint8_t)(value >> 8)};
  if (!m_i2c->Write(m_addr, &reg, 1, regval, 2))
  {
    PrintLog(LOGBTTRYCHRGR, "ERROR: Unable to write to battery charger, "
             "failed to ack at byte %" PRIu32 "\n", m_i2c->AckFail());
    m_i2c->Stop();
    return false;
  }
  m_i2c->Stop();

  return true;
}

/*! Enables forced telemetry readings, and reads the battery voltage from the
    vbat register. Telemetry readings can take about 25 milliseconds to become
    valid when the battery is not being charged. If SetForceTelemetry(true) is
    called before calling this function, other tasks can be performed while
    waiting for the telemetry readings to become valid.
    If low power mode is used, SetForceTelemetry(false) should be called before
    putting the microcontroller to sleep, to reduce idle current.
    \param voltage When not nullptr, the read voltage is returned here.
    \param print When true, the voltage is printed to the log.
    \param input When not nullptr, the charger input voltage is returned here.
    \param charging When not nullptr, if the battery is currently charging is returned here.
    \pararm batterycurrent When not nullptr, the battery charge current is returned here.
    \param timeout Timeout in milliseconds waiting for the telemetry to become valid.
    \return true if reading the voltage succeeded.
    \return false on any error.*/
bool CBatteryCharger::ReadBatteryVoltage(float* voltage, bool print /*= false*/, float* input /*= nullptr*/,
                                         bool* charging /*= nullptr*/, float* batterycurrent /*= nullptr*/,
                                         uint32_t timeout /*= TELEMETRY_TIMEOUT*/)
{
  if (voltage)
    *voltage = 0.0f;
  if (input)
    *input = 0.0f;
  if (charging)
    *charging = false;
  if (batterycurrent)
    *batterycurrent = 0.0f;

  //Limit the timeout to something reasonable.
  timeout = std::min((uint32_t)(10 * 60 * 1000), timeout);

  CLock lock(m_mutex);

  //enable forced battery telemetry readings, so that the charger will read
  //the battery voltage
  if (!SetForceTelemetry(true))
    return false;

  //wait for the telemetry readings to become valid
  int64_t  start = Millis();
  bool     valid = false;
  while (!(valid = TelemetryValid()) && (Millis() - start < timeout))
    DelayMS(1);

  //If the input voltage of the charger is around 3.6 volts, the telemetry will never
  //become valid, print an error to the log, but still read the telemetry registers,
  //because the readings are still good.
  if (!valid)
    PrintLog(LOGBTTRYCHRGR, "ERROR: battery telemetry did not become valid in %"
             PRIi64 " milliseconds\n", Millis() - start);

  if (voltage)
  {
    //read the battery voltage from the vbat register
    uint16_t vbat;
    if (!ReadRegister(LTC4162F_vbat, &vbat))
      return false;

    //the voltage in vbat is 192.4 microvolts per least significant bit,
    //convert it to volts
    *voltage = (float)vbat * (192.4f / 1000000.0f);
  }

  if (input)
  { //If requested, read the input voltage.
    int16_t vin;
    if (!ReadRegister(LTC4162F_vin, (uint16_t*)&vin))
      return false;

    //Convert the vin register to volts.
    *input = (float)vin * 0.001649f;
  }

  if (charging)
  { //If requested, read the en_chg bit from the system status register.
    uint16_t statusreg;
    if (!ReadRegister(LTC4162F_SYSTEM_STATUS_REG, &statusreg))
      return false;

    if (statusreg & LTC4162F_SYSTEM_STATUS_REG_en_chg)
      *charging = true;
  }

  if (batterycurrent)
  { //If requested, read the battery charge current.
    int16_t ibat;
    if (!ReadRegister(LTC4162F_ibat, (uint16_t*)&ibat))
      return false;

    //Convert the ibat register to voltage over the 0.032 ohm sense resistor,
    //then convert to current.
    *batterycurrent = ((float)ibat * 0.000001466f) / 0.032f;
  }

  if (print)
  { //If requested, print measured values to the log.
    CLogLock loglock;

    PrintLog(LOGBTTRYCHRGR, "Battery");

    if (voltage)
      PrintLog(LOGBTTRYCHRGR, " voltage: %f", *voltage);

    if (input)
      PrintLog(LOGBTTRYCHRGR, ", input voltage: %f", *input);

    if (charging)
      PrintLog(LOGBTTRYCHRGR, ", charging: %s", *charging ? "yes" : "no");

    if (batterycurrent)
      PrintLog(LOGBTTRYCHRGR, ", battery current: %f", *batterycurrent);

    PrintLog(LOGBTTRYCHRGR, "\n");
  }

  return true;
}

/*! Checks if the battery voltage is at least the required minimum.
    \param minimum The minimum battery voltage required.
    \param loglevel The log level used when printing to the log.
    \param task The name of the task that requires the voltage, used for log printing.
    \return true if the battery voltage is sufficient.
    \return true if the battery voltage can't be read.
    \return false if the battery voltage is insufficient.*/
bool CBatteryCharger::CheckBatteryVoltage(float minimum, uint32_t loglevel, const char* task)
{
  float voltage;
  if (!ReadBatteryVoltage(&voltage))
  {
    PrintLog(loglevel, "ERROR: unable to check if battery voltage is at least minimum of %.1f volts required for %s\n",
             minimum, task);
    return true;
  }

  if (voltage < minimum)
  {
    PrintLog(loglevel, "ERROR: battery voltage is %f, which is below minimum %.1f required for %s\n",
             voltage, minimum, task);
    return false;
  }

  PrintLog(loglevel, "Battery voltage is %f, which is above minimum voltage %.1f required for %s\n",
           voltage, minimum, task);

  return true;
}

bool CBatteryCharger::ReadChargerInfo(SBatteryInfo* batteryinfo, bool print /*= false*/,
                                      uint32_t timeout /*= TELEMETRY_TIMEOUT*/)
{
  return ReadBatteryVoltage(&batteryinfo->m_intvoltage, print,
                            &batteryinfo->m_extvoltage,
                            &batteryinfo->m_charging,
                            &batteryinfo->m_chargecurrent, timeout);
}

/*! Checks the LTC4162F_CHEM_CELLS_REG of the LTC4162-F battery charger.
    If it's set incorrectly, the wrong type of charger has been soldered
    on the pcb, or the charger has not been soldered correctly.
    In that case the charger should be disabled.
    \return true The charger has correct chemistry and cell count.
    \return false The charger has incorrect chemistry, incorrect cell count,
                  or does not communicate.*/
bool CBatteryCharger::CheckChemistry()
{
  CLock lock(m_mutex);

  //read the LTC4162F_CHEM_CELLS_REG register, bits 11:8 hold
  //the battery chemistry that the charger has been designed for
  uint16_t chemreg;
  if (!ReadRegister(LTC4162F_CHEM_CELLS_REG, &chemreg))
    return false;
  uint32_t chemistry = (chemreg >> 8) & 0xF;

  bool chargercorrect = true;

  //check if the LTC4162 is of the LiFePO chemistry type,
  //if not, the charger should be disabled
  bool enablecharger = true;
  if (chemistry == LTC4162_FAD || chemistry == LTC4162_FFS || chemistry == LTC4162_FST)
  {
    PrintLog(LOGBTTRYCHRGR, "LTC4162 has correct chemistry type %" PRIu32
             ":%s, which is LiFePO\n", chemistry, LTC4162ChemistryStr(chemistry));
  }
  else if (chemistry == LTC4162_LAD || chemistry == LTC4162_L42 ||
           chemistry == LTC4162_L41 || chemistry == LTC4162_L40)
  {
    PrintLog(LOGBTTRYCHRGR, "ERROR: LTC4162 has incorrect chemistry type %"
             PRIu32 ":%s, which is lithium-ion, disabling charger\n",
             chemistry, LTC4162ChemistryStr(chemistry));
    enablecharger  = false;
    chargercorrect = false;
  }
  else if (chemistry == LTC4162_SST || chemistry == LTC4162_SAD)
  {
    PrintLog(LOGBTTRYCHRGR, "ERROR: LTC4162 has incorrect chemistry type %"
             PRIu32 ":%s, which is lead–acid, disabling charger\n",
             chemistry, LTC4162ChemistryStr(chemistry));
    enablecharger  = false;
    chargercorrect = false;
  }
  else
  { //If the chemistry type is not known, enable the charger anyway,
    //because it's possible that the chemistry type has been misread.
    PrintLog(LOGBTTRYCHRGR, "ERROR: LTC4162 has unknown chemistry type %"
             PRIu32 "\n", chemistry);

    //Indicate that the charger does not report correct chemistry.
    chargercorrect = false;
  }

  //Check the cell count, it should be 0 while not charging, and 1 while charging.
  //Any other value probably means that pins 17 and 18 of the LTC4162 are not
  //soldered correctly.
  uint32_t cellcount = chemreg & 0xF;
  if (cellcount != 0 && cellcount != 1)
  {
    PrintLog(LOGBTTRYCHRGR, "ERROR: LTC4162 has invalid cell count %"
             PRIu32 "\n", cellcount);
    chargercorrect = false;
  }

  //enable/disable the charger based on the cell chemistry
  if (!SetCharger(enablecharger))
    return false;

  return chargercorrect;
}

/*! Enables or disables the charger using the suspend_charger bit in LTC4162F_CONFIG_BITS_REG.
    \param enabled set to true to enable the charger, set to false to disable it.
    \return true if setting the charger succeeded.
    \return false on any error.*/
bool CBatteryCharger::SetCharger(bool enabled)
{
  CLock lock(m_mutex);

  uint16_t configbits;
  if (!ReadRegister(LTC4162F_CONFIG_BITS_REG, &configbits))
    return false;

  //only bits 5:1 are valid, set all other bits to 0
  configbits &= 0x3E;

  if (enabled)
    configbits &= ~(LTC4162F_CONFIG_BITS_REG_suspend_charger);
  else
    configbits |= LTC4162F_CONFIG_BITS_REG_suspend_charger;

  if (!WriteRegister(LTC4162F_CONFIG_BITS_REG, configbits))
    return false;

  return true;
}

/*! Enables/disables forced telemetry using the force_telemetry_on bit in CONFIG_BITS_REG.
    \param enabled Set to true to enable forced telemetry, false to disable it.
    \param fast When true, set the telemetry_speed for fast telemetry readings.
    \param reset When true, always write CONFIG_BITS_REG, this triggers a telemetry read
                 if force telemetry is enabled. Since telemetry reads are done every 5 seconds
                 in the slow mode, this can be used to ensure that no telemetry read happens
                 for the next 5 seconds.
    \return true if enabling forced telemetry succeeded, false on any error.*/
bool CBatteryCharger::SetForceTelemetry(bool enabled, bool fast /*= false*/, bool reset /*= true*/)
{
  CLock lock(m_mutex);

  uint16_t configbits;
  if (!ReadRegister(LTC4162F_CONFIG_BITS_REG, &configbits))
    return false;

  //only bits 5:1 are valid, set all other bits to 0
  configbits &= 0x3E;

  uint16_t newconfigbits = configbits;
  if (enabled)
  {
    newconfigbits |= LTC4162F_CONFIG_BITS_REG_force_telemetry_on;
    if (fast)
      newconfigbits |= LTC4162F_CONFIG_BITS_REG_telemetry_speed;
  }
  else
  {
    newconfigbits &= ~(LTC4162F_CONFIG_BITS_REG_force_telemetry_on |
                       LTC4162F_CONFIG_BITS_REG_telemetry_speed);
  }

  //only write the configbits register when anything should change, because
  //writing force_telemetry_on seems to trigger a telemetry read immediately,
  //even when set to the slow mode, in which case a telemetry read should
  //be done by the LTC4162 only every 5 seconds
  if (reset || (newconfigbits != configbits))
  {
    if (!WriteRegister(LTC4162F_CONFIG_BITS_REG, newconfigbits))
      return false;
  }

  return true;
}

/*! Reads the telemetry_valid bit from the TELEMETRY_STATUS_REG.
    \return true if the current telemetry is valid.
    \return false if the current telemetry is not valid.
    \return false on any error.*/
bool CBatteryCharger::TelemetryValid()
{
  CLock lock(m_mutex);

  uint16_t telemetry;
  if (!ReadRegister(LTC4162F_TELEMETRY_STATUS_REG, &telemetry))
    return false;

  return telemetry & LTC4162F_TELEMETRY_STATUS_REG_telemetry_valid;
}

/*! Enables or disables the MPPT algorithm using the mppt_en bit in LTC4162F_CONFIG_BITS_REG.
    \param enabled set to true to enable the MPPT algorithm, set to false to disable it.
    \return true if setting the charger succeeded.
    \return false on any error.*/
bool CBatteryCharger::SetMPPT(bool enabled)
{
  CLock lock(m_mutex);

  uint16_t configbits;
  if (!ReadRegister(LTC4162F_CONFIG_BITS_REG, &configbits))
    return false;

  //only bits 5:1 are valid, set all other bits to 0
  configbits &= 0x3E;

  if (enabled)
    configbits |= LTC4162F_CONFIG_BITS_REG_mppt_en;
  else
    configbits &= ~(LTC4162F_CONFIG_BITS_REG_mppt_en);

  if (!WriteRegister(LTC4162F_CONFIG_BITS_REG, configbits))
    return false;

  return true;
}

/*! Checks if the max_charge_time_fault bit is set in the charger_state register.
    If so, it will reset the charge timer by disabling and enabling the charger.
    \param mininput The minimum required input voltage to reset the charger.
                    Use a negative value to always reset the charger if the
                    max_charge_time_fault bit is set.
    \return true If all communication with the charger succeeded.
    \return false If there was a communication error with the charger.*/
bool CBatteryCharger::CheckMaxChargeTimeFault(float mininput)
{
  uint16_t chargerstate;
  if (!ReadRegister(LTC4162F_charger_state, &chargerstate))
    return false;

  //Check if the max_charge_time_fault bit it is set, this means that the charger
  //has been charging longer than the firmware has allowed.
  bool reset = false;
  if (chargerstate & LTC4162F_charger_state_max_charge_time_fault)
  {
    if (mininput < 0.0f)
    { //Always reset the charger if the minimum required voltage is negative.
      PrintLog(LOGBTTRYCHRGR, "max_charge_time_fault set, no minimum input "
               "voltage required, resetting charger\n");
      reset = true;
    }
    else
    { //Read the external voltage, reset the charger if it's high enough.
      float input;
      if (!ReadBatteryVoltage(nullptr, false, &input))
        return false;

      if (input >= mininput)
        reset = true;

      PrintLog(LOGBTTRYCHRGR, "max_charge_time_fault set, input voltage: %.3f"
               ", required minimum input voltage: %.3f, "
               "%sresetting charger\n", input, mininput, reset ? "" : "not ");
    }
  }

  //Reset the charger if requested.
  //This resets the charge timer and the charger will start charging the battery.
  if (reset)
    TriggerRecharge();

  return true;
}

/*! Sets the type of source that the charger draws power from.
    This decides how the internal LiFePO battery is charged.
    \param powersource A value of enum EChargerPowerSource.
    \return true if setting the charger input succeeded.
    \return false on any error.*/
bool CBatteryCharger::SetChargerPowerSource(EChargerPowerSource powersource)
{
  CLock lock(m_mutex);

  //Default values.
  uint32_t    maxcvtime         = LTC4162F_max_cv_time_default;
  uint32_t    rechargethreshold = LTC4162F_v_recharge_lifepo4_default;
  uint32_t    undervoltage      = LTC4162F_input_undervoltage_setting_default;
  bool        mppt              = true;
  uint32_t    maxchargetime     = 0;
  float       mininput          = -1.0f;
  const char* inputstr          = "";

  if (powersource == ChargerPowerSource_Solar_PowerSupply)
  { //For a power supply or solar panel input, the charger defaults are used.
    inputstr          = "solar panel / power supply";
  }
  else if (powersource == ChargerPowerSource_Battery)
  { //For an external battery input, set the constant voltage time to 1 second,
    //which is the minimum value, setting it to 0 turns the timer off.
    //Set the recharge threshold to 3.2 volts.
    //This will start the recharge only when the internal LiFePO battery
    //is at around 20% charge, and the charging will stop immediately when the
    //LiFePO battery voltage has reached 3.6 volts.
    //This should should be the most efficient way to charge a LiFePO battery
    //from an external battery.
    inputstr          = "external battery";
    maxcvtime         = 1;
    rechargethreshold = LTC4162F_v_recharge_lifepo4_from_volt(3.2);
  }
  else if (powersource == ChargerPowerSource_Solar_12vBattery)
  { //When using a solar panel with 12 volt lead acid battery as input, set the
    //constant voltage time to 1 second, the recharge voltage to 3.29 volts,
    //and the input undervoltage lockout to 11.672 volts.
    //This will keep the internal LiFePO battery reasonably well charged without
    //constantly recharging it, and it prevents discharging the 12 volt
    //lead acid battery too far.
    //Max charge time is limited, because the charger will still draw
    //about 4 mA even when the input voltage drops below the lockout voltage.
    inputstr          = "solar panel with 12v battery";
    maxcvtime         = 1;
    rechargethreshold = LTC4162F_v_recharge_lifepo4_from_volt(3.29);
    undervoltage      = LTC4162F_input_undervoltage_setting_from_volt(11.672);
    mppt              = false; //Undervoltage only works with mppt disabled.
    maxchargetime     = 12 * 60; //Charge for max 12 hours.
    mininput          = 12.0f; //Reset the charge timer above this input voltage.
  }
  else
  {
    PrintLog(LOGBTTRYCHRGR, "ERROR: wrong setting %i for charger power source\n",
            (int)powersource);
    inputstr = "ERROR WRONG CHARGER POWER SOURCE";
  }

  PrintLog(LOGBTTRYCHRGR, "Power source set to %s, setting max cv time:%" PRIu32
           " second%s, recharge threshold:%.3f, mppt:%s, undervoltage:%.3f, "
           "max charge time:%" PRIu32 " minutes\n",
           inputstr, maxcvtime, maxcvtime > 1 ? "s" : "",
           LTC4162F_v_recharge_lifepo4_to_volt(rechargethreshold),
           mppt ? "on" : "off",
           LTC4162F_input_undervoltage_setting_to_volt(undervoltage),
           maxchargetime);

  if (!WriteRegister(LTC4162F_max_cv_time, maxcvtime))
    return false;

  if (!WriteRegister(LTC4162F_v_recharge_lifepo4, rechargethreshold))
    return false;

  if (!SetMPPT(mppt))
    return false;

  if (!WriteRegister(LTC4162F_input_undervoltage_setting, undervoltage))
    return false;

  if (!WriteRegister(LTC4162F_max_charge_time, maxchargetime))
    return false;

  if (!CheckMaxChargeTimeFault(mininput))
    return false;

  return true;
}

/*! Disables and enables the charger, which triggers a recharge.
    \return true Disabling and enabling the charger succeeded.
    \return false Communication with the charger failed.*/
bool CBatteryCharger::TriggerRecharge()
{
  bool success = SetCharger(false);
  DelayMS(10);
  success = success && SetCharger(true);

  return success;
}
