#include "cbortools.h"
#include "log.h"
#include "lock.h"
#include "clock.h"

/*! Can find a string element in a map, by calling cbor_value_map_find_value.
    Will print an error and return false if something is wrong.
    \param loglevel The loglevel to use when printing an error.
    \param rootname The string that describes the root element.
    \param root The map to find an element in.
    \param name The string that describes the element to find.
    \param value The found element is stored here.
    \param options Can be set to a bitwise OR of CBOROption_PrintNotFoundOff and CBOROption_SkipTag.
    \return true if the element is found.
    \return false if anything is wrong.*/
bool CBORFindMapValue(uint32_t loglevel, const char* rootname, CborValue* root, const char* name, CborValue* value, CBOROption options /*= CBOROption_None*/)
{
  bool printnotfoundoff = options & CBOROption_PrintNotFoundOff;
  bool skiptag = options & CBOROption_SkipTag;

  //cbor_value_map_find_value can only be used on a map element
  if (!cbor_value_is_map(root))
  {
    PrintLog(loglevel, "ERROR: %s element is not a map\n", rootname);
    return false;
  }

  //find the element in the map
  CborError retval;
  retval = cbor_value_map_find_value(root, name, value);

  //error checks
  if (retval != CborNoError)
  {
    PrintLog(loglevel, "ERROR: cbor_value_map_find_value on %s with %s returned %i:%s\n",
             rootname, name, (int)retval, cbor_error_string(retval));
    return false;
  }
  else if (!cbor_value_is_valid(value))
  {
    if (!printnotfoundoff)
      PrintLog(loglevel, "ERROR: %s element in %s map is not present\n", name, rootname);

    return false;
  }

  //the value might start with a tag, skip over the tag if requested
  if (skiptag)
  {
    retval = cbor_value_skip_tag(value);
    if (retval != CborNoError)
    {
      PrintLog(loglevel, "ERROR: cbor_value_skip_tag returned %i:%s after finding %s element in %s map\n",
               (int)retval, cbor_error_string(retval), name, rootname);
      return false;
    }
  }

  return true;
}

/*! Uses CBORFindMapValue to find an element in a map, and checks if it's a string.
    \param options Can be CBOROption_SkipTag to skip over strings starting with a tag,
                   for example tag 0 is a standard date/time string.*/
bool CBORFindMapString(uint32_t loglevel, const char* rootname, CborValue* root, const char* name, CborValue* value, CBOROption options /*= CBOROption_None*/)
{
  if (!CBORFindMapValue(loglevel, rootname, root, name, value, options))
    return false;

  if (!cbor_value_is_text_string(value))
  {
    CLogLock loglock;
    PrintLog(loglevel, "ERROR: %s element in %s map is not a string: ", name, rootname);
    cbor_value_to_pretty(stdout, value);
    PrintLog(loglevel, "\n");
    return false;
  }

  return true;
}

/*! Uses CBORFindMapValue to find an element in a map, and checks if it's an array.*/
bool CBORFindMapArray(uint32_t loglevel, const char* rootname, CborValue* root, const char* name, CborValue* value)
{
  if (!CBORFindMapValue(loglevel, rootname, root, name, value))
    return false;

  if (!cbor_value_is_array(value))
  {
    CLogLock loglock;
    PrintLog(loglevel, "ERROR: %s element in %s map is not an array: ", name, rootname);
    cbor_value_to_pretty(stdout, value);
    PrintLog(loglevel, "\n");
    return false;
  }

  return true;
}

/*! Uses CBORFindMapValue to find an element in a map, and checks if it's an integer.
    \param options Can be CBOROption_SkipTag to skip over strings starting with a tag,
                   for example tag 0 is a standard date/time string.*/
bool CBORFindMapInt(uint32_t loglevel, const char* rootname, CborValue* root, const char* name, int64_t* value, CBOROption options /*= CBOROption_None*/)
{
  CborValue element;
  if (!CBORFindMapValue(loglevel, rootname, root, name, &element, options))
    return false;

  //check if the value is an integer
  if (!cbor_value_is_integer(&element))
  {
    CLogLock loglock;
    PrintLog(loglevel, "ERROR: %s element in %s map is not an integer: ", name, rootname);
    cbor_value_to_pretty(stdout, &element);
    PrintLog(loglevel, "\n");
    return false;
  }

  //copy the integer to the value pointer
  CborError retval;
  retval = cbor_value_get_int64(&element, value);
  if (retval != CborNoError)
  {
    PrintLog(loglevel, "ERROR: unable to get %s value from %s map: %i:%s\n",
             name, rootname, (int)retval, cbor_error_string(retval));
    return false;
  }

  return true;

}

/*! Skips over tags using cbor_value_skip_tag and advances to the next value using cbor_value_advance.
    \param loglevel The log level to use when printing an error.
    \param rootname The name of the root map to use when printing an error.
    \param element The iterator to advance.
    \return true if advancing to the next element succeeds.
    \return false on any error.*/
static bool CBORAdvance(uint32_t loglevel, const char* rootname, CborValue* root, CborValue* element)
{
  CborError retval;

  //advance to the next element
  retval = cbor_value_advance(element);
  if (retval != CborNoError)
  {
    PrintLog(loglevel, "ERROR: unable to iterate over %s map, cbor_value_advance returned %i:%s\n",
             rootname, (int)retval, cbor_error_string(retval));
    return false;
  }

  //tag elements are not wanted, cbor_value_skip_tag keeps advancing until the next non-tag element
  retval = cbor_value_skip_tag(element);
  if (retval != CborNoError)
  {
    PrintLog(loglevel, "ERROR: unable to iterate over %s map, cbor_value_skip_tag returned %i:%s\n",
             rootname, (int)retval, cbor_error_string(retval));
    return false;
  }
  return true;
}

/*! Iterates over all values in a map, finds a value with a specific integer key.
    Will print an error and return false if something is wrong.
    \param loglevel The loglevel to use when printing an error.
    \param rootname The string that describes the root element.
    \param root The map to find an element in.
    \param key The integer that describes the element to find.
    \param element The found element is stored here.
    \param options A bitwise OR of one or more values of enum CBOROption.
    \return true if the element is found.
    \return false if anything is wrong or the element is not found.*/
bool CBORFindIntMapValue(uint32_t loglevel, const char* rootname, CborValue* root, int64_t key, CborValue* value, CBOROption options /*= CBOROption_None*/)
{
  bool printnotfoundoff = options & CBOROption_PrintNotFoundOff;
  bool findlast         = options & CBOROption_FindLast;

  //this function can only be used on a map element
  if (!cbor_value_is_map(root))
  {
    PrintLog(loglevel, "ERROR: %s element is not a map\n", rootname);
    return false;
  }

  CborError retval;
  CborValue element;
  retval = cbor_value_enter_container(root, &element);
  if (retval != CborNoError)
  {
    PrintLog(loglevel, "ERROR: unable to enter %s map, cbor_value_enter_container returned %i:%s\n",
             rootname, (int)retval, cbor_error_string(retval));
    return false;
  }

  //iterate over all elements in the map find the element with the correct key
  //when iterating over map elements, first the key is encountered, then the value
  bool found = false;
  while (!cbor_value_at_end(&element))
  {
    //tag elements are not wanted, cbor_value_skip_tag keeps advancing until the next non-tag element
    retval = cbor_value_skip_tag(&element);
    if (retval != CborNoError)
    {
      PrintLog(loglevel, "ERROR: unable to iterate over %s map, cbor_value_skip_tag returned %i:%s\n",
               rootname, (int)retval, cbor_error_string(retval));
      return false;
    }

    //only consider integer keys of the key:value pairs
    if (cbor_value_is_integer(&element))
    {
      //get the integer element key
      int64_t elementkey;
      retval = cbor_value_get_int64(&element, &elementkey);
      if (retval != CborNoError)
      {
        PrintLog(loglevel, "ERROR: unable to get integer key from %s map, cbor_value_get_int64 returned %i:%s\n",
                 rootname, (int)retval, cbor_error_string(retval));
        return false;
      }

      //advance to the next element
      if (!CBORAdvance(loglevel, rootname, root, &element))
        return false;

      //if this key is the correct key, element now points to the correct value
      if (elementkey == key)
      {
        *value = element;
        found = true;

        //if the first element with this key should be found, this function is done
        if (!findlast) 
          return true;
      }
    }
    else
    {//skip non integer keys of the key:value pairs
      if (!CBORAdvance(loglevel, rootname, root, &element))
        return false;
    }

    //skip the value of this key
    if (!CBORAdvance(loglevel, rootname, root, &element))
      return false;
  }

  if (found)
    return true; //*value points to the last value with the correct key

  if (!printnotfoundoff)
    PrintLog(loglevel, "ERROR: %" PRIi64 " element in %s map not present\n", key, rootname);

  //return false if the element with the specific key is not found
  return false;
}

/*! Uses CBORFindIntMapValue to find a value with a specific integer key.
    It checks if the value belonging to the key is an integer.
    Will print an error and return false if something is wrong.
    \param loglevel The loglevel to use when printing an error.
    \param rootname The string that describes the root element.
    \param root The map to find an element in.
    \param key The integer that describes the element to find.
    \param value The value of the key is returned here.
    \param options A bitwise OR of one or more values of enum CBOROption.
    \return true if the element is found.
    \return false if anything is wrong or the element is not found.*/
bool CBORFindIntMapInt(uint32_t loglevel, const char* rootname, CborValue* root, int64_t key, int64_t* value, CBOROption options /*= CBOROption_None*/)
{
  //find the element with the correct key
  CborValue element;
  if (!CBORFindIntMapValue(loglevel, rootname, root, key, &element, options))
    return false;

  //check if the value is an integer
  if (!cbor_value_is_integer(&element))
  {
    CLogLock loglock;
    PrintLog(loglevel, "ERROR: %" PRIi64 " element in %s map is not an integer: ", key, rootname);
    cbor_value_to_pretty(stdout, &element);
    PrintLog(loglevel, "\n");
    return false;
  }

  //copy the integer to the value pointer
  CborError retval;
  retval = cbor_value_get_int64(&element, value);
  if (retval != CborNoError)
  {
    PrintLog(loglevel, "ERROR: unable to get %" PRIi64 " value from %s map: %i:%s\n",
             key, rootname, (int)retval, cbor_error_string(retval));
    return false;
  }

  return true;
}

/*! Uses CBORFindIntMapValue to find a value with a specific integer key.
    It check if the value belonging to the key is a string.
    Will print an error and return false if something is wrong.
    \param loglevel The loglevel to use when printing an error.
    \param rootname The string that describes the root element.
    \param root The map to find an element in.
    \param key The integer that describes the element to find.
    \param value The string value of the key is returned here.
    \param options A bitwise OR of one or more values of enum CBOROption.
    \return true if the element is found.
    \return false if anything is wrong or the element is not found.*/
bool CBORFindIntMapString(uint32_t loglevel, const char* rootname, CborValue* root, int64_t key, CborValue* value, CBOROption options /*= CBOROption_None*/)
{
  //find the element with the correct key
  if (!CBORFindIntMapValue(loglevel, rootname, root, key, value, options))
    return false;

  //check if the value is a string
  if (!cbor_value_is_text_string(value))
  {
    CLogLock loglock;
    PrintLog(loglevel, "ERROR: %" PRIi64 " element in %s map is not a string: ", key, rootname);
    cbor_value_to_pretty(stdout, value);
    PrintLog(loglevel, "\n");
    return false;
  }

  return true;
}

/*! Uses CBORFindIntMapValue to find a value with a specific integer key.
    It checks if the value belonging to the key is a bool.
    Will print an error and return false if something is wrong.
    \param loglevel The loglevel to use when printing an error.
    \param rootname The string that describes the root element.
    \param root The map to find an element in.
    \param key The integer that describes the element to find.
    \param value The value of the key is returned here.
    \param options A bitwise OR of one or more values of enum CBOROption.
    \return true if the element is found.
    \return false if anything is wrong or the element is not found.*/
bool CBORFindIntMapBool(uint32_t loglevel, const char* rootname, CborValue* root, int64_t key, bool* value, CBOROption options /*= CBOROption_None*/)
{
  //find the element with the correct key
  CborValue element;
  if (!CBORFindIntMapValue(loglevel, rootname, root, key, &element, options))
    return false;

  //check if the value is an integer
  if (!cbor_value_is_boolean(&element))
  {
    CLogLock loglock;
    PrintLog(loglevel, "ERROR: %" PRIi64 " element in %s map is not a boolean: ", key, rootname);
    cbor_value_to_pretty(stdout, &element);
    PrintLog(loglevel, "\n");
    return false;
  }

  //copy the integer to the value pointer
  CborError retval;
  retval = cbor_value_get_boolean(&element, value);
  if (retval != CborNoError)
  {
    PrintLog(loglevel, "ERROR: unable to get %" PRIi64 " value from %s map: %i:%s\n",
             key, rootname, (int)retval, cbor_error_string(retval));
    return false;
  }

  return true;
}

/*! Uses CBORFindIntMapValue to find a value with a specific integer key.
    It check if the value belonging to the key is a byte string.
    Will print an error and return false if something is wrong.
    \param loglevel The loglevel to use when printing an error.
    \param rootname The string that describes the root element.
    \param root The map to find an element in.v
    \param key The integer that describes the element to find.
    \param value The string value of the key is returned here.
    \param options A bitwise OR of one or more values of enum CBOROption.
    \return true if the element is found.
    \return false if anything is wrong or the element is not found.*/
bool CBORFindIntMapByteString(uint32_t loglevel, const char* rootname, CborValue* root, int64_t key, CborValue* value, CBOROption options /*= CBOROption_None*/)
{
  //find the element with the correct key
  if (!CBORFindIntMapValue(loglevel, rootname, root, key, value, options))
    return false;

  //check if the value is a byte string
  if (!cbor_value_is_byte_string(value))
  {
    CLogLock loglock;
    PrintLog(loglevel, "ERROR: %" PRIi64 " element in %s map is not a byte string: ", key, rootname);
    cbor_value_to_pretty(stdout, value);
    PrintLog(loglevel, "\n");
    return false;
  }

  return true;
}

/*! Uses CBORFindIntMapValue to find a value with a specific integer key.
    It checks if the value belonging to the key is a float.
    Will print an error and return false if something is wrong.
    \param loglevel The loglevel to use when printing an error.
    \param rootname The string that describes the root element.
    \param root The map to find an element in.
    \param key The integer that describes the element to find.
    \param value The value of the key is returned here.
    \param options A bitwise OR of one or more values of enum CBOROption.
    \return true if the element is found.
    \return false if anything is wrong or the element is not found.*/
bool CBORFindIntMapFloat(uint32_t loglevel, const char* rootname, CborValue* root, int64_t key, float* value, CBOROption options /*= CBOROption_None*/)
{
  //find the element with the correct key
  CborValue element;
  if (!CBORFindIntMapValue(loglevel, rootname, root, key, &element, options))
    return false;

  //check if the value is an integer
  if (!cbor_value_is_float(&element))
  {
    CLogLock loglock;
    PrintLog(loglevel, "ERROR: %" PRIi64 " element in %s map is not a float: ", key, rootname);
    cbor_value_to_pretty(stdout, &element);
    PrintLog(loglevel, "\n");
    return false;
  }

  //copy the float to the value pointer
  CborError retval;
  retval = cbor_value_get_float(&element, value);
  if (retval != CborNoError)
  {
    PrintLog(loglevel, "ERROR: unable to get %" PRIi64 " value from %s map: %i:%s\n",
             key, rootname, (int)retval, cbor_error_string(retval));
    return false;
  }

  return true;
}

/*! Calls cbor_value_text_string_equals to do a string compare, will print any error to the log.
    \param loglevel The loglevel to use with PrintLog().
    \param element The element to do a string compare on.
    \param str The string to compare with.
    \param result The result of the string comparison is stored here.
    \return true if the string comparison succeeded, false if it failed.*/
bool CBORCompareString(uint32_t loglevel, const char* name, CborValue* element, const char* str, bool* result)
{
  *result = false; //failsafe

  if (!cbor_value_is_text_string(element))
  {
    CLogLock loglock;
    PrintLog(loglevel, "ERROR: element %s to compare with %s is not a string: ", name, str);
    cbor_value_to_pretty(stdout, element);
    PrintLog(loglevel, "\n");
    return false;
  }

  CborError retval = cbor_value_text_string_equals(element, str, result);
  if (retval != CborNoError)
  {
    PrintLog(loglevel, "ERROR: cbor_value_text_string_equals on element %s with %s returned %i:%s\n",
             name, str, (int)retval, cbor_error_string(retval));
    return false;
  }

  return true;
}

/*! Calulates the length of a string using cbor_value_calculate_string_length.
    It will print any error it encounters to the log.
    \param loglevel The loglevel to use with PrintLog().
    \param name The name of the element to use when printing to the log.
    \param element The element to calculate the string length of.
    \param length The length of the string is returned here.
    \return true if the string length was successfully calculated.
    \return false on any error.*/
bool CBORGetStringLength(uint32_t loglevel, const char* name, CborValue* element, uint32_t* length)
{
  *length = 0; //failsafe

  if (!cbor_value_is_text_string(element) && !cbor_value_is_byte_string(element))
  {
    CLogLock loglock;
    PrintLog(loglevel, "ERROR: element %s to get length of is not a string: ", name);
    cbor_value_to_pretty(stdout, element);
    PrintLog(loglevel, "\n");
    return false;
  }

  size_t lengthval;
  CborError retval = cbor_value_calculate_string_length(element, &lengthval);
  if (retval != CborNoError)
  {
    PrintLog(loglevel, "ERROR: getting string length of element %s returned %i:%s\n",
             name, (int)retval, cbor_error_string(retval));
    return false;
  }

  *length = lengthval;
  return true;
}

/*! Copies a CBOR string to a char array using cbor_value_copy_text_string, and prints any error to the log.
    \param loglevel The loglevel to use with PrintLog().
    \param name The name of the element to use when printing to the log.
    \param element The element to copy to the char array.
    \param buf The buffer to copy the CBOR string to.
    \param length A pointer to the length of the buffer in bytes, the length of the copied string is returned here.
    \return true if copying the string succeeded.
    \return false if anything failed.*/
bool CBORCopyString(uint32_t loglevel, const char* name, CborValue* element, char* buf, uint32_t* length)
{
  if (!cbor_value_is_text_string(element))
  {
    CLogLock loglock;
    PrintLog(loglevel, "ERROR: element %s to copy not a string: ", name);
    cbor_value_to_pretty(stdout, element);
    PrintLog(loglevel, "\n");
    return false;
  }

  size_t lengthval = *length;
  CborError retval = cbor_value_copy_text_string(element, buf, &lengthval, nullptr);
  if (retval != CborNoError)
  {
    PrintLog(loglevel, "ERROR: copying element %s to string returned %i:%s\n",
             name, (int)retval, cbor_error_string(retval));
    return false;
  }

  *length = lengthval;
  return true;
}

/*! Copies a CBOR byte string to a uint8_t array using cbor_value_copy_byte_string, and prints any error to the log.
    \param loglevel The loglevel to use with PrintLog().
    \param name The name of the element to use when printing to the log.
    \param element The element to copy to the char array.
    \param buf The buffer to copy the CBOR byte string to.
    \param length A pointer to the length of the buffer in bytes, the length of the copied byte string is returned here.
    \return true if copying the string succeeded.
    \return false if anything failed.*/
bool CBORCopyByteString(uint32_t loglevel, const char* name, CborValue* element, uint8_t* buf, uint32_t* length)
{
  if (!cbor_value_is_byte_string(element))
  {
    CLogLock loglock;
    PrintLog(loglevel, "ERROR: element %s to copy is not a byte string: ", name);
    cbor_value_to_pretty(stdout, element);
    PrintLog(loglevel, "\n");
    return false;
  }

  size_t lengthval = *length;
  CborError retval = cbor_value_copy_byte_string(element, buf, &lengthval, nullptr);
  if (retval != CborNoError)
  {
    PrintLog(loglevel, "ERROR: copying element %s to string returned %i:%s\n",
             name, (int)retval, cbor_error_string(retval));
    return false;
  }

  *length = lengthval;
  return true;
}

/*! Uses cbor_value_enter_container to enter a container, prints any error to the log.
    \param loglevel The loglevel to use when printing to the log.
    \param name The name of the container to use when printing to the log.
    \param container The container to enter.
    \param it The iterator is returned here when entering the container.
    \return true if entering the container succeeded.
    \return false if entering the container failed.*/
bool CBOREnterContainer(uint32_t loglevel, const char* name, CborValue* container, CborValue* it)
{
  CborError retval = cbor_value_enter_container(container, it);
  if (retval != CborNoError)
  {
    PrintLog(loglevel, "ERROR: unable to enter %s container\n", name);
    return false;
  }

  return true;
}

/*! Uses cbor_value_leave_container to leave a container which previously
    should have been opened with cbor_value_enter_container, prints any error to the log.
    \param loglevel The loglevel to use when printing to the log.
    \param name The name of the container to use when printing to the log.
    \param container The container to leave.
    \param it The iterator that was used when leaving the container previously.
    \return true if leaving the container succeeded.
    \param false if leaving the container failed.*/
bool CBORLeaveContainer(uint32_t loglevel, const char* name, CborValue* container, CborValue* it)
{
  CborError retval = cbor_value_leave_container(container, it);
  if (retval != CborNoError)
  {
    PrintLog(loglevel, "ERROR: unable to leave %s container\n", name);
    return false;
  }

  return true;
}

/*! Uses cbor_value_advance to advance an iterator from a container that
    was previously entered with cbor_value_enter_container.
    It will print any error to the log.
    \param loglevel The loglevel to use when printing to the log.
    \param name The container name to use when printing to the log.
    \param it The iterator to advance.
    \return true if advancing the iterator succeeded.
    \return false if advancding the iterator failed.*/
bool CBORValueAdvance(uint32_t loglevel, const char* name, CborValue* it)
{
  CborError retval = cbor_value_advance(it);
  if (retval != CborNoError)
  {
    PrintLog(loglevel, "ERROR: unable to advance in %s container\n", name);
    return false;
  }

  return true;
}

/*! Calls cbor_encoder_close_container() and checks its return value,
    will print an error if it's not CborNoError.
    \param loglevel The loglevel to use when printing.
    \param name The name of the element to use when printing.
    \param encoder A pointer to the encoder used with cbor_encoder_close_container().
    \param containter A pointer to the container used with cbor_encoder_close_container().
    \return true if cbor_encoder_close_container() succeeded.
    \return false if cbor_encoder_close_container() failed.*/
bool CBORCloseContainer(uint32_t loglevel, const char* name, CborEncoder* encoder, CborEncoder* container)
{
  CborError retval = cbor_encoder_close_container(encoder, container);

  if (retval == CborErrorOutOfMemory)
  {
    size_t bytesneeded = cbor_encoder_get_extra_bytes_needed(encoder);
    if (bytesneeded != 0)
    {
      PrintLog(loglevel, "ERROR: %s too large, need %i extra bytes to encode it\n",
               name, (int)bytesneeded);
      return false;
    }
  }

  if (retval != CborNoError)
  {
    PrintLog(loglevel, "ERROR: cbor_encoder_close_container on %s returned %i:%s\n",
             name, (int)retval, cbor_error_string(retval));
    return false;
  }

  return true;
}

/*! Gets an element from the root map, and parses it as an ISO 8601 date string.
    It then sets the internal RTC to this date.
    \param loglevel The loglevel to use when printing.
    \param rootname The string that describes the root element.
    \param root The map to find an element in.
    \param name The string that describes the element to find.
    \return true if parsing the date string succeeded.
    \return fales if anything failed.*/
bool CBORSetRTCFromDateField(uint32_t loglevel, const char* rootname, CborValue* root, const char* name)
{
  //find the element that holds the date string, use CBOROption_SkipTag in case
  //the string is tagged as a standard date/time string
  CborValue cbordate;
  if (!CBORFindMapString(loglevel, rootname, root, name, &cbordate, CBOROption_SkipTag))
    return false;

  uint32_t length;
  if (!CBORGetStringLength(loglevel, name, &cbordate, &length))
    return false;

  if (length > DATESTR_MAX_LENGTH)
  {
    PrintLog(loglevel, "ERROR: date string of length %" PRIu32
             " is too long, the maximum length allowed is %i\n",
             length, DATESTR_MAX_LENGTH);
    return false;
  }

  //allocate one extra byte and set the array to 0,
  //so that datestr is always a valid c-string
  char datestr[DATESTR_MAX_LENGTH + 1] = {};

  //copy the date string to a char array
  //length = DATESTR_MAX_LENGTH;
  if (!CBORCopyString(loglevel, name, &cbordate, datestr, &length))
    return false;

  //set the internal RTC based on the date string
  if (!SetRTCISO8601(datestr))
  {
    PrintLog(loglevel, "ERROR: date string %s is incorrect\n", datestr);
    return false;
  }

  return true;
}

/*! Checks the CBOR in the buffer, and prints it to stdout.
    The CBOR should at least have a root map.
    \param loglevel The loglevel to use when printing.
    \param prefix The string printed before the CBOR, should end with ": ", or be nullptr.
    \param buf The buffer that holds the CBOR.
    \param bufsize The size of the buffer that buf points to
    \return true if the CBOR is correct.
    \return false if the CBOR is incorrect, or does not have a root map.*/
bool CBORPrint(uint32_t loglevel, const char* prefix, uint8_t* buf, uint32_t bufsize)
{
  if (prefix == nullptr)
    prefix = "";

  CborError  retval;
  CborParser parser;
  CborValue  it;

  //parse and validate the CBOR in the buffer
  retval = cbor_parser_init(buf, bufsize, CborValidateBasic, &parser, &it);
  if (retval != CborNoError)
  {
    PrintLog(loglevel, "%sERROR: cbor_parser_init returned %i:%s\n",
             prefix, (int)retval, cbor_error_string(retval));
    return false;
  }

  retval = cbor_value_validate_basic(&it);
  if (retval != CborNoError)
  {
    PrintLog(loglevel, "%sERROR: cbor_value_validate_basic returned %i:%s\n",
             prefix, (int)retval, cbor_error_string(retval));
    return false;
  }

  if (!cbor_value_is_map(&it))
  {
    CLogLock loglock;
    PrintLog(loglevel, "%sERROR: the first CBOR element should always be a map: ",
             prefix);
    cbor_value_to_pretty(stdout, &it);
    PrintLog(loglevel, "\n");
    return false;
  }

  //print the CBOR to the log
  CLogLock loglock;
  PrintLog(loglevel, "%s", prefix);
  retval = cbor_value_to_pretty(stdout, &it);
  PrintLog(loglevel, "\n");
  if (retval != CborNoError)
  {
    PrintLog(loglevel, "%sERROR: cbor_value_to_pretty returned %i:%s\n",
             prefix, (int)retval, cbor_error_string(retval));
    return false;
  }

  return true;
}
