#define _GNU_SOURCE //for strptime in C++

#include "clock.h"
#include "log.h"
#include "lock.h"
#include "timeconversion.h"
#include "backupregisters.h"

#include <FreeRTOS.h>
#include <task.h>

#include <stm32f4xx.h>

#include <cstdlib>
#include <cassert>
#include <cstring>

/*! \return The frequency that timers are driven with.*/
uint32_t TimerClockFrequency()
{
  //get the clock frequency of the APB1, which is what clocks timer2,
  RCC_ClocksTypeDef clockfreqs;
  RCC_GetClocksFreq(&clockfreqs);

  //according to "Figure 21. Clock tree" on page 216 of the STM32F407 reference manual,
  //timers are clocked by the system clock, with a prescaler of 1 if the APBx prescaler is 1,
  //and a prescaler of 2 if the APBx prescaler is anything else
  uint32_t clockdiv;
  if (clockfreqs.SYSCLK_Frequency == clockfreqs.PCLK1_Frequency)
    clockdiv = 1;
  else
    clockdiv = 2;

  return clockfreqs.SYSCLK_Frequency / clockdiv;
}

/*! This function sets up timer2 as a 32 bit microsecond counter.*/
void ClockInit()
{
  //enable the clock for timer2
  RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM2, ENABLE);

  //initialize timer2 as a 32 bit upcounter clock at 1 tick per microsecond
  TIM_TimeBaseInitTypeDef timinit = {};
  timinit.TIM_Prescaler           = (TimerClockFrequency() / 1000000) - 1;
  timinit.TIM_CounterMode         = TIM_CounterMode_Up;
  timinit.TIM_Period              = 0xFFFFFFFF;
  timinit.TIM_ClockDivision       = TIM_CKD_DIV1;
  timinit.TIM_RepetitionCounter   = 0;
  TIM_TimeBaseInit(TIM2, &timinit);
  TIM_Cmd(TIM2, ENABLE);
}

/*! Backup domain access by calling EnableBackupAccess() should be enabled before
    calling this function.
    This function checks the LSERDY flag in the RCC_BDCR register,
    and the rtc initialized flag in the flags register,
    if one of those are cleared it assumes that a backup domain reset has happened.
    It will then do another backup domain reset in case the LSERDY flag is cleared
    for another reason,
    and then it sets up the RTC with the 32.768 external oscillator.
    Checking the LSERDY flag just by itself does not seem to be a reliable indicator
    that the RTC should be set up, because of that one bit in the flags register
    has been reserved to indicate if the RTC has been set up correctly by this code.*/
void RTCInit(bool force /*= false*/)
{
  bool lserdy = RCC_GetFlagStatus(RCC_FLAG_LSERDY);
  bool rtcinitialized = RTCInitializedFlagSet();

  if (lserdy && rtcinitialized && !force)
    return;

  PrintLog(LOGCLOCK, "Doing a backup domain reset, LSERDY:%i, RTC initialized:%s,"
           " force initialization: %s\n",(int)lserdy, rtcinitialized ? "yes" : "no",
           force ? "yes" : "no");

  //if the LSERDY flag is not set, it should mean that a backup doman reset has happened
  //do another backup domain reset in case the LSERDY flag is reset because of a different reason
  RCC_BackupResetCmd(ENABLE);
  for (volatile uint32_t i = 0; i < 1000000; i++);
  RCC_BackupResetCmd(DISABLE);
  PrintLog(LOGCLOCK, "Backup domain reset completed\n");

  //start the 32.768 external oscillator if it's not started, this might take a few seconds
  //this usually happens when VBAT has dropped to 0, because the supercapacitor is empty
  PrintLog(LOGCLOCK, "Waiting for 32.768 kHz oscillator to start\n");
  RCC_LSEConfig(RCC_LSE_ON);

  //wait for the 32.768 kHz oscillator to start
  uint32_t start = Micros();
  for(;;)
  {
    if (RCC_GetFlagStatus(RCC_FLAG_LSERDY) == SET)
    {
      PrintLog(LOGCLOCK, "32.768 kHz oscillator started\n");
      break;
    }
    else if (Micros() - start > (20 * 1000000))
    {
      PrintLog(LOGCLOCK, "ERROR: 32.768 kHz oscillator did not start in time\n");
      break;
    }
  }

  RCC_RTCCLKConfig(RCC_RTCCLKSource_LSE); //select the external 32.768 kHz oscillator as RTC clock source
  RCC_RTCCLKCmd(ENABLE);

  RTC_InitTypeDef rtcinit  = {};
  RTC_StructInit(&rtcinit); //sets 24 hour format with default prescalers for 32.768 kHz oscillator
  ErrorStatus rtcstate = RTC_Init(&rtcinit);

  if (rtcstate == ERROR)
    PrintLog(LOGCLOCK, "ERROR: unable to initialize RTC\n");
  else
    PrintLog(LOGCLOCK, "Successfully initialized RTC\n");

  //set the flag that the RTC has been correctly initialized
  SetRTCInitializedFlag(true);
}

void PrintRTCTime()
{
  char buf[DATESTR_MAX_LENGTH] = {};
  struct tm timeval;
  GetRTC(&timeval);
  PrintLog(LOGCLOCK, "Internal RTC time: %s", asctime_r(&timeval, buf));

  //print an error here if the 32.768 kHz oscillator is not running,
  //so that it ends up in the SPI flash log
  if (RCC_GetFlagStatus(RCC_FLAG_LSERDY) != SET)
    PrintLog(LOGCLOCK, "ERROR: 32.768 kHz oscillator is not running\n");
}

//! Number of seconds the RTC test will run.
#define RTC_TEST_TIME 5

/*! Tests if the RTC runs forward in time.
    \return true The RTC is working correctly.
    \return false The RTC is not working correctly.*/
bool TestRTC()
{
  PrintLog(LOGCLOCK, "Testing if internal RTC runs\n");

  //Get a timestamp from the RTC, wait RTC_TEST_TIME seconds,
  //get another RTC timestamp.
  uint32_t teststart = Micros();
  int64_t  start = GetRTCSeconds(nullptr);
  DelayMS(RTC_TEST_TIME * 1000);
  int64_t  elapsed = GetRTCSeconds(nullptr) - start;
  uint32_t testduration = Micros() - teststart;

  //Check if the rtc has moved forward RTC_TEST_TIME seconds.
  if (elapsed >= RTC_TEST_TIME - 1 && elapsed <= RTC_TEST_TIME + 1)
  {
    PrintLog(LOGCLOCK, "Internal RTC is working correctly, time increased by %" PRIi64
             " seconds in %" PRIu32 " microseconds\n", elapsed, testduration);
    return true;
  }
  else
  {
    PrintLog(LOGCLOCK, "ERROR: Internal RTC is not working correctly, time increased by %"
             PRIi64 " seconds in %" PRIu32 " microseconds\n", elapsed, testduration);
    return false;
  }
}

/*! Updates and returns a 64 bit signed counter of milliseconds
    elapsed since the system was started.
    The counter is based on the Micros() function.
    This counter can be used for general timing, and also
    to check how long the firmware has been running.
 \return Milliseconds elapsed since the system was started.*/
int64_t Millis()
{
  //disable interrupts to prevent other tasks from changing
  //the milliseconds and microseconds variables
  __disable_irq();

  static volatile int64_t  milliseconds; //milliseconds counter since the system was started
  static volatile uint32_t microseconds; //last value of Micros() that was added to milliseconds

  uint32_t now  = Micros();
  uint32_t diff = now - microseconds;

  milliseconds += diff / 1000;         //add the milliseconds elapsed since the previous function call
  microseconds  = now - (diff % 1000); //store the timestamp of this function call,

  int64_t millisvalue = milliseconds; //store the milliseconds value before exiting the critical section

  __enable_irq();

  return millisvalue;
}

/*! Blocking millisecond delay, based on the vTaskDelay() function.
    Requires the FreeRTOS kernel to run.
    This function can't be used from an interrupt or a critical section.
    \param milliseconds Number of milliseconds to wait.*/
void DelayMS(uint32_t milliseconds)
{
  vTaskDelay(pdMS_TO_TICKS(milliseconds));
}

/*! Blocking microsecond delay, based on the Micros() function.
    \param microseconds Number of microseconds to wait.*/
void DelayUS(uint32_t microseconds)
{
  uint32_t start = Micros();
  while ((Micros() - start) < microseconds);
}

static CRTCAdjustCallback* g_callback;

/*! Sets the internal RTC based on the time in the timeval argument.*/
void SetRTC(struct tm* timeval)
{
  //get the current RTC time
  tm now;
  GetRTC(&now);

  int64_t nowsecs  = TmToSecs(&now);
  int64_t newsecs  = TmToSecs(timeval);
  int64_t timediff = newsecs - nowsecs;

  //only set the RTC if the time is different from the time passed as argument
  //or if the RTC valid flag has not been set in the battery backed ram
  if (timediff != 0 || !RTCValidFlagSet())
  {
    char nowstr[DATESTR_MAX_LENGTH] = {};
    strftime(nowstr, sizeof(nowstr) - 1, "%c", &now);

    char newstr[DATESTR_MAX_LENGTH] = {};
    strftime(newstr, sizeof(newstr) - 1, "%c", timeval);

    PrintLog(LOGCLOCK, "Adjusting RTC by %" PRIi64 " second%s, from %s to %s\n",
             timediff, llabs(timediff) == 1 ? "" : "s", nowstr, newstr);

    RTC_TimeTypeDef time =
    {
      .RTC_Hours   = (uint8_t)timeval->tm_hour,
      .RTC_Minutes = (uint8_t)timeval->tm_min,
      .RTC_Seconds = (uint8_t)timeval->tm_sec,
      .RTC_H12     = RTC_H12_AM,
    };

    RTC_DateTypeDef date =
    {
      //the tm_wday member is days since sunday,
      //the RTC_Weekday member is 1 for monday and 7 for sunday
      .RTC_WeekDay = (uint8_t)(timeval->tm_wday == 0 ? 7 : timeval->tm_wday),
      //the tm_mon member is months since january,
      //the RTC_Month member is 1 for january and 12 for december
      .RTC_Month   = (uint8_t)(timeval->tm_mon + 1),
      .RTC_Date    = (uint8_t)timeval->tm_mday,
      //the RTC_Year member holds a value from 0 to 99
      //the tm_year member holds years since 1900
      .RTC_Year    = (uint8_t)(timeval->tm_year % 100),
    };

    //Inform an installed callback that the RTC is about to be adjusted.
    //This is used by CRTCScheduler.
    if (g_callback)
      g_callback->RTCAdjustCallback(timediff,
                                    CRTCAdjustCallback::PreAdjustCallback);

    //adjust the RTC time and date
    CCritical criticalsection;
    RTC_SetTime(RTC_Format_BIN, &time);
    RTC_SetDate(RTC_Format_BIN, &date);
    criticalsection.Leave();

    //set the flag in the battery backed ram that the RTC has been set to a valid time
    SetRTCValidFlag(true);

    //Inform an installed callback that the RTC has been adjusted. This is used
    //by CRTCScheduler.
    if (g_callback)
      g_callback->RTCAdjustCallback(timediff,
                                    CRTCAdjustCallback::PostAdjustCallback);
  }
  else
  {
    char nowstr[DATESTR_MAX_LENGTH];
    strftime(nowstr, sizeof(nowstr) - 1, "%c", &now);
    PrintLog(LOGCLOCK, "RTC set to %s, no adjustment needed.\n", nowstr);
  }
}

/*! Gets the date from the RTC and converts it to a struct tm.*/
void GetRTC(struct tm* timeval)
{
  //prevent anything from interrupting getting the RTC time and date,
  //so that the time and date values are read from the RTC as close to
  //each other as possible
  CCritical criticalsection;

  //wait until the RTC has copied the current time and date into the shadow registers
  RTC_WriteProtectionCmd(DISABLE);
  RTC_WaitForSynchro();

  //get the time from the RTC
  RTC_TimeTypeDef time;
  RTC_DateTypeDef date;
  RTC_GetTime(RTC_Format_BIN, &time);
  RTC_GetDate(RTC_Format_BIN, &date);

  //convert the time to a struct tm
  timeval->tm_sec  = time.RTC_Seconds;
  timeval->tm_min  = time.RTC_Minutes;
  timeval->tm_hour = time.RTC_Hours;
  timeval->tm_mday = date.RTC_Date;
  //the tm_mon member is months since january,
  //the RTC_Month member is 1 for january and 12 for december
  timeval->tm_mon  = date.RTC_Month - 1;
  //the RTC_Year member holds a value from 0 to 99
  //the tm_year member holds years since 1900
  timeval->tm_year = ((int)date.RTC_Year) + 100;
  //the tm_wday member is days since sunday,
  //the RTC_Weekday member is 1 for monday and 7 for sunday
  timeval->tm_wday = date.RTC_WeekDay % 7;
  timeval->tm_yday = 0; //not used by TmToSecs
}

/*! Returns the RTC time as a unix timestamp.*/
int64_t GetRTCSeconds(bool* valid)
{
  //get the RTC time and the RTC valid flag inside a critical section,
  //so that the RTC valid flag doesn't get set by a different task in
  //between getting the RTC time and getting the RTC valid flag
  CCritical criticalsection;

  //get the valid flag if requested
  if (valid)
    *valid = RTCValidFlagSet();

  //get the time and date from the RTC
  struct tm timeval;
  GetRTC(&timeval);

  //return the time as a unix timestamp
  return TmToSecs(&timeval);
}

/*! Sets the RTC from a unix timestamp.*/
void SetRTCSeconds(int64_t unixtime)
{
  struct tm timeval;
  SecsToTm(unixtime, &timeval);
  SetRTC(&timeval);
}

/*! Sets a pointer to a CRTCAdjustCallback who's RTCAdjustCallback method will be
    called after the RTC has been adjusted.*/
void SetRTCAdjustCallback(CRTCAdjustCallback* callback)
{
  g_callback = callback;
}

/*! Sets the RTC wakeup counter value, and sets the frequency to 1 hertz.
    \param count The count to reach before a wakeup event is generated.
                 Each tick takes 1 second.*/
void SetRTCWakeup(uint32_t count)
{
  if (count > RTC_WAKEUP_MAX)
    count = RTC_WAKEUP_MAX;

  PrintLog(LOGCLOCK, "Setting RTC wakeup timer to %" PRIu32 "h %02" PRIu32
           "m %02" PRIu32 "s\n", count / 3600, (count / 60) % 60, count % 60);

  CCritical criticalsection;

  //the RTC wakeup should be disabled before it can be adjusted
  RTC_WakeUpCmd(DISABLE);
  RTC_ClearFlag(RTC_FLAG_WUTF);

  if (count >= 0x10000)
  { //set the wakeup counter to 1 hertz, with 2^16 added to the counter value
    RTC_WakeUpClockConfig(RTC_WakeUpClock_CK_SPRE_17bits);
    count -= 0x10000;
  }
  else
  { //set the wakeup counter to 1 hertz
    RTC_WakeUpClockConfig(RTC_WakeUpClock_CK_SPRE_16bits);
  }

  //set the wakeup counter register and enable the wakeup counter
  RTC_SetWakeUpCounter(count);
  RTC_WakeUpCmd(ENABLE);
}

/*! Disables the RTC wakeup counter.*/
void DisableRTCWakeup()
{
  PrintLog(LOGCLOCK, "Disabling RTC wakeup timer\n");

  CCritical criticalsection;

  RTC_WakeUpCmd(DISABLE);
  RTC_ClearFlag(RTC_FLAG_WUTF);
}

/*! \return The state of the RTC_FLAG_WUTF flag.*/
bool RTCWakeupFlagSet()
{
  CCritical criticalsection;
  return RTC_GetFlagStatus(RTC_FLAG_WUTF);
}

/*! Clears the RTC wakeup flag, set by the RTC wakeup counter.*/
void ClearRTCWakeupFlag()
{
  CCritical criticalsection;
  RTC_ClearFlag(RTC_FLAG_WUTF);
}

/*! Sets the RTC alarm to a specific time of the day, and specific day of the month.
    If the RTC alarm is already set to the correct setting, the alarm will not
    be set again, however it will be enabled using RTC_AlarmCmd().
    \param rtcalarm RTC_Alarm_A or RTC_Alarm_B.
    \param timeval The alarm will be set to timeval->tm_hour and timeval->tm_min.*/
void SetRTCAlarm(uint32_t rtcalarm, struct tm* timeval)
{
  assert(rtcalarm == RTC_Alarm_A || rtcalarm == RTC_Alarm_B);

  //make sure all bytes including padding are set to zero
  RTC_AlarmTypeDef currentalarm;
  memset(&currentalarm, 0, sizeof(currentalarm));

  //get the current alarm information
  CCritical criticalsection;
  RTC_GetAlarm(RTC_Format_BIN, rtcalarm, &currentalarm);
  criticalsection.Leave();

  //make sure all bytes including padding are set to zero
  RTC_AlarmTypeDef wantedalarm = {};
  memset(&wantedalarm, 0, sizeof(wantedalarm));

  //create a struct about how the alarm should be configured
  wantedalarm.RTC_AlarmMask             = RTC_AlarmMask_None;
  wantedalarm.RTC_AlarmDateWeekDaySel   = RTC_AlarmDateWeekDaySel_Date;
  wantedalarm.RTC_AlarmDateWeekDay      = timeval->tm_mday;
  wantedalarm.RTC_AlarmTime.RTC_Hours   = timeval->tm_hour % 24;
  wantedalarm.RTC_AlarmTime.RTC_Minutes = timeval->tm_min % 60;
  wantedalarm.RTC_AlarmTime.RTC_Seconds = timeval->tm_sec % 60;
  wantedalarm.RTC_AlarmTime.RTC_H12     = RTC_H12_AM;

  const char* alarmstr = rtcalarm == RTC_Alarm_A ? "A" : "B";
  if (memcmp(&currentalarm, &wantedalarm, sizeof(currentalarm)) != 0)
  { //the current alarm setting is different from the wanted setting,
    //set the alarm to the wanted setting
    PrintLog(LOGCLOCK, "Setting RTC alarm %s to %02i:%02i:%02i, day of month: %i\n",
             alarmstr, timeval->tm_hour, timeval->tm_min, timeval->tm_sec, timeval->tm_mday);

    criticalsection.Enter();
    RTC_AlarmCmd(rtcalarm, DISABLE);
    RTC_SetAlarm(RTC_Format_BIN, rtcalarm, &wantedalarm);
    RTC_AlarmCmd(rtcalarm, ENABLE);
    criticalsection.Leave();
  }
  else
  { //the current alarm setting is the same as the wanted setting, don't touch
    PrintLog(LOGCLOCK, "RTC alarm %s already set to %02i:%02i:%02i, day of month: %i\n",
             alarmstr, timeval->tm_hour, timeval->tm_min, timeval->tm_sec, timeval->tm_mday);

    //enable the alarm, if not already enabled
    criticalsection.Enter();
    RTC_AlarmCmd(rtcalarm, ENABLE);
    criticalsection.Leave();
  }
}

/*! Disables the RTC alarm.
    \param rtcalarm The RTC alarm to disable.*/
void DisableRTCAlarm(uint32_t rtcalarm)
{
  assert(rtcalarm == RTC_Alarm_A || rtcalarm == RTC_Alarm_B);

  const char* alarmstr = rtcalarm == RTC_Alarm_A ? "A" : "B";
  PrintLog(LOGCLOCK, "Disabling RTC alarm %s\n", alarmstr);

  CCritical criticalsection;
  RTC_AlarmCmd(rtcalarm, DISABLE);
  RTC_ClearFlag(rtcalarm);
}

/*! \param rtcalarm RTC_Alarm_A or RTC_Alarm_B.
    \return The state of RTC_FLAG_ALRAF or RTC_FLAG_ALRBF.*/
bool RTCAlarmFlagSet(uint32_t rtcalarm)
{
  assert(rtcalarm == RTC_Alarm_A || rtcalarm == RTC_Alarm_B);

  bool alarm = false;

  CCritical criticalsection;

  if (rtcalarm == RTC_Alarm_A)
    alarm = RTC_GetFlagStatus(RTC_FLAG_ALRAF);
  else if (rtcalarm == RTC_Alarm_B)
    alarm = RTC_GetFlagStatus(RTC_FLAG_ALRBF);

  return alarm;
}

/*! Gets the status of RTC_FLAG_ALRAF, RTC_FLAG_ALRBF and RTC_FLAG_WUTF.
    \return A bitwise OR of RTC_FLAG_ALRAF, RTC_FLAG_ALRBF and RTC_FLAG_WUTF,
            depending on which flags are set.*/
uint32_t RTCAlarmFlags()
{
  CCritical criticalsection;

  uint32_t alarms = 0;

  if (RTC_GetFlagStatus(RTC_FLAG_ALRAF))
    alarms |= RTC_FLAG_ALRAF;
  if (RTC_GetFlagStatus(RTC_FLAG_ALRBF))
    alarms |= RTC_FLAG_ALRBF;
  if (RTC_GetFlagStatus(RTC_FLAG_WUTF))
    alarms |= RTC_FLAG_WUTF;

  return alarms;
}

/*! Clears the RTC alarm flag, set by the RTC alarm.
    \param rtcalarm RTC_Alarm_A or RTC_Alarm_B.*/
void ClearRTCAlarmFlag(uint32_t rtcalarm)
{
  assert(rtcalarm == RTC_Alarm_A || rtcalarm == RTC_Alarm_B);

  CCritical criticalsection;

  if (rtcalarm == RTC_Alarm_A)
    RTC_ClearFlag(RTC_FLAG_ALRAF);
  else if (rtcalarm == RTC_Alarm_B)
    RTC_ClearFlag(RTC_FLAG_ALRBF);
}

static uint32_t rtcbootflags;

/*! Reads the RTC alarm flags and stores them in the rtcbootflags variable.
    This information can later be used to check why the microcontroller has
    woken up from standby mode.*/
void StoreRTCAlarmBootFlags()
{
  CCritical criticalsection;

  if (RTC_GetFlagStatus(RTC_FLAG_WUTF))
    rtcbootflags |= RTC_FLAG_WUTF;

  if (RTC_GetFlagStatus(RTC_FLAG_ALRAF))
    rtcbootflags |= RTC_FLAG_ALRAF;

  if (RTC_GetFlagStatus(RTC_FLAG_ALRBF))
    rtcbootflags |= RTC_FLAG_ALRBF;
}

/*! \return rtcbootflags, which is a bitwise OR of:
            RTC_FLAG_WUTF
            RTC_FLAG_ALRAF
            RTC_FLAG_ALRBF
            These flags indicate which RTC alarm might have
            woke the microcontroller from standby.*/
uint32_t GetRTCAlarmBootFlags()
{
  return rtcbootflags;
}

/*  Checks if a flag is set in flags, clears the flag,
    returns a printable string describing the flag.
    \param flag Pointer to a uint32_t holding a bitmask of:
                RTC_FLAG_WUTF, RTC_FLAG_ALRAF and RTC_FLAG_ALRBF.
    \return A printable string describing the flag.
    \return nullptr if no flag is set.*/
const char* RTCAlarmBootFlagStr(uint32_t* flags)
{
  if (*flags & RTC_FLAG_WUTF)
  {
    *flags &= ~(RTC_FLAG_WUTF);
    return "RTC wakeup timer";
  }
  else if (*flags & RTC_FLAG_ALRAF)
  {
    *flags &= ~(RTC_FLAG_ALRAF);
    return "RTC alarm A";
  }
  else if (*flags & RTC_FLAG_ALRBF)
  {
    *flags &= ~(RTC_FLAG_ALRBF);
    return "RTC alarm B";
  }
  else
  {
    return nullptr;
  }
}

/*! Writes the RTC time as an ISO8601 compatible string, as UTC.
    \param buf The buffer to write the time to, it should be at least
               DATESTR_MAX_LENGTH bytes long.
    \param bufsize The maximum number of bytes to write to buf.
    \param timezone A string that is put at the end of the timestamp.
    \param offset Time in seconds added to the timestamp.
                  Use a negative value to get a timestamp from the past.*/
void GetRTCISO8601(char* buf, uint32_t bufsize, const char* timezone /*= "Z"*/,
                   int64_t offset /*= 0*/)
{
  tm timeval = {};
  GetRTC(&timeval);

  //Add an offset to the timestamp if requested.
  if (offset != 0)
    SecsToTm(TmToSecs(&timeval) + offset, &timeval);

  strftime(buf, bufsize, "%Y-%m-%dT%H:%M:%S", &timeval);
  buf[bufsize - 1] = 0;

  if (strlen(timezone) > 0)
  {
    strlcat(buf, timezone, bufsize);
    buf[bufsize - 1] = 0;
  }
}

/*! Sets the RTC from an ISO8601 compatible string.
    \param buf A date/time string in ISO8601 compatible format.
    \return true if the time string is correct and the RTC has been set.
    \return false if the time string is incorrect.*/
bool SetRTCISO8601(char* buf)
{
  tm timeval = {};
  if (strptime(buf, "%Y-%m-%dT%H:%M:%S", &timeval) == nullptr)
    return false;

  SetRTC(&timeval);
  return true;
}
