#include "crc32hw.h"
#include "lock.h"

#include <stm32f4xx.h>

SemaphoreHandle_t CCRC32Hw::m_mutex;
bool              CCRC32Hw::m_enabled;

CCRC32Hw::CCRC32Hw()
{
}

CCRC32Hw::~CCRC32Hw()
{
}

/*! Enables the clock for the hardware crc generator.*/
void CCRC32Hw::EnableClock()
{
  //use a test and test-and-set method, to avoid entering
  //a critical section when not needed
  if (!m_enabled)
  {
    CCritical criticalsection;
    if (!m_enabled)
    {
      RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_CRC, ENABLE);
      m_enabled = true;
    }
  }
}

/*! Creates a mutex inside a critical section,
    for locking access around the hardware crc32 generator.*/
void CCRC32Hw::SetupMutex()
{
  //Use a test and test-and-set method, to avoid entering
  //a critical section when not needed, but don't allocate
  //while inside the critical section.
  if (m_mutex == nullptr)
  {
    SemaphoreHandle_t mutex = xSemaphoreCreateRecursiveMutex();

    CCritical criticalsection;
    if (m_mutex == nullptr)
    {
      m_mutex = mutex;
      mutex = nullptr;
    }
    criticalsection.Leave();

    if (mutex)
      vSemaphoreDelete(mutex);
  }
}

/*! Uses the hardware crc32 generator to generate a crc32 of a block of memory,
    with the polynomial 0x4C11DB7, which is the same as used for ethernet.
    \param buf The memory to create a crc32 of.
    \param size The number of 32 bit unsigned integers to calculate the crc of.
    \return The calculated crc32.*/
uint32_t CCRC32Hw::Calculate(uint32_t* buf, uint32_t size)
{
  EnableClock();

  if (size <= 512)
  { //on small sizes use a critical section for atomic access,
    //512 bytes takes about 5 microseconds to calculate
    CCritical criticalsection;
    return Generate(buf, size);
  }
  else
  { //on large sizes use a mutex for atomic access
    SetupMutex();
    CLock lock(m_mutex);
    return Generate(buf, size);
  }
}

uint32_t CCRC32Hw::Generate(uint32_t* buf, uint32_t size)
{
  CRC_ResetDR();

  uint32_t* end = buf + size;
  while (buf != end)
    CRC->DR = *(buf++);

  return CRC->DR;
}
