#include "dma.h"
#include "lock.h"

#include <cassert>

/*! Array to hold pointers to objects derived from the CDmaIrq class,
    one index per DMA stream, there are 2 DMA controllers and 8 streams per DMA controller.*/
CDmaIrq *g_irqhandlers[16];

/*! Macro to generate the dma interrupt handler functions.
    The functions will call the DMAIRQHandler() method of the corresponding g_irqhandler object.
    The index it uses corresponds to what the g_dma.DMANum() method returns.
    \param dmanum Number of the DMA controller, 1 or 2.
    \param streamnum Number of the stream of the DMA controller, 0 to 7.*/
#define DMA_IRQ_HANDLER(dmanum, streamnum) \
  extern "C" void DMA ## dmanum ## _Stream ## streamnum ## _IRQHandler() \
  { \
    uint32_t irqnum = ((dmanum) - 1) * 8 + (streamnum); \
    if (g_irqhandlers[irqnum]) \
      g_irqhandlers[irqnum]->DMAIRQHandler(); \
  }

DMA_IRQ_HANDLER(1, 0); //!< extern "C" void DMA1_Stream0_IRQHandler();
DMA_IRQ_HANDLER(1, 1); //!< extern "C" void DMA1_Stream1_IRQHandler();
DMA_IRQ_HANDLER(1, 2); //!< extern "C" void DMA1_Stream2_IRQHandler();
DMA_IRQ_HANDLER(1, 3); //!< extern "C" void DMA1_Stream3_IRQHandler();
DMA_IRQ_HANDLER(1, 4); //!< extern "C" void DMA1_Stream4_IRQHandler();
DMA_IRQ_HANDLER(1, 5); //!< extern "C" void DMA1_Stream5_IRQHandler();
DMA_IRQ_HANDLER(1, 6); //!< extern "C" void DMA1_Stream6_IRQHandler();
DMA_IRQ_HANDLER(1, 7); //!< extern "C" void DMA1_Stream7_IRQHandler();
DMA_IRQ_HANDLER(2, 0); //!< extern "C" void DMA2_Stream0_IRQHandler();
DMA_IRQ_HANDLER(2, 1); //!< extern "C" void DMA2_Stream1_IRQHandler();
DMA_IRQ_HANDLER(2, 2); //!< extern "C" void DMA2_Stream2_IRQHandler();
DMA_IRQ_HANDLER(2, 3); //!< extern "C" void DMA2_Stream3_IRQHandler();
DMA_IRQ_HANDLER(2, 4); //!< extern "C" void DMA2_Stream4_IRQHandler();
DMA_IRQ_HANDLER(2, 5); //!< extern "C" void DMA2_Stream5_IRQHandler();
DMA_IRQ_HANDLER(2, 6); //!< extern "C" void DMA2_Stream6_IRQHandler();
DMA_IRQ_HANDLER(2, 7); //!< extern "C" void DMA2_Stream7_IRQHandler();

CDma g_dma; //! Global object used to interface with the DMA code.

CDma::CDma()
{
  m_streamlock = 0;
}

/*! Called from BoardInit(), enables the two DMA peripherals and creates the event group for locking the streams.*/
void CDma::Setup()
{
  //make sure this function is called only once
  assert(m_streamlock == 0);

  //enable the two DMA peripherals
  CCritical critsection;
  RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_DMA1, ENABLE);
  RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_DMA2, ENABLE);
  critsection.Leave();

  //create an event group, where each bit in the event group is used as a mutex for a dma stream
  //this saves allocating 16 mutexes
  m_streamlock = xEventGroupCreate();

  //set all bits to indicate that all streams are free
  xEventGroupSetBits(m_streamlock, 0xFFFF);
}

/*! Waits until the bit that belongs to this dma stream becomes set,
    and automatically clear it when it does, signaling that this stream is now locked.
    \param stream A pointer to a dma stream.*/
void CDma::LockStream(DMA_Stream_TypeDef* stream)
{
  assert(IS_DMA_ALL_PERIPH(stream));

  xEventGroupWaitBits(m_streamlock, DMABit(stream), pdTRUE, pdFALSE, portMAX_DELAY);
}

/*! Sets the bit that belongs to this dma stream, indicating that it's now unlocked.
    \param stream A pointer to a dma stream.*/
void CDma::UnlockStream(DMA_Stream_TypeDef* stream)
{
  assert(IS_DMA_ALL_PERIPH(stream));

  xEventGroupSetBits(m_streamlock, DMABit(stream));
}

/*! Installs an irq handler in g_irqhandlers, and enables the irq corresponding to the stream in the NVIC.
    \param stream The DMA stream of which the IRQ should be enabled.
    \param irqhandler Pointer to an object that derives from CDmaIrq, of which the DMAIRQHandler() method is called.
    \param priority Priority of the DMA interrupt.*/
void CDma::EnableIrq(DMA_Stream_TypeDef* stream, CDmaIrq* irqhandler, uint8_t priority)
{
  //make sure the stream is locked
  assert(IS_DMA_ALL_PERIPH(stream));
  assert((xEventGroupGetBits(m_streamlock) & DMABit(stream)) == 0);

  //use a critical section to protect the irq handlers and nvic registers
  CCritical critsection;

  //install the irq handler, it should be a pointer to a class that derives from CDmaIrq,
  //and implements the DMAIRQHandler() virtual method, this method is then
  //called from the corresponding irq handler function of the stream
  uint32_t dmanum = DMANum(stream);
  assert(g_irqhandlers[dmanum] == NULL);
  g_irqhandlers[dmanum] = irqhandler;

  //enable the IRQ that matches this stream
  NVIC_InitTypeDef nvicsettings =
  {
    .NVIC_IRQChannel                   = (uint8_t)DMAIRQn(stream),
    .NVIC_IRQChannelPreemptionPriority = (uint8_t)priority,
    .NVIC_IRQChannelSubPriority        = 0,
    .NVIC_IRQChannelCmd                = ENABLE,
  };
  NVIC_Init(&nvicsettings);
}

/*! Disables the IRQ of the DMA stream, and removes its IRQ handler.
    \param stream DMA stream of which to disable the interrupt.*/
void CDma::DisableIrq(DMA_Stream_TypeDef* stream)
{
  //make sure the stream is locked
  assert(IS_DMA_ALL_PERIPH(stream));
  assert((xEventGroupGetBits(m_streamlock) & DMABit(stream)) == 0);

  //use a critical section to protect the irq handlers and nvic registers
  CCritical critsection;

  //disable the IRQ that matches this stream
  NVIC_InitTypeDef nvicsettings =
  {
    .NVIC_IRQChannel                   = (uint8_t)DMAIRQn(stream),
    .NVIC_IRQChannelPreemptionPriority = 0,
    .NVIC_IRQChannelSubPriority        = 0,
    .NVIC_IRQChannelCmd                = DISABLE,
  };
  NVIC_Init(&nvicsettings);

  //remove the irq handler for this stream
  uint32_t dmanum = DMANum(stream);
  assert(g_irqhandlers[dmanum] != NULL);
  g_irqhandlers[dmanum] = NULL;
}

/*! Array to convert a dma stream to an index number, or an irq number.
    The order corresponds to what the DMA_IRQ_HANDLER macro uses, it should not be changed.*/
static const struct
{
  DMA_Stream_TypeDef* stream;
  IRQn_Type           irqn;
}
g_streams[] =
{
  { DMA1_Stream0, DMA1_Stream0_IRQn },
  { DMA1_Stream1, DMA1_Stream1_IRQn },
  { DMA1_Stream2, DMA1_Stream2_IRQn },
  { DMA1_Stream3, DMA1_Stream3_IRQn },
  { DMA1_Stream4, DMA1_Stream4_IRQn },
  { DMA1_Stream5, DMA1_Stream5_IRQn },
  { DMA1_Stream6, DMA1_Stream6_IRQn },
  { DMA1_Stream7, DMA1_Stream7_IRQn },
  { DMA2_Stream0, DMA2_Stream0_IRQn },
  { DMA2_Stream1, DMA2_Stream1_IRQn },
  { DMA2_Stream2, DMA2_Stream2_IRQn },
  { DMA2_Stream3, DMA2_Stream3_IRQn },
  { DMA2_Stream4, DMA2_Stream4_IRQn },
  { DMA2_Stream5, DMA2_Stream5_IRQn },
  { DMA2_Stream6, DMA2_Stream6_IRQn },
  { DMA2_Stream7, DMA2_Stream7_IRQn }
};

/*! Converts a DMA_Stream_TypeDef* into a bit number to be used with m_streamlock.
    \param stream A pointer to a dma stream.
    \return The bit that represents stream.*/
uint32_t CDma::DMABit(DMA_Stream_TypeDef* stream)
{
  assert(IS_DMA_ALL_PERIPH(stream));

  for (uint32_t i = 0; i < sizeof(g_streams) / sizeof(g_streams[0]); i++)
  {
    if (g_streams[i].stream == stream)
      return 1 << i;
  }

  assert(0);
  return 0;
}

/*! Converts a DMA_Stream_TypeDef* into its matching irq number to be used with the NVIC.
    \param stream A pointer to a dma stream.
    \return A value of IRQn_Type.*/
IRQn_Type CDma::DMAIRQn(DMA_Stream_TypeDef* stream)
{
  assert(IS_DMA_ALL_PERIPH(stream));

  for (uint32_t i = 0; i < sizeof(g_streams) / sizeof(g_streams[0]); i++)
  {
    if (g_streams[i].stream == stream)
      return g_streams[i].irqn;
  }

  assert(0);
  return (IRQn)0;
}

/*! Converts a DMA_Stream_TypeDef* into an index number to be used with the g_irqhandlers array.
    The index it returns corresponds to what the DMA_IRQ_HANDLER uses, it should not be changed.
    \param stream A pointer to a dma stream.
    \return The index of stream.*/
uint32_t CDma::DMANum(DMA_Stream_TypeDef* stream)
{
  assert(IS_DMA_ALL_PERIPH(stream));

  for (uint32_t i = 0; i < sizeof(g_streams) / sizeof(g_streams[0]); i++)
  {
    if (g_streams[i].stream == stream)
      return i;
  }

  assert(0);
  return 0;
}
