#include "externalrtc.h"
#include "log.h"
#include "clock.h"
#include "lock.h"

#define PCF8563_WRITE_ADDR            0xA2
#define PCF8563_READ_ADDR             0xA3

#define PCF8563_Control_status_1      0x00
#define PCF8563_Control_status_2      0x01
#define PCF8563_VL_seconds            0x02
#define PCF8563_Minutes               0x03
#define PCF8563_Hours                 0x04
#define PCF8563_Days                  0x05
#define PCF8563_Weekdays              0x06
#define PCF8563_Century_months        0x07
#define PCF8563_Years                 0x08
#define PCF8563_Minute_alarm          0x09
#define PCF8563_Hour_alarm            0x0A
#define PCF8563_Day_alarm             0x0B
#define PCF8563_Weekday_alarm         0x0C
#define PCF8563_CLKOUT_control        0x0D
#define PCF8563_Timer_control         0x0E
#define PCF8563_Timer                 0x0F

#define PCF8563_NUM_REGISTERS         16

#define PCF8563_TESTC                 (1 << 3) //!< TESTC bit in Control_status_1
#define PCF8563_VL                    (1 << 7) //!< VL bit in VL_seconds
#define PCF8563_TI_TP                 (1 << 4) //!< TI_TP bit in Control_status_2
#define PCF8563_AE                    (1 << 7) //!< AE bit in the alarm registers
#define PCF8563_TD_1_60TH             (3 << 0) //!< TD bits in Timer_control
#define PCF8563_AIE                   (1 << 1) //!< AIE bit in Control_status_2
#define PCF8563_AF                    (1 << 3) //!< AF bit in Control_status_2

#define PCF8563_HOUR_ALARM_MASK       0xBF //!< And mask to set the unused bit to 0.
#define PCF8563_CONTROL_STATUS_2_MASK 0x1F //!< And mask to set the unused bits to 0.
#define PCF8563_SECONDS_MASK          0x7F //!< And mask to set the unused bits to 0.
#define PCF8563_MINUTES_MASK          0x7F //!< And mask to set the unused bits to 0.
#define PCF8563_HOURS_MASK            0x3F //!< And mask to set the unused bits to 0.

#define PCF8563_RETRY_TIME            1000

/*! \param i2c Pointer to an instance of CI2C that should be set up previously.
    \param reboothold The pin to make high to prevent a reboot from happening on an RTC interrupt.*/
CExternalRTC::CExternalRTC()
{
  m_i2c        = nullptr;
  m_reboothold = {};
}

void CExternalRTC::Setup(CI2C* i2c, SGPIOPin reboothold,
                         bool checkvlow /*= true*/, bool forcereset /* = false*/)
{
  m_i2c        = i2c;
  m_reboothold = reboothold;

  //Check if the RTC has had a power cycle,
  //in which case all registers need to be set.
  if (checkvlow)
    CheckVlow();

  //If requested, reset all registers.
  if (forcereset)
  {
    PrintLog(LOGEXTRTC, "Force resetting all registers by request\n");
    ResetRegisters();
  }
}

/* Tests if the time increased by between 4 and 6 seconds when waiting for 5 seconds.
   \return true if the test passed.
   \return false if the test failed.*/
bool CExternalRTC::Test()
{
  PrintLog(LOGEXTRTC, "Testing if external RTC runs\n");

  //take a start timestamp of the external RTC
  uint32_t teststart = Micros();
  uint8_t  start;
  if (!RTCSeconds(&start))
  {
    PrintLog(LOGEXTRTC, "ERROR: Unable to test if external RTC is working\n");
    return false;
  }

  DelayMS(5000);

  //take an end timestamp of the RTC after waiting for 5 seconds
  uint8_t end;
  if (!RTCSeconds(&end))
  {
    PrintLog(LOGEXTRTC, "ERROR: Unable to test if external RTC is working\n");
    return false;
  }
  uint32_t testduration = Micros() - teststart;

  //calculate how far the external RTC has advanced, and take into account the
  //wrapping of the seconds register at 60
  int32_t elapsed = (int32_t)end - (int32_t)start;
  if (elapsed < 0)
    elapsed += 60;

  //if the RTC has advanced between 4 and 6 seconds while waiting for 5 seconds
  //assume it's working correctly
  if (elapsed >= 4 && elapsed <= 6)
  {
    PrintLog(LOGEXTRTC, "External RTC is working correctly, time increased by %" PRIi32
             " seconds in %" PRIu32 " microseconds\n", elapsed, testduration);
    return true;
  }
  else
  {
    PrintLog(LOGEXTRTC, "ERROR: External RTC is not working correctly, time increased by %" PRIi32
             " seconds in %" PRIu32 " microseconds\n", elapsed, testduration);
    return false;
  }
}

/* Sets up an interrupt on the PCF8563 RTC which triggers every 24 hours, its interrupt pin
   is connected to a one-shot circuit which turns off the power supply for about 1.5 second.*/
void CExternalRTC::SetupReset()
{
  //while interfacing with the external RTC,
  //disable the hardware power cycle circuit
  GPIO_Set(m_reboothold);

  //try a few times to set up the external RTC
  const uint32_t tries = 5;
  for (uint32_t i = 0; i < tries; i++)
  {
    //the PCF8563 might see the initial power on as a start condition
    //send a stop to make sure that it's ready to receive a start
    DelayMS(10);
    m_i2c->Stop();
    DelayMS(10);

    if (!CheckVlow() || !CheckAlarm())
    {
      DelayMS(PCF8563_RETRY_TIME);
      continue;
    }

    break;
  }

  GPIO_Reset(m_reboothold);
}

/*! Resets all registers to a default value, all time registers set to 0
    and all alarms disabled.*/
bool CExternalRTC::ResetRegisters()
{
  static const uint8_t registers[]
  {
    PCF8563_TESTC,     //Control_status_1
    PCF8563_TI_TP,     //Control_status_2
    0,                 //VL_seconds
    0,                 //Minutes
    0,                 //Hours
    0,                 //Days
    0,                 //Weekdays
    0,                 //Century_months
    0,                 //Years
    PCF8563_AE,        //Minute_alarm
    PCF8563_AE,        //Hour_alarm
    PCF8563_AE,        //Day_alarm
    PCF8563_AE,        //Weekday_alarm
    0,                 //CLKOUT_control
    PCF8563_TD_1_60TH, //Timer_control
    0,                 //Timer
  };

  //While interfacing with the external RTC,
  //disable the hardware power cycle circuit.
  GPIO_Set(m_reboothold);

  bool success = true;
  if (!WriteRegisters(0, registers, sizeof(registers)))
  {
    PrintLog(LOGEXTRTC, "ERROR: unable to reset registers of external RTC\n");
    success = false;
  }

  GPIO_Reset(m_reboothold);

  return success;
}

/*! Reads one or more registers from the PCF8563 RTC.
    \param reg The address of the first register to read.
    \param buf Where to store the register values, should be an array of uint8_t[count].
    \param count The number of registers to read.
    \return true if the read succeeded, false it it failed.*/
bool CExternalRTC::ReadRegisters(uint8_t reg, uint8_t* buf, uint32_t count)
{
  CLock lock(*m_i2c);

  bool success;

  //write the register address to the PCF8563, then send a stop command
  //this sets the internal pointer of the PCF8563 to the register address
  success = m_i2c->Write(PCF8563_WRITE_ADDR, (const uint8_t[]){reg});
  m_i2c->Stop();
  if (!success)
  {
    PrintLog(LOGEXTRTC, "ERROR: write to I2C address %02X register address %" PRIu8 " failed\n",
             PCF8563_WRITE_ADDR, reg);
    return false;
  }

  //do a read from the PCF8563, it will start reading at the register address
  //that was previously written to it
  success = m_i2c->Read(PCF8563_READ_ADDR, buf, count);
  m_i2c->Stop();
  if (!success)
  {
    PrintLog(LOGEXTRTC, "ERROR: read from I2C address %02X failed at byte %" PRIu32 "\n",
             PCF8563_READ_ADDR, m_i2c->AckFail());
    return false;
  }

  return true;
}

/*! Writes one or more registers of the PCF8563 RTC.
    \param reg The address of the first register to write.
    \param buf The register values to write, should be an array of uint8_t[count].
    \param count The number of registers to write.
    \return true if the write succeeded, false if it failed.*/
bool CExternalRTC::WriteRegisters(uint8_t reg, const uint8_t* buf, uint32_t count)
{
  CLock lock(*m_i2c);

  //first the register address is written to the PCF8563,
  //after that all the register values are written
  bool success = m_i2c->Write(PCF8563_WRITE_ADDR, &reg, 1, buf, count);
  if (!success)
    PrintLog(LOGEXTRTC, "ERROR: write to I2C address %02X failed at byte %" PRIu32 "\n",
             PCF8563_WRITE_ADDR, m_i2c->AckFail());
  m_i2c->Stop();

  return success;
}

/*! Checks the Vlow bit in the VL_second register, if it's set it means that
    the supply voltage of the PCF8563 has dropped below a certain value, and all the
    time registers have been set to an undefined state.
    This function will then reset all those registers to a default value,
    with all the time registers set to 0 and all alarms disabled.
    \return true if reading and writing the registers succeeded, false if anything failed.*/
bool CExternalRTC::CheckVlow()
{
  PrintLog(LOGEXTRTC, "Checking Vlow bit\n");

  uint8_t vlseconds;
  if (!ReadRegisters(PCF8563_VL_seconds, &vlseconds, 1))
    return false;

  if (vlseconds & PCF8563_VL)
  {
    PrintLog(LOGEXTRTC, "Vlow bit set, resetting RTC\n");
    if (!ResetRegisters())
      return false;
  }
  else
  {
    PrintLog(LOGEXTRTC, "Vlow bit not set, no reset needed\n");
  }

  return true;
}

bool CExternalRTC::CheckAlarm()
{
  //read all registers in one go, this always produces a correct timestamp
  uint8_t registers[PCF8563_NUM_REGISTERS];
  if (!ReadRegisters(0, registers, PCF8563_NUM_REGISTERS))
    return false;

  //print the current time the external RTC is set at
  PrintLog(LOGEXTRTC, "External RTC time: %02" PRIX8 ":%02" PRIX8 ":%02" PRIX8 "\n",
           registers[PCF8563_Hours]      & PCF8563_HOURS_MASK,
           registers[PCF8563_Minutes]    & PCF8563_MINUTES_MASK,
           registers[PCF8563_VL_seconds] & PCF8563_SECONDS_MASK);

  if (registers[PCF8563_Control_status_2] & PCF8563_AF)
    PrintLog(LOGEXTRTC, "Alarm interrupt occurred\n");

  //wanted settings, alarm interrupt enabled at 12:00
  //this will make the 24 hour reset happen outside office hours
  //because the RTC's time is initialized to 0:00
  uint8_t houralarm      = 0x12;
  uint8_t minutealarm    = 0x00;
  uint8_t controlstatus2 = PCF8563_TI_TP | PCF8563_AIE;

  //if the RTC is not set to the wanted settings, write the new settings
  //this also clears the interrupt alarm flag
  if ((registers[PCF8563_Minute_alarm])                                     != minutealarm ||
      (registers[PCF8563_Hour_alarm] & PCF8563_HOUR_ALARM_MASK)             != houralarm   ||
      (registers[PCF8563_Control_status_2] & PCF8563_CONTROL_STATUS_2_MASK) != controlstatus2)
  {
    PrintLog(LOGEXTRTC, "Setting alarm interrupt to %02" PRIX8 ":%02" PRIX8 "\n",
             houralarm, minutealarm);

    if (!WriteRegisters(PCF8563_Minute_alarm, &minutealarm, 1))
      return false;

    if (!WriteRegisters(PCF8563_Hour_alarm, &houralarm, 1))
      return false;

    if (!WriteRegisters(PCF8563_Control_status_2, &controlstatus2, 1))
      return false;
  }
  else
  {
    PrintLog(LOGEXTRTC, "Alarm interrupt is set to %02" PRIX8 ":%02" PRIX8 "\n",
             registers[PCF8563_Hour_alarm] & PCF8563_HOUR_ALARM_MASK,
             registers[PCF8563_Minute_alarm]);
  }

  return true;
}

/*! Reads the seconds register.
    \param seconds The value of the seconds register is returned here.
    \return true if reading the seconds register succeeded.
    \return false if reading the seconds register failed.*/
bool CExternalRTC::RTCSeconds(uint8_t* seconds)
{
  //read all registers in one go, this always produces a correct timestamp
  uint8_t registers[PCF8563_NUM_REGISTERS];
  if (!ReadRegisters(0, registers, PCF8563_NUM_REGISTERS))
    return false;

  //the seconds register is binary coded decimal, convert to binary
  uint8_t secondsreg = registers[PCF8563_VL_seconds] & PCF8563_SECONDS_MASK;
  *seconds = (secondsreg >> 4) * 10 + (secondsreg & 0xF);

  return true;
}
