#include "exti.h"
#include "lock.h"
#include "clock.h"

#include <cassert>
#include <cstddef>

CExti g_exti;

/*! Interrupt handler for EXTI line 0, gets called when a pin interrupt is configured for it.
    I calls into g_exti with the corresponding pin number.*/
extern "C" void EXTI0_IRQHandler()
{
  EXTI_ClearFlag(EXTI_Line0);
  g_exti.OnGPIOIrq(0);
}

/*! Interrupt handler for EXTI line 1, gets called when a pin interrupt is configured for it.
    I calls into g_exti with the corresponding pin number.*/
extern "C" void EXTI1_IRQHandler()
{
  EXTI_ClearFlag(EXTI_Line1);
  g_exti.OnGPIOIrq(1);
}

/*! Interrupt handler for EXTI line 2, gets called when a pin interrupt is configured for it.
    I calls into g_exti with the corresponding pin number.*/
extern "C" void EXTI2_IRQHandler()
{
  EXTI_ClearFlag(EXTI_Line2);
  g_exti.OnGPIOIrq(2);
}

/*! Interrupt handler for EXTI line 3, gets called when a pin interrupt is configured for it.
    I calls into g_exti with the corresponding pin number.*/
extern "C" void EXTI3_IRQHandler()
{
  EXTI_ClearFlag(EXTI_Line3);
  g_exti.OnGPIOIrq(3);
}

/*! Interrupt handler for EXTI line 4, gets called when a pin interrupt is configured for it.
    I calls into g_exti with the corresponding pin number.*/
extern "C" void EXTI4_IRQHandler()
{
  EXTI_ClearFlag(EXTI_Line4);
  g_exti.OnGPIOIrq(4);
}

/*! Called from the shared interrupt handlers for EXTI lines 5 to 9, and 10 to 15.
    It checks and clears all the corresponding EXTI interrupt bits first,
    then calls ingo g_exti with the corresponding pin number.
    This is to make sure that the next pin interrupt can be caught as soon as possible.
    \param irqlines Array of EXTI_Line defines.
    \param numirqlines Size of the irqlines array.
    \param irqstart Number of the first irq line in irqlines.*/
void OnSharedInterrupt(uint32_t* irqlines, uint32_t numirqlines, uint32_t irqstart)
{
  bool triggered[numirqlines] = {};

  //check if the EXTI line has been triggered, if so set a flag in triggered
  //and clear the bit in the EXTI
  for (uint32_t i = 0; i < numirqlines; i++)
  {
    if (EXTI_GetITStatus(irqlines[i]))
    {
      EXTI_ClearFlag(irqlines[i]);
      triggered[i] = true;
    }
  }

  //call into g_exti for each pin that had an interrupt
  for (uint32_t i = 0; i < numirqlines; i++)
  {
    if (triggered[i])
      g_exti.OnGPIOIrq(irqstart + i);
  }
}

/*! Interrupt handler for EXTI lines 5 to 9, gets called when a pin interrupt is configured for it.*/
extern "C" void EXTI9_5_IRQHandler()
{
  uint32_t irqlines[] =
  {
    EXTI_Line5,
    EXTI_Line6,
    EXTI_Line7,
    EXTI_Line8,
    EXTI_Line9
  };

  OnSharedInterrupt(irqlines, sizeof(irqlines) / sizeof(irqlines[0]), 5);
}

/*! Interrupt handler for EXTI lines 10 to 15, gets called when a pin interrupt is configured for it.*/
extern "C" void EXTI15_10_IRQHandler()
{
  uint32_t irqlines[] =
  {
    EXTI_Line10,
    EXTI_Line11,
    EXTI_Line12,
    EXTI_Line13,
    EXTI_Line14,
    EXTI_Line15
  };

  OnSharedInterrupt(irqlines, sizeof(irqlines) / sizeof(irqlines[0]), 10);
}

/*! Interrupt handler for EXTI line 17, which is the RTC alarm interrupt line.*/
extern "C" void RTC_Alarm_IRQHandler()
{
  EXTI_ClearFlag(EXTI_Line17);

  uint32_t alarms = RTCAlarmFlags();

  //RTC alarm A and alarm B have the same interrupt handler. Check which interrupt
  //has happened by checking the RTC alarm flags.
  if (alarms & RTC_FLAG_ALRAF)
    g_exti.OnRTCIrq(RTCIRQAlarmA);
  if (alarms & RTC_FLAG_ALRBF)
    g_exti.OnRTCIrq(RTCIRQAlarmB);
}

/*! Interrupt handler for EXTI line 22, which is the RTC wakeup interrupt line.*/
extern "C" void RTC_WKUP_IRQHandler()
{
  EXTI_ClearFlag(EXTI_Line22);
  g_exti.OnRTCIrq(RTCIRQWakeup);
}

CExti::CExti() : m_gpioirqhandlers {}, m_rtcirqhandlers {}
{
}

void CExti::EnableGPIOIrq(SGPIOPin pin, EXTITrigger_TypeDef type, CGPIOInterrupt* irqhandler)
{
  //protect the registers and irq handlers using a critical section
  CCritical critsection;

  //make sure that no irq handler has been installed yet for this pin
  uint32_t pinnum = GPIONumByPin(pin.pin);
  assert(m_gpioirqhandlers[pinnum].irqhandler == nullptr);
  assert(irqhandler != nullptr);

  //install the irq handler
  m_gpioirqhandlers[pinnum].pin        = pin;
  m_gpioirqhandlers[pinnum].irqhandler = irqhandler;

  //the EXTI has 16 gpio interrupt lines, one for each gpio on a port,
  //the port has to be selected using the system configuration controller, which is enabled in BoardInit()
  //this means that the whole microcontroller can only use 16 pins for interrupts at the same time
  //and those 16 pins have to be different pin numbers, even if they are on different ports
  SYSCFG_EXTILineConfig(EXTIPortSourceByPort(pin.port), GPIONumByPin(pin.pin));

  uint32_t extiline = ExtiLineByNum(pinnum);
  EXTI_ClearFlag(extiline); //clear the interrupt pending bit

  //enable the exti line in the requested mode
  EXTI_InitTypeDef extisettings =
  {
    .EXTI_Line    = extiline,
    .EXTI_Mode    = EXTI_Mode_Interrupt,
    .EXTI_Trigger = type,
    .EXTI_LineCmd = ENABLE
  };
  EXTI_Init(&extisettings);

  //enable the interrupt, if not already enabled
  SetupInterrupt(pinnum, true);
}

void CExti::DisableGPIOIrq(SGPIOPin pin)
{
  //protect the registers and irq handlers using a critical section
  CCritical critsection;

  //if no irq handler has been installed for this pin, stop
  uint32_t pinnum = GPIONumByPin(pin.pin);
  if (m_gpioirqhandlers[pinnum].irqhandler == nullptr)
    return;

  //disable the pin interrupt
  SetupInterrupt(pinnum, false);

  uint32_t extiline = ExtiLineByNum(pinnum);

  //disable the EXTI line
  EXTI_InitTypeDef extisettings =
  {
    .EXTI_Line    = extiline,
    .EXTI_Mode    = EXTI_Mode_Interrupt,
    .EXTI_Trigger = EXTI_Trigger_Rising,
    .EXTI_LineCmd = DISABLE
  };
  EXTI_Init(&extisettings);

  //clear the interrupt pending bit
  EXTI_ClearFlag(extiline); 

  //remove the irq handler
  m_gpioirqhandlers[pinnum].pin        = {};
  m_gpioirqhandlers[pinnum].irqhandler = nullptr;
}

/*! Enables the corresponding RTC IRQ in the EXIT and NVIC, and installs
    a pointer to an IRQ handler class, who's RTCIRQHandler() method will
    be called when the corresponding RTC IRQ happens.
    The IRQ in the RTC should be enabled separately using RTC_ITConfig().
    \param type The type of RTC IRQ to enable.
    \param irqhandler A pointer to an irq handler class to call.*/
void CExti::EnableRTCIrq(ERTCIRQType type, CRTCInterrupt* irqhandler)
{
  //protect the registers and irq handlers using a critical section
  CCritical critsection;

  //install the irq handler
  assert(m_rtcirqhandlers[type] == nullptr);
  assert(irqhandler != nullptr);
  m_rtcirqhandlers[type] = irqhandler;

  uint32_t extiline = ExtiLineByRTC(type);
  EXTI_ClearFlag(extiline); //clear the interrupt pending bit

  //enable the exti line in rising edge mode
  EXTI_InitTypeDef extisettings =
  {
    .EXTI_Line    = extiline,
    .EXTI_Mode    = EXTI_Mode_Interrupt,
    .EXTI_Trigger = EXTI_Trigger_Rising,
    .EXTI_LineCmd = ENABLE
  };
  EXTI_Init(&extisettings);

  //enable the EXTI interrupt in the NVIC
  NVIC_InitTypeDef nvicsettings =
  {
    .NVIC_IRQChannel                   = (uint8_t)ExtiIrqByRTC(type),
    .NVIC_IRQChannelPreemptionPriority = configLIBRARY_MAX_SYSCALL_INTERRUPT_PRIORITY,
    .NVIC_IRQChannelSubPriority        = 0,
    .NVIC_IRQChannelCmd                = ENABLE
  };
  NVIC_Init(&nvicsettings);
}

/*! Disables the corresponding RTC IRQ in the EXIT and NVIC, and removes
    the pointer to the IRQ handler class.
    The IRQ in the RTC should be disabled separately using RTC_ITConfig().
    \param type The type of RTC IRQ to disable.*/
void CExti::DisableRTCIrq(ERTCIRQType type)
{
  //protect the registers and irq handlers using a critical section
  CCritical critsection;

  //if no irq handler has been installed for this RTC IRQ, stop
  if (m_rtcirqhandlers[type] == nullptr)
    return;

  //disable the EXTI interrupt in the NVIC
  NVIC_InitTypeDef nvicsettings =
  {
    .NVIC_IRQChannel                   = (uint8_t)ExtiIrqByRTC(type),
    .NVIC_IRQChannelPreemptionPriority = configLIBRARY_MAX_SYSCALL_INTERRUPT_PRIORITY,
    .NVIC_IRQChannelSubPriority        = 0,
    .NVIC_IRQChannelCmd                = DISABLE
  };
  NVIC_Init(&nvicsettings);

  uint32_t extiline = ExtiLineByRTC(type);

  //disable the EXTI line
  EXTI_InitTypeDef extisettings =
  {
    .EXTI_Line    = extiline,
    .EXTI_Mode    = EXTI_Mode_Interrupt,
    .EXTI_Trigger = EXTI_Trigger_Rising,
    .EXTI_LineCmd = DISABLE
  };
  EXTI_Init(&extisettings);

  //clear the interrupt pending bit
  EXTI_ClearFlag(extiline); 

  //remove the irq handler
  m_rtcirqhandlers[type] = nullptr;
}

/*! Convers an exti line number to a value of @ref EXTI_Lines.
    \param num The number of the exti line.
    \return a value of @ref EXTI_Lines.*/
uint32_t CExti::ExtiLineByNum(uint32_t num)
{
  static const uint32_t extilines[] =
  {
    EXTI_Line0,
    EXTI_Line1,
    EXTI_Line2,
    EXTI_Line3,
    EXTI_Line4,
    EXTI_Line5,
    EXTI_Line6,
    EXTI_Line7,
    EXTI_Line8,
    EXTI_Line9,
    EXTI_Line10,
    EXTI_Line11,
    EXTI_Line12,
    EXTI_Line13,
    EXTI_Line14,
    EXTI_Line15,
    EXTI_Line16,
    EXTI_Line17,
    EXTI_Line18,
    EXTI_Line19,
    EXTI_Line20,
    EXTI_Line21,
    EXTI_Line22,
    EXTI_Line23
  };

  assert(num < sizeof(extilines) / sizeof(extilines[0]));

  return extilines[num];
}

/*! Convers an RTC interrupt type to a value of @ref EXTI_Lines.
    \param irqtype The interrupt type of the RTC.
    \return a value of @ref EXTI_Lines.*/
uint32_t CExti::ExtiLineByRTC(ERTCIRQType irqtype)
{
  if (irqtype == RTCIRQAlarmA || irqtype == RTCIRQAlarmB)
    return EXTI_Line17;
  else if (irqtype == RTCIRQWakeup)
    return EXTI_Line22;
  else
    assert(0);
}

/*! Converts a pin number to a value of @ref IRQn to be used with NVIC_Init().
    \param num The pin number, 0 to 15.
    \return An EXTI value of IRQn.*/
IRQn CExti::ExtiIrqByNum(uint32_t num)
{
  assert(num <= 15);

  if (num == 0)
    return EXTI0_IRQn;
  else if (num == 1)
    return EXTI1_IRQn;
  else if (num == 2)
    return EXTI2_IRQn;
  else if (num == 3)
    return EXTI3_IRQn;
  else if (num == 4)
    return EXTI4_IRQn;
  else if (num <= 9)
    return EXTI9_5_IRQn; //EXTI lines 5 to 9 share the same interrupt
  else if (num <= 15)
    return EXTI15_10_IRQn; //EXTI lines 10 to 15 share the same interrupt
  else
    assert(0);

  return (IRQn)0;
}

/*! Convers an RTC interrupt type to a value of of @ref IRQn to be used with NVIC_Init().
    \param rqtype The interrupt type of the RTC.
    \return a value of @ref EXTI_Lines.*/
IRQn CExti::ExtiIrqByRTC(ERTCIRQType irqtype)
{
  if (irqtype == RTCIRQAlarmA || irqtype == RTCIRQAlarmB)
    return RTC_Alarm_IRQn;
  else if (irqtype == RTCIRQWakeup)
    return RTC_WKUP_IRQn;
  else
    assert(0);
}

void CExti::SetupInterrupt(uint32_t pinnum, bool enable, uint8_t priority /*= configLIBRARY_MAX_SYSCALL_INTERRUPT_PRIORITY*/)
{
  if (pinnum <= 4)
  {
    //exti lines 0 to 4 have their own interrupt, so they can be enabled or disabled directly
    NVIC_InitTypeDef nvicsettings =
    {
      .NVIC_IRQChannel                   = (uint8_t)ExtiIrqByNum(pinnum),
      .NVIC_IRQChannelPreemptionPriority = priority,
      .NVIC_IRQChannelSubPriority        = 0,
      .NVIC_IRQChannelCmd                = enable ? ENABLE : DISABLE
    };
    NVIC_Init(&nvicsettings);
  }
  else if (pinnum <= 9)
  { //EXTI lines 5 to 9 share the same interrupt
    SetSharedInterrupt(5, 9, pinnum, enable, priority);
  }
  else if (pinnum <= 15)
  { //EXTI lines 10 to 15 share the same interrupt
    SetSharedInterrupt(10, 15, pinnum, enable, priority);
  }
}

/*! Checks if there is another interrupt handler installed on the EXTI lines start to end,
    if not, it will set the shared interrupt to the state of enable.
    \param start EXTI line to start the check at, should be 5 or 10.
    \param end EXTI line to end the check at, should be 9 or 15.
    \param pinnum Pin number of which to enable the EXTI line interrup for, should be from 0 to 15.
    \param enable Set to true to enable the interrupt, set to false to disable the interrupt.
    \param priority NVIC priority of the interrupt.*/
void CExti::SetSharedInterrupt(uint32_t start, uint32_t end, uint32_t pinnum, bool enable, uint8_t priority)
{
  assert((start == 5 && end == 9) || (start == 10 && end == 15));
  assert(pinnum <= 15);
  //check if there is another interrupt handler installed on the shared EXTI lines
  bool found = false;
  for (uint32_t i = start; i <= end; i++)
  {
    if (i == pinnum)
      continue;

    if (m_gpioirqhandlers[i].irqhandler)
    {
      found = true;
      break;
    }
  }

  //if there is no other irq handler installed on the shared EXTI lines,
  //it can be enabled/disabled
  if (!found)
  {
    NVIC_InitTypeDef nvicsettings =
    {
      .NVIC_IRQChannel                   = (uint8_t)ExtiIrqByNum(pinnum),
      .NVIC_IRQChannelPreemptionPriority = priority,
      .NVIC_IRQChannelSubPriority        = 0,
      .NVIC_IRQChannelCmd                = enable ? ENABLE : DISABLE
    };
    NVIC_Init(&nvicsettings);
  }
}

void CExti::OnGPIOIrq(uint32_t pinnum)
{
  assert(pinnum < PINS_PER_PORT);

  if (m_gpioirqhandlers[pinnum].irqhandler)
    m_gpioirqhandlers[pinnum].irqhandler->GPIOIRQHandler(m_gpioirqhandlers[pinnum].pin);
}

void CExti::OnRTCIrq(ERTCIRQType type)
{
  assert(type >= (ERTCIRQType)0);
  assert(type < NumRTCIRQs);

  if (m_rtcirqhandlers[type])
    m_rtcirqhandlers[type]->RTCIRQHandler(type);
}
