#include "filllevelsensor.h"
#include "clock.h"
#include "log.h"

#include <algorithm>
#include <cstring>

CFillLevelSensor::CFillLevelSensor()
{
  m_distance = -1.0f;
  ResetRead();
}

/* Sets up the usart to read data from the RS485 port.*/
void CFillLevelSensor::Setup(USART_TypeDef* usart)
{
  USART_InitTypeDef usartinit;
  USART_StructInit(&usartinit);
  usartinit.USART_BaudRate = 9600;
  usartinit.USART_Mode = USART_Mode_Rx;
  m_usart.Setup(usart, &usartinit, SENSOR_MSG_SIZE * 16);
}

void CFillLevelSensor::Close()
{
  m_usart.Close();
  m_distance = -1.0f;
  ResetRead();
}

/*! Maximum time waiting for the fill level sensor.*/
#define MAX_READ_TIME (60 * 60 * 1000)

/* Reads a distance value from the sensor.
   \param timeout Timeout in milliseconds waiting for data from the fill level sensor.
   \return true if a correct distance value was read.
   \return false if an error happened.*/
bool CFillLevelSensor::Read(uint32_t timeout /*= 1000*/)
{
  if (timeout > MAX_READ_TIME)
    timeout = MAX_READ_TIME;

  //Keep reading from the USART until a valid message is received.
  int64_t start = Millis();
  while (Millis() - start < timeout)
  {
    uint8_t  input;
    uint32_t bytesread = 0;
    int64_t  readtimeout = std::max((int64_t)timeout - (Millis() - start), (int64_t)1);
    uint32_t result = m_usart.Read(&input, 1, &bytesread, readtimeout);
    
    if (result == READRESULT_SUCCESS)
    {
      if (ProcessMessage(input))
        return true; //A valid distance value was read from the sensor over RS485.
    }
    else
    {
      if (result != READRESULT_ERROR_TIMEOUT)
        PrintLog(LOGFILLSENSOR, "ERROR: unable to read distance from sensor: %s\n",
                 m_usart.ResultToStr(result));

      ResetRead();
    }
  }

  m_distance = -1.0f;
  return false; //No valid distance value was read from the sensor over RS485.
}

void CFillLevelSensor::ResetRead()
{
  m_buffill = 0;
  memset(m_buf, 0, sizeof(m_buf));
}

/* Stores the last received byte in m_buf, then checks if m_buf holds
   a valid message from the ultrasonic sensor.
   \param input The last byte received over RS485.
   \return true if m_buf holds a valid message, m_distance then holds the last
                received distance in meters.
   \return false if m_buf does not hold a valid message.*/
bool CFillLevelSensor::ProcessMessage(uint8_t input)
{
  //Move all bytes in m_buf and write the last byte at the last position.
  memmove(m_buf, m_buf + 1, SENSOR_MSG_SIZE - 1);
  m_buf[SENSOR_MSG_SIZE - 1] = input;
  if (m_buffill < SENSOR_MSG_SIZE)
    m_buffill++;

  //Check if a complete message has been received, if the message starts with 0xFF,
  //and if the first 3 bytes added together is the same as the 4th byte, which means
  //that the checksum is also correct.
  if (m_buffill == SENSOR_MSG_SIZE &&
      m_buf[0] == 0xFF &&
      ((m_buf[0] + m_buf[1] + m_buf[2]) & 0xFF) == m_buf[3])
  { //There is a valid distance value in m_buf.
    //Convert from millimeters to meters.
    m_distance = (float)(((uint32_t)m_buf[1] << 8) | m_buf[2]) / 1000.0f;

    //Reset the current message in m_buf.
    ResetRead();
    
    PrintLog(LOGFILLSENSOR, "Distance read: %f\n", m_distance);

    //Indicate that a valid message has been read, and that its distance
    //is now stored in m_distance.
    return true;
  }
  else
  { //No valid distance value is in m_buf.
    return false;
  }
}
