#include "flashringbuffer.h"
#include "log.h"
#include "cryptokeys.h"
#include "random.h"

#include <cassert>
#include <cstring>
#include <cstddef>

#include <algorithm>

#include <stm32f4xx.h>

const char* ElementStateStr(ElementState state)
{
  if (state == ElementState_Null)
    return "null";
  else if (state == ElementState_Invalid)
    return "invalid";
  else if (state == ElementState_Free)
    return "free";
  else if (state == ElementState_Used)
    return "used";
  else
    return "ERROR UNKNOWN STATE";
}

CFlashRingbuffer::CFlashRingbuffer()
{
  m_enabled     = false;
  m_spiflash    = nullptr;
  m_position    = 0;
  m_elementsize = 0;
  m_numelements = 0;
  m_writepos    = 0;
  m_readpos     = 0;
  m_readoffset  = 0;
}

CFlashRingbuffer::~CFlashRingbuffer()
{
}

/*! Sets up a ringbuffer in the flash.
    \param spiflash A pointer to an instance that manages the flash, should already be set up.
    \param position The position of the ringbuffer inside the flash, should start at a multiple of the sector size.
    \param size The size of the ringbuffer inside the flash, should be a multiple of the sector size and the element size,
                and it should be at least two sectors big.
    \param elementsize The storage size per element of the ringbuffer, it should be a factor of one sector,
                       and it should not be larger than one sector.*/
void CFlashRingbuffer::Setup(CSPIFlash* spiflash, uint32_t position, uint32_t size, uint32_t elementsize)
{
  m_spiflash    = spiflash;
  m_position    = position;
  m_elementsize = elementsize;
  m_numelements = size / m_elementsize;

  assert(m_spiflash);
  if (!m_spiflash->Enabled())
  {
    PrintLog(LOGFLASHBUF, "ERROR: SPI flash not enabled\n");
    return;
  }

  //sanity checks
  assert((size % m_spiflash->SectorSize()) == 0);          //ringbuffer size should be a multiple of the sector size
  assert((size / m_spiflash->SectorSize()) >= 2);          //ringbuffer size should be at least two sectors
  assert((size % m_elementsize) == 0);                     //ringbuffer size should be a multiple of the element size
  assert((m_position % m_spiflash->SectorSize()) == 0);    //ringbuffer should start at a sector address
  assert(m_position < m_spiflash->Size());                 //position should be inside the flash
  assert(m_position + size <= m_spiflash->Size());         //size should fit inside the flash
  assert(m_elementsize <= m_spiflash->SectorSize());       //element size should fit inside one sector
  assert((m_spiflash->SectorSize() % m_elementsize) == 0); //element size should be a factor of sector size
  assert(m_elementsize >= sizeof(SRingbufferElement));     //element size should be able to hold at least one element

  //set enabled to true here, in case the setup calls any
  //method that checks m_enabled
  m_enabled = true;

  if (!CalculateBufferPositions())
  {
    PrintLog(LOGFLASHBUF, "ERROR: unable to calculate buffer positions\n");
    m_enabled = false;
  }
}

/* Calculates the read and write in the ringbuffer.
   \return true if everything succeeded, false if interfacing with the flash failed.*/
bool CFlashRingbuffer::CalculateBufferPositions()
{
  int32_t      readpos   = -1;
  int32_t      writepos  = -1;
  ElementState prevstate = ElementState_Null;

  PrintLog(LOGFLASHBUF, "Checking %" PRIu32 " elements starting at address %" PRIX32 "\n",
           m_numelements, ElementAddr(0));

  //iterate over all elements, find where the ringbuffer starts and ends
  for(uint32_t i = 0; i < m_numelements; i++)
  {
    //read the header of an element from the SPI flash
    SRingbufferElementHeader header;
    if (!ReadElementHeader(i, &header))
      return false;

    //check which state this element is in, used, free or invalid
    ElementState state = header.State(m_elementsize);

    //the read position should point to the first used element
    //the write position should point to the first free element after the last used element
    //the ringbuffer consists of a contiguous block of used/invalid elements, and a contigous block
    //of free elements, which might wrap over the end, that can be detected by checking if the current
    //element is different from the previous element, and if the current element is the same as the first element
    if (state == ElementState_Used)
    {
      if ((readpos == -1) || (readpos == 0 && prevstate != ElementState_Used))
        readpos = i;
    }
    else if (state == ElementState_Free)
    {
      if (writepos == -1 || (writepos == 0 && prevstate != ElementState_Free))
        writepos = i;
    }

    //if all positions are higher than zero, then they're
    //all valid and will not wrap over the end, the loop can be stopped
    if (readpos > 0 && writepos > 0)
      break;

    prevstate = state;
  }
  PrintLog(LOGFLASHBUF, "Done checking %" PRIu32 " elements at address %" PRIX32 "\n",
           m_numelements, ElementAddr(0));

  if (writepos == -1)
  {
    //if no free element is found, there is no way to check where the ringbuffer starts
    //in that case, erase the first sector and start there
    writepos = 0;
    uint32_t sector = m_position / m_spiflash->SectorSize();
    PrintLog(LOGFLASHBUF, "ERROR: no free element detected, erasing sector %" PRIu32 "\n", sector);
    if (!m_spiflash->Erase(sector, 1))
      return false;

    //set the read position to the next element after the write position,
    //and find the first valid element, if that is not found, assume the ringbuffer is empty
    readpos = writepos + 1;
    m_readpos = readpos;
    if (!NextReadElement())
      readpos = writepos;
  }

  //if no element with valid data is found, the ringbuffer is completely empty
  if (readpos == -1)
    readpos = writepos;

  PrintLog(LOGFLASHBUF, "Read position: %" PRIi32 ", write position: %" PRIi32 ", start address: %" PRIX32 "\n",
           readpos, writepos, ElementAddr(0));

  m_readpos  = readpos;
  m_writepos = writepos;

  return true;
}

/*! Returns the number of stored elements in the ringbuffer.*/
uint32_t CFlashRingbuffer::NumStoredElements()
{
  if (!m_enabled)
    return 0;

  if (m_readpos <= m_writepos)
    return (m_writepos - m_readpos);
  else
    return (m_writepos + m_numelements - m_readpos);
}

/* Returns the number of free spaces in the ringbuffer.*/
uint32_t CFlashRingbuffer::FreeSpace()
{
  return m_numelements - NumStoredElements();
}

/*! Calculcates the address of an element inside the flash memory.
    \param elementindex The index of the element, from 0 to m_numelements.
    \return The address of the element inside the flash memory.*/
uint32_t CFlashRingbuffer::ElementAddr(uint32_t elementindex)
{
  assert(elementindex < m_numelements);
  return elementindex * m_elementsize + m_position;
}

/*! Reads the header of an element.
    \param index The index of the element to read.
    \param header A pointer where to store the header of the element.
    \return true if reading the header succeeded, false if it failed.*/
bool CFlashRingbuffer::ReadElementHeader(uint32_t elementindex, SRingbufferElementHeader* header)
{
  if (!m_spiflash->Read(ElementAddr(elementindex), (uint8_t*)header, sizeof(SRingbufferElementHeader)))
  {
    PrintLog(LOGFLASHBUF, "ERROR: unable to read element %" PRIu32 " at address %" PRIu32 "\n",
             elementindex, ElementAddr(elementindex));
    return false;
  }

  return true;
}

/*! Advances m_readpos to the next valid element that can be read.
    \return true if advanding m_readpos succeeded, false if reading the flash memory failed.*/
bool CFlashRingbuffer::NextReadElement()
{
  while (m_readpos != m_writepos)
  {
    //read the header of the current element
    SRingbufferElementHeader header;
    if (!ReadElementHeader(m_readpos, &header))
      return false;

    //if the header of the current element is valid, this loop is done
    if (header.State(m_elementsize) == ElementState_Used)
    {
      PrintLog(LOGFLASHBUF, "First used element is at index %" PRIu32 "\n", m_readpos);
      return true;
    }

    //advance to the next element
    m_readpos = (m_readpos + 1) % m_numelements;
  }

  PrintLog(LOGFLASHBUF, "Did not find first used element\n");

  return false;
}

/*! Writes data to the end of the ringbuffer.
    \param data A pointer to the buffer to write.
    \param size The number of bytes to write, the maximum write size is m_elementsize - sizeof(SRingbufferElement).
    \return true if the write succeeded, false if it failed.*/
bool CFlashRingbuffer::Write(uint8_t* data, uint32_t size)
{
  if (!m_enabled)
  {
    PrintLog(LOGFLASHBUF, "ERROR: can't write, ringbuffer not enabled\n");
    return false;
  }

  //check if what needs to be written fits in one element
  if (!WriteSizeFits(size))
    return false;

  //check if something can be written at the current m_writepos
  if (!CheckWriteable())
    return false;

  //write the data to an element in the flash ringbuffer
  WriteElement(data, size);

  //update the write position
  m_writepos = (m_writepos + 1) % m_numelements;

  return true;
}

/*! Checks if what needs to be written fits into one element, including the required
    header, IV and HMAC, and after encryption has caused the size to grow upwards to a
    multiple of the AES block size.
    \param size The size of the data that needs to be written.
    \return true if the data fits into one element.
    \return false if the data does not fit into one element.*/
bool CFlashRingbuffer::WriteSizeFits(uint32_t size)
{
  uint32_t maxsize = m_elementsize - sizeof(SRingbufferElement);
  uint32_t encryptedsize = size / AES_BLOCK_SIZE * AES_BLOCK_SIZE;
  if ((size % AES_BLOCK_SIZE) != 0)
    encryptedsize++;

  if (encryptedsize > maxsize)
  {
    PrintLog(LOGFLASHBUF, "ERROR: being asked to write %" PRIu32 " bytes, but element can "
             "hold a maximum of %" PRIu32 " bytes\n", encryptedsize, maxsize);
    return false;
  }

  return true;
}

/*! Checks if something can be written at the current m_writepos, and that the position after that is erased.
    If m_writepos or the next position is the start of a sector it will erase the sector, this guarantees that
    there will always be one erased element in the ringbuffer, which is needed to calculate the read and write
    positions in the setup.
    \return true if something can be written at m_writepos.
    \return false if nothing can be written, or something failed.*/
bool CFlashRingbuffer::CheckWriteable()
{
  //check the current and the next write positions
  for (uint32_t i = 0; i < 2; i++)
  {
    uint32_t writepos = (m_writepos + i) % m_numelements;

    //if the write position is at the start of a sector, erase the sector first
    //the spi flash code will check if the sector actually needs erasing, by checking
    //if any bits are set to 0, this prevents unnecessary erases
    if ((ElementAddr(writepos) % m_spiflash->SectorSize()) == 0)
    {
      //only erase if the sector that is to be erased doesn't hold valid data
      if (FreeSpace() < (m_spiflash->SectorSize() / m_elementsize) - i)
      {
        PrintLog(LOGFLASHBUF, "ERROR: unable to write to position %" PRIu32 " address %" PRIX32 " because the %s sector should "
                              "be erased first, but there are still used elements stored inside, free space: %" PRIu32 " bytes\n",
                              m_writepos, ElementAddr(m_writepos), i == 0 ? "current" : "next", FreeSpace() * m_elementsize);
        return false;
      }

      //execute the erase
      uint32_t sector = ElementAddr(writepos) / m_spiflash->SectorSize();
      PrintLog(LOGFLASHBUF, "Erasing sector %" PRIu32 " (%s)\n", sector, i == 0 ? "current" : "next");
      bool alreadyerased;
      if (!m_spiflash->Erase(sector, 1, &alreadyerased))
        return false;

      if (alreadyerased)
        PrintLog(LOGFLASHBUF, "Sector %" PRIu32 " is already erased\n", sector);
      else
        PrintLog(LOGFLASHBUF, "Done erasing sector %" PRIu32 "\n", sector);
    }
  }

  //check if there enough free space to write the element, there should always be one
  //element free to indicate the write position at boot
  uint32_t freespace = FreeSpace();
  if (freespace < 2)
  {
    PrintLog(LOGFLASHBUF, "ERROR: unable to write to position %" PRIu32 " address %" PRIX32 " because the next position is not free\n",
             m_writepos, ElementAddr(m_writepos));
    return false;
  }

  return true;
}

/*! Writes an element to the ringbuffer at m_writepos.
    \param data The buffer to write.
    \param size The size of the buffer to write.
    \return true if the write succeeded, false if anything failed.*/
bool CFlashRingbuffer::WriteElement(uint8_t* data, uint32_t size)
{
  assert(size <= m_elementsize - sizeof(SRingbufferElement));

  //read the header of the write position to make sure it's free
  SRingbufferElementHeader header;
  if (!ReadElementHeader(m_writepos, &header))
    return false;

  if (header.State(m_elementsize) != ElementState_Free)
  {
    //this should not happen, if it does there is a software bug or the flash is corrupt
    PrintLog(LOGFLASHBUF, "ERROR: unable to write element %" PRIu32 " at address %" PRIX32 " because the position is not free\n",
             m_writepos, ElementAddr(m_writepos));
    return false;
  }

  //write the header to the flash
  header.Init(size);
  if (!m_spiflash->Write(ElementAddr(m_writepos), (uint8_t*)&header, sizeof(SRingbufferElementHeader)))
  {
    PrintLog(LOGFLASHBUF, "ERROR: unable to write element %" PRIu32 " at address %" PRIX32 "\n",
             m_writepos, ElementAddr(m_writepos));
    return false;
  }

  //encrypt and sign the data, and write it to the flash
  if (!WriteDataEncrypted(&header, data, size))
    return false;
  
  return true;
}

/*! Writes the initialization vector to the flash.*/
bool CFlashRingbuffer::WriteIV(uint8_t* iv)
{
  uint32_t writeaddr = ElementAddr(m_writepos) + offsetof(SRingbufferElement, iv);
  if (!m_spiflash->Write(writeaddr, iv, AES_BLOCK_SIZE))
  {
    PrintLog(LOGFLASHBUF, "ERROR: unable to write element %" PRIu32 " at address %" PRIX32 "\n",
             m_writepos, writeaddr);
    return false;
  }

  return true;
}

/*! Writes the SHA256 HMAC to the flash.*/
bool CFlashRingbuffer::WriteHMAC(uint8_t* hmac)
{
  uint32_t writeaddr = ElementAddr(m_writepos) + offsetof(SRingbufferElement, hmac);
  if (!m_spiflash->Write(writeaddr, hmac, SHA256_BLOCK_SIZE))
  {
    PrintLog(LOGFLASHBUF, "ERROR: unable to write element %" PRIu32 " at address %" PRIX32 "\n",
             m_writepos, writeaddr);
    return false;
  }

  return true;
}

/*! This function ecncrypts the data per 16 byte block, writes it to the flash, and feeds
    the encrypted data to the HMAC generator, this way an HMAC can be generated over the encrypted
    data without having to buffer the entire encrypted data in ram.
    \param header A pointer to the header of the element, will be used to create the HMAC.
    \param data The buffer holding the unencrypted data.
    \param size The size of the unencrypted data in bytes.
    \return true if writing the data to the flash succeeded, false if it failed.*/
bool CFlashRingbuffer::WriteDataEncrypted(SRingbufferElementHeader* header, uint8_t* data, uint32_t size)
{
  //get 16 random bytes to use as initialization vector, and write it to the flash
  uint8_t iv[AES_BLOCK_SIZE];
  CreateIV(iv);
  if (!WriteIV(iv))
    return false;

  //calculate how many times to run the AES-128 block cipher
  uint32_t rounds = size / AES_BLOCK_SIZE;
  if ((size % AES_BLOCK_SIZE) != 0)
    rounds++;

  uint32_t datapos = 0;
  uint8_t  encryptinbuf[AES_BLOCK_SIZE] = {};
  uint8_t  encryptoutbuf[AES_BLOCK_SIZE];

  //prepare the output buffer with the random data from the initialization vector,
  //for the first round of CBC
  memcpy(encryptoutbuf, iv, AES_BLOCK_SIZE);

  //initialize an HMAC context to generate an HMAC from the header and the encrypted data
  HMAC_SHA256_CTX hmacctx;
  HMAC_SHA256_Init(&hmacctx, FlashRingbufferSignKey, CRYPTO_KEY_SIZE);
  HMAC_SHA256_Update(&hmacctx, (uint8_t*)header, sizeof(SRingbufferElementHeader));

  for (uint32_t i = 0; i < rounds; i++)
  {
    //do a bitwise XOR with the encrypted data from the previous round,
    //and the current unencrypted data
    for (uint32_t j = 0; j < AES_BLOCK_SIZE; j++)
    {
      uint8_t databyte = 0;
      if (datapos < size)
      {
        databyte = data[datapos];
        datapos++;
      }

      encryptinbuf[j] = encryptoutbuf[j] ^ databyte;
    }

    //encrypt what is in the input buffer, and write it to the output buffer
    AES_ECB_encrypt(encryptinbuf, FlashRingbufferEncryptKey, encryptoutbuf, AES_BLOCK_SIZE);

    //write the output buffer to the flash
    uint32_t writeaddr = ElementAddr(m_writepos) + sizeof(SRingbufferElement) + i * AES_BLOCK_SIZE;
    if (!m_spiflash->Write(writeaddr, encryptoutbuf, AES_BLOCK_SIZE))
    {
      PrintLog(LOGFLASHBUF, "ERROR: unable to write element %" PRIu32 " at address %" PRIX32 "\n",
               m_writepos, writeaddr);
      return false;
    }

    //feed the encrypted data to the HMAC generator
    HMAC_SHA256_Update(&hmacctx, encryptoutbuf, AES_BLOCK_SIZE);
  }

  //write the generated HMAC to the flash
  uint8_t hmac[SHA256_BLOCK_SIZE];
  HMAC_SHA256_Final(&hmacctx, hmac);
  if (!WriteHMAC(hmac))
    return false;

  return true;
}

/*! Reads the first used element in the ringbuffer, should be called repeatedly to
    read new elements until it returns false.
    \param data A pointer to a buffer where the data is read into, should be m_elementsize.
    \param size The size of the read data is returned here.
    \param advance When true, advances the read position after a successful read.
                   The read position is always advanced after a failed read.
    \return true if the read succeeded, false if it failed.*/
bool CFlashRingbuffer::Read(uint8_t* data, uint32_t* size, bool advance /*= true*/)
{
  memset(data, 0, m_elementsize);
  *size = 0;

  if (!m_enabled)
  {
    PrintLog(LOGFLASHBUF, "ERROR: can't read, ringbuffer not enabled\n");
    return false;
  }

  for(;;)
  {
    //calculate the read position based on the first used element in the ringbuffer,
    //and the number of elements already read, if it's the same as the write position,
    //all elements in the ringbuffer are read
    uint32_t currentreadpos = (m_readpos + m_readoffset) % m_numelements;
    if (currentreadpos == m_writepos)
      return false;

    //read the current element into ram
    if (!m_spiflash->Read(ElementAddr(currentreadpos), data, m_elementsize))
    {
      PrintLog(LOGFLASHBUF, "ERROR: unable to read element %" PRIu32 " at address %" PRIX32 "\n",
               currentreadpos, ElementAddr(currentreadpos));
      return false;
    }

    //create a copy of the header of the current element, and check it
    SRingbufferElementHeader header(data);

    //check if the header is valid
    if (header.State(m_elementsize) != ElementState_Used)
    {
      if (header.startcode != RINGBUFFER_ELEMENT_NULLCODE)
      {
        PrintLog(LOGFLASHBUF, "ERROR: element %" PRIu32 " at address %" PRIX32 " has invalid header, clearing element\n",
                 currentreadpos, ElementAddr(currentreadpos));
        ClearElement(currentreadpos); //set the element start code to 0
      }
      m_readoffset++; //skip over this element if it's invalid
      continue;
    }

    //decrypt and verify the data of the current element, when successful, the raw data
    //is stored in the buffer, any header is lost
    if (!VerifyDecryptData(&header, data, size))
    {
      PrintLog(LOGFLASHBUF, "ERROR: element %" PRIu32 " at address %" PRIX32 " has invalid data, clearing element\n",
               currentreadpos, ElementAddr(currentreadpos));
      ClearElement(currentreadpos); //set the element start code to 0
      m_readoffset++; //skip over this element if it's invalid
      continue;
    }

    //increase the read position on a successful read only if requested
    if (advance)
      m_readoffset++;

    //stop here if the element that was read is valid
    return true;
  }
}

/*! Advances the read position by one.
    \return true if the read position could be advanced by one.
    \return false if the read position is the same as the write position.*/
bool CFlashRingbuffer::AdvanceReadPosition()
{
  if (!m_enabled)
    return false;

  //calculate the read position based on the first used element in the ringbuffer,
  //and the number of elements already read, if it's the same as the write position,
  //all elements in the ringbuffer are read
  uint32_t currentreadpos = (m_readpos + m_readoffset) % m_numelements;
  if (currentreadpos == m_writepos)
    return false;

  m_readoffset++;
  return true;
}

/*! Calculates and verifies the HMAC of the encrypted data, when correct, it will decrypt the data.
    The decrypted data will be written into the data buffer, starting at position 0.
    The header, hmac and iv are lost after that.
    \param header A pointer to a copy of the header of this element.
    \param data A pointer to a buffer holding the element.
    \param size The size of the decrypted data is returned here.
    \return true if the HMAC is correct, false if it's wrong.*/
bool CFlashRingbuffer::VerifyDecryptData(SRingbufferElementHeader* header, uint8_t* data, uint32_t* size)
{
  SRingbufferElement* element = (SRingbufferElement*)data;

  //calculate how many times to run the AES-128 block cipher
  uint32_t rounds = header->size / AES_BLOCK_SIZE;
  if ((header->size % AES_BLOCK_SIZE) != 0)
    rounds++;

  //calculate the HMAC over the header and the encrypted data
  uint8_t hmac[SHA256_BLOCK_SIZE];
  HMAC_SHA256_CTX hmacctx;
  HMAC_SHA256_Init(&hmacctx, FlashRingbufferSignKey, CRYPTO_KEY_SIZE);
  HMAC_SHA256_Update(&hmacctx, (uint8_t*)header, sizeof(SRingbufferElementHeader));
  HMAC_SHA256_Update(&hmacctx, data + sizeof(SRingbufferElement), rounds * AES_BLOCK_SIZE);
  HMAC_SHA256_Final(&hmacctx, hmac);

  //check if the calculated HMAC is the same as the stored HMAC
  //this loop should run in constant time, so that a timing attack
  //can't be used to guess which HMAC the message should have
  bool valid = true;
  for (uint32_t i = 0; i < SHA256_BLOCK_SIZE; i++)
    valid = valid && (hmac[i] == element->hmac[i]);

  //clear the calculated HMAC as soon as possible for security reasons
  memset(hmac, 0, SHA256_BLOCK_SIZE);

  if (!valid)
    return false; //abort if the HMAC is incorrect

  uint8_t* xorptr = element->iv; //initialize to the initialization vector of the element
  uint8_t  decryptoutbuf[AES_BLOCK_SIZE];

  //decrypt the data per 16 byte block, front to back, the decrypted data is written directly to the buffer
  //any information about the element is lost after this loop
  for (uint32_t i = 0; i < rounds; i++)
  {
    uint8_t* decryptinptr = data + sizeof(SRingbufferElement) + i * AES_BLOCK_SIZE;
    AES_ECB_decrypt(decryptinptr, FlashRingbufferEncryptKey, decryptoutbuf, AES_BLOCK_SIZE);

    for (uint32_t j = 0; j < AES_BLOCK_SIZE; j++)
      data[i * AES_BLOCK_SIZE + j] = decryptoutbuf[j] ^ xorptr[j];

    xorptr = decryptinptr;
  }

  *size = header->size;

  return true;
}

/*! Sets m_readoffset to 0 so that any elements that have been read can be read again.*/
void CFlashRingbuffer::ResetReadCounter()
{
  if (!m_enabled)
    return;

  m_readoffset = 0;
}

/*! Sets the start code of each read element to 0 to indicate that the element no longer holds valid data.*/
void CFlashRingbuffer::ClearReadElements()
{
  if (!m_enabled)
    return;

  for (uint32_t i = 0; i < m_readoffset; i++)
    ClearElement((m_readpos + i) % m_numelements);

  m_readpos = (m_readpos + m_readoffset) % m_numelements;
  m_readoffset = 0;
}

/*! Writes 0 to the start code of an element, to indicate that it no longer holds valid data.*/
bool CFlashRingbuffer::ClearElement(uint32_t elementpos)
{
  uint32_t nullcode = RINGBUFFER_ELEMENT_NULLCODE;
  if (!m_spiflash->Write(ElementAddr(elementpos), (uint8_t*)nullcode, sizeof(nullcode)))
  {
    PrintLog(LOGFLASHBUF, "ERROR: unable to clear element %" PRIu32 " at address %" PRIX32 "\n",
             elementpos, ElementAddr(elementpos));
    return false;
  }

  return true;
}
