#include <stm32f4xx.h>

#include "fwupdate.h"
#include "log.h"
#include "cryptokeys.h"
#include "internalflash.h"
#include "backupregisters.h"
#include "spiflashlog.h"
#include "sha256.h"
#include "aes.h"

#include <cassert>
#include <cstring>

#include <algorithm>

//TinyCBOR should be compiled with CBOR_PARSER_MAX_RECURSIONS set to a reasonable value,
//to protect against stack overruns, since each FreeRTOS stack is not very big
//to do this, just add CBOR_PARSER_MAX_RECURSIONS=16 in the preprocessor symbols
//for all project configurations in System workbench for STM32
#if !defined(CBOR_PARSER_MAX_RECURSIONS) || CBOR_PARSER_MAX_RECURSIONS > 16
  #warning CBOR_PARSER_MAX_RECURSIONS should be defined to 16 or lower for security!
#endif

#if defined(ACCELLO)
static const char* g_mainfilename = "Accello.bin";
#elif defined(ECAT)
static const char* g_mainfilename = "eCat.bin";
#elif defined(TMT)
static const char* g_mainfilename = "TMT.bin";
#else
  #error No filename defined for main firmware
#endif
static const char* g_updaterfilename = "Updater.bin";

CFwUpdate::CFwUpdate()
{
  m_spiflash = nullptr;
  Reset();
}

void CFwUpdate::Init(CSPIFlash* spiflash)
{
  m_spiflash = spiflash;
}

void CFwUpdate::Reset()
{
  m_fwupdatereserved = 0;
  m_bytesdownloaded  = 0;
  m_encryptedblock   = {};
  m_decryptedblock   = {};
  memset(m_signatureblockhash, 0, sizeof(m_signatureblockhash));
  m_signature        = {};
  memset(m_header, 0, sizeof(m_header));
  m_parser           = {};
  m_updaterfile      = {};
  m_mainfile         = {};
}

/*! Should be called before the download is started,
    it will erase the section of the flash that is reserved for firmware updates.
    \return true if the erase succeeded, false if it failed.*/
bool CFwUpdate::Prepare()
{
  assert(m_spiflash);

  Reset();

  //check if the flash is working
  if (!CheckFlashState())
    return false;

  //erase the section of the flash that is reserved for firmware updates
  uint32_t numsectors = m_fwupdatereserved / m_spiflash->SectorSize();
  PrintLog(LOGFWUPDATE, "Erasing %" PRIu32 " sectors of %" PRIu32 " bytes\n", numsectors, m_spiflash->SectorSize());
  if (!m_spiflash->Erase(FWUPDATE_ADDRESS / m_spiflash->SectorSize(), numsectors))
  {
    PrintLog(LOGFWUPDATE, "Erasing flash failed\n");
    return false;
  }
  PrintLog(LOGFWUPDATE, "Erasing flash done\n");

  return true;
}

/*! Returns the maximum block size this class can handle through the Download() method,
    set to the page size of the SPI flash, this has the best throughput, because the
    spi flash is writing while the firmware is getting data from the modem.
    \return The maximum block size this class can handle through the Download method.*/
uint32_t CFwUpdate::DownloadBlockSize()
{
  assert(m_spiflash);
  return m_spiflash->PageSize();
}

void CFwUpdate::StartDownload(const char* url, uint32_t httprspcode, uint32_t contentlength)
{
}

/*! Gets called through the CHTTPDownload class interface, size should be limited to DownloadBlockSize().
    \param buf The buffer to write to the SPI flash.
    \param size The size of the buffer.*/
void CFwUpdate::Download(const uint8_t* buf, uint32_t size)
{
  assert(m_spiflash);
  assert(size <= DownloadBlockSize());

  if (size + m_bytesdownloaded > m_fwupdatereserved)
  {
    PrintLog(LOGFWUPDATE, "ERROR: being asked to store %" PRIu32 " bytes, but reserved size is %" PRIu32 " bytes\n",
             m_bytesdownloaded + size, m_fwupdatereserved);
    return;
  }
  
  //write the buffer to the flash 
  if (!m_spiflash->Write(FWUPDATE_ADDRESS + m_bytesdownloaded, buf, size, false))
  {
    PrintLog(LOGFWUPDATE, "ERROR: unable to write flash\n");
    return;
  }

  m_bytesdownloaded += size;
}

bool CFwUpdate::CheckUpdateImage()
{
  //check if the flash is working
  if (!CheckFlashState())
    return false;

  if (!CheckSignature())
    return false;

  if (!CheckBlocks())
    return false;

  if (!CheckHeader())
    return false;

  return true;
}

/*! Checks if the flash is enabled, and calculates the size reserved for the firmware updates,
    rounded up to a multiple of the sector size.
    \return true if if the flash is enabled, false if it's not.*/
bool CFwUpdate::CheckFlashState()
{
  assert(m_spiflash);

  if (!m_spiflash->Enabled())
  {
    PrintLog(LOGFWUPDATE, "ERROR: SPI FLASH disabled\n");
    return false;
  }

  //make sure the firmware update address is a multiple of the sector size
  if ((FWUPDATE_ADDRESS % m_spiflash->SectorSize()) != 0)
  {
    PrintLog(LOGFWUPDATE, "ERROR: firmware update address %i is not a multiple of SPI flash sector size %" PRIu32 "\n",
             FWUPDATE_ADDRESS, m_spiflash->SectorSize());
    return false;
  }

  //round the reserved size for the firmware update down to a multiple of the sector size
  uint32_t numsectors = FWUPDATE_RESERVED / m_spiflash->SectorSize();
  m_fwupdatereserved = numsectors * m_spiflash->SectorSize();

  return true;
}

/*! Reads the first block, checks the HMAC, decrypts it, and checks the info inside it.
    \return true if the first block is a valid signature, false if it's not.*/
bool CFwUpdate::CheckSignature()
{
  PrintLog(LOGFWUPDATE, "Checking firmware signature\n");

  //read and verify the signature block
  if (!ReadVerifyDecryptBlock(0, &m_encryptedblock, &m_decryptedblock))
    return false;

  //store the signature
  memcpy(&m_signature, m_decryptedblock.block, sizeof(m_signature));

  //check if the signature has the correct version
  if (m_signature.version != FW_SIGNATURE_VERSION)
  {
    PrintLog(LOGFWUPDATE, "ERROR: Signature has incorrect version %" PRIu32 "\n", m_signature.version);
    return false;
  }

  PrintLog(LOGFWUPDATE, "Firmware image version: %" PRIu32 "\n", m_signature.version);
  PrintLog(LOGFWUPDATE, "Firmware image size: %" PRIu32 "\n", m_signature.size);

  //the firmware image size is the size of all the decrypted data together,
  //do some sanity checks on it
  if ((m_signature.size % FW_BLOCK_SIZE) != 0)
  {
    PrintLog(LOGFWUPDATE, "ERROR: Firmware image size should be a multiple of block size %i", FW_BLOCK_SIZE);
    return false;
  }
  else if (m_signature.size <= FW_SIGNATURE_SIZE + FW_HEADER_SIZE)
  {
    PrintLog(LOGFWUPDATE, "ERROR: Firmware image size is too small to hold any files\n");
    return false;
  }

  //check if the size in the signature is small enough to fit in the reserved space after encryption
  //if it's too large, the size in the signature is incorrect
  uint32_t encryptedsize = m_signature.size / FW_BLOCK_SIZE * FW_ENCRYPTED_BLOCK_SIZE;
  if (encryptedsize > m_fwupdatereserved)
  {
    PrintLog(LOGFWUPDATE, "ERROR: Firmware size is larger than reserved size %" PRIu32 "\n", m_fwupdatereserved);
    return false;
  }

  PrintLog(LOGFWUPDATE, "Firmware signature is correct\n");

  return true;
}

/*! Reads each block after the signature block, checks the HMAC,
    decrypts it, checks the block number, and uses the decrypted data of all blocks
    to check the signature HMAC.
    \return true if all checks succeed, false if any check failed.*/
bool CFwUpdate::CheckBlocks()
{
  uint32_t numblocks = m_signature.size / FW_BLOCK_SIZE;

  PrintLog(LOGFWUPDATE, "Checking %" PRIu32 " blocks\n", numblocks - 1);

  //start the SHA256 HMAC for the firmware signature check
  HMAC_SHA256Context hmacctx;
  HMAC_SHA256_Init(&hmacctx, FwSignKey, CRYPTO_KEY_SIZE);

  for (uint32_t i = 1; i < numblocks; i++)
  {
    //read and verify the firmware block
    if (!ReadVerifyDecryptBlock(i, &m_encryptedblock, &m_decryptedblock))
      return false;

    //feed the decrypted data to the HMAC generator
    HMAC_SHA256_Update(&hmacctx, m_decryptedblock.block, FW_BLOCK_SIZE);

    //the first blocks this loop reads is the header,
    //copy them into the header buffer
    uint32_t headerpos = (i - 1) * FW_BLOCK_SIZE;
    if (headerpos < FW_HEADER_SIZE)
      memcpy(m_header + headerpos, m_decryptedblock.block, FW_BLOCK_SIZE);
  }

  //check the generated HMAC against the HMAC in the signature
  uint8_t hash[FW_HASH_SIZE];
  HMAC_SHA256_Final(&hmacctx, hash);
  bool hashcorrect = CompareHMAC(hash, m_signature.signature);
  memset(hash, 0, sizeof(hash)); //erase the generated HMAC for security

  /* BIG FAT WARNING */
  // The calculated hash should never be printed,
  // if an attacker knows it, it does not provide any security at all.

  if (!hashcorrect)
  {
    PrintLog(LOGFWUPDATE, "ERROR: Firmware image has incorrect signature\n");
    return false;
  }

  PrintLog(LOGFWUPDATE, "All %" PRIu32 " blocks are correct\n", numblocks - 1);

  return true;
}

/* Checks the CBOR header which is copied into m_header by CheckBlocks().
   Inside the CBOR header is an array that describes each file in the firmware update image.
   Each file is checked against its SHA256 hash.
   \return true if the header is correct, false if it's not.*/
bool CFwUpdate::CheckHeader()
{
  PrintLog(LOGFWUPDATE, "Checking header\n");

  CborError retval;
  CborValue root;

  //initialize the CBOR parser
  retval = cbor_parser_init(m_header, FW_HEADER_SIZE, CborValidateBasic, &m_parser, &root);
  if (retval != CborNoError)
  {
    PrintLog(LOGFWUPDATE, "ERROR: unable to parse CBOR header: \"%s\"\n", cbor_error_string(retval));
    return false;
  }

  //the first element in the CBOR header should be a map,
  //in this map, all other things are stored
  if (!cbor_value_is_map(&root))
  {
    PrintLog(LOGFWUPDATE, "ERROR: first CBOR element should be a map: ");
    cbor_value_to_pretty(stdout, &root);
    PrintLog(LOGFWUPDATE, "\n");
    return false;
  }

  //open the root map
  CborValue it;
  retval = cbor_value_enter_container(&root, &it);
  if (retval != CborNoError)
  {
    PrintLog(LOGFWUPDATE, "ERROR: unable to enter root map: \"%s\"\n",
             cbor_error_string(retval));
    return false;
  }

  //iterate over all keys in the root map, find the FW_FILES key
  while (!cbor_value_at_end(&it))
  {
    if (!cbor_value_is_integer(&it))
    {
      PrintLog(LOGFWUPDATE, "ERROR: root map key is not an integer: ");
      cbor_value_to_pretty(stdout, &it);
      PrintLog(LOGFWUPDATE, "\n");
      return false;
    }

    //get the key of this map element
    int64_t key;
    retval = cbor_value_get_int64_checked(&it, &key);
    if (retval != CborNoError)
    {
      PrintLog(LOGFWUPDATE, "ERROR: unable to get key from root map: \"%s\"\n",
               cbor_error_string(retval));
      return false;
    }

    //get the value of this map element
    retval = cbor_value_advance(&it);
    if (retval != CborNoError)
    {
      PrintLog(LOGFWUPDATE, "ERROR: unable to advance to value for key %" PRIi64 " in root map: \"%s\"\n",
               key, cbor_error_string(retval));
      return false;
    }

    if (key == FW_FILES)
    { //in the root map, there should be an element with key FW_FILES and it should be an array
      if (!cbor_value_is_array(&it))
      {
        PrintLog(LOGFWUPDATE, "ERROR: key %i in root map should be an array: ", FW_FILES);
        cbor_value_to_pretty(stdout, &it);
        PrintLog(LOGFWUPDATE, "\n");
        return false;
      }

      CborValue files;
      retval = cbor_value_enter_container(&it, &files);
      if (retval != CborNoError)
      {
        PrintLog(LOGFWUPDATE, "ERROR: unable to enter files array: %s\n", cbor_error_string(retval));
        return false;
      }

      //check each file in the files array
      if (!CheckFiles(&files))
        return false;
    }

    //advance to the key of the next map element
    retval = cbor_value_advance(&it);
    if (retval != CborNoError)
    {
      PrintLog(LOGFWUPDATE, "ERROR: unable to advance to next key in root map: \"%s\"\n",
               cbor_error_string(retval));
      return false;
    }
  }

  PrintLog(LOGFWUPDATE, "Header is correct\n");

  return true;
}

/*! This function will try to read a block, verify the HMAC, decrypt the block and check the block number.
    It will to this up to 100 times, or until the block is valid, which is almost always the first time.
    It's possible that the firmware update image has been corrupted somewhere, or that there is data corruption
    between the SPI flash and the microcontroller, this is why the check is done at least 100 times before aborting.

    WARNING: This function should be called on block number 0 first, because the HMAC
             of that block is needed for the HMAC of all other blocks.

    \param blocknum The number of the block to read.
    \param encryptedblock Where to store the read block before decryption
    \param decryptedblock Where to store the read block after decryption
    \return true if the block was read and the HMAC and block number are correct, false if they're not.*/
bool CFwUpdate::ReadVerifyDecryptBlock(uint32_t blocknum, SEncryptedBlock* encryptedblock, SEncryptedBlock* decryptedblock)
{
  const uint32_t tries = 100;
  bool readfailed      = false;
  bool hmacfailed      = false;
  bool blocknumfailed  = false;

  for (uint32_t i = 0; i < tries; i++)
  {
    //read the block from SPI flash, if it fails, try again
    if (!ReadBlock(blocknum, encryptedblock))
    {
      readfailed = true;
      continue;
    }

    //verify the HMAC of the block, if it's incorrect, try again
    if (!VerifyBlockHMAC(blocknum, encryptedblock))
    {
      hmacfailed = true;
      continue;
    }

    //decrypt the block using AES decryption
    DecryptBlock(encryptedblock, decryptedblock);

    //check the block number of the block after decryption
    if (!VerifyBlockNum(decryptedblock, blocknum))
    {
      blocknumfailed = true;
      continue;
    }

    //if this point is reached, all checks have passed successfully
    return true;
  }

  if (readfailed)
    PrintLog(LOGFWUPDATE, "ERROR: failed to read block %" PRIu32 "\n", blocknum);
  if (hmacfailed)
    PrintLog(LOGFWUPDATE, "ERROR: block %" PRIu32 " is incorrect\n", blocknum);
  if (blocknumfailed)
    PrintLog(LOGFWUPDATE, "ERROR: block %" PRIu32 " has wrong block number\n", blocknum);

  return false;
}

/*! Reads a block of encrypted firmware from the SPI flash.
    \param blocnum The block number to read, which starts at 0.
    \param block The read block is returned here.
    \return true if reading the block succeeded, false if it failed.*/
bool CFwUpdate::ReadBlock(uint32_t blocknum, SEncryptedBlock* block)
{
  assert(m_spiflash);

  //check if the address of the block is within the region reserved for firmware updates
  uint32_t blockaddress = blocknum * FW_ENCRYPTED_BLOCK_SIZE;
  assert(blockaddress < m_fwupdatereserved);

  //read the block from the SPI flash
  if (!m_spiflash->Read(FWUPDATE_ADDRESS + blockaddress, (uint8_t*)block, FW_ENCRYPTED_BLOCK_SIZE))
    return false;

  return true;
}

/*! Calculates the HMAC of an encrypted block, and checks it against the HMAC inside the block.
    If blocknum is 0, it will store the HMAC in m_signatureblockhash, if blocknum is anything else,
    m_signatureblockhash will be hashed with the block.
    \param blocknum The number of the block.
    \param block The encrypted block to verify.
    \return true if the HMAC matches, false if it does not.*/
bool CFwUpdate::VerifyBlockHMAC(uint32_t blocknum, SEncryptedBlock* block)
{
  //calculate a SHA256 HMAC over the block
  uint8_t hash[FW_HASH_SIZE];
  HMAC_SHA256_CTX hmacctx;
  HMAC_SHA256_Init(&hmacctx, FwBlockSignKey, CRYPTO_KEY_SIZE);
  HMAC_SHA256_Update(&hmacctx, block->iv,       FW_IV_SIZE);
  HMAC_SHA256_Update(&hmacctx, block->block,    FW_BLOCK_SIZE);
  HMAC_SHA256_Update(&hmacctx, block->blocknum, FW_BLOCKNUM_SIZE);

  if (blocknum == 0)
  { //if this block is the signature block, store the calculated hash
    HMAC_SHA256_Final(&hmacctx, hash);
    memcpy(m_signatureblockhash, hash, FW_HASH_SIZE);
  }
  else
  { //for all other blocks, also hash the signature hash, so that it's certain
    //that all blocks belong together
    HMAC_SHA256_Update(&hmacctx, m_signatureblockhash, FW_HASH_SIZE);
    HMAC_SHA256_Final(&hmacctx, hash);
  }

  /* BIG FAT WARNING */
  // The calculated hash should never be printed,
  // if an attacker knows it, it does not provide any security at all.

  //check if the calculated hash matches the hash stored in the block
  bool hmacmatches = CompareHMAC(hash, block->hash);

  //erase the calculated hash for better security
  memset(hash, 0, sizeof(hash));

  return hmacmatches;
}

/*! Compares two HMACs in constant time, to prevent timing attacks where the HMAC can be guessed.
    \param hmac1 The first HMAC to compare.
    \param hmac2 The second HMAC to compare.
    \return true if both HMACs are equal.
    \return false if the HMACs differ.*/
bool CFwUpdate::CompareHMAC(uint8_t hmac1[FW_HASH_SIZE], uint8_t hmac2[FW_HASH_SIZE])
{
  bool valid = true;
  for (uint32_t i = 0; i < FW_HASH_SIZE; i++)
    valid = valid && (hmac1[i] == hmac2[i]);

  return valid;
}

/*! Decrypts a block using the firmware encryption key.
    \param encryptedblock The encrypted block to decrypt.
    \param decryptedblock Where the decrypted block is written to.*/
void CFwUpdate::DecryptBlock(SEncryptedBlock* encryptedblock, SEncryptedBlock* decryptedblock)
{
  AES_CBC_decrypt_buffer(decryptedblock->block, encryptedblock->block, FW_BLOCK_SIZE + FW_BLOCKNUM_SIZE,
                         FwEncryptKey, encryptedblock->iv);
}

/*! Verifies the block number of a decrypted block.
    \param decryptedblock The decrypted block to check the number of.
    \param num The number to check.
    \return true if the number matches, false if it does not.*/
bool CFwUpdate::VerifyBlockNum(SEncryptedBlock* decryptedblock, uint32_t num)
{
  uint32_t blocknum;
  memcpy(&blocknum, decryptedblock->blocknum, sizeof(blocknum));

  return blocknum == num;
}

/*! Iterates over the files array, checks if every element is a map,
    calls CheckFiles() for each element.
    \param files The opened files array.
    \return true if all elements are correct, false on any error.*/
bool CFwUpdate::CheckFiles(CborValue* files)
{
  CborError retval;

  uint32_t fileindex = 0;
  while (!cbor_value_at_end(files))
  {
    PrintLog(LOGFWUPDATE, "Checking file at index %" PRIu32 "\n", fileindex);

    //check if the array element is a map
    if (!cbor_value_is_map(files))
    {
      PrintLog(LOGFWUPDATE, "ERROR: at file index %" PRIu32 ", files array should be an array of maps: ", fileindex);
      cbor_value_to_pretty(stdout, files);
      PrintLog(LOGFWUPDATE, "\n");
      return false;
    }

    //open the array element
    CborValue file;
    retval = cbor_value_enter_container(files, &file);
    if (retval != CborNoError)
    {
      PrintLog(LOGFWUPDATE, "ERROR: unable to open element %" PRIu32 " of files array: %s\n",
               fileindex, cbor_error_string(retval));
      return false;
    }

    //check the file map
    if (!CheckFile(&file))
    {
      PrintLog(LOGFWUPDATE, "ERROR: file index %" PRIu32 " is not correct\n", fileindex);
      return false;
    }

    //leave the file map, this automatically advances to the next array element
    retval = cbor_value_leave_container(files, &file);
    if (retval != CborNoError)
    {
      PrintLog(LOGFWUPDATE, "ERROR: unable to close element %" PRIu32 " of files array: %s\n",
               fileindex, cbor_error_string(retval));
      return false;
    }

    PrintLog(LOGFWUPDATE, "File at index %" PRIu32 " is correct\n", fileindex);

    fileindex++;
  }

  //check if all needed files are present
  bool success = true;
  if (!m_updaterfile.present)
  {
    PrintLog(LOGFWUPDATE, "ERROR: Updater firmware file \"%s\" not present\n", g_updaterfilename);
    success = false;
  }
  if (!m_mainfile.present)
  {
    PrintLog(LOGFWUPDATE, "ERROR: Main firmware file \"%s\" not present\n", g_mainfilename);
    success = false;
  }

  return success;
}

/* This function gets passed a map which describes a file,
   it has at least present:
   filename
   file position
   file size
   file hash
   These all get checked, and if the file is needed for the firmware update,
   its parameters get stored.
   \param file The map that describes a file.
   \return true if the file map is correct, false if it's not.*/
bool CFwUpdate::CheckFile(CborValue* file)
{
  CborError retval;

  EFWFileType filetype = EFWFileType_None;
  SFileInfo   fileinfo;
  bool        present[NUM_FILE_MAP_ELEMENTS] = {};

  while (!cbor_value_at_end(file))
  {
    //file should point to the first key of the map element, which should be an integer
    if (!cbor_value_is_integer(file))
    {
      PrintLog(LOGFWUPDATE, "ERROR: file map key is not an integer: ");
      cbor_value_to_pretty(stdout, file);
      PrintLog(LOGFWUPDATE, "\n");
      return false;
    }

    //get the key for this map element
    int64_t key;
    retval = cbor_value_get_int64_checked(file, &key);
    if (retval != CborNoError)
    {
      PrintLog(LOGFWUPDATE, "ERROR: unable to get key from file map: \"%s\"\n",
               cbor_error_string(retval));
      return false;
    }

    //get the value of this map element
    retval = cbor_value_advance(file);
    if (retval != CborNoError)
    {
      PrintLog(LOGFWUPDATE, "ERROR: unable to advance to value for key %" PRIi64 " in file map: \"%s\"\n",
               key, cbor_error_string(retval));
      return false;
    }

    //check the value based on the key
    if (key == FW_FILENAME)
    {
      if (!CheckFileMapName(file, &filetype, fileinfo.name))
        return false;
      present[0] = true;
    }
    else if (key == FW_FILEPOSITION)
    {
      if (!CheckFileMapPosition(file, &fileinfo.pos))
        return false;
      present[1] = true;
    }
    else if (key == FW_FILESIZE)
    {
      if (!CheckFileMapSize(file, &fileinfo.size))
        return false;
      present[2] = true;
    }
    else if (key == FW_FILEHASH)
    {
      if (!CheckFileMapHash(file, fileinfo.hash))
        return false;
      present[3] = true;
    }

    //move to the next element in the map
    retval = cbor_value_advance(file);
    if (retval != CborNoError)
    {
      PrintLog(LOGFWUPDATE, "ERROR: unable to advance to next element in file map: %s\n",
               cbor_error_string(retval));
      return false;
    }
  }

  PrintLog(LOGFWUPDATE, "Filename: \"%.*s\", position: %" PRIu32 ", size: %" PRIu32 ", hash: ",
           (int)FW_FILENAME_MAX_LENGTH, fileinfo.name, fileinfo.pos, fileinfo.size);

  for (uint32_t i = 0; i < FW_HASH_SIZE; i++)
    PrintLog(LOGFWUPDATE, "%02" PRIX8, fileinfo.hash[i]);

  PrintLog(LOGFWUPDATE, "\n");

  //return an error if the file map is missing anything
  if (!CheckFileElementsPresent(present))
    return false;

  //return an error if there is something wrong with the file position or size
  if (!CheckFilePositionSize(&fileinfo, filetype))
    return false;

  //check if the hash in the header matches the file
  if (!VerifyFileHash(&fileinfo))
    return false;

  //store the file information for later use
  fileinfo.present = true;
  if (filetype == EFWFileType_Main_Firmware)
    m_mainfile = fileinfo;
  else if (filetype == EFWFileType_Updater_Firmware)
    m_updaterfile = fileinfo;

  return true;
}

/*! Checks if value is a string, compares it with known filenames, and stores the result in filetype.
    \param value The CBOR string to check.
    \param filetype The filetype is returned here.
    \param filename The filename is returned here.
    \return true if reading and comparing the value succeeded, false if it failed.*/
bool CFwUpdate::CheckFileMapName(CborValue* value, EFWFileType* filetype, char filename[FW_FILENAME_MAX_LENGTH])
{
  CborError retval;

  //check if the filename is a text string
  if (!cbor_value_is_text_string(value))
  {
    PrintLog(LOGFWUPDATE, "ERROR: filename in file map should be a text string: ");
    cbor_value_to_pretty(stdout, value);
    PrintLog(LOGFWUPDATE, "\n");
    return false;
  }

  //check if the length of the filename is correct
  size_t len;
  retval = cbor_value_calculate_string_length(value, &len);
  if (retval != CborNoError)
  {
    PrintLog(LOGFWUPDATE, "ERROR: unable to calculate string length of filename: %s\n", 
             cbor_error_string(retval));
    return false;
  }

  if (len > FW_FILENAME_MAX_LENGTH)
  {
    PrintLog(LOGFWUPDATE, "ERROR: filename length is too long: ");
    cbor_value_to_pretty(stdout, value);
    PrintLog(LOGFWUPDATE, "\n");
    return false;
  }

  //copy the filename into the buffer
  len = FW_FILENAME_MAX_LENGTH;
  retval = cbor_value_copy_text_string(value, filename, &len, nullptr);
  if (retval != CborNoError)
  {
    PrintLog(LOGFWUPDATE, "ERROR: unable to copy filename: %s\n", 
             cbor_error_string(retval));
    return false;
  }

  const char* filenames[] = {g_updaterfilename, g_mainfilename};

  //check if the filename is the name of the main firmware, or the name of the updater firmware
  //and store the result in *filetype
  for (uint32_t i = 0; i < sizeof(filenames) / sizeof(filenames[0]); i++)
  {
    bool result;
    retval = cbor_value_text_string_equals(value, filenames[i], &result);
    if (retval != CborNoError)
    {
      PrintLog(LOGFWUPDATE, "ERROR: unable to compare filename text string: %s\n", 
               cbor_error_string(retval));
      return false;
    }

    if (result)
    {
      if (strcmp(g_mainfilename, filenames[i]) == 0)
        *filetype = EFWFileType_Main_Firmware;
      else if (strcmp(g_updaterfilename, filenames[i]) == 0)
        *filetype = EFWFileType_Updater_Firmware;
      else
        *filetype = EFWFileType_None;
    }
  }

  return true;
}

/*! Checks if value is an integer, and stores it in filepos.
    \param value The CBOR integer to read.
    \param filepos The CBOR integer is returned here.
    \return true if value is an integer and it was read successfully, false if it failed.*/
bool CFwUpdate::CheckFileMapPosition(CborValue* value, uint32_t* filepos)
{
  CborError retval;
  *filepos = 0;

  if (!cbor_value_is_integer(value))
  {
    PrintLog(LOGFWUPDATE, "ERROR: file position is not an integer: ");
    cbor_value_to_pretty(stdout, value);
    PrintLog(LOGFWUPDATE, "\n");
  }

  int64_t pos;
  retval = cbor_value_get_int64_checked(value, &pos);
  if (retval != CborNoError)
  {
    PrintLog(LOGFWUPDATE, "ERROR: unable to read file position: %s\n", 
             cbor_error_string(retval));
    return false;
  }
  else if (pos < 0)
  {
    PrintLog(LOGFWUPDATE, "ERROR: invalid value %" PRIi64 " for file position\n", pos);
    return false;
  }

  *filepos = pos;

  return true;
}

/*! Checks if value is an integer, and stores it in filesize.
    \param value The CBOR integer to read.
    \param filepos The CBOR integer is returned here.
    \return true if value is an integer and it was read successfully, false if it failed.*/
bool CFwUpdate::CheckFileMapSize(CborValue* value, uint32_t* filesize)
{
  CborError retval;
  *filesize = 0;

  if (!cbor_value_is_integer(value))
  {
    PrintLog(LOGFWUPDATE, "ERROR: file size is not an integer: ");
    cbor_value_to_pretty(stdout, value);
    PrintLog(LOGFWUPDATE, "\n");
  }

  int64_t size;
  retval = cbor_value_get_int64_checked(value, &size);
  if (retval != CborNoError)
  {
    PrintLog(LOGFWUPDATE, "ERROR: unable to read file size: %s\n", 
             cbor_error_string(retval));
    return false;
  }
  else if (size < 0)
  {
    PrintLog(LOGFWUPDATE, "ERROR: invalid value %" PRIi64 " for file size\n", size);
    return false;
  }

  *filesize = size;

  return true;
}

/*! Checks if value is a byte string of size FW_HASH_SIZE.
    \param value The CBOR byte string to check.
    \param filehash Where to store the read byte string.
    \return true if the byte string can be read and has the correct length, false if not.*/
bool CFwUpdate::CheckFileMapHash(CborValue* value, uint8_t filehash[FW_HASH_SIZE])
{
  CborError retval;

  //check if the hash is a byte string
  if (!cbor_value_is_byte_string(value))
  {
    PrintLog(LOGFWUPDATE, "ERROR: file hash is not a byte string: ");
    cbor_value_to_pretty(stdout, value);
    PrintLog(LOGFWUPDATE, "\n");
  }

  //check if the byte string has the correct length
  size_t len;
  retval = cbor_value_calculate_string_length(value, &len);
  if (retval != CborNoError)
  {
    PrintLog(LOGFWUPDATE, "ERROR: unable to calculate file hash length: %s\n", 
             cbor_error_string(retval));
    return false;
  }

  if (len != FW_HASH_SIZE)
  {
    PrintLog(LOGFWUPDATE, "ERROR: file hash has wrong length %" PRIu32 ", expected %" PRIu32 "\n",
             (int32_t)len, (int32_t)FW_HASH_SIZE);
    return false;
  }

  //copy the byte string into filehash
  retval = cbor_value_copy_byte_string(value, filehash, &len, nullptr);
  if (retval != CborNoError)
  {
    PrintLog(LOGFWUPDATE, "ERROR: unable to read file hash: %s\n", 
             cbor_error_string(retval));
    return false;
  }

  return true;
}

bool CFwUpdate::CheckFileElementsPresent(bool present[NUM_FILE_MAP_ELEMENTS])
{
  //print which elements are not present
  if (present[0] == false)
    PrintLog(LOGFWUPDATE, "ERROR: filename not present\n");
  if (present[1] == false)
    PrintLog(LOGFWUPDATE, "ERROR: file position not present\n");
  if (present[2] == false)
    PrintLog(LOGFWUPDATE, "ERROR: file size not present\n");
  if (present[3] == false)
    PrintLog(LOGFWUPDATE, "ERROR: file hash not present\n");

  //return an error if any element is not present
  for (uint32_t i = 0; i < NUM_FILE_MAP_ELEMENTS; i++)
  {
    if (present[i] == false)
      return false;
  }

  return true;
}

/*! Checks if the file is within the firmware update image, and if the file starts at a block.
    \param filepos The position of the file within the firmware update image.
    \param filesize The size of the file within the firmware update image.
    \return true if the file size and position are correct, false if there is a problem.*/
bool CFwUpdate::CheckFilePositionSize(SFileInfo* fileinfo, EFWFileType filetype)
{
  //check the file position and file size
  if (fileinfo->pos >= m_signature.size)
  {
    PrintLog(LOGFWUPDATE, "ERROR: File position %" PRIu32 " is not correct, file would be outside the update image range 0-%" PRIu32 "\n",
             fileinfo->pos, m_signature.size);
    return false;
  }

  if (fileinfo->pos + fileinfo->size > m_signature.size)
  {
    PrintLog(LOGFWUPDATE, "ERROR: File size %" PRIu32 " is not correct, file would be outside the update image range 0-%" PRIu32 "\n",
             fileinfo->size, m_signature.size);
    return false;
  }

  if ((fileinfo->pos % FW_BLOCK_SIZE) != 0)
  {
    PrintLog(LOGFWUPDATE, "ERROR: File position %" PRIu32 " is not a multiple of block size %i\n",
             fileinfo->pos, FW_BLOCK_SIZE);
    return false;
  }

  //if the file is meant to go into the internal flash, check if it fits into the reserved space
  if (filetype == EFWFileType_Main_Firmware || filetype == EFWFileType_Updater_Firmware)
  {
    uint32_t maxsize;
    const char* fwname;
    if (filetype == EFWFileType_Main_Firmware)
    {
      fwname = "Main";
      maxsize = FlashSectorsSize(MAIN_FIRMWARE_START_SECTOR, MAIN_FIRMWARE_NUM_SECTORS);
    }
    else
    {
      fwname = "Updater";
      maxsize = FlashSectorsSize(UPDATER_FIRMWARE_START_SECTOR, UPDATER_FIRMWARE_NUM_SECTORS);
    }

    if (fileinfo->size > maxsize)
    {
      PrintLog(LOGFWUPDATE, "ERROR: %s firmware file size %" PRIu32 " is too large to fit "
                            "into internal flash, max reserved size is %" PRIu32 "\n",
               fwname, fileinfo->size, maxsize);
      return false;
    }
  }

  return true;
}

/*! Reads the file at the specified position and number of bytes, calculates a SHA256 hash,
    and verifies it against the hash from the CBOR header.
    \param fileinfo The information about the file to check.
    \return true if all data was read correctly and the hash is correct, false on any error.*/
bool CFwUpdate::VerifyFileHash(SFileInfo* fileinfo)
{
  //this should have been checked previously
  assert((fileinfo->pos % FW_BLOCK_SIZE) == 0);

  //calculate which blocks to read for this file
  uint32_t startblock = fileinfo->pos / FW_BLOCK_SIZE;
  uint32_t numblocks  = fileinfo->size / FW_BLOCK_SIZE;
  if ((fileinfo->size % FW_BLOCK_SIZE) != 0)
    numblocks++;

  //read blocks from the SPI flash, verify and decrypt them,
  //and feed their data to the SHA256 generator
  SHA256_CTX sha256ctx;
  WSS_sha256_init(&sha256ctx);
  uint32_t bytesread = 0;
  for (uint32_t i = startblock; i < startblock + numblocks; i++)
  {
    if (!ReadVerifyDecryptBlock(i, &m_encryptedblock, &m_decryptedblock))
      return false;

    uint32_t hashbytes = std::min(fileinfo->size - bytesread, (uint32_t)FW_BLOCK_SIZE);
    WSS_sha256_update(&sha256ctx, m_decryptedblock.block, hashbytes);

    bytesread += FW_BLOCK_SIZE;
  }

  uint8_t hash[FW_HASH_SIZE];
  WSS_sha256_final(&sha256ctx, hash);

  //check if the generated hash is the same as the hash from the CBOR header
  if (!CompareHMAC(fileinfo->hash, hash))
  {
    PrintLog(LOGFWUPDATE, "ERROR: hash of file at position %" PRIu32 " and size %" PRIu32 " is not correct\n",
             fileinfo->pos, fileinfo->size);
    return false;
  }

  return true;
}

bool CFwUpdate::InstallUpdateImage()
{
  //if UPDATER is not defined, the current firmware running is the main firmware,
  //that means that the updater firmware should be written to the internal flash
  //if UPDATER is defined, the updater firmware is running and the main firmware
  //should be written to the internal flash
#ifndef UPDATER
  SFileInfo*  fileinfo    = &m_updaterfile;
  const char* fwname      = "updater";
  uint32_t    startsector = UPDATER_FIRMWARE_START_SECTOR;
  uint32_t    numsectors  = UPDATER_FIRMWARE_NUM_SECTORS;
#else //UPDATER
  SFileInfo*  fileinfo    = &m_mainfile;
  const char* fwname      = "main";
  uint32_t    startsector = MAIN_FIRMWARE_START_SECTOR;
  uint32_t    numsectors  = MAIN_FIRMWARE_NUM_SECTORS;
#endif

  PrintLog(LOGFWUPDATE, "Writing %s firmware to internal flash\n", fwname);

  //wait for the debug usart to transmit all messages,
  //because no code will be executed when writing to the internal flash
  WaitLogWriteComplete(2000);

#ifdef UPDATER
  //before erasing the main firmware, set the flag that signals the boot firmware that the
  //main firmware is erased, this is in case the writing of the new main firmware gets interrupted,
  //the boot firmware will then start the updater firmware again
  if (!SetMainFirmwareErasedFlag(true))
    return false;
#endif //UPDATER

  if (!FlashEraseSectors(startsector, numsectors))
    return false;

  if (!WriteFileToInternalFlash(fileinfo, startsector, numsectors))
    return false;

#ifdef UPDATER
  //set the signal for the boot firmware that the main firmware is valid again
  SetMainFirmwareErasedFlag(false);
#endif //UPDATER

  PrintLog(LOGFWUPDATE, "Starting %s firmware\n", fwname);

#ifndef UPDATER
  //if this code runs from the main firmware, set the boot updater flag in the boot flag register of the RTC
  //this is a battery backed register, which keeps its value during a reset
  //this tells the boot firmware to start at the updater address
  SetBootUpdaterFlag(true);
#else
  //If this code runs from the firmware updater, set a flag in the battery backed
  //ram to indicate to the main firmware that new firmware has been written.
  //If the main firmware uses this flag, it should clear it when it's done using it.
  SetFirmwareUpdatedFlag(true);
#endif

  //wait for the debug usart to transmit all log messages before rebooting
  WaitLogWriteComplete(2000);

  //reboot to start the just written firmware
  NVIC_SystemReset();

  return false; //this point should never be reached
}

/*! Writes the file in the SPI flash that fileinfo points to, into the internal flash.
    \param fileinfo Information about where the file is stored in the SPI flash.
    \param startsector The sector where to start writing the file.
    \param numsector The number of sectors reserved for this file, this function asserts on an overrun,
                     because the size of the file should have been checked before.*/
bool CFwUpdate::WriteFileToInternalFlash(SFileInfo* fileinfo, uint32_t startsector, uint32_t numsectors)
{
  uint32_t maxsize = FlashSectorsSize(startsector, numsectors);
  assert(fileinfo->size <= maxsize);

  //calculate which blocks to read for this file
  uint32_t startblock = fileinfo->pos / FW_BLOCK_SIZE;
  uint32_t numblocks  = fileinfo->size / FW_BLOCK_SIZE;
  if ((fileinfo->size % FW_BLOCK_SIZE) != 0)
    numblocks++;

  //read the file from the SPI flash, and write it to the internal flash, one block at a time
  uint32_t byteswritten = 0;
  uint32_t writeaddress = FlashSectorAddress(startsector);
  for (uint32_t i = startblock; i < startblock + numblocks; i++)
  {
    if (!ReadVerifyDecryptBlock(i, &m_encryptedblock, &m_decryptedblock))
      return false;

    uint32_t writebytes = std::min(fileinfo->size - byteswritten, (uint32_t)FW_BLOCK_SIZE);
    if (!FlashWrite(writeaddress, m_decryptedblock.block, writebytes))
      return false;

    byteswritten += writebytes;
    writeaddress += writebytes;
  }

  //create a hash of the flash where the file has just been written to,
  //to verify that the write is correct
  uint8_t hash[FW_HASH_SIZE];
  SHA256_CTX sha256ctx;
  WSS_sha256_init(&sha256ctx);
  WSS_sha256_update(&sha256ctx, (uint8_t*)FlashSectorAddress(startsector), fileinfo->size);
  WSS_sha256_final(&sha256ctx, hash);

  if (!CompareHMAC(fileinfo->hash, hash))
  {
    PrintLog(LOGFWUPDATE, "ERROR: SHA256 hash verification of sector %" PRIu32 " size %" PRIu32 " failed\n",
             startsector, fileinfo->size);
    return false;
  }

  return true;
}

/*! Sets the main firmware erased flag.
    \param erased Set to true to set the flag, set to false to clear the flag.
    \return true if Setting the state of the flag succeeded, false if it failed.*/
bool CFwUpdate::SetMainFirmwareErasedFlag(bool erased)
{
  if (erased)
  { //signal to the boot firmware that the main firmware has been erased, by writing to 0 the
    //sector that has been reserved for the flag
    uint8_t erasedflag = 0;
    if (!FlashWrite(FlashSectorAddress(MAIN_FIRMWARE_ERASED_SECTOR), &erasedflag, sizeof(erasedflag)))
    {
      PrintLog(LOGFWUPDATE, "ERROR: Unable to set flag that the main firmware is erased\n");
      return false;
    }
  }
  else
  { //signal to the boot firmware that the main firmware is valid,
    //by erasing the entire sector that is reserved for this flag
    if (!FlashEraseSector(MAIN_FIRMWARE_ERASED_SECTOR))
    {
      PrintLog(LOGFWUPDATE, "ERROR: Unable to clear flag that the main firmware is erased\n");
      return false;
    }
  }

  return true;
}
