#include "gpio.h"
#include "gpiopin.h"
#include "log.h"
#include "inclstdint.h"
#include <stm32f4xx.h>
#include <cassert>
#include <stdio.h>

/*! The PortInit array should be added as an extern const in the main project,
    it describes how every gpio should be initialized.*/
extern const SGPIOStateInit PortInit[NUM_PORTS][PINS_PER_PORT];

/*! Converts a port number (0 to 12) to a GPIO_Typedef*.
\param portnum Number of the port, GPIOA = 0, GPIOK = 12.
\return Pointer to the GPIO_TypeDef* of the port.*/
GPIO_TypeDef* GPIOPortByNum(uint32_t portnum)
{
  static GPIO_TypeDef* const ports[] =
  {
    GPIOA,
    GPIOB,
    GPIOC,
    GPIOD,
    GPIOE,
    GPIOF,
    GPIOG,
    GPIOH,
    GPIOI,
    GPIOJ,
    GPIOK
  };

  assert(portnum < sizeof(ports) / sizeof(ports[0]));

  return ports[portnum];
}

/*! Converts a pin number (0 to 15) to a GPIO_Pin define from stm32f4xx_gpio.h.
\param portnum Number of the pin, 0 to 15.
\return GPIO_Pin of the corresponding pin.*/
uint16_t GPIOPinByNum(uint32_t pinnum)
{
  static const uint32_t pins[] =
  {
    GPIO_Pin_0,
    GPIO_Pin_1,
    GPIO_Pin_2,
    GPIO_Pin_3,
    GPIO_Pin_4,
    GPIO_Pin_5,
    GPIO_Pin_6,
    GPIO_Pin_7,
    GPIO_Pin_8,
    GPIO_Pin_9,
    GPIO_Pin_10,
    GPIO_Pin_11,
    GPIO_Pin_12,
    GPIO_Pin_13,
    GPIO_Pin_14,
    GPIO_Pin_15
  };

  assert(pinnum < sizeof(pins) / sizeof(pins[0]));

  return pins[pinnum];
}

/*! Converts a pin define (GPIO_Pin_0 to GPIO_Pin_15) to a number from 0 to 15.*/
uint32_t GPIONumByPin(uint16_t pin)
{
  static const uint32_t pins[] =
  {
    GPIO_Pin_0,
    GPIO_Pin_1,
    GPIO_Pin_2,
    GPIO_Pin_3,
    GPIO_Pin_4,
    GPIO_Pin_5,
    GPIO_Pin_6,
    GPIO_Pin_7,
    GPIO_Pin_8,
    GPIO_Pin_9,
    GPIO_Pin_10,
    GPIO_Pin_11,
    GPIO_Pin_12,
    GPIO_Pin_13,
    GPIO_Pin_14,
    GPIO_Pin_15
  };

  for (uint32_t i = 0; i < sizeof(pins) / sizeof(pins[0]); i++)
  {
    if (pin == pins[i])
      return i;
  }

  assert(0);
}

/*! Converts a GPIO_TypeDef (GPIOA to GPIOK) to a number from 0 to 10.*/
uint32_t GPIONumByPort(GPIO_TypeDef* port)
{
  static const GPIO_TypeDef* ports[] =
  {
    GPIOA,
    GPIOB,
    GPIOC,
    GPIOD,
    GPIOE,
    GPIOF,
    GPIOG,
    GPIOH,
    GPIOI,
    GPIOJ,
    GPIOK
  };

  for (uint32_t i = 0; i < sizeof(ports) / sizeof(ports[0]); i++)
  {
    if (port == ports[i])
      return i;
  }

  assert(0);
}

/*! Converts a port number (0 to 12) to a RCC_AHB1Periph define from stm32f4xx_rcc.h.
\param portnum Number of the pin, 0 to 15.
\return RCC_AHB1Periph of the corresponding pin.*/
static uint32_t GPIOGetPeriph(uint32_t portnum)
{
  static const uint32_t ports[] =
  {
    RCC_AHB1Periph_GPIOA,
    RCC_AHB1Periph_GPIOB,
    RCC_AHB1Periph_GPIOC,
    RCC_AHB1Periph_GPIOD,
    RCC_AHB1Periph_GPIOE,
    RCC_AHB1Periph_GPIOF,
    RCC_AHB1Periph_GPIOG,
    RCC_AHB1Periph_GPIOH,
    RCC_AHB1Periph_GPIOI,
    RCC_AHB1Periph_GPIOJ,
    RCC_AHB1Periph_GPIOK
  };

  assert(portnum < sizeof(ports) / sizeof(ports[0]));

  return ports[portnum];
}

/*! Converts a GPIO_TypeDef* to an EXTI_PortSourceGPIO to use with the SYSCFG_EXTILineConfig() function.
    \param port A pointer to the port to convert.
    \return a value of @ref SYSCFG_EXTI_Port_Sources.*/
uint32_t EXTIPortSourceByPort(GPIO_TypeDef* port)
{
  static const struct
  {
    GPIO_TypeDef* port;
    uint32_t      portsource;
  }
  ports[] =
  {
    { GPIOA, EXTI_PortSourceGPIOA },
    { GPIOB, EXTI_PortSourceGPIOB },
    { GPIOC, EXTI_PortSourceGPIOC },
    { GPIOD, EXTI_PortSourceGPIOD },
    { GPIOE, EXTI_PortSourceGPIOE },
    { GPIOF, EXTI_PortSourceGPIOF },
    { GPIOG, EXTI_PortSourceGPIOG },
    { GPIOH, EXTI_PortSourceGPIOH },
    { GPIOI, EXTI_PortSourceGPIOI },
    { GPIOJ, EXTI_PortSourceGPIOJ },
    { GPIOK, EXTI_PortSourceGPIOK }
  };

  for (uint32_t i = 0; i < sizeof(ports) / sizeof(ports[0]); i++)
  {
    if (ports[i].port == port)
      return ports[i].portsource;
  }

  assert(0);

  return 0;
}

static uint8_t wakeupstate = Bit_RESET;

/*! Initializes all GPIOs according to the PortInit array.*/
void GPIOInit()
{
  for (uint32_t port = 0; port < NUM_PORTS; port++)
  {
    //enable the clock for this port
    RCC_AHB1PeriphClockCmd(GPIOGetPeriph(port), ENABLE);

    //Initialize all pins for this port.
    for (uint32_t pin = 0; pin < PINS_PER_PORT; pin++)
      GPIOInitPinNum(port, pin);
  }

  //store the boot state of the wakeup pin
  wakeupstate = GPIO_ReadInput(WAKEUP);
  printf("GPIO INIT IS DONE\n");

}

/*! Initializes a pin to its default state, as described in the PortInit array.
    \param port A value of GPIOA to GPIOG.
    \param pin A value of GPIO_Pin_0 to GPIO_Pin_15.
    \param stateinit When not nullptr, initialize the pin this way,
                     otherwise use the PortInit array.*/
void GPIOInitPin(GPIO_TypeDef* port, uint32_t pin,
                 const SGPIOStateInit* stateinit /*= nullptr*/)
{
  GPIOInitPinNum(GPIONumByPort(port), GPIONumByPin(pin), stateinit);
}

/*! Initializes a pin to its default state, as described in the PortInit array.
    \param port A value of 0 to 6 representing GPIOA to GPIOG.
    \param pin A value of 0 to 15 representing GPIO_Pin_0 to GPIO_Pin_15.
    \param stateinit When not nullptr, initialize the pin this way,
                     otherwise use the PortInit array.*/
void GPIOInitPinNum(uint32_t port, uint32_t pin,
                    const SGPIOStateInit* stateinit /*= nullptr*/)
{
  if (stateinit == nullptr)
    stateinit = PortInit[port] + pin;

  //The GPIO_Pin field is a bit mask, if it's zero then skip over it,
  //because no GPIO is filled in.
  if (!IS_GPIO_PIN(stateinit->init.GPIO_Pin))
    return;

  GPIO_TypeDef* portptr = GPIOPortByNum(port);

  //Set the output state, low or high.
  GPIO_WriteBit(portptr, stateinit->init.GPIO_Pin, stateinit->state);

  //Set the alternate function.
  if (stateinit->af != -1)
    GPIO_PinAFConfig(portptr, pin, stateinit->af);

  //Initialize the pin.
  GPIO_InitTypeDef init = stateinit->init;
  GPIO_Init(portptr, &init);
}

/* \return Bit_SET if the wakeup pin is high during boot.
   \return Bit_RESET if the wakeup pin is low during boot.*/
uint8_t WakeupPinBootState()
{
  return wakeupstate;
}
