#include "guileds.h"

CGUILeds::CGUILeds()
{
  m_i2c     = nullptr;
  m_addr    = 0;
  m_enabled = false;
}

CGUILeds::~CGUILeds()
{
}

/*! Sets up the IO expander which drives the GUI leds.
    \param i2c The I2C bus the IO expander is connected to.
    \param addr The I2C address the IO expander is connected to.
    \param orientation The orientation of the display next to the GUI leds.
    \return true The IO expander has replied to all communication.
    \return false The IO expander does not reply on its I2C address.*/
bool CGUILeds::Setup(CI2C* i2c, uint8_t addr, EOrientation orientation /*= Orientation_Top*/)
{
  m_i2c         = i2c;
  m_addr        = addr;
  m_orientation = orientation;

  if (!m_ioexpander.Setup(m_i2c, m_addr, "guileds"))
    return false;

  if (!m_ioexpander.ResetToDefaults())
    return false;

  if (!m_ioexpander.SetAsOutput(IOEXP_ALL))
    return false;

  m_enabled = true;

  return true;
}

/*! Turns on the selected LEDs.
    \param leds The LEDs to turn on.*/
void CGUILeds::On(uint8_t leds)
{
  m_ioexpander.Set(Translate(leds));
}

/*! Turns off the selected LEDs.
    \param leds The LEDs to turn on.*/
void CGUILeds::Off(uint8_t leds)
{
  m_ioexpander.Reset(Translate(leds));
}

/*! Rotates the led function according to m_orientation.
    \param leds Bitmask of the leds that should be turned on or off.
    \return Translated bitmask of the leds that should be turned on or off.*/
uint8_t CGUILeds::Translate(uint8_t leds)
{
  uint8_t output = 0;
  if (m_orientation == Orientation_Top)
  {
    output = leds; //No translation needed.
  }
  else if (m_orientation == Orientation_Right)
  { //The display is on the right, rotate the LEDs to the left.
    if (leds & GUILED_TOP_LEFT_RED)
      output |= GUILED_BOT_LEFT_RED;
    if (leds & GUILED_TOP_LEFT_GRN)
      output |= GUILED_BOT_LEFT_GRN;

    if (leds & GUILED_TOP_RIGHT_RED)
      output |= GUILED_TOP_LEFT_RED;
    if (leds & GUILED_TOP_RIGHT_GRN)
      output |= GUILED_TOP_LEFT_GRN;

    if (leds & GUILED_BOT_LEFT_RED)
      output |= GUILED_BOT_RIGHT_RED;
    if (leds & GUILED_BOT_LEFT_GRN)
      output |= GUILED_BOT_RIGHT_GRN;

    if (leds & GUILED_BOT_RIGHT_RED)
      output |= GUILED_TOP_RIGHT_RED;
    if (leds & GUILED_BOT_RIGHT_GRN)
      output |= GUILED_TOP_RIGHT_GRN;
  }
  else if (m_orientation == Orientation_Bottom)
  { //The display is on the bottom, rotate the LEDs 180 degrees.
    if (leds & GUILED_TOP_LEFT_RED)
      output |= GUILED_BOT_RIGHT_RED;
    if (leds & GUILED_TOP_LEFT_GRN)
      output |= GUILED_BOT_RIGHT_GRN;

    if (leds & GUILED_TOP_RIGHT_RED)
      output |= GUILED_BOT_LEFT_RED;
    if (leds & GUILED_TOP_RIGHT_GRN)
      output |= GUILED_BOT_LEFT_GRN;

    if (leds & GUILED_BOT_LEFT_RED)
      output |= GUILED_TOP_RIGHT_RED;
    if (leds & GUILED_BOT_LEFT_GRN)
      output |= GUILED_TOP_RIGHT_GRN;

    if (leds & GUILED_BOT_RIGHT_RED)
      output |= GUILED_TOP_LEFT_RED;
    if (leds & GUILED_BOT_RIGHT_GRN)
      output |= GUILED_TOP_LEFT_GRN;
  }
  else if (m_orientation == Orientation_Left)
  { //The display is on the left, rotate the LEDs to the right.
    if (leds & GUILED_TOP_LEFT_RED)
      output |= GUILED_TOP_RIGHT_RED;
    if (leds & GUILED_TOP_LEFT_GRN)
      output |= GUILED_TOP_RIGHT_GRN;

    if (leds & GUILED_TOP_RIGHT_RED)
      output |= GUILED_BOT_RIGHT_RED;
    if (leds & GUILED_TOP_RIGHT_GRN)
      output |= GUILED_BOT_RIGHT_GRN;

    if (leds & GUILED_BOT_LEFT_RED)
      output |= GUILED_TOP_LEFT_RED;
    if (leds & GUILED_BOT_LEFT_GRN)
      output |= GUILED_TOP_LEFT_GRN;

    if (leds & GUILED_BOT_RIGHT_RED)
      output |= GUILED_BOT_LEFT_RED;
    if (leds & GUILED_BOT_RIGHT_GRN)
      output |= GUILED_BOT_LEFT_GRN;
  }

  return output;
}
