#include "http.h"

#include <algorithm>
#include <cstring>

/*! \param txbuf Buffer to read upload data from, may be the same as rxbuf.
    \param txbufsize The number of bytes to read from txbuf and upload.
    \param rxbuf The buffer to store downloaded bytes into, may be the same as txbuf.
    \param rxbfufsize The size of the buffer that rxbuf points to.*/
CHTTPTransceiver::CHTTPTransceiver(uint8_t* txbuf, uint32_t txbufsize, uint8_t* rxbuf, uint32_t rxbufsize)
{
  m_txbuf      = txbuf;
  m_txbufsize  = txbufsize;
  m_rxbuf      = rxbuf;
  m_rxbufsize  = rxbufsize;

  m_uploaded   = 0;
  m_downloaded = 0;
  m_overflow   = 0;
}

/*! Sets the tx buffer and number of bytes that should be uploaded.
    \param txbuf Pointer to the buffer that should be uploaded.
    \param txbufsize Number of bytes that should be uploaded.*/
void CHTTPTransceiver::SetTxBuf(uint8_t* txbuf, uint32_t txbufsize)
{
  m_txbuf     = txbuf;
  m_txbufsize = txbufsize;
  m_uploaded  = 0;
  m_overflow  = 0;
}

/*! Tells the modem class how many bytes to upload via HTTP POST.*/
uint32_t CHTTPTransceiver::UploadSize()
{
  return m_txbufsize;
}

/*! Called before an upload is started.*/
void CHTTPTransceiver::StartUpload(const char* url)
{
  m_uploaded = 0;
}

/*! Called by the modem class every time data should be uploaded.
    \param buf The buffer to copy bytes into that should be uploaded.
    \param bufsize The maximum number of bytes that can be copied into buf.
    \return The number of bytes that have been copied into buf.*/
uint32_t CHTTPTransceiver::Upload(uint8_t* buf, uint32_t bufsize)
{
  uint32_t uploadsize = std::min(bufsize, m_txbufsize - m_uploaded);
  memcpy(buf, m_txbuf + m_uploaded, uploadsize);
  m_uploaded += uploadsize;
  return uploadsize;
}

/*! Returns the preferred block size of data downloads.
    Because this class writes to a single ram buffer, return the size of that buffer.*/
uint32_t CHTTPTransceiver::DownloadBlockSize()
{
  return m_rxbufsize;
}

/*! Called before the download is started.*/
void CHTTPTransceiver::StartDownload(const char* url, uint32_t httprspcode, uint32_t contentlength)
{
  m_downloaded  = 0;
  m_overflow    = 0;
  m_httprspcode = httprspcode;
}

/*! Called for every block of data that is read from the modem.
    \param buf The buffer that has the downloaded data.
    \param size The number of downloaded bytes in buf.*/
void CHTTPTransceiver::Download(const uint8_t* buf, uint32_t size)
{
  uint32_t downloadsize = std::min(size, m_rxbufsize - m_downloaded);
  if (downloadsize > 0)
  {
    memcpy(m_rxbuf + m_downloaded, buf, downloadsize);
    m_downloaded += downloadsize;
  }

  //increase m_overflow with the number of bytes that could
  //not be stored in m_rxbuf
  if (downloadsize < size)
    m_overflow += size - downloadsize;
}
