#include "i2c.h"
#include "lock.h"
#include "clock.h"
#include "i2c.h"

#include <cassert>
#include <cstdlib>

//TODO: add doxygen documentation to this file

CI2C::CI2C()
{
  m_i2c     = nullptr;
  m_ackfail = 0;
  m_mutex   = nullptr;
}

CI2C::~CI2C()
{
  Close();
}

void CI2C::Setup(I2C_TypeDef* i2c, SGPIOPin sclpin, uint32_t clockspeed /*= DEFAULT_I2C_CLOCK_SPEED*/)
{
  //shut down the I2C peripheral if it was previously set up
  Close();

  m_mutex = xSemaphoreCreateRecursiveMutex();

  //check if i2c points to a valid I2C peripheral
  assert(i2c == I2C1 || i2c == I2C2 || i2c == I2C3);
  m_i2c = i2c;

  //Send pulses on the clock line to reset any I2C devices.
  //This must be done before the I2C peripheral is initialized.
  SendReset(sclpin);

  //enable the clock for the spi peripheral
  SetClock(ENABLE);

  //initialize the I2C peripheral
  I2C_InitTypeDef i2cinit;
  I2C_StructInit(&i2cinit);
  i2cinit.I2C_ClockSpeed = clockspeed; //use the default i2c clock speed
  I2C_Init(m_i2c, &i2cinit);
  I2C_Cmd(m_i2c, ENABLE);
}

void CI2C::Close()
{
  if (m_i2c)
  {
    //protect the RCC registers with a critical section
    CCritical critsection;

    //reset and turn off the I2C peripheral
    Stop();
    I2C_Cmd(m_i2c, DISABLE);
    I2C_DeInit(m_i2c);
    SetClock(DISABLE);
    m_i2c = nullptr;

    vSemaphoreDelete(m_mutex);
  }
}

/*! Sends clock pulses on the I2C bus which might reset stuck I2C devices.*/
void CI2C::SendReset(SGPIOPin sclpin)
{
  CLock lock(m_mutex);

  //Set the SCL pin to gpio.
  GPIO_InitTypeDef sclinit = { sclpin.pin, GPIO_Mode_OUT, GPIO_Speed_2MHz, GPIO_OType_OD, GPIO_PuPd_UP };
  CCritical critsection;
  GPIO_Init(sclpin.port, &sclinit);
  critsection.Leave();

  //Send pulses on the clock line to reset any I2C devices.
  // EDWIN : some people say more pulses are needed was 10 now 27
  for (uint32_t i = 0; i < 27; i++)
  {
    GPIO_Reset(sclpin);
    DelayUS(10);
    GPIO_Set(sclpin);
    DelayUS(10);
  }

  //Set the SCL pin back to I2C.
  sclinit.GPIO_Mode = GPIO_Mode_AF;
  critsection.Enter();
  GPIO_Init(sclpin.port, &sclinit);
  critsection.Leave();
}

void CI2C::SetClock(FunctionalState state)
{
  assert(m_i2c);

  //protect the RCC registers with a critical section
  CCritical critsection;

  //enable the clock for the I2C peripheral
  if (m_i2c == I2C1)
    RCC_APB1PeriphClockCmd(RCC_APB1Periph_I2C1, state);
  else if (m_i2c == I2C2)
    RCC_APB1PeriphClockCmd(RCC_APB1Periph_I2C2, state);
  else if (m_i2c == I2C3)
    RCC_APB1PeriphClockCmd(RCC_APB1Periph_I2C3, state);
}

#define TIMEOUT 1000000 //!< Timeout in microseconds waiting on the I2C bus.

//TODO: use interrupts in this function, so that the RTOS is not blocked
bool CI2C::Write(uint8_t address, const uint8_t* buf1, uint32_t size1, const uint8_t* buf2 /*= nullptr*/, uint32_t size2 /*= 0*/)
{
  CLock lock(m_mutex);

  m_ackfail = 0;

  //generate a start condition
  uint32_t start = Micros();
  I2C_GenerateSTART(m_i2c, ENABLE);
  while (I2C_CheckEvent(m_i2c, I2C_EVENT_MASTER_MODE_SELECT) != SUCCESS)
  {
    if (Micros() - start > TIMEOUT)
      return false;
  }

  //send a transmit command to the address
  start = Micros();
  I2C_Send7bitAddress(m_i2c, address, I2C_Direction_Transmitter);
  while (I2C_CheckEvent(m_i2c, I2C_EVENT_MASTER_TRANSMITTER_MODE_SELECTED) != SUCCESS)
  {
    if (!CheckAck())
      return false;

    if (Micros() - start > TIMEOUT)
      return false;
  }

  for (uint32_t b = 0; b < 2; b++)
  {
    //select which buffer to use
    const uint8_t* buf;
    uint32_t       size;
    if (b == 0)
    {
      buf  = buf1;
      size = size1;
    }
    else
    {
      if (buf2 == nullptr)
        break; //if buf2 is not passed, this loop is done

      buf  = buf2;
      size = size2;
    }

    for (uint32_t i = 0; i < size; i++)
    {
      //transmit the byte
      start = Micros();
      I2C_SendData(m_i2c, buf[i]);
      while (I2C_CheckEvent(m_i2c, I2C_EVENT_MASTER_BYTE_TRANSMITTED) != SUCCESS)
      {
        if (!CheckAck())
        {
          m_ackfail = i + 1; //store which byte failed to ack
          return false;
        }

        if (Micros() - start > TIMEOUT)
          return false;
      }
    }
  }

  return true;
}

//TODO: use interrupts in this function, so that the RTOS is not blocked
bool CI2C::Read(uint8_t address, uint8_t* buf, uint32_t size)
{
  CLock lock(m_mutex);

  m_ackfail = 0;

  //generate a start condition
  uint32_t start = Micros();
  I2C_GenerateSTART(m_i2c, ENABLE);
  while (I2C_CheckEvent(m_i2c, I2C_EVENT_MASTER_MODE_SELECT) != SUCCESS)
  {
    if (Micros() - start > TIMEOUT)
      return false;
  }

  //send a receive command to the address
  start = Micros();
  I2C_Send7bitAddress(m_i2c, address, I2C_Direction_Receiver);
  while (I2C_CheckEvent(m_i2c, I2C_EVENT_MASTER_RECEIVER_MODE_SELECTED) != SUCCESS)
  {
    if (!CheckAck())
      return false;

    if (Micros() - start > TIMEOUT)
      return false;
  }

  //every byte but the last should be acked by the receiver
  if (size > 1)
    I2C_AcknowledgeConfig(m_i2c, ENABLE);

  for (uint32_t i = 0; i < size; i++)
  {
    //wait for a byte to be received
    start = Micros();
    while (I2C_CheckEvent(m_i2c, I2C_EVENT_MASTER_BYTE_RECEIVED) != SUCCESS)
    {
      if (Micros() - start > TIMEOUT)
      {
        m_ackfail = i + 1; //store which byte timed out
        return false;
      }
    }

    //disable the ack if the next iteration is the last byte
    if (i + 1 == size)
      I2C_AcknowledgeConfig(m_i2c, DISABLE);

    //receive the byte
    buf[i] = I2C_ReceiveData(m_i2c);
  }

  return true;
}

/*! Checks the ack failed bit, if set it will clear it, generate an I2C stop and return false.*/
bool CI2C::CheckAck()
{
  if (I2C_GetFlagStatus(m_i2c, I2C_FLAG_AF) == SET)
  {
    I2C_ClearFlag(m_i2c, I2C_FLAG_AF);
    Stop();
    return false;
  }

  return true;
}

void CI2C::Stop()
{
  CLock lock(m_mutex);

  //generate a stop condition
  I2C_GenerateSTOP(m_i2c, ENABLE);
  DelayMS(1);
  I2C_GenerateSTOP(m_i2c, DISABLE);
}

/* \return A printable string describing the I2C peripheral.*/
const char* CI2C::Name()
{
  if (m_i2c == I2C1)
    return "I2C1";
  else if (m_i2c == I2C2)
    return "I2C2";
  else if (m_i2c == I2C3)
    return "I2C3";
  else
    return "ERROR UNKNOWN I2C PERIPHERAL";
}
