#include "internalflash.h"
#include "log.h"
#include "lock.h"

#include <cassert>

static SemaphoreHandle_t g_flashmutex;

/*! Creates a mutex to lock access to the flash interface.
    This prevents multiple tasks from erasing or writing flash at the same time.*/
void InitFlashMutex()
{
  //Use a test and test-and-set method to assign a mutex to g_flashmutex.
  //But don't allocate it from inside a critical section.
  if (g_flashmutex == nullptr)
  {
    SemaphoreHandle_t mutex = xSemaphoreCreateRecursiveMutex();

    CCritical criticalsection;
    if (g_flashmutex == nullptr)
    {
      g_flashmutex = mutex;
      mutex = nullptr;
    }
    criticalsection.Leave();

    if (mutex)
      vSemaphoreDelete(mutex);
  }
}

/*! Returns a handle to g_flashmutex, can be used with CLock to synchronize
    access to the internal flash. The flash is divided into two banks,
    a read from one bank cannot occur at the same time as a write or erase on
    the same bank, if that happens a bus stall occurs.*/
SemaphoreHandle_t FlashMutex()
{
  InitFlashMutex();

  return g_flashmutex;
}

/*! Converts a FLASH_Status to a human readable string.
    \param status The status to convert.
    \return A human readable string constant.*/
const char* FlashStatusStr(FLASH_Status status)
{
  if (status == FLASH_BUSY)
    return "busy";
  else if (status == FLASH_ERROR_RD)
    return "read protection enabled";
  else if (status == FLASH_ERROR_PGS)
    return "FLASH_ERROR_PGS"; //not used anywhere, unknown what it means
  else if (status == FLASH_ERROR_PGP)
    return "FLASH_ERROR_PGP"; //not used anywhere, unknown what it means
  else if (status == FLASH_ERROR_PGA)
    return "FLASH_ERROR_PGA"; //not used anywhere, unknown what it means
  else if (status == FLASH_ERROR_WRP)
    return "write protection enabled";
  else if (status == FLASH_ERROR_PROGRAM)
    return "programming error";
  else if (status == FLASH_ERROR_OPERATION)
    return "operation error";
  else if (status == FLASH_COMPLETE)
    return "operation complete";
  else
    return "ERROR UNKNOWN STATUS";
}

static const struct
{
  uint32_t id;
  uint32_t addr;
  uint32_t size;
}
sectors[] =
{
  //Bank 1
  { FLASH_Sector_0,   0x08000000, 16  * 1024 },
  { FLASH_Sector_1,   0x08004000, 16  * 1024 },
  { FLASH_Sector_2,   0x08008000, 16  * 1024 },
  { FLASH_Sector_3,   0x0800C000, 16  * 1024 },
  { FLASH_Sector_4,   0x08010000, 64  * 1024 },
  { FLASH_Sector_5,   0x08020000, 128 * 1024 },
  { FLASH_Sector_6,   0x08040000, 128 * 1024 },
  { FLASH_Sector_7,   0x08060000, 128 * 1024 },
  { FLASH_Sector_8,   0x08080000, 128 * 1024 },
  { FLASH_Sector_9,   0x080A0000, 128 * 1024 },
  { FLASH_Sector_10,  0x080C0000, 128 * 1024 },
  { FLASH_Sector_11,  0x080E0000, 128 * 1024 },
  //Bank 2
  { FLASH_Sector_12,  0x08100000, 16  * 1024 },
  { FLASH_Sector_13,  0x08104000, 16  * 1024 },
  { FLASH_Sector_14,  0x08108000, 16  * 1024 },
  { FLASH_Sector_15,  0x0810C000, 16  * 1024 },
  { FLASH_Sector_16,  0x08110000, 64  * 1024 },
  { FLASH_Sector_17,  0x08120000, 128 * 1024 },
  { FLASH_Sector_18,  0x08140000, 128 * 1024 },
  { FLASH_Sector_19,  0x08160000, 128 * 1024 },
  { FLASH_Sector_20,  0x08180000, 128 * 1024 },
  { FLASH_Sector_21,  0x081A0000, 128 * 1024 },
  { FLASH_Sector_22,  0x081C0000, 128 * 1024 },
  { FLASH_Sector_23,  0x081E0000, 128 * 1024 },
};

/*! Converts a sector number from 0 to 23 to a FLASH_Sector define.
    \param sectornum She number of the sector.
    \return A value of @FLASH_Sectors.*/
uint16_t FlashNumToSectorID(uint32_t sectornum)
{
  assert(sectornum < sizeof(sectors) / sizeof(sectors[0]));
  return sectors[sectornum].id;
}

/*! Returns the memory address of the flash sector.
    \param sectornum The number of the sector.
    \return The memory address of the sector.*/
uint32_t FlashSectorAddress(uint32_t sectornum)
{
  assert(sectornum < sizeof(sectors) / sizeof(sectors[0]));
  return sectors[sectornum].addr;
}

/*! Returns the combined size of one or more sectors.
    \param startsector The sector to start counting at.
    \param numsectors The number of sectors to count the size of.*/
uint32_t FlashSectorsSize(uint32_t startsector, uint32_t numsectors)
{
  uint32_t size = 0;
  for (uint32_t i = startsector; i < startsector + numsectors; i++)
  {
    assert(i < sizeof(sectors) / sizeof(sectors[0]));
    size += sectors[i].size;
  }

  return size;
}

/*! Returns the size of one sector.
    \param sector The sector to return the size of.
    \return the size of the sector in bytes.*/
uint32_t FlashSectorSize(uint32_t sector)
{
  assert(sector < sizeof(sectors) / sizeof(sectors[0]));
  return sectors[sector].size;
}

/*! Checks if all bits of a sector are set to 1.
    \param sectornum The number of the sector to check.
    \return true if all bits the sector are 1, false if one or more bits are 0.*/
bool FlashSectorIsErased(uint32_t sectornum)
{
  assert(sectornum < sizeof(sectors) / sizeof(sectors[0]));

  for (uint32_t i = sectors[sectornum].addr;
       i < sectors[sectornum].addr + sectors[sectornum].size;
       i += 4)
  {
    if (*((uint32_t*)i) != 0xFFFFFFFF)
      return false;
  }

  return true;
}

/*! Checks if a sector is already erased, if not will execute the erase action.
    \param sectornum The sector number to erase.
    \return true if the sector is already erased, or if the erase suceeded, false if the erase failed.*/
bool FlashEraseSector(uint32_t sectornum)
{
  InitFlashMutex();
  CLock lock(g_flashmutex);

  if (FlashSectorIsErased(sectornum))
    return true;

  FLASH_Unlock();
  FLASH_Status status = FLASH_EraseSector(FlashNumToSectorID(sectornum), VoltageRange_3);
  FLASH_Lock();

  if (status != FLASH_COMPLETE)
  {
    PrintLog(LOGINTFLASH, "ERROR: unable to erase sector %" PRIu32 ": %" PRIi32":%s\n",
             sectornum, (int32_t)status, FlashStatusStr(status));
    return false;
  }

  return true;
}

/* Erases several flash sectors.
   \param startsector The sector to start erasing at.
   \param numsectors The number of sectors to erase.
   \return true if the erasing succeeded, false if it failed.*/
bool FlashEraseSectors(uint32_t startsector, uint32_t numsectors)
{
  InitFlashMutex();
  CLock lock(g_flashmutex);

  for (uint32_t i = startsector; i < startsector + numsectors; i++)
  {
    if (!FlashEraseSector(i))
      return false;
  }

  return true;
}


/*! Writes data to the internal flash.
    \param address The address of the internal flash to write to.
    \param buf The buffer to write to the internal flash.
    \param numbytes The number of bytes to write to the flash.*/
bool FlashWrite(uint32_t address, uint8_t* buf, uint32_t numbytes)
{
  InitFlashMutex();
  CLock lock(g_flashmutex);

  uint32_t writeaddress = address;
  uint32_t endaddress   = address + numbytes;
  uint8_t* readptr      = buf;

  FLASH_Status status = FLASH_COMPLETE;
  while (writeaddress < endaddress)
  {
    if ((endaddress - writeaddress >= 4) && (writeaddress % 4 == 0))
    { //if there are at least 4 bytes to write, and the write address is
      //aligned to 4 bytes, use FLASH_ProgramWord to program 4 bytes at a time
      FLASH_Unlock();
      status = FLASH_ProgramWord(writeaddress, *((uint32_t*)readptr));
      FLASH_Lock();
      writeaddress += 4;
      readptr      += 4;
    }
    else
    { //use FLASH_ProgramByte to program a single byte at a time
      FLASH_Unlock();
      status = FLASH_ProgramByte(writeaddress, *readptr);
      FLASH_Lock();
      writeaddress++;
      readptr++;
    }

    if (status != FLASH_COMPLETE)
    {
      PrintLog(LOGINTFLASH, "ERROR: unable to write address %" PRIu32 ", %" PRIi32":%s\n",
               writeaddress, (int32_t)status, FlashStatusStr(status));
      return false;
    }
  }

  return true;
}
