#include "ioexpander.h"
#include "log.h"
#include "lock.h"

#include <cctype>
#include <cassert>
#include <cstring>

//MCP23008 registers
#define MCP23008_IODIR   0x00
#define MCP23008_IPOL    0x01
#define MCP23008_GPINTEN 0x02
#define MCP23008_DEFVAL  0x03
#define MCP23008_INTCON  0x04
#define MCP23008_IOCON   0x05
#define MCP23008_GPPU    0x06
#define MCP23008_INTF    0x07
#define MCP23008_INTCAP  0x08
#define MCP23008_GPIO    0x09
#define MCP23008_OLAT    0x0A

#define MCP23008_INTPOL  (1 << 1) //!< INTPOL bit in the IOCON register

CIOExpander::CIOExpander()
{
  m_i2c  = nullptr;
  m_name = nullptr;
}

CIOExpander::~CIOExpander()
{
}

/*! Stores pointers and the IO expander's address, and checks if the MCP23008 responds.
    \param i2c A pointer to a CI2C instance that has been set up.
    \param addr The address of the MCP23008 IO expander.
    \paran name The name of the IO expander, used for logging, should be at least 1 character.*/
bool CIOExpander::Setup(CI2C* i2c, uint8_t addr, const char* name)
{
  m_i2c  = i2c;
  m_addr = addr;
  m_name = name;

  assert(strlen(m_name) > 0);

  CLock lock(*m_i2c);

  //start by sending a stop condition, to reset all communication
  m_i2c->Stop();

  //test if there is an I2C device at the correct address
  if (m_i2c->Write(m_addr, nullptr, 0))
  {
    PrintLog(LOGIOEXPANDER, "%c%s IO expander responded at address %" PRIu8 "\n",
             toupper(m_name[0]), m_name + 1, m_addr);
    m_i2c->Stop();
  }
  else
  {
    PrintLog(LOGIOEXPANDER, "ERROR: %c%s IO expander is not responding at address %" PRIu8 "\n",
             toupper(m_name[0]), m_name + 1, m_addr);
    m_i2c->Stop();
    return false;
  }

  return true;
}

/*! Reads a register from the MCP23008 IO expander.
    \param reg The register to read.
    \param value The register value is stored here after a successful read.
    \return true if reading the register succeeded.
    \return false if reading the register failed.*/
bool CIOExpander::ReadRegister(uint8_t reg, uint8_t* value)
{
  CLock lock(*m_i2c);

  //first write the register address to the MCP23008
  if (!m_i2c->Write(m_addr, &reg, 1))
  {
    PrintLog(LOGIOEXPANDER, "ERROR: Unable to write to %s IO expander, failed to ack at byte %" PRIu32 "\n",
             m_name, m_i2c->AckFail());
    m_i2c->Stop();
    return false;
  }

  //read from the MCP23008, the first byte read is the register value
  if (!m_i2c->Read(m_addr, value, 1))
  {
    PrintLog(LOGIOEXPANDER, "ERROR: Unable to read from %s IO expander, failed to ack at byte %" PRIu32 "\n",
             m_name, m_i2c->AckFail());
    m_i2c->Stop();
    return false;
  }
  m_i2c->Stop();

  return true;
}

/*! Writes a register on the MCP23008 IO expander.
    \param reg The register to write.
    \pararm value The value to write to the register.
    \return true if writing the register succeeded.
    \return false if writing the register failed.*/
bool CIOExpander::WriteRegister(uint8_t reg, uint8_t value)
{
  CLock lock(*m_i2c);

  //write the register address and the register value
  //this sets the register to the wanted value
  if (!m_i2c->Write(m_addr, &reg, 1, &value, 1))
  {
    PrintLog(LOGIOEXPANDER, "ERROR: Unable to write to %s IO expander, failed to ack at byte %" PRIu32 "\n",
             m_name, m_i2c->AckFail());
    m_i2c->Stop();
    return false;
  }
  m_i2c->Stop();

  return true;
}

/*! Sets the value of specific pin bits in a specific register.
    \param reg The register to set bits in.
    \param pins A bitwise OR of the IOEXP_GP pin definitions.
    \param value The value to set the bits to.
    \param regname The name of the register, used for logging.*/
bool CIOExpander::SetPinsInReg(uint8_t reg, uint8_t pins, bool value, const char* regname)
{
  uint8_t regval;
  if (!ReadRegister(reg, &regval))
  {
    PrintLog(LOGIOEXPANDER, "ERROR: Unable to read %s register from %s IO expander\n",
             regname, m_name);
    return false;
  }

  if (value)
    regval |= pins;
  else
    regval &= ~pins;

  if (!WriteRegister(reg, regval))
  {
    PrintLog(LOGIOEXPANDER, "ERROR: Unable to write %s register on %s IO expander\n",
             regname, m_name);
    return false;
  }

  return true;
}

/*! Sets all registers to default values, with the exception of the INTCAP register.
    \return true if setting all registers to default values succeeded.
    \return false if anything failed.*/
bool CIOExpander::ResetToDefaults()
{
  //set the IOCON register to its default value
  if (!WriteRegister(MCP23008_IOCON, 0))
    return false;

  //disable anything that can generate an interrupt
  if (!DisableInterrupt(IOEXP_ALL))
    return false;
  if (!SetDefaultCompare(IOEXP_ALL, false))
    return false;
  if (!SetInterruptOnChange(IOEXP_ALL))
    return false;

  //set all pins as input
  if (!SetAsInput(IOEXP_ALL))
    return false;
  if (!SetInputPolarity(IOEXP_ALL, false))
    return false;
  if (!DisablePullup(IOEXP_ALL))
    return false;

  //make the output latch of all pins low
  if (!Reset(IOEXP_ALL))
    return false;

  return true;
}

/*! Sets pins on the MCP23008 IO expander as input, by setting their
    corresponding bits in the IODIR register to 1.
    \param pins A bitwise OR of the IOEXP_GP pin definitions.
    \return true if setting the pin as input succeeded, false if it failed.*/
bool CIOExpander::SetAsInput(uint8_t pins)
{
  return SetPinsInReg(MCP23008_IODIR, pins, true, "IODIR");
}

/* Sets pins on the MCP23008 IO expander as output, by setting their
   corresponding bits in the IODIR register to 0.
   \param pins A bitwise OR of the IOEXP_GP pin definitions.
   \return true if setting the pin as input succeeded, false if it failed.*/
bool CIOExpander::SetAsOutput(uint8_t pins)
{
  return SetPinsInReg(MCP23008_IODIR, pins, false, "IODIR");
}

/*! Reads the GPIO register of the MCP23008 IO expander, and checks if a pin is high or low.
    \param pin A value of IOEXP_GP0 to IOEXP_GP7, only one should be passed.
    \param value The value of the pin is returned here on a successful read.
    \return true if reading the GPIO register succeeded.
    \return false if reading the GPIO register failed.*/
bool CIOExpander::Read(uint8_t pin, bool* value)
{
  uint8_t gpioreg;
  if (!ReadRegister(MCP23008_GPIO, &gpioreg))
  {
    PrintLog(LOGIOEXPANDER, "ERROR: Unable to read GPIO register from %s IO expander\n", m_name);
    return false;
  }

  *value = (gpioreg & pin) == pin;

  return true;
}

/*! Reads the GPIO register of the MCP23008 IO expander, and returns it.
    \param gpio The value of the GPIO register is returned here.
    \return true if reading the GPIO register succeeded.
    \return false if reading the GPIO register failed.*/
bool CIOExpander::ReadGPIO(uint8_t* gpio)
{
  if (!ReadRegister(MCP23008_GPIO, gpio))
  {
    PrintLog(LOGIOEXPANDER, "ERROR: Unable to read GPIO register from %s IO expander\n", m_name);
    return false;
  }

  return true;
}

/*! Sets one or more pins of the MCP23008 IO expander to the high state,
    by writing to the OLAT register.
    \param pins A bitwise OR of IOEXP_GP0 to IOEXP_GP7.
    \return true if setting the pins succeeded.
    \return false if setting the pins failed.*/
bool CIOExpander::Set(uint8_t pins)
{
  return SetPinsInReg(MCP23008_OLAT, pins, true, "OLAT");
} 

/*! Sets one or more pins of the MCP23008 IO expander to the low state,
    by writing to the OLAT register.
    \param pins A bitwise OR of IOEXP_GP0 to IOEXP_GP7.
    \return true if setting the pins succeeded.
    \return false if setting the pins failed.*/
bool CIOExpander::Reset(uint8_t pins)
{
  return SetPinsInReg(MCP23008_OLAT, pins, false, "OLAT");
} 

/*! Sets the input polarity of one or more IO pins.
    \param pins A bitwise OR of IOEXP_GP0 to IOEXP_GP7.
    \param inverted Set to true to invert the read value of the pin.
    \return true if setting the input polarity succeeded.
    \return false if setting the input polarity failed.*/
bool CIOExpander::SetInputPolarity(uint8_t pins, bool inverted)
{
  return SetPinsInReg(MCP23008_IPOL, pins, inverted, "IPOL");
}

/*! Enables the interrupt for one or more IO pins.
    \param pins A bitwise OR of IOEXP_GP0 to IOEXP_GP7.
    \return true if enabling the interrupt succeeded.
    \return false if enabling the interrupt failed.*/
bool CIOExpander::EnableInterrupt(uint8_t pins)
{
  return SetPinsInReg(MCP23008_GPINTEN, pins, true, "GPINTEN");
}

/*! Disables the interrupt for one or more IO pins.
    \param pins A bitwise OR of IOEXP_GP0 to IOEXP_GP7.
    \return true if disabling the interrupt succeeded.
    \return false if disabling the interrupt failed.*/
bool CIOExpander::DisableInterrupt(uint8_t pins)
{
  return SetPinsInReg(MCP23008_GPINTEN, pins, false, "GPINTEN");
}

/*! Sets the default value of one or more pins to compare with.
    If the pin is not this value, an interrupt is generated.
    \param pins A bitwise OR of IOEXP_GP0 to IOEXP_GP7.
    \param value The default value to compare with.
    \return true if setting the default compare succeeded.
    \return false if setting the default compare failed.*/
bool CIOExpander::SetDefaultCompare(uint8_t pins, bool value)
{
  return SetPinsInReg(MCP23008_DEFVAL, pins, value, "DEFVAL");
}

/*! Configures one or more pins to generated an interrupt when its value changes.
    \param pins A bitwise OR of IOEXP_GP0 to IOEXP_GP7.
    \return true if setting the pins to interrupt on change succeeded.
    \return false if setting the pins to interrupt on change failed.*/
bool CIOExpander::SetInterruptOnChange(uint8_t pins)
{
  return SetPinsInReg(MCP23008_INTCON, pins, false, "INTCON");
}

/*! Configures one or more pins to generated an interrupt when its value is not
    its corresponding bit in the DEFVAL register.
    \param pins A bitwise OR of IOEXP_GP0 to IOEXP_GP7.
    \return true if setting the pins to interrupt on change succeeded.
    \return false if setting the pins to interrupt on change failed.*/
bool CIOExpander::SetInterruptOnNotDefault(uint8_t pins)
{
  return SetPinsInReg(MCP23008_INTCON, pins, true, "INTCON");
}

/*! Enables the pullup resistors for one or more pins.
    \param pins A bitwise OR of IOEXP_GP0 to IOEXP_GP7.
    \return true if enabling the pullup resistors succeeded.
    \return false if enabling the pullup resistors failed.*/
bool CIOExpander::EnablePullup(uint8_t pins)
{
  return SetPinsInReg(MCP23008_GPPU, pins, true, "GPPU");
}

/*! Disables the pullup resistors for one or more pins.
    \param pins A bitwise OR of IOEXP_GP0 to IOEXP_GP7.
    \return true if disabling the pullup resistors succeeded.
    \return false if disabling the pullup resistors failed.*/
bool CIOExpander::DisablePullup(uint8_t pins)
{
  return SetPinsInReg(MCP23008_GPPU, pins, false, "GPPU");
}

/* Reads the INTF register and returns if pin is currently causing an interrupt.
   \param pins A bitwise OR of IOEXP_GP0 to IOEXP_GP7.
   \param value The value of the corresponding bit in INTF is returned here.
   \return true if reading the interrupt flag succeeded.
   \return false if reading the interrupt flag failed.*/
bool CIOExpander::GetInterruptFlag(uint8_t pin, bool* value)
{
  uint8_t intfreg;
  if (!ReadRegister(MCP23008_INTF, &intfreg))
  {
    PrintLog(LOGIOEXPANDER, "ERROR: Unable to read INTF register from %s IO expander\n", m_name);
    return false;
  }

  *value = (intfreg & pin) == pin;

  return true;
}

/*! Returns the value of the interrupt capture register,
    this indicates which pin has caused the interrupt.
    Reading this register, or reading the GPIO register,
    clears the interrupt capture register.
    \param value The value of the interrupt capture register is returned here.
    \return true if reading the interrupt capture register succeeded.
    \return false if reading the interrupt capture register failed.*/
bool CIOExpander::GetInterruptCapture(uint8_t* value)
{
  if (!ReadRegister(MCP23008_INTCAP, value))
  {
    PrintLog(LOGIOEXPANDER, "ERROR: Unable to read INTCAP register from %s IO expander\n", m_name);
    return false;
  }

  return true;
}

/* Sets or clears the IOPOL bit in the IOCON register.
   If the IOPOL bit is set, the INT pin is inverted.
   \param inverted
   \return true if setting the IOPOL bit succeeded.
   \return false if setting the IOPOL bit failed.*/
bool CIOExpander::SetOutputPolarity(bool inverted)
{
  return SetPinsInReg(MCP23008_IOCON, MCP23008_INTPOL, inverted, "IOCON");
}
