#include "keyvaluestorage.h"
#include "internalflash.h"
#include "log.h"
#include "cbortools.h"
#include "lock.h"

#include <cstring>
#include <algorithm>

/* Two sectors are reserved for key-value pairs.
   Every time a key-value pair is changed, it's written at the end of the data of one sector,
   and a new SKeyValueHash struct is written before the previous one,
   starting at the end of the sector.
   When that sector is full, all key-value pairs are copied to the other sector,
   with duplicates filtered, and then the first sector is erased.
   This way, there is always a working copy of key-value pairs in the flash,
   and the number of erases is minimized.*/

#define KEY_VALUE_PAIRS_SECTOR_INVALID -1

/*! Value of the first 4 bytes of the sector when it's erased.*/
#define SECTOR_ERASED ((uint32_t)(0xFFFFFFFF))

#include <cassert>
#include <cstring>

CKeyValueStorage::CKeyValueStorage()
{
  m_keyvalueinfo = nullptr;
  m_numkeyvalues = 0;
  m_sector1      = KEY_VALUE_PAIRS_SECTOR_INVALID;
  m_sector2      = KEY_VALUE_PAIRS_SECTOR_INVALID;
  m_loglevel     = 0;
  m_mutex        = 0;
  m_inswitch     = false;
  ResetState();
}

CKeyValueStorage::~CKeyValueStorage()
{
}

/*! Checks if key-value pairs are stored in one of two sectors, and uses the most recent one.
    If no key-value pairs are stored, it will initialize the first sector to hold a CBOR map.
    \param keyvalueinfo An array of SKeyValueInfo that describes the key-value pairs.
    \param numkeyvalues The number of entries in keyvalueinfo.
    \param sector1 The first sector to use to store key-value pairs.
    \param sector2 The second sector to use to store key-value pairs.
    \param loglevel The loglevel to use when printing to the log.*/
void CKeyValueStorage::Init(const SKeyValueInfo* keyvalueinfo, uint32_t numkeyvalues,
                            uint32_t sector1, uint32_t sector2, int32_t loglevel)
{
  m_keyvalueinfo = keyvalueinfo;
  m_numkeyvalues = numkeyvalues;
  m_sector1      = sector1;
  m_sector2      = sector2;
  m_loglevel     = loglevel;

  m_mutex = xSemaphoreCreateRecursiveMutex();
  CLock lock(m_mutex);

  if (CheckStoredKeyValuePairs())
    return; //valid stored key-value pairs found

  if (RecoverStoredKeyValuePairs())
  {
    if (CheckStoredKeyValuePairs())
      return; //previously stored key-value pairs have been recovered successfully
  }

  //if no valid stored key-value pairs are found,
  //initialize the two sectors
  if (!InitStoredKeyValuePairs())
  {
    ResetState();
    return;
  }

  //the flash sectors were just initialized, there should be valid
  //stored key-value pairs found
  if (!CheckStoredKeyValuePairs())
  {
    PrintLog(m_loglevel, "ERROR: sector %" PRIu32 " initialized but no valid "
             "key-value pairs found, disabling custom key-value pairs\n", m_sector);
    ResetState();
  }
}

void CKeyValueStorage::ResetState()
{
  m_sector    = KEY_VALUE_PAIRS_SECTOR_INVALID;
  m_addr      = nullptr;
  m_size      = 0;
  m_writepos  = 0;
  m_hashptr   = nullptr;
  m_parser    = {};
  m_root      = {};
  m_writehash = true;
}

/*! Converts a key to an index of m_keyvalueinfo.
   \param key The key of which to get the index.
   \return The index of the key in m_keyvalueinfo.*/
uint32_t CKeyValueStorage::KeyIndex(uint32_t key)
{
  for (uint32_t i = 0; i < m_numkeyvalues; i++)
  {
    if (m_keyvalueinfo[i].key == key)
      return i;
  }

  assert(0);
  return 0;
}

/*! Converts a key to its corresponding SKeyValueInfo.*/
const SKeyValueInfo& CKeyValueStorage::KeyInfo(uint32_t key)
{
  return m_keyvalueinfo[KeyIndex(key)];
}

/*! Checks in which sector the key-value pairs are stored.
    Initializes the CBOR decoder and the write position.
    \return true if valid key-value pairs are found.
    \return false if no or invalid key-value pairs are found.*/
bool CKeyValueStorage::CheckStoredKeyValuePairs()
{
  if (!ChooseSector())
    return false;

  if (!CheckStoredHash())
    return false;

  if (!InitDecoder())
    return false;

  if (!CheckWritePosition())
    return false;

  return true;
}

/*! Chooses which sector to use for key-value pairs, and stores it in m_sector.
    The choice is made from the first 4 bytes, which is used as a sequence number.
    \return true if a sector has been chosen.
    \return if a sector has not been chosen, and InitSector() should be used.*/
bool CKeyValueStorage::ChooseSector()
{
  m_sector = KEY_VALUE_PAIRS_SECTOR_INVALID;

  uint32_t* sector1ptr = (uint32_t*)FlashSectorAddress(m_sector1);
  uint32_t* sector2ptr = (uint32_t*)FlashSectorAddress(m_sector2);

  if (*sector1ptr == SECTOR_ERASED && *sector2ptr == SECTOR_ERASED)
  {
    PrintLog(m_loglevel, "No key-value pairs are stored\n");
    return false; //both sectors are empty
  }

  //choose the sector that has the highest sequence number,
  //and it not erased
  if (*sector1ptr == SECTOR_ERASED)
    m_sector = m_sector2;
  else if (*sector2ptr == SECTOR_ERASED)
    m_sector = m_sector1;
  else if (*sector1ptr > *sector2ptr)
    m_sector = m_sector1;
  else if (*sector2ptr > *sector1ptr)
    m_sector = m_sector2;

  PrintLog(m_loglevel, "Using key-value pairs in sector %" PRIu32
           ", which has sequence number %" PRIu32 "\n", m_sector,
           *(uint32_t*)FlashSectorAddress(m_sector));

  m_addr = (uint8_t*)FlashSectorAddress(m_sector);
  m_size = FlashSectorSize(m_sector);

  return true;
}

/*! Reads the last stored hash at the end of the sector, compares it with a hash
    generated from stored key-value pairs at the beginning of the sector.
    \return true if the stored hash matches the generated hash.
    \return false if the stored hash is incorrect, or no hash is stored.*/
bool CKeyValueStorage::CheckStoredHash()
{
  //each time a key-value pair is stored, a new hash is written at the end of the sector
  //each new hash is written in front of the old hash, and there should always be
  //an erased block of size sizeof(SKeyValueHash) in front of the hashes
  //this loop finds the erased block
  SKeyValueHash* hashptr = ((SKeyValueHash*)(m_addr + m_size)) - 1;
  while ((uint8_t*)hashptr >= (m_addr + sizeof(uint32_t) + 1) && !hashptr->IsErased())
    hashptr--;

  //sanity check, hashptr should now point to an erased hash block
  if ((uint8_t*)hashptr < (m_addr + sizeof(uint32_t) + 1) || !hashptr->IsErased())
  {
    PrintLog(m_loglevel, "ERROR: did not find erased hash block in sector %" PRIu32 "\n", m_sector);
    return false;
  }

  //increment the pointer to find the first stored hash
  //if the pointer now points to m_addr + m_size, no hash is stored
  hashptr++;
  if ((uint8_t*)hashptr >= m_addr + m_size)
  {
    PrintLog(m_loglevel, "ERROR: sector %" PRIu32 " has no hash stored\n", m_sector);
    return false;
  }

  //sanity check, the number in the size field should not be bigger than the
  //number of bytes between the stored hash and the start address of the sector
  uint32_t maxstored = (uintptr_t)hashptr - (uintptr_t)m_addr;
  if (hashptr->size > maxstored)
  {
    PrintLog(m_loglevel, "ERROR: found %" PRIu32 " in size field in hash of sector %" PRIu32
             " but max size left before hash is %" PRIu32 " bytes\n", hashptr->size, m_sector, maxstored);
    return false;
  }

  //generate a hash of the bytes stored in the sector
  SKeyValueHash hash;
  CreateHash(m_addr, hashptr->size, &hash);

  //check if the stored hash is the same as the generated hash
  if (!std::equal(hash.hash, hash.hash + KEY_VALUE_HASH_SIZE, hashptr->hash))
  {
    CLogLock loglock;
    PrintLog(m_loglevel, "ERROR: stored hash ");
    for (uint32_t i = 0; i < KEY_VALUE_HASH_SIZE; i++)
      PrintLog(m_loglevel, "%02" PRIX8, hashptr->hash[i]);

    PrintLog(m_loglevel, " in sector %" PRIu32 " is not the same as generated hash ", m_sector);

    for (uint32_t i = 0; i < KEY_VALUE_HASH_SIZE; i++)
      PrintLog(m_loglevel, "%02" PRIX8, hash.hash[i]);

    PrintLog(m_loglevel, "\n");

    return false;
  }

  PrintLog(m_loglevel, "Hash of sector %" PRIu32 " at position %" PRIuPTR " is correct\n",
           m_sector, (uintptr_t)hashptr - (uintptr_t)m_addr);

  m_hashptr  = hashptr;
  m_writepos = hashptr->size;

  PrintLog(m_loglevel, "Write position: %" PRIu32 "\n", m_writepos);

  return true;
}

/*! Sets up a CBOR decoder on the selected sector.
    \return true if setting up the CBOR decoder succeeed.
    \return false if setting up the CBOR decoder failed.*/
bool CKeyValueStorage::InitDecoder()
{
  //initialize the parser and check its return value
  CborError retval;
  retval = cbor_parser_init(m_addr + sizeof(uint32_t), m_size - sizeof(uint32_t), CborValidateBasic, &m_parser, &m_root);
  if (retval != CborNoError)
  {
    PrintLog(m_loglevel, "ERROR: cbor_parser_init returned %i:%s\n",
             (int)retval, cbor_error_string(retval));
    return false;
  }

  if (!cbor_value_is_map(&m_root))
  { //the first setting element should always be a map
    CLogLock loglock;
    PrintLog(m_loglevel, "ERROR: first key-value pairs element is not a map: ");
    cbor_value_to_pretty(stdout, &m_root);
    PrintLog(m_loglevel, "\n");
    return false;
  }

  return true;
}

/*! Uses the CBOR decoder to find the break byte (0xFF) of the CBOR root map.
    It then checks if the break byte is at m_addr[m_writepos], which is where it should be.
    \return true if the break byte is found and it's at m_addr[m_writepos].
    \return false on any error.*/
bool CKeyValueStorage::CheckWritePosition()
{
  CborError retval;

  //use a new CborValue to advance past the root map, this avoids changing m_root
  CborValue end = m_root;
  retval = cbor_value_advance(&end);
  if (retval != CborNoError)
  {
    PrintLog(m_loglevel, "ERROR: cbor_value_advance on the root map returned %i:%s\n",
             (int)retval, cbor_error_string(retval));
    return false;
  }

  //sanity check, there should be no more CBOR left after advancing past the root map
  if (!cbor_value_at_end(&end))
  {
    PrintLog(m_loglevel, "ERROR: cbor_value_at_end returned false after advancing past the root map\n");
    return false;
  }

  //end.ptr now points to one byte after the root map
  //the write position should be set to one byte before that
  uint32_t writepos = (uintptr_t)end.ptr - (uintptr_t)m_addr - 1;
  if (writepos >= m_size)
  {
    PrintLog(m_loglevel, "ERROR: calculated invalid write position %" PRIu32
             " from start pointer 0x%" PRIXPTR " and end pointer 0x%" PRIXPTR "\n",
             writepos, (uintptr_t)m_addr, (uintptr_t)end.ptr);
    return false;
  }

  //the write position should match with the size stored together with the hash
  if (writepos != m_writepos)
  {
    PrintLog(m_loglevel, "ERROR: calculated write position %" PRIu32 " does not match "
             "stored size %" PRIu32 "\n", writepos, m_writepos);
    return false;
  }

  //the root map is a map of indefinite length, meaning that it ends at a break byte (0xFF)
  if (m_addr[writepos] != 0xFF)
  {
    PrintLog(m_loglevel, "ERROR: %02" PRIX8 "character found at write position, instead of 0xFF\n",
             m_addr[writepos]);
    return false;
  }

  //any new key-value pairs should be written at m_addr + m_writepos
  //as CBOR key-value pairs, for as long as one 0xFF byte remains at the end
  //to close the CBOR root map

  return true;
}

/*! Checks the sequence numbers of the sectors, places them in the
    sectors aray, highest sequence number first. A sector that is erased is
    not placed in the sectors array.
    The sector array then holds the sectors that should be checked
    for recovery, in order of highest sequence number.
    \param sectors The sectors that should be checked for recovery are
                   returned here.
    \return true if one or more sectors should be checked for recovery.
    \return false if no sector should be checked for recovery.*/
bool CKeyValueStorage::GetRecoverSectors(int32_t sectors[2])
{
  sectors[0] = KEY_VALUE_PAIRS_SECTOR_INVALID;
  sectors[1] = KEY_VALUE_PAIRS_SECTOR_INVALID;

  uint32_t* sector1ptr = (uint32_t*)FlashSectorAddress(m_sector1);
  uint32_t* sector2ptr = (uint32_t*)FlashSectorAddress(m_sector2);

  if (*sector1ptr == SECTOR_ERASED && *sector2ptr == SECTOR_ERASED)
    return false; //there is nothing that can be recovered

  //the two sectors should be checked in order of
  //sequence number, highest first, and an erased sector should not be checked
  if (*sector1ptr == SECTOR_ERASED)
  {
    sectors[0] = m_sector2;
  }
  else if (*sector2ptr == SECTOR_ERASED)
  {
    sectors[0] = m_sector1;
  }
  else if (*sector1ptr > *sector2ptr)
  {
    sectors[0] = m_sector1;
    sectors[1] = m_sector2;
  }
  else if (*sector2ptr > *sector1ptr)
  {
    sectors[0] = m_sector2;
    sectors[1] = m_sector1;
  }

  return true;
}

/*! Copies key-value pairs from one sector to another, with a sequence number that is one higher.
    A new hash is generated over the new data and written to the end of the sector.
    The source sector is erased after copying.
    \param source The source sector to copy data from.
    \param size The size of the key-value data to copy, this should not include the 0xFF CBOR break byte.
    \return true if copying the data succeeded.
    \return false if copying the data failed.*/
bool CKeyValueStorage::CopySectorData(int32_t source, uint32_t size)
{
  uint8_t* sourceaddress = (uint8_t*)FlashSectorAddress(source);

  //the target is the other sector
  uint32_t target = (source == m_sector1) ? m_sector2 : m_sector1;
  uint32_t targetaddress = FlashSectorAddress(target);
  uint32_t targetsize = FlashSectorSize(target);

  PrintLog(m_loglevel, "Copying over key-value pairs from sector %"
           PRIu32 " to sector %" PRIu32 "\n", source, target);

  //erase the target sector so that it can hold new key-value pairs
  if (!FlashEraseSector(target))
    return false;

  //write a new sequence number to the target sector
  uint32_t sequencenr = *(uint32_t*)sourceaddress + 1;
  if (!FlashWrite(targetaddress, (uint8_t*)&sequencenr, sizeof(sequencenr)))
    return false;

  //copy over the existing key-value pairs to the target sector
  if (!FlashWrite(targetaddress + sizeof(uint32_t), sourceaddress + sizeof(uint32_t), size - sizeof(uint32_t)))
    return false;

  //create a new hash from the key-value pairs that were just stored
  SKeyValueHash hash;
  CreateHash((uint8_t*)targetaddress, size, &hash);

  //write the hash to the target sector
  if (!FlashWrite(targetaddress + targetsize - sizeof(hash), (uint8_t*)&hash, sizeof(hash)))
    return false;

  //erase the old key-value pairs
  if (!FlashEraseSector(source))
    return false;

  PrintLog(m_loglevel, "Done copying over key-value pairs from sector %"
           PRIu32 " to sector %" PRIu32 "\n", source, target);

  return true;
}

/*! Finds the most recent valid stored key-value pairs, by checking the stored hashes.
    It will then copy those key-value pairs to the other sector, with a higher sequence number.
    This will recover key-value pairs in case storing them was interrupted previously.
    \return true if recovering key-value pairs succeeded.
    \return false if recovering key-value pairs failed.*/
bool CKeyValueStorage::RecoverStoredKeyValuePairs()
{
  int32_t sectors[2];
  if (!GetRecoverSectors(sectors))
    return false; //all sectors are erased, can't recover

  PrintLog(m_loglevel, "Trying to recover previously stored key-value pairs\n");

  for (uint32_t i = 0; i < 2; i++)
  {
    if (sectors[i] == KEY_VALUE_PAIRS_SECTOR_INVALID)
      continue;

    uint8_t* address      = (uint8_t*)FlashSectorAddress(sectors[i]);
    uint32_t size         = FlashSectorSize(sectors[i]);
    SKeyValueHash* hashptr = ((SKeyValueHash*)(address + size)) - 1;

    //find the erased block of size sizeof(SKeyValueHash) that should be before
    //all the stored hashes, it should be written after the sequence number and CBOR
    while (((uint8_t*)hashptr >= (address + sizeof(uint32_t) + 1)) && !hashptr->IsErased())
      hashptr--;

    //sanity check, hashptr should now point to an erased hash block
    if ((uint8_t*)hashptr < (address + sizeof(uint32_t) + 1) || !hashptr->IsErased())
    {
      PrintLog(m_loglevel, "ERROR: did not find erased hash block in sector %" PRIu32 "\n", sectors[i]);
      continue;
    }

    //check all stored hashes, find the latest stored hash that is still valid
    for(;;)
    {
      //increment the pointer to find the next stored hash
      //if the pointer now points to address + size, all hashes have been checked
      hashptr++;
      if ((uint8_t*)hashptr >= address + size)
      {
        PrintLog(m_loglevel, "ERROR: sector %" PRIu32 " has no valid hash stored\n", sectors[i]);
        break;
      }

      //the hash is created over the data that is stored before the hash in the sector,
      //check that the size field in the hash block is set to a correct value
      uint32_t hashpos = (uintptr_t)hashptr - (uintptr_t)address;
      if ((hashptr->size < sizeof(uint32_t) + 1) || (hashptr->size > hashpos))
      {
        PrintLog(m_loglevel, "ERROR: sector %" PRIu32 ", hash block at position %" PRIu32
                 " has invalid size %" PRIu32 " stored\n", sectors[i], hashpos, hashptr->size);
        continue;
      }

      //create a new hash over the stored data
      SKeyValueHash hash;
      CreateHash(address, hashptr->size, &hash);

      //check if the generated hash matches the hash in the hash block
      if (std::equal(hash.hash, hash.hash + KEY_VALUE_HASH_SIZE, hashptr->hash))
      {
        PrintLog(m_loglevel, "Sector %" PRIu32 " hash valid hash stored at position %"
                 PRIu32 " with size %" PRIu32 "\n", sectors[i], hashpos, hashptr->size);

        //if the hash is correct, then the sector has valid key-value pairs stored, 
        //copy over the key-value pairs to the other sector
        return CopySectorData(sectors[i], hashptr->size);
      }
      else
      {
        PrintLog(m_loglevel, "ERROR: Sector %" PRIu32 " hash invalid hash stored at position %"
                 PRIu32 " with size %" PRIu32 "\n", sectors[i], hashpos, hashptr->size);
      }
    }
  }

  PrintLog(m_loglevel, "Unable to recover previously stored key-value pairs\n");

  return false;
}

/*! Erases the two flash sectors reserved for key-value pairs,
    and initializes the first sector to hold key-value pairs.
    \return true if erasing and initializing the two sectors succeeded.
    \return false on any error.*/
bool CKeyValueStorage::InitStoredKeyValuePairs()
{
  //erase the two sectors if no valid key-value pairs are found
  PrintLog(m_loglevel, "No valid stored key-value pairs found,"
           " resetting to factory defaults\n");

  if (!FlashEraseSector(m_sector1))
    return false;
  if (!FlashEraseSector(m_sector2))
    return false;

  //initialize the first sector to hold a CBOR map
  PrintLog(m_loglevel, "Initializing sector %" PRIi32 " to hold key-value pairs\n",
           m_sector1);
  if (!InitSector(m_sector1, 0))
  {
    PrintLog(m_loglevel, "ERROR: unable to initialize sector %" PRIu32
             " to hold key-value pairs, disabling custom key-value pairs\n", m_sector);
    return false;
  }

  //create a hash over the sequence number and CBOR map stored in the flash,
  //don't include the 0xFF break byte, because it will be overwritten when
  //new key-value pairs are stored
  uint32_t address = FlashSectorAddress(m_sector1);
  uint32_t size = sizeof(uint32_t) + 1;
  SKeyValueHash hash;
  CreateHash((uint8_t*)address, size, &hash);

  //write the hash and size to the end of the flash sector
  uint32_t hashposition = address + FlashSectorSize(m_sector1) - sizeof(hash);
  if (!FlashWrite(hashposition, (uint8_t*)&hash, sizeof(hash)))
    return false;

  return true;
}

/*! Erases and initializes an internal flash sector as a CBOR map.
    \param sector The sector to initialize.
    \param sequencenr The sequence number to write to the sector.
    \return true if initializing the sector succeeded.
    \return false if initializing the sector failed.*/
bool CKeyValueStorage::InitSector(int32_t sector, uint32_t sequencenr)
{
  assert(sector != KEY_VALUE_PAIRS_SECTOR_INVALID);

  if (!FlashEraseSector(sector))
    return false;

  //create a byte with the major type set to map, and the additional
  //type value set to indefinite length, this creates a CBOR map
  //that ends at a break character, which is 0xFF
  //this way, key-value pairs can be written to the flash sector
  //for as long as there is at least one erased byte to end with
  uint8_t  map = CborMapType | 0x1f;
  
  uint32_t address = FlashSectorAddress(sector);

  //write the header byte to position 4, which is where the CBOR map starts
  if (!FlashWrite(address + sizeof(uint32_t), &map, sizeof(map)))
    return false;

  //write the sequence number to the flash, to indicate that
  //this sector holds the most recent CBOR
  if (!FlashWrite(address, (uint8_t*)&sequencenr, sizeof(sequencenr)))
    return false;

  return true;
}

/*! \param key The key of the key-value pair.
    \return The stored value of this key pair.
    \return The factory default, if no value is stored for this key.*/
bool CKeyValueStorage::GetBoolean(uint32_t key)
{
  const SKeyValueInfo& keyvalueinfo = KeyInfo(key);
  assert(keyvalueinfo.type == CborBooleanType);

  CLock lock(m_mutex);

  //only try to read the stored CBOR if the parser has been initialized
  if (m_sector != KEY_VALUE_PAIRS_SECTOR_INVALID)
  {
    bool value;
    if (CBORFindIntMapBool(m_loglevel, "key-value pairs flash root", &m_root, key, &value,
                           CBOROption_PrintNotFoundOff | CBOROption_FindLast))
      return value;
  }

  //if the key is not stored in the flash, return the default value
  return keyvalueinfo.defval.b;
}

/*! \param key The key of the key-value pair.
    \return The stored value of this key.
    \return The factory default, if no value is stored for this key.*/
int64_t CKeyValueStorage::GetInteger(uint32_t key)
{
  const SKeyValueInfo& keyvalueinfo = KeyInfo(key);
  assert(keyvalueinfo.type == CborIntegerType);

  CLock lock(m_mutex);

  //only try to read the stored CBOR if the parser has been initialized
  if (m_sector != KEY_VALUE_PAIRS_SECTOR_INVALID)
  {
    int64_t value;
    if (CBORFindIntMapInt(m_loglevel, "key-value pairs flash root", &m_root, key, &value,
                          CBOROption_PrintNotFoundOff | CBOROption_FindLast))
      return value;
  }

  //if the key is not stored in the flash, return the default value
  return keyvalueinfo.defval.i;
}

/*! \param key The key of the key-value pair.
    \param buf The buffer in which the key's value is returned.
    \param bufsize The size of the buffer that buf points to.*/
void CKeyValueStorage::GetString(uint32_t key, char* buf, uint32_t bufsize)
{
  const SKeyValueInfo& keyvalueinfo = KeyInfo(key);
  assert(keyvalueinfo.type == CborTextStringType);

  CLock lock(m_mutex);

  //only try to read the stored CBOR if the parser has been initialized
  if (m_sector != KEY_VALUE_PAIRS_SECTOR_INVALID)
  {
    CborValue value;
    if (CBORFindIntMapString(m_loglevel, "key-value pairs flash root", &m_root, key, &value,
                             CBOROption_PrintNotFoundOff | CBOROption_FindLast))
    {
      uint32_t length = bufsize;
      if (CBORCopyString(m_loglevel, keyvalueinfo.name, &value, buf, &length))
      {
        buf[bufsize - 1] = 0;
        return;
      }
    }
  }
   
  //if the key is not stored in the flash, copy the default value
  strncpy(buf, keyvalueinfo.defval.s, bufsize);
  buf[bufsize - 1] = 0;
}  

/*! \param key The key of the key-value pair.
    \return The stored value of this key.
    \return The factory default, if no value is stored for this key.*/
float CKeyValueStorage::GetFloat(uint32_t key)
{
  const SKeyValueInfo& keyvalueinfo = KeyInfo(key);
  assert(keyvalueinfo.type == CborFloatType);

  CLock lock(m_mutex);

  //only try to read the stored CBOR if the parser has been initialized
  if (m_sector != KEY_VALUE_PAIRS_SECTOR_INVALID)
  {
    float value;
    if (CBORFindIntMapFloat(m_loglevel, "key-value pairs flash root", &m_root, key, &value,
                            CBOROption_PrintNotFoundOff | CBOROption_FindLast))
      return value;
  }

  //if the key is not stored in the flash, return the default value
  return keyvalueinfo.defval.f;
}

/*! Checks if the new value is different from the current value,
    if so it will store the new value into the internal flash.
    \param The key of the key-value pair.
    \param value The new value of the key-value pair.
    \return true if the key-value pair has been written.
    \return false if the key-value pair has not been written,
            because it did not change, or because something went wrong.*/
bool CKeyValueStorage::SetBoolean(uint32_t key, bool value)
{  
  const SKeyValueInfo& keyvalueinfo = KeyInfo(key);
  assert(keyvalueinfo.type == CborBooleanType);

  CLock lock(m_mutex);

  //only try to write the new key-value pair if the CBOR parser has been initialized
  if (m_sector == KEY_VALUE_PAIRS_SECTOR_INVALID)
    return false;

  //check if the new value is different from the old value
  bool oldvalue = GetBoolean(key);
  if (oldvalue != value)
  {
    if (!m_inswitch)
    { //the value only actually changed when not switching sectors
      PrintLog(m_loglevel, "Boolean %s(%s) changed from %s to %s\n",
               keyvalueinfo.name, keyvalueinfo.ServerField(),
               oldvalue ? "true" : "false", value ? "true" : "false");
    }

    //encode the new value into key-value stored as CBOR
    uint8_t* encodebuf = EncodeBuf();
    CborEncoder encoder;
    cbor_encoder_init(&encoder, encodebuf, KEY_VALUE_ENCODE_BUF_SIZE, 0);
    cbor_encode_int(&encoder, key);
    cbor_encode_boolean(&encoder, value);
    size_t cborsize = cbor_encoder_get_buffer_size(&encoder, encodebuf);

    //write the CBOR to the internal flash
    if (!WriteCBOR(encodebuf, cborsize))
    {
      PrintLog(m_loglevel, "ERROR: unable to store boolean %s(%s)\n",
               keyvalueinfo.name, keyvalueinfo.ServerField());
      return false;
    }

    return true;
  }

  return false;
}

/*! Checks if the new value is different from the current value,
    if so it will store the new value into the internal flash.
    \param The key of the key-value pair.
    \param value The new value of the key-value pair.
    \return true if the key-value pair has been written.
    \return false if the key-value pair has not been written,
            because it did not change, or because something went wrong.*/
bool CKeyValueStorage::SetInteger(uint32_t key, int64_t value)
{
  const SKeyValueInfo& keyvalueinfo = KeyInfo(key);
  assert(keyvalueinfo.type == CborIntegerType);

  CLock lock(m_mutex);

  //only try to write the new key-value pair if the CBOR parser has been initialized
  if (m_sector == KEY_VALUE_PAIRS_SECTOR_INVALID)
    return false;

  //check if the new value is different from the old value
  int64_t oldvalue = GetInteger(key);
  if (oldvalue != value)
  {
    if (!m_inswitch)
    { //the value only actually changed when not switching sectors
      PrintLog(m_loglevel, "Integer %s(%s) changed from %" PRIi64 " to %" PRIi64 "\n",
               keyvalueinfo.name, keyvalueinfo.ServerField(), oldvalue, value);
    }

    //encode the new value into key-value stored as CBOR
    uint8_t* encodebuf = EncodeBuf();
    CborEncoder encoder;
    cbor_encoder_init(&encoder, encodebuf, KEY_VALUE_ENCODE_BUF_SIZE, 0);
    cbor_encode_int(&encoder, key);
    cbor_encode_int(&encoder, value);
    size_t cborsize = cbor_encoder_get_buffer_size(&encoder, encodebuf);

    //write the CBOR to the internal flash
    if (!WriteCBOR(encodebuf, cborsize))
    {
      PrintLog(m_loglevel, "ERROR: unable to store integer %s(%s)\n",
               keyvalueinfo.name, keyvalueinfo.ServerField());
      return false;
    }

    return true;
  }

  return false;
}

/*! Checks if the new value is different from the current value,
    if so it will store the new value into the internal flash.
    \param The key of the key-value pair.
    \param value The new value of the key-value pair.
    \return true if the key-value pair has been written.
    \return false if the key-value pair has not been written,
            because it did not change, or because something went wrong.*/
bool CKeyValueStorage::SetString(uint32_t key, const char* value)
{
  const SKeyValueInfo& keyvalueinfo = KeyInfo(key);
  assert(keyvalueinfo.type == CborTextStringType);

  CLock lock(m_mutex);

  //only try to write the new key-value pair if the CBOR parser has been initialized
  if (m_sector == KEY_VALUE_PAIRS_SECTOR_INVALID)
    return false;

  //check if the new value is different from the old value
  //can't use m_stringbuf for this, it might hold the new value
  GetString(key, m_oldstrvalue, STRING_VALUE_MAX_LENGTH);
  if (strncmp(m_oldstrvalue, value, STRING_VALUE_MAX_LENGTH) != 0)
  {
    if (!m_inswitch)
    { //the value only actually changed when not switching sectors
      PrintLog(m_loglevel, "String %s(%s) changed from \"%s\" to \"%s\"\n",
               keyvalueinfo.name, keyvalueinfo.ServerField(),
               m_oldstrvalue, value);
    }

    //encode the new value into key-value stored as CBOR
    uint8_t* encodebuf = EncodeBuf();
    CborEncoder encoder;
    cbor_encoder_init(&encoder, encodebuf, KEY_VALUE_ENCODE_BUF_SIZE, 0);
    cbor_encode_int(&encoder, key);
    cbor_encode_text_stringz(&encoder, value);
    size_t cborsize = cbor_encoder_get_buffer_size(&encoder, encodebuf);

    //write the CBOR to the internal flash
    if (!WriteCBOR(encodebuf, cborsize))
    {
      PrintLog(m_loglevel, "ERROR: unable to store string %s(%s)\n",
               keyvalueinfo.name, keyvalueinfo.ServerField());
      return false;
    }

    return true;
  }

  return false;
}

/*! Checks if the new value is different from the current value,
    if so it will store the new value into the internal flash.
    \param key The key of the key-value pair.
    \param value The new value of the key-value pair.
    \return true if the key-value pair has been written.
    \return false if the key-value pair has not been written,
            because it did not change, or because something went wrong.*/
bool CKeyValueStorage::SetFloat(uint32_t key, float value)
{
  const SKeyValueInfo& keyvalueinfo = KeyInfo(key);
  assert(keyvalueinfo.type == CborFloatType);

  CLock lock(m_mutex);

  //only try to write the new key-value pair if the CBOR parser has been initialized
  if (m_sector == KEY_VALUE_PAIRS_SECTOR_INVALID)
    return false;

  //check if the new value is different from the old value
  float oldvalue = GetFloat(key);
  if (oldvalue != value)
  {
    if (!m_inswitch)
    { //the value only actually changed when not switching sectors
      PrintLog(m_loglevel, "Float %s(%s) changed from %g to %g\n",
               keyvalueinfo.name, keyvalueinfo.ServerField(), oldvalue, value);
    }

    //encode the new value into key-value stored as CBOR
    uint8_t* encodebuf = EncodeBuf();
    CborEncoder encoder;
    cbor_encoder_init(&encoder, encodebuf, KEY_VALUE_ENCODE_BUF_SIZE, 0);
    cbor_encode_int(&encoder, key);
    cbor_encode_float(&encoder, value);
    size_t cborsize = cbor_encoder_get_buffer_size(&encoder, encodebuf);

    //write the CBOR to the internal flash
    if (!WriteCBOR(encodebuf, cborsize))
    {
      PrintLog(m_loglevel, "ERROR: unable to store float %s(%s)\n",
               keyvalueinfo.name, keyvalueinfo.ServerField());
      return false;
    }

    return true;
  }

  return false;
}

/*! Gets each stored value, converts it to a CBOR format that the server wants,
    and adds it to the root map.
    \param root The map to add the CBOR key-value pairs to.*/
void CKeyValueStorage::ToServerCBOR(CborEncoder* root)
{
  CLock lock(m_mutex);

  for (uint32_t i = 0; i < m_numkeyvalues; i++)
  {
    const SKeyValueInfo& keyvalueinfo = m_keyvalueinfo[i];

    //if serverfield is nullptr, this key-value pair should not be sent to the server
    if (keyvalueinfo.serverfield == nullptr)
      continue;

    //create key-value pairs, where the key is how the server describes the value
    cbor_encode_text_stringz(root, keyvalueinfo.serverfield);

    //add the value of the key-value pair, depending on what type it is
    if (keyvalueinfo.type == CborIntegerType)
    {
      cbor_encode_int(root, GetInteger(keyvalueinfo.key));
    }
    else if (keyvalueinfo.type == CborTextStringType)
    {
      GetString(keyvalueinfo.key, m_stringbuf, STRING_VALUE_MAX_LENGTH);
      cbor_encode_text_stringz(root, m_stringbuf);
    }
    else if (keyvalueinfo.type == CborBooleanType)
    {
      cbor_encode_boolean(root, GetBoolean(keyvalueinfo.key));
    }
    else if (keyvalueinfo.type == CborFloatType)
    {
      cbor_encode_float(root, GetFloat(keyvalueinfo.key));
    }
    else
    {
      assert(0);
    }
  }
}

/*! Accepts CBOR received from the server, and stores any values that have changed.
    \param root Pointer to the root map that holds the key-value pairs.*/
void CKeyValueStorage::FromServerCBOR(CborValue* root)
{
  PrintLog(m_loglevel, "Applying new values from server\n");
  
  CLock lock(m_mutex);

  //disable writing hashes until the loop is done
  //this prevents writing a new hash for every value that has changed,
  //instead, a new hash will be written after the loop is done
  m_writehash = false;
  bool pairwritten = false;

  //iterate over all existing key-value pairs, check if they're present
  //in the CBOR received from the server
  for (uint32_t i = 0; i < m_numkeyvalues; i++)
  {
    const SKeyValueInfo& keyvalueinfo = m_keyvalueinfo[i];

    //if serverfield is nullptr, this key-value pair is not sent from the server
    if (keyvalueinfo.serverfield == nullptr)
      continue;

    CborValue it;
    if (!CBORFindMapValue(m_loglevel, "root", root, keyvalueinfo.serverfield,
                          &it, CBOROption_PrintNotFoundOff | CBOROption_SkipTag))
    {
      PrintLog(m_loglevel, "WARNING: key %s(%s) is not present\n",
               keyvalueinfo.name, keyvalueinfo.ServerField());
      continue;
    }

    //if the key-value pair is present, store it
    if (StoreKeyValuePair(keyvalueinfo, &it, "the server"))
      pairwritten = true;
  }

  //write a new hash if a value has changed
  m_writehash = true;
  if (pairwritten)
    WriteHash();
}

/*! Checks if the type of the key-value pair matches the type of the CBOR value field,
    if it's correct, the correct function is called to store the key-value pair.
    \param keyinfo Information about the key-value pair that should be stored.
    \param it Pointer to the value received from the server.
    \param source Name of the source, used for logging.
    \return true if the key-value pair has been written.
    \return false if the key-value pair has not been written,
            because it did not change, or because something went wrong.*/
bool CKeyValueStorage::StoreKeyValuePair(const SKeyValueInfo& keyinfo, CborValue* it, const char* source)
{
  CborError retval = CborUnknownError;

  if (keyinfo.type == CborIntegerType)
  {
    //check if the CBOR value received from the server is an integer
    if (!cbor_value_is_integer(it))
    {
      CLogLock loglock;
      PrintLog(m_loglevel, "ERROR: %s(%s) is an integer, but did not receive an integer from %s: ",
               keyinfo.name, keyinfo.ServerField(), source);
      cbor_value_to_pretty(stdout, it);
      PrintLog(m_loglevel, "\n");
      return false;
    }

    //convert the CBOR integer to int64_t
    int64_t intval;
    retval = cbor_value_get_int64(it, &intval);
    if (retval != CborNoError)
    {
      PrintLog(m_loglevel, "ERROR: cbor_value_get_int64 returned %i:%s on %s:(%s)\n",
               (int)retval, cbor_error_string(retval), keyinfo.name, keyinfo.ServerField());
      return false;
    }

    //store the new key-value pair, if it has changed from the current value
    return SetInteger(keyinfo.key, intval);
  }
  else if (keyinfo.type == CborTextStringType)
  {
    //check if the CBOR value received from the server is a string
    if (!cbor_value_is_text_string(it))
    {
      CLogLock loglock;
      PrintLog(m_loglevel, "ERROR: %s(%s) is a string, but did not receive a string from %s: ",
               keyinfo.name, keyinfo.ServerField(), source);
      cbor_value_to_pretty(stdout, it);
      PrintLog(m_loglevel, "\n");
      return false;
    }

    //copy the string setting to a buffer
    uint32_t length = STRING_VALUE_MAX_LENGTH;
    if (!CBORCopyString(m_loglevel, keyinfo.name, it, m_stringbuf, &length))
    {
      PrintLog(m_loglevel, "ERROR: unable to copy %s(%s) to buffer of size %i\n",
               keyinfo.name, keyinfo.ServerField(), STRING_VALUE_MAX_LENGTH);
      return false;
    }

    //store the new key-value pair, if it has changed from the current value
    return SetString(keyinfo.key, m_stringbuf);
  }
  else if (keyinfo.type == CborBooleanType)
  {
    //check if the CBOR value received from the server is a boolean
    if (!cbor_value_is_boolean(it))
    {
      CLogLock loglock;
      PrintLog(m_loglevel, "ERROR: %s(%s) is a boolean, but did not receive a boolean from %s: ",
               keyinfo.name, keyinfo.ServerField(), source);
      cbor_value_to_pretty(stdout, it);
      PrintLog(m_loglevel, "\n");
      return false;
    }

    //convert the CBOR bool to C++ bool
    bool boolval;
    retval = cbor_value_get_boolean(it, &boolval);
    if (retval != CborNoError)
    {
      PrintLog(m_loglevel, "ERROR: cbor_value_get_boolean returned %i:%s on %s:(%s)\n",
               (int)retval, cbor_error_string(retval), keyinfo.name, keyinfo.ServerField());
      return false;
    }

    //store the new key-value pair, if it has changed from the current value
    return SetBoolean(keyinfo.key, boolval);
  }
  else if (keyinfo.type == CborFloatType)
  {
    //CBOR supports 16, 32 and 64 bit floats,
    //convert what the server sends to a 32 bit float.
    float floatval = 0.0f;
    if (cbor_value_is_float(it))
    {
      retval = cbor_value_get_float(it, &floatval);
    }
    else if (cbor_value_is_double(it))
    {
      double doubleval = 0.0;
      retval = cbor_value_get_double(it, &doubleval);
      floatval = doubleval;
    }
    else if (cbor_value_is_half_float(it))
    {
      //Store the bits of the half precision float value in a uint16_t.
      uint16_t halffloatval = 0;
      retval = cbor_value_get_half_float(it, &halffloatval);

      //Convert the sign bit from a half precision float to a single precision float.
      uint32_t sign = ((uint32_t)halffloatval & 0x8000) << 16;

      //Extract the 5 bit exponent with offset 15 of the half precision float,
      //and convert it to an 8 bit exponent with offset 127 for a single precision float.
      uint32_t exponent = ((((uint32_t)halffloatval >> 10) & 0x1F) + (127 - 15)) << 23;

      //Extract the 10 fraction bits from the half precision float,
      //and store them in the 10 most significant bits of
      //the 23 bit fraction of the single precision float.
      uint32_t fraction = ((uint32_t)halffloatval & 0x3FF) << 13;

      //Assemble the bits as a full single precision float.
      uint32_t floatbits = sign | exponent | fraction;
      memcpy(&floatval, &floatbits, sizeof(floatval));
    }
    else
    {
      CLogLock loglock;
      PrintLog(m_loglevel, "ERROR: %s(%s) is a float, but did not receive a float from %s: ",
               keyinfo.name, keyinfo.ServerField(), source);
      cbor_value_to_pretty(stdout, it);
      PrintLog(m_loglevel, "\n");
      return false;
    }

    if (retval != CborNoError)
    {
      PrintLog(m_loglevel, "ERROR: unable to get float value, error %i:%s on %s:(%s)\n",
               (int)retval, cbor_error_string(retval), keyinfo.name, keyinfo.ServerField());
      return false;
    }

    //store the new key-value pair, if it has changed from the current value
    return SetFloat(keyinfo.key, floatval);
  }
  else
  {
    assert(0);
    return false;
  }
}

/*! Checks if a number of CBOR bytes can be written to the flash,
    while leaving one 0xFF byte at the end.
    \param size The number of CBOR bytes to write.
    \return true if the number of CBOR bytes can be written.
    \return false if the number of CBOR bytes can't be written.*/
bool CKeyValueStorage::CanWrite(uint32_t size)
{
  CLock lock(m_mutex);

  if (m_sector == KEY_VALUE_PAIRS_SECTOR_INVALID)
    return false;

  //check if there is enough free space to write the data, a hash block, and leave one section
  //of size sizeof(SKeyValueHash) erased, that section acts as both a
  //CBOR break byte (0xFF), and indicates where the last hash is stored
  uint32_t freespace = (uintptr_t)m_hashptr - (uintptr_t)m_addr - m_writepos;
  if (size + (sizeof(SKeyValueHash) * 2) >= freespace)
    return false;

  //check if all bytes needed are currently set to 0xFF, meaning that they're erased
  uint8_t* flashptr = m_addr + m_writepos;
  uint8_t* end = flashptr + size + 1; //+1 to include the CBOR break byte
  while (flashptr != end)
  {
    if (*(flashptr++) != 0xFF)
      return false;
  }

  return true;
}

/*! Writes a buffer filled with CBOR to the flash, at the current write position.
    \param buf Pointer to the buffer to write to the flash.
    \param size Size of the buffer to write to the flash.
    \return true if writing the buffer succeeded, false if writing the buffer failed.*/
bool CKeyValueStorage::WriteCBOR(uint8_t* buf, uint32_t size)
{
  CLock lock(m_mutex);

  if (m_sector == KEY_VALUE_PAIRS_SECTOR_INVALID)
    return false;

  //check if there is enough free space available to write the CBOR,
  //switch sector if there is not enough free space
  if (!CanWrite(size))
  {
    if (!SwitchSector())
    {
      PrintLog(m_loglevel, "ERROR: unable to switch sector\n");
      ResetState(); //reset the state variables, they're invalid now
      return false;
    }

    if (!CanWrite(size))
    {
      PrintLog(m_loglevel, "ERROR: no space available to write %" PRIu32
               " CBOR bytes, even after switching sector\n", size);
      return false;
    }
  }

  if (!FlashWrite((uint32_t)(m_addr + m_writepos), buf, size))
    return false;

  m_writepos += size;

  if (!WriteHash())
    return false;

  return true;
}

/*! Creates a hash of all stored key-value pairs, excluding the break byte that ends the map,
    and writes it in front of all other stored hashes.
    \return true if writing the hash succeeded.
    \return false on any error.*/
bool CKeyValueStorage::WriteHash()
{
  CLock lock(m_mutex);

  if (m_sector == KEY_VALUE_PAIRS_SECTOR_INVALID)
    return false;

  if (!m_writehash)
    return true;

  //create a hash of all key-value pair bytes
  SKeyValueHash hash;
  CreateHash(m_addr, m_writepos, &hash);

  //the new hash should be written before the old hash
  m_hashptr--;

  PrintLog(m_loglevel, "Writing hash at position %" PRIuPTR " of sector %" PRIu32 "\n",
           (uintptr_t)m_hashptr - (uintptr_t)m_addr, m_sector);

  //write the new hash
  if (!FlashWrite((uint32_t)m_hashptr, (uint8_t*)&hash, sizeof(hash)))
    return false;

  return true;
}

/*! Calls CopySectorData() to switch key-value pairs from one sector to the other.
    Should be called when the currently used sector has run out of space.
    Because this calls the regular key-value pair store fuctions, it's protected against
    begin called recursively, although that should never happen.
    \return true if switching the sector succeeded.
    \return false on any error.*/
bool CKeyValueStorage::SwitchSector()
{
  CLock lock(m_mutex);

  if (m_inswitch)
  {
    PrintLog(m_loglevel, "ERROR: SwitchSector() called recursively\n");
    return false;
  }

  m_inswitch = true;
  bool success = CopySectorKeyValuePairs();
  m_inswitch = false;

  return success;
}

/*! Copies key-value pairs from the current sector to the other sector.
    It will only store key-value pairs that are different from the factory default,
    and it will store each key-value pair only once. This frees up storage space.
    \return true if copying data to the other sector succeeded.
    \return false if anything failed.*/
bool CKeyValueStorage::CopySectorKeyValuePairs()
{
  uint32_t target;
  if (m_sector == m_sector1)
    target = m_sector2;
  else if (m_sector == m_sector2)
    target = m_sector1;
  else
    assert(0);

  uint32_t freespace = (uintptr_t)m_hashptr - (uintptr_t)m_addr - m_writepos;
  PrintLog(m_loglevel, "Free space on sector %" PRIu32 ": %" PRIu32
           " bytes, moving key-value pairs to sector %" PRIu32 "\n",
           m_sector, freespace, target);

  //initialize the target sector, with a sequence number one higher
  //than the current sector, to indicate that it holds newer key-value pairs
  if (!InitSector(target, (*(uint32_t*)m_addr) + 1))
    return false;

  //switch over to the other sector
  m_sector   = target;
  m_addr     = (uint8_t*)FlashSectorAddress(m_sector);
  m_size     = FlashSectorSize(m_sector);
  m_writepos = sizeof(uint32_t) + 1;

  //initialize so that at the next call of WriteHash(), the first hash is written
  //at the end of the sector
  m_hashptr = (SKeyValueHash*)(m_addr + m_size);

  //save the old parser structs before they're overwritten by InitDecoder()
  CborParser sourceparser = m_parser;
  CborValue  sourceroot   = m_root;
  sourceroot.parser = &sourceparser;

  if (!InitDecoder())
    return false;

  if (!CheckWritePosition())
    return false;

  //disable writing hashes until the loop is done,
  //instead one hash is written after the loop,
  //this prevents writing a hash per key-value pairs
  m_writehash = false;

  //copy the key-value pairs from the old sector to the new sector
  for (uint32_t i = 0; i < m_numkeyvalues; i++)
  {
    const SKeyValueInfo& keyvalueinfo = m_keyvalueinfo[i];
    CborValue value;
    if (!CBORFindIntMapValue(m_loglevel, "source sector", &sourceroot,
                             keyvalueinfo.key, &value,
                             CBOROption_PrintNotFoundOff | CBOROption_FindLast))
      continue;

    StoreKeyValuePair(keyvalueinfo, &value, "the old sector");
  }

  //after all key-value pairs have been copied, write a hash to the sector
  //this indicates that this sector now holds valid key-value pairs
  m_writehash = true;
  if (!WriteHash())
    return false;

  return true;
}

/*! Returns a pointer to a buffer that can be used for encoding CBOR into.
    If a sector switch is being done, a pointer to a different encode buffer is returned,
    because the first one is already in use.
    By making these buffers member variables, putting them on the stack is avoided,
    this is preferred because the stack of FreeRTOS tasks isn't very big.*/
uint8_t* CKeyValueStorage::EncodeBuf()
{
  CLock lock(m_mutex);
  if (!m_inswitch)
    return m_encodebuf[0];
  else
    return m_encodebuf[1];
}

/*! Gets the value of each key, and prints it to the log.
    This takes about 200 milliseconds, so it should only be used when necessary.*/
void CKeyValueStorage::PrintKeyValuePairs()
{
  CLogLock loglock;

  //for all keys, calculate the maximum length of the name, and the server field,
  //this is then used as a minimum field width for PrintLog(), to increase readability
  int namewidth = 0;
  int fieldwidth = 0;
  for (uint32_t i = 0; i < m_numkeyvalues; i++)
  {
    int length = strlen(m_keyvalueinfo[i].name);
    if (length > namewidth)
      namewidth = length;

    length = strlen(m_keyvalueinfo[i].serverfield);
    if (length > fieldwidth)
      fieldwidth = length;
  }

  //print each key-value pair to the log, in the form:
  //<name>(<serverfield): value
  for (uint32_t i = 0; i < m_numkeyvalues; i++)
  {
    const SKeyValueInfo& keyvalueinfo = m_keyvalueinfo[i];
    if (keyvalueinfo.type == CborBooleanType)
    {
      bool value = GetBoolean(keyvalueinfo.key);
      PrintLog(m_loglevel, "%*s(%*s): %s\n", namewidth, keyvalueinfo.name,
               fieldwidth, keyvalueinfo.ServerField(), value ? "true" : "false");
    }
    else if (keyvalueinfo.type == CborIntegerType)
    {
      int64_t value = GetInteger(keyvalueinfo.key);
      PrintLog(m_loglevel, "%*s(%*s): %" PRIi64 "\n", namewidth, keyvalueinfo.name,
               fieldwidth, keyvalueinfo.ServerField(), value);
    }
    else if (keyvalueinfo.type == CborTextStringType)
    {
      GetString(keyvalueinfo.key, m_stringbuf, STRING_VALUE_MAX_LENGTH);
      PrintLog(m_loglevel, "%*s(%*s): \"%s\"\n", namewidth, keyvalueinfo.name,
               fieldwidth, keyvalueinfo.ServerField(), m_stringbuf);
    }
    else if (keyvalueinfo.type == CborFloatType)
    {
      float value = GetFloat(keyvalueinfo.key);
      PrintLog(m_loglevel, "%*s(%*s): %g\n", namewidth, keyvalueinfo.name,
               fieldwidth, keyvalueinfo.ServerField(), value);
    }
    else
    {
      assert(0);
    }
  }
}

/*! Creates a SHA256 over a section of memory, and copies the first part of the hash
    into the hash member of the SKeyValueHash struct. It also sets its size member.
    \param address Pointer to the memory to create a SHA256 over.
    \param size Number of bytes to create a SHA256 over.
    \param hash The first part of the SHA256 hash is returned here.*/
void CKeyValueStorage::CreateHash(uint8_t* address, uint32_t size, SKeyValueHash* hash)
{
  uint8_t hashbuf[SHA256_BLOCK_SIZE];
  SHA256_CTX ctx;
  WSS_sha256_init(&ctx);
  WSS_sha256_update(&ctx, address, size);
  WSS_sha256_final(&ctx, hashbuf);

  hash->size = size;
  memcpy(hash->hash, hashbuf, KEY_VALUE_HASH_SIZE);
}

/*! \return A c-string describing the name of the key.*/
const char* CKeyValueStorage::Name(uint32_t key)
{
  const SKeyValueInfo& keyvalueinfo = KeyInfo(key);
  return keyvalueinfo.name;
}
