#include "linkedlist.h"

#include <cassert>

CListElement::CListElement()
{
  m_prev = nullptr;
  m_next = nullptr;
}

CListElement::~CListElement()
{
  assert(m_prev == nullptr);
  assert(m_next == nullptr);
}

/*! Adds an element to the start of the list, after this list.m_start will point to this element.*/
void CListElement::AddToListAtStart(CList& list)
{
  //make sure this element has not already been added to a list
  assert(m_prev == nullptr);
  assert(m_next == nullptr);

  if (list.m_start == nullptr || list.m_end == nullptr)
  {
    //on an empty list, m_start and m_end should be nullptr
    //and m_size should be 0
    assert(list.m_start == nullptr);
    assert(list.m_end   == nullptr);
    assert(list.m_size  == 0);

    //set the start and end pointers to this element
    list.m_start = this;
    list.m_end   = this;
  }
  else
  {
    assert(list.m_start->m_prev == nullptr);
    assert(list.m_size > 0);

    m_next         = list.m_start; //make the start element the next element
    list.m_start   = this;         //make this element the start element
    m_next->m_prev = this;         //make this element the previous element of the next element
    m_prev         = nullptr;      //make sure this element has no previous element
  }

  list.m_size++;
}

/*! Adds an element to the end of the list, after this list.m_end will point to this element.*/
void CListElement::AddToListAtEnd(CList& list)
{
  //make sure this element has not already been added to a list
  assert(m_prev == nullptr);
  assert(m_next == nullptr);

  if (list.m_start == nullptr || list.m_end == nullptr)
  {
    //on an empty list, m_start and m_end should be nullptr
    //and m_size should be 0
    assert(list.m_start == nullptr);
    assert(list.m_end   == nullptr);
    assert(list.m_size  == 0);

    //set the start and end pointers to this element
    list.m_start = this;
    list.m_end   = this;
  }
  else
  {
    assert(list.m_end->m_next == nullptr);
    assert(list.m_size != 0);

    m_prev         = list.m_end; //make the end element the previous element
    list.m_end     = this;       //make this element the end element
    m_prev->m_next = this;       //make this element the next element of the previous element
    m_next         = nullptr;    //make sure this element has no next element
  }

  list.m_size++;
}

/*! Adds this element before the element passed as argument.*/
void CListElement::AddToListBeforeElement(CList& list, CListElement* element)
{
  //make sure this element is not part of a list
  assert(m_prev == nullptr);
  assert(m_next == nullptr);

  //make sure the element exists
  assert(element);

  //insert this element into the list
  m_next = element;
  m_prev = element->m_prev;

  //if element was the first element on the list, update the pointer
  //to this element
  if (list.m_start == element)
  {
    list.m_start = this;
    assert(element->m_prev == nullptr);
  }

  //adjust pointer on the previous and next element to point to this element
  m_next->m_prev = this;
  if (m_prev)
    m_prev->m_next = this;

  list.m_size++;
}

/*! Adds this element after the element passed as argument.*/
void CListElement::AddToListAfterElement(CList& list, CListElement* element)
{
  //make sure this element is not part of a list
  assert(m_prev == nullptr);
  assert(m_next == nullptr);

  //make sure the element exists
  assert(element);

  //insert this element into the list
  m_next = element->m_next;
  m_prev = element;

  //if element was the last element on the list, update the pointer
  //to this element
  if (list.m_end == element)
  {
    list.m_end = this;
    assert(element->m_next == nullptr);
  }

  //adjust pointer on the previous and next element to point to this element
  if (m_next)
    m_next->m_prev = this;
  m_prev->m_next = this;

  list.m_size++;
}

void CListElement::RemoveFromList(CList& list)
{
  //make sure this element has been added to a list before
  assert(list.m_start != nullptr);
  assert(list.m_end   != nullptr);
  assert(list.m_size > 0);

  if (list.m_start == this)
  {
    //if this element is the first element, make sure it has no previous element
    assert(m_prev == nullptr);

    if (m_next == nullptr)
    {
      //if this element has no next element, make sure that it's the only element
      assert(list.m_end == this);
      assert(list.m_size == 1);

      //make the list empty
      list.m_start = nullptr;
      list.m_end   = nullptr;
    }
    else
    {
      //if this element has a next element, make sure the list has more than one element
      assert(list.m_size > 1);

      //remove this element from the start of the list
      m_next->m_prev = nullptr;
      list.m_start   = m_next;
    }
  }
  else if (list.m_end == this)
  {
    //if this element is at the end of the list, make sure it doesnt' have a next element
    //make sure that it does have a previous element, and that it's not the only element on the list
    //because if it is the only element on the list, then list.m_start should point to this element,
    //in which case the previous branch should have removed this element
    assert(m_next == nullptr);
    assert(m_prev != nullptr);
    assert(list.m_size > 1);

    //remove this element from the end of the list
    m_prev->m_next = nullptr;
    list.m_end     = m_prev;
  }
  else
  {
    //if this is an element somewhere in the middle of the list, make sure
    //that the list has more than 2 elements
    assert(list.m_size > 2);

    //remove this element from the middle of the list
    m_next->m_prev = m_prev;
    m_prev->m_next = m_next;
  }

  m_next = nullptr;
  m_prev = nullptr;
  list.m_size--;
}
