#include "log.h"
#include "lock.h"
#include "clock.h"
#include "usart.h"
#include "spiflashlog.h"

#include <stm32f4xx.h>

#include <FreeRTOS.h>
#include <semphr.h>
#include <queue.h>

#include <cstdio>
#include <cstdarg>
#include <cstring>

/*! Usart class used for debug printing.*/
static CUsart             g_usart;

/*! Created in SetupDebugUsart(), it's locked from PrintLog() and from __io_putchar().
    When using PrintLog(), it will protect individual PrintLog() calls from being interrupted
    by other calls to PrintLog() and printf().*/
static SemaphoreHandle_t  g_usartlock;

/*! Set to true when FreeRTOS is started, signals if data for DEBUGUSART should be written to g_usart,
    or directly to the usart. When data is written to g_usart, it will only be written to DEBUGUSART when FreeRTOS is running. */
static bool               g_asyncusart;

/*! Set from PrintLog() and read from __io_putchar() after each newline is printed,
    it's the log level of the currently printed line.*/
static int                g_loglevel = LOGSTDOUT;

/*! Set from SetApplicationLogLevels() and read from __io_putchar() before the first character of each line.*/
static const char* const* g_apploglevels;

/*! Set from SetApplicationLogLevels() and read from __io_putchar() before the first character of each line.*/
static int                g_appnumloglevels;

/*! Set from PrintLogDecode, will decode unprintable characters into printable words.*/
static bool               g_decodenonprintable;

/*! Initializes DEBUGUSART at 115k2 8n1, which by default is configured on pins PA9 and PA10 in GPIOInit(). */
void SetupDebugUsart()
{
  if (g_usart.IsEnabled())
    return;

  //create a mutex for DEBUGUSART
  g_usartlock = xSemaphoreCreateRecursiveMutex();

  //setup DEBUGUSART as 115k2 8n1
  USART_InitTypeDef usartinit;
  USART_StructInit(&usartinit);
  usartinit.USART_BaudRate = 115200;
  usartinit.USART_Mode = USART_Mode_Tx;
  g_usart.Setup(DEBUGUSART, &usartinit);
}

/*! Switches the debug usart between writing bytes directly
    and writing bytes to g_usart and using the tx interrupt.
    Using g_usart requires a running FreeRTOS kernel,
    it's recommended to call this function from 
    \param async true for async mode, false for direct mode.*/
void SetDebugUsartAsync(bool async)
{
  if (g_usartlock)
  {
    CLock lock(g_usartlock);
    g_asyncusart = async;
  }
  else
  {
    g_asyncusart = async;
  }
}

/*! prints a string directly to DEBUGUSART, used from error interrupts like the hardfault handler. */
void DEBUGUSARTprintf(const char* fmt, ...)
{
  if (!g_usart.IsEnabled())
    return;

  va_list args;
  va_start(args, fmt);
  int bufsize = vsnprintf(NULL, 0, fmt, args);
  va_end(args);

  char buf[bufsize + 1];
  va_list args2;
  va_start(args2, fmt);
  vsnprintf(buf, bufsize + 1, fmt, args2);
  va_end(args2);

  //test bit 7 of the CR1 register for a while, this is the transmitter register empty interrupt bit
  //if this is set, a line from PrintLog() or printf() is probably being sent out the serial port
  //this should not be interrupted
  for (volatile uint32_t i = 0; i < 1000000; i++)
  {
    if (!(DEBUGUSART->CR1 & (1 << 7)))
      break;
  }

  //disable all interrupts so that nothing will interrupt this write to the usart
  __disable_irq();

  //disable the transmitter register empty interrupt
  USART_ITConfig(DEBUGUSART, USART_IT_TXE, DISABLE); 

  char* bufptr = buf;
  while (*bufptr)
  {
    if (*bufptr == '\n') //send a carriage return before a newline character
    {
      while (USART_GetFlagStatus(DEBUGUSART, USART_FLAG_TXE) == RESET); //wait for tranmission register empty
      USART_SendData(DEBUGUSART, '\r');
    }

    while (USART_GetFlagStatus(DEBUGUSART, USART_FLAG_TXE) == RESET); //wait for tranmission register empty
    USART_SendData(DEBUGUSART, *bufptr);
    bufptr++;
  }
  while (USART_GetFlagStatus(DEBUGUSART, USART_FLAG_TC) == RESET); //wait for transmission complete

  //enable interrupts again
  __enable_irq();
}

/*! Prints a line using vprintf, but does it inside a lock of g_usartlock, this prevents multiple calls from interrupting each other's lines.
\param level The log level, as defined in log.h and loglevels.h
\param fmt printf() format string
\param ... printf() variable arguments list */
int PrintLog(int level, const char* fmt, ...)
{
  if (!g_usart.IsEnabled())
    return 0;

  CLock lock(g_usartlock);

  //make sure any write to the stdout and stderr FILE* has completed
  fflush(stdout);
  fflush(stderr);

  int oldloglevel = g_loglevel;
  g_loglevel = level;

  va_list args;
  va_start(args, fmt);
  int retval = vprintf(fmt, args);
  va_end(args);

  fflush(stdout); //make sure all bytes are written before releasing the lock

  g_loglevel = oldloglevel;

  return retval;
}

/*! Like Printlog(), but will set g_decodenonprintable during the print,
    so that non printable ASCII characters will be decoded to printable words.
    The printed line will then always be terminated with a newline.
    \param level The log level, as defined in log.h and loglevels.h
    \param fmt printf() format string
    \param ... printf() variable arguments list */
int PrintLogDecode(int level, const char* fmt, ...)
{
  if (!g_usart.IsEnabled())
    return 0;

  //make sure any write to the stdout FILE* has completed
  fflush(stdout); 

  CLock lock(g_usartlock);

  //enable decoding of non printable characters
  g_decodenonprintable = true;

  g_loglevel = level;

  va_list args;
  va_start(args, fmt);
  int retval = vprintf(fmt, args);
  va_end(args);

  //make sure all bytes are written before changing
  //g_decodenonprintable and releasing the lock
  fflush(stdout); 

  //disable decoding of non printable characters, and print a newline
  g_decodenonprintable = false;
  printf("\n");

  return retval;
}

/*! Prints a buffer as ascii hex.
    \param level The loglevel to use when printing.
    \param buf A pointer to a buffer to print.
    \param size The number of bytes to print.
    \param suffix When not nullptr, string to print after the hex string.*/
void PrintLogHexBuf(int level, const uint8_t* buf, uint32_t size, const char* suffix)
{
  CLogLock loglock;
  for (uint32_t i = 0; i < size; i++)
    PrintLog(level, "%02" PRIX8, buf[i]);

  if (suffix)
    PrintLog(level, "%s", suffix);
}

/*! Writes a character to g_usart when the async usart is enabled,
    writes a character directly to the usart when the async usart is disabled.*/
static void WriteDebugUsart(char ch)
{
  CLock lock(g_usartlock);

  if (g_asyncusart)
  {
    g_usart.Write((uint8_t*)&ch, 1, 1000);
  }
  else
  {
    //wait for tranmission register empty, then write the character to the transmit register
    while (USART_GetFlagStatus(DEBUGUSART, USART_FLAG_TXE) == RESET);
    USART_SendData(DEBUGUSART, ch);
  }
}

/*! Writes a character to the debug usart, and to the log in the spi flash.
    \param ch The character to write.*/
static void WriteDebugChar(char ch)
{
  WriteDebugUsart(ch);
  g_spiflashlog.WriteDebugLogChar(ch);
}

/*! Converts a 64 bit millisecond timestamp to ascii, and prints a decimal separator.
    This is used instead of sprintf, because that uses a rather large amount of stack.
    \param num A 64 bit millisecond timestamp to convert.
    \param str A pointer to a char array where the string is written.
    \return The length of the string written to str, excluding the null byte.*/
static uint32_t MillisToAscii(uint64_t num, char* str)
{
  //convert each decimal to ascii, for at least 5 characters,
  //three are needed for milliseconds, one for the decimal separator, and at least one for seconds
  uint64_t base = 10;
  uint32_t pos = 0;
  while (num || pos < 5)
  {
    str[pos] = (num % base) + '0';
    num /= base;
    pos++;
    if (pos == 3)
    { //print the decimal separator after three characters
      str[pos] = '.';
      pos++;
    }
  }
  str[pos] = 0;

  //the decimals are written in reverse order, so reverse them
  char x;
  for (uint32_t i = 0; i < pos / 2; i++)
  {
    x = str[pos - i - 1];
    str[pos - i - 1] = str[i];
    str[i] = x;
  }

  return pos;
}

/*! Time in microseconds after a newline that an extra newline is printed,
    this makes the debug output more readable.*/
//#define LINEBREAK_TIME 2000000
#define LINEBREAK_TIME 0 //set to a nonzero value to enable extra newlines after a timeout

static void UsartPutChar(int ch)
{
  static int      lastchar;
  static bool     startofline;
  static uint32_t lastnewline;

  //hold a lock so that if a timestamp is printed, it's not interrupted by a printf
  CLock lock(g_usartlock);

  //if the previous character was a newline, write a millisecond timestamp out the serial port
  //and print the log level
  if (startofline)
  {
    //convert a millisecond timestamp to ascii
    int64_t  now = Millis();
    char     timestr[32] = {};
    uint32_t timestrlen = MillisToAscii(now, timestr);

    //left pad the ascii timestamp with spaces if it's too short
    const uint32_t mintimelen = 8;
    if (mintimelen > timestrlen)
    {
      for (uint32_t i = 0; i < mintimelen - timestrlen; i++)
        WriteDebugChar(' ');
    }

    //write the timestamp out the serial port
    for (uint32_t i = 0; i < timestrlen; i++)
      WriteDebugChar(timestr[i]);

    //put a space between the timestamp and the log line
    WriteDebugChar(' ');

    //convert the log level to a string
    const char* loglevelstr = "UNKNOWN";
    static const char* loglevels[] =
    { 
      "STDOUT",
      "MAIN",
      "INFO",
      "ERROR",
      "CLOCK",
      "TASK",
      "ATCMD",
      "APNLIST",
      "SPIFLASH",
      "FWUPDATE",
      "INTFLASH",
      "FLASHLOG",
      "FLASHBUF",
      "MODEM",
      "RFM9xW",
      "EXTRTC",
      "OLEDDPY",
      "IOEXPANDER",
      "BTTRYCHRGR",
      "FILLSENSOR",
      "RTCSCHED",
    };

    //positive log levels are supplied by the application via the SetApplicationLogLevels() function
    //negative log levels are library log levels
    if (g_loglevel >= 0 && g_loglevel < g_appnumloglevels)
      loglevelstr = g_apploglevels[g_loglevel];
    else if (-g_loglevel - 1 < (int32_t)(sizeof(loglevels) / sizeof(loglevels[0])))
      loglevelstr = loglevels[-g_loglevel - 1];
    
    uint32_t loglevelstrlen = strlen(loglevelstr);

    //pad the log level string to size LOGLEVELPADLEN with spaces
    if (loglevelstrlen < LOGLEVELPADLEN)
    {
      for (uint32_t i = 0; i < LOGLEVELPADLEN - loglevelstrlen; i++)
        WriteDebugChar(' ');
    }

    //write the log level string out the serial port
    for (uint32_t i = 0; i < loglevelstrlen; i++)
      WriteDebugChar(loglevelstr[i]);

    WriteDebugChar(':');

    startofline = false;
  }

  //if this character is a newline, and the previous character was not a
  //carriage return, send a carriage return first
  if (ch == '\n')
  {
    if (lastchar != '\r')
      UsartPutChar('\r');

    //signal for the next call to UsartPutChar() that this character is a newline
    //by storing a nonzero microsecond timestamp
#if LINEBREAK_TIME > 0
    lastnewline = Micros();
    if (lastnewline == 0)
      lastnewline = 1;
#endif

    //signal to the next call of this function that this character has ended the line
    startofline = true;
  }
  else if (lastnewline)
  {
    uint32_t now = Micros();
    uint32_t diff = now - lastnewline;
    lastnewline = 0; //prevent infinite recursion

    //if the previous character was a newline, and no character was printed for some time
    //print an extra newline, this increases the readability of the serial output
    if (diff > LINEBREAK_TIME)
      UsartPutChar('\n');

    lastnewline = 0; //signal that the last character was not a newline
  }

  WriteDebugChar(ch);

  lastchar = ch;
}

/*! Converts a 32 bit integer to ascii.
    This is used instead of sprintf, because that uses a rather large amount of stack.
    \param num A 32 bit millisecond timestamp to convert.
    \param str A pointer to a char array where the string is written.
    \return The length of the string written to str, excluding the null byte.*/
static uint32_t IntegerToAscii(uint32_t num, char* str)
{
  if (num == 0)
  {
    str[0] = '0';
    str[1] = 0;
    return 1;
  }
  else
  {
    //convert each decimal to ascii
    uint64_t base = 10;
    uint32_t pos = 0;
    while (num)
    {
      str[pos] = (num % base) + '0';
      num /= base;
      pos++;
    }
    str[pos] = 0;

    //the decimals are written in reverse order, so reverse them
    char x;
    for (uint32_t i = 0; i < pos / 2; i++)
    {
      x = str[pos - i - 1];
      str[pos - i - 1] = str[i];
      str[i] = x;
    }

    return pos;
  }
}

/*! Checks if ch is a printable character, if not, it will print a word describing the character instead.
    \param ch The character that should be printed.*/
static void UsartPutCharDecode(uint8_t ch)
{
  //ASCII control characters 0 to 31
  static const char *controlchars[] =
  {
    "<NUL>",
    "<SOH>",
    "<STX>",
    "<ETX>",
    "<EOT>",
    "<ENQ>",
    "<ACK>",
    "<BEL>",
    "<BS>",
    "<HT>",
    "<LF>",
    "<VT>",
    "<FF>",
    "<CR>",
    "<SO>",
    "<SI>",
    "<DLE>",
    "<DC1>",
    "<DC2>",
    "<DC3>",
    "<DC4>",
    "<NAK>",
    "<SYN>",
    "<ETB>",
    "<CAN>",
    "<EM>",
    "<SUB>",
    "<ESC>",
    "<FS>",
    "<GS>",
    "<RS>",
    "<US>",
  };

  const char* printstr = nullptr;

  char valuestr[32];
  if (ch < 32)
  {
    printstr = controlchars[ch];
  }
  else if (ch == 127)
  {
    printstr = "<DEL>";
  }
  else if (ch > 127)
  { //if the character is not one of the ASCII controll characters,
    //convert it to a number within <>
    valuestr[0] = '<';
    uint32_t valuelen = IntegerToAscii(ch, valuestr + 1);
    valuestr[valuelen + 1] = '>';
    valuestr[valuelen + 2] = 0;
    printstr = valuestr;
  }

  if (printstr)
  {
    const char* printptr = printstr;
    while (*printptr)
      UsartPutChar(*(printptr++));
  }
  else
  {
    UsartPutChar(ch);
  }
}

/*! Called for each character that should be written to stdout. It will write them to g_usart,
    inside a lock held on g_QeueLock, this prevents this function from interrupting lines printed from PrintLog by other tasks.
    \param ch Character that should be written to stdout.*/
extern "C" int __io_putchar(int ch)
{
  if (!g_usart.IsEnabled())
    return ch;

  //if g_decodenonprintable is set, decode any non printable character into a printable word
  if (g_decodenonprintable)
    UsartPutCharDecode(ch);
  else
    UsartPutChar(ch);

  return ch;
}

//! Locks g_usartlock, useful for printing a single line with multiple calls to PrintLog()
void LockLog()
{
  xSemaphoreTakeRecursive(g_usartlock, portMAX_DELAY);
}

void UnlockLog()
{
  xSemaphoreGiveRecursive(g_usartlock);
}

void SetApplicationLogLevels(const char* const* levels, int numlevels)
{
  g_apploglevels    = levels;
  g_appnumloglevels = numlevels;
}

void WaitLogWriteComplete(uint32_t timeout)
{
  CLogLock loglock;
  g_usart.WaitWriteComplete(timeout);
}
