#include "modembg96.h"
#include "log.h"
#include "clock.h"
#include "lock.h"
#include "apnlist.h"
#include "taskhandler.h"

#include <algorithm>
#include <cstring>
#include <cstdio>
#include <cctype>
#include <cassert>

#define XSTR(a) STR(a)
#define STR(a)  #a

CModemBG96::CModemBG96() : m_atcmd(m_usart)
{
  m_usartptr = nullptr;
  m_pwrkey   = {};
  m_status   = {};
  m_reset    = {};
  m_rssi     = -1;

  memset(&m_siminfo, 0, sizeof(m_siminfo));
  memset(m_imei,     0, sizeof(m_imei));
  memset(m_manufact, 0, sizeof(m_manufact));
  memset(m_model,    0, sizeof(m_model));
  memset(m_revision, 0, sizeof(m_revision));
  memset(m_appver,   0, sizeof(m_appver));

  ResetState();
}

CModemBG96::~CModemBG96()
{
}

void CModemBG96::ResetState()
{
  m_type          = ModemType_UNKNOWN;
  m_activated     = false;
  m_needsreset    = false;
  m_statussignal  = 0;
  m_ber           = 0;

  //Clear the SIM info except the ICCID, because the ICCID should be logged
  //if connecting to the network failed.
  memset(m_siminfo.imsi, 0, sizeof(m_siminfo.imsi));
  memset(m_siminfo.spn, 0, sizeof(m_siminfo.spn));

  memset(m_fnn,      0, sizeof(m_fnn));
  memset(m_snn,      0, sizeof(m_snn));
  memset(m_spn,      0, sizeof(m_snn));
  memset(m_rplmn,    0, sizeof(m_rplmn));
  memset(m_lacstr,   0, sizeof(m_lacstr));
  m_lac = 0;
  memset(m_cistr,    0, sizeof(m_cistr));
  m_ci = 0;
  m_needsfwupdate = false;
}

/*! Sets the usart and the pins the modem is connected to.*/
void CModemBG96::SetPeripherals(USART_TypeDef* usartptr, const SGPIOPin pwrkey, const SGPIOPin status, const SGPIOPin reset)
{
  m_usartptr = usartptr;
  m_pwrkey   = pwrkey;
  m_status   = status;
  m_reset    = reset;
}

/*! Sets up the modem, will call Close() if it fails.
    \param factoryreset Perform a factory reset using AT+QPRTPARA.
                        After that the modem will be powered off
                        and the setup will fail.
    \param allowfwupdate When true, allow updating the modem firmware.
    \return true if setting up the modem succeeded, false if it failed.*/
bool CModemBG96::Setup(bool factoryreset /*= false*/, bool allowfwupdate /*= true*/)
{
  PrintLog(LOGMODEM, "Setting up modem\n");

  Close();

  PreSetup();

  //set m_activated to true before setting up the modem,
  //so that if setting up the modem fails, the call to Close()
  //after that will still close and power down the modem
  m_activated = true;

  if (SetupModem(factoryreset, allowfwupdate))
  {
    PrintLog(LOGMODEM, "Modem set up successfully\n");
    return true;
  }
  else
  {
    if (!factoryreset)
      PrintLog(LOGMODEM, "ERROR: Setting up modem failed\n");

    Close();
    return false;
  }

  return true;
}

/*! Called before SetupModem(), resets state that is no longer valid.
    This state is not reset in ResetState(), because it might still be needed
    if setting up an internet connection fails.*/
void CModemBG96::PreSetup()
{
  m_rssi = -1;
  memset(m_imei,     0, sizeof(m_imei));
  memset(m_manufact, 0, sizeof(m_manufact));
  memset(m_model,    0, sizeof(m_model));
  memset(m_revision, 0, sizeof(m_revision));
  memset(m_appver,   0, sizeof(m_appver));
}

/*! Calls all functions needed to set up the modem/
    \param factoryreset Perform a factory reset using AT+QPRTPARA and return false.
    \param allowfwupdate When true, perform a firmware update if needed.
    \return true if setting up the modem succeeded,
            false if any step of setting up the modem failed.*/
bool CModemBG96::SetupModem(bool factoryreset, bool allowfwupdate)
{
  //make sure the modem is powered off, so that the next
  //power on will do a full reset of the modem
  if (!SetPower(false))
    return false;

  //power on the modem
  if (!SetPower(true))
    return false;

  if (!SetupBaudrate())
    return false;

  //wait for any firmware update still in progress
  if (!CheckFirmwareUpdateInProgress())
    return false;

  if (!SetFlowControl())
    return false;

  if (!SetErrorMessageFormat())
    return false;

  if (!GetModemInfo())
    return false;

  if (!FactoryReset(factoryreset))
    return false; //Factory reset happened, modem should be powered off.

  //Don't scan NB-IOT networks, it takes too long.
  SetNetworkScanMode();

  //read the ICCID from the SIM, this needs to be done before entering the PIN,
  //because the ICCID is used to select the PIN
  if (!ReadSIMICCID())
    return false;

  //the SIM PIN should be entered before getting any SIM info,
  //otherwise the modem returns error on any command that requests SIM info
  if (!EnterSIMPIN())
    return false;

  //setting the operator selection can only happen when the SIM is fully functional
  if (!SetOperatorSelection())
    return false;

  if (!ReadSIMIMSI())
    return false;

  //read the service provider name from the sim
  //if this fails it's not a fatal error, it's only needed to select an apn
  //some providers don't write their spn to the sim
  ReadSIMSPN();

  if (!WaitNetworkRegister())
    return false;

  if (!GetSignalQuality())
    return false;

  if (!GetNetworkInfo())
    return false;

  if (!GetNetworkType())
    return false;

  //Still try to set up the connection if the packet service attach fails.
  //It's not known if this failure means that the connection doesn't work.
  WaitPacketServiceAttach();

  if (!ActivatePDPContext())
    return false;

  if (!GetIPAddress())
    return false;

  if (!UpdateFirmware(allowfwupdate))
    return false;

  return true;
}

/*! Shuts down and powers off the modem.
    \return true The modem has been powered off.
    \return false The modem could not be powered off.*/
bool CModemBG96::Close()
{
  if (!m_activated)
    return true;

  PrintLog(LOGMODEM, "Shutting down modem\n");

  //send a power down command
  if (m_usart.IsEnabled())
  {
    //deactivate the PDP context before powering off the modem
    //providers don't like it when you leave the PDP context in the active state
    DeactivatePDPContext();

    //disable phone functionality before powering down
    SetPhoneFunctionality(false);

    //shut down the modem with the AT+QPOWD command
    if (m_atcmd.ATQPOWD(1) != ATCmdResult_Success)
    {
      PrintLog(LOGMODEM, "ERROR: unable to shut down modem:\n");
      m_atcmd.PrintError();
    }
  }

  //make sure the modem is powered off, by waiting for the status pin to rise
  //if it doesn't rise, power off the modem using the PWRKEY pin
  bool poweredoff = SetPower(false, true);

  m_usart.Close();

  ResetState();
  m_atcmd.ResetState();

  return poweredoff;
}

/*! Powers the modem off.*/
void CModemBG96::PowerOff()
{
  SetPower(false);
}

/*! Interrupt handler for the modem status pin, used for detecting power on and power off.*/
void CModemBG96::GPIOIRQHandler(SGPIOPin pin)
{
  //if the FreeRTOS task is blocked on waiting for the semaphore
  //and a lower priority task is currently running, use portYIELD_FROM_ISR
  //to switch to the task waiting on the semaphore
  if (m_statussignal)
  { //signal to the task that the gpio interrupt has happened
    BaseType_t taskwoken = pdFALSE;
    xSemaphoreGiveFromISR(m_statussignal, &taskwoken);
    portYIELD_FROM_ISR(taskwoken);
  }
}

//! Timeout in milliseconds for powering on or off.
#define POWER_TIMEOUT  300000
//! Timeout in milliseconds to wait for the status pin to change.
#define STATUS_TIMEOUT  30000

/*! Tries to power on or off the modem by toggling the PWRKEY pin,
    if the modem is already on when power on is requested,
    or the modem is already off when a power off is requested,
    this function will do nothing.
    \param on true when the modem should be powered on
           false when it should be powered off.
    \param waitfirst when true, wait for the status pin to change first,
           used when sending AT+QPOWD before calling this function.
    \return true if the requested power state was achieved, false if it was not.*/
bool CModemBG96::SetPower(bool on, bool waitfirst /*= false*/)
{
  PrintLog(LOGMODEM, "Powering %s modem\n", on ? "on" : "off");

  //Create a semaphore to signal from the gpio interrupt handler to the current task.
  CCritical criticalsection;
  m_statussignal = xSemaphoreCreateBinary();
  criticalsection.Leave();

  //Add the modem status pin as an edge interrupt pin,
  //from the gpio interrupt, m_statussignal is signaled.
  g_exti.EnableGPIOIrq(m_status, on ? EXTI_Trigger_Falling : EXTI_Trigger_Rising, this);

  bool powered;
  bool triedreset = false;
  int64_t powerstart = Millis();
  for(;;)
  {
    //Check if the modem has the requested power state.
    powered = !GPIO_ReadInput(m_status);
    if (on == powered)
    {
      PrintLog(LOGMODEM, "Modem is powered %s\n", on ? "on" : "off");
      break;
    }

    //Abort if the modem takes too long to respond.
    uint32_t elapsed = Millis() - powerstart;
    if (elapsed > POWER_TIMEOUT)
      break;

    //If half the power timeout elapsed and the modem has still not responded,
    //toggle the reset pin.
    //Delay is chosen to be compatible with Quectel BG96 and BG95.
    if (!triedreset && elapsed > POWER_TIMEOUT / 2)
    {
      PrintLog(LOGMODEM, "Modem is not responding to PWRKEY pin, toggling RESET pin\n");
      GPIO_Set(m_reset);
      DelayMS(2500);
      GPIO_Reset(m_reset);
      DelayMS(30000); //give the modem some time to settle
    }

    if (waitfirst)
    { //If waitfirst is true, for the first iteration,
      //just wait for the status pin to go to the correct level.
      PrintLog(LOGMODEM, "Modem is powered %s, waiting for power %s\n",
               powered ? "on" : "off", powered ? "off" : "on");
      waitfirst = false;
    }
    else
    {
      PrintLog(LOGMODEM, "Modem is powered %s, toggling PWRKEY pin\n",
               powered ? "on" : "off");

      //Toggle the PWRKEY pin to change the modem power state.
      //The delay is chosen to be compatible with the Quectel BG96 and BG95.
      GPIO_Set(m_pwrkey);
      DelayMS(750);
      GPIO_Reset(m_pwrkey);
    }

    //Wait for the modem status pin to change.
    //m_statussignal is signaled by GPIOIRQHandler()
    //on an edge on the modem status pin.
    const char* edgestr = on ? "falling" : "rising";
    PrintLog(LOGMODEM, "Waiting for %s edge on STATUS pin\n", edgestr);
    if (xSemaphoreTake(m_statussignal, pdMS_TO_TICKS(STATUS_TIMEOUT)))
    {
      PrintLog(LOGMODEM, "Detected %s edge on STATUS pin\n", edgestr);

      //After powering on, wait for the modem to boot.
      if (on)
        DelayMS(2000);
    }
    else
    {
      PrintLog(LOGMODEM, "ERROR: Did not detect %s edge on STATUS pin\n", edgestr);
    }
  }
  
  //Remove the interrupt handler on the modem status pin.
  g_exti.DisableGPIOIrq(m_status);

  //Delete the status signal.
  criticalsection.Enter();
  vSemaphoreDelete(m_statussignal);
  m_statussignal = 0;
  criticalsection.Leave();

  if (on != powered)
    PrintLog(LOGMODEM, "ERROR: Could not power %s modem\n", on ? "on" : "off");

  //Return if the requested power state was reached.
  return on == powered;
}

#define DEFAULT_BAUDRATE   115200 //!< Default baud rate of the Quectel BG96.
#define PREFERRED_BAUDRATE 921600 //!< Fastest baud rate of the Quectel BG96.

/*! Detects the current baud rate of the BG96, and sets it to the maximum.
    \return true if detecting and setting the baud rate succeeded,
            false if it failed.*/
bool CModemBG96::SetupBaudrate()
{
  //check what the current modem baud rate is, start at the default modem baud rate
  uint32_t baudrate = DetectBaudrate(DEFAULT_BAUDRATE);
  if (baudrate == 0)
    return false;

  //if the detected baud rate is not the preferred baud rate, change it
  if (baudrate != PREFERRED_BAUDRATE)
  {
    PrintLog(LOGMODEM, "Modem is at baud rate %" PRIu32 ", setting the baud rate to %i\n",
             baudrate, PREFERRED_BAUDRATE);

    if (m_atcmd.ATIPR(PREFERRED_BAUDRATE) == ATCmdResult_Success)
    {
      DelayMS(1000); //Give the modem some time to switch baud rate.
    }
    else
    {
      PrintLog(LOGMODEM, "ERROR: unable to set modem baud rate:\n");
      m_atcmd.PrintError();
    }
  }

  //detect the baud rate again, start at the preferred baud rate
  return DetectBaudrate(PREFERRED_BAUDRATE) != 0;
}

/*! Detects the modem baud rate by sending an AT&F command,
    and checking if there is a valid reply.
    The AT&F command sets all parameters to the factory default,
    which makes sure that command echo is enabled,
    and that the response format is set to the long codes.
    When this function returns, it leaves the usart correctly configured.
    \param baudratefirst When not 0, this baud rate is checked first.
    \return The detected baud rate, or 0 when no valid baud rate was detected.*/
uint32_t CModemBG96::DetectBaudrate(uint32_t baudratefirst)
{
  PrintLog(LOGMODEM, "Detecting baud rate\n");

  //The modem might be set to auto baud, or to a specific baud rate.
  //Test these baudrates in order until the correct baud rate is found
  //or until the modem has responsed to the auto baud.
  static const uint32_t baudrates[] =
  {
    921600,
    460800,
    230400,
    115200,
    57600,
    38400,
    19200,
    9600,
    4800,
    2400,
    1200,
  };

  if (baudratefirst != 0 && CheckBaudrate(baudratefirst))
    return baudratefirst;

  for (uint32_t i = 0; i < sizeof(baudrates) / sizeof(baudrates[0]); i++)
  {
    if (baudratefirst != 0 && baudrates[i] != baudratefirst)
    {
      if (CheckBaudrate(baudrates[i]))
        return baudrates[i];
    }
  }

  PrintLog(LOGMODEM, "ERROR: Could not detect modem baud rate\n");

  return 0;
}

//!< The number of times the AT&F command is sent to the modem before giving up.
#define CHECK_RETRIES 5

/*! Checks the baud rate by setting up the usart, and sending the AT&F command
    to reset the modem to factory defaults, it then checks if it gets a valid reply.
    \param baudrate The baud rate to test.
    \return true if the modem responded at this baud rate, false if it did not.*/
bool CModemBG96::CheckBaudrate(uint32_t baudrate)
{
  PrintLog(LOGMODEM, "Trying baud rate %" PRIu32 "\n", baudrate);

  //Set up the usart in 8n1 mode, with hardware flow control enabled,
  //but only on the RTS line, because the modem doesn't yet have flow control enabled.
  //This does guarantee that when flow control is enabled on the modem,
  //it will transmit data to the usart, and the usart will always transmit data to the
  //modem even if it does not have flow control enabled.
  m_usartinit =
  {
    .USART_BaudRate            = baudrate,
    .USART_WordLength          = USART_WordLength_8b,
    .USART_StopBits            = USART_StopBits_1,
    .USART_Parity              = USART_Parity_No,
    .USART_Mode                = USART_Mode_Rx | USART_Mode_Tx,
    .USART_HardwareFlowControl = USART_HardwareFlowControl_RTS,
  };

  //Make sure the usart has the same buffer size as the CATCmd class,
  //so that the usart can buffer a complete message with echo enabled.
  m_usart.Setup(m_usartptr, &m_usartinit, AT_BUFSIZE);

  for (uint32_t i = 0; i < CHECK_RETRIES; i++)
  {
    if (m_atcmd.ATF() == ATCmdResult_Success)
    {
      PrintLog(LOGMODEM, "Modem responded correctly at baud rate %" PRIu32 "\n",
               baudrate);
      return true;
    }
    else
    {
      PrintLog(LOGMODEM, "Modem did not respond correctly at baud rate %" PRIu32 "\n",
               baudrate);
      m_atcmd.PrintError();
    }

    DelayMS(1000); //Give the modem some time to settle before trying again.
  }

  //close the usart if the baudrate is not valid,
  //this prevents further commands being sent to it when powering it down
  m_usart.Close();

  return false;
}

/*! Uses the AT+IFC command to enable flow control.
    \return true if setting flow control succeeded, false if it failed.*/
bool CModemBG96::SetFlowControl()
{
  PrintLog(LOGMODEM, "Enabling flow control\n");

  //enable RTS and CTS flow control on the BG96
  if (m_atcmd.ATIFC(2, 2) != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to enable flow control:\n");
    m_atcmd.PrintError();
    return false;
  }

  //enable RTS and CTS flow control on the usart
  m_usartinit.USART_HardwareFlowControl = USART_HardwareFlowControl_RTS_CTS;
  m_usart.Close();
  m_usart.Setup(m_usartptr, &m_usartinit, AT_BUFSIZE);

  return true;
}

/*! Uses the AT+CMEE command to set the modem error messages to verbose values.
    \return true if the command succeeded, false if it failed.*/
bool CModemBG96::SetErrorMessageFormat()
{
  PrintLog(LOGMODEM, "Setting modem message format to verbose\n");

  if (m_atcmd.ATCMEE(2) != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to set modem message format:\n");
    m_atcmd.PrintError();
    return false;
  }

  return true;
}

/*! Performs a factory reset using the AT+QPRTPARA command.
    \param factoryreset When true, perform a factory reset.
    \return true No factory reset needed.
    \return false Factory reset happened, modem should be powered off.*/
bool CModemBG96::FactoryReset(bool factoryreset)
{
  if (factoryreset)
  {
    PrintLog(LOGMODEM, "Performing factory reset\n");

    if (m_atcmd.ATQPRTPARA() != ATCmdResult_Success)
    {
      PrintLog(LOGMODEM, "ERROR: unable to perform modem factory reset:\n");
      m_atcmd.PrintError();
    }

    DelayMS(2000);

    return false; //Factory reset done, modem should be powered off.
  }

  return true; //No factory reset needed.
}

/*! Enables/disables phone functionality using the AT+CFUN command.
    \return true if setting the phone functionality succeeded.
    \return false if anything failed.*/
bool CModemBG96::SetPhoneFunctionality(bool on)
{
  PrintLog(LOGMODEM, "%s phone functionality\n", on ? "Enabling" : "Disabling");

  //enable phone functionality
  if (m_atcmd.ATCFUN(on ? 1 : 0, 0) != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to %s phone functionality:\n",
             on ? "enable" : "disable");
    m_atcmd.PrintError();
    return false;
  }

  //give the SIM some time to initialize, otherwise the modem RETURNS error
  //when trying to request PIN information
  if (on)
    DelayMS(2000);

  return true;
}

/*! Sets the scan mode to only scan for LTE Cat-M1 and GPRS networks.
    Don't scan for NB-IOT, because it takes too long.
    \return true if the commands succeeded.
    \return false if anything failed.*/
bool CModemBG96::SetNetworkScanMode()
{
  PrintLog(LOGMODEM, "Reading network scan mode\n");

  //Request the current scan mode.
  if (m_atcmd.ATQCFG("\"iotopmode\"") != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to read network search category:\n");
    m_atcmd.PrintError();
    return false;
  }

  const char* line;
  uint32_t    linelength;
  line = m_atcmd.ReplyLine(0, &linelength);

  //The reply from the modem is: +QCFG: "iotopmode",<mode>
  //Where <mode> can be:
  //  0 LTE Cat M1 on BG96 and eMTC on BG95
  //  1 NB-IoT
  //  2 LTE Cat M1 on BG96 and eMTC on BG95, and NB-IoT
  //The default is 2, which means that the modem will spend a lot of time scanning
  //for NB-IoT networks, which is unnecessary.
  uint32_t mode;
  if (sscanf(line, "+QCFG: \"iotopmode\",%" SCNu32, &mode) != 1)
  {
    PrintLogDecode(LOGMODEM, "ERROR: wrong reply to requesting network search "
                   "category: %.*s", (int)linelength, line);
    return false;
  }

  if (mode == 0)
  {
    PrintLog(LOGMODEM, "Network scan mode already set to LTE Cat M1 only\n");
    return true; //Mode is correct, no change necessary.
  }

  PrintLog(LOGMODEM, "Setting network scan mode to LTE Cat M1 only\n");

  if (m_type == ModemType_BG96)
  { //WARNING: the second argument should always be 0, which means the setting
    //takes effect after a reboot. Otherwise a modem bug is triggered where the
    //connection doesn't work for a few minutes.
    if (m_atcmd.ATQCFG("\"iotopmode\",0,0") != ATCmdResult_Success)
    {
      PrintLog(LOGMODEM, "ERROR: unable to set network search category to LTE Cat M1:\n");
      m_atcmd.PrintError();
      return false;
    }
  }
  else if (m_type == ModemType_BG95)
  { //Second argument is 1, to apply the change immediately, otherwise the BG95
    //will take longer than the timeout to connect to a network.
    if (m_atcmd.ATQCFG("\"iotopmode\",0,1") != ATCmdResult_Success)
    {
      PrintLog(LOGMODEM, "ERROR: unable to set network search category to LTE Cat M1:\n");
      m_atcmd.PrintError();
      return false;
    }
  }
  else
  {
    PrintLog(LOGMODEM, "ERROR: unknown modem type %i\n", (int)m_type);
    return false;
  }

  return true;
}

/* Checks the operator selection using AT+COPS?, if it's not set to auto,
   it will set it to auto using AT+COPS=0. The reason for this is that if
   AT+COPS=0 is executed, the previously selected operator is cleared.
   So AT+COPS=0 should only be executed when needed.
    \return true if the command succeeded, false if it failed.*/
bool CModemBG96::SetOperatorSelection()
{
  //Request the current operator selection mode.
  if (m_atcmd.ATCOPS_Query() == ATCmdResult_Success)
  {
    const char* line;
    uint32_t    linelength;
    line = m_atcmd.ReplyLine(0, &linelength);

    //Parse the operator selection line. It should have the format:
    //+COPS: <mode>
    //Where mode should be set to 0 for auto.
    uint32_t mode;
    if (sscanf(line, "+COPS: %" SCNu32, &mode) == 1)
    {
      if (mode == 0)
      {
        PrintLog(LOGMODEM, "Operator selection is set to auto\n");
        return true;
      }
      else
      {
        PrintLogDecode(LOGMODEM, "ERROR: operator selection is not set to auto: %.*s",
                       (int)linelength, line);
      }
    }
    else
    {
      PrintLogDecode(LOGMODEM, "ERROR: wrong reply to requesting operator selection: %.*s",
                     (int)linelength, line);
    }
  }
  else
  {
    PrintLog(LOGMODEM, "ERROR: unable to get operator selection:\n");
    m_atcmd.PrintError();
  }

  PrintLogDecode(LOGMODEM, "Setting operator selection to auto\n");

  //Set the operator selection to auto.
  if (m_atcmd.ATCOPS_Set(0) != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to set operator selection to auto:\n");
    m_atcmd.PrintError();
    return false;
  }

  return true;
}

/*! Gets the manufacturer, model, firmware version and IMEI from the modem.
    \return true if getting all info succeeds, false if anything fails.*/
bool CModemBG96::GetModemInfo()
{
  const char* line;
  uint32_t    linelength;

  PrintLog(LOGMODEM, "Getting modem info\n");

  //get the manufacturer
  if (m_atcmd.ATGMI() != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to get modem manufacturer:\n");
    m_atcmd.PrintError();
    return false;
  }
  line = m_atcmd.ReplyLine(0, &linelength);
  PrintLogDecode(LOGMODEM, "Manufacturer: %.*s", (int)linelength, line);

  //store the manufacturer as a C-string
  strncpy(m_manufact, line, std::min(linelength, (uint32_t)sizeof(m_manufact)));
  m_manufact[sizeof(m_manufact) - 1] = 0;

  //get the model
  if (m_atcmd.ATGMM() != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to get modem model:\n");
    m_atcmd.PrintError();
    return false;
  }
  line = m_atcmd.ReplyLine(0, &linelength);
  PrintLogDecode(LOGMODEM, "Model: %.*s", (int)linelength, line);

  //store the model as a C-string
  strncpy(m_model, line, std::min(linelength, (uint32_t)sizeof(m_model)));
  m_model[sizeof(m_model) - 1] = 0;

  //Set the m_type variable based on the manufacturer and model strings.
  m_type = ModemType_UNKNOWN;
  const char* quectel = "Quectel";
  if (strncasecmp(m_manufact, quectel, strlen(quectel)) == 0)
  {
    const char* bg96 = "BG96";
    const char* bg95 = "BG95";
    if (strncasecmp(m_model, bg96, strlen(bg96)) == 0)
    {
      m_type = ModemType_BG96;
      PrintLog(LOGMODEM, "Modem recognized as: Quectel BG96\n");
    }
    else if (strncasecmp(m_model, bg95, strlen(bg95)) == 0)
    {
      m_type = ModemType_BG95;
      PrintLog(LOGMODEM, "Modem recognized as: Quectel BG95\n");
    }
  }

  if (m_type == ModemType_UNKNOWN)
    PrintLog(LOGMODEM, "ERROR: modem type not recognized\n");

  //get the revision
  if (m_atcmd.ATGMR() != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to get modem revision:\n");
    m_atcmd.PrintError();
    return false;
  }
  line = m_atcmd.ReplyLine(0, &linelength);
  PrintLogDecode(LOGMODEM, "Revision: %.*s", (int)linelength, line);

  //store the revision as a C-string
  strncpy(m_revision, line, std::min(linelength, (uint32_t)sizeof(m_revision)));
  m_revision[sizeof(m_revision) - 1] = 0;

  //get the application version
  //not every modem firmware version supports this command
  if (m_atcmd.ATQAPPVER() == ATCmdResult_Success)
  {
    //store the application version as a C-string
    line = m_atcmd.ReplyLine(0, &linelength);
    sscanf(line, "+QAPPVER: %32[^\r]", m_appver);
    m_appver[sizeof(m_appver) - 1] = 0;
    PrintLogDecode(LOGMODEM, "App version: %s", m_appver);
  }
  else
  {
    m_appver[0] = 0;
    PrintLog(LOGMODEM, "ERROR: unable to get modem application version:\n");
    m_atcmd.PrintError();
  }

  //get the IMEI
  if (m_atcmd.ATGSN() != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to get IMEI:\n");
    m_atcmd.PrintError();
    return false;
  }
  line = m_atcmd.ReplyLine(0, &linelength);
  PrintLogDecode(LOGMODEM, "IMEI: %.*s", (int)linelength, line);

  if (linelength > sizeof(m_imei) - 1)
  {
    PrintLog(LOGMODEM, "ERROR: IMEI too long\n");
    return false;
  }

  //store the IMEI as a C-string
  strncpy(m_imei, line, std::min(linelength, (uint32_t)sizeof(m_imei)));
  m_imei[sizeof(m_imei) - 1] = 0;

  return true;
}

/*! Gets the IMSI and ICCID from the SIM.
    \return true if getting all info succeeds, false if something failed.*/
bool CModemBG96::ReadSIMIMSI()
{
  const char* line;
  uint32_t    linelength;

  //get the IMSI
  if (m_atcmd.ATCIMI() != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to get IMSI:\n");
    m_atcmd.PrintError();
    return false;
  }
  line = m_atcmd.ReplyLine(0, &linelength);
  PrintLogDecode(LOGMODEM, "IMSI: %.*s", (int)linelength, line);

  if (linelength > sizeof(m_siminfo.imsi) - 1)
  {
    PrintLog(LOGMODEM, "ERROR: IMSI too long\n");
    return false;
  }

  strncpy(m_siminfo.imsi, line, std::min(linelength, (uint32_t)sizeof(m_siminfo.imsi)));
  m_siminfo.imsi[sizeof(m_siminfo.imsi) - 1] = 0;

  return true;
}

/*! Reads the ICCID using the AT+QCCID command.
    \return true if reading the ICCID succeeded, false if it failed.*/
bool CModemBG96::ReadSIMICCID()
{
  m_siminfo.iccid[0] = 0; //always terminate the ICCID as C-string

  const char* line;
  uint32_t    linelength;

  //use the AT+QCCID command to read the ICCID
  if (m_atcmd.ATQCCID() != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to get ICCID:\n");
    m_atcmd.PrintError();
    return false;
  }
  line = m_atcmd.ReplyLine(0, &linelength);

  //check if the ICCID reply from the modem starts with "+QCCID: "
  const char* prefix = "+QCCID: ";
  uint32_t prefixlen = strlen(prefix);
  if (linelength < prefixlen || strncmp(line, prefix, prefixlen) != 0)
  {
    PrintLogDecode(LOGMODEM, "ERROR: expected ICCID to start with \"%s\", but received \"%.*s\"",
                   prefix, (int)linelength, line);
    return false;
  }

  PrintLogDecode(LOGMODEM, "ICCID: %.*s", (int)(linelength - prefixlen), line + prefixlen);

  if (linelength - prefixlen > sizeof(m_siminfo.iccid) - 1)
  {
    PrintLog(LOGMODEM, "ERROR: ICCID too long\n");
    return false;
  }

  strncpy(m_siminfo.iccid, line + prefixlen, std::min(linelength - prefixlen, (uint32_t)sizeof(m_siminfo.iccid)));
  m_siminfo.iccid[sizeof(m_siminfo.iccid) - 1] = 0;

  return true;
}

static int hex2int(char ch)
{
  if (ch >= '0' && ch <= '9')
    return ch - '0';
  if (ch >= 'A' && ch <= 'F')
    return ch - 'A' + 10;
  if (ch >= 'a' && ch <= 'f')
    return ch - 'a' + 10;
  return -1;
}

/*! Reads the service provider name from the SIM.
    \return true if reading the service provider name succeeded, false if it failed.*/
bool CModemBG96::ReadSIMSPN()
{
  const char* spn;
  uint32_t    spnlen;

  //make sure that the stored spn is a valid C-string
  m_siminfo.spn[0] = 0;

  //use the AT+CRSM command to do a read at file 0x6F46 of 17 bytes,
  //which is where the service provider name (spn) is stored
  //not all providers write this to the sim
  //this is specified in document ETSI TS 131 102, section 4.2.12 EF SPN (Service Provider Name)
  if (m_atcmd.ATCRSM_ReadBinary(0x6F46, 17, &spn, &spnlen) != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to read SPN from SIM:\n");
    m_atcmd.PrintError();
    return false;
  }

  //the spn bytes are returned as hex by the modem,
  //skip over the first 00, because it's the display condition byte
  //convert the rest of the bytes to ascii, until FF is encountered,
  //which is the value of unused bytes
  for (uint32_t i = 2; i < spnlen; i += 2)
  {
    //convert two hex characters to an integer
    int value = 0;
    for (uint32_t j = 0; j < 2; j++)
    {
      value <<= 4;
      int charvalue = hex2int(spn[i + j]);
      if (charvalue == -1)
      {
        PrintLogDecode(LOGMODEM, "ERROR: invalid character %c at position %" PRIu32 " in SPN from SIM: %.*s",
                       spn[i + j], i + j + 1, (int)spnlen, spn);
        return false;
      }
      value |= charvalue;
    }

    if (value != 0xFF)
    { //store the converted byte, and always terminate as C-string
      m_siminfo.spn[i / 2 - 1] = value;
      m_siminfo.spn[i / 2] = 0;
    }
    else
    { //if 0xFF is read, it's an unused byte, so stop converting
      break;
    }
  }

  PrintLogDecode(LOGMODEM, "SPN:  \"%s\"", m_siminfo.spn);

  return true;
}

/*! Checks if the SIM needs a pin, and if it does, it will try 0000.
    \return true if the SIM does not need a pin, or if 0000 is accepted.
            false if anything fails.*/
bool CModemBG96::EnterSIMPIN()
{
  PrintLog(LOGMODEM, "Checking if the SIM needs a PIN\n");

  bool ready;
  if (!CheckSIMPIN(&ready))
    return false;

  if (ready)
  {
    PrintLog(LOGMODEM, "SIM does not need a PIN\n");
    return true;
  }

  //check the number of PIN retries, if it's only 1, don't try to enter the PIN
  uint32_t pincounter;
  if (!GetNumPINRetries(&pincounter))
    return false;
  
  if (pincounter <= 1)
  {
    PrintLog(LOGMODEM, "ERROR: number of pin retries left: %" PRIu32 ", not entering PIN\n", pincounter);
    return false;
  }

  PrintLog(LOGMODEM, "Number of PIN retries left: %" PRIu32 "\n", pincounter);

  //enter the SIM PIN and check if it's correct
  CAPNList apnlist;
  const char* simpin = apnlist.GetPIN(m_siminfo.iccid);
  if (m_atcmd.ATCPIN_Set(simpin) == ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "PIN %s accepted\n", simpin);

    //the modem needs some time to initialize the SIM
    //without this delay, getting any information about the SIM returns ERROR
    DelayMS(1000);

    return true;
  }
  else
  {
    PrintLog(LOGMODEM, "ERROR: PIN %s not accepted:\n", simpin);
    m_atcmd.PrintError();
    return false;
  }
}

/*! Uses the AT+CPIN? command to check if the SIM needs a PIN.
    \param ready Returns true if the SIM is ready to operate.
    \return true If the sim is ready to operate, or needs a PIN,
            false if anything else was returned.*/
bool CModemBG96::CheckSIMPIN(bool* ready)
{
  const char* line;
  uint32_t    linelength;

  *ready = false;

  //check if the SIM needs a PIN
  if (m_atcmd.ATCPIN_Query() != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to get SIM PIN information:\n");
    m_atcmd.PrintError();
    return false;
  }
  line = m_atcmd.ReplyLine(0, &linelength);

  if (strncmp(line, "+CPIN: READY", linelength) == 0)
  { //if +CPIN: READY is returned, the SIM does not need a PIN
    *ready = true;
    return true;
  }
  else if (strncmp(line, "+CPIN: SIM PIN", linelength) == 0)
  { //if +CPIN: SIM PIN is returned, the SIM needs a PIN code
    return true;
  }
  else
  { //consider any other response an error
    PrintLogDecode(LOGMODEM, "ERROR: getting SIM PIN returned: \"%.*s\"", (int)linelength, line);
    return false;
  }
}

/*! Uses the AT+QPINC command to get the number of PIN retries from the SIM.
    \param pincounter The number of PIN retries is returned here.
    \return true if getting the number of PIN retries succeeded, false if it failed.*/
bool CModemBG96::GetNumPINRetries(uint32_t* pincounter)
{
  //request the number of PIN retries
  if (m_atcmd.ATQPINC("SC") != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to request number of PIN retries:\n");
    m_atcmd.PrintError();
    return false;
  }

  const char* line;
  uint32_t    linelength;
  line = m_atcmd.ReplyLine(0, &linelength);

  //the reply is something like:
  //+QPINC: "SC",3,10
  //the first number is the number of PIN retries, the second number is the number of PUK retries
  if (sscanf(line, "+QPINC: \"SC\",%" SCNu32, pincounter) != 1)
  {
    PrintLogDecode(LOGMODEM, "ERROR: wrong reply to number of PIN retries: %.*s", (int)linelength, line);
    return false;
  }

  return true;
}

//! Maximum time to wait for the modem to register to the network.
#define REGISTERED_STATE_TIMEOUT 300000

/*! Checks the registration state using the AT+CREG? command.
    It will wait for REGISTERED_STATE_TIMEOUT milliseconds for the state to change to registered before giving up.
    \return true if registering to the network succeeded, false if anything failed.*/
bool CModemBG96::WaitNetworkRegister()
{
  //states as returned by AT+CREG?, copied from the BG96 AT commands manual
  static const char *statestrs[] =
  {
    "Not registered. MT is not currently searching an operator to register to",
    "Registered, home network",
    "Not registered, but MT is currently trying to attach or searching an operator to register to",
    "Registration denied",
    "Unknown",
    "Registered, roaming"
  };

  //Enable network registration and location information, this also enables
  //unsollicited return codes, which are currently not handled. They might cause
  //errors when executing AT commands, so they are retried on error.
  m_atcmd.ATCREG_Set(2);

  PrintLog(LOGMODEM, "Waiting for network register\n");
  int64_t start = Millis();
  int32_t oldstat = -1;
  do
  {
    //request the current registration state
    if (m_atcmd.ATCREG_Query() != ATCmdResult_Success)
    {
      PrintLog(LOGMODEM, "ERROR: unable to get network registration status:\n");
      m_atcmd.PrintError();
      DelayMS(1000);
      continue;
    }
    const char* line;
    uint32_t    linelength;
    line = m_atcmd.ReplyLine(0, &linelength);

    //scan the reply from the AT+CREG? command
    memset(m_lacstr, 0, sizeof(m_lacstr));
    memset(m_cistr,  0, sizeof(m_cistr));
    uint32_t n;
    uint32_t stat;
    int      numparms = sscanf(line, "+CREG: %" SCNu32 ",%" SCNu32 ",\"%" XSTR(LAC_SIZE)
                               "[^\"]\",\"%" XSTR(CI_SIZE) "[^\"]\"",
                               &n, &stat, m_lacstr, m_cistr);
    m_lacstr[LAC_SIZE] = 0;
    m_cistr[CI_SIZE]   = 0;

    if (numparms < 2)
    {
      PrintLogDecode(LOGMODEM, "ERROR: invalid network registration status: \"%.*s\"", (int)linelength, line);
      DelayMS(1000);
      continue;
    }

    //print the state if it changed
    if (stat != (uint32_t)oldstat)
    {
      if (stat < sizeof(statestrs) / sizeof(statestrs[0]))
        PrintLog(LOGMODEM, "Registration state: %s\n", statestrs[stat]);
      else
        PrintLog(LOGMODEM, "INVALID REGISTRATION STATE %" PRIu32 "\n", stat);

      oldstat = stat;
    }

    //1 = registered to home network, 5 = registered to roaming network
    if (stat == 1 || stat == 5)
    {
      //convert the location area code(lac) and cell id (ci) from hex to decimal
      m_lac = 0;
      m_ci  = 0;
      sscanf(m_lacstr, "%" SCNx32, &m_lac);
      sscanf(m_cistr, "%" SCNx32, &m_ci);

      PrintLog(LOGMODEM, "Registered to %s network, LAC: 0x%s(%" PRIu32
               "), CID: 0x%s(%" PRIu32 ")\n",
               stat == 1 ? "home" : "roaming", m_lacstr, m_lac, m_cistr, m_ci);
      m_atcmd.ATCREG_Set(0); //disable unsollicited return code
      return true;
    }

    DelayMS(1000);
  }
  while (Millis() - start < REGISTERED_STATE_TIMEOUT);

  PrintLog(LOGMODEM, "Timeout registering to the network\n");
  m_atcmd.ATCREG_Set(0); //disable unsollicited return code
  return false;
}

//! Maximum time in milliseconds to wait for the modem to attach to the packet service.
#define PS_STATE_TIMEOUT 300000

/*! Uses the AT+CGATT command to check if the modem has been attached to the packet service.
    \return true The mode is attached to the packet service.
    \return false On a timeout or an error happened.*/
bool CModemBG96::WaitPacketServiceAttach()
{
  PrintLog(LOGMODEM, "Waiting for packet service attach\n");
  int64_t start = Millis();
  while (Millis() - start < PS_STATE_TIMEOUT)
  {
    if (m_atcmd.ATCGATT_Query() == ATCmdResult_Success)
    {
      const char* line;
      uint32_t    linelength;
      line = m_atcmd.ReplyLine(0, &linelength);
  
      //The reply of the AT+CGATT command indicates the attached status.
      int32_t state;
      int numparms = sscanf(line, "+CGATT: %" SCNi32, &state);
      if (numparms == 1 && state == 1)
      {
        PrintLog(LOGMODEM, "Attached to packet service\n");
        return true;
      }
  
      DelayMS(1000);
    }
    else
    {
      PrintLog(LOGMODEM, "ERROR: unable to get packet service attached status:\n");
      m_atcmd.PrintError();
      return false;
    }
  }

  PrintLog(LOGMODEM, "ERROR: timeout attaching to packet service\n");
  return false;
}

/*! Timeout in milliseconds to wait for a valid quality number,
    this is set to 5 minutes because it's possible that if this is unknown,
    the modem also doesn't have a working connection.*/
#define QUALITY_TIMEOUT 300000

/*! Gets the received signal strength indication and bit error rate using the AT+CSQ command.
    \return true if getting all information succeeded, false if anything failed.*/
bool CModemBG96::GetSignalQuality()
{
  bool    first = true;
  int64_t start = Millis();
  do
  {
    //read the rssi and the bit error rate
    if (!UpdateRSSI())
      return false;

    //a value of 99 indicated not known, if that happens try again
    //don't test the bit error rate for 99, because that value is often not known
    if (m_rssi != 99)
      break;
    
    if (first)
    {
      PrintLog(LOGMODEM, "Waiting for network signal quality information\n");
      first = false;
    }

    DelayMS(1000);
  }
  while (Millis() - start < QUALITY_TIMEOUT);

  PrintLog(LOGMODEM, "Network quality: RSSI: %" PRIi32 ", bit error rate: %" PRIu32 "\n", m_rssi, m_ber);

  return true;
}

/*! Gets the full network name (fnn), short network name (snn), service provider name (spn)
    and registered public land mobile network (rplmn) from the modem.
    \return true if getting all info succeeded, false if anything failed.*/
bool CModemBG96::GetNetworkInfo()
{
  const char* line = "";
  uint32_t    linelength = 0;

  const uint32_t tries = 10;
  bool success = false;
  for (uint32_t i = 0; i < tries; i++)
  { //get the name of the network that the modem registered to
    //try this several times, it might fail at first
    EATCmdResult result = m_atcmd.ATQSPN();
    if (result == ATCmdResult_Success)
    {
      success = true;
      line = m_atcmd.ReplyLine(0, &linelength);
      break;
    }
    else
    {
      PrintLog(LOGMODEM, "ERROR: unable to get name of registered network:\n");
      m_atcmd.PrintError();
      if (result == ATCmdResult_Reply_Not_OK)
        DelayMS(5000); //if the modem reply was not OK, retry
      else
        break;         //on any other error, abort
    }
  }

  if (!success)
    return false;

  //scans: +QSPN: "fnn","snn","spn",0,"rplmn"
  //each string is read including the last double quote,
  //because sscanf needs to read at least one character
  //so %[^,] is used to read a string until a comma is read,
  //except for the last string where %[^\r] is used to read a string until a carriage return
  const char* scanstr = "+QSPN: \"%" XSTR(FNN_SIZE) "[^,],\"%" XSTR(SNN_SIZE) "[^,],\"%"
                        XSTR(SPN_SIZE) "[^,],%*i,\"%" XSTR(RPLMN_SIZE) "[^\r]";
  int numparms = sscanf(line, scanstr, m_fnn, m_snn, m_spn, m_rplmn);

  //make sure all parameters are valid C-strings
  m_fnn[FNN_SIZE]     = 0;
  m_snn[SNN_SIZE]     = 0;
  m_spn[SPN_SIZE]     = 0;
  m_rplmn[RPLMN_SIZE] = 0;

  //remove the last character from all parameters,
  //because it's a double quote
  if (strlen(m_fnn) > 0)
    m_fnn[strlen(m_fnn) - 1] = 0;
  if (strlen(m_snn) > 0)
    m_snn[strlen(m_snn) - 1] = 0;
  if (strlen(m_spn) > 0)
    m_spn[strlen(m_spn) - 1] = 0;
  if (strlen(m_rplmn) > 0)
    m_rplmn[strlen(m_rplmn) - 1] = 0;

  if (numparms != 4)
  {
    PrintLogDecode(LOGMODEM, "ERROR: invalid registered network string: \"%.*s\"", (int)linelength, line);
    return false;
  }

  //the rplmn is make of the three decimal digit mobile country code,
  //and the two or three decimal digit mobile network code
  uint32_t rplmnlength = strlen(m_rplmn);
  if (rplmnlength != 5 && rplmnlength != 6)
  {
    PrintLogDecode(LOGMODEM, "ERROR: invalid RPLMN, length should be 5 or 6, but got string of length %" PRIu32 ": \"%s\"",
                   rplmnlength, m_rplmn);
    return false;
  }

  //check if the rplmn is completely numerical
  for (uint32_t i = 0; i < rplmnlength; i++)
  {
    if (!isdigit(m_rplmn[i]))
    {
      PrintLogDecode(LOGMODEM, "ERROR: invalid non numerical character %c at position %" PRIu32 " in the RPLMN \"%s\"",
                     m_rplmn[i], i + 1, m_rplmn);
      return false;
    }
  }

  PrintLog(LOGMODEM, "Registered to:\n");
  PrintLogDecode(LOGMODEM, "FNN:   \"%s\"", m_fnn);
  PrintLogDecode(LOGMODEM, "SNN:   \"%s\"", m_snn);
  PrintLogDecode(LOGMODEM, "SPN:   \"%s\"", m_spn);
  PrintLogDecode(LOGMODEM, "RPLMN: \"%s\"", m_rplmn);

  return true;
}

#define MODESTRLEN 32
#define OPERSTRLEN 32
#define BANDSTRLEN 32

/*! Uses the AT+QCSQ command to check if a GSM, CAT-M1 or CAT-NB1 is used.
    \return true if getting the information succeeded, false if it failed.*/
bool CModemBG96::GetNetworkType()
{
  const char* line;
  uint32_t    linelength;

  //get the network type and quality
  if (m_atcmd.ATQCSQ() != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to get network type and signal quality:\n");
    m_atcmd.PrintError();
    return false;
  }

  line = m_atcmd.ReplyLine(0, &linelength);

  //get the parameters from the reply of the AT+QCSQ command
  char mode[MODESTRLEN + 1] = {};
  int32_t rssi;
  int32_t rsrp;
  int32_t sinr;
  int32_t rsrq;
  int numparms = sscanf(line, "+QCSQ: \"%" XSTR(MODESTRLEN) "[^\"]\",%" SCNi32 ",%" SCNi32 ",%" SCNi32 ",%" SCNi32,
                        mode, &rssi, &rsrp, &sinr, &rsrq);

  //parse the parameters of the AT+QCSQ command,
  //in case of GSM mode, only the RSSI is reported,
  //in case of CAT-M1 and CAT-NB1, the RSSI, RSRP, SINR and RSRQ are reported
  //The Quectel BG96 returns "CAT-M1" and "CAT-NB1".
  //The Quectel BG95 returns "eMTC" and "NBIoT".
  if (numparms == 2 && strcmp(mode, "GSM") == 0)
  {
    PrintLog(LOGMODEM, "Access type: GSM, RSSI: %" PRIi32 "\n", rssi);
  }
  else if (numparms == 5 &&
           ((strcmp(mode, "CAT-M1") == 0) || (strcmp(mode, "eMTC") == 0)))
  {
    PrintLog(LOGMODEM, "Access type: LTE Cat M1, RSSI: %" PRIi32 ", RSRP: %"
             PRIi32 ", SINR: %" PRIi32 ", RSRQ: %" PRIi32 "\n",
             rssi, rsrp, sinr, rsrq);
  }
  else if (numparms == 5 &&
           ((strcmp(mode, "CAT-NB1") == 0) || (strcmp(mode, "NBIoT") == 0)))
  {
    PrintLog(LOGMODEM, "Access type: LTE Cat NB1, RSSI: %" PRIi32 ", RSRP: %"
             PRIi32 ", SINR: %" PRIi32 ", RSRQ: %" PRIi32 "\n",
             rssi, rsrp, sinr, rsrq);
  }
  else
  {
    PrintLogDecode(LOGMODEM, "ERROR: unknown network type string: \"%.*s\"", (int)linelength, line);
    return false;
  }

  //get the network type, operator, band and channel
  if (m_atcmd.ATQNWINFO() != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to get network information:\n");
    m_atcmd.PrintError();
    return false;
  }

  line = m_atcmd.ReplyLine(0, &linelength);

  char    oper[OPERSTRLEN + 1] = {};
  char    band[BANDSTRLEN + 1] = {};
  int32_t channel = 0;

  //parse the parameters of the AT+ATQNWINFO command,
  //its output is: +QNWINFO: "mode","oper","band",channel
  numparms = sscanf(line, "+QNWINFO: \"%" XSTR(MODESTRLEN) "[^\"]\",\"%"
                    XSTR(OPERSTRLEN) "[^\"]\",\"%" XSTR(BANDSTRLEN) "[^\"]\",%" SCNi32,
                    mode, oper, band, &channel);

  if (numparms != 4)
  {
    PrintLogDecode(LOGMODEM, "ERROR: invalid AT+ATQNWINFO returned invalid reply: \"%.*s\"",
                   (int)linelength, line);
    return false;
  }

  PrintLog(LOGMODEM, "Access technology: %s, operator: %s, band: %s, channel: %" PRIi32 "\n",
           mode, oper, band, channel);

  return true;
}

/*! Picks an APN based on the SPN and the IMSI of the SIM,
    then sets up a PDP context with that APN, and tries to activate it.
    If activating the PDP context fails, it will try the next APN in the list.
    \return true if activating the PDP context succeeds, false if it failed.*/
bool CModemBG96::ActivatePDPContext()
{
  CAPNList apnlist;
  const _apn* apn;

  PrintLog(LOGMODEM, "Setting up PDP context\n");

  //find the apn (access point name) info based on
  //spn  (service provider name)
  //imsi (international mobile subscriber identity)
  while ((apn = apnlist.GetApn(m_siminfo.spn, m_siminfo.imsi, m_siminfo.iccid)) != nullptr)
  {
    CLogLock loglock;
    PrintLog(LOGMODEM, "Selected APN: \"%s\"", apn->apn);
    if (apn->user && strlen(apn->user) > 0)
      PrintLog(LOGMODEM, " user: \"%s\"", apn->user);
    if (apn->password && strlen(apn->password) > 0)
      PrintLog(LOGMODEM, " password: \"%s\"", apn->password);
    PrintLog(LOGMODEM, "\n");
    loglock.Leave();

    //try to enable the PDP context several times before trying a new APN
    const uint32_t tries = 3;
    const uint32_t retrydelay = 2000;
    for (uint32_t i = 1; i <= tries; i++)
    {
      PrintLog(LOGMODEM, "Setting up PDP context try %" PRIu32 "\n", i);

      //set the APN of PDP context 1, so that it has internet access
      uint32_t contexttype = 1; //IPv4
      const char* user     = apn->user ? apn->user : "";
      const char* password = apn->password ? apn->password : "";
      if (m_atcmd.ATQICSGP_Set(1, &contexttype, apn->apn, user, password) != ATCmdResult_Success)
      {
        PrintLog(LOGMODEM, "ERROR: unable to set up PDP context:\n");
        m_atcmd.PrintError();
        DelayMS(retrydelay);
        continue; //try the next one in the list
      }

      //activate the PDP context
      //use a timeout of 5 minutes, because this can take a long time
      m_atcmd.SetReadTimeout(300000);
      EATCmdResult result = m_atcmd.ATQIACT_Set(1);
      m_atcmd.ResetReadTimeout();

      if (result == ATCmdResult_Success)
      { //if the PDP context was successfully activated, this function is done
        PrintLog(LOGMODEM, "PDP context successfully activated\n");
        return true;
      }
      else
      { //if there was an error activating the PDP context, try the next APN in the list
        PrintLog(LOGMODEM, "ERROR: unable to activate PDP context:\n");
        m_atcmd.PrintError();
        DelayMS(retrydelay);
        continue;
      }
    }
  }

  PrintLogDecode(LOGMODEM, "ERROR: could not find working APN based on SPN \"%s\" and IMSI \"%s\"",
                 m_siminfo.spn, m_siminfo.imsi);
  return false;
}


/*! Deactivates the PDP context that was previously activated with m_atcmd.ATQIACT_Set().
    \return true if deactivating the PDP context succeeded, false if it failed.*/
bool CModemBG96::DeactivatePDPContext()
{
  PrintLog(LOGMODEM, "Deactivating PDP context\n");

  EATCmdResult result = m_atcmd.ATQIDEACT(1);

  if (result == ATCmdResult_Success)
  { //if the PDP context was successfully deactivated, this function is done
    PrintLog(LOGMODEM, "PDP context successfully deactivated\n");
    return true;
  }
  else
  { //if there was an error deactivating the PDP context, try the next APN in the list
    PrintLog(LOGMODEM, "ERROR: unable to deactivate PDP context:\n");
    m_atcmd.PrintError();
    return false;
  }
}

/*! Uses the AT+QIACT command to get the IP address of PDP context 1.
    \return true if getting the IP address succeeded, false if it failed.*/
bool CModemBG96::GetIPAddress()
{
  const char* line;
  uint32_t    linelength;

  if (m_atcmd.ATQIACT_Query())
  {
    PrintLog(LOGMODEM, "ERROR: unable to read IP address:\n");
    m_atcmd.PrintError();
    return false;
  }

  //request information about all context IDs
  uint32_t numlines = m_atcmd.ReplyNumLines();
  for (uint32_t i = 0; i < numlines - 1; i++)
  {
    line = m_atcmd.ReplyLine(i, &linelength);

    uint32_t contextid;
    uint32_t contextstate;
    uint32_t contexttype;
    int numparms = sscanf(line, "+QIACT: %" SCNu32 ",%" SCNu32 ",%" SCNu32,
                          &contextid, &contextstate, &contexttype);

    //the reply is something like:
    //+QIACT: 1,1,1,"1.2.3.4"
    //the first three integers should always be present
    if (numparms < 3)
    {
      PrintLogDecode(LOGMODEM, "Invalid PDP context line: \"%.*s\"", (int)linelength, line);
      continue;
    }

    //only PDP context 1 should be connected to the internet
    if (contextid != 1)
      continue;

    if (contextstate == 0)
    {
      PrintLog(LOGMODEM, "ERROR: PDP context 1 is deactivated\n");
      return false;
    }

    //the ip address starts at the first double quote
    const char* ipaddrstart = (const char*)memchr(line, '"', linelength);
    if (ipaddrstart == nullptr)
    {
      PrintLogDecode(LOGMODEM, "Invalid PDP context line, did not find starting quote: \"%.*s\"", (int)linelength, line);
      continue;
    }

    ipaddrstart++; //skip over the first double quote

    //find the number of characters until the terminating double quote
    int ipaddrlen = strcspn(ipaddrstart, "\"");

    PrintLogDecode(LOGMODEM, "IP address: %.*s", ipaddrlen, ipaddrstart);
    return true;
  }

  PrintLog(LOGMODEM, "ERROR: did not get IP address from PDP context 1\n");
  return false;
}

/*! Checks if the modem firmware is at a revision that should be updated.
    The modem firmware update files are supplied by http://top-electronics.com/
    The files are delta update files, they upgrade from one specific revision to another.
    \param allowfwupdate Firmware update is only done when this is true.
    \return true if the firmware was not updated and the modem can be used normally.
    \return false if the modem was updated and rebooted.*/
bool CModemBG96::UpdateFirmware(bool allowfwupdate)
{
  if (!allowfwupdate)
  {
    PrintLog(LOGMODEM, "Updating firmware disabled\n");
    return true;
  }

  //check if there is a firmware update file for the current modem firmware revision
  const char* url = nullptr;
  if (strncmp(m_revision, "BG96MAR02A07M1G", REVISION_SIZE) == 0 &&
      strncmp(m_appver, "01.010.01.010", APPVER_SIZE) == 0)
  { //updater file to go from BG96MAR02A07M1G app version 01.010.01.010 to BG96MAR02A08M1G app version 01.013.01.013
    url = "http://rsoffice.net/firmware/Modem/Quectel%20BG96/BG96MAR02A07M1G_01.010.01.010-BG96MAR02A08M1G_01.013.01.013.bin";
  }
  else if (strncmp(m_revision, "BG96MAR02A08M1G", REVISION_SIZE) == 0 &&
           strncmp(m_appver, "01.012.01.012", APPVER_SIZE) == 0)
  { //updater file to go from BG96MAR02A08M1G app version 01.012.01.012 to BG96MAR02A08M1G app version 01.013.01.013
    url = "http://rsoffice.net/firmware/Modem/Quectel%20BG96/BG96MAR02A08M1G_01.012.01.012-BG96MAR02A08M1G_01.013.01.013.bin";
  }
  else
  {
    PrintLog(LOGMODEM, "Not updating firmware\n");
    return true;
  }

  m_needsfwupdate = true;

  PrintLog(LOGMODEM, "Updating firmware with %s\n", url);

  //execute the firmware update over the air command
  if (m_atcmd.ATQFOTADL(url) != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: Unable to update firmware:\n");
    m_atcmd.PrintError();
    return true; //modem should be used normally if this fails
  }

  //return false if the firmware updated completed, because the modem has
  //powered down and should be set up again
  return !WaitFirmwareUpdateComplete();
}

/*! Maximum number of milliseconds a modem firmware update can take.*/
#define MODEM_FW_UPDATE_TIME (7200 * 1000)

/*! Waits for a firmware update to complete after it has been started with AT+QFOTADL
    \return true if the firmware updated has completed successfully.
    \return false on any error.*/
bool CModemBG96::WaitFirmwareUpdateComplete()
{
  //check the progress of the firmware update over the air
  int64_t start = Millis();
  for(;;)
  {
    int64_t now = Millis();
    int64_t elapsed = now - start;
    if (elapsed > MODEM_FW_UPDATE_TIME)
    {
      PrintLog(LOGMODEM, "ERROR: Timeout updating firmware\n");
      return false;
    }

    //read a line from the modem, that starts and ends with \r\n
    EATCmdResult result;
    result = m_atcmd.ATQFOTADL_State(MODEM_FW_UPDATE_TIME - elapsed);
    if (result != ATCmdResult_Success)
    {
      PrintLog(LOGMODEM, "ERROR: Unable to read firmware update state:\n");
      m_atcmd.PrintError();
      return false;
    }

    const char* line;
    uint32_t    linelength;
    line = m_atcmd.ReplyLine(0, &linelength);

    //read a string that is:
    //+QIND: "<indication>","<state>",<value>
    char    indication[32] = {};
    char    state[32] = {};
    int32_t value;
    int     numparms;
    numparms = sscanf(line, "+QIND: \"%31[^\"]\",\"%31[^\"]\",%" SCNi32, indication, state, &value);

    //sscanf should have read at least the 2 strings
    //indication should always be FOTA at this point
    if (numparms < 2 || strcmp(indication, "FOTA") != 0)
    {
      PrintLogDecode(LOGMODEM, "ERROR: Received wrong firmware update state: %.*s",
                     (int)linelength, line);
      return false;
    }

    CLogLock loglock;
    PrintLog(LOGMODEM, "Firmware update state: %s", state);
    if (numparms > 2)
      PrintLog(LOGMODEM, ":%" PRIu32, value);
    PrintLog(LOGMODEM, "\n");
    loglock.Leave();
    
    if (strcmp(state, "HTTPEND") == 0)
    { //when the HTTP request has finished, check the status, it should be 0 for no error
      if (numparms <= 2)
      {
        PrintLogDecode(LOGMODEM, "ERROR: Wrong HTTPEND state in firmware update state: %.*s", (int)linelength, line);
        return false;
      }
      else if (value != 0)
      {
        PrintLog(LOGMODEM, "ERROR: Firmware update ended with HTTP error: %" PRIi32 "\n", value);
        return false;
      }
    }
    else if (strcmp(state, "END") == 0)
    { //when the firmware update has finished, check the status, it should be 0 for no error
      if (numparms <= 2)
      {
        PrintLogDecode(LOGMODEM, "ERROR: Wrong END state in firmware update state: %.*s", (int)linelength, line);
        return false;
      }
      else if (value != 0)
      {
        PrintLog(LOGMODEM, "ERROR: Firmware update ended with error: %" PRIi32 "\n", value);
        return false;
      }
      else
      {
        PrintLog(LOGMODEM, "Firmware update complete\n");
        m_needsfwupdate = false;

        //after the modem update has finished, it will power down by itself,
        //wait for the modem to power down, then close the usart to prevent any
        //commands being sent to it
        SetPower(false, true);
        m_usart.Close();

        return true; //the modem has finished updating
      }             
    }
  }

  //This point is never reached, but gcc complains about no return.
  return false;
}

/*! Number of milliseconds to check if a firmware update is in progress.*/
#define MODEM_FW_UPDATE_CHECK_TIME (5 * 1000)

/*! Checks if there's a firmware update in progress,
    if so it will wait for it to finish.
    \return true if the modem setup should be continued.
    \return false if the modem has updated its firmware and should be set up again.*/
bool CModemBG96::CheckFirmwareUpdateInProgress()
{
  PrintLog(LOGMODEM, "Checking if firmware update is in progress\n");

  int64_t start = Millis();
  for(;;)
  {
    int64_t now = Millis();
    int64_t elapsed = now - start;
    if (elapsed > MODEM_FW_UPDATE_CHECK_TIME)
      break;

    //read a line from the modem, that starts and ends with \r\n
    EATCmdResult result;
    result = m_atcmd.ATQFOTADL_State(MODEM_FW_UPDATE_CHECK_TIME - elapsed);
    if (result != ATCmdResult_Success)
      continue;

    const char* line;
    uint32_t    linelength;
    line = m_atcmd.ReplyLine(0, &linelength);

    //check if the read line starts with +QIND: "FOTA"
    //which means that a firmware update is in progress
    char indication[32] = {};
    int  numparms;
    numparms = sscanf(line, "+QIND: \"%31[^\"]\"", indication);
    if (numparms == 1 && strcmp(indication, "FOTA") == 0)
    { //return false if the firmware updated completed, because the modem has
      //powered down and should be set up again
      PrintLog(LOGMODEM, "Firmware update in progress, waiting for it to finish\n");
      return !WaitFirmwareUpdateComplete();
    }
  }

  PrintLog(LOGMODEM, "Firmware update is not in progress\n");
  return true;
}

/*! Does a HTTP GET request to a url.
    \param url The url to do the get request with.
    \param timeout The timeout for the GET request in seconds.
    \return true if the GET request succeeded, false if it failed.*/
bool CModemBG96::HTTPGet(const char* url, uint32_t timeout, CHTTPDownload* httpdownload, bool logpercentages /*= false*/)
{
  if (!m_activated)
  {
    PrintLogDecode(LOGMODEM, "ERROR: modem not activated, can't perform "
                   "HTTP GET: %s", url);
    return false;
  }

  PrintLogDecode(LOGMODEM, "HTTP GET: %s", url);

  //set up the context ids and the URL
  if (!SetupHTTPRequest(url))
    return false;

  //start the HTTP GET request
  uint32_t httprspcode;
  uint32_t contentlength;
  if (!StartHTTPGET(timeout, &httprspcode, &contentlength))
    return false;

  //download the data returned by the GET request
  //TODO; abort if the content-length is larger than the max download size
  if (!HTTPDownload(url, timeout, httpdownload, logpercentages, httprspcode, contentlength))
    return false;

  //receive the final data from the AT+QHTTPREAD command
  uint32_t httpresult;
  if (!FinishHTTPDownload(&httpresult))
    return false;

  PrintLog(LOGMODEM, "Finished downloading %" PRIu32 " bytes with modem result %" PRIu32 "\n",
           contentlength, httpresult);

  return true;
}

bool CModemBG96::HTTPPost(const char* url, uint32_t timeout, CHTTPPost* httppost, bool logpercentages /*= false*/)
{
  if (!m_activated)
  {
    PrintLogDecode(LOGMODEM, "ERROR: modem not activated, can't perform "
                   "HTTP POST: %s", url);
    return false;
  }

  PrintLogDecode(LOGMODEM, "HTTP POST: %s", url);

  uint32_t uploadsize = httppost->UploadSize();

  //set up the context ids and the URL
  if (!SetupHTTPRequest(url))
    return false;

  //set the content type of the http post data to application/octet-stream
  if (!SetHTTPContentType(2))
    return false;

  //start the HTTP POST
  if (!StartHTTPPOST(uploadsize, timeout))
    return false;

  //upload all data needed for the HTTP POST
  if (!HTTPUpload(url, timeout, httppost, logpercentages, uploadsize))
    return false;

  //check the reply from the modem after all data has finished uploading
  //TODO: abort if the content-length is larger than the max download size
  uint32_t httprspcode;
  uint32_t contentlength;
  if (!FinishHTTPPOST(timeout, &httprspcode, &contentlength))
    return false;

  //download the data returned by the POST request
  if (!HTTPDownload(url, timeout, httppost, logpercentages, httprspcode, contentlength))
    return false;

  //receive the final data from the AT+QHTTPREAD command
  uint32_t httpresult;
  if (!FinishHTTPDownload(&httpresult))
    return false;

  PrintLog(LOGMODEM, "Finished uploading %" PRIu32 " bytes and downloading %" PRIu32
           " bytes with modem result %" PRIu32 "\n",
           uploadsize, contentlength, httpresult);

  return true;
}

/*! Sets up the context ids and the HTTP url.*/
bool CModemBG96::SetupHTTPRequest(const char* url)
{
  //use PDP context 1 for HTTP
  if (m_atcmd.ATQHTTPCFG("contextid", 1) != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to set HTTP PDP context:\n");
    m_atcmd.PrintError();
    return false;
  }

  //use SSL context 1 for SSL
  if (m_atcmd.ATQHTTPCFG("sslctxID", 1) != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to set HTTP SSL context:\n");
    m_atcmd.PrintError();
    return false;
  }

  //set the URL to use for HTTP
  if (m_atcmd.ATQHTTPURL(url) != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to set HTTP URL:\n");
    m_atcmd.PrintError();
    return false;
  }

  return true;
}

/*! Sets the content type of the HTTP POST data.
    \param content_type:
           0 application/x-www-form-urlencoded
           1 text/plain
           2 application/octet-stream
           3 multipart/form-data
     \return true if setting the content type succeeded, false if it failed.*/
bool CModemBG96::SetHTTPContentType(uint32_t content_type)
{
  if (m_atcmd.ATQHTTPCFG("contenttype", content_type) != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to set HTTP content_type:\n");
    m_atcmd.PrintError();
    return false;
  }

  return true;
}

/*! Starts the HTTP GET request, which should have been set up before using SetupHTTPRequest().
    Checks the response of the HTTP GET request.
    \param timeout The timeout in seconds of the GET request.
    \param httprspcode The HTTP response code is returned here.
    \param contentlength The HTTP content length field is returned here.
    \return true if the response was correct, false if it was not.*/
bool CModemBG96::StartHTTPGET(uint32_t timeout, uint32_t* httprspcode, uint32_t* contentlength)
{
  //start the HTTP GET request
  if (m_atcmd.ATQHTTPGET(timeout) != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: HTTP GET request:\n");
    m_atcmd.PrintError();
    return false;
  }

  const char* line;
  uint32_t    linelength;
  line = m_atcmd.ReplyLine(0, &linelength);

  //parse the modem response of the GET request
  uint32_t err;
  int numparms = sscanf(line, "+QHTTPGET: %" SCNu32 ",%" SCNu32 ",%" SCNu32, &err, httprspcode, contentlength);
  if (numparms != 3)
  {
    PrintLogDecode(LOGMODEM, "ERROR: missing content length in AT+QHTTPGET: \"%.*s\"", (int)linelength, line);
    return false;
  }

  if (err != 0)
  {
    PrintLogDecode(LOGMODEM, "ERROR: %" PRIu32 " AT+QHTTPGET: \"%.*s\"", err, (int)linelength, line);
    return false;
  }

  PrintLog(LOGMODEM, "HTTP response: %" PRIu32 ", content-length: %" PRIu32 "\n", *httprspcode, *contentlength);
  return true;
}

/*! Sets up everything for the HTTP POST, and starts the request.
    \param url The URL of the GET request.
    \return true if starting the HTTP GET succeeded, false if it failed.*/
bool CModemBG96::StartHTTPPOST(uint32_t size, uint32_t timeout)
{
  //start the HTTP POST request
  if (m_atcmd.ATQHTTPPOST(size, timeout, timeout) != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: HTTP POST request:\n");
    m_atcmd.PrintError();
    return false;
  }

  return true;
}

bool CModemBG96::FinishHTTPPOST(uint32_t timeout, uint32_t* httprspcode, uint32_t* contentlength)
{
  //wait for the end of the AT+QHTTPPOST command
  if (m_atcmd.ATQHTTPPOST_Finish(timeout) != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to finish upload:\n");
    m_atcmd.PrintError();
    return false;
  }

  const char* line;
  uint32_t    linelength;
  line = m_atcmd.ReplyLine(0, &linelength);

  //parse the modem response of the POST request
  uint32_t err;
  int numparms = sscanf(line, "+QHTTPPOST: %" SCNu32 ",%" SCNu32 ",%" SCNu32, &err, httprspcode, contentlength);
  if (numparms != 3)
  {
    PrintLogDecode(LOGMODEM, "ERROR: missing content length in AT+QHTTPPOST: \"%.*s\"", (int)linelength, line);
    return false;
  }

  if (err != 0)
  {
    PrintLogDecode(LOGMODEM, "ERROR: AT+QHTTPPOST return error %" PRIu32 ": \"%.*s\"", err, (int)linelength, line);
    return false;
  }

  PrintLog(LOGMODEM, "HTTP response: %" PRIu32 ", content-length: %" PRIu32 "\n", *httprspcode, *contentlength);
  return true;
}

/*! Reads data from httppost->Upload(), and writes it to the usart.
    StartHTTPPOST() should have been called before this.
    \param url The url where the POST request is done.
    \param timeout Timeout of each upload block in seconds.
    \param httppost A pointer to a class derived from CHTTPPost.
    \param logpercentages When true, logs every 10% uploaded, with the speed in kilobyte per second.
    \param size The number of bytes to upload.
    \return true if the upload succeeded, false if anything failed.*/
bool CModemBG96::HTTPUpload(const char* url, uint32_t timeout, CHTTPPost* httppost,
                                bool logpercentages, uint32_t size)
{
  httppost->StartUpload(url);

  //use the buffer that m_atcmd has allocated for reads,
  //this saves ram by not having to allocate an extra buffer
  uint8_t* buf     = m_atcmd.GetReadBuf();
  uint32_t bufsize = m_atcmd.GetReadBufSize();

  uint32_t start        = Millis();
  uint32_t byteswritten = 0;
  uint32_t logcompleted = 0;
  uint32_t bytesspeed   = 0;
  uint32_t timespeed    = start;
  while (byteswritten < size)
  {
    //read a block from the HTTP POST uploader
    uint32_t readsize = std::min(bufsize, size - byteswritten);
    uint32_t writesize = httppost->Upload(buf, readsize);
    assert(writesize <= bufsize);
    if (writesize == 0)
    {
      PrintLog(LOGMODEM, "ERROR: read 0 bytes from http uploader\n");
      return false;
    }

    //write the block to the usart, the modem will send it as POST data to the server
    if (m_atcmd.WriteUsart(buf, writesize, timeout) != ATCmdResult_Success)
    {
      //if there is a problem uploading data to the modem,
      //the modem should be reset, because it's unknown what state it's in
      m_needsreset = true;

      PrintLogDecode(LOGMODEM, "ERROR: uploading %s:", url);
      m_atcmd.PrintError();
      return false;
    }

    byteswritten += writesize;

    //if it's requested to output the percentage uploaded to the log,
    //calculate how far it's along from 1 to 10, and print only when it changes
    if (logpercentages)
    {
      uint32_t completed = byteswritten * 10 / size;
      if (completed != logcompleted)
      {
        logcompleted = completed;

        //calculate how fast this 10% was uploaded in kilobytes per second
        uint32_t now = Millis();
        float elapsed = (float)(now - timespeed) / 1000.0f;
        if (elapsed > 0.1f)
        { //only print the speed when meaningful
          float speed = ((float)(byteswritten - bytesspeed) / elapsed) / 1024.0f;
          PrintLog(LOGMODEM, "Upload %" PRIu32 "%% complete, speed:%6.2f kilobyte/s\n", logcompleted * 10, speed);
        }
        else
        {
          PrintLog(LOGMODEM, "Upload %" PRIu32 "%% complete\n", logcompleted * 10);
        }

        timespeed = now;
        bytesspeed = byteswritten;

      }
    }
  }

  float secs = (float)(Millis() - start) / 1000.0f;
  if (secs > 0.1f) //only print the speed when meaningful
    PrintLog(LOGMODEM, "Uploaded %" PRIu32 " bytes in %.3f second%s, speed:%6.2f kilobyte/s\n",
             size, secs, secs >= 2.0f ? "s" : "", (float)size / secs / 1024.0f);
  else
    PrintLog(LOGMODEM, "Uploaded %" PRIu32 " bytes in %.3f second%s\n", size, secs, secs >= 2.0f ? "s" : "");

  return true;
}

bool CModemBG96::HTTPDownload(const char* url, uint32_t timeout, CHTTPDownload* httpdownload, bool logpercentages,
                                 uint32_t httprspcode, uint32_t contentlength)
{
  //start the HTTP GET request
  if (m_atcmd.ATQHTTPREAD_Start(timeout) != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to read HTTP data:\n");
    m_atcmd.PrintError();
    return false;
  }

  //calculate at which block size the CHTTPDownload class will receive the downloaded HTTP data
  uint32_t blocksize = std::min(httpdownload->DownloadBlockSize(), (uint32_t)AT_BUFSIZE);

  //inform the CHTTPDownload class that a download is about to start
  httpdownload->DownloadBegin(url, httprspcode, contentlength);

  //read data from the modem, and pass it to the CHTTPDownload class in blocks
  uint32_t start = Millis();
  uint32_t bytesread = 0;
  uint32_t logcompleted = 0;
  uint32_t bytesspeed = 0;
  uint32_t timespeed = start;
  while (bytesread < contentlength)
  {
    //read a block from the usart
    uint32_t readsize = std::min(blocksize, contentlength - bytesread);
    if (m_atcmd.ReadUsart(readsize, timeout) != ATCmdResult_Success)
    {
      //if there is a problem reading the data from the modem,
      //the modem should be reset, because it's unknown if the download
      //has stopped, or is just really slow and more data might come in
      m_needsreset = true;

      PrintLogDecode(LOGMODEM, "ERROR: downloading %s:", url);
      m_atcmd.PrintError();
      return false;
    }
    bytesread += m_atcmd.ReplyLen();

    //send the block to the CHTTPDownload class for further processing
    httpdownload->Download((uint8_t*)m_atcmd.Reply(), m_atcmd.ReplyLen());

    //if it's requested to output the percentage downloaded to the log,
    //calculate how far it's along from 1 to 10, and print only when it changes
    if (logpercentages)
    {
      uint32_t completed = bytesread * 10 / contentlength;
      if (completed != logcompleted)
      {
        logcompleted = completed;

        //calculate how fast this 10% was downloaded in kilobytes per second
        uint32_t now = Millis();
        float elapsed = (float)(now - timespeed) / 1000.0f;
        if (elapsed > 0.1f)
        { //only print the speed when meaningful
          float speed = ((float)(bytesread - bytesspeed) / elapsed) / 1024.0f;
          PrintLog(LOGMODEM, "Download %" PRIu32 "%% complete, speed:%6.2f kilobyte/s\n", logcompleted * 10, speed);
        }
        else
        {
          PrintLog(LOGMODEM, "Download %" PRIu32 "%% complete\n", logcompleted * 10);
        }

        timespeed = now;
        bytesspeed = bytesread;

      }
    }
  }

  float secs = (float)(Millis() - start) / 1000.0f;
  if (secs > 0.1f) //only print the speed when meaningful
    PrintLog(LOGMODEM, "Downloaded %" PRIu32 " bytes in %.3f second%s, speed:%6.2f kilobyte/s\n",
             contentlength, secs, secs >= 2.0f ? "s" : "", (float)contentlength / secs / 1024.0f);
  else
    PrintLog(LOGMODEM, "Downloaded %" PRIu32 " bytes in %.3f second%s\n",
             contentlength, secs, secs >= 2.0f ? "s" : "");

  return true;
}

/*! Waits for the finish of the AT+QHTTPREAD command.
    \return true if the command was successfully finished, false if it was not.*/
bool CModemBG96::FinishHTTPDownload(uint32_t* httpresult)
{
  //wait for the finish of the AT+QHTTPREAD command
  if (m_atcmd.ATQHTTPREAD_Finish() != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to finish download:\n");
    m_atcmd.PrintError();
    return false;
  }

  //check the result code of the AT+QHTTPREAD command
  uint32_t numparms = sscanf(m_atcmd.Reply(), "%" SCNu32, httpresult);
  if (numparms != 1)
  {
    PrintLogDecode(LOGMODEM, "ERROR: wrong HTTP download result: \"%s\"", m_atcmd.Reply());
    return false;
  }
  else if (*httpresult != 0)
  {
    PrintLogDecode(LOGMODEM, "ERROR: HTTP download ended with result %" PRIu32 "\n", *httpresult);
    return false;
  }

  return true;
}

/*! Reads the RSSI from the modem and stores it in m_rssi.
    \return true if reading the RSSI from the modem succeeded.
    \return false if reading the RSSI from the modem failed.*/
bool CModemBG96::UpdateRSSI()
{
  if (!m_activated)
  {
    PrintLog(LOGMODEM, "ERROR: unable to get signal quality report,"
             " modem not activated\n");
    return false;
  }

  //get the signal quality
  if (m_atcmd.ATCSQ() != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to get signal quality report:\n");
    m_atcmd.PrintError();
    return false;
  }

  uint32_t    linelength;
  const char* line = m_atcmd.ReplyLine(0, &linelength);

  //scan the reply from the AT+CSQ command
  int numparms = sscanf(line, "+CSQ: %" SCNi32 ",%" SCNu32, &m_rssi, &m_ber);

  if (numparms != 2)
  {
    PrintLogDecode(LOGMODEM, "ERROR: invalid signal quality string: \"%.*s\"",
                   (int)linelength, line);
    m_rssi = 0;
    m_ber = 0;
    return false;
  }

  return true;
}

/*! Enable GNSS positioning using the AT+QGPS command.
    GNSS positioning should be disabled after getting a GNSS fix, otherwise
    no data communication is possible on the Quectel BG95.
    \return true if the command succeeded, false if it failed.*/
bool CModemBG96::EnableGNSS()
{
  PrintLog(LOGMODEM, "Enabling GNSS\n");

  if (m_type == ModemType_BG96)
  {
    uint32_t gnssmode   = 1;  //Stand-alone.
    uint32_t fixmaxtime = 30; //Default.
    uint32_t fixmaxdist = 50; //Default.
    uint32_t fixcount   = 0;  //Continous positioning.
    uint32_t fixrate    = 1;  //Default.

    EATCmdResult result = m_atcmd.ATQGPS_Set_BG96(gnssmode, &fixmaxtime,
                                                  &fixmaxdist, &fixcount, &fixrate);
    if (result != ATCmdResult_Success)
    {
      PrintLog(LOGMODEM, "ERROR: unable to enable GNSS:\n");
      m_atcmd.PrintError();
      return false;
    }
  }
  else if (m_type == ModemType_BG95)
  {
    uint32_t gnssmode = 1; //Stand-alone.
    uint32_t accuracy = 3; //High accuracy.
    uint32_t fixcount = 0; //Continuous positioning.
    uint32_t fixrate  = 1; //Interval between positioning.
    uint32_t hepe     = 0; //No accuracy threshold.

    //Enable the GNSS receiver. The Quectel BG95 defaults to GNSS priority mode,
    //which means that no data communication is possible while the GNSS receiver
    //is active.
    EATCmdResult result;
    result = m_atcmd.ATQGPS_Set_BG95(gnssmode, &accuracy,
                                     &fixcount, &fixrate, &hepe);
    if (result != ATCmdResult_Success)
    {
      PrintLog(LOGMODEM, "ERROR: unable to enable GNSS:\n");
      m_atcmd.PrintError();
      return false;
    }
  }
  else
  {
    PrintLog(LOGMODEM, "ERROR: unknown modem type %i, can't enable GNSS\n",
             (int)m_type);

    return false;
  }

  return true;
}

/*! Disables GNSS using the AT+QGPSEND command.
    \return true Disabling GNSS succeeded.
    \return false Disabling GNSS failed.*/
bool CModemBG96::DisableGNSS()
{
  PrintLog(LOGMODEM, "Disabling GNSS\n");

  EATCmdResult result = m_atcmd.ATQGPSEND();

  if (result != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to disable GNSS\n");
    m_atcmd.PrintError();
    return false;
  }

  //Give the modem some time to disable GNSS, and resume normal data transmission.
  DelayMS(1000);

  return true;
}

/*! Upper limit of the GNSS fix time in seconds.*/
#define LOCATION_MAX_FIX_TIME (60 * 60)

/*! Uses the AT+QGPSLOC command to read the location from the GNSS receiver of the modem.
    \param fixtime Max time in seconds waiting for the GNSS location to become valid.
    \param latitude The GNSS latitude is returned here.
    \pararm longitude The GNSS longitude is returned here.
    \return true If there was a new location returned by the modem.
    \return false If the modem did not return a location.*/
bool CModemBG96::GetLocation(uint32_t fixtime, double* latitude, double* longitude)
{
  if (!m_activated)
  {
    PrintLog(LOGMODEM, "ERROR: unable to get GNSS location,"
             " modem not activated\n");
    return false;
  }

  //Enable GNSS positioning. This should only be enabled when a location fix
  //is needed, and should be disabled right after the fix, because the Quectel
  //BG95 can't do data communication when the GNSS receiver is active.
  if (!EnableGNSS())
  {
    DisableGNSS();
    return false;
  }

  *latitude  = 0.0;
  *longitude = 0.0;

  fixtime = std::min(fixtime, (uint32_t)LOCATION_MAX_FIX_TIME);

  PrintLog(LOGMODEM, "Waiting for GNSS fix for %" PRIu32 " seconds\n", fixtime);

  EATCmdResult result;

  int64_t start = Millis();
  do
  {
    //Get the GNSS coordinates in decimal degrees.
    if ((result = m_atcmd.ATQGPSLOC(2)) == ATCmdResult_Success)
    {
      //After a successful GNSS location fix, wait some time in case
      //data from more satellites is received.
      DelayMS(10000);

      //Read the coordinates again.
      if ((result = m_atcmd.ATQGPSLOC(2)) == ATCmdResult_Success)
      {
        const char* line;
        uint32_t    linelength;
        line = m_atcmd.ReplyLine(0, &linelength);

        //Parse the coordinates, the output from the modem is:
        //UTC,latitude,longitude
        double utc = 0.0;
        int numparms = sscanf(line, "+QGPSLOC: %lf,%lf,%lf",
                              &utc, latitude, longitude);
        if (numparms == 3)
        {
          PrintLog(LOGMODEM, "GNSS location: %f,%f\n", *latitude, *longitude);
          DisableGNSS(); //The BG95 can't do data communication with GNSS active.
          return true;
        }
        else
        {
          PrintLogDecode(LOGMODEM, "ERROR: wrong GNSS information: %.*s",
                         (int)linelength, line);
        }
      }
    }

    DelayMS(1000);
  }
  while (((Millis() - start) / 1000) < fixtime);

  if (result != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to get GNSS location:\n");
    m_atcmd.PrintError();
  }
  else
  {
    PrintLog(LOGMODEM, "ERROR: timeout getting GNSS location\n");
  }

  DisableGNSS(); //The BG95 can't do data communication with GNSS active.

  return false;
}

/*! Uses the AT+QGDCNT? to read from the modem how many
    bytes are sent and received.
    \param tx The number of transmitted bytes is returned here.
    \param rx The number of received bytes is returned here.
    \return true if reading the number of bytes from the modem succeeded.
    \return false if reading the number of bytes from the modem failed.*/
bool CModemBG96::GetBytesTransferred(uint64_t* tx, uint64_t* rx)
{
  *tx = 0;
  *rx = 0;

  if (!m_activated)
  {
    PrintLog(LOGMODEM, "ERROR: unable to get packet data counter,"
             " modem not activated\n");
    return false;
  }

  //Get the tx and rx byte counters.
  if (m_atcmd.ATQGDCNT_Query() != ATCmdResult_Success)
  {
    PrintLog(LOGMODEM, "ERROR: unable to get packet data counter:\n");
    m_atcmd.PrintError();
    return false;
  }

  uint32_t    linelength;
  const char* line = m_atcmd.ReplyLine(0, &linelength);

  //scan the reply from the AT+QGDCNT command
  int numparms = sscanf(line, "+QGDCNT: %" SCNu64 ",%" SCNu64, tx, rx);

  if (numparms != 2)
  {
    PrintLogDecode(LOGMODEM, "ERROR: invalid packet data counter string: \"%.*s\"",
                   (int)linelength, line);
    *tx = 0;
    *rx = 0;
    return false;
  }

  return true;
}
