#include "oleddisplay.h"
#include "log.h"
#include "lock.h"

#include <cstring>
#include <cassert>

#include "cp437font8x8.h"

COLEDDisplay::COLEDDisplay()
{
  m_i2c     = nullptr;
  m_working = false;
  m_addr    = 0;
}

COLEDDisplay::~COLEDDisplay()
{
}

/*! Sets up an OLED display with a SH1106 display driver using I2C.
    \param i2c a pointer to a CI2C instance, which has been set up to use an I2C peripheral.
    \return true if setting up the display succeeded, false if it failed.*/
bool COLEDDisplay::Setup(CI2C* i2c)
{
  m_i2c     = i2c;
  m_working = false;

  if (!FindAddr())
    return false;

  if (!Init())
    return false;

  return true;
}

/*! Tries to write to every address on the I2C bus,
    if an address is acked it will use that as the display address.
    \return true if the display address was found, false if it was not.*/
bool COLEDDisplay::FindAddr()
{
  CLock lock(*m_i2c);

  for (uint32_t i = 0; i < 256; i += 2)
  {
    if (m_i2c->Write(i, nullptr, 0))
    {
      m_addr = i;
      m_i2c->Stop();
      PrintLog(LOGOLEDDPY, "Found display at address %" PRIX32 "\n", m_addr);
      return true;
    }
  }
  m_i2c->Stop();

  PrintLog(LOGOLEDDPY, "ERROR: Could not detect display address\n");
  return false;
}

/*! Writes buf using m_i2c to address m_addr, and checks the return value of m_i2c.Write(),
    if it's false, it will print which byte wasn't acked, and it will return false after that.*/
#define W(buf)\
  do\
  {\
    if (!m_i2c->Write(m_addr, buf))\
    {\
      PrintLog(LOGOLEDDPY, "ERROR: Writing %" PRIu32 " bytes to address %" PRIu32 " failed to ack at byte %" PRIu32 "\n",\
               (uint32_t)sizeof(buf), m_addr, m_i2c->AckFail());\
      m_working = false;\
      m_i2c->Stop();\
      return false;\
    }\
  }\
  while(false)

/*! Initializes an OLED display with the SH1106 display driver.
    \return true if initializing the OLED display succeeded.
    \return false if initializing the OLED display failed.*/
bool COLEDDisplay::Init()
{
  PrintLog(LOGOLEDDPY, "Initializing display\n");

  CLock lock(*m_i2c);

  //this initialization sequence is copied from the u8g2 arduino library
  //it appears that the protocol for the SH1106 driver is not public,
  //but it's almost the same as the SSD1306
  W(((const uint8_t[]){0x00, 0xAE})); //turn the display off
  W(((const uint8_t[]){0x00, 0xD5})); //clock divide ratio
  W(((const uint8_t[]){0x00, 0x80})); //oscillator frequency
  W(((const uint8_t[]){0x00, 0xA8})); //multiplex ratio
  W(((const uint8_t[]){0x00, 0x3F})); //multiplex ratio
  W(((const uint8_t[]){0x00, 0xD3})); //display offset
  W(((const uint8_t[]){0x00, 0x00})); //display offset
  W(((const uint8_t[]){0x00, 0x40})); //display start line 0
  W(((const uint8_t[]){0x00, 0x20})); //page addressing mode
  W(((const uint8_t[]){0x00, 0x02})); //page addressing mode
  W(((const uint8_t[]){0x00, 0xA1})); //segment remap a0/a1
  W(((const uint8_t[]){0x00, 0xC8})); //c0: scan dir normal, c8: reverse
  W(((const uint8_t[]){0x00, 0xDA})); //com pin HW config, sequential com pin config (bit 4), disable left/right remap (bit 5)
  W(((const uint8_t[]){0x00, 0x12})); //com pin HW config, sequential com pin config (bit 4), disable left/right remap (bit 5)
  W(((const uint8_t[]){0x00, 0x81})); //[2] set contrast control
  W(((const uint8_t[]){0x00, 0xCF})); //[2] set contrast control
  W(((const uint8_t[]){0x00, 0xD9})); //[2] pre-charge period 0x022/f1
  W(((const uint8_t[]){0x00, 0xF1})); //[2] pre-charge period 0x022/f1
  W(((const uint8_t[]){0x00, 0xDB})); //vcomh deselect level
  W(((const uint8_t[]){0x00, 0x40})); //vcomh deselect level
  W(((const uint8_t[]){0x00, 0x2E})); //Deactivate scroll
  W(((const uint8_t[]){0x00, 0xA4})); //output ram to display
  W(((const uint8_t[]){0x00, 0xA6})); //none inverted normal display mode

  //set m_working to true here so that the Draw() method works
  m_working = true;

  //the pixels are at a random state at start,
  //set them all to black
  Clear();
  Draw();

  W(((const uint8_t[]){0x00, 0xAF})); //turn the display on

  m_i2c->Stop();

  return true;
}

void COLEDDisplay::Clear()
{
  memset(m_buf, 0, sizeof(m_buf));
}

/*! Draws m_buf onto the OLED display.
    \return true if drawing m_buf onto the OLED display succeeded.
    \return false if drawing m_buf onto the OLED display failed.*/
bool COLEDDisplay::Draw()
{
  if (!m_working)
    return false;

  CLock lock(*m_i2c);

  //since each byte is 8 pixels high, and the display has 64 lines,
  //this divides the display into 8 groups of 8 lines
  for (uint32_t y = 0; y < OLED_DPY_HEIGHT / 8; y++)
  {
    W(((const uint8_t[]){0x00, 0x40}));     //set line offset to 0
    W(((const uint8_t[]){0x00, 0x10}));
    W(((const uint8_t[]){0x00, 0x02}));     //set column address
    W(((const uint8_t[]){0x00, (uint8_t)(0xB0 + y)}));

    //the display is 128 columns wide, and 16 byte can be written in one call
    //so write 8 times 16 bytes for 128 columns in total, where each byte is a vertical line of 8 pixels
    for (uint32_t x = 0; x < OLED_DPY_WIDTH / 16; x++)
    {
      if (!m_i2c->Write(m_addr, (const uint8_t[]){0x40}, 1, m_buf + y * 128 + x * 16, 16))
      {
        PrintLog(LOGOLEDDPY, "ERROR: Writing %i bytes to address %" PRIu32 " failed to ack at byte %" PRIu32 "\n",\
                 16, m_addr, m_i2c->AckFail());
        m_working = false;
        m_i2c->Stop();
        return false;
      }
    }
  }

  m_i2c->Stop();

  return true;
}

/*! Draws a line of ASCII text on m_buf, which can then be
    shown on the OLED display using the Draw() method.
    \param text The ascii text to draw.
    \param line The line number to draw on, 0 is the top line,
                OLED_DPY_LINES - 1 is the bottom line.*/
void COLEDDisplay::DrawText(const char* text, uint32_t line)
{
  assert(line < OLED_DPY_LINES);

  //limit to the maximum number of characters per line
  uint32_t len = strnlen(text, OLED_DPY_LINESIZE);

  //create a line of text padded with spaces up to the full display size
  char displayline[OLED_DPY_LINESIZE];
  memcpy(displayline, text, len);
  if (len < OLED_DPY_LINESIZE)
    memset(displayline + len, ' ', OLED_DPY_LINESIZE - len);

  //the SH1106 OLED driver has an internal buffer, where one byte is a column of 8 pixels,
  //this is used to draw lines of text, with a font that is 8 pixels high
  //the font is stored as 8 bytes per glyph, so to show one char on the display,
  //8 bytes should be copied front the font glyph into the buffer
  uint8_t* bufptr = m_buf + line * OLED_DPY_WIDTH;
  for (uint32_t i = 0; i < OLED_DPY_LINESIZE; i++)
  {
    uint8_t c = displayline[i];
    const uint8_t* glyph = cp437font8x8[c];
    memcpy(bufptr, glyph, OLED_DPY_CHAR_WIDTH);
    bufptr += OLED_DPY_CHAR_WIDTH;
  }
}
