#include "random.h"
#include "lock.h"
#include "sha256.h"
#include "backupregisters.h"
#include "aes.h"
#include "clock.h"

#include <stm32f4xx.h>

#include <FreeRTOS.h>
#include <semphr.h>

#include <cassert>
#include <cstdlib>
#include <cstring>
#include <algorithm>

static SemaphoreHandle_t g_rnglock;
static unsigned int g_seed;

/*! Initializes the random number generator.*/
void InitRNG()
{
  //enable the random number generator
  RCC_AHB2PeriphClockCmd(RCC_AHB2Periph_RNG, ENABLE);
  RNG_Cmd(ENABLE);

  //create a mutex for the random number generator
  g_rnglock = xSemaphoreCreateRecursiveMutex();

  //throw away the first random number, it should not be used
  GetRandomNumber();

  //initialize the seed for rand_r()
  g_seed = GetRandomNumber();
}

/*! Gets a random number from the random number generator.*/
uint32_t GetRandomNumber()
{
  assert(g_rnglock);
  CLock lock(g_rnglock);

  //on an error, clear the error flag and reset the random number generator
  if (RNG_GetFlagStatus(RNG_FLAG_CECS) || RNG_GetFlagStatus(RNG_FLAG_SECS))
  {
    RNG_ClearFlag(RNG_FLAG_CECS);
    RNG_ClearFlag(RNG_FLAG_SECS);
    RNG_Cmd(DISABLE);
    DelayUS(1000);
    RNG_Cmd(ENABLE);
  }

  while (!RNG_GetFlagStatus(RNG_FLAG_DRDY)); //wait until there is a random number available
  return RNG_GetRandomNumber();              //return a random number from the random number generator
}

/*! Returns a value generated by rand_r().*/
uint32_t GetPseudoRandomNumber()
{
  CCritical criticalsection;
  return rand_r(&g_seed);
}

/*! Fills iv with random bytes which can be used as initialization vector for AES-CBC.
    \param iv A pointer to a buffer where the result is written.
    \param size The size of iv. Should not be larger than SHA256_BLOCK_SIZE.*/
void CreateIV(uint8_t* iv, uint32_t size /*= AES_BLOCK_SIZE*/)
{
  //create a SHA256 from several sources to use as IV
  SHA256_CTX sha256ctx;
  WSS_sha256_init(&sha256ctx);

  //add the seed from the battery backed ram that was generated the previous time
  uint32_t seed = GetRandomSeed();
  WSS_sha256_update(&sha256ctx, (uint8_t*)&seed, sizeof(seed));

  //add the current date as a unix timestamp
  int64_t secs = GetRTCSeconds(nullptr);
  WSS_sha256_update(&sha256ctx, (uint8_t*)&secs, sizeof(secs));

  //add the current microsecond clock
  uint32_t microseconds = Micros();
  WSS_sha256_update(&sha256ctx, (uint8_t*)&microseconds, sizeof(microseconds));

  //add the current millisecond clock
  int64_t milliseconds = Millis();
  WSS_sha256_update(&sha256ctx, (uint8_t*)&milliseconds, sizeof(milliseconds));

  //add input from the hardware random number generator
  for (uint32_t i = 0; i < 32; i++)
  {
    uint32_t random = GetRandomNumber();
    WSS_sha256_update(&sha256ctx, (uint8_t*)&random, sizeof(random));
  }

  //store the new SHA256
  uint8_t hash[SHA256_BLOCK_SIZE];
  WSS_sha256_final(&sha256ctx, hash);

  //use the last 4 bytes of the SHA256 as seed for the next time
  seed = ((uint32_t)hash[SHA256_BLOCK_SIZE - 1] << 24) | ((uint32_t)hash[SHA256_BLOCK_SIZE - 2] << 16) |
         ((uint32_t)hash[SHA256_BLOCK_SIZE - 3] << 8)  | ((uint32_t)hash[SHA256_BLOCK_SIZE - 4]);
  SetRandomSeed(seed);

  //Use the SHA256 result as initialization vector.
  memcpy(iv, hash, std::min((uint32_t)SHA256_BLOCK_SIZE, size));
}
