#include "rfm9xw.h"
#include "clock.h"
#include "log.h"

#include <cmath>

CRFM9xW::CRFM9xW()
{
  ResetState();
}

CRFM9xW::~CRFM9xW()
{
}

/*! Initializes the SPI peripheral, and checks if the RFM9xW is working.
    \param devicestr A c-string that describes the device, for the log printing.
    \param spi The SPI peripheral to use.
    \param nss The slave select pin.
    \param rst The reset pin.
    \param dio0 The DIO0 pin.
    \param irqhandler A pointer to a class derived from CRFM9xWInterrupt, on which the CRFM9xWInterruptHandler()
                      method is called on a rising edge of DIO0.*/
bool CRFM9xW::Setup(const char* devicestr, SPI_TypeDef* spi, SGPIOPin nss, SGPIOPin rst, SGPIOPin dio0, CRFM9xWInterrupt* irqhandler)
{
  m_activated  = true;
  m_devicestr  = devicestr;
  m_nss        = nss;
  m_rst        = rst;
  m_dio0       = dio0;
  m_irqhandler = irqhandler;

  Deselect();

  //set up the SPI peripheral in master mode 0
  m_spi.Setup(spi, SPI_BaudRatePrescaler_16);

  //reset the RFM9xW
  GPIO_Reset(m_rst);
  DelayMS(10);
  GPIO_Set(m_rst);
  DelayMS(10);

  //check if the RFM9xW is working by checking the version register
  if (!CheckVersion())
  {
    Close();
    return false;
  }

  //initialize the timestamp for periodic checking
  m_lastcheck = Micros();

  //store which frequency the RFM9xW is set at
  m_frequency = GetFrequency();

  return true;
}

void CRFM9xW::Close()
{
  if (m_activated)
  {
    //put the RFM9xW into the reset state
    GPIO_Reset(m_rst);

    Deselect();
    m_spi.Close();

    DisableDIO0Interrupt();

    ResetState();
  }
}

/* Resets all state variables.*/
void CRFM9xW::ResetState()
{
  m_activated  = false;
  m_devicestr  = nullptr;
  m_nss        = {};
  m_rst        = {};
  m_dio0       = {};
  m_irqhandler = nullptr;
  m_frequency  = 0;
}

#define NSS_DELAY 10 //!< 10 microseconds delay

/*! Makes the NSS pin low, wrapped in a delay. */
void CRFM9xW::Select()
{
  assert(m_activated);

  DelayUS(NSS_DELAY);
  GPIO_Reset(m_nss);
  DelayUS(NSS_DELAY);
}

/*! Makes the NSS pin high, wrapped in a delay. */
void CRFM9xW::Deselect()
{
  assert(m_activated);

  DelayUS(NSS_DELAY);
  GPIO_Set(m_nss);
  DelayUS(NSS_DELAY);
}

/*! Reads a register from the RFM9xW.*/
uint8_t CRFM9xW::ReadRegister(uint8_t address)
{
  assert(m_activated);

  uint8_t output[2] = {(uint8_t)(address & 0x7f), 0x00};
  uint8_t input[2]  = {};

  Select();
  m_spi.Transceive(output, input, 2);
  Deselect();

  return input[1];
}

/*! Writes a register of the RFM9xW.
    \param address The register address in the RFM9xW to write to.
    \param value The value to write to address.
    \param check A value of @ref ECheckWrittenValue, when set to CheckWrittenValue it will
    read back the register after the write, to make sure that the correct value has been written to the register.*/
bool CRFM9xW::WriteRegister(uint8_t address, uint8_t value, ECheckWrittenValue check /*= CheckWrittenValue*/)
{
  assert(m_activated);

  uint8_t output[2] = {(uint8_t)(address | WRITE_CMD), value};
  uint8_t input[2];

  Select();
  m_spi.Transceive(output, input, 2);
  Deselect();

  if (check == CheckWrittenValue)
  {
    uint8_t readvalue = ReadRegister(address);
    if (readvalue != value)
    {
      PrintLog(LOGRFM9xW, "ERROR: %s: wrote value %02" PRIX8 " to register %02" PRIX8 " but read value %02" PRIX8 " back\n",
               m_devicestr, value, address, readvalue);
      return false;
    }
  }

  return true;
}

/*! Reads the version register, and check if it has the correct value.
    \return true if the value is correct, false if it's not.*/
bool CRFM9xW::CheckVersion()
{
  assert(m_activated);

  uint8_t versionreg = ReadRegister(REG_VERSION);

  if (versionreg == RFM9xW_VERSION)
  {
    return true;
  }
  else
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: version register has value %" PRIX8 ", but should be %" PRIX8 "\n",
             m_devicestr, versionreg, RFM9xW_VERSION);
    return false;
  }
}

//! How many microseconds between checks.
#define CHECK_PERIOD (60 * 1000000)

/* Periodically checks if the version register of the RFM9xW returns the correct value,
   and that the mode is set correctly.
   If that is not the case, there is something wrong with the RFM9xW, it should be re-initialized.
   \param mode The mode that the RFM9xW should be in, use ModeRxCont for LoRa, and ModeRx for OOK.
   \return true if the RFM9xW is working correctly, false if it's not.*/
bool CRFM9xW::CheckOperation(EMode mode)
{
  assert(m_activated);

  uint32_t now = Micros();

  if (now - m_lastcheck < CHECK_PERIOD)
    return true;

  m_lastcheck = now;

  if (!CheckVersion())
    return false;

  EMode currentmode = GetMode();
  if (currentmode != mode)
  {
    PrintLog(LOGRFM9xW, "ERROR: %s mode is %i, should be %i\n", m_devicestr, (int)currentmode, (int)mode);
    return false;
  }

  return true;
}

/*! Returns how many milliseconds to wait before doing the next check.
    A millisecond is always added, to make sure that when this time is waited for,
    the check always happens when the CheckOperation() method is called.*/
uint32_t CRFM9xW::NextCheckDelay()
{
  uint32_t elapsed = Micros() - m_lastcheck;
  if (elapsed < CHECK_PERIOD)
    return (CHECK_PERIOD - elapsed) / 1000 + 1;
  else
    return 1;
}

/*! Sets the mode of the RFM9xW base on the mode parameter.
    \param mode A value of @ref EMode.*/
bool CRFM9xW::SetMode(enum EMode mode)
{
  assert(m_activated);

  CHECK_MODE(mode);

  //read the current mode register
  uint8_t modereg = ReadRegister(REG_OP_MODE);

  //set the mode bits to 0, and apply the new mode
  modereg &= ~MODE_MASK;
  modereg |= (uint8_t)mode;

  //change the mode
  WriteRegister(REG_OP_MODE, modereg, NoCheckWrittenValue);

  //wait some time until the new mode is set
  uint32_t start = Micros();
  uint8_t  currentmodereg;
  do
  {
    currentmodereg = ReadRegister(REG_OP_MODE);
    if (currentmodereg == modereg)
      return true;

    DelayMS(10);
  }
  while (Micros() - start < 10000000);

  PrintLog(LOGRFM9xW, "ERROR: %s: unable to change mode register to %" PRIX8 ", mode register is %" PRIX8 "\n",
           m_devicestr, modereg, currentmodereg);

  return false;
}

/*! Returns the mode bits of the RegOpMode register, converted to an enum EMode.*/
EMode CRFM9xW::GetMode()
{
  assert(m_activated);

  uint8_t modereg = ReadRegister(REG_OP_MODE);
  return (EMode)(modereg & MODE_MASK);
}

/*! Sets the modulation to LoRa or OOK, by writing the opmode register.
    \param modulation The modulation that should be selected.
    \return true if setting the modulation succeeded, false if it failed.*/
bool CRFM9xW::SetModulation(EModulation modulation)
{
  assert(m_activated);

  //read the current mode register
  uint8_t modereg = ReadRegister(REG_OP_MODE);

  //set the correct bits based on the selected modulation
  if (modulation == Modulation_LoRa)
  {
    modereg |= OP_MODE_LORA;
    modereg &= ~(OP_MODE_MODULATION_OOK);
  }
  else if (modulation == Modulation_OOK)
  {
    modereg &= ~(OP_MODE_LORA);
    modereg |= OP_MODE_MODULATION_OOK;
  }
  else
  {
    assert(0);
  }

  //change the mode
  WriteRegister(REG_OP_MODE, modereg, NoCheckWrittenValue);

  //check if the mode has been set correctly
  uint8_t mask = OP_MODE_LORA | OP_MODE_MODULATION_OOK;
  uint8_t currentmodereg = ReadRegister(REG_OP_MODE);
  if ((modereg & mask) != (currentmodereg & mask))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to change mode register to %" PRIX8 ", mode register is %" PRIX8 "\n",
             m_devicestr, modereg, currentmodereg);
    return false;
  }

  return true;
}

/*! 32 MHz crystal oscillator of the RFM9xW divided by 2^19,
    used for translating between the RF frequency and the RF frequency register.*/
#define FREQ_STEP 61.03515625

/*! Sets the RFM9xW RF frequency.
    \param frequency The RF Frequency in hertz.*/
bool CRFM9xW::SetFrequency(uint32_t frequency)
{
  assert(m_activated);

  //the value for the RegFrf register should be the frequency divided by the crystal oscillator frequency divided by 2^19.
  uint32_t regfrf = lround((double)frequency / FREQ_STEP);

  //write the new frequency, in order of most to least significant byte, because
  //a change in frequency happens only after the least significant byte is written
  if (!WriteRegister(REG_FRF_MSB, (regfrf >> 16) & 0xFF) ||
      !WriteRegister(REG_FRF_MID, (regfrf >>  8) & 0xFF) ||
      !WriteRegister(REG_FRF_LSB, (regfrf >>  0) & 0xFF))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to set frequency to %" PRIu32 "\n", m_devicestr, frequency);
    return false;
  }

  //low frequency mode should be enabled for frequencies below 525 MHz
  //low frequency mode should be disabled for frequencies above 779 MHz
  if (!SetLowFrequencyMode(frequency < 652000000))
    return false;

  PrintLog(LOGRFM9xW, "%s frequency set to %f MHz\n", m_devicestr, 
           (double)lround((double)regfrf * FREQ_STEP) / 1000000.0);

  //store the frequency for future RSSI calculations
  m_frequency = frequency;

  return true;
}

/* Reads the frequency registers, and calculates which frequency the RFM9xW is set to.
   \return The frequency the RFM9xW is set to.*/
uint32_t CRFM9xW::GetFrequency()
{
  assert(m_activated);

  uint32_t regfreq = (((uint32_t)ReadRegister(REG_FRF_MSB)) << 16) |
                     (((uint32_t)ReadRegister(REG_FRF_MID)) << 8)  |
                      ((uint32_t)ReadRegister(REG_FRF_LSB));

  uint32_t frequency = lround((double)regfreq * FREQ_STEP);

  return frequency;
}

/*! Sets low frequency mode in the mode register.
    It will not write the mode register if the low frequency mode already has the correct value.
    \param on true to set low frequency mode on, false to turn it off.
    \param true if setting the mode succeeded, false if it failed.*/
bool CRFM9xW::SetLowFrequencyMode(bool on)
{
  assert(m_activated);

  //read the current mode register
  uint8_t modereg = ReadRegister(REG_OP_MODE);

  //if the low frequency mode is already at the requested mode,
  //don't do anything
  if (on == !!(modereg & OP_MODE_LOW_FREQUENCY_MODE_ON))
    return true;

  //set or clear the low frequency mode bit
  if (on)
    modereg |= OP_MODE_LOW_FREQUENCY_MODE_ON;
  else
    modereg &= ~(OP_MODE_LOW_FREQUENCY_MODE_ON);

  //change the mode
  WriteRegister(REG_OP_MODE, modereg, NoCheckWrittenValue);

  //check if the mode has been set correctly
  uint8_t mask = OP_MODE_LOW_FREQUENCY_MODE_ON;
  uint8_t currentmodereg = ReadRegister(REG_OP_MODE);
  if ((modereg & mask) != (currentmodereg & mask))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to turn low frequency mode %s, tried to set mode register to %" PRIX8
                        ", mode register is %" PRIX8 "\n", m_devicestr, on ? "on" : "off", modereg, currentmodereg);
    return false;
  }

  return true;
}

/*! Reads the low frequency mode bit in the RegOpMode register.
    \return true if the low frequency mode bit is set, false if it's cleared.*/
bool CRFM9xW::GetLowFrequencyMode()
{
  assert(m_activated);

  //read the current mode register
  uint8_t modereg = ReadRegister(REG_OP_MODE);

  //return if the low frequency bit is on or off
  return (modereg & OP_MODE_LOW_FREQUENCY_MODE_ON) == OP_MODE_LOW_FREQUENCY_MODE_ON;
}

/*! Sets the transmit power.
    \param dbm The transmit power in dBm, ranges from 2 to 17.*/
bool CRFM9xW::SetTransmitPower(uint32_t dbm)
{
  assert(m_activated);

  assert(dbm >= 2 && dbm <= 17);
  uint8_t regpaconfig = PA_BOOST | (dbm - 2);
  if (!WriteRegister(REG_PA_CONFIG, regpaconfig))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to set transmit power\n", m_devicestr);
    return false;
  }

  return true;
}

/*! Sets the the automatic gain control for the low noise amplifier.
    \param enable Whether to enable agc for the lna.*/
bool CRFM9xW::SetLoRaLnaAgc(bool enable)
{
  assert(m_activated);

  //the agc bit for LoRa is in the RegModemConfig3 register
  uint8_t regmodemconfig3 = ReadRegister(REG_MODEM_CONFIG_3);
  if (enable)
    regmodemconfig3 |= LORA_LNA_AGC_BIT;
  else
    regmodemconfig3 &= ~(LORA_LNA_AGC_BIT);

  if (!WriteRegister(REG_MODEM_CONFIG_3, regmodemconfig3))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to set LoRa AGC\n", m_devicestr);
    return false;
  }

  return true;
}

/*! Sets the RF bandwidth in LoRa mode.
    \param A value of @ref ELoRaSignalBandwidth.
    \return true if setting the bandwidth succeeded, false if it failed.*/
bool CRFM9xW::SetLoRaBandwidth(ELoRaSignalBandwidth bandwidth)
{
  assert(m_activated);

  CHECK_BANDWIDTH(bandwidth);

  uint8_t regmodemconfig1 = ReadRegister(REG_MODEM_CONFIG_1);

  regmodemconfig1 &= ~(LORA_SIGNAL_BANDWIDTH_MASK);
  regmodemconfig1 |= (uint8_t)bandwidth;

  if (!WriteRegister(REG_MODEM_CONFIG_1, regmodemconfig1))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to set LoRa bandwidth\n", m_devicestr);
    return false;
  }

  return true;
}

/*! Sets the LoRa coding rate, LoRa uses this for forward error correction.
    \param codingrate A value of @ref ELoRaCodingRate.
    \return true if setting the coding rate succeeded, false if it failed.*/
bool CRFM9xW::SetLoRaCodingRate(ELoRaCodingRate codingrate)
{
  assert(m_activated);

  CHECK_CODING_RATE(codingrate);

  uint8_t regmodemconfig1 = ReadRegister(REG_MODEM_CONFIG_1);
  regmodemconfig1 &= ~(LORA_CODING_RATE_MASK);
  regmodemconfig1 |= (uint8_t)codingrate;

  if (!WriteRegister(REG_MODEM_CONFIG_1, regmodemconfig1))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to set LoRa coding rate\n", m_devicestr);
    return false;
  }

  return true;
}

/*! Sets the header mode, implicit or explicit, in implicit header mode, the header is not sent by the RFM9xW.
    \param headermode A value of @ref ELoRaHeaderMode.
    \return true if setting the header mode succeeded, false if it failed.*/
bool CRFM9xW::SetLoRaHeaderMode(ELoRaHeaderMode headermode)
{
  assert(m_activated);

  CHECK_HEADER_MODE(headermode);

  uint8_t regmodemconfig1 = ReadRegister(REG_MODEM_CONFIG_1);
  regmodemconfig1 &= ~(LORA_HEADER_MODE_MASK);
  regmodemconfig1 |= (uint8_t)headermode;

  if (!WriteRegister(REG_MODEM_CONFIG_1, regmodemconfig1))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to set LoRa header mode\n", m_devicestr);
    return false;
  }

  return true;
}

/*! Sets the crc mode, when it's on the RFM9xW will generate a crc on transmit, and it will check it on receive.
    \param crcmode A value of @ref ELoRaCrcMode.
    \return true if setting the crc mode succeeded, false if it failed.*/
bool CRFM9xW::SetLoRaCrcMode(ELoRaCrcMode crcmode)
{
  assert(m_activated);

  CHECK_CRC_MODE(crcmode);

  uint8_t regmodemconfig2 = ReadRegister(REG_MODEM_CONFIG_2);
  regmodemconfig2 &= ~(LORA_CRC_MODE_MASK);
  regmodemconfig2 |= (uint8_t)crcmode;

  if (!WriteRegister(REG_MODEM_CONFIG_2, regmodemconfig2))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to set LoRa crc mode\n", m_devicestr);
    return false;
  }

  return true;
}

/*! Sets the loRa spreading factor.
    \param spreadingfactor A value of @ref ELoRaSpreadingFactor.
    \return true if setting the spreading factor succeeded, false if it failed.*/
bool CRFM9xW::SetLoRaSpreadingFactor(ELoRaSpreadingFactor spreadingfactor)
{
  assert(m_activated);

  CHECK_SPREADING_FACTOR(spreadingfactor);

  uint8_t regmodemconfig2 = ReadRegister(REG_MODEM_CONFIG_2);
  regmodemconfig2 &= ~(LORA_SPREADING_FACTOR_MASK);
  regmodemconfig2 |= (uint8_t)spreadingfactor;

  if (!WriteRegister(REG_MODEM_CONFIG_2, regmodemconfig2))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to set LoRa spreading factor\n", m_devicestr);
    return false;
  }

  return true;
}

/*! Sets the base addresses of the RX and TX LoRa packets in the 256 byte fifo.
    \param txbaseaddr The address of the TX packet.
    \param rxbaseaddr The address of the RX packet.
    \return true if setting the base addresses succeeded, false if it failed.*/
bool CRFM9xW::SetLoRaBaseAddresses(uint8_t txbaseaddr, uint8_t rxbaseaddr)
{
  assert(m_activated);

  if (!WriteRegister(REG_FIFO_TX_BASE_ADDR, txbaseaddr))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to set LoRa tx base address\n", m_devicestr);
    return false;
  }

  if (!WriteRegister(REG_FIFO_RX_BASE_ADDR, rxbaseaddr))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to set LoRa rx base address\n", m_devicestr);
    return false;
  }

  return true;
}

/*! Gets the rssi of the last received LoRa packet.
    \return The rssi of the last received packet in dBm.*/
float CRFM9xW::GetLoRaPacketRssi()
{
  assert(m_activated);

  float regpktrssivalue = ReadRegister(REG_PKT_RSSI_VALUE);

  //the offset is -164 when using the LF port, and -157 when using the HF port
  float offset;
  if (m_frequency < 652000000)
    offset = -164.0f;
  else
    offset = -157.0f;

  return offset + regpktrssivalue;
}

/*! Sets the mode of the DIO0 pin of the RFM9xW.
    \param dio0mode A value of @ref EDIO0Mode.
    \return true if setting the DIO0 mode succeeded, false if it failed.*/
bool CRFM9xW::SetLoRaDIO0Mode(EDIO0Mode dio0mode)
{
  assert(m_activated);

  CHECK_DIO0_MODE(dio0mode);

  uint8_t regdiomapping1 = ReadRegister(REG_DIO_MAPPING_1);
  regdiomapping1 &= ~(DIO0_MODE_MASK);
  regdiomapping1 |= (uint8_t)dio0mode;

  if (!WriteRegister(REG_DIO_MAPPING_1, regdiomapping1))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to set LoRa DIO0 mode\n", m_devicestr);
    return false;
  }

  return true;
}

/* Returns if the CrcOnPayload bit is set in the RegHopChannel register,
   this indicates if the received LoRa packet has a crc.*/
bool CRFM9xW::GetLoRaCrcOnPayload()
{
  assert(m_activated);

  uint8_t reghopchannel = ReadRegister(REG_HOP_CHANNEL);
  return !!(reghopchannel & CRC_ON_PAYLOAD);
}

/* Returns the value of RegRxNbBytes, which indicates the size of the received LoRa packet.*/
uint8_t CRFM9xW::GetLoRaPacketSize()
{
  assert(m_activated);

  return ReadRegister(REG_RX_NB_BYTES);
}

/*! Returns the value of FifoRxCurrentAddr, which is the start address of the received LoRa packet.*/
uint8_t CRFM9xW::GetLoRaPacketAddress()
{
  assert(m_activated);

  return ReadRegister(REG_FIFO_RX_CURRENT_ADDR);
}

/*! Sets the fifo pointer for LoRa mode, this is where in the fifo bytes can be read or written.
    \param fifoptr The new value of the fifo pointer.
    \return true if setting the fifo pointer succeeded, false if it failed.*/
bool CRFM9xW::SetLoRaFifoPtr(uint8_t fifoptr)
{
  assert(m_activated);

  if (!WriteRegister(REG_FIFO_ADDR_PTR, fifoptr))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to set fifo pointer to %" PRIu8 "\n", m_devicestr, fifoptr);
    return false;
  }

  return true;
}

/*! Reads data from the RFm9xW fifo into buf.
    \param buf The buffer to write data into.
    \param size The number of bytes to write into buf.*/
void CRFM9xW::ReadLoRaFifo(uint8_t* buf, uint8_t size)
{
  assert(m_activated);
  assert(buf);
  assert(size > 0);

  uint8_t output[1] = {REG_FIFO};

  //write REG_FIFO to the RFM9xW, to tell it to start reading from the fifo
  //after that, read bytes from the RFM9xW into buf
  Select();
  m_spi.Transceive(output, nullptr, 1);
  m_spi.Transceive(nullptr, buf, size);
  Deselect();
}

/*! Sets the length of the LoRa payload to be transmitted.*/
bool CRFM9xW::SetLoRaPacketSize(uint8_t packetsize)
{
  assert(m_activated);

  if (!WriteRegister(REG_PAYLOAD_LENGTH_LORA, packetsize))
  {
    PrintLog(LOGRFM9xW, "%s: ERROR: unable to set packet size to %" PRIu8 "\n",
             m_devicestr, packetsize);
    return false;
  }

  return true;
}

/*! Writes data to the fifo.*/
void CRFM9xW::WriteLoRaFifo(uint8_t* buf, uint8_t size)
{
  assert(m_activated);
  assert(buf);
  assert(size > 0);

  uint8_t output[1] = {REG_FIFO | WRITE_CMD};

  //write REG_FIFO to the RFM9xW with the write bit set,
  //to tell it to start writing to the fifo
  //after that, write buf to the fifo
  Select();
  m_spi.Transceive(output, nullptr, 1);
  m_spi.Transceive(buf, nullptr, size);
  Deselect();
}

/* Checks the signal detected bit of RegModemStat, this bit is set to 1 when a valid LoRa preamble is detected.
   \return true if receiving a LoRa packet is happening.*/
bool CRFM9xW::GetLoRaIsReceiving()
{
  uint8_t regmodemstat = ReadRegister(REG_MODEM_STAT);
  return ((regmodemstat & LORA_SIGNAL_DETECTED_BIT) == LORA_SIGNAL_DETECTED_BIT);
}

/*! Sets the the automatic gain control for the low noise amplifier.
    \param enable Whether to enable agc for the lna.*/
bool CRFM9xW::SetOOKLnaAgc(bool enable)
{
  assert(m_activated);

  //the agc bit for OOK is in the RegRxConfig register
  uint8_t regrxconfig = ReadRegister(REG_RX_CONFIG);
  if (enable)
    regrxconfig |= OOK_LNA_AGC_BIT;
  else
    regrxconfig &= ~(OOK_LNA_AGC_BIT);

  if (!WriteRegister(REG_RX_CONFIG, regrxconfig))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to set OOK AGC\n", m_devicestr);
    return false;
  }

  return true;
}

/*! Enables or disables the bit synchronizer.
    \param bitsynchronizer a value of EBitSynchronizer.
    \return true if setting the bit synchronizer succeeded, false if it failed.*/
bool CRFM9xW::SetOOKBitSynchronizer(EBitSynchronizer bitsynchronizer)
{
  assert(m_activated);

  CHECK_BIT_SYNCHRONIZER(bitsynchronizer);

  uint8_t regookpeak = ReadRegister(REG_OOK_PEAK);
  regookpeak &= ~(OOK_BIT_SYNCHRONIZER_MASK);
  regookpeak |= (uint8_t)bitsynchronizer;

  if (!WriteRegister(REG_OOK_PEAK, regookpeak))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to set OOK bit synchronizer\n", m_devicestr);
    return false;
  }

  return true;
}

/*! Sets the threshold mode of the OOK reception.
    \param threshmode a value of @ref EOOKThreshType.
    \return true if setting the threshold mode succeeded, false if it failed.*/
bool CRFM9xW::SetOOKThresholdMode(EOOKThreshType threshmode)
{
  assert(m_activated);

  CHECK_THRESH_TYPE(threshmode);

  uint8_t regookpeak = ReadRegister(REG_OOK_PEAK);
  regookpeak &= ~(OOK_THRESH_TYPE_MASK);
  regookpeak |= (uint8_t)threshmode;
  if (!WriteRegister(REG_OOK_PEAK, regookpeak))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to set OOK threshold mode\n", m_devicestr);
    return false;
  }

  return true;
}

/* Sets the threshold level for fixed threshold mode, and the floor threshold for peak threshold mode.
   \param threshold The threshold level.*/
bool CRFM9xW::SetOOKFixedThreshold(uint8_t threshold)
{
  assert(m_activated);

  if (!WriteRegister(REG_OOK_FIX, threshold))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to set OOK fixed threshold\n", m_devicestr);
    return false;
  }

  return true;
}
/* Set the period of decrement of the RSSI threshold in the OOK demodulator.
   \param decr a value of @ref EOOKPeakThreshDec.
   \return true if setting the period of decrement succeeded, false if it failed.*/
bool CRFM9xW::SetOOKPeakThreshDec(EOOKPeakThreshDec decr)
{
  assert(m_activated);

  CHECK_OOK_PEAK_THRESH_DEC(decr);

  uint8_t regookavg = ReadRegister(REG_OOK_AVG);
  regookavg &= ~(OOK_PEAK_THRESH_DEC_MASK);
  regookavg |= (uint8_t)decr;

  if (!WriteRegister(REG_OOK_AVG, regookavg))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to set OOK fixed threshold\n", m_devicestr);
    return false;
  }

  return true;
}

/*! Sets the OOK data mode.
    \param datamode A value of @ref EOOKDataMode.
    \return true if setting the OOK data mode succeeded, false if it failed.*/
bool CRFM9xW::SetOOKDataMode(EOOKDataMode datamode)
{
  assert(m_activated);

  CHECK_DATA_MODE(datamode);

  uint8_t regpacketconfig2 = ReadRegister(REG_PACKET_CONFIG_2);
  regpacketconfig2 &= ~(DATA_MODE_MASK);
  regpacketconfig2 |= (uint8_t)(datamode);

  if (!WriteRegister(REG_PACKET_CONFIG_2, regpacketconfig2))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to set OOK data mode\n", m_devicestr);
    return false;
  }

  return true;
}

/*! Sets the bandwidth of the OOK receiver.
    \param bw A value of @ref ERxBw.
    \return true if setting the bandwidth succeeded, false if it failed.*/
bool CRFM9xW::SetOOKRxBw(ERxBw bw)
{
  assert(m_activated);

  CHECK_RX_BW(bw);
  if (!WriteRegister(REG_RX_BW, (uint8_t)bw))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to set OOK rx bandwidth\n", m_devicestr);
    return false;
  }

  return true;
}

/*! Sets the rssi smoothing of the OOK receiver.
    \param rssismoothing a value of @ref ERssiSmoothing.
    \return true if setting the rssi smoothing succeeded, false if it failed.*/
bool CRFM9xW::SetOOKRssiSmoothing(ERssiSmoothing rssismoothing)
{
  assert(m_activated);

  CHECK_RSSI_SMOOTHING(rssismoothing);

  uint8_t regrssiconfig = ReadRegister(REG_RSSI_CONFIG);
  regrssiconfig &= ~(RSSI_SMOOTHING_MASK);
  regrssiconfig |= rssismoothing;

  if (!WriteRegister(REG_RSSI_CONFIG, regrssiconfig))
  {
    PrintLog(LOGRFM9xW, "ERROR: %s: unable to set OOK RSSI smoothing\n", m_devicestr);
    return false;
  }

  return true;
}

/* Returns the measured rssi in OOK reception.*/
float CRFM9xW::GetOOKRssi()
{
  assert(m_activated);

  float regrssivalue = ReadRegister(REG_RSSI_VALUE);
  return -regrssivalue * 0.5f;
}

/*! Enables the pin interrupt on dio0 via the EXTI controller,
    the CRFM9xWInterruptHandler() method is then called on m_irqhandler.*/
void CRFM9xW::EnableDIO0Interrupt()
{
  g_exti.EnableGPIOIrq(m_dio0, EXTI_Trigger_Rising, this);
}

/*! When enabled, disables the pin interrupt on DIO0.*/
void CRFM9xW::DisableDIO0Interrupt()
{
  g_exti.DisableGPIOIrq(m_dio0);
}

/*! When EnableDIO0Interrupt() is called, this method is called from g_exti when
    a rising edge happens on the dio0 pin.*/
void CRFM9xW::GPIOIRQHandler(SGPIOPin pin)
{
  if (m_irqhandler)
    m_irqhandler->CRFM9xWInterruptHandler();
}

/* Returns the value of the irq flags register, and it will write back
   the value it has written to it, to clear all irq flags.*/
uint8_t CRFM9xW::GetIRQFlags()
{
  uint8_t irqflags = ReadRegister(REG_IRQ_FLAGS);
  WriteRegister(REG_IRQ_FLAGS, irqflags, NoCheckWrittenValue);

  return irqflags;
}
