#include "rtcscheduler.h"
#include "crc32hw.h"
#include "backupregisters.h"
#include "log.h"
#include "timeconversion.h"
#include "lock.h"

#include <cstring>
#include <cassert>
#include <algorithm>

/* Sets a valid timestamp for this schedule entry.
   \param timestamp A unix timestamp for when this entry happens.
   \param countdown When true, the entry is a countdown timer, which means that
                    the timestamp is adjusted when the RTC time is adjusted.*/
void CRTCScheduleEntry::Set(int64_t timestamp, bool countdown)
{
  Reset();
  m_timestamp = timestamp;
  m_flags     = RTCSCHEDULEFLAG_VALID;
  if (countdown)
    m_flags |= RTCSCHEDULEFLAG_COUNTDOWN;
  SetCRC();
}

/*! Resets all fields to 0.*/
void CRTCScheduleEntry::Reset()
{
  memset(this, 0, sizeof(*this));
}

/*! Checks if the valid flag is set, if so checks the CRC.
    \param index The index of this entry, used for logging.
    \return true The valid flag is set and the CRC is correct.
    \return false The valid flag is not set.*/
bool CRTCScheduleEntry::Valid(uint32_t index /*= MAX_RTCSCHEDULE_ENTRIES*/)
{
  if (m_flags & RTCSCHEDULEFLAG_VALID)
  {
    uint32_t crc = GenerateCRC();
    if (crc == m_crc)
    {
      return true;
    }
    else
    {
      if (index < MAX_RTCSCHEDULE_ENTRIES)
        PrintLog(LOGRTCSCHED, "ERROR: Resetting RTC schedule entry %" PRIu32
                 " because of invalid CRC %08" PRIX32 ", should be %08" PRIX32 "\n",
                 index, m_crc, crc);
      else
        PrintLog(LOGRTCSCHED, "ERROR: Resetting RTC schedule entry"
                 " because of invalid CRC %08" PRIX32 ", should be %08" PRIX32 "\n",
                 m_crc, crc);

      //Reset all fields if the CRC is not correct.
      Reset();
      return false;
    }
  }
  else
  {
    return false;
  }
}

/*! Adjusts the timestamp with the specified timediff. Is used on countdown
    alarms after the RTC has been adjusted.
    \param timediff The difference in seconds to adjust the timestamp with.*/
void CRTCScheduleEntry::AdjustTimestamp(int64_t timediff)
{
  m_timestamp += timediff;
  SetCRC();
}

/*! Marks an entry as expired, this is done when the RTC time is newer than the
    timestamp of this entry.
    \param index The index of this entry, used for logging.*/
void CRTCScheduleEntry::SetExpired()
{
  m_flags |= RTCSCHEDULEFLAG_EXPIRED;
  SetCRC();
}

/*! Marks an entry as handled, this is the responsibility of the code that is
    using the entry. The code should detect that the entry has expired, do its
    corresponding processing, then mark the entry as handled.*/
void CRTCScheduleEntry::SetHandled()
{
  m_flags |= RTCSCHEDULEFLAG_HANDLED;
  SetCRC();
}

void CRTCScheduleEntry::SetCRC()
{
  m_crc = GenerateCRC();
}

/*! \return A CRC32 calculated over the CRTCScheduleEntry object.*/
uint32_t CRTCScheduleEntry::GenerateCRC()
{
  return CCRC32Hw::Calculate(*this, CRCGenSize());
}

CRTCScheduler::CRTCScheduler() : m_callbacks {}, m_signal(0), m_mutex(0)
{
  //The RTC schedule entries are stored in the battery backed SRAM.
  m_entries = (CRTCScheduleEntry*)BackupSRAM();
}

void CRTCScheduler::Setup()
{
  //Check the CRC of all entries.
  for (uint32_t i = 0; i < MAX_RTCSCHEDULE_ENTRIES; i++)
    m_entries[i].Valid(i);

  //Create a binary semaphore to signal from the RTC interrupt that the RTC
  //alarm expired.
  m_signal = xSemaphoreCreateBinary();

  //Create a mutex for locking access to member variables.
  m_mutex = xSemaphoreCreateRecursiveMutex();

  //Enable the alarm B interrupt, this is needed so that the RTCIRQHandler()
  //method is called when the RTC alarm B happens, and it's needed to wake the
  //microcontroller up from standby.
  RTC_ITConfig(RTC_IT_ALRB, ENABLE);
  g_exti.EnableRTCIrq(RTCIRQAlarmB, this);

  //Set a callback pointer so that the RTCAdjustCallback() method is called when
  //the RTC time is adjusted through the SetRTC() function.
  SetRTCAdjustCallback(this);

  //Call virtual method in derived class. The Accello uses this to register the
  //scheduler as a low power task.
  PostSetup();
}

void CRTCScheduler::Start()
{
  Create("RTCScheduler");
}

void CRTCScheduler::Process()
{
  //Clear any pending signal so that the loop doesn't run twice when this
  //task is started.
  xSemaphoreTake(m_signal, 0);

  for(;;)
  {
    //Call virtual method in derived class. The Accello uses this to inhibit
    //entering low power mode.
    PreProcess();

    //Always clear the RTC alarm B flag, as long as it's set the microcontroller
    //will not enter low power mode.
    RTC_ClearFlag(RTC_FLAG_ALRBF);

    //Check if any alarm has expired, if so call its corresponding callback.
    ProcessAlarms();

    //Set RTC alarm B to the earliest alarm that has not expired yet.
    //If the microcontroller is running when RTC alarm B happens,
    //the RTCIRQHandler() method will be called. If the microcontroller is in
    //low power mode, RTC alarm B will wake it up.
    SetNextAlarm();

    //Call virtual method in derived class. The Accello uses this to allow
    //entering low power mode.
    PostProcess();

    //Wait for RTC alarm B to happen.
    xSemaphoreTake(m_signal, portMAX_DELAY);
  }
}

/*! Checks if an alarm has expired, if so marks it as expired and calls its
    corresponding callback.*/
void CRTCScheduler::ProcessAlarms()
{
  bool expired;

  do //Keep looping until there are no more expired alarms.
  {
    CLock lock(m_mutex);
    expired = false;

    int64_t now = GetRTCSeconds(nullptr);
    for (uint32_t i = 0; i < MAX_RTCSCHEDULE_ENTRIES; i++)
    {
      CRTCScheduleEntry& entry = m_entries[i];

      if (entry.Valid(i) && !entry.Expired() && now >= entry.Timestamp())
      {
        char alarmstr[DATESTR_MAX_LENGTH] = {};
        ToISO8601(entry.Timestamp(), alarmstr, sizeof(alarmstr));

        PrintLog(LOGRTCSCHED, "Marking RTC schedule entry %" PRIu32
                 " with timestamp %s as expired\n",
                 i, alarmstr);

        entry.SetExpired();

        //Leave the lock before calling the callback, because the callback might
        //take a long time to complete.
        lock.Leave();

        if (m_callbacks[i])
          m_callbacks[i]->RTCSchedulerCallback(i);

        //Restart the for loop, because some alarm might have changed when the
        //callback returns.
        expired = true;
        break;
      }
    }
  }
  while (expired);
}

//! Maximum hours in the future that the next alarm should be set at.
#define MAX_ALARM_IN_FUTURE 36

/*! Sets RTC alarm B to the earliest alarm entry has has not yet expired.*/
void CRTCScheduler::SetNextAlarm()
{
  CLock lock(m_mutex);

  int64_t now = GetRTCSeconds(nullptr);

  //Limit how far in the future the alarm can be set.
  //This acts as a safeguard, in case something went wrong in calculating the alarm.
  int64_t nextalarm = now + (MAX_ALARM_IN_FUTURE * 60 * 60);
  uint32_t nextindex = MAX_RTCSCHEDULE_ENTRIES;

  //Find the earliest alarm that is not yet expired, RTC alarm B will be set
  //to this time.
  for (uint32_t i = 0; i < MAX_RTCSCHEDULE_ENTRIES; i++)
  {
    CRTCScheduleEntry& entry = m_entries[i];

    if (entry.Valid(i) && !entry.Expired() && entry.Timestamp() < nextalarm)
    {
      nextalarm = entry.Timestamp();
      nextindex = i;
    }
  }

  if (nextindex < MAX_RTCSCHEDULE_ENTRIES)
  {
    char alarmstr[DATESTR_MAX_LENGTH] = {};
    ToISO8601(m_entries[nextindex].Timestamp(), alarmstr, sizeof(alarmstr));

    PrintLog(LOGRTCSCHED, "Next alarm is index %" PRIu32 " with timestamp %s\n",
             nextindex, alarmstr);
  }
  else
  {
    PrintLog(LOGRTCSCHED, "No alarms scheduled, setting next alarm %i hours in "
             "the future\n", MAX_ALARM_IN_FUTURE);
  }

  //Make sure that the next alarm is always at least 3 seconds into the future,
  //because the RTC alarm only triggers when the RTC changes to the alarm time.
  //If the RTC is already at the alarm time, the alarm does not trigger.
  nextalarm = std::max(nextalarm, now + 3);

  //Set RTC alarm B to the next time that the microcontroller should process
  //the scheduled alarms. If the microcontroller is in low power mode the alarm
  //will wake it up.
  tm timeval;
  SecsToTm(nextalarm, &timeval);
  SetRTCAlarm(RTC_Alarm_B, &timeval);
}

/*! Adjusts countdown alarms with timediff. Called when the RTC time is adjusted.
    \param timediff Number of seconds that the RTC has been adjusted.
                    Positive when the RTC had been adjusted forwards.
                    Negative when the RTC has been adjusted backwards.*/
void CRTCScheduler::AdjustCountDownAlarms(int64_t timediff)
{
  CLock lock(m_mutex);

  for (uint32_t i = 0; i < MAX_RTCSCHEDULE_ENTRIES; i++)
  {
    CRTCScheduleEntry& entry = m_entries[i];

    if (entry.Valid(i) && entry.CountDown())
    {
      CLogLock loglock;
      PrintLog(LOGRTCSCHED, "Adjusting countdown alarm %" PRIu32 " from ", i);

      char alarmstr[DATESTR_MAX_LENGTH] = {};
      ToISO8601(entry.Timestamp(), alarmstr, sizeof(alarmstr));
      PrintLog(LOGRTCSCHED, "%s to ", alarmstr);

      entry.AdjustTimestamp(timediff);

      ToISO8601(entry.Timestamp(), alarmstr, sizeof(alarmstr));
      PrintLog(LOGRTCSCHED, "%s\n", alarmstr);
    }
  }
}

/*! Sets a scheduled alarm.
    \param index The index of the alarm, should be lower than MAX_RTCSCHEDULE_ENTRIES.
    \param timeval The time that the alarm expires, as a struct tm.*/
void CRTCScheduler::SetAlarm(uint32_t index, tm* timeval)
{
  SetAlarm(index, timeval, TmToSecs(timeval), false);
}

/*! Sets a scheduled alarm.
    \param index The index of the alarm, should be lower than MAX_RTCSCHEDULE_ENTRIES.
    \param timestamp The time that the alarm expires, as a unix timestamp.*/
void CRTCScheduler::SetAlarm(uint32_t index, int64_t timestamp)
{
  tm timeval;
  SecsToTm(timestamp, &timeval);
  SetAlarm(index, &timeval, timestamp, false);
}

/*! Sets a countdown alarm in seconds.
    \param index The index of the alarm, should be lower than MAX_RTCSCHEDULE_ENTRIES.
    \param seconds The number of seconds until the alarm should trigger.*/
void CRTCScheduler::SetCountDown(uint32_t index, int64_t seconds)
{
  //Get the RTC timestamp while holding a lock on m_mutex, this avoids a race
  //condition when the RTC time is changed by the SetRTC() function. m_mutex is
  //also locked by SetRTC() through the RTCAdjustCallback() method.
  CLock lock(m_mutex);

  int64_t timestamp = GetRTCSeconds(nullptr) + seconds;

  tm timeval;
  SecsToTm(timestamp, &timeval);
  SetAlarm(index, &timeval, timestamp, true);
}

/*! Sets a scheduled alarm.
    \param index The index of the alarm, should be lower than MAX_RTCSCHEDULE_ENTRIES.
    \param timeval The time that the alarm expires, as a struct tm.
    \param timestamp The time that the alarm expires, as a unix timestamp.
    \param countdown When true, the alarm is a countdown timer, which means that
                     the timestamp is adjusted when the RTC time is adjusted.*/
void CRTCScheduler::SetAlarm(uint32_t index, tm* timeval, int64_t timestamp,
                             bool countdown)
{
  assert(index < MAX_RTCSCHEDULE_ENTRIES);
  
  char alarmstr[DATESTR_MAX_LENGTH] = {};
  ToISO8601(timeval, alarmstr, sizeof(alarmstr));

  PrintLog(LOGRTCSCHED, "Setting alarm %" PRIu32 " to %s\n", index, alarmstr);

  CLock lock(m_mutex);
  m_entries[index].Set(timestamp, countdown);

  //Set RTC alarm from this task, because it's possible that the microcontroller
  //will enter low power mode shortly after this function exits, in which case the
  //CRTCScheduler task will not have enough time to set RTC alarm B.
  SetNextAlarm();

  //Signal to the CRTCScheduler task that an alarm has been set.
  xSemaphoreGive(m_signal);
}

/*! Disables a scheduled alarm.
    \param index The index of the alarm to disable.*/
void CRTCScheduler::DisableAlarm(uint32_t index)
{
  assert(index < MAX_RTCSCHEDULE_ENTRIES);

  CLock lock(m_mutex);

  if (m_entries[index].Valid(index))
  {
    PrintLog(LOGRTCSCHED, "Disabling alarm %" PRIu32 "\n", index);

    m_entries[index].Reset();
    SetNextAlarm();

    //Signal to the CRTCScheduler task that an alarm has been disabled.
    xSemaphoreGive(m_signal);
  }
}

/*! Sets a callback which is called when the alarm happens.
    \param index The index of the alarm, should be lower than MAX_RTCSCHEDULE_ENTRIES.
    \param callback Pointer to an object of which RTCSchedulerCallback() is called,
                    or nullptr to remove the callback.*/
void CRTCScheduler::SetCallback(uint32_t index, CRTCSchedulerCallback* callback)
{
  assert(index < MAX_RTCSCHEDULE_ENTRIES);

  CLock lock(m_mutex);
  m_callbacks[index] = callback;
}

/*! \param index The index of the alarm to check.
    \return If the alarm is expired.*/
bool CRTCScheduler::AlarmExpired(uint32_t index)
{
  assert(index < MAX_RTCSCHEDULE_ENTRIES);

  CLock lock(m_mutex);

  CRTCScheduleEntry& entry = m_entries[index];

  return entry.Valid(index) && entry.Expired();
}

/*! Called from the RTC alarm interrupt, via the g_exti object.*/
void CRTCScheduler::RTCIRQHandler(ERTCIRQType irqtype)
{
  xSemaphoreGiveFromISR(m_signal, nullptr);
}

/*! Called from the SetRTC() function, before and after the RTC is adjusted.
    \param timediff The difference in seconds between the old and new value of
                    the RTC. A negative value means the RTC has been set backwards.
                    A positive value means the RTC has been set forwards.
    \param callbacktype PreAdjustCallback when the callback is called before
                        the RTC is adjusted.
                        PostAdjustCallback when the callback is called after
                        the RTC is adjusted.
                        Both are called by the SetRTC() function.*/
void CRTCScheduler::RTCAdjustCallback(int64_t timediff,
                                      RTCAdjustCallbackType callbacktype)
{
  if (callbacktype == PreAdjustCallback)
  { //Lock m_mutex before the RTC is adjusted, this is to avoid a race condition
    //when setting a countdown alarm.
    xSemaphoreTakeRecursive(m_mutex, portMAX_DELAY);
  }
  else if (callbacktype == PostAdjustCallback)
  {
    AdjustCountDownAlarms(timediff);
    SetNextAlarm();

    xSemaphoreGiveRecursive(m_mutex);

    xSemaphoreGive(m_signal);
  }
}
