#include "safeheap.h"

#include <FreeRTOS.h>
#include <task.h>
#include <semphr.h>
#include "lock.h"

/*! Mutex for reentrant safe calls to heap functions, which are defined
    in this file. The following linker flags should be passed
    to make this work: -Wl,-wrap,malloc,-wrap,free,-wrap,calloc,-wrap,realloc,-wrap,reallocarray
    The linker then prefixes the original functions with __real_ and prefixes calls to
    the original functions with __wrap_.
    This way, a call to malloc() actually calls __wrap_malloc(), from which a mutex is locked first,
    and then from that function, __real_malloc() is called to do the allocation.
*/
static SemaphoreHandle_t g_malloclock;

//prototypes for the renamed versions of the heap functions
extern "C"
{
  void* __real_malloc(size_t size);
  void  __real_free(void* ptr);
  void* __real_calloc(size_t nmemb, size_t size);
  void* __real_realloc(void *ptr, size_t size);
  void* __real_reallocarray(void *ptr, size_t nmemb, size_t size);
}

void InitSafeHeap()
{
  //call to make sure that the function exist,
  //if they don't, this causes a linker error
  __real_free(__real_malloc(0));
  __real_free(__real_calloc(0, 0));
  __real_free(__real_realloc(nullptr, 0));
  __real_free(__real_reallocarray(nullptr, 0, 0));

  //create a mutex for reentrant safe calls to the heap functions,
  //see the comment above where g_malloclock is defined
  g_malloclock = xSemaphoreCreateRecursiveMutex();
}

//Locking a mutex only works if the mutex has been allocated and the kernel is running.
bool CanLock()
{
  return g_malloclock && xTaskGetSchedulerState() == taskSCHEDULER_RUNNING;
}

//reentrant safe wrappers for the heap functions
extern "C" void* __wrap_malloc(size_t size)
{
  CLock lock(g_malloclock, CanLock());
  return __real_malloc(size);
}

extern "C" void __wrap_free(void* ptr)
{
  CLock lock(g_malloclock, CanLock());
  __real_free(ptr);
}

extern "C" void* __wrap_calloc(size_t nmemb, size_t size)
{
  CLock lock(g_malloclock, CanLock());
  return __real_calloc(nmemb, size);
}

extern "C" void* __wrap_realloc(void *ptr, size_t size)
{
  CLock lock(g_malloclock, CanLock());
  return __real_realloc(ptr, size);
}

extern "C" void* __wrap_reallocarray(void *ptr, size_t nmemb, size_t size)
{
  CLock lock(g_malloclock, CanLock());
  return __real_reallocarray(ptr, nmemb, size);
}
