#include "spi.h"
#include "lock.h"
#include "clock.h"

#include <cstddef>
#include <cassert>
#include <cstdio>

CSpi::CSpi() {
    m_spinss = SPI_NSS_Hard;
    m_spi = NULL;
    m_txstream = 0;
    m_rxstream = 0;
    m_txchannel = 0;
    m_rxchannel = 0;
    m_lastdisable = 0;
    m_rxdone = 0;
}

void CSpi::Setup(SPI_TypeDef *spi, uint16_t baudrateprescaler /*= SPI_BaudRatePrescaler_256*/,
                 uint16_t spinss /*= SPI_NSS_Soft*/,
                 uint16_t pol /*= SPI_CPOL_Low*/, uint16_t pha /*= SPI_CPHA_1Edge*/) {

    assert(IS_SPI_ALL_PERIPH(spi));
    assert(IS_SPI_BAUDRATE_PRESCALER(baudrateprescaler));
    assert(IS_SPI_NSS(spinss));
//
//  4. Issue: SPI cannot transmit data after switching the clock
//reason SPI design difference between GD32 and STM32
//solution Close SPI before switching the SPI clock and enable SPI after
//the switching is completed
//type GD32 all series
//tips
//5. Issue: SPI can send data and cannot receive data
//reason Receive port is configure as in floating mode in normal, while,
//receiving port of ST , which is configure as push pull output,
//could work in normal.
//solution Check the mode of MISO pin is floating input or not.
//type GD32 all series

    Close(); // doet niks, want m_spi not set

    //creat a binary semaphore for signaling that the rx is done
    m_rxdone = xSemaphoreCreateBinary();

    //use a critical section to protect the state of CSpi and the APB and SPI registers
    CCritical criticalsection;

    m_spinss = spinss;

    //store a pointer to the spi peripheral, the index of the spi peripheral
    //the tx and rx dma streams, and the tx and rx dma request channels
    m_spi = spi;
    if (m_spi == SPI1) {
        m_txstream = DMA2_Stream3;
        m_rxstream = DMA2_Stream2;
        m_txchannel = DMA_Channel_3;
        m_rxchannel = DMA_Channel_3;
    } else if (m_spi == SPI2) {
        m_txstream = DMA1_Stream4;
        m_rxstream = DMA1_Stream3;
        m_txchannel = DMA_Channel_0;
        m_rxchannel = DMA_Channel_0;
    } else if (m_spi == SPI3) {
        m_txstream = DMA1_Stream7;
        m_rxstream = DMA1_Stream2;
        m_txchannel = DMA_Channel_0;
        m_rxchannel = DMA_Channel_0;
    } else if (m_spi == SPI4) { // Edwin : we use SPI4 (on GD genaamd SPI3)
        /*This library uses only USART TX DMA. Default DMA streams and channels are in table below:

  SPIx	DMA	    DMA TX Stream	DMA TX Channel	DMA RX Stream	DMA RX Channel
  SPI1	DMA2	DMA Stream 3	DMA Channel 3	DMA Stream 2	DMA Channel 3
  SPI2	DMA1	DMA Stream 4	DMA Channel 0	DMA Stream 3	DMA Channel 0
  SPI3	DMA1	DMA Stream 5	DMA Channel 0	DMA Stream 0	DMA Channel 0
  *** SPI4	DMA2	DMA Stream 1	DMA Channel 4	DMA Stream 0	DMA Channel 4
  SPI5	DMA2	DMA Stream 6	DMA Channel 7	DMA Stream 5	DMA Channel 7
  SPI6	DMA2	DMA Stream 5	DMA Channel 1	DMA Stream 6	DMA Channel 0
         * */
        m_txstream = DMA2_Stream1;
        m_rxstream = DMA2_Stream0;
        m_txchannel = DMA_Channel_4;
        m_rxchannel = DMA_Channel_4;
    } else if (m_spi == SPI5) {
        m_txstream = DMA2_Stream4;
        m_rxstream = DMA2_Stream5;
        m_txchannel = DMA_Channel_2;
        m_rxchannel = DMA_Channel_7;
    } else if (m_spi == SPI6) { //WARNING: this shares a stream with SPI5, so only one can be active at one time
        m_txstream = DMA2_Stream5;
        m_rxstream = DMA2_Stream6;
        m_txchannel = DMA_Channel_1;
        m_rxchannel = DMA_Channel_1;
    }

    //enable the clock for the SPI peripheral
    //   SetClock(DISABLE);
    //   vTaskDelay(1);
    SetClock(ENABLE);
    //  vTaskDelay(1);

    SPI_I2S_DeInit(m_spi); // edwin added reset of device I think it needs to be done here, the clock is active now

    //initialize the spi peripheral in master mode 0, with hardware controlled slave select
    SPI_InitTypeDef spiinit =
            {
                    .SPI_Direction         = SPI_Direction_2Lines_FullDuplex,
                    .SPI_Mode              = SPI_Mode_Master,
                    .SPI_DataSize          = SPI_DataSize_8b,
                    .SPI_CPOL              = pol,
                    .SPI_CPHA              = pha,
                    .SPI_NSS               = m_spinss,
                    .SPI_BaudRatePrescaler = baudrateprescaler,
                    .SPI_FirstBit          = SPI_FirstBit_MSB,
                    .SPI_CRCPolynomial     = 7
            };

    //initialize the spi peripheral
    SPI_Init(m_spi, &spiinit);

    //enable the slave select output to go low when the spi peripheral is enabled with SPI_Cmd()
    SPI_SSOutputCmd(m_spi, ENABLE);

    //initialize the disable timestamp
    m_lastdisable = Micros();
}

void CSpi::Close() {
    CCritical criticalsection;

#ifdef GD32F427
    if (m_spi == SPI4 || m_spi == SPI5 || m_spi == SPI6) {
        // bitbang it

        //reset the internal state
        m_spi = NULL;
        m_txstream = 0;
        m_rxstream = 0;
        m_txchannel = 0;
        m_rxchannel = 0;

        criticalsection.Leave();
        vSemaphoreDelete(m_rxdone);
        return;
    }
#endif

    if (m_spi) {
        //reset the SPI peripheral to the default state
        SPI_I2S_DeInit(m_spi);

        //disable the clock for the SPI peripheral
        SetClock(DISABLE);

        //reset the internal state
        m_spi = NULL;
        m_txstream = 0;
        m_rxstream = 0;
        m_txchannel = 0;
        m_rxchannel = 0;

        criticalsection.Leave();
        vSemaphoreDelete(m_rxdone);
    }
}

#define SPI_IRQ_PRIORITY (configLIBRARY_MAX_SYSCALL_INTERRUPT_PRIORITY)

/*! Delay in microseconds to wait after enabling/disabling the SPI peripheral,
    this makes the NSS line to go/high.*/
#define NSS_DELAY 5

/*! Minimum number of microseconds between an spi disable, and the next enable,
    this is needed because the hardware NSS pin is open drain with a pull up resistor,
    this has a rather large rise time.*/
#define NSS_RISE_DELAY 50

void CSpi::Transceive(const uint8_t *output, uint8_t *input, uint32_t numbytes) {

    assert(m_spi);
    // bitbang it
    uint32_t SCK_PORT = 0;
    uint32_t MISO_PORT = 0;
    uint32_t MOSI_PORT = 0;
//    uint32_t CS_PORT = 0;

    uint16_t SCK_PIN = 0;
    uint16_t MOSI_PIN = 0;
    uint16_t MISO_PIN = 0;
 //   uint16_t CS_PIN = 0;

#ifdef GD32F427

    // EDWIN HACK
    if (m_spi == SPI4 || m_spi == SPI5 || m_spi == SPI6) {

        if (m_spi == SPI4) {
            SCK_PORT = (uint32_t) GPIOE;
            MISO_PORT = (uint32_t) GPIOE;
            MOSI_PORT = (uint32_t) GPIOE;
            // CS_PORT = (uint32_t) GPIOE;
            // --- test ---
            SCK_PIN = GPIO_Pin_2;
            MISO_PIN = GPIO_Pin_5;
            MOSI_PIN = GPIO_Pin_6;
            //  CS_PIN = GPIO_Pin_11;
            // GPIO_WriteBit	(	GPIO_TypeDef * 	GPIOx,uint16_t 	GPIO_Pin,BitAction 	BitVal);
//PE2  ST25R3911B SCK
//PE4  ST25R3911B IRQ
//PE5  ST25R3911B MISO
//PE6  ST25R3911B MOSI
//PE11 ST25R3911B CS
        } else if (m_spi == SPI5) {
            // todo
            // deze is voor 6D sensor
            SCK_PORT = (uint32_t) GPIOE;
            MISO_PORT = (uint32_t) GPIOF;
            MOSI_PORT = (uint32_t) GPIOF;
            // CS_PORT = (uint32_t) GPIOE;
            // --- test ---
            SCK_PIN = GPIO_Pin_7;
            MISO_PIN = GPIO_Pin_8;
            MOSI_PIN = GPIO_Pin_9;
            // CS_PIN = GPIO_Pin_11;
        } else if (m_spi == SPI6) {
            // todo
        } else {
            // can not get here
        }
        //  GPIO_WriteBit((GPIO_TypeDef *) SCK_PORT, SCK_PIN, Bit_RESET); // clock must be low mode 1

        //  TEST1  GPIO_WriteBit((GPIO_TypeDef *) CS_PORT, CS_PIN, Bit_RESET); // chip select
        //  TEST1 DelayUS(NSS_DELAY); //give the NSS pin some time to go low

        //   GPIO_WriteBit((GPIO_TypeDef *) SCK_PORT, SCK_PIN, Bit_SET); // hoog
// digitalWrite(CS_PORT, LOW);        // SS low
        // slaveData = bitBangData(sendData); // data transmission
        // output = bit_bang_spi_edwin(input,numbytes);
        for (uint32_t b = 0; b < numbytes; b++) {

            uint8_t _receive = 0;
            uint8_t _transmit = (output == nullptr) ? 0 : output[b];
            uint8_t bit;
            for (uint8_t i = 0x80; i != 0; i >>= 1)  // 8 bits in a byte
            {
                DelayUS(1);
                GPIO_WriteBit((GPIO_TypeDef *) SCK_PORT, SCK_PIN, Bit_SET);
                bit = (_transmit & i);
                GPIO_WriteBit((GPIO_TypeDef *) MOSI_PORT, MOSI_PIN, bit ? Bit_SET : Bit_RESET);

                DelayUS(1);
                GPIO_WriteBit((GPIO_TypeDef *) SCK_PORT, SCK_PIN, Bit_RESET); // LAAG

                uint8_t rb = GPIO_ReadInputDataBit((GPIO_TypeDef *) MISO_PORT, MISO_PIN);
                if (rb == Bit_SET) {
                    _receive |= i;
                    //    printf("1\n");
                } else {
                    //   printf("0");
                }

            }
            if (input != nullptr)
                input[b] = _receive;
            // printf("%d",_receive);

        }
        //digitalWrite(SSPin, HIGH);       // SS high again
        //   GPIO_WriteBit((GPIO_TypeDef *) SCK_PORT, SCK_PIN, Bit_RESET);
        // TEST1      DelayUS(1);
        // TEST1  GPIO_WriteBit((GPIO_TypeDef *) CS_PORT, CS_PIN, Bit_SET);
// TEST 1 DelayUS(1);
        return;
    }
#endif
// if not GD chip or SPI1 2 or 3 just do normal thing EDWIN

    g_dma.LockStream(m_txstream);
    g_dma.LockStream(m_rxstream);

    //settings for the tx dma stream, if output is set, write it to the spi data register
    //if output is NULL, write nullout to the spi data register, numbytes times
    uint8_t nullout = 0;
    DMA_InitTypeDef txdmainit =
            {
                    .DMA_Channel            = m_txchannel,
                    .DMA_PeripheralBaseAddr = (uint32_t) &m_spi->DR,
                    .DMA_Memory0BaseAddr    = output ? (uint32_t) output : (uint32_t) &nullout,
                    .DMA_DIR                = DMA_DIR_MemoryToPeripheral,
                    .DMA_BufferSize         = numbytes,
                    .DMA_PeripheralInc      = DMA_PeripheralInc_Disable,
                    .DMA_MemoryInc          = output ? DMA_MemoryInc_Enable : DMA_MemoryInc_Disable,
                    .DMA_PeripheralDataSize = DMA_PeripheralDataSize_Byte,
                    .DMA_MemoryDataSize     = DMA_MemoryDataSize_Byte,
                    .DMA_Mode               = DMA_Mode_Normal,
                    .DMA_Priority           = DMA_Priority_Medium,
                    .DMA_FIFOMode           = DMA_FIFOMode_Disable,
                    .DMA_FIFOThreshold      = DMA_FIFOThreshold_1QuarterFull,
                    .DMA_MemoryBurst        = DMA_MemoryBurst_Single,
                    .DMA_PeripheralBurst    = DMA_PeripheralBurst_Single,
            };

    //settings for the rx dma stream, if input is set, read from the spi data register,
    //and write to the input pointer, otherwise write to the nullin variable
    uint8_t nullin = 0;
    DMA_InitTypeDef rxdmainit =
            {
                    .DMA_Channel            = m_rxchannel,
                    .DMA_PeripheralBaseAddr = (uint32_t) &m_spi->DR,
                    .DMA_Memory0BaseAddr    = input ? (uint32_t) input : (uint32_t) &nullin,
                    .DMA_DIR                = DMA_DIR_PeripheralToMemory,
                    .DMA_BufferSize         = numbytes,
                    .DMA_PeripheralInc      = DMA_PeripheralInc_Disable,
                    .DMA_MemoryInc          = input ? DMA_MemoryInc_Enable : DMA_MemoryInc_Disable,
                    .DMA_PeripheralDataSize = DMA_PeripheralDataSize_Byte,
                    .DMA_MemoryDataSize     = DMA_MemoryDataSize_Byte,
                    .DMA_Mode               = DMA_Mode_Normal,
                    .DMA_Priority           = DMA_Priority_Medium,
                    .DMA_FIFOMode           = DMA_FIFOMode_Disable,
                    .DMA_FIFOThreshold      = DMA_FIFOThreshold_1QuarterFull,
                    .DMA_MemoryBurst        = DMA_MemoryBurst_Single,
                    .DMA_PeripheralBurst    = DMA_PeripheralBurst_Single,
            };

    g_dma.EnableIrq(m_rxstream, this, SPI_IRQ_PRIORITY);

    //use a critical section to protect the DMA and SPI registers
    CCritical critsection;

    //make sure that enough time has passed since the last disable
    //so that the open drain NSS pin has enough time to rise using its pullup resistor
    if (m_spinss == SPI_NSS_Hard)
        while (Micros() - m_lastdisable < NSS_RISE_DELAY);

    //enable the SPI peripheral, this makes the NSS line go low
    SPI_Cmd(m_spi, ENABLE);

    if (m_spinss == SPI_NSS_Hard)
        DelayUS(NSS_DELAY); //give the NSS pin some time to go low
    //reset the TX and RX DMA streams EDWIN WANTS RESET FIRST
    //  DMA_DeInit(m_txstream);
    //  DMA_DeInit(m_rxstream);
    //  errata sheet? mogelijk TransferComplete en HalfComplete flags resetten...?
    //  DMA_ClearFlag(m_rxstream,DMA_FLAG_TCIF4 |DMA_FLAG_HTIF4); // channel 4
    //  DMA_ClearFlag(m_txstream,DMA_FLAG_TCIF4|DMA_FLAG_HTIF4); // channel 4

    //intialize the TX and RX DMA streams
    DMA_Init(m_txstream, &txdmainit);
    DMA_Init(m_rxstream, &rxdmainit);

    //  enable the transfer complete interrupt on the RX DMA stream
    //  DMA_ITConfig(m_rxstream, DMA_IT_TC, DISABLE);
    //  enable the transfer complete interrupt on the RX DMA stream
    DMA_ITConfig(m_rxstream, DMA_IT_TC, ENABLE);

    //enable the dma streams
    DMA_Cmd(m_txstream, ENABLE);
    DMA_Cmd(m_rxstream, ENABLE);

    // EDWIN reads... clear all DMA interrupts voor dat we enable doen, anders komt er geen interrupt
    // TODO TODO EDWIN  DMA_ClearITPendingBit(m_rxstream,DMA_IT_TCIF)
    // enable DMA mode for TX and RX in the SPI peripheral
    // this will start the DMA transfer
    // SPI_I2S_DMACmd(m_spi, SPI_I2S_DMAReq_Tx | SPI_I2S_DMAReq_Rx, DISABLE); // edwins
    SPI_I2S_DMACmd(m_spi, SPI_I2S_DMAReq_Tx | SPI_I2S_DMAReq_Rx, ENABLE);

    //leave the critical section while waiting for the transfer complete interrupt on the RX stream
    critsection.Leave();

    //wait until the RX DMA stream is done
    xSemaphoreTake(m_rxdone, portMAX_DELAY); //  hang EDWIN
    while (!DMA_GetFlagStatus(m_rxstream, DMA_FLAG_TCIF2));

    critsection.Enter();

    //disable DMA mode for TX and RX in the SPI peripheral
    SPI_I2S_DMACmd(m_spi, SPI_I2S_DMAReq_Tx | SPI_I2S_DMAReq_Rx, DISABLE);

    //reset the TX and RX DMA streams
    DMA_DeInit(m_txstream);
    DMA_DeInit(m_rxstream);

    //disable the SPI peripheral after a small delay
    if (m_spinss == SPI_NSS_Hard)
        DelayUS(NSS_DELAY);

    SPI_Cmd(m_spi, DISABLE);
    m_lastdisable = Micros();

    //disable the dma interrupt and unlock the streams
    critsection.Leave();
    g_dma.DisableIrq(m_rxstream);
    g_dma.UnlockStream(m_rxstream);
    g_dma.UnlockStream(m_txstream);

}

/* Bit bang it
 * Start
1 Set the SS pin low to begin communication
2 Set the pin for Master Out Slave In (MOSI) to the first bit of the data to be sent
3 Set the clock pin (SCK) high so data is transmitted by the master and received by the slave
4 Read the state of the Master in Slave Out (MISO) to receive the first bit of data from slave
5 Set SCK Low, so data can be sent on the next rising edge
6 Go to 2 till all data bits have been transmitted.
7 Set the SS pin High to stop transmission.
8 Stop*/

void CSpi::SetClock(FunctionalState state) {
    assert(IS_SPI_ALL_PERIPH(m_spi));

    if (m_spi == SPI1)
        // 0x40013000
        RCC_APB2PeriphClockCmd(RCC_APB2Periph_SPI1, state);
    else if (m_spi == SPI2)
        // 0x40003800
        RCC_APB1PeriphClockCmd(RCC_APB1Periph_SPI2, state);
    else if (m_spi == SPI3)
        // 0x40003c00
        RCC_APB1PeriphClockCmd(RCC_APB1Periph_SPI3, state);
    else if (m_spi == SPI4) {
        // 0x40013400
#ifndef GD32F427
        RCC_APB2PeriphClockCmd(RCC_APB2Periph_SPI4, state);
#endif
    } else if (m_spi == SPI5) {
        // 0x40015000
#ifndef GD32F427
        RCC_APB2PeriphClockCmd(RCC_APB2Periph_SPI5, state);
#endif
    } else if (m_spi == SPI6) {
        // 0x40015400
#ifndef GD32F427
        RCC_APB2PeriphClockCmd(RCC_APB2Periph_SPI6, state);
#endif
    }
}

void CSpi::DMAIRQHandler() {
    //this interrupt is only called on transfer complete on the RX DMA stream
    //so signal that the transfer is complete and disable the interrupt
    //the TC flag on this stream is not cleared, because that is done in the DMA_DeInit function
    DMA_ITConfig(m_rxstream, DMA_IT_TC, DISABLE);
    assert(m_rxdone);

    //the SPI task should be blocked on the binary semaphore,
    //use portYIELD_FROM_ISR to unblock it immediately when this ISR is finished
    BaseType_t taskwoken = pdFALSE;
    xSemaphoreGiveFromISR(m_rxdone, &taskwoken);
    portYIELD_FROM_ISR(taskwoken);
}
