#include "spiflash.h"
#include "clock.h"
#include "log.h"
#include "spiflashaddrs.h"

#include <cassert>
#include <algorithm>

CSPIFlash::CSPIFlash()
{
  m_enabled = false;
}

CSPIFlash::~CSPIFlash()
{
}

/*! Sets up the SPI flash memory, can only be called when not enabled.
    If the SPI flash memory needs to be set up a second time, Close() should be called first.
    \param spiptr A pointer to an SPI peripheral.
    \param cs The chip select pin.
    \param rset The reset pin.
    \param wp The write protect pin.
    \param hold The hold pin.
    \return true if setting up the spi flash memory succeeded.
    \return false if setting up the spi flash memory failed.*/
bool CSPIFlash::Setup(SPI_TypeDef* spiptr, SGPIOPin cs, SGPIOPin reset, SGPIOPin wp, SGPIOPin hold)
{
  assert(!m_enabled);

  m_spiptr  = spiptr;
  m_cs      = cs;
  m_reset   = reset;
  m_wp      = wp;
  m_hold    = hold;

  //create a mutex to make access to the spi peripheral atomic
  m_mutex   = xSemaphoreCreateRecursiveMutex();

  //set the state to enabled before doing the setup,
  //because other methods might check the enabled state
  //and only perform actions when it's set to true
  m_enabled = true;

  if (!SetupInternal())
  {
    PrintLog(LOGSPIFLASH, "ERROR: unable to setup SPI FLASH\n");
    Close();
    return false;
  }

  return true;
}

/*! Disables the SPI flash memory.
    This method can always be called, even in the disabled state.*/
void CSPIFlash::Close()
{
  if (m_enabled)
  {
    CloseInternal();
    m_enabled = false;
    vSemaphoreDelete(m_mutex);
    m_mutex = 0;
  }
}

/*! Writes a test pattern to the region reserved for testing, and reads it back.
    \return true The SPI flash is working correctly.
    \return false The SPI flash is not working correctly.*/
bool CSPIFlash::Test()
{
  PrintLog(LOGSPIFLASH, "Testing if SPI flash works, erasing test section\n");

  if (!Enabled())
  {
    PrintLog(LOGSPIFLASH, "ERROR: SPI flash not enabled\n");
    return false;
  }

  //Make sure the section reserved for testing is a multiple of the sector size.
  assert((TEST_ADDRESS % SectorSize()) == 0);
  assert((TEST_RESERVED % SectorSize()) == 0);
  uint32_t sector = TEST_ADDRESS / SectorSize();
  uint32_t numsectors = TEST_RESERVED / SectorSize();

  //Erase the section reserved for performing a test.
  if (!Erase(sector, numsectors))
  {
    PrintLog(LOGSPIFLASH, "ERROR: erasing sectors for test failed\n");
    return false;
  }

  //Create a buffer filled with a test pattern.
  const uint32_t bufsize = 256;
  uint8_t writebuf[bufsize];
  for (uint32_t i = 0; i < bufsize; i++)
    writebuf[i] = i;

  //Fill one whole sector of the SPI flash with the test pattern.
  PrintLog(LOGSPIFLASH, "Writing test pattern to SPI flash\n");
  uint32_t byteswritten = 0;
  while (byteswritten < TEST_RESERVED)
  {
    uint32_t writesize = std::min(bufsize, TEST_RESERVED - byteswritten);
    if (!Write(TEST_ADDRESS + byteswritten, writebuf, writesize))
    {
      PrintLog(LOGSPIFLASH, "ERROR: writing test pattern to SPI flash failed\n");
      return false;
    }
    byteswritten += writesize;
  }

  //Read the test pattern back and check it.
  PrintLog(LOGSPIFLASH, "Reading back test pattern from SPI flash\n");
  uint8_t  readbuf[bufsize];
  uint32_t bytesread = 0;
  while (bytesread < TEST_RESERVED)
  {
    uint32_t readsize = std::min(bufsize, TEST_RESERVED - bytesread);
    if (!Read(FWUPDATE_ADDRESS + bytesread, readbuf, readsize))
    {
      PrintLog(LOGSPIFLASH, "ERROR: reading test pattern from SPI flash failed\n");
      return false;
    }

    for (uint32_t i = 0; i < readsize; i++)
    {
      if (readbuf[i] != writebuf[i])
      {
        PrintLog(LOGSPIFLASH, "ERROR: read test pattern byte %" PRIX8
                 " from SPI flash at position %" PRIu32
                 ", but value %" PRIX8 " has been written there\n",
                 readbuf[i], bytesread + i, writebuf[i]);
        return false;
      }
    }

    bytesread += readsize;
  }

  //Erase the test section after using it.
  if (!Erase(sector, numsectors))
  {
    PrintLog(LOGSPIFLASH, "Erasing SPI flash after writing test pattern failed\n");
    return false;
  }

  //Check if the test section has been erased after using it.
  if (!IsErased(TEST_ADDRESS, TEST_RESERVED))
  {
    PrintLog(LOGSPIFLASH, "ERROR: test section has not been erased fully "
             "after executing an erase\n");
    return false;
  }

  //Report that the SPI flash is working correctly.
  PrintLog(LOGSPIFLASH, "SPI Flash passed all tests\n");

  return true;
}

#define CS_DELAY 10 //! 10 microseconds chip select delay.

void CSPIFlash::Select()
{
  assert(m_enabled);

  DelayUS(CS_DELAY);
  GPIO_Reset(m_cs);
  DelayUS(CS_DELAY);
}

void CSPIFlash::Deselect()
{
  assert(m_enabled);

  DelayUS(CS_DELAY);
  GPIO_Set(m_cs);
  DelayUS(CS_DELAY);
}
