#include "spiflash_S25FL512S.h"
#include "s25fl512s.h"
#include "clock.h"
#include "log.h"
#include "lock.h"

#include <algorithm>
#include <cassert>

CSPIFlash_S25FL512S::CSPIFlash_S25FL512S()
{
}

CSPIFlash_S25FL512S::~CSPIFlash_S25FL512S()
{
}

/*! Called from CSPIFlash::Setup() to set up the SPI peripheral and the flash memory.
    \return true if setting up the flash memory succeeded, false if it failed.*/
bool CSPIFlash_S25FL512S::SetupInternal()
{
  assert(m_enabled);

  //set up the SPI peripheral with the highest speed possible,
  //in mode 0 with software chip select
  m_spi.Setup(m_spiptr, SPI_BaudRatePrescaler_2);

  //make all data pins high
  GPIO_Set(m_cs);
  GPIO_Set(m_wp);
  GPIO_Set(m_hold);
  GPIO_Set(m_reset);

  DelayMS(1);

  //it's possible that the SPI flash is currently doing a write
  //in that case, wait for the write to finish, because any other
  //command will fail
  WaitWriteFinished();

  //read the manufacturer id and the device id
  static const uint8_t out[] = {S25FL512S_READ_ID_CMD, 0x00, 0x00, 0x00, 0x00, 0x00};
  uint8_t in[sizeof(out)]    = {};

  Select();
  m_spi.Transceive(out, in, sizeof(out));
  Deselect();

  PrintLog(LOGSPIFLASH, "S25FL512S manufacturer: 0x%02" PRIX8 ", device: 0x%02" PRIX8 "\n",
           in[4], in[5]);

  //check the manufacturer and device ids, if they're not valid,
  //the spi flash chip is probably not working correctly
  if (in[4] != S25FL512S_MANUFACTURER_ID)
  {
    PrintLog(LOGSPIFLASH, "ERROR: Wrong manufacturer id 0x%02" PRIX8 ", expected 0x%02" PRIX8 "\n",
             in[4], S25FL512S_MANUFACTURER_ID);
    return false;
  }

  if (in[5] != S25FL512S_DEVICE_ID)
  {
    PrintLog(LOGSPIFLASH, "ERROR: Wrong device id 0x%02" PRIX8 ", expected 0x%02" PRIX8 "\n",
             in[5], S25FL512S_DEVICE_ID);
    return false;
  }

  return true;
}

void CSPIFlash_S25FL512S::CloseInternal()
{
  assert(m_enabled);

  m_spi.Close();

  //make all data pins low, so that the flash chip can be powered off
  GPIO_Reset(m_cs);
  GPIO_Reset(m_wp);
  GPIO_Reset(m_hold);
  GPIO_Reset(m_reset);
}

/* Erases one or more 256 kilobyte sectors.
   \param sector The first sector to erase.
   \param numsectors The number of sectors to erase.
   \param isalreadyerased When not nullptr, returns if the last sector that was checked is already erased,
          can be used for useful logging when erasing single sectors multiple times.
   \return true if erasing the sectors succeeded, false if it failed.*/
bool CSPIFlash_S25FL512S::Erase(uint32_t sector, uint32_t numsectors, bool* isalreadyerased /*= nullptr*/)
{
  assert(m_enabled);

  //Sanity check, erasing 0 sectors is probably wrong.
  if (numsectors == 0)
    PrintLog(LOGSPIFLASH, "WARNING: asked to erase 0 sectors at start sector %"
             PRIu32 "\n", sector);

  CLock lock(*this);

  if (isalreadyerased)
    *isalreadyerased = false;

  //wait for any previous write to finish
  if (!WaitWriteFinished())
  {
    PrintLog(LOGSPIFLASH, "ERROR: can't erase %" PRIu32 "sectors starting at sector %" PRIu32 "\n",
             numsectors, sector);
    return false;
  }

  for (uint32_t i = 0; i < numsectors; i++)
  {
    //don't erase the sector if it's already erased
    if (IsSectorErased((sector + i)))
    {
      if (isalreadyerased)
        *isalreadyerased = true;

      continue;
    }

    if (isalreadyerased)
      *isalreadyerased = false;

    //set the Write Enable Latch (WEL) to enable writes
    EnableWrite();

    //calculate the address of the sector that should be erased
    uint32_t address   = (sector + i) * SectorSize();
    uint8_t erasecmd[] = {S25FL512S_SECTOR_ERASE_4_BYTE_ADDR_CMD, (uint8_t)(address >> 24),
                          (uint8_t)(address >> 16), (uint8_t)(address >> 8), (uint8_t)address};

    //execute the erase command
    Select();
    m_spi.Transceive(erasecmd, nullptr, sizeof(erasecmd));
    Deselect();

    //wait until the erase command has completed
    if (!WaitWriteFinished())
    {
      PrintLog(LOGSPIFLASH, "ERROR: unable to erase sector %" PRIu32 "\n", sector + i);
      return false;
    }
  }

  return true;
}

/*! Reads bytes from the flash memory.
    \param address The address to start reading.
    \param buf The buffer to write read data into.
    \param size The Number of bytes to read from the flash memory and to write into buf.
    \return true if reading succeeded, false if it failed.*/
bool CSPIFlash_S25FL512S::Read(uint32_t address, uint8_t* buf, uint32_t size)
{
  assert(m_enabled);

  CLock lock(*this);

  //wait for any previous write to finish
  if (!WaitWriteFinished())
  {
    PrintLog(LOGSPIFLASH, "ERROR: can't read %" PRIu32 "bytes starting at address %" PRIu32 "\n",
             size, address);
    return false;
  }

  //send a fast read command with a 4 byte address, and one dummy byte afterwards,
  //as is required with the default latency code setting of 00 in the CR1 register
  uint8_t readcmd[] = {S25FL512S_FAST_READ_4_BYTE_ADDR_CMD, (uint8_t)(address >> 24),
                       (uint8_t)(address >> 16), (uint8_t)(address >> 8), (uint8_t)address, 0};

  Select();
  m_spi.Transceive(readcmd, nullptr, sizeof(readcmd));
  m_spi.Transceive(nullptr, buf, size);
  Deselect();

  return true;
}

/*! Writes from buf to the flash memory, one page at a time.
    It's advised to start writing at a 16 byte boundary, and to write a multiple of 16 bytes,
    because the automatic ECC of the S25FL512S works at 16 byte blocks,
    and will only work if only a single write per block is done.
    \param address The address to start writing at.
    \param buf The buffer that should be written to the flash memory.
    \param size The number of bytes that should be written to the flash memory.
    \param waitfinish Set to true to wait for this write to complete, set to false to immediately return,
                      the next call to this function will wait for the write to finish first as well.
                      can be used to do other things while the flash memory is busy writing.
    \return true if writing the flash succeeded, false if it failed.*/
bool CSPIFlash_S25FL512S::Write(uint32_t address, const uint8_t* buf, uint32_t size, bool waitfinish /*= true*/)
{
  assert(m_enabled);

  CLock lock(*this);

  //writing the S25FL512S can only be done per 512 byte page,
  //after a write to each page, a new page can only be written when that write has finished

  uint32_t pageoffset = address % PageSize();
  if (pageoffset != 0)
  { //if address does not point to a page boundary, write everything up to the next page boundary
    uint32_t writesize = std::min(PageSize() - pageoffset, size);
    if (!WritePage(address, buf, writesize))
      return false;

    address += writesize;
    buf     += writesize;
    size    -= writesize;
  }

  if (size == 0)
    return true;

  //calculate the number of pages to write, rounded up
  uint32_t numpages = size / PageSize();
  if ((size % PageSize()) != 0)
    numpages++;

  //write the rest of the data, one page at a time
  for (uint32_t i = 0; i < numpages; i++)
  {
    uint32_t writesize = std::min(PageSize(), size);
    if (!WritePage(address, buf, writesize))
      return false;

    address += writesize;
    buf     += writesize;
    size    -= writesize;
  }

  if (waitfinish)
  {
    //wait for the page write to finish
    if (!WaitWriteFinished())
    {
      PrintLog(LOGSPIFLASH, "ERROR: unable to write %" PRIu32 " bytes at address %" PRIu32 "\n",
               address, size);
      return false;
    }
  }

  return true;
}

/*! Finds a byte set to a certain value in a certain section of the SPI flash.
    \param address The address to start reading at.
    \param size The number of bytes to read and check.
    \param match The value to check for.
    \param matchtype MatchType_Normal = do a direct match, MatchType_Not = match if the read byte is any other value (!= operator).
    \return The address of the matched byte, or -1 if no match is found.*/
int32_t CSPIFlash_S25FL512S::FindByte(uint32_t address, uint32_t size, uint8_t match, EMatchType matchtype /*= MatchType_Normal*/)
{
  assert(m_enabled);

  CLock lock(*this);

  //send a fast read command with a 4 byte address, and one dummy byte afterwards,
  //as is required with the default latency code setting of 00 in the CR1 register
  uint8_t readcmd[] = {S25FL512S_FAST_READ_4_BYTE_ADDR_CMD, (uint8_t)(address >> 24),
                       (uint8_t)(address >> 16), (uint8_t)(address >> 8), (uint8_t)address, 0};

  Select();

  //transmit the read command
  m_spi.Transceive(readcmd, nullptr, sizeof(readcmd));

  //check if the entire sector is 0xFF, if only one bit is set to 0, the sector is not erased
  uint32_t       bytesread = 0;
  const uint32_t bufsize   = 128;
  uint8_t        buf[bufsize];
  for (uint32_t i = 0; i < size; i += bufsize)
  { //read the sector in small pieces, so that only a small buffer is needed
    uint32_t readsize = std::min(size - bytesread, bufsize);
    m_spi.Transceive(nullptr, buf, readsize);

    for (uint32_t j = 0; j < readsize; j++)
    {
      //check if the read byte is the same as match for MatchType_Normal,
      //or if it's not the same for MatchType_not
      if ((matchtype == MatchType_Normal && buf[j] == match) ||
          (matchtype == MatchType_Not    && buf[j] != match))
      {
        Deselect();
        return address + i + j; //return the address of this byte
      }
    }

    bytesread += readsize;
  }

  Deselect();

  return -1; //return -1 if no match is found
}

/*! Checks if all bytes of a section of the SPI flash are 0xFF, which indicates an erased state.
    \param address The address of the SPI flash to start reading at.
    \param size The number of bytes to read.
    \return true if the section is erased, false if it's not.*/
bool CSPIFlash_S25FL512S::IsErased(uint32_t address, uint32_t size)
{
  assert(m_enabled);
  return FindByte(address, size, 0xFF, MatchType_Not) == -1;
}

/*! Checks if all bytes of a sector of the SPI flash are 0xFF, which indicates an erased state.
    \param sector The sector to check.
    \return true if the section is erased, false if it's not.*/
bool CSPIFlash_S25FL512S::IsSectorErased(uint32_t sector)
{
  assert(m_enabled);
  return IsErased(sector * SectorSize(), SectorSize());
}

/*! Sets the Write Enable Latch (WEL) in Status Register 1 (SR1).
    The WEL should be set before any write or erase command,
    and is automatically cleared after a successful write or erase.*/
void CSPIFlash_S25FL512S::EnableWrite()
{
  static const uint8_t writeenablecmd[] = {S25FL512S_WRITE_ENABLE_CMD};

  Select();
  m_spi.Transceive(writeenablecmd, nullptr, sizeof(writeenablecmd));
  Deselect();
}

/*! Reads SR1 and returns if the write in progress (WIP) bit is set.
    If there was an error during programming or erasing, WIP will always
    be set, and the error will be signaled in the E_ERR or P_ERR bits.
    \param e_err the E_ERR bit of SR1 can be returned here.
    \param p_err the P_ERR bit of SR1 can be returned here.
    \return true if the WIP bit in SR1 is set, false if it's cleared.*/
bool CSPIFlash_S25FL512S::WriteInProgress(bool* e_err, bool* p_err)
{
  static const uint8_t readsr1cmd[] = {S25FL512S_READ_STATUS_REG1_CMD, 0x00};
  uint8_t in[sizeof(readsr1cmd)]    = {};

  Select();
  m_spi.Transceive(readsr1cmd, in, sizeof(readsr1cmd));
  Deselect();

  if (e_err)
    *e_err = (in[1] & S25FL512S_SR1_ERERR) == S25FL512S_SR1_ERERR;

  if (p_err)
    *p_err = (in[1] & S25FL512S_SR1_PGERR) == S25FL512S_SR1_PGERR;

  return (in[1] & S25FL512S_SR1_WIP) == S25FL512S_SR1_WIP;
}

/*! Executes a single page write command, a single page is 512 bytes.
    Only a single page can be written at at time, and any writes will stay within the page.
    \param address The address where the data should be written.
    \param buf The buffer that should be written to address.
    \param size The number of bytes that should be written to address.
    \return true if the write succeeded, false if it failed.*/
bool CSPIFlash_S25FL512S::WritePage(uint32_t address, const uint8_t* buf, uint32_t size)
{
  //wait for the previous page write to finish
  if (!WaitWriteFinished())
  {
    PrintLog(LOGSPIFLASH, "ERROR: unable to write %" PRIu32 " bytes at address %" PRIu32 "\n",
             address, size);
    return false;
  }

  //set the Write Enable Latch (WEL) to enable writes
  EnableWrite();

  uint8_t writecmd[] = {S25FL512S_PAGE_PROG_4_BYTE_ADDR_CMD, (uint8_t)(address >> 24),
                        (uint8_t)(address >> 16), (uint8_t)(address >> 8), (uint8_t)address};

  //send a page write command, and the bytes that should be written to the page
  Select();
  m_spi.Transceive(writecmd, nullptr, sizeof(writecmd));
  m_spi.Transceive(buf, nullptr, size);
  Deselect();

  return true;
}

bool CSPIFlash_S25FL512S::WaitWriteFinished()
{
  //wait for the previous write to finish
  uint32_t start = Millis();
  bool     e_err;
  bool     p_err;
  while(WriteInProgress(&e_err, &p_err))
  {
    if (e_err || p_err)
    { //return false if a programming error occurred
      PrintLog(LOGSPIFLASH, "ERROR: flash memory signaled a write error, e_err:%i p_err:%i\n", e_err, p_err);
      return false;
    }

    if (Millis() - start > 3000)
    { //return false if the page programming has not completed after 3 seconds,
      //the maximum time it should take is 340 microseconds for a page write,
      //and 2600 milliseconds for an erase
      PrintLog(LOGSPIFLASH, "ERROR: timeout waiting for write to complete\n");
      return false;
    }
  }

  return true;
}
