#include "spiflashlog.h"
#include "log.h"
#include "lock.h"
#include "stm32wss.h"
#include "backupregisters.h"

#include <cassert>
#include <algorithm>

CSPIFlashLog g_spiflashlog;

CSPIFlashLog::CSPIFlashLog()
{
  m_enabled    = false;
  m_spiflash   = nullptr;
  m_queue      = nullptr;
  m_mutex      = 0;
  m_linesignal = 0;
  m_writeaddr  = 0;
  m_free       = 0;
}

CSPIFlashLog::~CSPIFlashLog()
{
}

void CSPIFlashLog::Setup(CSPIFlash* spiflash)
{
  if (!spiflash->Enabled())
  {
    PrintLog(LOGFLASHLOG, "ERROR: not enabling SPI flash log, SPI flash not enabled\n");
    return;
  }

  m_spiflash   = spiflash;
  m_queue      = xQueueCreate(LOGQUEUESIZE, sizeof(char));
  m_mutex      = xSemaphoreCreateRecursiveMutex();
  m_linesignal = xSemaphoreCreateCounting(LOGQUEUESIZE, 0);

  //set the enabled flag only after everything has been allocated
  m_enabled = true;

  //make sure that the reserved address and size for the log are on sector boundaries
  assert((LOGFLASH_ADDRESS % m_spiflash->SectorSize()) == 0);
  assert((LOGFLASH_RESERVED % m_spiflash->SectorSize()) == 0);

  PrintLog(LOGFLASHLOG, "****** SPI FLASH LOG ENABLED ******\n");

  //if the log write address and free space are stored in the battery backed
  //registers, print their values
  uint32_t writeaddr = GetSPIFlashLogBackupWriteAddr();
  uint32_t freespace = GetSPIFlashLogBackupFreeSpace();
  if (writeaddr != 0 && freespace != 0)
    PrintLog(LOGFLASHLOG, "Battery backed state: write address:%" PRIu32 ", free space:%" PRIu32 "\n",
             writeaddr - 1, freespace - 1);
  else
    PrintLog(LOGFLASHLOG, "Battery backed state: not stored\n");

#if EXPORT_LOG
  //If enabled during compile-time, read the stored log from the SPI-flash and
  //write it out the serial port. It will start at the last known read address,
  //and print the entire memory block reserved for the log. This makes sure that
  //the entire log is exported, even if the battery backed counters are wrong.
  uint32_t bufsize = 1024;
  uint8_t buf[bufsize + 1];
  for (uint32_t i = 0; i < LOGFLASH_RESERVED; i += bufsize)
  {
    uint32_t readaddr = ((writeaddr + freespace + i) % LOGFLASH_RESERVED) + LOGFLASH_ADDRESS;
    m_spiflash->Read(readaddr, buf, bufsize);
    buf[bufsize] = 0;
    DEBUGUSARTprintf("%s", buf);
    IWDG_ReloadCounter();
  }

  for(;;)
    IWDG_ReloadCounter();
#endif

  //create a task that writes log characters to the SPI flash
  Create("FlashLog");
}

/*! Timeout in milliseconds for writing debug characters to the log queue,
    because it takes about half a second to erase a log sector, a one second timeout
    should be enough, this will avoid a complete freeze in case of a deadlock.*/
#define DEBUG_WRITE_TIMEOUT 1000

/*! This method is eventually called for every character written to stdout, through
    __io_putchar(), which is defined in log.cpp*/
void CSPIFlashLog::WriteDebugLogChar(char ch)
{
  if (!m_enabled)
    return;

  //lock to make the access to m_queue and m_linesignal atomic
  //because the order in which they are signaled matters
  CLock lock(m_mutex);

  //if a newline is written, or the queue is very full, signal the spi flash log task
  //to read data from the queue and to write it to the spi flash
  if (ch == '\n' || uxQueueMessagesWaiting(m_queue) >= LOGQUEUETRIGGER)
  {
    if (uxQueueSpacesAvailable(m_queue) == 0)
    { //if the queue is completely full, signal the spi flash log task first
      //otherwise the write to the queue will cause a deadlock
      xSemaphoreGive(m_linesignal);
      xQueueSendToBack(m_queue, &ch, pdMS_TO_TICKS(DEBUG_WRITE_TIMEOUT));
    }
    else
    { //if there is space in the queue, write the character first,
      //then signal the spi flash log task, so that it always reads a complete line
      xQueueSendToBack(m_queue, &ch, pdMS_TO_TICKS(DEBUG_WRITE_TIMEOUT));
      xSemaphoreGive(m_linesignal);
    }
  }
  else
  { //write single characters to the queue, without signaling the spi flash log
    xQueueSendToBack(m_queue, &ch, pdMS_TO_TICKS(DEBUG_WRITE_TIMEOUT));
  }
}

/*! This method is called by the spi flash log task, it will read bytes from m_queue
    and write them to the reserved log section on the SPI flash.
    WARNING: this task can't write to the log, it will cause deadlocks.*/
void CSPIFlashLog::Process()
{
  //read the log write address and free space from the battery backed registers,
  //if they're invalid, read the reserved spi flash section to
  //check where the log write should start, and the size of the already written log
  if (!ReadBackupWriteAddressAndFreeSpace())
    FindLogWriteAddressAndSize();

  //check if there is enough free space for the log
  CheckErase();

  uint8_t buf[512];
  for(;;)
  {
    //wait until the semaphore is signaled, that means that there is a full line in the queue to write
    //if the semaphore has not been signaled for one second, check if there is data in the queue anyway
    xSemaphoreTake(m_linesignal, pdMS_TO_TICKS(1000));

    while (uxQueueMessagesWaiting(m_queue) > 0)
    {
      //lock to protect m_spiflash, m_free and m_writeaddr
      CLock lock(*m_spiflash);

      //calculate how many bytes can be written until the end of the log section
      //if enough bytes are in the queue that m_writeaddr has to be wrapped,
      //this loop will iterate twice
      uint32_t bytestoend = LOGFLASH_ADDRESS + LOGFLASH_RESERVED - m_writeaddr;

      //read until the buffer is full, the queue is empty, enough bytes are read
      //to write to the end of the log section, or until the log section on the flash is almost full
      //there should always be at least one byte in the log flash section that is 0xFF, because that
      //signals where any new bytes should be written
      uint32_t maxreadbytes = std::min(std::min((uint32_t)sizeof(buf), bytestoend), m_free - 1);
      uint32_t bytesread = 0;
      while (bytesread < maxreadbytes && xQueueReceive(m_queue, buf + bytesread, 0))
      {
        //make sure that 0xFF is never written, because that is what marks free bytes
        if (buf[bytesread] == 0xFF)
          buf[bytesread] = '?';

        bytesread++;
      }

#if EXPORT_LOG
      continue; //Disable writing the flash log when the export is enabled.
#endif

      //clear the write address and free space from the battery backed registers
      //in case of a power loss during the spi write
      SetSPIFlashLogBackupWriteAddr(0);
      SetSPIFlashLogBackupFreeSpace(0);

      //write the log bytes to the SPI flash
      m_spiflash->Write(m_writeaddr, buf, bytesread);

      //update the write address, wrap to the beginning if the end was hit
      m_writeaddr += bytesread;
      if (m_writeaddr >= LOGFLASH_ADDRESS + LOGFLASH_RESERVED)
        m_writeaddr -= LOGFLASH_RESERVED;

      //update the number of free log space
      m_free -= bytesread;

      //write the new write address and free space as +1 encoded
      SetSPIFlashLogBackupWriteAddr(m_writeaddr + 1);
      SetSPIFlashLogBackupFreeSpace(m_free + 1);

      //check if a sector of the log space should be erased
      CheckErase();
    }
  }
}

/*! Reads the write address and free space counters from the reserved battery backed registers,
    checks their values, and stores them in m_writeaddr and m_free.
    \return true if the values are correct, false if FindLogWriteAddressAndSize() should be called instead.*/
bool CSPIFlashLog::ReadBackupWriteAddressAndFreeSpace()
{
  //read the write address and the free space from the battery backed registers
  uint32_t writeaddr = GetSPIFlashLogBackupWriteAddr();
  uint32_t freespace = GetSPIFlashLogBackupFreeSpace();

  //if they're not set to 0, they should hold valid values, that are +1 encoded
  //the reason for the +1 encoding is that if battery backed power is lost,
  //all battery backed registers are cleared to 0
  if (writeaddr != 0 && freespace != 0)
  {
    writeaddr--;
    freespace--;

    //check if the write address and free space registers hold valid values
    if (writeaddr >= LOGFLASH_ADDRESS &&
        writeaddr <= LOGFLASH_ADDRESS + LOGFLASH_RESERVED &&
        freespace <= LOGFLASH_RESERVED)
    {
      m_writeaddr = writeaddr;
      m_free      = freespace;
      return true;
    }
  }

  return false;
}

/*! The section of the SPI flash that is reserved for log writing is used as a ringbuffer,
    this function finds the section that is set to 0xFF, which is where the new log line should be written.*/
void CSPIFlashLog::FindLogWriteAddressAndSize()
{
  CLock lock(*m_spiflash);

  m_writeaddr = LOGFLASH_ADDRESS; //default write address of the log section of the SPI flash

  //find the first byte that is set to 0xFF
  int32_t logwriteaddr = m_spiflash->FindByte(LOGFLASH_ADDRESS, LOGFLASH_RESERVED, 0xFF);

  if (logwriteaddr == -1)
  { //if no byte at 0xFF could be found, erase the first sector, the log write starts there
    m_spiflash->Erase(LOGFLASH_ADDRESS / m_spiflash->SectorSize(), 1);

    //the free space is now exactly one sector
    m_free = m_spiflash->SectorSize();
  }
  else if (logwriteaddr == LOGFLASH_ADDRESS)
  { //if the first byte is 0xFF, it's possible that the log write address starts somewhere in the middle
    //so find the first byte that is not 0xFF, that is where the log starts
    int32_t logstartaddr = m_spiflash->FindByte(LOGFLASH_ADDRESS, LOGFLASH_RESERVED, 0xFF, MatchType_Not);

    if (logstartaddr != -1)
    { //find the first byte that is 0xFF after the start of the log, if that exists,
      //that is where the log write should start
      int32_t searchsize = LOGFLASH_RESERVED - (logstartaddr - LOGFLASH_ADDRESS);
      int32_t logendaddr = m_spiflash->FindByte(logstartaddr, searchsize, 0xFF);
      if (logendaddr != -1)
      {
        m_writeaddr = logendaddr; //where the log write should start

        //calculate the size of the log, and how many bytes are still free
        int32_t logsize = logendaddr - logstartaddr;
        m_free = LOGFLASH_RESERVED - logsize;
      }
      else
      { //if all bytes from the start of the log to the end of the log section are not 0xFF,
        //then the free bytes are all bytes before the log start address
        m_free = logstartaddr - LOGFLASH_ADDRESS;
      }
    }
    else
    { //if all bytes are 0xFF, they are all free for writing
      m_free = LOGFLASH_RESERVED;
    }
  }
  else
  { //if the first byte at 0xFF is somewhere in the middle, the log write starts there
    m_writeaddr = logwriteaddr;

    //find the first byte after the first 0xFF that is not 0xFF, the log starts there
    int32_t searchsize = LOGFLASH_RESERVED - (logwriteaddr - LOGFLASH_ADDRESS);
    int32_t logstartaddr = m_spiflash->FindByte(m_writeaddr, searchsize, 0xFF, MatchType_Not);
    if (logstartaddr != -1)
    {
      //the number of free bytes is in between the log bytes
      m_free = logstartaddr - logwriteaddr;
    }
    else
    { //if all bytes until the end of the log section are 0xFF, they are all free for writing
      m_free = LOGFLASH_RESERVED - (logwriteaddr - LOGFLASH_ADDRESS);
    }
  }

  //write the calculated write address and free space as +1 encoded
  SetSPIFlashLogBackupWriteAddr(m_writeaddr + 1);
  SetSPIFlashLogBackupFreeSpace(m_free + 1);
}

/*! Checks how much free space in the SPI flash log section there is,
    if it's below a certain point, the sector that holds the oldest log data is erased.
    \param sectorfree When true, make sure that there is at least one free sector.*/
void CSPIFlashLog::CheckErase(bool sectorfree /*= false*/)
{
  if (!m_enabled)
    return;

  //lock to protect m_spiflash, m_free and m_writeaddr
  CLock lock(*m_spiflash);

  //if sectorfree is true, check if there is at least one full sector free
  //otherwise, check if there is at least one full queue size free
  if ((sectorfree && (m_free < m_spiflash->SectorSize())) || (m_free < LOGQUEUESIZE))
  {
    //calculate where the first byte of the log is
    uint32_t logstartaddr = m_writeaddr + m_free;
    if (logstartaddr >= LOGFLASH_ADDRESS + LOGFLASH_RESERVED)
      logstartaddr -= LOGFLASH_RESERVED;

    uint32_t sector;
    if ((logstartaddr % m_spiflash->SectorSize()) == 0)
    {
      //the log should always start at a multiple of the sector size
      sector = logstartaddr / m_spiflash->SectorSize();
    }
    else
    {
      //if the log doesn't start at a multiple of the sector size,
      //something went very wrong, in that case erase the first sector and reset the state
      sector      = LOGFLASH_ADDRESS / m_spiflash->SectorSize();
      m_writeaddr = LOGFLASH_ADDRESS;
      m_free      = 0;

      SetSPIFlashLogBackupWriteAddr(0);
      SetSPIFlashLogBackupFreeSpace(0);
    }

    //erase the sector and update the free space counter
    m_spiflash->Erase(sector, 1);
    m_free += m_spiflash->SectorSize();

    //store the new free space value in the battery backed register as +1 encoded
    SetSPIFlashLogBackupFreeSpace(m_free + 1);

    //print to the log that the SPI flash log sector is being erased,
    //but only if this method is called from a different task,
    //because logging from the SPI flash log task will cause a deadlock
    if (m_task != xTaskGetCurrentTaskHandle())
    {
      //don't hold a lock on m_spiflash while printing to the log, it might deadlock with the spi flash log task
      lock.Leave();
      PrintLog(LOGFLASHLOG, "SPI flash log has %" PRIu32 " bytes free, erased sector %" PRIu32 "\n",
               m_free, sector);
    }
  }
  else
  {
    if (m_task != xTaskGetCurrentTaskHandle())
    {
      //don't hold a lock on m_spiflash while printing to the log, it might deadlock with the spi flash log task
      lock.Leave();
      PrintLog(LOGFLASHLOG, "SPI flash log has %" PRIu32 " bytes free, not erasing\n", m_free);
    }
  }
}
