#include "stm32wss.h"

#include <FreeRTOS.h>
#include <task.h>
#include <timers.h>

#include "log.h"
#include "lock.h"
#include "clock.h"
#include "random.h"
#include "gpio.h"
#include "dma.h"
#include "safeheap.h"
#include "spiflashlog.h"
#include "crc32hw.h"
#include "internalflash.h"
#include "backupregisters.h"
#include "aes.h"
#include "internalflash.h"
#include "otpaddrs.h"

#include <cerrno>
#include <cstdlib>
#include <cassert>
#include <cstdio>
#include <cstring>
#include <cctype>

void ProgramDeviceID();
void PreSchedulerStart();
void Startup();
void HeartBeat();

void SystemHeartBeat(TimerHandle_t timer);
#define HEARTBEAT_INTERVAL pdMS_TO_TICKS(500)

//! Set to true when the SOFTWARE_REBOOTTIME_FLAG battery backed flag is set at boot.
static bool g_softwarereboottimehappened;

/*! Should be called from main(), sets up everything to the point where the FreeRTOS kernel is started.*/
void BoardInit()
{
  //set up the watchdog timer, do this first so that everything else gets protected by it
  //the watchdog timer is also set up from the boot firmware, it's set up here again to be safe
  SetupWatchdog();

  //Ensure all priority bits are assigned as preemption priority bits.
  //This is necessary for the FreeRTOS port of the Cortex M4
  NVIC_PriorityGroupConfig(NVIC_PriorityGroup_4);

  //enable the clock to the system configuration controller
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_SYSCFG, ENABLE);

  //Store in ram why the microcontroller booted.
  ResetFlagsStore();

  //initialize the hardware random number generator
  InitRNG();

  //enable the clock for the hardware crc generator
  CCRC32Hw::EnableClock();

  //initialize all gpio pins, this should be done before setting up the debug usart
  GPIOInit();

  //initialize the 32 bit microsecond clock
  ClockInit();

  //check if a device ID should be programmed
  ProgramDeviceID();

  //initialize the debug usart, and initialize the RTC right after
  //don't change the order of these calls, they depend on each other
  SetupDebugUsart();
  PrintLog(LOGMAIN, "\n********** BOARD STARTED **********\n");

  //Enable access to the backup domain and start the RTC.
  EnableBackupAccess();
  RTCInit();
  StoreRTCAlarmBootFlags();
 // GPIOInit(); // nog een keer
  //Store in ram if the current boot happened because of a software time reboot.
  g_softwarereboottimehappened = ReadBackupFlag(SOFTWARE_REBOOTTIME_FLAG);
  SetBackupFlag(SOFTWARE_REBOOTTIME_FLAG, false);

  //set up the DMA controller
  g_dma.Setup();

  //callback into the main program, where it can set things up before the FreeRTOS scheduler is started
  PreSchedulerStart();

  //create the heartbeat timer, this will signal the idle task that the FreeRTOS kernel is working
  TimerHandle_t heartbeattimer = xTimerCreate("SystemHeartBeat", HEARTBEAT_INTERVAL, pdTRUE, NULL, SystemHeartBeat);
  xTimerStart(heartbeattimer, HEARTBEAT_INTERVAL);

  //initialize the heap locking for the FreeRTOS kernel
  InitSafeHeap();

  //Allocate needed mutexes.
  CCRC32Hw::SetupMutex();
  SetupAESMutex();
  InitFlashMutex();

  //run the FreeRTOS kernel, this will also enable interrupts
  vTaskStartScheduler();

  //the FreeRTOS kernel should not exit, print an error in case it does
  DEBUGUSARTprintf("\n********* SCHEDULER EXITED ********\n");
  for(;;);
}

/*! Writes a string to DEBUGUSART.
    \param str A pointer to a string to write to DEBUGUSART.*/
static void WriteUSARTStr(const char* str)
{
  uint32_t length = strlen(str);
  for (uint32_t i = 0; i < length; i++)
  { //write a byte to the usart
    while (!USART_GetFlagStatus(DEBUGUSART, USART_FLAG_TXE));
    USART_SendData(DEBUGUSART, str[i]);
  }
}

/* Clears any pending input bytes from DEBUGUSART.*/
static void ClearUsartInput()
{
  uint32_t start = Micros();
  do
  {
    while (USART_GetFlagStatus(DEBUGUSART, USART_FLAG_RXNE))
      USART_ReceiveData(DEBUGUSART);
  }
  while (Micros() - start < 100000);
}

/*! Reads a line from DEBUGSART and stores it in str.
    \param str Pointer to a buffer to write the line to.
    \param maxlength The maximum number of bytes to write to str.*/
static void ReadUSARTLine(char* str, uint32_t maxlength)
{
  //clear any pending input bytes, so that any data sent accidentally is ignored
  ClearUsartInput();

  str[0] = 0;
  uint32_t bytesread = 0;
  for(;;)
  {
    //wait for an input byte, reset the watchdog timer while waiting
    while (!USART_GetFlagStatus(DEBUGUSART, USART_FLAG_RXNE))
      IWDG_ReloadCounter();

    char input = USART_ReceiveData(DEBUGUSART);

    //terminate on a line ending or carriage return
    if (input == '\n' || input == '\r')
    { //echo back \r\n
      while (!USART_GetFlagStatus(DEBUGUSART, USART_FLAG_TXE));
      USART_SendData(DEBUGUSART, '\r');
      while (!USART_GetFlagStatus(DEBUGUSART, USART_FLAG_TXE));
      USART_SendData(DEBUGUSART, '\n');
      return;
    }

    //echo the received byte to the sender
    while (!USART_GetFlagStatus(DEBUGUSART, USART_FLAG_TXE));
    USART_SendData(DEBUGUSART, input);

    //store the received byte and null-terminate str
    str[bytesread] = input;
    bytesread++;
    str[bytesread] = 0;

    if (bytesread + 1 == maxlength)
      return;
  }
}

/*! Reads a decimal value from DEBUGUSART, interprets it as hexadecimal,
    and stores it intval.
    \return true if the value is correct.
    \return false if the value is incorrect.*/
bool ReadUSARTInteger(uint32_t* intval)
{
  char buf[32];
  ReadUSARTLine(buf, sizeof(buf));
  
  //a 32 bit hexadecimal number should be entered, more than 8 characters is too much
  if (strlen(buf) > 8)
    return false;

  //check if all read characters are correct hexadecimal values
  char* bufptr = buf;
  while (*bufptr)
  {
    if (!isdigit(*(bufptr++)))
      return false;
  }
  
  int retval = sscanf(buf, "%" SCNx32, intval);
  return retval == 1;
}

/*! Checks if a device ID received over the USART is valid.
    \return true The device ID is valid.
    \return false the device ID is not valid.*/
bool CheckDeviceIDValid(uint32_t deviceid)
{
  if (deviceid < 0x1000000 || deviceid >= 0x2000000)
  {
    WriteUSARTStr("\r\nERROR: Device ID should be between 1000000 and 2000000\r\n");
    return false;
  }

  return true;
}

/*! Checks if a device ID should be programmed in the OTP.
    If so, it will ask for it to be input over the debug usart.
    It will ask for the device ID twice, to make sure it's entered correctly.*/
void ProgramDeviceID()
{
#ifdef UPDATER
  return; //don't program the device ID from the updater firmware
#endif

  //if the debug usart is ever changed, the call to RCC_APB2PeriphClockCmd
  //should be changed to enable a different clock
  assert(DEBUGUSART == USART1);

  if (*(uint32_t*)DEVICE_ID_PROGRAM != DEVICE_ID_INVALID)
    return; //Device ID is already programmed.

  //enable the clock for the debug USART
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_USART1, ENABLE);

  //setup DEBUGUSART as 115k2 8n1
  USART_InitTypeDef usartinit;
  USART_StructInit(&usartinit);
  usartinit.USART_BaudRate = 115200;
  USART_Init(DEBUGUSART, &usartinit);
  USART_Cmd(DEBUGUSART, ENABLE);

  //delay to allow the terminal software on the PC to start
  //after the firmware programming has completed
  DelayUS(5000000);

  for(;;)
  {
    char buf[128];
    sprintf(buf, "\r\nDevice ID not written\r\nEnter device ID:\r\n");
    WriteUSARTStr(buf);

    //read the device ID from the usart
    uint32_t deviceid1;
    if (!ReadUSARTInteger(&deviceid1) || !CheckDeviceIDValid(deviceid1))
    {
      WriteUSARTStr("\r\nERROR: wrong device ID entered\r\n");
      continue;
    }

    sprintf(buf, "Read device ID %" PRIX32 "\r\nEnter device ID again:\r\n", deviceid1);
    WriteUSARTStr(buf);

    //read the device ID again from the usart
    uint32_t deviceid2;
    if (!ReadUSARTInteger(&deviceid2) || !CheckDeviceIDValid(deviceid2))
    {
      WriteUSARTStr("\r\nERROR: wrong device ID entered\r\n");
      continue;
    }

    //check if they match, if so it can be safely written to the OTP
    if (deviceid1 != deviceid2)
    {
      sprintf(buf, "\r\nERROR: device IDs %" PRIX32 " and %" PRIX32 " don't match\r\n",
              deviceid1, deviceid2);
      WriteUSARTStr(buf);
      continue;
    }

    //check if the device id is not a reserved ID
    if (deviceid1 == DEVICE_ID_INVALID ||
        (deviceid1 >= DEVICE_ID_RESERVED_START && deviceid1 <= DEVICE_ID_RESERVED_END))
    {
      sprintf(buf, "\r\nERROR: device ID %" PRIX32 " is a reserved value\r\n", deviceid1);
      WriteUSARTStr(buf);
      continue;
    }

    sprintf(buf, "\r\nWriting device ID %" PRIX32 "\r\n", deviceid1);
    WriteUSARTStr(buf);

    //write the device ID to the OTP, and lock the corresponding OTP section
    uint8_t lock = 0;
    FlashWrite(DEVICE_ID_PROGRAM, (uint8_t*)&deviceid1, sizeof(deviceid1));
    FlashWrite(DEVICE_ID_PROGRAM_LOCK, &lock, sizeof(lock));

    //reboot the microcontroller
    DelayUS(500000);
    NVIC_SystemReset();
  }
}

/*! This callback is called from FreeRTOS, before any other tasks is started.*/
extern "C" void vApplicationDaemonTaskStartupHook()
{
  //switch the usart to async mode, because all other write to the usart will be from a FreeRTOS task at after this
  SetDebugUsartAsync(true);

  //callback into the main program, where it can set up other things before starting the FreeRTOS tasks
  Startup();
}

/*! Sets up the watchdog timer with the maximum prescaler value and maximum counter value.
    This lets the watchdog timer expire after roughly 32.7 seconds.*/
void SetupWatchdog()
{
  IWDG_WriteAccessCmd(IWDG_WriteAccess_Enable);

  while(IWDG_GetFlagStatus(IWDG_FLAG_PVU));
  IWDG_SetPrescaler(IWDG_Prescaler_256);

  while(IWDG_GetFlagStatus(IWDG_FLAG_RVU));
  IWDG_SetReload(0xFFF);

  IWDG_Enable();
}

static int64_t g_reboottime;

/*! Sets a time of Millis() since the system started at which NVIC_SystemReset() is called.
    \param rebootmillis The time after which the system is rebooted, use 0 to disable.*/
void SetRebootTime(int64_t rebootmillis)
{
  g_reboottime = rebootmillis;
}

/*! \return true if the current reboot happened because
                 of the time expired by SetRebootTime().
    \return false if the current reboot did not happen because of
                  the time set by SetRebootTime().*/
bool SoftwareRebootTimeHappened()
{
  return g_softwarereboottimehappened;
}

static volatile int64_t g_heartbeat;

/*! Callback function for a timer set up in main(),
    it will set g_heartbeat to Millis(), signaling the FreeRTOS idle task to reset the watchdog timer.
    WARNING: This function must never block.
\param xTimer Handle to the FreeRTOS blink timer.*/
void SystemHeartBeat(TimerHandle_t timer)
{
  //update the 64 bit millisecond timer
  int64_t now = Millis();

  //signal to the application idle hook that the heartbeat timer is running
  CCritical criticalsection;
  g_heartbeat = now;
  criticalsection.Leave();

  //if a reboot time is set, check if it's time to reboot
  if (g_reboottime && now >= g_reboottime)
  {
    SetBackupFlag(SOFTWARE_REBOOTTIME_FLAG, true);
    NVIC_SystemReset();
  }

  //call the application heartbeat
  HeartBeat();
}

/*! Callback function for the FreeRTOS idle hook. it gets called when nothing else is scheduled to run.
    In this function, the watchdog timer is reset using IWDG_ReloadCounter() when
    g_heartbeat is set to Millis() from the heartbeat timer.
    This way, the watchdog timer makes sure that not all cpu cycles are used by one or more tasks,
    and that FreeRTOS is still running the scheduler.
    Also, it guarantees that the Millis() function is ticking forward in time.
    WARNING: This function must never block, because the FreeRTOS kernel always needs
             at least one task that is ready to run, and not blocking on anything.
             So it must not block on any queue/semaphore/delay etc.*/
extern "C" void vApplicationIdleHook()
{
#if configUSE_IDLE_HOOK != 1 || configIDLE_SHOULD_YIELD != 1
  #error configUSE_IDLE_HOOK and configIDLE_SHOULD_YIELD should both be set to 1
#endif

  static int64_t prevheartbeat;

  //Use test and test-and-set to prevent entering a
  //critical section when not needed.
  if (g_heartbeat > prevheartbeat)
  {
    CCritical criticalsection;
    if (g_heartbeat > prevheartbeat)
    {
      IWDG_ReloadCounter(); //reset the watchdog timer
      prevheartbeat = g_heartbeat;
    }
  }
}

/*! Array which is used as heap, this has to be a global so that the linker will allocate memory for it.
    This is required if malloc() and free() need to be used together with FreeRTOS.
    configTOTAL_HEAP_SIZE is a #define in FreeRTOSConfig.h, it sets the size of the heap.*/
static uint8_t g_heap[configTOTAL_HEAP_SIZE];

/*! This function is called by the heap algorithm when more ram is needed.
    \param incr the number of bytes that should be extra allocated for the heap.
    \return A pointer to the start of the block of ram that was extra allocated, or -1 if the allocation failed.*/
extern "C" caddr_t _sbrk(int incr)
{
  static uint32_t heapsize = 0;
  static uint8_t* heapend = g_heap;

  //check if the heap can grow by incr bytes
  if (heapsize + incr > configTOTAL_HEAP_SIZE)
  {
    errno = ENOMEM;
    return (caddr_t)-1;
  }

  uint8_t* heapptr = heapend;
  heapsize += incr;
  heapend += incr;

  return (caddr_t)heapptr;
}

/*! Called in case of a fatal error, will print an error message and reset the microcontroller.*/
void FatalError()
{
  DEBUGUSARTprintf("\n*********** FATAL ERROR ***********\n");

  __disable_irq(); //prevent anything else from running at this point

  //Wait for the watchdog timer reset. This is used so that at the next boot,
  //a watchdog timer reset can be detected. If this happens too much,
  //a factory reset can possibly be done.
  DelayUS(120 * 1000000);

  //In case the watchdog timer reset doesn't happen.
  NVIC_SystemReset();
  for(;;);
}

/*! Called by FreeRTOS when a task overflows the stack.
    \param pxTask a pointer to the task that overflowed the stack.
    \param pcTaskName the name of the task that overflowed the stack.*/
extern "C" void vApplicationStackOverflowHook(TaskHandle_t *pxTask, signed char *pcTaskName)
{
#if configCHECK_FOR_STACK_OVERFLOW != 2
  #error configCHECK_FOR_STACK_OVERFLOW should be set to 2
#endif
  DEBUGUSARTprintf("\n********* STACK OVERFLOW **********\n");
  DEBUGUSARTprintf("Task %s overflowed the stack\n", pcTaskName);
  FatalError();
}

/*! Called by FreeRTOS when malloc failed.*/
extern "C" void vApplicationMallocFailedHook()
{
#if configUSE_MALLOC_FAILED_HOOK != 1
  #error configUSE_MALLOC_FAILED_HOOK should be set to 1
#endif
  DEBUGUSARTprintf("\n********** MALLOC FAILED **********\n");
  FatalError();
}

/*! Called from HardFault_Handler(), using asm.
    If link time optimization is enabled, the symbol for this function is
    optimized away, because it's not called by any C/C++ code, this causes
    a linker error. To prevent this, the used attribute is set.*/
extern "C" void __attribute__((used))
  HardFault_Handler_C(uint32_t* hardfault_args)
{
  uint32_t stacked_r0;
  uint32_t stacked_r1;
  uint32_t stacked_r2;
  uint32_t stacked_r3;
  uint32_t stacked_r12;
  uint32_t stacked_lr;
  uint32_t stacked_pc;
  uint32_t stacked_psr;

  //Exception stack frame
  stacked_r0 = ((uint32_t) hardfault_args[0]);
  stacked_r1 = ((uint32_t) hardfault_args[1]);
  stacked_r2 = ((uint32_t) hardfault_args[2]);
  stacked_r3 = ((uint32_t) hardfault_args[3]);
 
  stacked_r12 = ((uint32_t) hardfault_args[4]);
  stacked_lr  = ((uint32_t) hardfault_args[5]);
  stacked_pc  = ((uint32_t) hardfault_args[6]);
  stacked_psr = ((uint32_t) hardfault_args[7]);
 
  //when needed, the PC (program counter) value can be used to check where
  //the crash happened, using the addr2line tool, example:
  //addr2line -e Accello.elf 0x803aeae

  DEBUGUSARTprintf("\n******** HARDFAULT HANDLER ********\n");
  DEBUGUSARTprintf("R0   = %" PRIX32 "\n", stacked_r0);
  DEBUGUSARTprintf("R1   = %" PRIX32 "\n", stacked_r1);
  DEBUGUSARTprintf("R2   = %" PRIX32 "\n", stacked_r2);
  DEBUGUSARTprintf("R3   = %" PRIX32 "\n", stacked_r3);
  DEBUGUSARTprintf("R12  = %" PRIX32 "\n", stacked_r12);
  DEBUGUSARTprintf("LR   = %" PRIX32 "\n", stacked_lr);
  DEBUGUSARTprintf("PC   = %" PRIX32 "\n", stacked_pc);
  DEBUGUSARTprintf("PSR  = %" PRIX32 "\n", stacked_psr);
  DEBUGUSARTprintf("BFAR = %" PRIX32 "\n", (*((uint32_t*)(0xE000ED38))));
  DEBUGUSARTprintf("CFSR = %" PRIX32 "\n", (*((uint32_t*)(0xE000ED28))));
  DEBUGUSARTprintf("HFSR = %" PRIX32 "\n", (*((uint32_t*)(0xE000ED2C))));
  DEBUGUSARTprintf("DFSR = %" PRIX32 "\n", (*((uint32_t*)(0xE000ED30))));
  DEBUGUSARTprintf("AFSR = %" PRIX32 "\n", (*((uint32_t*)(0xE000ED3C))));

  FatalError();
}

/*! Called on a hardfault from the Cortex M4.
    This function will set up a pointer to call HardFault_Handler_C(),
    which can then extract the values of the registers during the crash.
    Copied from: https://nathan.vertile.com/blog/2016/05/04/stm32-hardfault-debugging/
    */
extern "C" void __attribute__ ((naked)) HardFault_Handler()
{
  asm volatile(
  "  movs r0,#4\n"
  "  movs r1, lr\n"
  "  tst r0, r1\n"
  "  beq _MSP\n"
  "  mrs r0, psp\n"
  "  b _HALT\n"
  "_MSP:\n"
  "  mrs r0, msp\n"
  "_HALT:\n"
  "  ldr r1,[r0,#20]\n"
  "  b HardFault_Handler_C\n"
  "  bkpt #0\n"
  );
}

/*! Called by the C library when assert() wants to exit.*/
extern "C" void _exit(int status)
{
  //make sure any message from assert() has been written
  LockLog();
  fflush(stdout);
  fflush(stderr);
  UnlockLog();
  WaitLogWriteComplete(1000);

  DEBUGUSARTprintf("\n******* EXIT FUNCTION CALLED ******\n");
  FatalError();
  for(;;); //prevent a compile warning, this function should not return
}

/*! Called by the assert_param() macro.
    \param file __FILE__ macro.
    \param line __LINE__ macro.*/
extern "C" void assert_failed(uint8_t* file, uint32_t line)
{
  DEBUGUSARTprintf("\n******** ASSERTION FAILED *********\n");
  DEBUGUSARTprintf("file:%s line:%" PRIu32 "\n", file, line);
  FatalError();
}

static uint32_t g_resetflags;

void ResetFlagsStore()
{
  //store the reset flags from the RCC into g_resetflags, so that the rcc
  //reset flags can be cleared using RCC_ClearFlag(), because otherwise
  //they are only cleared by a power reset
  //the reset flags can later be retrieved using the ResetFlags() function
  if (RCC_GetFlagStatus(RCC_FLAG_SFTRST) == SET)
    g_resetflags |= ResetFlagMask(RCC_FLAG_SFTRST);
  if (RCC_GetFlagStatus(RCC_FLAG_IWDGRST) == SET)
    g_resetflags |= ResetFlagMask(RCC_FLAG_IWDGRST);
  if (RCC_GetFlagStatus(RCC_FLAG_WWDGRST) == SET)
    g_resetflags |= ResetFlagMask(RCC_FLAG_WWDGRST);
  if (RCC_GetFlagStatus(RCC_FLAG_LPWRRST) == SET)
    g_resetflags |= ResetFlagMask(RCC_FLAG_LPWRRST);
  if (RCC_GetFlagStatus(RCC_FLAG_PORRST) == SET)
    g_resetflags |= ResetFlagMask(RCC_FLAG_PORRST);
  if (RCC_GetFlagStatus(RCC_FLAG_BORRST) == SET)
    g_resetflags |= ResetFlagMask(RCC_FLAG_BORRST);
  if (RCC_GetFlagStatus(RCC_FLAG_PINRST) == SET)
    g_resetflags |= ResetFlagMask(RCC_FLAG_PINRST);

  //clear all reset reason flags
  RCC_ClearFlag();
}

/*! Turns an RCC reset reason flag into a single set bit.*/
uint32_t ResetFlagMask(uint32_t reason)
{
  return 1 << (reason & 0x1F);
}

/*! Prints the reset reasons stored in g_resetflags.*/
void PrintResetReason()
{
  PrintLog(LOGMAIN, "Reset reasons:\n");
  uint32_t    resetflags = ResetFlags();
  const char* resetstr;
  while ((resetstr = ResetFlagStr(&resetflags)))
    PrintLog(LOGMAIN, "  %s\n", resetstr);
}

uint32_t ResetFlags()
{
  return g_resetflags;
}

/*! Converts a reset flag to a printable string, and clears the flag from resetflags.
    \param resetflags A pointer to a variable that is initialized with the return value of ResetFlags().
    \return A pointer to a string literal, or nullptr when there are no more reset flags.*/
const char* ResetFlagStr(uint32_t* resetflags)
{
  if (*resetflags & ResetFlagMask(RCC_FLAG_SFTRST))
  {
    *resetflags &= ~(ResetFlagMask(RCC_FLAG_SFTRST));
    return "software request";
  }
  else if (*resetflags & ResetFlagMask(RCC_FLAG_IWDGRST))
  {
    *resetflags &= ~(ResetFlagMask(RCC_FLAG_IWDGRST));
    return "independent watchdog timer";
  }
  else if (*resetflags & ResetFlagMask(RCC_FLAG_WWDGRST))
  {
    *resetflags &= ~(ResetFlagMask(RCC_FLAG_WWDGRST));
    return "windowed watchdog timer";
  }
  else if (*resetflags & ResetFlagMask(RCC_FLAG_LPWRRST))
  {
    *resetflags &= ~(ResetFlagMask(RCC_FLAG_LPWRRST));
    return "low-power management";
  }
  else if (*resetflags & ResetFlagMask(RCC_FLAG_PORRST))
  {
    *resetflags &= ~(ResetFlagMask(RCC_FLAG_PORRST));
    return "power-on";
  }
  else if (*resetflags & ResetFlagMask(RCC_FLAG_BORRST))
  {
    *resetflags &= ~(ResetFlagMask(RCC_FLAG_BORRST));
    return "brown-out";
  }
  else if (*resetflags & ResetFlagMask(RCC_FLAG_PINRST))
  {
    *resetflags &= ~(ResetFlagMask(RCC_FLAG_PINRST));
    return "reset pin";
  }
  else
  {
    return nullptr;
  }
}

/*! Used to print boot information to the SPI flash log.*/
void PrintBootInfo()
{
#ifdef UPDATER
  PrintLog(LOGMAIN, "***** UPDATER FIRMWARE STARTED ****\n");
#elif defined(ACCELLO)
  PrintLog(LOGMAIN, "***** ACCELLO FIRMWARE STARTED *****\n");
#elif defined(ECAT)
  PrintLog(LOGMAIN, "****** ECAT FIRMWARE STARTED ******\n");
#elif defined(TMT)
  PrintLog(LOGMAIN, "******* TMT FIRMWARE STARTED ******\n");
#else
  #error No firmware type defined
#endif

  PrintLog(LOGMAIN, "Firmware version: %s\n", FirmwareVersion());
  PrintLog(LOGMAIN, "Device ID: %s\n", DeviceIDStr());
  PrintLog(LOGMAIN, "SystemCoreClock: %" PRIu32 "\n", SystemCoreClock);
  PrintResetReason();
  PrintRTCTime();
}

/*! Returns the contents of OTP memory DEVICE_ID_ADDR2 if it's not 0xFFFFFFFF,
    otherwise returns the contents of OTP memory DEVICE_ID_ADDR1.
    The device ID is written to DEVICE_ID_ADDR1 in production, however if the device
    ID ever needs to be changed, a new device ID can be written to DEVICE_ID_ADDR2.*/
uint32_t DeviceID()
{
  if (*(uint32_t*)DEVICE_ID_ADDR2 != DEVICE_ID_INVALID)
    return *(uint32_t*)DEVICE_ID_ADDR2;
  else
    return *(uint32_t*)DEVICE_ID_ADDR1;
}

/*! Returns the device ID as ASCII hex.*/
const char* DeviceIDStr()
{
  static char deviceidstr[9];

  CCritical criticalsection;
  if (strlen(deviceidstr) == 0)
    snprintf(deviceidstr, sizeof(deviceidstr), "%" PRIX32, DeviceID());

  return deviceidstr;
}

/*! Returns the firmware version that should be defined in FreeRTOSConfig.h.*/
const char* FirmwareVersion()
{
#ifndef configFIRMWARE_VERSION
  #error configFIRMWARE_VERSION should be defined in FreeRTOSConfig.h
#endif

  return configFIRMWARE_VERSION;
}

/*! This function should never be called, it only exists to prevent a linker
    error when link time optimization is enabled.
    The vTaskSwitchContext() function is called from asm in FreeRTOS, but
    link time optimization removes the symbol if the function is not referenced
    from at least one other file, and it does not check if the function is
    called from asm.*/
static void __attribute__((used)) NoOptimize()
{
  vTaskSwitchContext();
}
