//this code is copied from the Musl project at http://www.musl-libc.org/
//the reason this is used instead of the standard C time functions,
//is because the current version of newlib has a 32 bit time_t,
//which will wrap at 03:14:08 UTC on 19 January 2038

#define _GNU_SOURCE //for strptime in C++

#include "timeconversion.h"
#include <climits>
#include <cstring>

/* 2000-03-01 (mod 400 year, immediately after feb29 */
#define LEAPOCH (946684800LL + 86400*(31+29))

#define DAYS_PER_400Y (365*400 + 97)
#define DAYS_PER_100Y (365*100 + 24)
#define DAYS_PER_4Y   (365*4   + 1)

int SecsToTm(int64_t t, struct tm *tm)
{
	int64_t days, secs, years;
	int remdays, remsecs, remyears;
	int qc_cycles, c_cycles, q_cycles;
	int months;
	int wday, yday, leap;
	static const char days_in_month[] = {31,30,31,30,31,31,30,31,30,31,31,29};

	/* Reject time_t values whose year would overflow int */
	if (t < INT_MIN * 31622400LL || t > INT_MAX * 31622400LL)
		return -1;

	secs = t - LEAPOCH;
	days = secs / 86400;
	remsecs = secs % 86400;
	if (remsecs < 0) {
		remsecs += 86400;
		days--;
	}

	wday = (3+days)%7;
	if (wday < 0) wday += 7;

	qc_cycles = days / DAYS_PER_400Y;
	remdays = days % DAYS_PER_400Y;
	if (remdays < 0) {
		remdays += DAYS_PER_400Y;
		qc_cycles--;
	}

	c_cycles = remdays / DAYS_PER_100Y;
	if (c_cycles == 4) c_cycles--;
	remdays -= c_cycles * DAYS_PER_100Y;

	q_cycles = remdays / DAYS_PER_4Y;
	if (q_cycles == 25) q_cycles--;
	remdays -= q_cycles * DAYS_PER_4Y;

	remyears = remdays / 365;
	if (remyears == 4) remyears--;
	remdays -= remyears * 365;

	leap = !remyears && (q_cycles || !c_cycles);
	yday = remdays + 31 + 28 + leap;
	if (yday >= 365+leap) yday -= 365+leap;

	years = remyears + 4*q_cycles + 100*c_cycles + 400LL*qc_cycles;

	for (months=0; days_in_month[months] <= remdays; months++)
		remdays -= days_in_month[months];

	if (months >= 10) {
		months -= 12;
		years++;
	}

	if (years+100 > INT_MAX || years+100 < INT_MIN)
		return -1;

	tm->tm_year = years + 100;
	tm->tm_mon = months + 2;
	tm->tm_mday = remdays + 1;
	tm->tm_wday = wday;
	tm->tm_yday = yday;

	tm->tm_hour = remsecs / 3600;
	tm->tm_min = remsecs / 60 % 60;
	tm->tm_sec = remsecs % 60;

	return 0;
}

int MonthToSecs(int month, int is_leap)
{
	static const int secs_through_month[] = {
		0, 31*86400, 59*86400, 90*86400,
		120*86400, 151*86400, 181*86400, 212*86400,
		243*86400, 273*86400, 304*86400, 334*86400 };
	int t = secs_through_month[month];
	if (is_leap && month >= 2) t+=86400;
	return t;
}

int64_t YearToSecs(int64_t year, int *is_leap)
{
	if (year-2ULL <= 136) {
		int y = year;
		int leaps = (y-68)>>2;
		if (!((y-68)&3)) {
			leaps--;
			if (is_leap) *is_leap = 1;
		} else if (is_leap) *is_leap = 0;
		return 31536000*(y-70) + 86400*leaps;
	}

	int cycles, centuries, leaps, rem;

	int leap = 0;
	if (!is_leap)
		is_leap = &leap;
	cycles = (year-100) / 400;
	rem = (year-100) % 400;
	if (rem < 0) {
		cycles--;
		rem += 400;
	}
	if (!rem) {
		*is_leap = 1;
		centuries = 0;
		leaps = 0;
	} else {
		if (rem >= 200) {
			if (rem >= 300) centuries = 3, rem -= 300;
			else centuries = 2, rem -= 200;
		} else {
			if (rem >= 100) centuries = 1, rem -= 100;
			else centuries = 0;
		}
		if (!rem) {
			*is_leap = 0;
			leaps = 0;
		} else {
			leaps = rem / 4U;
			rem %= 4U;
			*is_leap = !rem;
		}
	}

	leaps += 97*cycles + 24*centuries - *is_leap;

	return (year-100) * 31536000LL + leaps * 86400LL + 946684800 + 86400;
}

int64_t TmToSecs(const struct tm *tm)
{
	int is_leap;
	int64_t year = tm->tm_year;
	int month = tm->tm_mon;
	if (month >= 12 || month < 0) {
		int adj = month / 12;
		month %= 12;
		if (month < 0) {
			adj--;
			month += 12;
		}
		year += adj;
	}
	int64_t t = YearToSecs(year, &is_leap);
	t += MonthToSecs(month, is_leap);
	t += 86400LL * (tm->tm_mday-1);
	t += 3600LL * tm->tm_hour;
	t += 60LL * tm->tm_min;
	t += tm->tm_sec;
	return t;
}

/*! Converts a timestamp is ISO8601 format to a struct tm.
    \param buf A buffer containing a timestamp in ISO8601 format.
    \param timeval The time converted is returned here.
    \return true if the conversion was successful.
    \return false if the conversion failed.*/
bool FromISO8601(char* buf, struct tm* timeval)
{
  return strptime(buf, "%Y-%m-%dT%H:%M:%S", timeval) != nullptr;
}

/*! Converts a struct tm to a timestamp in ISO8601 format.
    \param timeval The time to convert.
    \param buf The ISO8601 time is returned here.
    \param buf The size of the buffer that buf points to.
    \param timezone A string that is put at the end of the timestamp.*/
void ToISO8601(struct tm* timeval, char* buf, uint32_t bufsize, const char* timezone /*= "Z"*/)
{
  strftime(buf, bufsize, "%Y-%m-%dT%H:%M:%S", timeval);
  buf[bufsize - 1] = 0;

  if (strlen(timezone) > 0)
  {
    strlcat(buf, timezone, bufsize);
    buf[bufsize - 1] = 0;
  }
}

/*! Converts a unix timestamp to a timestamp in ISO8601 format.
    \param t The time to convert.
    \param buf The ISO8601 time is returned here.
    \param buf The size of the buffer that buf points to.
    \param timezone A string that is put at the end of the timestamp.*/
void ToISO8601(int64_t t, char* buf, uint32_t bufsize, const char* timezone /*= "Z"*/)
{
  tm timeval;
  SecsToTm(t, &timeval);

  ToISO8601(&timeval, buf, bufsize, timezone);
}
