#include "usart.h"
#include "lock.h"
#include "clock.h"

#include <cassert>
#include <cstring>

static CUsart* g_usartirqhandlers[8];

/*! IRQ handlers for each usart, calls into the corresponding CUsart using g_usartirqhandlers.
    They all have the weak attribute set, so that they can be overridden in the application code.*/
extern "C"
{
  void USART1_IRQHandler()
  {
    if (g_usartirqhandlers[0])
      g_usartirqhandlers[0]->IRQHandler();
  }

  void USART2_IRQHandler()
  {
    if (g_usartirqhandlers[1])
      g_usartirqhandlers[1]->IRQHandler();
  }

  void USART3_IRQHandler()
  {
    if (g_usartirqhandlers[2])
      g_usartirqhandlers[2]->IRQHandler();
  }

  void UART4_IRQHandler()
  {
    if (g_usartirqhandlers[3])
      g_usartirqhandlers[3]->IRQHandler();
  }

  void UART5_IRQHandler()
  {
    if (g_usartirqhandlers[4])
      g_usartirqhandlers[4]->IRQHandler();
  }

  void USART6_IRQHandler()
  {
    if (g_usartirqhandlers[5])
      g_usartirqhandlers[5]->IRQHandler();
  }

  void UART7_IRQHandler()
  {
    if (g_usartirqhandlers[6])
      g_usartirqhandlers[6]->IRQHandler();
  }

  void UART8_IRQHandler()
  {
    if (g_usartirqhandlers[7])
      g_usartirqhandlers[7]->IRQHandler();
  }
}

/*! Converts usart into a number from 0 to 7.
    \return The index number of the usart.*/
static uint32_t UsartNum(USART_TypeDef* usart)
{
  if (usart == USART1)
    return 0;
  else if (usart == USART2)
    return 1;
  else if (usart == USART3)
    return 2;
  else if (usart == UART4)
    return 3;
  else if (usart == UART5)
    return 4;
  else if (usart == USART6)
    return 5;
  else if (usart == UART7)
    return 6;
  else if (usart == UART8)
    return 7;
  else
    assert(0);

  return 0;
}

/*! Installs a pointer in g_usartirqhandlers at the index corresponding with usart.
    \param usart A pointer to the corresponding usart.
    \param irqhandler A pointer to a class derived from CUsart.*/
void InstallUsartIRQHandler(USART_TypeDef* usart, CUsart* irqhandler)
{
  uint32_t usartnum = UsartNum(usart);
  assert(usartnum < sizeof(g_usartirqhandlers) / sizeof(g_usartirqhandlers[0]));
  assert(g_usartirqhandlers[usartnum] == nullptr);
  g_usartirqhandlers[usartnum] = irqhandler;
}

/*! Removes the pointer installed with InstallUsartIRQHandler()*/
void RemoveUsartIRQHandler(USART_TypeDef* usart)
{
  uint32_t usartnum = UsartNum(usart);
  assert(usartnum < sizeof(g_usartirqhandlers) / sizeof(g_usartirqhandlers[0]));
  assert(g_usartirqhandlers[usartnum] != nullptr);
  g_usartirqhandlers[usartnum] = nullptr;
}

CUsart::CUsart()
{
  m_usart            = nullptr;
  m_queuesize        = 0;
  m_rxqueue          = 0;
  m_txqueue          = 0;
  m_txcompletesignal = 0;
  m_rtsflowcontrol   = false;
}

CUsart::~CUsart()
{
  Close();
}

void CUsart::Setup(USART_TypeDef* usart, USART_InitTypeDef* init, uint32_t queuesize /*= USART_QUEUE_SIZE*/)
{
  Close();

  //enforce a minimum queue size, so that there is always one free space
  //when the rx interrupt is enabled for flow control
  assert(queuesize >= 2);

  m_usart = usart;
  m_queuesize = queuesize;

  //create a queue for receiving if the usart has rx mode enabled
  //it's created as 2 bytes per index, so that error flags can be passed for each byte
  //there is always one byte more allocated than needed, so that the READRESULT_ERROR_RX_QUEUE_FULL
  //flag can be passed, while still allowing all data to be buffered
  if (init->USART_Mode & USART_Mode_Rx)
  {
    m_rxqueue = xQueueCreate(queuesize + 1, sizeof(uint16_t));

    //store if the usart will have rts flow control enabled, because when it is,
    //the rx not empty interrupt should be disabled when the rx queue is full
    m_rtsflowcontrol = (init->USART_HardwareFlowControl & USART_HardwareFlowControl_RTS) ==
                        USART_HardwareFlowControl_RTS;
  }

  //create a queue for transmitting if the usart has tx mode enabled
  if (init->USART_Mode & USART_Mode_Tx)
  {
    m_txqueue          = xQueueCreate(queuesize, sizeof(uint8_t));
    m_txcompletesignal = xSemaphoreCreateBinary();
  }

  //use a critical section to protect the NVIC and usart registers
  CCritical critsection;

  //enable the clock for the usart
  SetClock(ENABLE);

  //initialize the usart with the specified parameters
  USART_Init(m_usart, init);

  //enable interrupts for the usart
  SetIrq(ENABLE);

  //enable the usart
  USART_Cmd(m_usart, ENABLE);
}

void CUsart::SetOneBitMode(FunctionalState newstate)
{
  assert(m_usart);
  USART_OneBitMethodCmd(m_usart, newstate);
}

void CUsart::Close()
{
  //use a critical section to protect the NVIC and usart registers
  CCritical critsection;

  if (m_usart == nullptr)
    return;

  //disable the usart
  USART_Cmd(m_usart, DISABLE);

  //deinit the usart
  SetIrq(DISABLE);
  USART_DeInit(m_usart);
  SetClock(DISABLE);

  critsection.Leave();

  //delete the queues
  if (m_rxqueue)
  {
    vQueueDelete(m_rxqueue);
    m_rxqueue = 0;
  }

  m_rtsflowcontrol = false;

  if (m_txqueue)
  {
    vQueueDelete(m_txqueue);
    m_txqueue = 0;
  }

  if (m_txcompletesignal)
  {
    vSemaphoreDelete(m_txcompletesignal);
    m_txcompletesignal = 0;
  }

  m_usart = nullptr;
}

/*! Converts m_usart into the corresponding IRQn for the NVIC.
    \return The interrupt number for the usart.*/
IRQn CUsart::UsartIrqn()
{
  if (m_usart == USART1)
    return USART1_IRQn;
  else if (m_usart == USART2)
    return USART2_IRQn;
  else if (m_usart == USART3)
    return USART3_IRQn;
  else if (m_usart == UART4)
    return UART4_IRQn;
  else if (m_usart == UART5)
    return UART5_IRQn;
  else if (m_usart == USART6)
    return USART6_IRQn;
  else if (m_usart == UART7)
    return UART7_IRQn;
  else if (m_usart == UART8)
    return UART8_IRQn;
  else
    assert(0);

  return (IRQn)0;
}

/*! Enables or disables the irq for the usart.
    \param newstate State of the irq, set to ENABLE or DISABLE.*/
void CUsart::SetIrq(FunctionalState newstate)
{
  //use a critical section to protect the nvic registers
  CCritical critsection;

  //when enabling the usart interrupt, first set the interrupt handler pointer to this class
  if (newstate == ENABLE)
    InstallUsartIRQHandler(m_usart, this);

  //set the interrupt for the usart
  NVIC_InitTypeDef nvicsettings =
  {
    .NVIC_IRQChannel                   = (uint8_t)UsartIrqn(),
    .NVIC_IRQChannelPreemptionPriority = configLIBRARY_MAX_SYSCALL_INTERRUPT_PRIORITY,
    .NVIC_IRQChannelSubPriority        = 0,
    .NVIC_IRQChannelCmd                = newstate,
  };
  NVIC_Init(&nvicsettings);

  //set the usart RX interrupt
  USART_ITConfig(m_usart, USART_IT_RXNE, newstate);

  //when disabling the usart interrupt, remove the interrupt handler pointer last
  if (newstate == DISABLE)
    RemoveUsartIRQHandler(m_usart);
}

/*! Enables or disables the clock for m_usart.
    \param newstate The state of the clock, ENABLE or DISABLE.*/
void CUsart::SetClock(FunctionalState newstate)
{
  if (m_usart == USART1)
    RCC_APB2PeriphClockCmd(RCC_APB2Periph_USART1, newstate);
  else if (m_usart == USART2)
    RCC_APB1PeriphClockCmd(RCC_APB1Periph_USART2, newstate);
  else if (m_usart == USART3)
    RCC_APB1PeriphClockCmd(RCC_APB1Periph_USART3, newstate);
  else if (m_usart == UART4)
    RCC_APB1PeriphClockCmd(RCC_APB1Periph_UART4, newstate);
  else if (m_usart == UART5)
    RCC_APB1PeriphClockCmd(RCC_APB1Periph_UART5, newstate);
  else if (m_usart == USART6)
    RCC_APB2PeriphClockCmd(RCC_APB2Periph_USART6, newstate);
  else if (m_usart == UART7)
    RCC_APB1PeriphClockCmd(RCC_APB1Periph_UART7, newstate);
  else if (m_usart == UART8)
    RCC_APB1PeriphClockCmd(RCC_APB1Periph_UART8, newstate);
  else
    assert(0);
}

/*! Receives data from the rx queue, and will turn on the rx interrupt if rts flow control
    is enabled and there is space in the rx queue.
    \param pvBuffer A pointer to a uint16_t to write the data to.
    \param xTicksToWait The number of ticks to wait for the rx queue to have data available.
    \return pdTRUE if an item was successfully received from the queue, otherwise pdFALSE.*/
BaseType_t CUsart::RxQueueReceive(void *pvBuffer, TickType_t xTicksToWait)
{
  BaseType_t receive = xQueueReceive(m_rxqueue, pvBuffer, xTicksToWait);
  if (m_rtsflowcontrol && receive)
  { //check the space available in the rx queue and set the rx interrupt from inside a critical section,
    //because the usart rx interrupt disables the interrupt when the queue is full,
    //the critical section prevents that from happening in between the queue check and the interrupt enable
    CCritical critsection;
    if (uxQueueSpacesAvailable(m_rxqueue) >= 3)
      USART_ITConfig(m_usart, USART_IT_RXNE, ENABLE);
  }

  return receive;
}

/*! Clears all received data in the rx queue.*/
void CUsart::ClearInput()
{
  if (m_rxqueue)
    xQueueReset(m_rxqueue);
}

/*! Reads bytes from the usart.
    \param buf The buffer to write received bytes into.
    \param bufsize The size of the buffer that buf points to.
    \param bytesread The number of bytes written into buf.
    \param timeout The timeout in milliseconds of the entire read.
    \param end When not -1, end the read if this character is read.
    \return: 
            READRESULT_SUCCESS when the read has succeeded,
            READRESULT_ERROR_TIMEOUT when the read has timed out,
            or a bitwire OR of:
            READRESULT_ERROR_OVERRUN when the uart interrupt did not read quickly enough.
            READRESULT_ERROR_NOISE when the uart detected a wrong state on the line during one of the bits.
            READRESULT_ERROR_FRAMING when the usart detected desynchronization, or a break character.
            READRESULT_ERROR_PARITY when parity is enabled and the uart detected a parity error.
            READRESULT_ERROR_RX_QUEUE_FULL when all space in the rx queue has been used.*/
uint32_t CUsart::Read(uint8_t* buf, uint32_t bufsize, uint32_t* bytesread, uint32_t timeout, int end /*= -1*/)
{
  assert(bufsize > 0);

  //doing a read is invalid when rx mode is disabled
  assert(m_rxqueue);

  *bytesread = 0;

  //keep copying bytes from the queue into the buffer, for as long as there are bytes to copy
  //this is done first, so that with a timeout of 0, all queued bytes are still read
  uint16_t input;
  while (*bytesread < bufsize && RxQueueReceive(&input, 0))
  {
    //store the received byte in the buffer
    buf[*bytesread] = input & READDATA_MASK;
    (*bytesread)++;

    //if there was an error receiving this byte, return it
    if ((input & READRESULT_MASK) != 0)
      return input & READRESULT_MASK;
    else if (end != -1 && (input & READDATA_MASK) == end)
      return READRESULT_SUCCESS; //return success when the end character is received
  }

  if (*bytesread == bufsize)
    return READRESULT_SUCCESS; //return success when the buffer is full

  //receive bytes until the buffer is full, the end character is received, or the timeout is reached
  int64_t start = Millis();
  for(;;)
  {
    uint32_t elapsed = Millis() - start;
    if (elapsed >= timeout)
    {
      return READRESULT_ERROR_TIMEOUT; //return timeout when the timeout has elapsed
    }
    else if (RxQueueReceive(&input, pdMS_TO_TICKS(timeout - elapsed)))
    {
      //store the received byte in the buffer
      buf[*bytesread] = input & READDATA_MASK;
      (*bytesread)++;

      //if there was an error receiving this byte, return it
      if ((input & READRESULT_MASK) != 0)
        return input & READRESULT_MASK;
      else if (end != -1 && (input & READDATA_MASK) == end)
        return READRESULT_SUCCESS; //return success when the end character is received
      else if (*bytesread == bufsize)
        return READRESULT_SUCCESS; //return success when the buffer is full
    }
    else
    {
      return READRESULT_ERROR_TIMEOUT; //return timeout when RxQueueReceive with timeout failed
    }
  }

  //prevents a compile warning, should never be reached
  assert(0);
  return READRESULT_SUCCESS;
}
/*! Waits for \r\n to be received, then writes received bytes into buf until an end response is received.
    The buffer holds all bytes of the received reply, and is always terminated as a C-string
    \param buf The buffer to write received bytes into.
    \param bufsize The size of the buffer that buf points to.
    \param bytesread The number of bytes written into buf, excluding the terminating 0 character.
    \param timeout The timeout in milliseconds of the entire read.
    \param start when not null, the string that signals the start of the reply, when null, \r\n is used.
    \param end When not null, the string that signals the end of the reply.
    \return: 
            READRESULT_SUCCESS when the read has succeeded,
            READRESULT_ERROR_TIMEOUT when the read has timed out,
            READRESULT_ERROR_TOO_LARGE when the modem reply does not fit into buf,
            or a bitwire OR of:
            READRESULT_ERROR_OVERRUN when the uart interrupt did not read quickly enough.
            READRESULT_ERROR_NOISE when the uart detected a wrong state on the line during one of the bits.
            READRESULT_ERROR_FRAMING when the usart detected desynchronization, or a break character.
            READRESULT_ERROR_PARITY when parity is enabled and the uart detected a parity error.
            READRESULT_ERROR_RX_QUEUE_FULL when all space in the rx queue has been used.*/
uint32_t CUsart::ReadModemReply(char* buf, uint32_t bufsize, uint32_t* bytesread, uint32_t timeout,
                                const char* start /*= nullptr*/, const char* end /*= nullptr*/)
{
  //doing a read is invalid when rx mode is disabled
  assert(m_rxqueue);

  if (bufsize == 0)
    return READRESULT_ERROR_TOO_LARGE;

  //initialize the buffer as C-string
  *bytesread = 0;
  buf[0] = 0;

  int64_t readstart = Millis();

  if (start == nullptr)
    start = "\r\n";

  uint32_t startlen = strlen(start);

  if (startlen > 0)
  { //wait for the starting string
    uint32_t retval = WaitForModemReplyStart(start, timeout);
    if (retval != 0)
      return retval; //return an error if WaitForModemReplyStart() failed

    //make sure the buffer is large enough for the starting \r\n as C-string
    if (bufsize < startlen + 1) 
      return READRESULT_ERROR_TOO_LARGE;

    //write the start string into the buffer
    strcpy(buf, start);
    *bytesread = startlen;
  }

  for(;;)
  {
    if (*bytesread + 1 >= bufsize)       //return an error if the buffer is full
      return READRESULT_ERROR_TOO_LARGE; //and it does not end with a modem result

    uint32_t elapsed = Millis() - readstart;
    if (elapsed >= timeout) //return an error if the data was not received on time
      return READRESULT_ERROR_TIMEOUT;

    uint16_t input;
    if (!RxQueueReceive(&input, pdMS_TO_TICKS(timeout - elapsed)))
      return READRESULT_ERROR_TIMEOUT; //return an error if no data was received from the queue

    //if the uart had an error receiving, return an error
    if (input & READRESULT_MASK)
      return input & READRESULT_MASK;

    //store the received byte
    buf[*bytesread] = input & READDATA_MASK;
    (*bytesread)++;
    buf[*bytesread] = 0; //always terminate as C-string

    if (EndsWithModemResult(buf, *bytesread, end))
      break; //if it ends with a modem result, the complete reply has been read
  }

  return READRESULT_SUCCESS;
}

/*! Waits for the start string to be received on the usart.
    \param timeout The maximum time in milliseconds waiting for the start string.
    \return 0 if the start string is received, 
            READRESULT_ERROR_TIMEOUT if the start string was not received on time,
            or a bitwise OR of:
            READRESULT_ERROR_OVERRUN when the uart interrupt did not read quickly enough.
            READRESULT_ERROR_NOISE when the uart detected a wrong state on the line during one of the bits.
            READRESULT_ERROR_FRAMING when the usart detected desynchronization, or a break character.
            READRESULT_ERROR_PARITY when parity is enabled and the uart detected a parity error.
            READRESULT_ERROR_RX_QUEUE_FULL when all space in the rx queue has been used.*/
uint32_t CUsart::WaitForModemReplyStart(const char* start, uint32_t timeout)
{
  int64_t  readstart = Millis();
  uint32_t startlen  = strlen(start);
  uint32_t readpos   = 0;
  for (;;)
  {
    uint32_t elapsed = Millis() - readstart;
    if (elapsed >= timeout) //return an error if the data was not received on time
      return READRESULT_ERROR_TIMEOUT;

    uint16_t input;
    if (!RxQueueReceive(&input, pdMS_TO_TICKS(timeout - elapsed)))
      return READRESULT_ERROR_TIMEOUT; //return an error if no data was received from the queue

    //if the uart had an error receiving, return an error
    if (input & READRESULT_MASK)
      return input & READRESULT_MASK;

    char readbyte = input & READDATA_MASK;
    if (readbyte == start[readpos])
    {
      readpos++;
      if (readpos == startlen)
        break; //the start string is received
    }
    else
    { //if a non-matching character of the start string is received,
      //check instead if it's the first character of the start string,
      //in which case the counter should be set to 1
      if (readbyte == start[0])
        readpos = 1;
      else
        readpos = 0;
    }
  }

  return 0;
}

/*! Checks if buf ends in a modem result.
    \param buf A pointer to a buffer where data received from the modem is stored.
    \param bytesread Number of bytes stored in buf.
    \param end When not null, the string to check for as end.
    \return true if buf ends in a modem result, false if it doesn't.*/
bool CUsart::EndsWithModemResult(const char* buf, uint32_t bytesread, const char* end)
{
  static const char *modemresults[] =
  {
    "\r\nOK\r\n",
    "\r\nCONNECT\r\n",
    "\r\nRING\r\n",
    "\r\nNO CARRIER\r\n",
    "\r\nERROR\r\n",
    "\r\nNO DIALTONE\r\n",
    "\r\nBUSY\r\n",
    "\r\nNO ANSWER\r\n",
  };

  const char** results;
  uint32_t     numresults;

  //if end is not a nullptr, check if buf ends with end
  if (end)
  {
    results = &end;
    numresults = 1;
  }
  else
  { //if end is a nullptr, check if buf ends with one of the modem results
    results = modemresults;
    numresults = sizeof(modemresults) / sizeof(modemresults[0]);
  }

  //check if buf ends with one of the results
  for (uint32_t i = 0; i < numresults; i++)
  {
    uint32_t resultlen = strlen(results[i]);

    //check if there are enough bytes in buf for the result string
    if (bytesread >= resultlen)
    {
      //check if buf ends in this modem result string
      if (strncmp(buf + (bytesread - resultlen), results[i], resultlen) == 0)
        return true;
    }
  }

  if (end == nullptr)
  {
    //check if buf ends with a CME error, which looks like:
    // \r\n+CME ERROR: Error string\r\n
    const char* cmeresult = "\r\n+CME ERROR:";
    uint32_t cmeresultlen = strlen(cmeresult);
    //check if buf is long enough to hold a CME error, and check if it ends with \r\n
    if (bytesread >= cmeresultlen + 2 && buf[bytesread - 2] == '\r' && buf[bytesread - 1] == '\n')
    {
      //search backwards for the preceding \r\n
      uint32_t state = 0;
      int32_t  pos   = -1;
      for (int32_t i = (int32_t)bytesread - 3; i >= 0; i--)
      {
        if (state == 0)
        { //find a newline
          if (buf[i] == '\n')
            state = 1;
        }
        else if (state == 1)
        { //check if a carriage return is found after the newline
          if (buf[i] == '\r')
          {
            pos = i;
            break;
          }
          else if (buf[i] != '\n')
          {
            state = 0;
          }
        }
      }

      //check if at the \r\n, the string matches \r\n+CME ERROR:
      if (pos != -1 && bytesread - pos >= cmeresultlen + 2)
      {
        if (strncmp(buf + pos, cmeresult, cmeresultlen) == 0)
          return true;
      }
    }
  }

  return false;
}

/*! Writes bytes to the usart.
    \param buf The buffer write to the usart.
    \param bufsize The size of the buffer to write to the usart.
    \param timeout Timeout in milliseconds of the entire write, 0 for infinite timeout.
    \return WRITERESULT_SUCCESS if the write succeeded, WRITERESULT_ERROR_TIMEOUT if the write failed.*/
uint32_t CUsart::Write(const uint8_t* buf, uint32_t bufsize, uint32_t timeout /*= 0*/)
{
  //doing a write is invalid when tx mode is disabled
  assert(m_txqueue);

  //write as many bytes from the buffer into the queue before checking the timeout
  uint32_t byteswritten = 0;
  while (byteswritten < bufsize && xQueueSendToBack(m_txqueue, buf + byteswritten, 0))
  {
    byteswritten++;

    //enable the transmitter register empty interrupt so that the usart will transmit the data
    USART_ITConfig(m_usart, USART_IT_TXE, ENABLE);
  }

  if (byteswritten == bufsize)
    return WRITERESULT_SUCCESS; //return success if all bytes are written to the queue

  //write bytes until 
  int64_t start = Millis();
  for(;;)
  {
    uint32_t elapsed = Millis() - start;
    if (timeout > 0 && elapsed >= timeout)
    {
      return WRITERESULT_ERROR_TIMEOUT; //return timeout when the timeout has elapsed
    }
    else if (xQueueSendToBack(m_txqueue, buf + byteswritten, timeout > 0 ? pdMS_TO_TICKS(timeout - elapsed) : portMAX_DELAY))
    {
      //enable the transmitter register empty interrupt so that the usart will transmit the data
      USART_ITConfig(m_usart, USART_IT_TXE, ENABLE);

      byteswritten++;
      if (byteswritten == bufsize)
        return WRITERESULT_SUCCESS; //return success if all bytes are written to the queue
    }
    else
    {
      return WRITERESULT_ERROR_TIMEOUT; //return timeout if the xQueueSendToBack() call with timeout failed
    }
  }

  //prevents a compile warning, should never be reached
  assert(0);
  return WRITERESULT_SUCCESS;
}

/*! Writes str to the usart, without the terminating null character.
    \param str a C-string to write to the usart.
    \param timeout Timeout in milliseconds to wait for the write to finish.
    \return The return value of CUsart::Write().*/
uint32_t CUsart::WriteStr(const char* str, uint32_t timeout /*= 0*/)
{
  return Write((uint8_t*)str, strlen(str), timeout);
}

/*! Waits until m_txqueue becomes empty.
    \param timeout Timeout in milliseconds waiting on the queue to become empty.
    \return true if the queue became empty in the timeout, false if it did not.*/
bool CUsart::WaitWriteComplete(uint32_t timeout /*= 0*/)
{
  assert(m_txqueue);
  assert(m_txcompletesignal);

  if (timeout == 0)
  {
    //keep waiting on m_txcompletesignal while m_txqueue still holds data
    while (uxQueueMessagesWaiting(m_txqueue) > 0)
      xSemaphoreTake(m_txcompletesignal, portMAX_DELAY);
  }
  else
  {
    //keep looping while m_txqueue has data, and the timeout has not elapsed
    int64_t start = Millis();
    while (uxQueueMessagesWaiting(m_txqueue) > 0)
    {
      int64_t elapsed = Millis() - start;
      if (elapsed >= timeout)
        break;

      xSemaphoreTake(m_txcompletesignal, pdMS_TO_TICKS(timeout - elapsed));
    }
  }

  return uxQueueMessagesWaiting(m_txqueue) == 0;
}

/*! Converts a usart result to a human readable string.
    \param result The usart result code.
    \return A string describing the result code.*/
const char* CUsart::ResultToStr(uint32_t result)
{
  //multiple bits can be set by the uart, in that case just print the first error
  if ((result & READRESULT_SUCCESS) == READRESULT_SUCCESS)
    return "read successful";
  else if ((result & READRESULT_ERROR_TIMEOUT) == READRESULT_ERROR_TIMEOUT)
    return "read timed out";
  else if ((result & READRESULT_ERROR_TOO_LARGE) == READRESULT_ERROR_TOO_LARGE)
    return "reply too large";
  else if ((result & READRESULT_ERROR_OVERRUN) == READRESULT_ERROR_OVERRUN)
    return "data overrun";
  else if ((result & READRESULT_ERROR_NOISE) == READRESULT_ERROR_NOISE)
    return "line noise";
  else if ((result & READRESULT_ERROR_FRAMING) == READRESULT_ERROR_FRAMING)
    return "framing error";
  else if ((result & READRESULT_ERROR_PARITY) == READRESULT_ERROR_PARITY)
    return "parity error";
  else if ((result & READRESULT_ERROR_RX_QUEUE_FULL) == READRESULT_ERROR_RX_QUEUE_FULL)
    return "rx queue full";
  else if ((result & WRITERESULT_SUCCESS) == WRITERESULT_SUCCESS)
    return "write successful";
  else if ((result & WRITERESULT_ERROR_TIMEOUT) == WRITERESULT_ERROR_TIMEOUT)
    return "write timed out";
  else
    return "ERROR UNKNOWN USART RESULT";
}

void CUsart::IRQHandler()
{
  //read the status register, and then the data register
  //when the receive register not empty bit is set
  //this clears the bits for overrun error and framing error
  uint16_t statusreg = m_usart->SR;

  if (statusreg & USART_FLAG_RXNE)
  {
    //read a byte from the data register if it's available
    uint16_t data = m_usart->DR & 0xFF;

    //pass error flags to the reading task
    if (statusreg & USART_FLAG_ORE)
      data |= READRESULT_ERROR_OVERRUN;
    if (statusreg & USART_FLAG_NE)
      data |= READRESULT_ERROR_NOISE;
    if (statusreg & USART_FLAG_FE)
      data |= READRESULT_ERROR_FRAMING;
    if (statusreg & USART_FLAG_PE)
      data |= READRESULT_ERROR_PARITY;

    uint32_t messageswaiting = uxQueueMessagesWaitingFromISR(m_rxqueue);

    //if rts flow control is enabled, and the rx queue is almost full,
    //disable the rx interrupt, so that the rts line will stay high and block
    //the other side from transmitting
    if (m_rtsflowcontrol && messageswaiting + 2 >= m_queuesize)
      USART_ITConfig(m_usart, USART_IT_RXNE, DISABLE);

    //the rx queue always has one more space allocated than needed
    //if this data is going into that last space, signal an error
    if (uxQueueMessagesWaitingFromISR(m_rxqueue) == m_queuesize)
      data |= READRESULT_ERROR_RX_QUEUE_FULL;

    //if the modem task is blocked on reading the RX queue,
    //make sure it wakes up immediately
    if (m_rxqueue)
    {
      BaseType_t taskwoken = pdFALSE;
      xQueueSendToBackFromISR(m_rxqueue, &data, &taskwoken);
      portYIELD_FROM_ISR(taskwoken);
    }
  }

  //check the transmission complete flag, it's important that this is checked
  //before the transmission register empty flag, because this code block might disable
  //the transmission complete interrupt, while the next block enables it
  if (statusreg & USART_FLAG_TC)
  {
    //disable the transmission complete interrupt
    USART_ITConfig(m_usart, USART_IT_TC, DISABLE);

    //if the usart signaled that the transmission is complete,
    //and the queue is empty, signal the writing task
    if (m_txcompletesignal)
    {
      //if a high priority task is waiting on this signal,
      //wake it up immediately
      BaseType_t taskwoken = pdFALSE;
      xSemaphoreGiveFromISR(m_txcompletesignal, &taskwoken);
      portYIELD_FROM_ISR(taskwoken);
    }
  }

  if (statusreg & USART_FLAG_TXE)
  {
    if (m_txqueue)
    {
      uint8_t out;
      if (xQueueReceiveFromISR(m_txqueue, &out, NULL) == pdPASS)
      {
        USART_ITConfig(m_usart, USART_IT_TC, ENABLE); //enable the transmission complete interrupt
        USART_SendData(m_usart, out); //transmit the byte out the serial port
      }

      if (xQueueIsQueueEmptyFromISR(m_txqueue))
        USART_ITConfig(m_usart, USART_IT_TXE, DISABLE); //all done, disable the transmit register empty interrupt
    }
    else
    { //disable the transmit interrupt if no tx queue is allocated, should not happen
      USART_ITConfig(m_usart, USART_IT_TXE, DISABLE);
    }
  }
}
